/*
 6502struct.h
  6502 CPU state structure

  Revised: 2013.07.02

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __X6502STRUCT_H
 #define NTSC_CPU		1789772.7272727272727272 /* 312500*7*9/11 */
 #define PAL_CPU		1662607.125		 /* 1477873*9/8 */

 #define N_FLAG			0x80	/* Negative */
 #define V_FLAG			0x40	/* oVerflow */
 #define B_FLAG			0x10	/* Break (X bit on 65816) */
 #define D_FLAG			0x08	/* Decimal */
 #define I_FLAG			0x04	/* Interrupt disable */
 #define Z_FLAG			0x02	/* Zero */
 #define C_FLAG			0x01	/* Carry */
 /* 65816 flags (for completeness) */
 #define M_FLAG			0x20	/* accuMulator (6502 PLPs this on) */
 #define X_FLAG			0x10	/* indeX 16-bit select */

typedef struct X6502u32 /* for immediate 32-bit extension */
{
	uint32		PC;
	uint32		A;
	uint32		X;
	uint8		Y, S, BP, P;
	uint32		IA;	/* last internal 16 bit interrupt read address */
} X6502u32;

typedef struct X6502struct/*(32)*/
{
#if BIG_ENDIAN
	uint16		PCzero, PC;
	uint8		Azero[3], A;
	uint8		Xzero[3], X;
	uint8		Y, S, BP, P;
	uint8		IAz0, IAz1, IAx2, IAx3;
#else
	uint16		PC, PCzero;
	uint8		A, Azero[3];
	uint8		X, Xzero[3];
	uint8		Y, S, BP, P;
	uint8		IAx3, IAx2, IAz1, IAz0;
#endif
	int32		tcount;  /* temporary cycle counter. */
	int32		count;
	uint16		intlow;  /* interrupt bits, emulator-specific. */
} X6502struct, X6502;

typedef union X6502U
{
	X6502		low;
	X6502u32	u32;
} X6502U;
#endif
