/*
 FCEU.h

 * MacFCEU 0.8.5
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _FCEUH

/* ---- constants */

#define FCEU_VERSION_NUMERIC	9821 /*0.8.5+*/

#define DECLFR(x)		uint8 FP_FASTAPASS(1) x (uint32 A)
#define DECLFW(x)		void FP_FASTAPASS(2) x (uint32 A, uint8 V)

 /* twelfth-root-two */
//#define M_12R2 (double)1.059463094359295264562

#define GI_RESETM2		1
#define GI_POWER		2
#define GI_CLOSE		3

#define JOY_A			1
#define JOY_B			2
#define JOY_SELECT		4
#define JOY_START		8
#define JOY_UP			0x10
#define JOY_DOWN		0x20
#define JOY_LEFT		0x40
#define JOY_RIGHT		0x80

/* ---- typedefs */

typedef struct FCEUS
{
	uint8	PAL; /* removed unused NetworkPlay; redundant: FCEUnetplay */
	uint8	GameGenie; /* some moved, reduced sizes; 2011.02.08 */
#if HAVE_HQ
	uint8	soundq, lowpass;
#else
	uint8 reserved[2];
#endif
	int	SnapName;
	int	SoundVolume; /* percentage; [-65536,65536) */
	int	SndVolF; /* precalculated multiplier ((volume<<15)/25) */

	/* Current first and last rendered scanlines. */
	int	FirstSLine, LastSLine;

	/* Driver code(user)-specified first and last rendered scanlines. 
	 Usr*SLine[0] is for NTSC, Usr*SLine[1] is for PAL.
	*/
	int	UsrFirstSLine[2];
	int	UsrLastSLine[2];
	int	UsrVBLSLines[2]; /* added; 2010.01.04 */
	/* rates here are 28:4 fixed-point samples per second */
	uint32	OrgRate; /* 111860.75 max., 0=off */
	uint32	SndRate;
	int32	SndRateF; /* (94<<23)/OrgRate */
	uint16	soundactivech; /* SN5F76xdntba */
	int16	soundrectamp, soundtndamp; /* 256=1.0 */
	uint16	soundnscale; /* 2=1.0 */
	uint16	soundtscale; /* 3=1.0 */
	uint16	sounddcfl; /* fixed point 7:9; FCEU=93 */
} FCEUS;

/* ---- globals */

extern struct FCEUGI *FCEUGameInfo;

extern FCEUS FSettings;

extern int fceuindbg;

extern uint8 RateFixFlags; /* 1.6 NTSC/PAL, adjust DMC rates on speed */

extern uint64 timestampbase;

extern uint8 RAM[0x800];
extern uint8 GameMemBlock[131072];

extern readfunc ARead[0x10000];
extern writefunc BWrite[0x10000];

extern void (*GameInterface)(int h);
extern void (*GameStateRestore)(int version);

extern uint8 pale;
extern uint8 vsdip;

extern int GameAttributes;
extern uint8 PAL;

/* ---- prototypes */

/* made public for checking purposes */
DECLFW(BNull);
DECLFR(ANull);

void ResetGameLoaded(void);

void FCEU_MemoryRand(uint8 *ptr,
	uint32 size);
writefunc FASTAPASS(1) GetWriteHandler(int32 a);
readfunc FASTAPASS(1) GetReadHandler(int32 a);

void FASTAPASS(3) SetReadHandler(int32 start,
	int32 end,
	readfunc f);
void FASTAPASS(3) SetWriteHandler(int32 start,
	int32 end,
	writefunc f);
/*you have to pass the correct final address for the custom-incremental ones;
 they set from end
*/
void FASTAPASS(4) SetReadHandlerI(int d,
	int32 start,
	int32 end,
	readfunc f);
void FASTAPASS(4) SetWriteHandlerI(int d,
	int32 start,
	int32 end,
	writefunc f);

void FCEU_ResetVidSys(void);

void ResetMapping(void);
void ResetNES(void);
void PowerNES(void);

void FCEU_PrintError(const char *format, ...);
void FCEU_printf(const char *format, ...);
void FCEU_DispMessage(const char *format, ...);

#if 0
void SetNESDeemph(uint8 d,
	int force);
#endif

void DrawTextTrans(uint8 *dstpix,
	uint32 width,
	const uint8 *textmsg,
	uint8 fgcolor);

#ifdef FRAMESKIP
//void FCEU_PutImageDummy(void);
#endif

 #define _FCEUH
#endif
