/*
 GameGenie.c
  Game Genie[TM] source originally in cart.c/fceu.c

 * MacFCEU 0.8a7
 *
 * Copyright notice for this file:
 *  cart.c (C) 2002 Xodnizel
 *  fceu.c (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "GameGenie.h"
#include "cart.h"
#include "memory.h"
#include "general.h"

uint8		GenieStage = 0;

readfunc	*AReadG;
writefunc	*BWriteG;

uint8		*VPageG[8];

static uint8	*GENIEROM = 0;

void FCEU_KillGenie()
{
	if (GENIEROM)
		{FCEU_free(GENIEROM); GENIEROM = 0;}
}

static int AllocGenieRW(void)
{
	if ((AReadG=(readfunc*)FCEU_malloc(0x8000*sizeof(readfunc)))==0)
		return 0;
	if ((BWriteG=(writefunc*)FCEU_malloc(0x8000*sizeof(writefunc)))==0)
		{FCEU_free(AReadG); AReadG = 0; return 0;}
	return 1;
}

void FlushGenieRW(void)
{
	int32		x;

	//X6502_DebugLevel(~4);

	if (AReadG)
	{
		x = 0;
		do
		{
			ARead[x+0x8000] = AReadG[x];
			BWrite[x+0x8000] = BWriteG[x];
		} while(++x<0x8000);

		FCEU_free(AReadG); AReadG = 0;
		FCEU_free(BWriteG); BWriteG = 0;
	}
}

int		modcon;
uint8		genieval[3];
uint8		geniech[3];
uint32		genieaddr[3];

static FILE *OpenGenieFile(void)
{
	FILE		*fp = 0;
	char		*fn;

	if ((fn=FCEU_MakeFName(FCEUMKF_GGROM,0,0))==0
			|| CheckFileOrDirExists(fn)!=1)
		FCEUD_Message("Couldn't locate Game Genie ROM image.\n");
	else if ((fp=FCEUD_UTF8fopen(fn,"rb"))==0)
		FCEU_PrintError("Error opening Game Genie ROM image.");

	if (fn)
		free(fn);

	return fp;
}

/* Called when a game(file) is opened successfully. */
void OpenGenie()
{
	FILE		*fp;
	int		x;

	if (GENIEROM)
		goto end;

	if ((GENIEROM=(uint8*)FCEU_malloc(4096+1024))==0)
		return; /* probably no memory for error message either */
	if ((fp=OpenGenieFile())==0)
		goto fail;

	if (fread(GENIEROM,1,16,fp)!=16)
	{
	grerr:	FCEU_PrintError("Error reading from Game Genie ROM image.");
	fail:	if (fp)
			fclose(fp);
		FCEU_KillGenie(); /* fix; 2010.03.26 */
		return;
	}

	if (GENIEROM[0]==0x4E)	/* iNES ROM */
	{
		if (fread(GENIEROM,1,4096,fp)!=4096)
			goto grerr;
		if (fseek(fp,16384-4096,SEEK_CUR))
			goto grerr;
		if (fread(GENIEROM+4096,1,256,fp)!=256)
			goto grerr;
	} else if (fread(GENIEROM+16,1,4352-16,fp)!=(4352-16))
		goto grerr;

	fclose(fp);

	/* Workaround for the FCE Ultra CHR page size only being 1KB */
	x = 0;
	do
		FCEU_memcpy(GENIEROM+4096+(x<<8),GENIEROM+4096,256);
	while(++x<4);
end:
	GenieStage = 1;
}

/* Called when a game is closed. */
void CloseGenie()
{
	/* No good reason to free() the Game Genie ROM image data. */
	GenieStage = 0;
	FlushGenieRW();
	VPageR = VPage;
}

static DECLFR(GenieRead)
{
	return GENIEROM[A&4095];
}

static DECLFW(GenieWrite)
{
	switch(A)
	{
	case 0x800C:
	case 0x8008:
	case 0x8004:
		genieval[(A-4)>>2&3] = V;
		break;
	case 0x800B:
	case 0x8007:
	case 0x8003:
		geniech[(A-3)>>2&3] = V;
		break;
	case 0x800A:
	case 0x8006:
	case 0x8002:
		A-= 2, A&= 0xF, A>>= 2;
		genieaddr[A]&= 0xFF00;
		genieaddr[A]|= V;
		break;
	case 0x8009:
	case 0x8005:
	case 0x8001:
		--A, A&= 0xF, A>>= 2;
		genieaddr[A]&= 0xFF;
		genieaddr[A]|= (V|0x80)<<8;
		break;
	case 0x8000:
		if (!V)
			FixGenieMap();
		else if ((modcon=V^=0xFF)==(0x71^0xFF))
			modcon = 0;
	}
}

static readfunc	GenieBackup[3];

static DECLFR(GenieFix1)
{
	uint8		r = GenieBackup[0](A);

	if (modcon&2)	// No check
		return genieval[0];
	else if (r==geniech[0])
		return genieval[0];
	return r;
}

static DECLFR(GenieFix2)
{
	uint8		r = GenieBackup[1](A);

	if (modcon&4)	// No check
		return genieval[1];
	else if (r==geniech[1])
		return genieval[1];
	return r;
}

static DECLFR(GenieFix3)
{
	uint8		r = GenieBackup[2](A);

	if (modcon&8)	// No check
		return genieval[2];
	else if (r==geniech[2])
		return genieval[2];
	return r;
}

void FixGenieMap()
{
	int		x;

	GenieStage = 2;

	x = 0;
	do
		VPage[x] = VPageG[x];
	while(++x<8);

	VPageR = VPage;
	FlushGenieRW();

	x = 0;
	do
		if ((modcon>>(4+x))&1)
		{
			static readfunc	tmp[3] = {GenieFix1,GenieFix2,GenieFix3};
			unsigned	A = genieaddr[x];

			GenieBackup[x] = GetReadHandler(A);
			/* double not-necessary: bit #15 is always on here! */
			//if (!(A&0x1FFF)) /* RAM cheat (not necessary - all >=0x8000) */
			//	X6502_DebugLevel(4);
			SetReadHandler(A,A,tmp[x]);
		}
	while(++x<3);
}

void GeniePower()
{
	uint32		x;

	if (!GenieStage)
		return;

	GenieStage = 1;
	x = 0;
	do
		genieval[x] = geniech[x] = 0xFF, genieaddr[x] = 0xFFFF;
	while(++x<3);
	modcon = 0;

	SetWriteHandler(0x8000,0xFFFF,GenieWrite);
	SetReadHandler(0x8000,0xFFFF,GenieRead);

	x = 0;
	do
		VPage[x] = GENIEROM+4096-0x400*x;
	while(++x<8);

	if (AllocGenieRW())
		VPageR = VPageG;
	else
		GenieStage = 2;
}
