/*
 PPU.h

 * MacFCEU 0.8a7
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* ---- constants */

/* number of PPU cycles/rendered pixels per scanline - 256 rendered, 85 invisible */
#define LCYCS			341

/* PPU type bits */
#define RP2C03B			(1<<0)		/* bog standard RGB palette */
#define RP2C03G			(1<<1)		/* similar palette, may have 1 color changed */
#define RP2C04_0001		(1<<2)		/* scrambled palette + new colors */
#define RP2C04_0002		(1<<3)		/* different scrambling, different new colors */
#define RP2C04_0003		(1<<4)		/* (similar to above) */
#define RP2C04_0004		(1<<5)		/* (similar to above) */
#define RC2C03B			(1<<6)		/* seems identical to RP2C03B */
#define RC2C03C			(1<<7)		/* 1 changed color or so */
#define RC2C05_01		(1L<<8)		/* all five of these have the normal palette...*/
#define RC2C05_02		(1L<<9)		/*...but with different bits returned on 2002 */
#define RC2C05_03		(1L<<10)
#define RC2C05_04		(1L<<11)
#define RC2C05_05		(1L<<12)
 /* 13-15 are currently invalid */
#define PPU2C02			(1L<<16)
#define PPU2C07			(1L<<17)

#define RCandP2C03B		(RP2C03B|RC2C03B)

/* ---- globals */

extern void FP_FASTAPASS(1) (*PPU_hook)(uint32 A);
extern void (*GameHBIRQHook)(void), (*GameHBIRQHook2)(void);

/* For cart.c and banksw.h, mostly */
extern uint8 NTARAM[0x800], *vnapage[4];
extern uint8 PPUNTARAM, PPUCHRRAM;

extern uint8 PALRAM[0x20], PPUvoidcolor;

extern int scanline;
extern uint32 PPUVFrameCounter;

/* ---- prototypes */

void FCEUPPU_Init(void);
void FCEUPPU_Reset(void);
void FCEUPPU_Power(void);
int FCEUPPU_Loop(int skip);

void FCEUPPU_LineUpdate();
void FCEUPPU_SetVideoSystem(int which); /* PAL: 1; NTSC: 0 */

void FCEUPPU_SaveState(void);
void FCEUPPU_LoadState(int version);

/* call this after modifying PALRAM; 2010.04.15 */
void FCEUPPU_UpdatePalette(void);

int FCEUPPU_GetPixel(void);
