/*
 UNIF.c
  UNIF format file

 * MacFCEU 0.8a8
 *
 * Copyright notice for this file:
 *  unif.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* TODO:  Battery backup file saving, mirror force		*/
/* **INCOMPLETE**						*/
/* Override stuff: CHR RAM instead of CHR ROM, mirroring.	*/

#include <stdio.h>
#include <ctype.h>

#include "driver.h"
#include "FCEU.h"
#include "cart.h"
#include "UNIF.h"
#include "general.h"
#include "state.h"
#include "file.h"
#include "memory.h"
#include "input.h"
#include "utility/endian.h"
#include "utility/MD5.h"

#include "board.h"

#ifdef env68k
 #pragma segment files
#endif

static CartInfo UNIFCart;

//static int	vramo;
static int	mirrortodo;
static uint8	*boardname, *sboardname;

uint8	*UNIFchrrama = 0;
static uint32	CHRRAMSize;

typedef struct UNIF_HEADER
{
	char	ID[4];
	uint32	info;
} UNIF_HEADER;

static UNIF_HEADER	unhead, uchead;

static uint8	*malloced[32] = {0};
static uint32	mallocedsizes[32] = {0};

static int FixRomSize(uint32 size, uint32 minimum)
{
	int	x = 1;

	if (size<minimum)
		return minimum;
	while(x<size)
		x+= x;
	return x;
}

static void FreeUNIF(void) //was FreeUNIF, ResetUNIF
{
	int	x;

//	vramo = 0;
	mirrortodo = 0;
	dwmemset(&UNIFCart,0,sizeof(UNIFCart));

	if (UNIFchrrama)
		{FCEU_free(UNIFchrrama); UNIFchrrama = 0;}
	if (boardname)
		{FCEU_free(boardname); boardname = 0;}
	for(x=0; x<32; ++x)
		if (malloced[x])
			{FCEU_free(malloced[x]); malloced[x] = 0;}
}

static void MooMirroring(void)
{
	static uint8	exntar[2048];

	UNIFCart.mirror = mirrortodo; //this was not set!; 2010.03.15

	if (mirrortodo<4)
		SetupCartMirroring(mirrortodo,1,0);
	else if (mirrortodo==4)
	{
		SetupCartMirroring(4,1,exntar);
		AddExState(exntar,2048,0,"EXNR");
	} else
		SetupCartMirroring(0,0,0);
}

static int DoMirroring(FCEUFILE *fp)
{
	static const char	*stuffo[6] =
		{"Horizontal","Vertical","$2000","$2400",
		"\"Four-screen\"","Controlled by Mapper Hardware"};
	int		t;

	if ((t=FCEU_fgetc(fp))<0)
		return 0;
	mirrortodo = t;

	if (t<6)
		FCEU_printf(" Name/Attribute Table Mirroring: %s\n",
			stuffo[t]);

	return 1;
}

static int NAME(FCEUFILE *fp)
{
	char		namebuf[102];
	int		t, index;

	FCEU_printf(" Name: ");
	index = 0;

	while((t=FCEU_fgetc(fp))>0)
		if (index<100)
			namebuf[index++] = t;
	namebuf[index++] = 0;
	FCEU_printf("%s\n",
		namebuf);

	if (FCEUGameInfo->name==0)
		if ((FCEUGameInfo->name=FCEU_malloc(index))==0)
			return 0;
		else
			FCEU_memcpy(FCEUGameInfo->name,namebuf,index);

	return 1;
}

static int DINF(FCEUFILE *fp)
{
	char		name[100], method[100];
	int		t, d, m, y;

	//for each FCEU_fgetc, ==EOF was used; 2010.01.05
	if (FCEU_fread(name,1,100,fp)!=100
			|| (d=FCEU_fgetc(fp))<0
			|| (m=FCEU_fgetc(fp))<0
			|| (y=FCEU_fgetc(fp))<0
			|| (t=FCEU_fgetc(fp))<0)
		return 0;
	y|= t<<8;
	if (FCEU_fread(method,1,100,fp)!=100)
		return 0;
	name[99] = method[99] = 0;
	FCEU_printf(
" Dumped by: %s\n"
" Dumped with: %s\n"
" Dumped on: %u.%02u.%02u"
		,name,method,y,m,d); //in: uniform date; out: English month name

	return 1;
}

static int CTRL(FCEUFILE *fp)
{
	int		t;

	if ((t=FCEU_fgetc(fp))<0)
		return 0;
/* The information stored in this byte isn't very helpful, but it's
	better than nothing...maybe.
*/

	if (t&1)
		FCEUGameInfo->input[0] = FCEUGameInfo->input[1] = SI_GAMEPAD;
	else
		FCEUGameInfo->input[0] = FCEUGameInfo->input[1] = SI_NONE;

	if (t&2)
		FCEUGameInfo->input[1] = SI_ZAPPER;
	//else if(t&0x10) FCEUGameInfo->input[1]=SI_POWERPAD;

	return 1;
}

static int TVCI(FCEUFILE *fp)
{
	int		t;

	if ((t=FCEU_fgetc(fp))<0)
		return 0;
	if (t<=2)
	{
		char	*stuffo[3] = {"NTSC","PAL","NTSC and PAL"};

		if (t==0)
			FCEUGameInfo->vidsys = GIV_NTSC;
		else if (t==1)
			FCEUGameInfo->vidsys = GIV_PAL;
		FCEU_printf(" TV Standard Compatibility: %s\n",
			stuffo[t]);
	}
	return 1;
}

static int EnableBattery(FCEUFILE *fp)
{
	FCEU_printf(" Battery-backed.\n");
	if (FCEU_fgetc(fp)<0)
		return 0;
	UNIFCart.battery = 1;
	return 1;
}

#ifdef moo
static int PCK(FCEUFILE *fp)
{
 int z;
 z=uchead.ID[3]-'0';
 if (z<0 || z>15) return(0);
}
#endif

static int LoadPRG(FCEUFILE *fp)
{
	int		z, t;

	if ((unsigned)(z=uchead.ID[3]-'0')>15)
		return 0;

	FCEU_printf(" PRG ROM %d size: %d\n",
		z,(int)uchead.info);

	if (malloced[z]!=0)
		FCEU_free(malloced[z]);
	t = FixRomSize(uchead.info,2048);
	ROM_size+= t; //add to prg
	if ((malloced[z]=(uint8*)FCEU_malloc(t))==0)
		return 0;

	mallocedsizes[z] = t;
	dwmemset(malloced[z]+uchead.info,-1,t-uchead.info);
	if (FCEU_fread(malloced[z],1,uchead.info,fp)!=uchead.info)
	{
		FCEU_printf("Read Error!\n");
		return 0;
	}

	SetupCartPRGMapping(z,malloced[z],t,0);
	return 1;
}

static int SetBoardName(FCEUFILE *fp)
{
	if ((boardname=(uint8*)FCEU_malloc(uchead.info+1))==0
			|| FCEU_fread(boardname,1,uchead.info,fp)!=uchead.info)
		return 0;
	boardname[uchead.info] = 0;
	FCEU_printf(" Board name: %s\n",
		boardname);

	sboardname = boardname;
	if (boardname[3]=='-' //skip some possible leading characters (redundancy)
			&&(*boardname=='B'
				&&((boardname[1]=='M' && boardname[2]=='C')
				|| (boardname[1]=='T' && boardname[2]=='L'))
			|| *boardname=='H'
				&& (boardname[1]=='V' && boardname[2]=='C')
			|| *boardname=='N'
				&& (boardname[1]=='E' && boardname[2]=='S')
			|| *boardname=='U'
				&& (boardname[1]=='N' && boardname[2]=='L')))
		sboardname+= 4;

	return 1;
}

static int LoadCHR(FCEUFILE *fp)
{
	int		z, t;

	if ((unsigned)(z=uchead.ID[3]-'0')>15)
		return 0;
	FCEU_printf(" CHR ROM %d size: %d\n",
		z,(int)uchead.info);

	if (malloced[16+z]!=0)
		FCEU_free(malloced[16+z]);

	t = FixRomSize(uchead.info,8192);
	VROM_size+= t; //add to chr
	if ((malloced[16+z]=(uint8*)FCEU_malloc(t))==0)
		return 0;

	mallocedsizes[16+z] = t;
	dwmemset(malloced[16+z]+uchead.info,-1,t-uchead.info);
	if (FCEU_fread(malloced[16+z],1,uchead.info,fp)!=uchead.info)
	{
		FCEU_printf("Read Error!\n");
		return 0;
	}

	SetupCartCHRMapping(z,malloced[16+z],t,0);

	return 1;
}

/* ---- mapping */

typedef struct BFMAPPING
{
	char	ID[4];
	int (*init)(FCEUFILE *fp);
} BFMAPPING;

static BFMAPPING bfunc[] =
{
	{"CTRL", CTRL},
	{"TVCI", TVCI},
	{"BATR", EnableBattery},
	{"MIRR", DoMirroring},
	{"PRG",  LoadPRG},
	{"CHR",  LoadCHR},
	//{"CCK",  CCK},
	//{"PCK",  PCK},
	{"NAME", NAME},
	{"MAPR", SetBoardName},
	{"DINF", DINF},
	{0,0}
};

int LoadUNIFChunks(FCEUFILE *fp);
int LoadUNIFChunks(FCEUFILE *fp)
{
	int		x, t;

	while((t=FCEU_fread(&uchead.ID,1,4,fp))>=4)
	{
		const BFMAPPING	*l = bfunc;
		char		c; //no more memcmp; 2010.01.05

		if (FCEU_read32le(&uchead.info,fp)==0)
			return 0;

		t = 0, x = sizeof(bfunc)/sizeof(bfunc[0]);
		do
		{
			if ((c=l->ID[0])==uchead.ID[0]
					&& ((c=l->ID[1])==0 || c==uchead.ID[1])
					&& ((c=l->ID[2])==0 || c==uchead.ID[2])
					&& ((c=l->ID[3])==0 || c==uchead.ID[3]))
			{
				if (l->init(fp)==0)
					return 0;
				t = 1;
				break;
			}
		} while(++l,--x!=0);

		if (t==0) //not found; skip data
			if (FCEU_fseek(fp,uchead.info,SEEK_CUR))
				return 0;
	}

	return t>0 ? 0:1;
}

//only case insensitive for uppercase in the board list; 2010.03.16
static int SameBoardName(const void *s1,
	const void *s2)
{
	const uint8	*a = s1, *b = s2;
	uint8		c;

	do
	{
		if ((c=*a++)==0)
			return *b==0;
	} while(*b++==c || toupper(b[-1])==c);

	return 0;
}

static int InitializeBoard(void)
{
	const BMAPPING	*l;
	uint8		f;

	if (sboardname==0)
		return -2;

	l = FCEU_bmap, f = sboardname[0];
	do
		if (l->name[0]==f && SameBoardName(1+l->name,1+sboardname))
		{
			if (malloced[16]==0 && (l->flags&(BMCFLAG_CHRROK|BMCFLAG_16KCHRR)))
			{
				CHRRAMSize = (l->flags&BMCFLAG_16KCHRR) ?
					0x4000:0x2000;
				if (VROM_size<CHRRAMSize)
					VROM_size = CHRRAMSize;
				if ((UNIFchrrama=(uint8*)FCEU_malloc(CHRRAMSize))!=0)
				{
					SetupCartCHRMapping(0,UNIFchrrama,CHRRAMSize,1);
					AddExState(UNIFchrrama,CHRRAMSize,0,"CHRR");
				} else
					return -1;
			}

			if (l->flags&BMCFLAG_FORCE4)
				mirrortodo = 4;
			MooMirroring();
			l->init(&UNIFCart);

			return 1;
		}
	while((++l)->name!=0 && l->mapper>-2);

	FCEU_PrintError("Board type not supported.");
	return 0;
}

static void UNIFGI(int h)
{
	switch(h)
	{
	case GI_RESETM2:
		if (UNIFCart.Reset)
			UNIFCart.Reset();
		break;
	case GI_POWER:
		if (UNIFCart.Power)
			UNIFCart.Power();
		if (UNIFchrrama)
			dwmemset(UNIFchrrama,0,0x2000); //CHRRAMSize);
		break;
	case GI_CLOSE:
		FCEU_SaveGameSave(&UNIFCart);
		if (UNIFCart.Close)
			UNIFCart.Close();
		FreeUNIF();
	}
}

static void DoMD5(void)
{
	struct md5_context md5;
	int x;

	md5_starts(&md5);

	x = 0;
	do
		if (malloced[x])
			md5_update(&md5,malloced[x],mallocedsizes[x]);
	while(++x<32);

	md5_finish(&md5,UNIFCart.MD5);
	FCEU_printf(" ROM MD5:  0x%s\n",
		md5_asciistr(UNIFCart.MD5,0));

	FCEU_memcpy(FCEUGameInfo->MD5,UNIFCart.MD5,sizeof(UNIFCart.MD5));
}

int UNIFLoad(const char *name,
	FCEUFILE *fp)
{
(void)name;
	FCEU_fseek(fp,0,SEEK_SET);
	FCEU_fread(&unhead,1,4,fp);
	if (unhead.ID[0]!='U' || unhead.ID[1]!='N'
			|| unhead.ID[2]!='I' || unhead.ID[3]!='F')
		return 0;

	ResetCartMapping();
	ResetExState(0,0);
	FreeUNIF();

	ROM_size = VROM_size = 0;

	if (!FCEU_read32le(&unhead.info,fp)
			|| FCEU_fseek(fp,0x20,SEEK_SET)<0
			|| LoadUNIFChunks(fp)<=0
			|| ROM_size==0)
		goto aborto;

	ROM_size>>= 14, VROM_size>>= 13;

	DoMD5();

	if (InitializeBoard()>0)
	{
		FCEU_LoadGameSave(&UNIFCart);
		GameInterface = UNIFGI;
		return 1;
	}
aborto:
	FreeUNIF();
	return 0;
}
