/*
 WAVE.c
  WAVE file recording

 * MacFCEU 0.8a9
 *
 * See AUTHORs for original credits
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "driver.h"
#include "FCEU.h"
#include "fsound.h"
#include "WAVE.h"

#ifdef env68k
 #pragma segment files
#endif

uint32		WAVE_DefaultRate = 22050;

uint8		FCEU_soundlogrec = 0;
static FILE	*FCEU_soundlog = 0;
static long	wsize;
static uint8	toupdate = 1;
static uint8	wavedepth;

static char	whead[] =
/*00*/	"RIFF\0\0\0\0" /* RIFF size */
/*08*/	"WAVEfmt "
/*10*/	"\x10\0\0\0" "\1" /*PCM*/ "\0" "\1" /* monophonic */ "\0"
/*18*/	"\0\0\0\0" /* rate */ "\0\0\0\0" /* pack size */
/*20*/	"\0\0\0\0data" /* depth.. */
/*28*/	"\0\0\0"; /* sound total - extra null for C str */

int FCEU_WaveUpdateHeader(void)
{
	uint32		o;

	FCEU_soundlogrec = 0;
	if (!FCEU_soundlog)
		return -1;
	if (!toupdate || (o=ftell(FCEU_soundlog))<0)
		return 0;
	//yet another overflow bug fix -- flush at seek w/ MPW ioctl EOF error?
	//ARGH!!--When will this bug go away?!; 2012.03.01
	fseek(FCEU_soundlog,wsize>1048 ? wsize-1024:wsize,SEEK_SET);
	fflush(FCEU_soundlog);
	fseek(FCEU_soundlog,0,SEEK_SET);
	whead[4] = o-=8, whead[5] = o>>8, whead[6] = o>>16, whead[7] = o>>24;
	/* update rate in case the APU code adjusted it */
	o = (FSettings.OrgRate+8)>>4;
	whead[24] = o, whead[25] = o>>8, whead[26] = o>>16, whead[27] = o>>24;
	o<<= wavedepth>>4;
	whead[28] = o, whead[29] = o>>=8, whead[30] = o>>=8, whead[31] = o>>8;

	o = wsize;
	whead[40] = o, whead[41] = o>>=8, whead[42] = o>>=8, whead[43] = o>>8;
	/* this must be done all at once in case of EOF */
	if (fwrite(whead,sizeof(whead),1,FCEU_soundlog)!=1)
		return -3;
	fflush(FCEU_soundlog);
	fseek(FCEU_soundlog,0,SEEK_END);

	toupdate = 0;

	return 1;
}

/* Checking whether the file exists before wiping it out is left up to the
   reader..err...I mean, the driver code, if it feels so inclined (I don't feel so).
*/
int FCEU_WriteWaveData(int16 *s,
	int c)
{
	static int16	temp[13471];

	int16		*dest;
	int32		req, written;

	FCEU_soundlogrec = 1;
	if (c>sizeof(temp)) /* larger than what temp buffer can hold; clip count */
		c = sizeof(temp);
	if (!FCEU_soundlog || c<=0)
		return 0;

	dest = temp;

	if (wavedepth==8)
		do
			*((uint8*)dest)++ = *s++>>8^0x80;
		while(--c);
	else /* otherwise assume 16 */
		do
#if BIG_ENDIAN
			((uint8*)dest)[1] = *((uint8*)s)++,
			((uint8*)dest)[0] = *((uint8*)s)++;
#else
			*dest = *s++;
#endif
		while(++dest,--c);
	req = (uint8*)dest-(uint8*)temp;
	if ((written=fwrite(temp,1,req,FCEU_soundlog))!=req)
	{ /* oh, crap */
		FCEUI_EndWaveRecord();
		return -1;
	}
	wsize+= written;
	toupdate = 1;

	return 1;
}

int FCEUI_EndWaveRecord(void)
{
	int		r;

	if (!FCEU_soundlog)
		return 0;
	r = FCEU_WaveUpdateHeader();

	fclose(FCEU_soundlog);
	FCEU_soundlog = 0;

	return r ? r:1;
}


int FCEUI_BeginWaveRecord(char *fn,
	int format)
{
	uint32		r;

	if ((FCEU_soundlog=FCEUD_UTF8fopen(fn,"wb"))==NULL)
		return 0;
	wsize = 0;

	/* put rate into header */
	if ((r=FSettings.OrgRate)==0)
	{
		FCEUD_Message("internal error - blank rate\n");
		exit(1);
	}
	r = (r+8)>>4;
	whead[24] = r, whead[25] = r>>8, whead[26] = r>>16, whead[27] = r>>24;

	if (format==8)
		wavedepth = 8;
	else
		wavedepth = 16;
	/* pack size (per second), depth */
	r<<= wavedepth>>4;
	whead[28] = r, whead[29] = r>>=8, whead[30] = r>>=8, whead[31] = r>>8;
	whead[32] = wavedepth>>3, whead[33] = 0;
	whead[34] = wavedepth, whead[35] = 0;

	return toupdate=fwrite(whead,sizeof(whead),1,FCEU_soundlog)==1;
}
