/*
 boards/8237.c

 * MacFCEU 0.5.5
 *
 * Copyright notice for this file:
 *  8237.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "unlicensed.h"

static uint8 cmdin;
static uint8 cmd;
static uint8 regs[8];
static uint8 master, chrm;

static int32 IRQCount, IRQa;

static void DoPRG(void)
{
 if(master&0x80)
 {
  if(master&0x20)
  {
   setprg32(0x8000,master>>1&7);
  } else
  {
   setprg16(0x8000,master&0xF);
   setprg16(0xC000,master&0xF);
  }
 } else
 {
  int b8, bc;

  setprg8(0xA000,regs[4]);
  setprg8(0xE000,~0);
  b8 = bc =~1;
  if(cmd&0x40)
   bc = regs[2];
  else
   b8 = regs[2];
  setprg8(0x8000,b8);
  setprg8(0xC000,bc);
 }
}

static void DoCHR(void)
{
 uint32 base=(cmd&0x80)<<5;
 int orie=(chrm&0x4)<<6;

 setchr2(0x0000^base,(orie|regs[0])>>1); //K
 setchr2(0x0800^base,(orie|regs[3])>>1); //43

 setchr1(0x1000,orie|regs[1]);
 setchr1(0x1400,orie|regs[5]);
 setchr1(0x1800,orie|regs[6]);
 setchr1(0x1C00,orie|regs[7]);
}

static DECLFW(UNL8237Write)
{
	switch(A&0xF000)
	{
	case 0xF000:IRQCount=V;return;
	case 0xE000:X6502_IRQEnd(FCEU_IQEXT);return;
	}
	if (A==0x5000)
	{
		master=V;
		DoPRG();
		DoCHR();
		return;
	} else if (A==0x5001)
	{
		chrm=V;
		DoCHR();
		return;
	}
	switch(A&0xE000)
	{
	case 0x8000:
		setmirror(~(V|(V>>7))&1);
		break;
	case 0xA000:
		cmd = V, cmdin = 1;
		DoPRG();
		DoCHR();
		break;
	case 0xC000:
		if (!cmdin)
			break;
		regs[cmd&7] = V;
		DoPRG();
		DoCHR();
		cmdin = 0;
	}
}

static void UNL8237Reset(void)
{
	int	x;

	for(x=0; x<8; ++x)
		regs[x] = 0;
	master = chrm = cmd = cmdin = IRQCount = IRQa = 0;
	SetCartRead(0x8000,0xFFFF);
	SetWriteHandler(0x5000,0xFFFF,UNL8237Write);
	DoPRG();
	DoCHR();
}

static void UNL8237hook(void)
{
	if (IRQCount)
		if (--IRQCount==0)
			X6502_IRQBegin(FCEU_IQEXT);
}

static void UNL8237Restore(int v)
{
(void)v;
	DoPRG();
	DoCHR();
}

void UNL_8237_Init(CartInfo *info)
{
	info->Power = UNL8237Reset;
	GameStateRestore = UNL8237Restore;
	GameHBIRQHook = UNL8237hook;
	AddExState(regs,8,0,"REGS");
	AddExState(&IRQCount,4,1,"IRQC");
	AddExState(&IRQa,1,0,"IRQA");
	AddExState(&master,1,0,"MAST");
	AddExState(&chrm,1,0,"CHRM");
	AddExState(&cmd,1,0,"CMD");
	AddExState(&cmdin,1,0,"CMDI");
}
