/*
 boards/AE.c
  Active Enterprises

 * MacFCEU 0.6.6
 *
 * Copyright notice for the original:
 *  228.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "latch.h"

/* some of these chips have 2 bytes of RAM?; 2010.05.29 */
static uint8	IRAM[4];

/* ---- mapper 228: Action 52 */
/* older state files are not compatible */

static void M228Sync(void)
{
	uint32		A = latch_addr;
	uint32		bank = (A>>7&0x1F)+(A>>7&A>>8&0x10);

//	bank = A>>7&0x3F;
//	if (!(~bank&0x30))
//		bank-= 0x10;
//	bank+= bank, bank+= (A>>6)&(A>>5)&1;
//	setprg16(0x8000,bank);
//	setprg16(0xC000,bank+(~A>>5&1));

	if (A&0x20)
	{
		bank = (bank<<2)|(A>>5&2);
		setprg16(0x8000,bank);
		setprg16(0xC000,bank);
	} else
		setprg32(0x8000,bank);
	setchr8((latch_byte&3)|(A<<2&0x3C));
	setmirror(~A>>13&1);
}

static DECLFR(A52R5)
{
	return IRAM[A&3]&15; /* only 16 bits used */
}

static DECLFW(A52W5)
{
	IRAM[A&3] = V;
}

static void A52Power(void)
{
	Latch_Power();
	SetReadHandler(0x5000,0x5FFF,A52R5);
	SetWriteHandler(0x5000,0x5FFF,A52W5);
}

void AE_A52_Init(CartInfo *info)
{
	Latch_Init(info,M228Sync,0,0x8000,0x8000FFFFU);
	info->Power = A52Power; //Reset
	AddExState(IRAM,4,0,"IRAM");
}
