/*
 boards/BMCResetBased.c

 * MacFCEU 0.5.5
 *
 * Copyright notice for this file:
 *  addrlatch.cpp (C) 2006 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "latch.h"
#include "unlicensed.h"

static uint8	dipsw;

static void BMC4in1Reset(void)
{
	++dipsw, dipsw&= 3;
}

/* ---- D1038/VT5201 */

/*	VT5201 dip#
	 0  1  2  3
	 6/15/35/43-in-1
	12/66/77/88-in-1
	28/46/63/118-in-1
	54/64/74/84-in-1
	55/65/75/85-in-1
	65/75/85/95-in-1
*/

static void BMCD1038Sync(void)
{
	uint32		a = latch_addr, b = ~a>>7&1;

	setprg16(0x8000,(a>>4)&~b);
	setprg16(0xC000,(a>>4)|b);
	setchr8(a);
	setmirror(~a>>3&1);
}

static DECLFR(BMCD1038Read)
{
	if (latch_addr&0x100)
		return dipsw;
	else
		return CartBR(A);
}

static void BMCD1038Power(void)
{
	dipsw = 0;
	Latch_Power();
}

void BMC_D1038_Init(CartInfo *info)
{
	Latch_Init(info,BMCD1038Sync,BMCD1038Read,0x8000,0x8000FFFFU);
	info->Power = BMCD1038Power;
	info->Reset = BMC4in1Reset;
	AddExState(&dipsw,1,0,"DIPS");
}

/* ---- mapper 60(P64K,C32K): Reset-based 4-in-1 */

static void ResetBased4in1Sync(int v)
{
(void)v;
	setprg16(0x8000,dipsw);
	setprg16(0xC000,dipsw);
	setchr8(dipsw);
}

static void ResetBased4in1Reset(void)
{
	BMC4in1Reset();
	ResetBased4in1Sync(0);
}

static void ResetBased4in1Power(void)
{
	dipsw = 0;
	ResetBased4in1Sync(0);
}

void BMC_ResetBased4in1_Init(CartInfo *info)
{
	if (ROM_size!=64>>4 || VROM_size!=32>>4)
		{BMC_D1038_Init(info); return;}
	info->Power = ResetBased4in1Power;
	info->Reset = ResetBased4in1Reset;
	GameStateRestore = ResetBased4in1Sync;
	AddExState(&dipsw,1,0,"DIPS");
}
