/*
 boards/H2288.c

 * MacFCEU 0.6.4
 *
 * Copyright notice for this file:
 *  h2288.c Copyright (C) 2003 Xodnizel
 *  (mmc3.cpp version probably written by CaH4e3; has no stated Copyright)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC3and6.h"
#include "unlicensed.h"

/* ---- mapper 123: H2288 */

//SuperGame Lion King also uses the same...
// but an integer has zero hits on data cache
#define Kay_H2288_Security	0x42765130

static void H2288PW(uint32 A,
	uint8 V)
{
	if ((MMC3EXPREGS[0]&0x40)==0)
		setprg8(A,V&0x3F);
}

static DECLFW(H2288_write5)
{
	unsigned	r = MMC3EXPREGS[0];

	A&= 1;
	if (MMC3EXPREGS[A]==V)
		return;
	MMC3EXPREGS[A] = V;

	if (r&0x40)
	{
		A = (r&5)|(r>>2&0xA);
		r = r>>1&1;
		setprg16(0x8000,A&~r);
		setprg16(0xC000,A|r);
	} else
		FixMMC3PRG(MMC3cmd);
}

static DECLFW(H2288_write8000)
{
(void)A;
	MMC3_CMDWrite(0x8000,
		(V&0xC0)|(Kay_H2288_Security>>(V<<4&0x70)&15));
}

static DECLFR(H2288_read)
{
	int	h = A>>8;
//there's no reason anymore to use XFrame.DB;
//the invalid page is read in indexed mode, XFrame.DB should always be (A>>8).
	return (h&~1)|(1&~h&A);
}

static void H2288Power(void)
{
	MMC3EXPREGS[0] = MMC3EXPREGS[1] = 0;
	GenMMC3Power();
	SetReadHandler(0x5000,0x5FFF,H2288_read);
	SetCartRead(0x8000,0xFFFF);
	SetWriteHandler(0x5800,0x5FFF,H2288_write5);
	SetWriteHandlerI(2,0x8000,0x8FFE,H2288_write8000); //MMC3 does the odd
}

void Kay_H2288_Init(CartInfo *info)
{
	GenMMC3_Init(info,256,256,0,0);
	info->Power = H2288Power;
	MMC3pwrap = H2288PW;
	AddExState(MMC3EXPREGS,2,0,"EXPR");
}
