/*
 boards/JYCompany.c
  mappers 90, 209... multiplication

 * MacFCEU 0.8a8
 *
 * Copyright notice for this file:
 *  90.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "JYCompany.h"

static uint8	is209;

static uint8	IRQMode;	// from $C001
static uint8	IRQPre;		// from $C004 and IRQXOR
static uint8	IRQCount;	// from $C005 and IRQXOR
static uint8	IRQXOR;		// Loaded from $C006
static uint8	IRQa;		// $C002, $C003, and $C000

uint8	JYCmul[2];
uint8	JYCregie, JYCtekker;
uint8	JYCcntl[4];
uint16	JYCnames[4];

static uint8	prgb[4];
static uint8	chrlow[8];
static uint8	chrhigh[8];

static SFORMAT Tek_StateRegs[] =
{
	{&IRQXOR,1,"IRQX"},
	{&IRQCount,1,"IRQC"},
	{&IRQa,1,"IRQa"},
	{JYCmul,2,"MUL"},
	{&JYCregie,1,"REGI"},
	{JYCcntl,4,"TKCO"},
	{prgb,4,"PRGB"},
	{chrlow,4,"CHRL"},
	{chrhigh,8,"CHRH"},
	{&JYCnames[0],2|FCEUSTATE_RLSB,"NMS0"},
	{&JYCnames[1],2|FCEUSTATE_RLSB,"NMS1"},
	{&JYCnames[2],2|FCEUSTATE_RLSB,"NMS2"},
	{&JYCnames[3],2|FCEUSTATE_RLSB,"NMS3"},
	{&JYCtekker,1,"TEKR"},
	{0}
};

static DECLFR(tekread)
{
	int		x = A;

	//switch(A)
	if ((x&=0x803)==0) //0x5000
		return JYCtekker|(A>>8&~JYCtekker);
	else if ((x-=0x800)==0) //0x5800
		return JYCmul[0]*JYCmul[1];
	else if (--x==0) //0x5801
		return (JYCmul[0]*JYCmul[1])>>8;
	else if ((x-=2)==0) //0x5803
		return JYCregie;

	return A>>8; //XFrame.DB;
}

static void mira(void)
{
	int	x;

	if (JYCcntl[0]&0x20 && is209)
	{
		x = 0;
		if (JYCcntl[0]&0x40)	// Name tables are ROM-only
			do
				setntamem(CHRptr[0]+((JYCnames[x]&CHRmask1[0])<<10),0,x);
			while(++x<4);
		else			// Name tables can be RAM or ROM.
			do
				if ((JYCcntl[2]^JYCnames[x])&0x80)
					setntamem(CHRptr[0]+(((JYCnames[x])&CHRmask1[0])<<10),0,x);
				else	// RAM selected.
					setntamem(NTARAM+((JYCnames[x]&1)<<10),1,x);
			while(++x<4);
		return;
	}

	//switch(tkcom[1]&3)
	x = JYCcntl[1]&3; //VH01
	if ((x&2)==0)
		x^= 1;
	setmirror(x);
}

static void tekprom(void)
{
	switch(JYCcntl[0]&3)
	{
	case 1: // 16 KB
		setprg16(0x8000,prgb[0]);
		setprg16(0xC000,prgb[2]);
		break;
	case 2: //2 = 8 KB ??
		if (JYCcntl[0]&4)
			setprg8(0xE000,prgb[3]);
		else
		{
			if (JYCcntl[0]&0x80)
				setprg8(0x6000,prgb[3]);
			setprg8(0xE000,~0);
		}
		setprg8(0x8000,prgb[0]);
		setprg8(0xA000,prgb[1]);
		setprg8(0xC000,prgb[2]);
		break;
	case 0:
	case 3:
		setprg8(0x8000,prgb[0]);
		setprg8(0xA000,prgb[1]);
		setprg8(0xC000,prgb[2]);
		setprg8(0xE000,prgb[3]);
	}
}

static void tekvrom(void)
{
	int		x = 0;

	switch(JYCcntl[0]&0x18)
	{
	case 0x00: // 8KB
		setchr8(chrlow[0]|(chrhigh[0]<<8));
		break;
	case 0x08: // 4KB
		setchr4(0<<10,chrlow[0]|(chrhigh[0]<<8));
		setchr4(4<<10,chrlow[4]|(chrhigh[4]<<8));
		break;
	case 0x10: // 2KB
		do
			setchr2(x<<10,chrlow[x]|(chrhigh[x]<<8));
		while((x+=2)<8);
		break;
	case 0x18: // 1KB
		do
			setchr1(x<<10,chrlow[x]|(chrhigh[x]<<8));
		while(++x<8);
		break;
	}
}

static DECLFW(JYCwrite5)
{
	if ((A&=3)==0) //0x5800
		JYCmul[0] = V;
	else if (--A==0) //0x5801
		JYCmul[1] = V;
	else if ((A-=2)==0) //0x5803
		JYCregie = V;
}

static DECLFW(JYCwrite8)
{
	prgb[A&3] = V;
	tekprom();
}

static DECLFW(JYCwrite9A)
{
	if (A&0x1000) //9
		chrlow[A&7] = V;
	else //A
		chrhigh[A&7] = V;
	tekvrom();
}

static DECLFW(JYCwriteB)
{
	if (A&4)
		JYCnames[A&=3]&= 0x00FF, JYCnames[A]|= V<<8;
	else
		JYCnames[A&=3]&= 0xFF00, JYCnames[A]|= V;
	mira();
}

static DECLFW(JYCwriteC)
{
	switch(A&7)
	{
	case 0: //note that there is an IRQ toggle
		if ((IRQa=V&1)==0)
			X6502_IRQEnd(FCEU_IQEXT);
		break;
	case 1: //oh, man the real thing is complicated; scale adjustment...
		IRQMode = V;
		break;
	case 2:
		IRQa = 0;
		X6502_IRQEnd(FCEU_IQEXT);
		break;
	case 3:
		IRQa = 1;
		break;
	case 4: IRQPre = V^IRQXOR; break;
	case 5: IRQCount = V^IRQXOR; break;
	case 6: IRQXOR = V; break;
	}
}

static DECLFW(JYCwriteD)
{
	JYCcntl[A&=3] = V;
	tekprom();
	tekvrom();
	mira();
}

static void CCL(void)
{
	int		ct = IRQMode>>6;

	if (ct==1) // Count Up
	{
		if (++IRQCount==0 && IRQa)
			X6502_IRQBegin(FCEU_IQEXT);
	} else if (ct==2) // Count down
		if (--IRQCount==0xFF && IRQa)
			X6502_IRQBegin(FCEU_IQEXT);
}

static void ClockCounter(void)
{
	int		ct = IRQMode>>6;
	uint8		premask;

	if (IRQMode&4)
		premask = 7;
	else
		premask = 0xFF;

	if (ct==1) // Count up
	{
		if ((++IRQPre&premask)==0)
			CCL();
	} else if (ct==2) // Count down
		if ((--IRQPre&premask)==premask)
			CCL();
}

static void SLWrap(void)
{
	int		x = 8;

	do
		ClockCounter();	// 8 PPU A10 0->1 transitions per scanline(usually)
	while(--x!=0);
}

/* ---- mapper 90: J.Y.Company type A */

static void JYCReset()
{
	JYCtekker =~JYCtekker;
}

static void JYCRestore(int v)
{
(void)v;
	mira();
}

static void JYCPower(void)
{
	SetWriteHandler(0x5800,0x5FFF,JYCwrite5);
	SetWriteHandler(0x8000,0x8FFF,JYCwrite8);
	SetWriteHandler(0x9000,0xAFFF,JYCwrite9A);
	SetWriteHandler(0xB000,0xBFFF,JYCwriteB);
	SetWriteHandler(0xC000,0xCFFF,JYCwriteC);
	SetWriteHandler(0xD000,0xDFFF,JYCwriteD);
	SetReadHandler(0x5000,0x5FFF,tekread);
	SetCartRead(0x6000,0xFFFF);

	JYCmul[0] = JYCmul[1] = JYCregie = 0;//0xFF;
	JYCcntl[0] = JYCcntl[1] = JYCcntl[2] = JYCcntl[3] = 0;//0xFF;

	dwmemset(prgb,-1,sizeof(prgb));
	dwmemset(chrlow,-1,sizeof(chrlow));
	dwmemset(chrhigh,-1,sizeof(chrhigh));
	dwmemset(JYCnames,0,sizeof(JYCnames));

	JYCtekker = 0;
	prgb[0] = prgb[1] = prgb[2] = prgb[3] = 0xFF;
	tekprom();
	tekvrom();
}

void JYCA_Init(CartInfo *info)
{
	is209 = 0;
	info->Power = JYCPower;
	info->Reset = JYCReset;
	GameHBIRQHook2 = SLWrap;
	GameStateRestore = JYCRestore;
	AddExState(Tek_StateRegs,~0,0,0);
}

/* ---- mapper 209: J.Y.Company type B */

void JYCB_Init(CartInfo *info)
{
	is209 = 1;
	info->Power = JYCPower;
	info->Reset = JYCReset;
	GameHBIRQHook2 = SLWrap;
	GameStateRestore = JYCRestore;
	AddExState(Tek_StateRegs,~0,0,0);
}
