/*
 board/MMC1.c
  covers mappers 1, 155

 * MacFCEU 0.8a8
 *
 * Copyright notice for this file:
 *  mmc1.c Copyright (C) 1998 BERO
 *  mmc1.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC1.h"
#include "../cheat.h"

uint8	MMC1DReg[4];

static uint8	Buffer, BufferShift;

static uint8	*WRAM = 0;
static uint8	is155, MMC1opts;

void (*MMC1CHRHook4)(uint32 A,
	uint8 V);
void (*MMC1PRGHook16)(uint32 A,
	uint8 V);

static DECLFW(MBWRAM)
{
	if ((MMC1DReg[3]&0x10)==0 || is155)
		Page[A>>11][A] = V;	// WRAM is enabled.
}

static DECLFR(MAWRAM)
{
	if (MMC1DReg[3]&0x10 && !is155)
		return A>>8;	//XFrame.DB -- WRAM is disabled
	return Page[A>>11][A];
}

void MMC1CHR()
{
	if (MMC1opts&4)
		if (MMC1DReg[0]&0x10)
			setprg8r(0x10,0x6000,(MMC1DReg[1]>>4)&1);
		else
			setprg8r(0x10,0x6000,(MMC1DReg[1]>>3)&1);
	if (MMC1DReg[0]&0x10)
	{
		MMC1CHRHook4(0x0000,MMC1DReg[1]);
		MMC1CHRHook4(0x1000,MMC1DReg[2]);
	} else
	{
		MMC1CHRHook4(0x0000,MMC1DReg[1]&0xFE);
		MMC1CHRHook4(0x1000,MMC1DReg[1]|1);
	}
}

void MMC1PRG()
{
	uint8		offs = MMC1DReg[1]&0x10;

	switch(MMC1DReg[0]&0xC)
	{
	case 0x0:
	case 0x4:
		MMC1PRGHook16(0x8000,offs+=MMC1DReg[3]&~1);
		MMC1PRGHook16(0xc000,++offs);
		break;
	case 0x8:
		MMC1PRGHook16(0xC000,MMC1DReg[3]+offs);
		MMC1PRGHook16(0x8000,offs);
		break;
	case 0xC:
		MMC1PRGHook16(0x8000,MMC1DReg[3]+offs);
		MMC1PRGHook16(0xC000,0xF+offs);
	}
}

void MMC1MIRROR()
{
	int	x = ~MMC1DReg[0]&3; //01VH -> HV10

	if (x&2)
		x^= 1;
	setmirror(x);
}

static uint64 lreset;

static DECLFW(MMC1_write)
{
	uint64	t, u; /* for 64 bit backward-compatibility */

	int	n = (A>>13)-4;

	/* The MMC1 is busy so ignore the write. */
	/* As of version FCE Ultra 0.81, the timestamp is only
	   increased before each instruction is executed(in other words
	   precision isn't that great), but this should still work to
	   deal with 2 writes in a row from a single RMW instruction.
	*/
	//if(lreset+2>timestampbase+timestamp) return;

	t = timestampbase; (void)uint64addu32(t,timestamp);
	u = lreset; (void)uint64addu32(u,2);
	if (int64gt(u,t))
		return;
	if (V&0x80)
	{
		MMC1DReg[0]|= 0xC;
		BufferShift = Buffer = 0;
		MMC1PRG();
		lreset = t;
		return;
	}

	Buffer|= (V&1)<<BufferShift++;

	if (BufferShift!=5)
		return;
	MMC1DReg[n] = Buffer;
	BufferShift = Buffer = 0;

	//switch(n)
	if (n==0)
		MMC1MIRROR();
	if (n!=3)
		MMC1CHR();
	if (n!=2)
		MMC1PRG();
}

static void MMC1_Restore(int v)
{
(void)v;
	MMC1MIRROR();
	MMC1CHR();
	MMC1PRG();
	_int64xset(lreset,0);	/* timestamp(base) is not stored in save states. */
}

static void MMC1CMReset(void)
{
	Buffer = BufferShift = 0;
	MMC1DReg[0] = 0x1F;
	MMC1DReg[1]=0;
	MMC1DReg[2]=0;		// Should this be something other than 0?
	MMC1DReg[3]=0;

	MMC1MIRROR();
	MMC1CHR();
	MMC1PRG();
}

static int DetectMMC1WRAMSize(uint32 crc32)
{
	switch(crc32)
	{
	case 0xc6182024: /* Romance of the 3 Kingdoms */
	case 0x2225c20f: /* Genghis Khan */
	case 0x4642dda6: /* Nobunaga's Ambition */
	case 0x29449ba9: /* ""	"" (J) */
	case 0x2b11e0b0: /* ""	"" (J) */
		FCEU_printf(
" >8KB external WRAM present.  Use UNIF if you hack the ROM image.\n");
		return 16;

	default: return 8;
	}
}

void GenMMC1Power()
{
	_int64xset(lreset,0);
	if (MMC1opts&1)
	{
		FCEU_CheatSetRAM(8,0x6000,WRAM);
		if (MMC1opts&4)
			dwmemset(WRAM,0,8192);
		else if (!(MMC1opts&2))
			dwmemset(WRAM,0,8192);
	}

	SetWriteHandler(0x8000,0xFFFF,MMC1_write);
	SetCartRead(0x8000,0xFFFF);

	if (MMC1opts&1)
	{
		SetReadHandler(0x6000,0x7FFF,MAWRAM);
		SetWriteHandler(0x6000,0x7FFF,MBWRAM);
		setprg8r(0x10,0x6000,0);
	}

	MMC1CMReset();
}

void GenMMC1Close()
{
	DeleteCHRRAM(0);

	if (WRAM)
		FCEU_free(WRAM);
	WRAM = 0;
}

static void defprg16(uint32 A, uint8 V)
{
	setprg16(A,V);
}

static void defchr4(uint32 A, uint8 V)
{
	setchr4(A,V);
}

void GenMMC1Init(CartInfo *info,
	int prg,
	int chr,
	int wram,
	int battery)
{
	is155 = 0;

	info->Close = GenMMC1Close;
	MMC1PRGHook16 = defprg16;
	MMC1CHRHook4 = defchr4;
	MMC1opts = 0;
	PRGmask16[0]&= (prg>>14)-1;
	CHRmask4[0]&= (chr>>12)-1;
	CHRmask8[0]&= (chr>>13)-1;

	info->Power = GenMMC1Power;

	if (wram)
	{
		WRAM = (uint8*)FCEU_gmalloc(wram*1024);
		MMC1opts|= 1;
		if (wram>8)
			MMC1opts|= 4;
		SetupCartPRGMapping(0x10,WRAM,wram*1024,1);
		AddExState(WRAM,wram*1024,0,"WRAM");

		if (battery)
		{
			MMC1opts|= 2;
			info->SaveGame[0] = WRAM+(MMC1opts&4 ? 8192:0);
			info->SaveGameLen[0] = 8192;
		}
	}

	if (!chr)
		if (NewCHRRAM(0,8192))
			AddExState(CHRptr[0],CHRsize[0],0,"CHRR");

	GameStateRestore = MMC1_Restore;
	AddExState(MMC1DReg,4,0,"DREG");
	AddExState(&lreset,8,1,"LRST");
}

/* ---- mapper 1 */

void SxROM_Init(CartInfo *info)
{
	int	ws = DetectMMC1WRAMSize(info->CRC32);

	GenMMC1Init(info,512,256,ws,info->battery);
}

/* ---- mapper 155: MMC1A */
/* Same as mapper 1, without respect for WRAM enable bit. */

void MMC1A_Init(CartInfo *info)
{
	GenMMC1Init(info,512,256,8,info->battery);
	is155 = 1;
}

/* ---- generic */

void SAROM_Init(CartInfo *info)
{
	GenMMC1Init(info,128,64,8,info->battery);
}

void SBROM_Init(CartInfo *info)
{
	GenMMC1Init(info,128,64,0,0);
}

void SCROM_Init(CartInfo *info)
{
	GenMMC1Init(info,128,128,0,0);
}

void SEROM_Init(CartInfo *info)
{
	GenMMC1Init(info,32,64,0,0);
}

/* P128/256, C<=64, W<=8 */
void SFROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,64,0,info->battery);
}

void SGROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,0,0,0);
}

void SHROM_Init(CartInfo *info)
{
	GenMMC1Init(info,32,128,0,0);
}

void SJROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,64,8,info->battery);
}

void SKROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,64,8,info->battery);
}

void SLROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,128,0,0);
}

void SL1ROM_Init(CartInfo *info)
{
	GenMMC1Init(info,128,128,0,0);
}

void SL2ROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,256,0,0);
}

void SNROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,0,8,info->battery);
}

void SOROM_Init(CartInfo *info)
{
	GenMMC1Init(info,256,0,16,info->battery);
}
