/*
 boards/MMC3and6.c
  MMC3 (there are actually 3 revisions), and MMC6(HKROM)

 * MacFCEU 0.8a8
 *
 * Copyright notice for this file:
 *  mmc3.c Copyright (C) 1998 BERO
 *  mmc3.c Copyright (C) 2003 Xodnizel
 *  Mapper 12 code Copyright (C) 2003 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC3and6.h"
#include "../cheat.h"

uint8		MMC3cmd, MMC3IRQReload, MMC3resetmode, MMC3A000B, MMC3A001B;
uint8		MMC3IRQCount, MMC3IRQLatch, MMC3IRQa;

uint8		MMC3DReg[8];
uint8		MMC3EXPREGS[8];	// For bootleg games, mostly.

void (*MMC3pwrap)(uint32 A,
	uint8 V);
void (*MMC3cwrap)(uint32 A,
	uint8 V);
void (*MMC3mwrap)(uint8 V);

static uint8	PPUCHRBus;
static uint8	TKSMIR[8];

uint8		*MMC3WRAM;
static uint32	WRAMsize;

static uint8	MMC3opts = 0; //bit #1: battery backup (don't zero out)

static SFORMAT MMC3_StateRegs[] =
{
	{MMC3DReg,8,"REGS"},
	{&MMC3resetmode,1,"RMOD"},
	{&MMC3cmd,1,"CMD"},
	{&MMC3A000B,1,"A000"},
	{&MMC3A001B,1,"A001"},
	{&MMC3IRQReload,1,"IRQR"},
	{&MMC3IRQCount,1,"IRQC"},
	{&MMC3IRQLatch,1,"IRQL"},
	{&MMC3IRQa,1,"IRQA"},
	{0}
};

void FixMMC3PRG(int V)
{
	if (V&0x40)
	{
		MMC3pwrap(0x8000,~1);
		MMC3pwrap(0xC000,MMC3DReg[6]);
	} else
	{
		MMC3pwrap(0x8000,MMC3DReg[6]);
		MMC3pwrap(0xC000,~1);
	}
	MMC3pwrap(0xA000,MMC3DReg[7]);
	MMC3pwrap(0xE000,~0);
}

void FixMMC3CHR(int V)
{
	int		cbase = V<<5&0x1000;

	MMC3cwrap(cbase,MMC3DReg[0]&~1);
	MMC3cwrap(cbase^0x400,MMC3DReg[0]|1);
	MMC3cwrap(cbase^0x800,MMC3DReg[1]&~1);
	MMC3cwrap(cbase^0xC00,MMC3DReg[1]|1);

	MMC3cwrap(cbase^0x1000,MMC3DReg[2]);
	MMC3cwrap(cbase^0x1400,MMC3DReg[3]);
	MMC3cwrap(cbase^0x1800,MMC3DReg[4]);
	MMC3cwrap(cbase^0x1C00,MMC3DReg[5]);
}

DECLFW(MMC3_CMDWrite)
{
	A&= 0xE001;
	if ((A-=0x8000)==0)
	{
		if ((V^MMC3cmd)&0x40) //(V&0x40!=MMC3cmd&0x40)
			FixMMC3PRG(V);
		if ((V^MMC3cmd)&0x80)
			FixMMC3CHR(V);
		MMC3cmd = V;
	} else if (!--A) //0x8001
	{
		int	cbase = (MMC3cmd&0x80)<<5, x = MMC3cmd&7;

		MMC3DReg[x] = V;
		switch(x)
		{
		case 0:
			MMC3cwrap(cbase^0x000,V&~1);
			MMC3cwrap(cbase^0x400,V|1);
			break;
		case 1:
			MMC3cwrap(cbase^0x800,V&~1);
			MMC3cwrap(cbase^0xC00,V|1);
			break;
		case 2:
			MMC3cwrap(cbase^0x1000,V);
			break;
		case 3:
			MMC3cwrap(cbase^0x1400,V);
			break;
		case 4:
			MMC3cwrap(cbase^0x1800,V);
			break;
		case 5:
			MMC3cwrap(cbase^0x1C00,V);
			break;
		case 6:
			if (MMC3cmd&0x40)
				MMC3pwrap(0xC000,V);
			else
				MMC3pwrap(0x8000,V);
			break;
		case 7:
			MMC3pwrap(0xA000,V);
		}
	} else if ((A-=0xA000-0x8001)==0)
	{
		if (MMC3mwrap)
			MMC3mwrap(V&1);
	} else if (!--A) //0xA001
		MMC3A001B = V; //WRAM security set
}

DECLFW(MMC3_IRQWrite)
{
	if (!(A&0xC000))
		return;
	if ((A&=0x2001)==0) //0xC000
		MMC3IRQLatch = V;
	else if (!--A) //0xC001
		MMC3IRQReload = 1;
	else if ((MMC3IRQa=A-=0xE000-0xC001)==0) //0xE00x
		X6502_IRQEnd(FCEU_IQEXT);
}

static void ClockMMC3Counter(void)
{
	int	count;

	if ((count=MMC3IRQCount)==0 || MMC3IRQReload)
		MMC3IRQCount = MMC3IRQLatch, MMC3IRQReload = 0;
	else
		--MMC3IRQCount;
	if (count && !MMC3IRQCount && MMC3IRQa)
		XFrame.intlow|= FCEU_IQEXT; //X6502_IRQBegin
}

static void MMC3_hb_KickMasterHack(void)
{
	if (scanline==238)
		ClockMMC3Counter();
	ClockMMC3Counter();
}

static void MMC3_hb_PALStarWarsHack(void)
{
	if (scanline==240)
		ClockMMC3Counter();
	ClockMMC3Counter();
}

/*
static void ClockMMC6Counter(void)
..
*/

void GenMMC3Restore(int v)
{
(void)v;
	MMC3mwrap(MMC3A000B&1);
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void GENPWRAP(uint32 A,
	uint8 V)
{
	setprg8(A,V&0x3F);
}

static void GENMWRAP(uint8 V)
{
	MMC3A000B = V, V^= 1;
	setmirror(V);
}

static void GENNOMWRAP(uint8 V)
{
	MMC3A000B = V;
}

static DECLFW(MBWRAM)
{
	if (MMC3WRAM)
		MMC3WRAM[A-0x6000] = V;
}

static DECLFR(MAWRAM)
{
	return MMC3WRAM[A-0x6000];
}

static DECLFW(MBWRAMMMC6)
{
	if (MMC3WRAM)
		MMC3WRAM[A&0x3FF] = V;
}

static DECLFR(MAWRAMMMC6)
{
	return MMC3WRAM[A&0x3FF];
}

/* ---- general */

void GenMMC3Reset()
{
	MMC3IRQCount = MMC3IRQLatch = MMC3IRQa = MMC3cmd = 0;

	MMC3DReg[0] = MMC3DReg[6] = 0;
	MMC3DReg[1] = 2;
	MMC3DReg[2] = 4;
	MMC3DReg[3] = 5;
	MMC3DReg[4] = 6;
	MMC3DReg[5] = 7;
	MMC3DReg[7] = 1;

	FixMMC3PRG(0);
	FixMMC3CHR(0);
}

void GenMMC3Power()
{
	MMC3A001B = MMC3A000B = 0;
	setmirror(~MMC3A000B&1);

	SetWriteHandler(0x8000,0xBFFF,MMC3_CMDWrite);
	SetWriteHandler(0xC000,0xFFFF,MMC3_IRQWrite);
	SetCartRead(0x8000,0xFFFF);

	if (WRAMsize)
	{
		if (WRAMsize==1024)
		{
			FCEU_CheatSetRAM(1,0x7000,MMC3WRAM);
			SetReadHandler(0x7000,0x7FFF,MAWRAMMMC6);
			SetWriteHandler(0x7000,0x7FFF,MBWRAMMMC6);
		} else
		{
			FCEU_CheatSetRAM(WRAMsize>>10,0x6000,MMC3WRAM);
			SetReadHandler(0x6000,0x6000-1+WRAMsize,MAWRAM);
			SetWriteHandler(0x6000,0x6000-1+WRAMsize,MBWRAM);
		}
		if (!(MMC3opts&2))
			dwmemset(MMC3WRAM,0,WRAMsize);
	}

	GenMMC3Reset();

	EraseCHRRAM(0,0);
	EraseCHRRAM(0x10,0);
}

void GenMMC3Close()
{
	if (MMC3WRAM)
		{FCEU_free(MMC3WRAM); MMC3WRAM = 0;}
	DeleteCHRRAM(0);
	DeleteCHRRAM(0x10);
	MMC3EXPREGS[5] = 0; /* for debug.c */
	MMC3opts = 0; /* re-initialize */
}

static void SetIRQHook(uint32 crc)
{
	void (*hook)(void);

	switch(crc)
	{
	case 0x5104833E: // Kick Master
		hook = MMC3_hb_KickMasterHack;
		break;
	case 0x5A6860F1:
	case 0xAE280E20: // Shougi Meikan '92/'93
		hook = MMC3_hb_KickMasterHack;
		break;
	case 0xFCD772EB: // PAL Star Wars, similar problem as Kick Master.
		hook = MMC3_hb_PALStarWarsHack;
		break;
	default: hook = ClockMMC3Counter;
	}
	GameHBIRQHook = hook;
}

void GenMMC3_Init(CartInfo *info,
	int prg,
	int chr,
	int wram,
	int battery)
{
	MMC3pwrap = GENPWRAP;
	MMC3cwrap = setchr1;
	MMC3mwrap = GENMWRAP;

	PRGmask8[0]&= (prg>>13)-1;
	CHRmask1[0]&= (chr>>10)-1;
	CHRmask2[0]&= (chr>>11)-1;

	if ((WRAMsize=wram*1024)!=0)
	{
		MMC3WRAM = (uint8*)FCEU_gmalloc(WRAMsize);
		AddExState(MMC3WRAM,WRAMsize,0,"WRAM");
	}
	if (battery)
	{
		MMC3opts|= 2;
		info->SaveGame[0] = MMC3WRAM;
		info->SaveGameLen[0] = WRAMsize;
	}
	if (!chr)
		if (NewCHRRAM(0,8192))
			AddExState(CHRptr[0],CHRsize[0],0,"CHRR");
	info->Power = GenMMC3Power;
	info->Reset = GenMMC3Reset;
	info->Close = GenMMC3Close;
	SetIRQHook(info->CRC32);
	GameStateRestore = GenMMC3Restore;
	AddExState(MMC3_StateRegs,~0,0,0);
}

/* ---- mapper 12: Rex Soft DBZ5 */

static void M12CW(uint32 A,
	uint8 V)
{
	setchr1(A,(MMC3EXPREGS[A>>12&1]<<8)+V);
}

static DECLFW(M12Write)
{
(void)A;
	MMC3EXPREGS[0] = V&1;
	MMC3EXPREGS[1] = V>>4&1;
}

static void M12Power(void)
{
	MMC3EXPREGS[0] = MMC3EXPREGS[1] = 0;
	GenMMC3Power();
	SetWriteHandler(0x4100,0x5FFF,M12Write);
}

void RexSoftDBZ5_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3cwrap = M12CW;
	info->Power = M12Power;
	AddExState(MMC3EXPREGS,2,0,"EXPR");
}

/* ---- mapper 37: ZZ */

static void M37PW(uint32 A,
	uint8 V)
{
	V&= (MMC3EXPREGS[0]<<1)|7;
	V|= MMC3EXPREGS[0]<<2&0x10;
	if (!(uint8)(~MMC3EXPREGS[0]&3))
		V|= 8;
	setprg8(A,V);
}

static void M37CW(uint32 A,
	uint8 V)
{
	uint32		NV = V&0x7F;

	NV|= MMC3EXPREGS[0]<<5&0x80;
	setchr1(A,NV);
}

static DECLFW(M37_write)
{
(void)A;
	V&= 7;
	if (MMC3EXPREGS[0]==V)
		return;
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M37Reset(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Reset();
}

static void M37Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M37_write);
}

void CustomPAL_ZZ_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3pwrap = M37PW;
	MMC3cwrap = M37CW;
	info->Reset = M37Reset;
	info->Power = M37Power;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 47: QJ */

static void M47PW(uint32 A,
	uint8 V)
{
	V&= 0xF;
	V|= MMC3EXPREGS[0]<<4;
	setprg8(A,V);
}

static void M47CW(uint32 A,
	uint8 V)
{
	uint32	NV = V;

	NV&= 0x7F;
	NV|= MMC3EXPREGS[0]<<7;
	setchr1(A,NV);
}

static DECLFW(M47Write)
{
(void)A;
	MMC3EXPREGS[0] = V&1;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M47Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M47Write);
	SetReadHandler(0x6000,0x7FFF,0);
}

void QJ_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3pwrap = M47PW;
	MMC3cwrap = M47CW;
	info->Power = M47Power;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 115: Kasing (standard) */

static void M115PW(uint32 A,
	uint8 V)
{
	setprg8(A,V);
	if (MMC3EXPREGS[0]&0x80)
		setprg16(0x8000,MMC3EXPREGS[0]&7);
}

static void M115CW(uint32 A,
	uint8 V)
{
	setchr1(A,(uint32)V|((MMC3EXPREGS[1]&1)<<8));
}

static DECLFW(M115Write)
{
	if (A==0x6000)
		MMC3EXPREGS[0] = V;
	else if (A==0x6001)
		MMC3EXPREGS[1] = V;
	FixMMC3PRG(MMC3cmd);
}

static void M115Power(void)
{
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M115Write);
	SetReadHandler(0x6000,0x7FFF,0);
}

void Kasing_STD_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,0,0);
	MMC3cwrap = M115CW;
	MMC3pwrap = M115PW;
	info->Power = M115Power;
	AddExState(MMC3EXPREGS,2,0,"EXPR");
}

/* ---- TxxROM boards */

static void FP_FASTAPASS(1) TKSPPU(uint32 A)
{
	A&= 0x1FFF;
	A>>= 10;
	PPUCHRBus = A;
	setmirror(MI_0+TKSMIR[A]);
}

static void TKSWRAP(uint32 A,
	uint8 V)
{
	TKSMIR[A>>10] = V>>7;
	setchr1(A,V&0x7F);
	if (PPUCHRBus==A>>10)
		setmirror(MI_0+(V>>7));
}

static int	hackm4 = 0; /* For Karnov, maybe others.  BLAH.  Stupid iNES format.*/

static void M4Power(void)
{
	GenMMC3Power();
	MMC3A000B =~hackm4&1;
	setmirror(hackm4);
}

/* mapper 4 */
void TxROM_Init(CartInfo *info)
{
	int	ws = 8;

	if (info->CRC32==0x93991433 || info->CRC32==0xAF65AA84)
	{
		FCEU_printf(
"Low-G-Man can not work normally in the iNES format.\n"
"This game has been recognized by its CRC32 value,\n"
"and the appropriate changes will be made so it will run.\n"
//"If you wish to hack this game, you should use the UNIF format for your hack.\n"
"If you wish to modify this game, you should use the UNIF format.\n"
"\n");
		ws = 0;
	}

	GenMMC3_Init(info,512,256,ws,info->battery);
	info->Power = M4Power;
	hackm4 = info->mirror;
}

void TEROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,32,32,0,0);
}

void TFROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,64,0,0);
}

void TGROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,0,0,0);
}

void TKROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
}

void TLROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,0,0);
}

void TSROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,0);
}

void TLSROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,0);
	MMC3cwrap = TKSWRAP;
	MMC3mwrap = GENNOMWRAP;
	PPU_hook = TKSPPU;
	AddExState(&PPUCHRBus,1,0,"PPUC");
}

/* mapper 118 */
void TKSROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3cwrap = TKSWRAP;
	MMC3mwrap = GENNOMWRAP;
	PPU_hook = TKSPPU;
	AddExState(&PPUCHRBus,1,0,"PPUC");
}

/* mapper 119 */
static void TQWRAP(uint32 A,
	uint8 V)
{
	setchr1r(V>>2&0x10,A,V&0x3F);
}

void TQROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,64,0,0);
	MMC3cwrap = TQWRAP;
	if (NewCHRRAM(0x10,8192))
		AddExState(CHRptr[0x10],CHRsize[0x10],0,"CHRR");
}

/* MMC6 board */
void HKROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,512,1,info->battery);
}
