/*
 boards/MMC5.c

 MacFCEU 0.8.5

 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* None of this code should use any of the iNES bank switching wrappers. */

#include "mapinc.h"
#include "MMC5.h"
#include "../cheat.h"

#include <stdlib.h>

/* according to Nestopia, this it's bit #0 */
#define RAW_DISABLE		0x41

static void (*sfun)(int P);
static void (*psfun)(void);

void MMC5RunSound(int32 Count);
#if HAVE_HQ
void MMC5RunSoundHQ(void);
#endif
//static INLINE void MMC5SPRVROM_BANK1(uint32 A,uint32 V)
#define MMC5SPRVROM_BANK1(A,V) do{uint32 _V=V; \
	{if(CHRptr[0]) {_V=(V)&CHRmask1[0];MMC5SPRVPage[(A)>>10]=&CHRptr[0][_V<<10]-(A);}}}while(0)
static INLINE void MMC5BGVROM_BANK1(uint32 A,uint32 V)
	{if(CHRptr[0]) {V&=CHRmask1[0];MMC5BGVPage[(A)>>10]=&CHRptr[0][(V)<<10]-(A);}}

static INLINE void MMC5SPRVROM_BANK2(uint32 A,uint32 V)
	{if(CHRptr[0]) {V&=CHRmask2[0];MMC5SPRVPage[(A)>>10]=MMC5SPRVPage[((A)>>10)+1]=&CHRptr[0][(V)<<11]-(A);}}
static INLINE void MMC5BGVROM_BANK2(uint32 A,uint32 V)
	{if(CHRptr[0]) {V&=CHRmask2[0];MMC5BGVPage[(A)>>10]=MMC5BGVPage[((A)>>10)+1]=&CHRptr[0][(V)<<11]-(A);}}

static INLINE void MMC5SPRVROM_BANK4(uint32 A,uint32 V)
	{if(CHRptr[0]) {V&=CHRmask4[0];MMC5SPRVPage[(A)>>10]=MMC5SPRVPage[((A)>>10)+1]= MMC5SPRVPage[((A)>>10)+2]=MMC5SPRVPage[((A)>>10)+3]=&CHRptr[0][(V)<<12]-(A);}}
static INLINE void MMC5BGVROM_BANK4(uint32 A,uint32 V)
	{if(CHRptr[0]) {V&=CHRmask4[0];MMC5BGVPage[(A)>>10]=MMC5BGVPage[((A)>>10)+1]=MMC5BGVPage[((A)>>10)+2]=MMC5BGVPage[((A)>>10)+3]=&CHRptr[0][(V)<<12]-(A);}}

static INLINE void MMC5SPRVROM_BANK8(uint32 V)
	{if(CHRptr[0]) {V&=CHRmask8[0];MMC5SPRVPage[0]=MMC5SPRVPage[1]=MMC5SPRVPage[2]=MMC5SPRVPage[3]=MMC5SPRVPage[4]=MMC5SPRVPage[5]=MMC5SPRVPage[6]=MMC5SPRVPage[7]=&CHRptr[0][(V)<<13];}}
static INLINE void MMC5BGVROM_BANK8(uint32 V)
	{if(CHRptr[0]) {V&=CHRmask8[0];MMC5BGVPage[0]=MMC5BGVPage[1]=MMC5BGVPage[2]=MMC5BGVPage[3]=MMC5BGVPage[4]=MMC5BGVPage[5]=MMC5BGVPage[6]=MMC5BGVPage[7]=&CHRptr[0][(V)<<13];}}

static uint8 PRGBanks[4];
static uint8 WRAMPage;
static uint8 WRAMMaskEnable[2];
static uint16 CHRBanks[8+4]; /* was uint8 */
#define CHRBanksA CHRBanks
#define CHRBanksB (CHRBanks+8)

/* for backward-compatible state files */
static uint8 CHRBanksL[8+4];
static uint8 CHRBanksH[8+4]; /* independent? */

uint8	MMC5ABMode;                /* A=0, B=1 */

static uint8 IRQScanline, IRQEnable;
static uint8 NTAMirroring, NTFill, ATFill;

			/* FCEUX name */
uint8 MMC5Ex;		/* MMC5Hack */
uint8 MMC5VROMMask;	/* MMC5VROMMask */
uint8 *MMC5ExRAM;	/* MMC5HackExNTARAMPtr */
uint8 *MMC5VROMPtr;	/* MMC5HackVROMPTR */
uint8 MMC5ExMode;	/* MMC5HackCHRMode */
uint8 MMC5SplitCtrl;	/* MMC5HackSPMode */
uint8 MMC5CHRHigh;	/* MMC50x5130 */
uint8 MMC5SplitY;	/* MMC5HackSPScroll */
uint8 MMC5SplitBank;	/* MMC5HackSPPage */

static uint8 MMC5IRQR;
static uint8 MMC5LineCounter;
static uint8 mmc5psize, mmc5vsize;
static uint8 mul[2];

static uint8 *WRAM;
static uint8 *MMC5fill;

static uint8 MMC5WRAMsize;
static uint8 MMC5WRAMIndex[8];

static uint8 MMC5ROMWrProtect[4];
static uint8 MMC5MemIn[5];

static void MMC5CHRA(void);
static void MMC5CHRB(void);

typedef struct __cartdata
{
	uint32		crc32;
	uint8		size;
} cartdata;


// ETROM seems to have 16KB of WRAM, ELROM seems to have 8KB
// EWROM seems to have 32KB of WRAM

#define MMC5_NOCARTS 14
cartdata MMC5CartList[] =
{
	{0x9c18762b,2},         /* L'Empereur */
	{0x26533405,2},
	{0x6396b988,2},

	{0xaca15643,2},        /* Uncharted Waters */
	{0xfe3488d1,2},        /* Dai Koukai Jidai */

	{0x15fe6d0f,2},        /* BKAC             */
	{0x39f2ce4b,2},        /* Suikoden              */

	{0x8ce478db,2},        /* Nobunaga's Ambition 2 */
	{0xeee9a682,2},

	{0x1ced086f,2},        /* Ishin no Arashi */

	{0xf540677b,4},        /* Nobunaga...Bushou Fuuun Roku */

	{0x6f4e4312,4},        /* Aoki Ookami..Genchou */

	{0xf011e490,4},        /* Romance of the 3 Kingdoms 2 */
	{0x184c2124,4},        /* Sangokushi 2 */
};


static int DetectMMC5WRAMSize(uint32 crc32)
{
	int		x = 0;

	do
		if (crc32==MMC5CartList[x].crc32)
		{
			FCEU_printf(" >8KB external WRAM present.  Use UNIF if you hack the ROM image.\n");
			return MMC5CartList[x].size*8;
		}
	while(++x<GET_ARRAY_COUNT(MMC5CartList));
	return 8;
}

static void BuildWRAMSizeTable(void)
{
	int		x = 0;

	do
		switch(MMC5WRAMsize)
		{
		case 0: MMC5WRAMIndex[x] = 255; break;
		case 1: MMC5WRAMIndex[x] = x>3 ? 255:0; break;
		case 2: MMC5WRAMIndex[x] = (x&4)>>2; break;
		case 4: MMC5WRAMIndex[x] = x>3 ? 255:(x&3); break;
		//case 8: MMC5WRAMIndex[x] = x;
		}
	while(++x<8);
}

static void MMC5CHRA(void)
{
	uint32		x;

	switch(mmc5vsize&3)
	{
	case 0: setchr8(CHRBanksA[7]);
		MMC5SPRVROM_BANK8(CHRBanksA[7]);
		break;
	case 1: setchr4(0x0000,CHRBanksA[3]);
		setchr4(0x1000,CHRBanksA[7]);
		MMC5SPRVROM_BANK4(0x0000,CHRBanksA[3]);
		MMC5SPRVROM_BANK4(0x1000,CHRBanksA[7]);
		break;
	case 2: setchr2(0x0000,CHRBanksA[1]);
		setchr2(0x0800,CHRBanksA[3]);
		setchr2(0x1000,CHRBanksA[5]);
		setchr2(0x1800,CHRBanksA[7]);
		MMC5SPRVROM_BANK2(0x0000,CHRBanksA[1]);
		MMC5SPRVROM_BANK2(0x0800,CHRBanksA[3]);
		MMC5SPRVROM_BANK2(0x1000,CHRBanksA[5]);
		MMC5SPRVROM_BANK2(0x1800,CHRBanksA[7]);
		break;
	case 3: x = 0;
		do
		{
			setchr1(x<<10,CHRBanksA[x]);
			MMC5SPRVROM_BANK1(x<<10,CHRBanksA[x]);
		} while(++x<8);
	}
}

static void MMC5CHRB(void)
{
	uint32		x;

	switch(mmc5vsize&3)
	{
	case 0: setchr8(CHRBanksB[3]);
		MMC5BGVROM_BANK8(CHRBanksB[3]);
		break;
	case 1: setchr4(0x0000,CHRBanksB[3]);
		setchr4(0x1000,CHRBanksB[3]);
		MMC5BGVROM_BANK4(0x0000,CHRBanksB[3]);
		MMC5BGVROM_BANK4(0x1000,CHRBanksB[3]);
		break;
	case 2: setchr2(0x0000,CHRBanksB[1]);
		setchr2(0x0800,CHRBanksB[3]);
		setchr2(0x1000,CHRBanksB[1]);
		setchr2(0x1800,CHRBanksB[3]);
		MMC5BGVROM_BANK2(0x0000,CHRBanksB[1]);
		MMC5BGVROM_BANK2(0x0800,CHRBanksB[3]);
		MMC5BGVROM_BANK2(0x1000,CHRBanksB[1]);
		MMC5BGVROM_BANK2(0x1800,CHRBanksB[3]);
		break;
	case 3:	x = 0;
		do
		{
			setchr1(x<<10,CHRBanksB[x&3]);
			MMC5BGVROM_BANK1(x<<10,CHRBanksB[x&3]);
		} while(++x<8);
	}
}

static void FASTAPASS(2) MMC5WRAM(uint32 A,
	uint32 V)
{
	//printf("%02x\n",V);
	V = MMC5WRAMIndex[V&7];
	if (V!=255)
	{
		setprg8r(0x10,A,V);
		MMC5MemIn[(A-0x6000)>>13] = 1;
	} else
		MMC5MemIn[(A-0x6000)>>13] = 0;
}

static void MMC5PRG(void)
{
	int		x;

	if ((x=mmc5psize&3)==0)
	{
		MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] =
			MMC5ROMWrProtect[2] = MMC5ROMWrProtect[3] = 1;
		setprg32(0x8000,(PRGBanks[1]&0x7F)>>2);
		x = 1;
		do
			MMC5MemIn[x] = 1;
		while(++x<5);
	} else if (!--x)
	{
		if (PRGBanks[1]&0x80)
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 1;
			setprg16(0x8000,PRGBanks[1]>>1);
			MMC5MemIn[1] = MMC5MemIn[2] = 1;
		} else
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 0;
			MMC5WRAM(0x8000,PRGBanks[1]&(7&0xFE));
			MMC5WRAM(0xA000,(PRGBanks[1]&(7&0xFE))+1);
		}
		MMC5MemIn[3] = MMC5MemIn[4] = 1;
		MMC5ROMWrProtect[2] = MMC5ROMWrProtect[3] = 1;
		setprg16(0xC000,(PRGBanks[3]&0x7F)>>1);
	} else if (!--x)
	{
		if (PRGBanks[1]&0x80)
		{
			MMC5MemIn[1] = MMC5MemIn[2] = 1;
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 1;
			setprg16(0x8000,(PRGBanks[1]&0x7F)>>1);
		} else
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 0;
			MMC5WRAM(0x8000,PRGBanks[1]&(7&0xFE));
			MMC5WRAM(0xA000,(PRGBanks[1]&(7&0xFE))+1);
		}
		if (PRGBanks[2]&0x80)
		{
			MMC5ROMWrProtect[2] = 1, MMC5MemIn[3] = 1;
			setprg8(0xC000,PRGBanks[2]&0x7F);
		} else
		{
			MMC5ROMWrProtect[2] = 0;
			MMC5WRAM(0xC000,PRGBanks[2]&7);
		}
		MMC5MemIn[4] = 1, MMC5ROMWrProtect[3] = 1;
		setprg8(0xE000,PRGBanks[3]&0x7F);
	} else
	{
		x = 0;
		do
			if (PRGBanks[x]&0x80)
			{
				MMC5ROMWrProtect[x] = 1;
				setprg8(0x8000+(x<<13),PRGBanks[x]&0x7F);
				MMC5MemIn[1+x] = 1;
			} else
			{
				MMC5ROMWrProtect[x] = 0;
				MMC5WRAM(0x8000+(x<<13),PRGBanks[x]&7);
			}
		while(++x<3);
		MMC5MemIn[4] = 1;
		MMC5ROMWrProtect[3] = 1;
		setprg8(0xE000,PRGBanks[3]&0x7F);
	}
}

static DECLFW(MMC5W5100) //Mapper5_write
{
	uint32		t;
	int		x;

	switch(A)
	{
	case 0x5105:
		NTAMirroring = V;
		x = 0;
		do
			switch((V>>x+x)&3)
			{
			case 0: PPUNTARAM|= 1<<x,vnapage[x] = NTARAM; break;
			case 1: PPUNTARAM|= 1<<x,vnapage[x] = NTARAM+0x400; break;
			case 2: PPUNTARAM|= 1<<x,vnapage[x] = MMC5ExRAM; break;
			case 3: PPUNTARAM&=~(1<<x),vnapage[x] = MMC5fill; break;
			}
		while(++x<4);
		break;
	case 0x5113:
		WRAMPage = V;
		MMC5WRAM(0x6000,V&7);
		break;
	case 0x5100:
		mmc5psize = V;
		MMC5PRG();
		break;
	case 0x5101:
		mmc5vsize = V;
		if (!MMC5ABMode)
			{MMC5CHRB(); MMC5CHRA();}
		else
			{MMC5CHRA(); MMC5CHRB();}
		break;
	case 0x5114:
	case 0x5115:
	case 0x5116:
	case 0x5117:
		PRGBanks[A&3] = V;
		MMC5PRG();
		break;
	case 0x5120:
	case 0x5121:
	case 0x5122:
	case 0x5123:
	case 0x5124:
	case 0x5125:
	case 0x5126:
	case 0x5127:
		MMC5ABMode = 0;
		CHRBanksL[A-0x5120] = V;
		CHRBanksH[A-0x5120] = MMC5CHRHigh;
		CHRBanksA[A-0x5120] = V | MMC5CHRHigh<<8;
		MMC5CHRA();
		break;
	case 0x5128:
	case 0x5129:
	case 0x512A:
	case 0x512B: /* FCEUX doesn't do the high here (it should) */
		MMC5ABMode = 1;
		CHRBanksL[A-0x5128+8] = V;
		CHRBanksH[A-0x5128+8] = MMC5CHRHigh;
		CHRBanksB[A-0x5128] = V | MMC5CHRHigh<<8;
		MMC5CHRB();
		break;
	case 0x5102:
		WRAMMaskEnable[0] = V;
		break;
	case 0x5103:
		WRAMMaskEnable[1] = V;
		break;
	case 0x5104:
		MMC5ExMode = V&3;
		break;
	case 0x5106:
		if (V!=NTFill)
		{
			t = V|(V<<8), t|= V<<16;
			FCEU_dwmemset(MMC5fill,t,0x3c0);
		}
		NTFill = V;
		break;
	case 0x5107:
		if (V!=ATFill)
		{
			ATFill = V;
			V|= (V<<2)|(V<<4)|(V<<6);
			t = V|(V<<8), t|= t<<16;
			FCEU_dwmemset(MMC5fill+0x3C0,t,0x40);
		}
		break;
	case 0x5130:
		FCEUPPU_LineUpdate(); /* 2013.05.29 */
		MMC5CHRHigh = V&3;
	}
}

static DECLFW(MMC5W5200)
{
	switch(A)
	{
	case 0x5200: MMC5SplitCtrl = V; break;
	case 0x5201: if (V>=240) /* 2013.05.29 */
			V-= 16;
		MMC5SplitY = V>>3&0x1F;
		break;
	case 0x5202: MMC5SplitBank = V&0x3F; break;
	case 0x5203: X6502_IRQEnd(FCEU_IQEXT); IRQScanline = V; break;
	case 0x5204: X6502_IRQEnd(FCEU_IQEXT); IRQEnable = V&0x80; break;
	case 0x5205: mul[0] = V; break;
	case 0x5206: mul[1] = V;
	}
}

static DECLFR(MMC5_ReadROMRAM)
{
	if (MMC5MemIn[(A-0x6000)>>13])
		return Page[A>>11][A];
	else
		return A>>8;
}

static DECLFW(MMC5_WriteROMRAM)
{
	if (A>=0x8000 && MMC5ROMWrProtect[(A-0x8000)>>13])
		return;
	if (MMC5MemIn[(A-0x6000)>>13])
		if (((WRAMMaskEnable[0]&3)|((WRAMMaskEnable[1]&3)<<2))==6)
			Page[A>>11][A] = V;
}

static DECLFW(MMC5_ExRAMWr)
{
	if (MMC5ExMode!=3)
		MMC5ExRAM[A&0x3FF] = V;
}

static DECLFR(MMC5_ExRAMRd)
{
	if (MMC5ExRAM && MMC5ExMode&2)
		return MMC5ExRAM[A&0x3FF];
	return A>>8;
}

static DECLFR(MMC5R5204) //MMC5_read
{
	uint8		x = MMC5IRQR;
(void)A;
	if (!fceuindbg)
		{X6502_IRQEnd(FCEU_IQEXT); MMC5IRQR&= 0x40;}
	return x;
}

static DECLFR(MMC5Rmul)
{
	if (A&=1) /*0x5205*/
		return mul[0]*mul[1];
	else
		return mul[0]*mul[1]>>8;
}

static void MMC5Synco(void)
{
	MMC5PRG();
	MMC5W5100(0x5101,mmc5vsize); /*AB*/
	MMC5W5100(0x5105,NTAMirroring);
	MMC5W5100(0x5106,NTFill);
	MMC5W5100(0x5107,ATFill);
	MMC5WRAM(0x6000,WRAMPage&7);
	X6502_IRQEnd(FCEU_IQEXT);
	MMC5ExMode&= 3, MMC5CHRHigh&= 3;
}

void MMC5_hb(int scanline)
{
	if (scanline==240)
	{
		MMC5LineCounter = 0, MMC5IRQR = 0x40;
		return;
	}
	if (MMC5LineCounter<240)
	{
		if (MMC5LineCounter==IRQScanline)
		{
			MMC5IRQR|= 0x80;
			if (IRQEnable&0x80)
				X6502_IRQBegin(FCEU_IQEXT);
		}
		++MMC5LineCounter;
	}
	if (MMC5LineCounter==240)
		MMC5IRQR = 0;
}

static void MMC5_StateRestore(int v)
{
	if (v<9820)
		memset(CHRBanksH,0,sizeof(CHRBanksH));
	v = 0;
	do
		CHRBanksA[v] = CHRBanksL[v]|CHRBanksH[v]<<8;
	while(++v<8);
	v = 0;
	do
		CHRBanksB[v] = CHRBanksL[v+8]|CHRBanksH[v+8]<<8;
	while(++v<4);
	MMC5Synco();
}

typedef struct
{
	uint16		wl[2];
	uint8		env[2];
	uint8		enable;
	uint8		running;
	uint8		raw, rawcontrol;
	int32		dcount[2];
	int32		BC[3];
	int32		vcount[2];
} MMC5APU;

static MMC5APU	MMC5Sound;

static int32	volume = 256;

void FCEU_MMC5SetVolume(int16 v)
{
#if HAVE_HQ
	if (v<0 && FSettings.soundq)
		v =-v;
#endif
	volume = (v<<(8-2))/25;
}

static DECLFW(MMC5W5) //Mapper5_SW
{
	GameExpSound.Fill[SCMMC5] = MMC5RunSound;
#if HAVE_HQ
	GameExpSound.HiFill[SCMMC5] = MMC5RunSoundHQ;
#endif
	switch(A-=0x5000)
	{
	case 0:
	case 4://printf("%04x:$%02x\n",A,V&0x30);
		A>>= 2;
		if (sfun)
			sfun(A);
		MMC5Sound.env[A] = V;
		break;
	case 2:
	case 6:	A>>= 2;
		if (sfun)
			sfun(A);
		MMC5Sound.wl[A]&=~0x00FF;
		MMC5Sound.wl[A]|= V&0xFF;
		break;
	case 3:
	case 7:	A>>= 2;
		MMC5Sound.wl[A]&=~0x700;
		MMC5Sound.wl[A]|= V<<8&0x700;
		MMC5Sound.running|= 1u<<A;
		break;
	case 0x10:
		if (psfun)
			psfun();
		MMC5Sound.rawcontrol = V;
		break;
	case 0x11:
		if (psfun)
			psfun();
		MMC5Sound.raw = V;
		break;
	case 0x15:
		if (sfun)
			{sfun(0); sfun(1);}
		MMC5Sound.running&= V;
		MMC5Sound.enable = V;
		//printf("%02x\n",V);
	}
}

static int	tal[4] = {1,2,4,6};

static void Do5SQ(int P)
{
	int32		V, start, end, wl;

	start = MMC5Sound.BC[P], end = (SOUNDTS<<16)/soundtsinc;
	if (end<=start)
		return;
	MMC5Sound.BC[P] = end;

	wl = MMC5Sound.wl[P]+1;
	if (wl>=8 && MMC5Sound.running&(P+1))
	{
		int32		amp, rthresh;
		int		dc,vc;

		wl<<= 18;
		amp = (MMC5Sound.env[P]&15)*volume;
		rthresh = tal[MMC5Sound.env[P]>>6]; //&0xC0)>>6

		dc = MMC5Sound.dcount[P];
		vc = MMC5Sound.vcount[P];

		for(V=start; V<end; ++V)
		{
			if (dc<rthresh)
				FCEUWave[V>>4]+= amp;
			vc-= nesincsize;
			while(vc<=0)
				vc+= wl, dc = (dc+1)&7;
		}
		MMC5Sound.dcount[P] = dc;
		MMC5Sound.vcount[P] = vc;
	}
}

static void Do5PCM()
{
	int32		V, start, end, amp;

	start = MMC5Sound.BC[2], end = (SOUNDTS<<16)/soundtsinc;
	if (end<=start)
		return;
	MMC5Sound.BC[2] = end, V = start;

	if (!(MMC5Sound.rawcontrol&RAW_DISABLE) && (amp=MMC5Sound.raw*volume>>3)!=0)
		do
			FCEUWave[V>>4]+= amp;
		while(++V<end);
}

#if HAVE_HQ
static void Do5SQHQ(int P)
{
	int32		V, wl;

	wl = MMC5Sound.wl[P]+1;

	if (wl>=8 && (MMC5Sound.running&(P+1)))
	{
		int32		end = SOUNDTS;
		int		amp, rthresh, dc, vc;

		wl+= wl;
		amp = (MMC5Sound.env[P]&0xF)*volume;
		rthresh = tal[MMC5Sound.env[P]>>6]; //&0xC0)>>6
		dc = MMC5Sound.dcount[P];
		vc = MMC5Sound.vcount[P];
		end = SOUNDTS;
		for(V=MMC5Sound.BC[P]; V<end; ++V)
		{
			if (dc<rthresh)
				FCEUWave[V]+= amp;
			--vc;
			if (vc<=0)   /* Less than zero when first started. */
				vc = wl, dc = (dc+1)&7;
		}
		MMC5Sound.dcount[P] = dc;
		MMC5Sound.vcount[P] = vc;
	}
	MMC5Sound.BC[P] = SOUNDTS;
}

static void Do5PCMHQ()
{
	int32		V, amp;

	if (!(MMC5Sound.rawcontrol&RAW_DISABLE) && (amp=MMC5Sound.raw*volume>>3)!=0)
		for(V=MMC5Sound.BC[2]; V<SOUNDTS; ++V)
			FCEUWave[V]+= amp;
	MMC5Sound.BC[2] = SOUNDTS;
}

void MMC5RunSoundHQ(void)
{
	Do5SQHQ(0);
	Do5SQHQ(1);
	Do5PCMHQ();
}

static void MMC5HiSync(int32 ts)
{
	int32		*d = MMC5Sound.BC;

	*d = d[1] = d[2] = ts;
}
#endif

void MMC5RunSound(int32 Count)
{
	int32		*d;

	Do5SQ(0);
	Do5SQ(1);
	Do5PCM();
	d = MMC5Sound.BC;
	*d = d[1] = d[2] = Count;
}

static void Mapper5_ESI(void)
{
	GameExpSound.RChange[SCMMC5] = Mapper5_ESI;
	if (FSettings.SndRate)
	{
	#if HAVE_HQ
		if (FSettings.soundq>=1)
			sfun = Do5SQHQ, psfun = Do5PCMHQ;
		else
	#endif
			sfun = Do5SQ, psfun = Do5PCM;
	} else
		sfun = 0, psfun = 0;
	memset(MMC5Sound.BC,0,sizeof(MMC5Sound.BC));
	memset(MMC5Sound.vcount,0,sizeof(MMC5Sound.vcount));
#if HAVE_HQ
	GameExpSound.HiSync[SCMMC5] = MMC5HiSync;
#endif
}

void NSFMMC5_Init(void)
{
	GenMMC5Close(); //2013.05.17

	memset(&MMC5Sound,0,sizeof(MMC5Sound));
	mul[0] = mul[1] = 0;
	MMC5ExRAM = (uint8*)FCEU_gmalloc(1024);
	Mapper5_ESI();
	SetWriteHandler(0x5C00,0x5FEF,MMC5_ExRAMWr);
	SetReadHandler(0x5C00,0x5FEF,MMC5_ExRAMRd);
	MMC5ExMode = 2;
	SetWriteHandler(0x5000,0x5015,MMC5W5);
	SetWriteHandler(0x5205,0x5206,MMC5W5200);
	SetReadHandler(0x5205,0x5206,MMC5Rmul);
}

void GenMMC5Close(void)
{
	if (MMC5ExRAM)
		{free(MMC5ExRAM); MMC5ExRAM = 0;}
	if (WRAM) //2013.05.17
		{free(WRAM); WRAM = 0;}
	MMC5Ex = 0; //2013.05.17
}

static void GenMMC5Reset(void)
{
	int		x = 0;

	do
		PRGBanks[x] = 0xFF, CHRBanksB[x] = x; /*~0;*/
	while(++x<4);
	x = 0;
	do
	
		CHRBanksA[x] = x; /*~0;*/
	while(++x<8);
	WRAMMaskEnable[0] = WRAMMaskEnable[1] =~0;

	mmc5psize = mmc5vsize = 3;
	/* more inits */
	MMC5ExMode = MMC5VROMMask = MMC5SplitCtrl = MMC5CHRHigh =
		MMC5SplitY = MMC5SplitBank = 0;

	NTAMirroring = NTFill = ATFill = 0xFF;

	MMC5Synco();

	SetWriteHandler(0x4020,0x5BFF,MMC5W5100);
	SetReadHandler(0x5204,0x5204,MMC5R5204); //0x4020,0x5bff,MMC5_read; 2013.05.17

	SetWriteHandler(0x5C00,0x5FFF,MMC5_ExRAMWr);
	SetReadHandler(0x5C00,0x5FFF,MMC5_ExRAMRd);

	SetWriteHandler(0x6000,0xFFFF,MMC5_WriteROMRAM);
	SetReadHandler(0x6000,0xFFFF,MMC5_ReadROMRAM);

	SetWriteHandler(0x5000,0x5015,MMC5W5);
	SetWriteHandler(0x5200,0x5206,MMC5W5200);
	SetReadHandler(0x5205,0x5206,MMC5Rmul);

	//GameHBIRQHook=MMC5_hb;
	FCEU_CheatSetRAM(8,0x6000,WRAM);
	FCEU_CheatSetRAM(1,0x5C00,MMC5ExRAM);

	MMC5Sound.rawcontrol = 1; /* initially disabled */
}

static SFORMAT MMC5_StateRegs[] =
{
//	{ &mmc5psize, 1, "PSIZ"},
//	{ &mmc5vsize, 1, "VSIZ"},
	{ PRGBanks, 4, "PRGB"},
	{ CHRBanksL, 8, "CHRA"},
	{ CHRBanksL+8, 4, "CHRB"},
	{ CHRBanksH, 8+4, "CHRH"}, /* 2013.05.29 */

	{ &WRAMPage, 1, "WRMP"},
	{ WRAMMaskEnable, 2, "WRME"},
	{ &MMC5ABMode, 1, "ABMD"},
	{ &IRQScanline, 1, "IRQS"},
	{ &IRQEnable, 1, "IRQE"},
	{ &MMC5ExMode, 1, "CHRM"},
	{ &NTAMirroring, 1, "NTAM"},
	{ &NTFill, 1, "NTFL"},
	{ &ATFill, 1, "ATFL"},

	{ &MMC5Sound.wl[0], 2|FCEUSTATE_RLSB, "SDW0"},
	{ &MMC5Sound.wl[1], 2|FCEUSTATE_RLSB, "SDW1"},
	{ MMC5Sound.env, 2, "SDEV"},
	{ &MMC5Sound.enable, 1, "SDEN"},
	{ &MMC5Sound.running, 1, "SDRU"},
	{ &MMC5Sound.raw, 1, "SDRW"},
	{ &MMC5Sound.rawcontrol, 1, "SDRC"},

	{ mul, 2, "MUL"}, /* 2013.05.29 */
	{0}
};

static void GenMMC5_Init(CartInfo *info,
	int wsize,
	int battery)
{
	GenMMC5Close(); //2013.05.17
	if (wsize)
	{
		WRAM = (uint8*)FCEU_gmalloc(wsize*1024);
		SetupCartPRGMapping(0x10,WRAM,wsize*1024,1);
		AddExState(WRAM, wsize*1024, 0, "WRAM");
	}

	MMC5fill = (uint8*)FCEU_gmalloc(1024);
	MMC5ExRAM = (uint8*)FCEU_gmalloc(1024);

	AddExState(MMC5_StateRegs, ~0, 0, 0);
	AddExState(WRAM, wsize*1024, 0, "WRAM");
	AddExState(MMC5ExRAM, 1024, 0, "ERAM");
	AddExState(&MMC5SplitCtrl, 1, 0, "SPLM");
	AddExState(&MMC5SplitY, 1, 0, "SPLS");
	AddExState(&MMC5SplitBank, 1, 0, "SPLP");
	AddExState(&MMC5CHRHigh, 1, 0, "5130");

	MMC5WRAMsize = wsize/8;
	BuildWRAMSizeTable();
	GameStateRestore = MMC5_StateRestore;
	info->Power = GenMMC5Reset;
	info->Close = GenMMC5Close; //2013.05.17

	if (battery)
	{
		info->SaveGame[0] = WRAM;
		if (wsize<=16)
			info->SaveGameLen[0] = 8192;
		else
			info->SaveGameLen[0] = 32768;
	}

	MMC5VROMMask = CHRmask4[0]&0x3F;
	MMC5Ex = 1;
	MMC5VROMPtr = CHRptr[0];
	MMC5ExMode = 0;
	MMC5SplitCtrl = MMC5SplitY = MMC5SplitBank = 0;
	Mapper5_ESI();
}

void ExROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,DetectMMC5WRAMSize(info->CRC32),info->battery);
}

// ELROM seems to have 0KB of WRAM
// EKROM seems to have 8KB of WRAM
// ETROM seems to have 16KB of WRAM
// EWROM seems to have 32KB of WRAM

// ETROM and EWROM are battery-backed, EKROM isn't.

void ETROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,16,info->battery);
}

void ELROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,0,0);
}

void EWROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,32,info->battery);
}

void EKROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,8,info->battery);
}
