/*
 boards/N106.c
  covers mappers 19 (Namcot163), 210 (Namcot???)

 * MacFCEU 0.8a9
 *
 * Copyright notice for this file:
 *  n106.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "Namcot.h"
#include "../cheat.h"

static uint16	IRQCount;
static uint8	IRQa;

static uint8	*WRAM = 0;
static uint8	IRAM[128];

static DECLFR(AWRAM)
{
	return WRAM[A-0x6000];
}

static DECLFW(BWRAM)
{
	if (WRAM)
		WRAM[A-0x6000] = V;
}

static uint8	NTAPage[4];
static uint8	dopol;
static uint8	CHRsel;
/*static uint8	gorko;*/

static void NamcoSound(int32 Count);
static void NamcoSoundHack(void);
static void DoNamcoSound(int32 *Wave,
	int Count);
static void DoNamcoSoundHQ(void);
static void SyncHQ(int32 ts);

static uint8	is210;	/* Lesser mapper. */
uint8		N163active;

static uint8	PRG[3];
static uint8	CHR[8];

static SFORMAT N1xxStateRegs[] =
{
	{IRAM,128,"IRAM"},
	{&IRQCount,2|FCEUSTATE_RLSB,"IRQC"}, /* these two were missing */
	{&IRQa,1,"IRQA"},

	{PRG,3,"PRG"},
	{CHR,8,"CHR"},
	{NTAPage,4,"NTA"},
	{&dopol,1,"IRDP"}, /* I also decided to add these... */
	{&CHRsel,1,"CHRG"},
	{0}
};

static void SyncPRG(void)
{
	setprg8(0x8000,PRG[0]);
	setprg8(0xA000,PRG[1]);
	setprg8(0xC000,PRG[2]);
	setprg8(0xE000,0x3F);
}

static void FP_FASTAPASS(1) NamcoIRQHook(int a)
{
	if (IRQa && (IRQCount+=a)>=0x7FFF)
	{
		X6502_IRQBegin(FCEU_IQEXT);
		IRQa = 0, IRQCount = 0x7FFF;
	}
}

static DECLFR(N163R4800)
{
	uint8		ret = IRAM[dopol&0x7F];
(void)A;
	/* Maybe I should call NamcoSoundHack() here? */
	if (!fceuindbg && dopol&0x80)
		dopol = (dopol&0x80)|((dopol+1)&0x7F);
	return ret;
}

static DECLFR(N163R5000)
{
	if (A&0x800)
		return IRQCount>>8;
	return IRQCount;
}

static void FASTAPASS(2) DoNTARAMROM(int w,
	uint8 V)
{
        NTAPage[w] = V;
        //if(V>=0xE0)
	// setntamem(NTARAM+((V&1)<<10), 1, w);

	//printf("%d, %02x\n",w,V);
	//printf("%02x, %02x\n",((gorko>>(6+(w>>1)))&1),V);
	if (V>=0xE0)
		setntamem(NTARAM+((V&1)<<10),1,w);
	else
	{
		V&= CHRmask1[0];
		setntamem(CHRptr[0]+(V<<10),0,w);
	}
}

static void FixNTAR(void)
{
	int		x = 0;

	do
		DoNTARAMROM(x,NTAPage[x]);
	while(++x<4);
}

static void FASTAPASS(2) DoCHRRAMROM(int x,
	uint8 V)
{
	CHR[x] = V;

	if (!is210 && !((CHRsel>>(x>>2)+6)&1) && V>=0xE0)
	{
		//printf("BLAHAHA: %d, %02x\n",x,V);
		//setchr1r(0x10,x<<10,V&7);
	} else
	{
		//printf("Noha: %d, %02x\n",x,V);
		setchr1(x<<10,V);
	}
}

static void FixCRR(void)
{
	int		x = 0;

	do
		DoCHRRAMROM(x,CHR[x]);
	while(++x<8);
}

static DECLFW(N163WC0D8)
{
	DoNTARAMROM((A-0xC000)>>11,V);
}

static uint32	FreqCache[8]; /* 18-bit */
static uint32	EnvCache[8]; /* 23-bit scale */
static uint32	LengthCache[8]; /* 6-bit */
static uint32	volume = 576716;

static void FixCache(int a,
	int V)
{
	int		w = a>>3&7;

	if (!(a&=7))
		FreqCache[w]&=~0x000000FF, FreqCache[w]|= V;
	else if (a==2)
		FreqCache[w]&=~0x0000FF00, FreqCache[w]|= V<<8;
	else if (a==4)
		FreqCache[w]&=~0x00030000, FreqCache[w]|= (V&3)<<16,
			LengthCache[w] = (8-(V>>2&7))<<2;
	else if (a==7)
		EnvCache[w] = (V&0xF)*volume;
}

static DECLFW(N163W4800)
{
(void)A;
	//printf("Yahaoo: %02x, %02x\n",dopol&0x7F,V);
	//puts("Hmm");
	if (dopol&0x40)
	{
		if (FSettings.SndRate)
		{
			if (!(~FSettings.soundactivech&(0x20|64u<<SCN106)))
				NamcoSoundHack();
			GameExpSound.Fill[SCN106] = NamcoSound;
			//GameExpSound.NeoFill=DoNamcoSound;
			GameExpSound.HiFill[SCN106] = DoNamcoSoundHQ;
			GameExpSound.HiSync[SCN106] = SyncHQ;
		}
		FixCache(dopol,V);
	}
	IRAM[dopol&0x7F] = V;

	if (dopol&0x80)
		dopol = (dopol&0x80)|((dopol+1)&0x7f);
}

static DECLFW(N163W5000)
{
	if (A&0x800) /* 0x5800-5FFF */
		IRQCount&= 0x00FF, IRQCount|= (V&0x7F)<<8, IRQa = V&0x80;
	else /* 0x5000-57FF */
		IRQCount&= 0xFF00, IRQCount|= V;
	X6502_IRQEnd(FCEU_IQEXT);
}

static DECLFW(N163W8000)
{
	DoCHRRAMROM((A-0x8000)>>11,V);
}

static DECLFW(N163WE000)
{
	if (A&0x800)
	{
		CHRsel = V&0xC0;
		FixCRR();
		PRG[1] = V&0x3F;
	} else
		/*gorko = V&0xC0,*/ PRG[0] = V&0x3F;
	SyncPRG();
}

static DECLFW(N163WF000)
{
	if (A&0x800)
		dopol = V;
	else
	{
		PRG[2] = V&0x3F;
		SyncPRG();
	}
}

static int	dwave = 0;

static void NamcoSoundHack(void)
{
	int32		a;
#if HAVE_HQ
	if (FSettings.soundq)
		{DoNamcoSoundHQ(); return;}
#endif
	if ((a=(((SOUNDTS<<16)/soundtsinc)>>4)-dwave)==0)
		return;
	DoNamcoSound(&FCEUWave[dwave],a);
	dwave+= a;
}

static void NamcoSound(int32 Count)
{
	int32		a;
(void)Count;
	if ((a=(((SOUNDTS<<16)/soundtsinc)>>4)-dwave)==0)
		return;
	DoNamcoSound(&FCEUWave[dwave],a);
	dwave = 0;
}

/* these are not saved to state */
static uint32	PlayIndex[8];
static int32	vcount[8];
static int32	CVBC;

#define TOINDEX	(16+1)

// 16:15
static void SyncHQ(int32 ts)
{
	CVBC = ts;
}

/* Things to do:
	1	Read freq low
	2	Read freq mid
	3	Read freq high
	4	Read envelope
	...?
*/

static void DoNamcoSoundHQ(void)
{
	int32		P, V, R = IRAM[0x7F]>>4&7;
	int32		cyclesuck = ((R)+1)*15;

	for(P=7; P>=7-R; --P)
		if (IRAM[0x44+(P<<3)]&0xE0 && (IRAM[0x47+(P<<3)]&0xF))
		{
			uint32	*pix = PlayIndex+P, freq, duff, lengo, envelope;
			int32	vco = vcount[P], x = IRAM[0x46+(P<<3)];

			freq = FreqCache[P];
			envelope = EnvCache[P];
			lengo = LengthCache[P];
			duff = (x+(*pix>>TOINDEX))&0xFF;
			duff = ((IRAM[duff>>1]>>(duff<<2&4)&15)*envelope)>>16;
			for(V=CVBC<<1; V<SOUNDTS<<1; ++V)
			{
				FCEUWave[V>>1]+= duff;
				if (!vco)
				{
					*pix+= freq;
					while((*pix>>TOINDEX)>=lengo)
						*pix-= lengo<<TOINDEX;
					duff = (x+(*pix>>TOINDEX))&0xFF;
					duff = ((IRAM[duff>>1]>>(duff<<2&4)&15)*envelope)>>16;
					vco = cyclesuck;
				}
				--vco;
			}
			vcount[P] = vco;
		}
	CVBC = SOUNDTS;
}

static void DoNamcoSound(int32 *Wave,
	int Count)
{
	int		P, V, R = IRAM[0x7F]>>4&7;
	double		num = FSettings.SndRate*(R+1)
				*(45.0*16384/16*(1<<15+8)/21477272);

	Count<<= 4;
	for(P=7; P>=7-R; --P)
	{
		int32		inc, vco, x;
		uint32		*pix, freq, duff, lengo, envelope;

		if (!(IRAM[0x44+(P<<3)]&0xE0) || !(IRAM[0x47+(P<<3)]&0xF))
			continue;
		vco = vcount[P];
		freq = FreqCache[P];
		envelope = EnvCache[P];
		lengo = LengthCache[P];

		if (!freq)
			{/*printf("Ack");*/ continue;}

		//int c = (IRAM[0x7F]>>4&7)+1;
		//inc = (long double)(FSettings.SndRate<<15)/((long double)freq*
		//	21477272/((long double)0x400000*c*45));
		inc = num/freq;
		x = IRAM[0x46+(P<<3)], pix = PlayIndex+P;
		duff = (x+*pix)&0xFF;
		duff = ((IRAM[duff>>1]>>(duff<<2&4)&15)*envelope)>>16-1;//+4;
		for(V=0; V<Count; ++V)
		{
			if (vco>=inc)
			{
				if (++*pix>=lengo)
					*pix = 0;
				vco-= inc;
				duff = (x+*pix)&0xFF;
				duff = ((IRAM[duff>>1]>>(duff<<2&4)&15)*envelope)>>16-1;//+4;
			}
			Wave[V>>4]+= duff;
			vco+= 0x8000;
		}
		vcount[P] = vco;
	}
}

static void FixSndCache(void)
{
	int		x = 0x40;

	do
		FixCache(x,IRAM[x]);
	while(++x<0x80);
}

void FCEU_N163SetVolume(int v)
{
#if HAVE_HQ
	if (FSettings.soundq && v<0)
		v =-v;
#endif
	volume = v*(576716>>2)/25;
	FixSndCache();
}

static void Namcot163Restore(int v)
{
#pragma unused(v)
	SyncPRG();
	FixNTAR();
	FixCRR();
	FixSndCache();
}

static void M19SC(void)
{
	if (FSettings.SndRate)
		Namcot163_ESI();
}

void Namcot163_ESI(void)
{
	N163active = 1;
	GameExpSound.RChange[SCN106] = M19SC;
	memset(vcount,0,sizeof(vcount));
	memset(PlayIndex,0,sizeof(PlayIndex));
	CVBC = 0;
}

void NSFN106_Init()
{ /* no IRQ triggering */
	dwmemset(IRAM,0,128);
	SetReadHandler(0x4800,0x4FFF,N163R4800);
	SetWriteHandler(0x4800,0x4FFF,N163W4800);
	SetWriteHandler(0xF800,0xFFFF,N163WF000);
	Namcot163_ESI();
}

static uint8	battery = 0;

void GenN1xxClose()
{
	N163active = 0;
	if (WRAM) /* all of those 8KBs add up... */
		{FCEU_free(WRAM); WRAM = 0;}
}

void GenN1xxPower()
{
	SetCartRead(0x8000,0xFFFF);
	SetWriteHandler(0x4800,0x4FFF,N163W4800); /*start was 0x4020*/
	SetWriteHandler(0x5000,0x5FFF,N163W5000);
	SetWriteHandler(0x8000,0xB800,N163W8000);
	SetWriteHandler(0xE000,0xEFFF,N163WE000);
	SetWriteHandler(0xF800,0xFFFF,N163WF000);

	if (!is210)
	{
		SetWriteHandler(0xC000,0xDFFF,N163WC0D8);
		SetReadHandler(0x4800,0x4FFF,N163R4800);
		SetReadHandler(0x5000,0x5FFF,N163R5000);
		NTAPage[0] = NTAPage[1] = NTAPage[2] = NTAPage[3] = 0xFF;
		FixNTAR();
	}

	SetReadHandler(0x6000,0x7FFF,AWRAM);
	SetWriteHandler(0x6000,0x7FFF,BWRAM);
	FCEU_CheatSetRAM(8,0x6000,WRAM);

	N163active = 1, CHRsel = 0xFF;
	SyncPRG();
	FixCRR();

	if (!battery)
	{
		dwmemset(WRAM,0,8192);
		dwmemset(IRAM,0,128);
	}

	FixSndCache();
}

void Namcot163_Init(CartInfo *info)
{
	if (!WRAM && (WRAM=FCEU_gmalloc(8192))==0)
		return;
	dwmemset(IRAM,0,128);
	is210 = 0;
	battery = info->battery;

	info->Power = GenN1xxPower;
	info->Close = GenN1xxClose;
	MapIRQHook = NamcoIRQHook;
	GameStateRestore = Namcot163Restore;
	GameExpSound.RChange[SCN106] = M19SC;

	if (FSettings.SndRate)
		Namcot163_ESI();

	if (info->battery)
	{
		info->SaveGame[0] = WRAM;
		info->SaveGameLen[0] = 8192;
		info->SaveGame[1] = IRAM;
		info->SaveGameLen[1] = 128;
	}

	AddExState(WRAM,8192,0,"WRAM");
	AddExState(N1xxStateRegs,~0,0,0);
}

static void M210Restore(int v)
{
(void)v;
	SyncPRG();
	FixCRR();
}

void NamcotM210_Init(CartInfo *info)
{
	if (!WRAM && (WRAM=FCEU_gmalloc(8192))==0)
		return;
	dwmemset(IRAM,0,128);
	is210 = 1;
	info->Power = GenN1xxPower;
	info->Close = GenN1xxClose;
	GameStateRestore = M210Restore;
	AddExState(WRAM,8192,0,"WRAM");
	AddExState(N1xxStateRegs,~0,0,0);
}
