/*
 boards/Sachen.c
  needs to be updated

 * MacFCEU 0.7
 *
 * Copyright notice for this file:
 *  sachen.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "latch.h"
#include "Sachen.h"

static uint8	type, latch[8];
#define cmd	latch_byte
#define dip	latch[7]

static void SAPower(void)
{
	int		x;
	writefunc	wf;

	Latch_Power();
	wf = GetWriteHandler(0x4100);
 	x = 0x4300;
	do /* finish */
		SetWriteHandler(x,x+0xFF,wf);
	while((x+=0x200)<0x6000);
}

static void TCU0xPower(void)
{
	int		x;
	writefunc	wf;

	Latch_Power();
	wf = GetWriteHandler(0x4102);
 	x = 0x4102;
	do /* finish */
		SetWriteHandlerI(4,x,x+0xFC,wf);
	while((x+=0x200)<0x10000);
}

/* ---- mapper 150: Sachen 74LS374N */

#define COPYRIGHT_VALUE		(1^1) //hard-coded, sorry

static DECLFR(S74LS374NRead)
{
	if ((A&0xF101)!=0x4100 || !type)
		return A>>8;
	return (~(cmd&7)&0x3F)^(dip ? COPYRIGHT_VALUE:1);
}

static void GenSachenMir(uint8 mir)
{
	mir^= 1;
	if ((mir&=3)==3)
		setmirrorw(0,1,1,1);
	else
		setmirror(mir);
}

static void S74LS374NSync(void)
{
	uint16		v;

	if ((v=latch_addr)==0)
	{
		setprg32(0x8000,latch[0]);
		setchr8(latch[1]);
		GenSachenMir(latch[2]);
		return;
	}
	switch((cmd&7)|type)
	{
	case 0:
		setchr8(latch[1]=3);
		v = 0;
		goto sprg;
	case 8|2:
		setprg32(0x8000,latch[0]=v&1);
	case 2:
		latch[1]&=~8U;
		setchr8(latch[1]|=v<<3&8);
		break;
	case 4:
		latch[1]&=~1U;
		setchr8(latch[1]|=v&1);
		break;
	case 5:
		v&= 1;
		goto sprg;
	case 6:
		latch[1]&=~6U, v+= v, v&= 6;
		setchr8(latch[1]|=v);
		break;
	case 7:
		GenSachenMir(latch[2]=v&1);
		break;
	case 8|4:
		latch[1]&=~4U;
		setchr8(latch[1]|=v<<2&4);
		break;
	case 8|5:
		v&= 7;
	sprg:	setprg32(0x8000,latch[0]=v);
		break;
	case 8|6:
		latch[1]&=~3U;
		setchr8(latch[1]|=v&3);
		break;
	case 8|7:
		GenSachenMir(latch[2]=v>>1&3);
	}
	latch_addr = 0;
}

static void S74LS374NPower(void)
{
	latch[0] = latch[1] = latch[2] = dip = 0;
	SAPower();
}

static void S74LS374NReset(void)
{
	dip^= 1; /* copyright: "Sachen" / "Sachen & Hacker" */
}

void S74LS374B_Init(CartInfo *info)
{
	type = 8;
	Latch_Init(info,S74LS374NSync,S74LS374NRead,1<<16,0x410041FFU);
	info->Power = S74LS374NPower;
	info->Reset = S74LS374NReset;
	AddExState(latch,3,0,"LATC");
	AddExState(&dip,1,0,"DIP");
}

/* ---- mapper 133: SA-72008 */

static void SA72008Sync()
{
	setprg32(0x8000,latch_byte>>2&1);
	setchr8(latch_byte&3);
}

void SA72008_Init(CartInfo *info)
{
	Latch_Init(info,SA72008Sync,0,1<<23,0x410041FFU);
	info->Power = SAPower;
}

/* ---- mapper 136: TCU-02 */

static DECLFR(TCU02Read)
{
	if ((A&0xF101)!=0x4100)
		return A>>8;
	return latch[0];
}

static void TCU02Sync(void)
{
	setchr8(latch[0]=(latch_byte&0x30)|((latch_byte+3)&0xF));
}

void TCU02_Init(CartInfo *info)
{
	Latch_Init(info,TCU02Sync,TCU02Read,0,0x41024102U);
	info->Power = TCU0xPower;
}

/* ---- mapper 137: Sachen 8259D */

static void S8259Sync(void)
{
	int		x, a;

	if ((a=latch_addr)!=0)
	{
		latch[x=cmd&7] = a;
		if (x==7)
			goto mir;
	}

	setprg32(0x8000,latch[5]&7);

	x = a = 0;
	if (!UNIFchrrama) //No CHR RAM?  Then BS'ing is ok.
		if (type==0x9D)
		{
			setchr1(0,latch[0]&7);
			setchr1(0x400,(latch[1]&7)|(latch[4]<<4&0x10));
			setchr1(0x800,(latch[2]&7)|(latch[4]<<3&0x10));
			setchr1(0xC00,(latch[3]&7)|(latch[4]<<2&0x10)|(latch[6]<<3&8));
		} else
		{
			unsigned	h = latch[4]<<3&0x38;
			unsigned	c = type==0x9C ? 2:0;
			unsigned	s = type==0x9A ? 1:c;
			unsigned	x = ~latch[7]&1;

			setchr2(0,(latch[0]&7)+h<<s);
			setchr2(0x800,(latch[x]&7)+h<<s|(type!=0x9B));
			setchr2(0x1000,(latch[x+x]&7)+h<<s|c);
			setchr2(0x1800,(latch[x+x+x]&7)+h<<s|(type==0x9A ? 1:c|(type==0x9C)));
		}
mir:
	GenSachenMir(latch[7]&1 ? 0:latch[7]>>1&3);
end:
	latch_addr = 0;
}

static void S8259Power(void)
{
	int		x = sizeof(latch);

	do
		latch[x-1] = 0;
	while(--x!=0);
	cmd = x;
	//if(UNIFchrrama)
	setchr8(x);

	SAPower();
}

void S8259D_Init(CartInfo *info)
{
	Latch_Init(info,S8259Sync,0,1<<16,0x410041FFU);
	info->Power = S8259Power;
	AddExState(latch,8,0,"LATC");
	type = 0x9D;
}

/* ---- mapper 138: Sachen 8259B */

void S8259B_Init(CartInfo *info)
{
	S8259D_Init(info);
	type = 0x9B;
}

/* ---- mapper 139: Sachen 8259C */

void S8259C_Init(CartInfo *info)
{
	S8259D_Init(info);
	type = 0x9B;
}

/* ---- mapper 141: Sachen 8259A */

void S8259A_Init(CartInfo *info)
{
	S8259D_Init(info);
	type = 0x9A;
}

/* ---- mapper 145: SA-72007 */

static void SA72007Sync()
{
	setprg32(0x8000,0);
	setchr8(latch_byte>>7);
}

void SA72007_Init(CartInfo *info)
{
	Latch_Init(info,SA72007Sync,0,0x800000U,0x410041FFU);
	info->Power = SAPower;
}

/* ---- mapper 146: SA-0161 */

static void SA0161MSync()
{
	setprg32(0x8000,latch_byte>>3&1);
	setchr8(latch_byte&7);
}

void SA0161M_Init(CartInfo *info)
{
	Latch_Init(info,SA0161MSync,0,0x800000U,0x410041FFU);
	info->Power = SAPower;
}

/* ---- mapper 147: TCU-01 */

static void TCU01Sync(void)
{
	setprg32(0x8000,(latch[0]>>6&2)|(latch[0]>>2&1));
	setchr8(latch[0]>>3&0xF);
}

void TCU01_Init(CartInfo *info)
{
	Latch_Init(info,TCU01Sync,0,0,0x41024102U);
	info->Power = TCU0xPower;
}

/* ---- mapper 148: SA-0037 */

static void SA0037Sync()
{
	setprg32(0x8000,latch_byte>>3); //&1
	setchr8(latch_byte&7);
}

void SA0037_Init(CartInfo *info)
{
	Latch_Init(info,SA0037Sync,0,0x800000U,0x8000FFFFU);
}

/* ---- mapper 149: SA-0036 */

static void SA0036Sync()
{
	setprg32(0x8000,0);
	setchr8(latch_byte>>7);
}

void SA0036_Init(CartInfo *info)
{
	Latch_Init(info,SA0036Sync,0,0x800000U,0x8000FFFFU);
}

/* ---- mapper 160: SA-009 */

static void SA009Sync()
{
	setprg32(0x8000,0);
	setchr8(latch_byte&1);
}

void SA009_Init(CartInfo *info)
{
	Latch_Init(info,SA009Sync,0,0x800000U,0x410041FFU);
}

/* ---- mapper 243: SA-009 */

void S74LS374A_Init(CartInfo *info)
{
	S74LS374B_Init(info);
	type = 0;
}
