/*
 boards/Super24.c
  BMC Super24-in-1

 * MacFCEU 0.6.4
 *
 * Copyright notice for this file:
 *  super24.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC3and6.h"
#include "unlicensed.h"

static int32	IRQCount,IRQLatch;
#define sizer	MMC3EXPREGS[0]
#define bigbnk1	MMC3EXPREGS[1]
#define bigbnk2	MMC3EXPREGS[2]

static const uint8	masko8[8] = {63,31,15,1,3,0,0,0};
//static int masko1[8]={511,255,127,7,7,0,0,0};

static void swsetprg8(uint32 A, uint8 V)
{
	V&= masko8[sizer&7];
	V|= bigbnk1*2;
	setprg8(A,V); //setprg8r((V/64)&15
}

static void swsetchr1(uint32 A, uint8 V)
{
	if (sizer&0x20)
		setchr1r(0x10,A,V);
	else
	{
//V&=masko1[sizer&7];
		V|= bigbnk2<<3&0xF00;
		setchr1r(V>>5&1,A,V); //V>>9&15
	}
}

//would be MMC3_hb_KickMasterHack if it wasn't for Count/Latch being 4 bytes...
static void Sup24_hb(void)
{
	MMC3IRQReload = 0;
	if (scanline==238)
		X6502_IRQBegin(FCEU_IQEXT);
	if (IRQCount>=0 && --IRQCount<0)
		if(MMC3IRQa)
		{
			MMC3IRQReload = 1;
			X6502_IRQBegin(FCEU_IQEXT);
			//printf("IRQ: %d,%d\n",scanline,timestamp);
		}
}

//left alone so the ExState's would work...
static DECLFW(Sup24IRQWrite)
{
	switch(A&0xE001)
	{
	case 0xc000:
		IRQLatch=V;
		if (MMC3IRQReload==1)
			IRQCount=IRQLatch;
		break;
	case 0xc001:
		MMC3IRQReload=1;
		IRQCount=IRQLatch;
		break;
	case 0xE000:
		MMC3IRQa=0;
		X6502_IRQEnd(FCEU_IQEXT);
		if (MMC3IRQReload==1)
			{IRQCount=IRQLatch;}
		break;
	case 0xE001:
		MMC3IRQa=1;
		if (MMC3IRQReload==1)
			{IRQCount=IRQLatch;}
		break;
	}
}

static DECLFW(Super24Write)
{
	switch(A)
	{
	case 0x5FF0:
		sizer = V;
		FixMMC3PRG(MMC3cmd);
		FixMMC3CHR(MMC3cmd);
		break;
	case 0x5FF1:
		bigbnk1 = V;
		FixMMC3PRG(MMC3cmd);
		break;
	case 0x5FF2:
		bigbnk2 = V;
		FixMMC3CHR(MMC3cmd);
	}
}

static void Super24Power(void)
{
	IRQCount = IRQLatch = 0;
	sizer = 0x24;
	bigbnk1 = 159;
	bigbnk2 = 0;

	GenMMC3Power(); //inits MMC3IRQa, MMC3cmd

	SetWriteHandler(0x5000,0x5002,Super24Write);
//	SetWriteHandler(0x8000,0xBFFF,MMC3_CMDWrite);
	SetWriteHandler(0xC000,0xFFFF,Sup24IRQWrite);
	SetCartRead(0x8000,0xFFFF);
}

static SFORMAT Sup24StateRegs[] =
{
	{GameMemBlock,8192,"CHRR"},
	{MMC3DReg,8,"DREG"},
	{&MMC3cmd,1,"CMD"}, //cmd and reload weren't even recorded!
	{&MMC3A000B,1,"A000"},
	{&MMC3IRQReload,1,"IRQR"},
	{&IRQCount,4|FCEUSTATE_RLSB,"IRQC"},
	{&IRQLatch,4|FCEUSTATE_RLSB,"IQL1"},
	{&MMC3IRQa,1,"IRQA"},
	{&sizer,1,"SIZA"},
	{&bigbnk1,1,"BIG1"},
	{&bigbnk2,1,"BIG2"},
	{0}
};

void Super24_Init(CartInfo *info)
{
	info->Power = Super24Power;
	info->Reset = GenMMC3Reset;
	GameStateRestore = GenMMC3Restore;
	MMC3pwrap = swsetprg8;
	MMC3cwrap = swsetchr1;
	GameHBIRQHook = Sup24_hb;
	SetupCartCHRMapping(0x10,GameMemBlock,8192,1);
	AddExState(Sup24StateRegs,~0,0,0);
}
