/*
 boards/Supervision.c

 * MacFCEU 0.6.4
 *
 * Copyright notice for this file:
 *  supervision.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "unlicensed.h"

#define CHRRAM	(GameMemBlock+16)

static void DoSuper(void)
{
	setprg8r(GameMemBlock[0]>>2&3,0x6000,(GameMemBlock[0]<<4&0x30)|0xF);
	if (GameMemBlock[0]&0x10)
	{
		setprg16r(GameMemBlock[0]>>2&3,0x8000,
			(GameMemBlock[0]<<3&0x18)|(GameMemBlock[1]&7));
		setprg16r(GameMemBlock[0]>>2&3,0xC000,
			(GameMemBlock[0]<<3&0x18)|7);
	} else
		setprg32r(4,0x8000,0);

	setmirror(~GameMemBlock[0]>>5&1);
}

static DECLFW(SuperWrite)
{
(void)A;
	if (!(GameMemBlock[0]&0x10))
	{
		GameMemBlock[0] = V;
		DoSuper();
	}
}

static DECLFW(SuperHi)
{
(void)A;
	GameMemBlock[1] = V; 
	DoSuper();
}

static void SuperReset(void)
{
	SetWriteHandler(0x6000,0x7FFF,SuperWrite);
	SetWriteHandler(0x8000,0xFFFF,SuperHi);
	SetCartRead(0x6000,0xFFFF);  
	GameMemBlock[0] = GameMemBlock[1] = 0;
	setprg32r(4,0x8000,0);
	setvram8(CHRRAM);
	FCEU_dwmemset(CHRRAM,0,8192);
}

static void SuperRestore(int version)
{
(void)version;
	DoSuper();
}

void Supervision16_Init(CartInfo *info)
{
	info->Power = SuperReset;
	GameStateRestore = SuperRestore;
	AddExState(&GameMemBlock[0],1,0,"L1");
	AddExState(&GameMemBlock[1],1,0,"L2");
	AddExState(CHRRAM,8192,0,"CHRR");
}
