/*
 boards/Taito.c
  mappers 80, 207

 * MacFCEU 0.8a8
 *
 * Copyright notice for this file:
 *  80.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "Taito.h"

static uint8	versB, lastA;

static uint8	IRAM[128];
static uint8	PRGbank[4]; /* mimics older form for backward compatibility */
static uint8	mapb[32];
#define Mirror		mapb[0]
#define Security	mapb[1]
#define CHRbank		(mapb+8)
#define CCache		(mapb+16)	/* rebuilt in Sync */

static void TaitoX1_005_Sync(int x)
{
	if (x>=0)
	{
		switch(x)
		{
		case 6:
		case 7:
		default: x = -1;
		case 0: setchr2(0x0000,CHRbank[0]>>1 /*&0x3F*/); if (x>=0) break;
		case 1: setchr2(0x0800,CHRbank[1]>>1 /*&0x3F*/); if (x>=0) break;
		case 2: setchr1(0x1000,CHRbank[2]); if (x>=0) break;
		case 3: setchr1(0x1400,CHRbank[3]); if (x>=0) break;
		case 4: setchr1(0x1800,CHRbank[4]); if (x>=0) break;
		case 5: setchr1(0x1C00,CHRbank[5]);
		}
	}

	if (versB)
	{
		CCache[0] = CCache[1] = CHRbank[0]>>7;
		CCache[2] = CCache[3] = CHRbank[1]>>7;

		do
			CCache[4+x] = CHRbank[2+x]>>7;
		while(++x<4);

		setmirror(MI_0+CCache[lastA]);
	} else
		setmirror(Mirror&1); //MIRROR_SET2
}

static DECLFW(M80W7EF0)
{
	switch(A)
	{
	case 0x7EF0:
	case 0x7EF1:
	case 0x7EF2:
	case 0x7EF3:
	case 0x7EF4:
	case 0x7EF5: CHRbank[A&=7] = V; TaitoX1_005_Sync(A); break;
	case 0x7EF6:
	case 0x7EF7: Mirror = V; TaitoX1_005_Sync(-1); break;
	case 0x7EF8:
	case 0x7EF9: Security = V; break;
	case 0x7EFA:
	case 0x7EFB: setprg8(0x8000,PRGbank[0]=V); break;
	case 0x7EFC:
	case 0x7EFD: setprg8(0xA000,PRGbank[1]=V); break;
	case 0x7EFE:
	case 0x7EFF: setprg8(0xC000,PRGbank[2]=V); break;
	}
}

static DECLFW(M80W7F00)
{
	IRAM[A&0x7F] = V;
}

static DECLFR(M80R7E00)
{
	if (A<=0x7EF9)
		return Security;
	else /* battery RAM has security check */
		return Security!=0xA3 ? A>>8:IRAM[A&0x7F];
}

static void TaitoX1_005_Restore(int v)
{
(void)v;
	TaitoX1_005_Sync(8);
	setprg8(0x8000,PRGbank[0]);
	setprg8(0xA000,PRGbank[1]);
	setprg8(0xC000,PRGbank[2]);
}

static void TaitoX1_005_Power(void)
{
	int		i;

	dwmemset(IRAM,0,128);

	i = 0;
	do
		CHRbank[i] = i, PRGbank[i] = i;
	while(++i<3);
	do
		CHRbank[i] = i;
	while(++i<8);

	TaitoX1_005_Restore(0);
	setprg8(0xE000,PRGbank[3]=~0);

	SetWriteHandler(0x7EF0,0x7EFF,M80W7EF0);
	SetWriteHandler(0x7F00,0x7FFF,M80W7F00);
	SetReadHandler(0x7EF8,0x7EF9,M80R7E00);
	SetReadHandler(0x7F00,0x7FFF,M80R7E00);
}

/* ---- mapper 80: Taito X1-005 version A */

void TaitoX1_005A_Init(CartInfo *info)
{
	versB = 0;
	if (info->battery)
		info->SaveGame[0] = IRAM, info->SaveGameLen[0] = 128;
	info->Power = TaitoX1_005_Power;
	GameStateRestore = TaitoX1_005_Restore;
	AddExState(mapb,32,0,"MPBY");
	AddExState(PRGbank,4,0,"PBL");
	AddExState(IRAM,128,0,"IRAM");
}

/* ---- mapper 207: Taito X1-005 version B */

static void FP_FASTAPASS(1) TaitoX1_005_PPUhook(uint32 A)
{
	static uint8	last = ~0;
	uint8		z;

	if (A>=0x2000)
		return;

	lastA = A>>=10;

	z = CCache[A];
	if (z!=last)
	{
		setmirror(MI_0+z);
		last = z;
	}
}

void TaitoX1_005B_Init(CartInfo *info)
{
	TaitoX1_005A_Init(info);
	versB = 1;
	PPU_hook = TaitoX1_005_PPUhook;
}
