/*
 boards/Tengen.c
  mappers 64, 158

 * MacFCEU 0.5.5
 *
 * Copyright notice for this file:
 *  tengen.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "Tengen.h"

uint8	TengenDReg[11]; //0:C0,C2,C4,C5,C6,C7,P0,P1,C1,C3,F:P2
uint8	TengenCmd;

static uint8	mir, IRQreload, IRQmode;
static uint8	IRQCount, IRQLatch, IRQa; //IRQCount's *Not* set by a write
static int	acount;

enum
{
	DRchr0,
	DRchr2,
	DRchr4,
	DRchr5,
	DRchr6,
	DRchr7,
	DRprg0, //6
	DRprg1,
	DRchr1, //8
	DRchr3,

	DRprg2
};

static SFORMAT Rambo_StateRegs[] =
{
	{&TengenCmd, 1, "CMD"},
	{&mir, 1, "MIR"},
	{&IRQreload, 1, "RMOD"},
	{&IRQmode, 1, "IRQM"},
	{&IRQCount, 1, "IRQC"},
	{&IRQa, 1, "IRQA"},
	{&IRQLatch, 1, "IRQL"},
	{TengenDReg, 11, "DREG"},
	{0}
};

static uint8	domirA;

//static void FP_FASTAPASS(2) (*setchr1wrap)(unsigned int A, unsigned int V);

#define HBlank() do{\
	if(IRQCount==0||IRQreload) \
		IRQCount=IRQLatch+IRQreload,IRQreload=0; \
	else if (--IRQCount==0&&IRQa) \
		X6502_IRQBegin(FCEU_IQEXT); \
}while(0)

static void FP_FASTAPASS(1) RAMBO1_IRQHook(int a)
{
	if (!IRQmode)
		return;
	acount+= a;
	while(acount>=4)
	{
		acount-= 4;
//		if (--IRQCount==0xFF && IRQa)
//			X6502_IRQBegin(FCEU_IQEXT);
		HBlank();
	}
}

//pins A12/M2:
//according to quietust, this is seems to occur at PPU cycle 264, not 25x
static void RAMBO1_hb(void)
{
	if (IRQmode)
		return;
	if (scanline==240)
		return; /* hmm.  Maybe that should be an mmc3-only call in fce.c. */
//	rmode = 0;
//	if (--IRQCount==0xFF && IRQa)
//	{
//		rmode = 1;
//		X6502_IRQBegin(FCEU_IQEXT);
//	}
	HBlank();
}

static void RAMBO1CHR(void) //appears to be correct now; 0x0000,0x1000
{
	int		x, chrxor = TengenCmd<<5&0x1000;

	if (TengenCmd&0x20)
	{
		setchr1(chrxor^0x000,TengenDReg[DRchr0]);
		setchr1(chrxor^0x400,TengenDReg[DRchr1]);
		setchr1(chrxor^0x800,TengenDReg[DRchr2]);
		setchr1(chrxor^0xC00,TengenDReg[DRchr3]);
	} else
	{
		setchr1(chrxor^0x000,x=TengenDReg[DRchr0]&0xFE);
		setchr1(chrxor^0x400,++x);
		setchr1(chrxor^0x800,x=TengenDReg[DRchr1]&0xFE);
		setchr1(chrxor^0xC00,++x);
	}

	x = 4;
	do
		setchr1(chrxor^(x*0x400),TengenDReg[(DRchr4-4)+x]);
	while(++x<8);
}

static void (*TengenCHR)(void) = RAMBO1CHR;

static void RAMBO1PRG(void)
{
	if (TengenCmd&0x40)
	{ //rolled order
		setprg8(0x8000,TengenDReg[DRprg2]);
		setprg8(0xA000,TengenDReg[DRprg0]);
		setprg8(0xC000,TengenDReg[DRprg1]);
	} else
	{
		setprg8(0x8000,TengenDReg[DRprg0]);
		setprg8(0xA000,TengenDReg[DRprg1]);
		setprg8(0xC000,TengenDReg[DRprg2]);
	}
}

static DECLFW(RAMBO1_write8)
{
	int	x;

	if ((A&1)==0)
		{TengenCmd = V; return;}
	x = TengenCmd&15;
	if (x<10)
	{
		TengenDReg[x] = V;
		if ((unsigned)(x-DRprg0)<=1)
			goto prg;
		TengenCHR();
	} else if (x==15)
	{
		TengenDReg[DRprg2] = V;
	prg:
		RAMBO1PRG();
	}
}

static DECLFW(RAMBO1_writeA)
{
	if (A&1)
		return;
	mir = V&1;
	if (domirA)
		setmirror(mir^1);
}

static DECLFW(RAMBO1_writeC)
{
	if ((A&=1)==0)
	{
		IRQLatch = V;
//		if (IRQreload==1)
//			IRQCount = IRQLatch;
	} else
	{
		IRQreload = 1;
//		IRQCount = IRQLatch;
		IRQmode = V&1;
	}
}

static DECLFW(RAMBO1_writeE)
{
(void)V;
	if ((IRQa=A&1)==0)
		X6502_IRQEnd(FCEU_IQEXT);
//	if (IRQreload==1)
//		IRQCount = IRQLatch;
}

static void RAMBO1_Restore(int v)
{
(void)v;
	TengenCHR();
	RAMBO1PRG();
	if (domirA)
		setmirror(mir^1); //the invert is correct
}

static void RAMBO1Power(void)
{
	int		x = 0;

	TengenCmd = mir = 0;
	TengenDReg[DRprg0] = TengenDReg[DRprg1] = TengenDReg[DRprg2] =~0;
	TengenDReg[DRchr0] = 0;
	TengenDReg[DRchr1] = 1;
	TengenDReg[DRchr2] = 2;
	TengenDReg[DRchr3] = 3;
	x = 4;
	do
		TengenDReg[(DRchr4-4)+x] = x;
	while(++x<8);

	RAMBO1_Restore(0);
	setprg8(0xE000,~0);

	SetWriteHandler(0x8000,0x8FFF,RAMBO1_write8);
	SetWriteHandler(0xA000,0xAFFF,RAMBO1_writeA);
	SetWriteHandler(0xC000,0xCFFF,RAMBO1_writeC);
	SetWriteHandler(0xE000,0xEFFF,RAMBO1_writeE);
}

static void RAMBO1_Init(CartInfo *info,
	uint8 domir)
{
	domirA = domir;

	info->Power = RAMBO1Power;
	TengenCHR = RAMBO1CHR;
	GameHBIRQHook = RAMBO1_hb;
	MapIRQHook = RAMBO1_IRQHook;
	GameStateRestore = RAMBO1_Restore;
	AddExState(Rambo_StateRegs,~0,0,0); //called on Power before?
}

/* ---- mapper 64: Tengen 800032 */

void Tengen800032_Init(CartInfo *info)
{
	RAMBO1_Init(info,1);
}

/* ---- mapper 158: Tengen 800037 */

/*
static int MirCache[8];
static unsigned	int PPUCHRBus;

static void FP_FASTAPASS(2) MirWrap(unsigned int A, unsigned int V)
{
	MirCache[A>>10] = V>>7&1;
	if (PPUCHRBus==A>>10)
		setmirror(MI_0+(V>>7&1));
	setchr1(A,V);
}

static void FP_FASTAPASS(1) MirrorFear(uint32 A)
{
	A = A>>10&7;
	PPUCHRBus = A;
	setmirror(MI_0+MirCache[A]);
}
*/

static void CHR8_00037(void)
{
	RAMBO1CHR();
	if (TengenCmd&0x80)
		setmirrorw(TengenDReg[DRchr2]>>7^1,
			TengenDReg[DRchr3]>>7^1,
			TengenDReg[DRchr4]>>7^1,
			TengenDReg[DRchr5]>>7^1);
	else
		setmirrorw(TengenDReg[DRchr0]>>7^1,
			TengenDReg[DRchr0]>>7^1,
			TengenDReg[DRchr1]>>7^1,
			TengenDReg[DRchr1]>>7^1);
}

void Tengen800037_Init(CartInfo *info)
{
	RAMBO1_Init(info,0);
	TengenCHR = CHR8_00037;
//	setchr1wrap = MirWrap;
//	PPU_hook = MirrorFear;
}

/*
 MIMIC1 boards T800002, T800004, T800030 are actually Sunsoft4...
*/
