/*
 boards/Waixing.c

 * MacFCEU 0.7
 *
 * Copyright notice for this file:
 *  164.c Copyright (C) 2002 Xodnizel
 *  mmc3.c Copyright (C) 1998 BERO
 *  mmc3.c Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC3and6.h"
#include "Waixing.h"

/* type F has only one CHR bank (I think) */
/* type I is a simplified version of type A; type A has as much as 4KB of PRGRAM */

/* ---- mapper 74: Waixing Type A */

static void M74PW(uint32 a,
	uint8 v)
{
	setprg8(a,v&0x3F);
}

static void M74CW(uint32 a,
	uint8 v)
{
	int		x = 8; //PRGsize[0]>>16&8;

	if (v==x || v==++x)
		setchr1r(0x10,a,v);
	else
		setchr1(a,v);
}

void Waixing_A_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3cwrap = M74CW;
	MMC3pwrap = M74PW;
	if (NewCHRRAM(0x10,2048))
		AddExState(CHRptr[0x10],CHRsize[0x10],0,"CHRR");
}

/* ---- mapper 165: Waixing SH2 */

static void M165CW(uint32 A,
	uint8 V)
{
	if (V<4)
		setchr1r(0x10,A,V);
	else
		setchr1(A,V);
}

void Waixing_SH2_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,128,8,info->battery);
	MMC3cwrap = M165CW;
	if (NewCHRRAM(0x10,8192))
		AddExState(CHRptr[0x10],CHRsize[0x10],0,"chrr");
}

/* ---- mapper 245: Waixing-H */

static void M245CW(uint32 A, uint8 V)
{
	if (!UNIFchrrama) //added; 2010.04.12
		setchr1(A,V&7);
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
}

static void M245PW(uint32 A, uint8 V)
{
	setprg8(A,(V&0x3F)|((MMC3EXPREGS[0]&2)<<5));
}

static void M245Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
}

void Waixing_H_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	info->Power = M245Power;
	MMC3cwrap = M245CW;
	MMC3pwrap = M245PW;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 249: Waixing +Security */

static void M249PW(uint32 A, uint8 V)
{
	if (MMC3EXPREGS[0]&2)
		if (V<0x20)
			V = (V&1)|(V>>3&2)|(V>>1&4)|(V<<2&8)|(V<<2&0x10);
		else
		{
			V-= 0x20;
			V = (V&3)|(V>>1&4)|(V>>4&8)|(V>>2&0x10)
				|(V<<3&0x20)|(V<<2&0xC0);
		}
	setprg8(A,V);
}

static void M249CW(uint32 A, uint8 V)
{
	if (MMC3EXPREGS[0]&2)
		V = (V&3)|(V>>1&4)|(V>>4&8)|(V>>2&0x10)
			|(V<<3&0x20)|(V<<2&0xC0);
	setchr1(A,V);
}

static DECLFW(M249Write5)
{
(void)A;
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M249Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
	SetWriteHandler(0x5000,0x5000,M249Write5);
}

void Waixing_Security_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3cwrap = M249CW;
	MMC3pwrap = M249PW;
	info->Power = M249Power;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 164: Waixing FFV */
/* so far virtually unchanged from 0.5.3 */

static SFORMAT StateRegs[] =
{
	{&MMC3cmd,1,"CMD"},
	{MMC3DReg,8,"DREG"},
	{0}
};

static void M164Restore(int v)
{
	if (v>9814)
		;
}

static DECLFW(M164Write)
{
	if ((A&0x7300)==0x5000)
		setprg32(0x8000,V);
}

static uint8	*WRAM = 0;

static DECLFR(AWRAM)
{
	return WRAM[A-0x6000];
}

static DECLFW(BWRAM)
{
	if (WRAM)
		WRAM[A-0x6000] = V;
}

static void M164Close(void)
{
	if (WRAM)
		{FCEU_free(WRAM); WRAM = 0;}
}

static void M164Power(void)
{
	setchr8(0);
	setprg32(0x8000,~0);
	MMC3cmd = 0;
	memset(MMC3DReg,0,8); //blah
//	Sync();
	SetCartRead(0x8000,0xFFFF);
	SetWriteHandler(0x4020,0xFFFF,M164Write);
	SetReadHandler(0x6000,0x7FFF,AWRAM);
	SetWriteHandler(0x6000,0x7FFF,BWRAM);
}

void Waixing_FFV_Init(CartInfo *info)
{
	WRAM = (uint8*)FCEU_gmalloc(8192);
	info->Power = M164Power;
	info->Close = M164Close;
	GameStateRestore = M164Restore;
	AddExState(&StateRegs,~0,0,0);
}

/* ---- mapper 252: (San Guo Zhi)

static void M252Power(void)
{
}

void Waixing_SGZ_Init(CartInfo *info)
{
	info->Power = M252Power;
} */
