/*
 boards/boards.c
  what addresses the whole lot; only NROM MMC is implemented

 * MacFCEU 0.8a4
 *
 * Copyright notice for this file:
 *  simple.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../types.h"
#include "../FCEU.h"
#include "../cart.h"
#include "../board.h"
#include "Bandai.h"
#include "JYCompany.h"
#include "latch.h"
#include "MMC1.h"
#include "MMC3and6.h"
#include "MMC5.h"
#include "Namcot.h"
#include "Sachen.h"
#include "Taito.h"
#include "Tengen.h"
#include "unlicensed.h"
#include "Waixing.h"

/* ---- the basic MMC */

/* some carts require 4KB, 8KB (FPBASIC) WRAM */

static void NROMPower(void)
{
	setprg16r(0x10,0x6000,0);
	setprg16(0x8000,0);
	setprg16(0xC000,~0);
	setchr8(0);
	SetCartRead(0x6000,0xFFFF);
	SetWriteHandler(0x6000,0x7FFF,CartBW);
}

static void NROM_Init(CartInfo *info)
{
	info->Power = NROMPower;
	SetupCartPRGMapping(0x10,GameMemBlock,8192,1);
	if (info->battery)
		info->SaveGame[0] = GameMemBlock, info->SaveGameLen[0] = 8192;
}

/* ---- the list */

BMAPPING FCEU_bmap[] =
{
/* MMC1 */
 {"SAROM",-1, SAROM_Init, 0},
 {"SBROM",-1, SBROM_Init, 0},
 {"SCROM",-1, SCROM_Init, 0},
 {"SEROM",-1, SEROM_Init, 0},
 {"SGROM",-1, SGROM_Init, 0},
 {"SKROM",-1, SKROM_Init, 0},
 {"SLROM",-1, SLROM_Init, 0},
 {"SL1ROM",-1, SL1ROM_Init, 0},
 {"SL2ROM",-1, SL2ROM_Init, 0},
 {"SNROM",-1, SNROM_Init, 0},
 {"SOROM",-1, SOROM_Init, 0},
 {"EVENT", 105, MMC1_Event_Init, 0},

/* MMC3 */
 {"TGROM",-1, TGROM_Init, 0},
 {"TR1ROM",-1, TFROM_Init, BMCFLAG_FORCE4},
 {"TEROM",-1, TEROM_Init, 0},
 {"TFROM",-1, TFROM_Init, 0},
 {"TLROM",-1, TLROM_Init, 0},
 {"TKROM",-1, TKROM_Init, 0},
 {"TSROM",-1, TSROM_Init, 0},
 {"TLSROM",-1, TLSROM_Init, 0},
 {"TKSROM", 118, TKSROM_Init, 0},
 {"TQROM", 119, TQROM_Init, 0}, /* Has CHR ROM and CHR RAM by default.  Hmm. */
 {"TVROM",-1, TLROM_Init, BMCFLAG_FORCE4},

/* MMC6 */
 {"HKROM",-1, HKROM_Init, 0},

/* MMC5 */
 {"EWROM",-1, EWROM_Init, 0},
 {"EKROM",-1, EKROM_Init, 0},
 {"ELROM",-1, ELROM_Init, 0},
 {"ETROM",-1, ETROM_Init, 0},

/* NROM */
 {"NROM", 0, NROM_Init, 0},
 {"NROM-128", 0, NROM_Init, 0},
 {"NROM-256", 0, NROM_Init, 0},
 {"RROM", 0, NROM_Init, 0},
 {"RROM-128", 0, NROM_Init, 0},

/* xxxROM */
 {"UNROM", 2, UxROM_Init, 0},
 {"UN1ROM", 94, UN1ROM_Init, 0},
 {"CNROM", 3, CNROM_Init, 0},
 {"ANROM", 7, AxROM_Init, 0},
 {"AN1ROM", 7, AxROM_Init, 0},
 {"CPROM", 13, CPROM_Init, BMCFLAG_16KCHRR},
 {"GNROM", 66, GNROM_Init, 0},
 {"MHROM",-1, MHROM_Init, 0},
 {"SROM", 0, NROM_Init, 0},
 {"STROM", 0, NROM_Init, 0},

/* other HVC/NES */
 {"QJ", 47, QJ_Init, 0},

/* AVE carts */
// { "MB-91", MB91_Init, 0},	// DeathBots
// { "NINA-06", NINA06_Init, 0}, // F-15 City War
// { "NINA-03", NINA03_Init, 0}, // Tiles of Fate
// { "NINA-001", NINA001_Init, 0}, // Impossible Mission 2
// { "AVE-NINA-01", NINA001_Init, 0},
 {"AVE-NINA-06", 79, AVE_NINA06_Init, 0},

/* Jaleco */
 {"JALECO-JF-14", 140, JF14x_Init, 0},
 {"JALECO-JF-16", 78, JF16_Init, 0},

/* AGCI */
 {"AGCI-47516", 11, AGCI47516_Init, 0}, //not completely sure about this
 {"AGCI-50282", 144, AGCI50282_Init, 0},

/* J.Y.Company */
 {"TEK90", 90, JYCA_Init, 0},

/* Namcot */
 {"NAMCOT-163", 19, Namcot163_Init, 0},
 {"NAMCOT-3301", 0, NROM_Init, 0},
 {"NAMCOT-3302", 0, NROM_Init, 0},
 {"NAMCOT-3303", 0, NROM_Init, 0},
 {"NAMCOT-3305", 0, NROM_Init, 0},
 {"NAMCOT-3311", 0, NROM_Init, 0},
 {"NAMCOT-3401", 206, DxROM_Init, 0},
 {"NAMCOT-3405", 206, DxROM_Init, 0},
 {"NAMCOT-3406", 206, DxROM_Init, 0},
 {"NAMCOT-3407", 206, DxROM_Init, 0},
 {"NAMCOT-3411", 0, NROM_Init, 0},
 {"NAMCOT-3413", 206, DxROM_Init, 0},
 {"NAMCOT-3414", 206, DxROM_Init, 0},
 {"NAMCOT-3415", 206, DxROM_Init, 0},
 {"NAMCOT-3416", 206, DxROM_Init, 0},
 {"NAMCOT-3417", 206, DxROM_Init, 0},
 {"NAMCOT-3425", 95, Namcot3425_Init, 0},
 {"NAMCOT-3451", 206, DxROM_Init, 0},
 {"DEROM", 206, DxROM_Init, 0},
 {"DE1ROM", 206, DxROM_Init, 0},
 {"DRROM", 206, DxROM_Init, 0},

/* Sachen Carts */
 {"TC-U01-1.5M", 147, TCU01_Init, 0},
 {"SA-002", 136, TCU02_Init, 0},
 {"SACHEN-8259A", 141, S8259A_Init, BMCFLAG_CHRROK},
 {"SACHEN-8259B", 138, S8259B_Init, BMCFLAG_CHRROK},
 {"SACHEN-8259C", 139, S8259C_Init, BMCFLAG_CHRROK},
 {"SACHEN-8259D", 137, S8259D_Init, BMCFLAG_CHRROK},
 {"SACHEN-74LS374N", 150, S74LS374B_Init, 0},
 {"SA-016-1M", 146, SA0161M_Init, 0},
 {"SA-72007", 145, SA72007_Init, 0},
 {"SA-72008", 133, SA72008_Init, 0},
 {"SA-0036", 149, SA0036_Init, 0},
 {"SA-0037", 148, SA0037_Init, 0},
 {"SA-009", 160, SA009_Init, 0}, //maybe

/* Taito/Discrete */
 {"TAITO-74*139/74", 87, Discrete_74_139_74_Init, 0},
 {"TAITO-74*161/161/32", 152, Discrete_74_161_161_32B_Init, 0}, //type B: special mirroring
 {"TAITO-CNROM-128", 3, CNROM_Init, 0},
 {"TAITO-NROM-128", 0, NROM_Init, 0},
 {"TAITO-NROM-256", 0, NROM_Init, 0},
 {"TAITO-X1-005", 80, TaitoX1_005A_Init, 0},

/* Tengen */
 {"TENGEN-800002", 0, NROM_Init, 0},
 {"TENGEN-800032", 64, Tengen800032_Init, 0},
 {"TENGEN-800037", 158, Tengen800037_Init, 0},

/* Active Enterprises */
 {"MLT-ACTION52", 228, AE_A52_Init, 0},

 {"PAL-ZZ", 37, CustomPAL_ZZ_Init, 0},

/* duplicate basic */
 {"BANDAI-NROM-128", 0, NROM_Init, 0},
 {"BANDAI-NROM-256", 0, NROM_Init, 0},
 {"IREM-NROM-128", 0, NROM_Init, 0},
 {"IREM-NROM-256", 0, NROM_Init, 0},
 {"KONAMI-CNROM", 3, CNROM_Init, 0},
 {"KONAMI-NROM-128", 0, NROM_Init, 0},
 {"SETA-NROM-128", 0, NROM_Init, 0},
 {"SUNSOFT-NROM-256", 0, NROM_Init, 0},

/* other */
 {"MARIO1-MALEE2", 55, BTL_MALEE_Init, 0},
 {"SUPERVISION16IN1", 53, Supervision16_Init, 0},
 {"NOVELDIAMOND9999999IN1", 54, Novel_Init, 0},
 {"SUPER24IN1SC03",-1, Super24_Init, 0},
 {"SMB2J", 43, BTL_SMB2jC_Init, 0},

 {"H2288", 123, Kay_H2288_Init, 0},
 {"8237",-1, UNL_8237_Init, 0}, //SuperGame Pocahontas 2

 {"D1038",-1, BMC_D1038_Init, 0},
 {"CC-21",-1, UNL_CC21_Init, 0},
 {"SUPERHIK8IN1", 45, BMC_SuperHeroXin1_Init, 0},

 {"22211", 132, UNL_22211A_Init, 0}, //maybe

/* generic BMCs - terminate with -2 mapper */
 {"190IN1",-2, BMC_190in1_Init, 0},

/* no UNIF string */
 {"SxROM (MMC1)", 1, SxROM_Init},
 {"TxROM (MMC3)", 4, TxROM_Init},
 {"ExROM (MMC5)", 5, ExROM_Init},
 {"RexSoftDBZ5", 12, RexSoftDBZ5_Init},
 {"BNROM", 34, BNROM_Init},
 {"Taito-74*161/138", 38, Discrete_74_161_138_Init},
 {"BMC SuperBig 7-in-1", 44, BMC_SuperBig7in1_Init},
 {"BMC SuperHIK 4-in-1", 49, BMC_SuperHIK4in1_Init},
 {"BMC Ball Games 11-in-1", 51, BMC_BallGames11in1_init},
 {"BMC Mario Party 7-in-1", 52, BMC_MarioParty7in1_Init},
 {"T3H53", 59, T3H53_Init},
 {"BMC Reset Based 4-in-1", 60, BMC_ResetBased4in1_Init},
 {"Taito-74*161/161/32-A", 70, Discrete_74_161_161_32A_Init},
 {"Waixing-A", 74, Waixing_A_Init},
 {"Sunsoft 2a", 93, Sunsoft2A_Init},
 {"HES Std.", 113, HES_STD_Init},
 {"Kasing Std.", 115, Kasing_STD_Init},
 {"SomeriTeam SL12", 116, SomeriTeam_SL12_Init},
 {"Kay Panda Prince", 121, Kay_PandaPrince_Init},
 {"MMC1-A", 155, MMC1A_Init},
 {"Waixing FFV", 164, Waixing_FFV_Init},
 {"Waixing SH2", 165, Waixing_SH2_Init},
 {"Unlicensed 22211B", 172, UNL_22211B_Init},
 {"Unlicensed 22211C", 173, UNL_22211C_Init},
 {"KOF96", 187, KOF96_Init},
 {"Bandai Karaoke", 188, Bandai_Karaoke_Init},
 {"Taito X1-005B", 207, TaitoX1_005B_Init},
 {"G37017", 208, G37017_Init},
 {"JYC-B", 209, JYCB_Init},
 {"N106-relative", 210, NamcotM210_Init}, //???
 {"JYC-C", 211, JYCB_Init}, //J.Y.Company type C, almost identical to JYCB
 {"BMC Golden Game xx0-in-1", 235, BMC_GoldenGamexx0in1_Init},
 {"CandE SHLZ", 240, CandE_SHLZ_Init},
 {"S74LS374A", 243, S74LS374A_Init},
 {"Waixing H", 245, Waixing_H_Init},
 {"Waixing Security", 249, Waixing_Security_Init},
 {"Nitra TDA", 250, Nitra_TDA_Init},
 {"Bootleg Pikachu Y2K", 254, BTL_PikachuY2K_Init},
 {0}
};
