/*
 cart.c

 * MacFCEU 0.8a3
 *
 * Copyright notice for this file:
 *  cart.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "driver.h"
#include "FCEU.h"
#include "PPU.h"

#include "cart.h"
#include "memory.h"

#include "general.h"

/*
   This file contains all code for coordinating the mapping in of the
 address space external to the NES.
   It's also (ab)used by the NSF code.
*/

uint32	ROM_size, VROM_size; /* totals (prev. in ines.c); *16KB PRG, *8KB CHR */

uint8	*Page[32], *VPage[8];
uint8	**VPageR = VPage;
uint8	*MMC5SPRVPage[8];
uint8	*MMC5BGVPage[8];

uint8	PRGIsRAM[32]; /* This page is/is not PRG RAM. */

/* 16 are (sort of) reserved for UNIF/iNES and 16 to map other stuff. */
uint8	CartCHRram[32];
uint8	CartPRGram[32];

uint8	*PRGptr[32];
uint8	*CHRptr[32];

uint32	PRGsize[32];
uint32	CHRsize[32];

uint32	PRGmask2[32];
uint32	PRGmask4[32];
uint32	PRGmask8[32];
uint32	PRGmask16[32];
uint32	PRGmask32[32];

uint32	CHRmask1[32];
uint32	CHRmask2[32];
uint32	CHRmask4[32];
uint32	CHRmask8[32];

/* can't we make it 16 entries instead of 32? */
DECLFR(CartBR00) {return Page[0][A];} DECLFR(CartBR08) {return Page[1][A];}
DECLFR(CartBR10) {return Page[2][A];} DECLFR(CartBR18) {return Page[3][A];}
DECLFR(CartBR20) {return Page[4][A];} DECLFR(CartBR28) {return Page[5][A];}
DECLFR(CartBR30) {return Page[6][A];} DECLFR(CartBR38) {return Page[7][A];}
DECLFR(CartBR40) {return Page[8][A];} DECLFR(CartBR48) {return Page[9][A];}
DECLFR(CartBR50) {return Page[10][A];} DECLFR(CartBR58) {return Page[11][A];}
DECLFR(CartBR60) {return Page[12][A];} DECLFR(CartBR68) {return Page[13][A];}
DECLFR(CartBR70) {return Page[14][A];} DECLFR(CartBR78) {return Page[15][A];}
DECLFR(CartBR80) {return Page[16][A];} DECLFR(CartBR88) {return Page[17][A];}
DECLFR(CartBR90) {return Page[18][A];} DECLFR(CartBR98) {return Page[19][A];}
DECLFR(CartBRA0) {return Page[20][A];} DECLFR(CartBRA8) {return Page[21][A];}
DECLFR(CartBRB0) {return Page[22][A];} DECLFR(CartBRB8) {return Page[23][A];}
DECLFR(CartBRC0) {return Page[24][A];} DECLFR(CartBRC8) {return Page[25][A];}
DECLFR(CartBRD0) {return Page[26][A];} DECLFR(CartBRD8) {return Page[27][A];}
DECLFR(CartBRE0) {return Page[28][A];} DECLFR(CartBRE8) {return Page[29][A];}
DECLFR(CartBRF0) {return Page[30][A];} DECLFR(CartBRF8) {return Page[31][A];}

void FASTAPASS(2) SetCartRead(uint16 a,
	uint16 e)
{
	uint16		v, n;

	if (a>e)
		return;
	v = a&0xF800;
	switch(v>>8+3)
	{
	case 0: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR00);
		if (n) break; a = ++v;
	case 1: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR08);
		if (n) break; a = ++v;
	case 2: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR10);
		if (n) break; a = ++v;
	case 3: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR18);
		if (n) break; a = ++v;
	case 4: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR20);
		if (n) break; a = ++v;
	case 5: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR28);
		if (n) break; a = ++v;
	case 6: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR30);
		if (n) break; a = ++v;
	case 7: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR38);
		if (n) break; a = ++v;
	case 8: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR40);
		if (n) break; a = ++v;
	case 9: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR48);
		if (n) break; a = ++v;
	case 10: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR50);
		if (n) break; a = ++v;
	case 11: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR58);
		if (n) break; a = ++v;
	case 12: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR60);
		if (n) break; a = ++v;
	case 13: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR68);
		if (n) break; a = ++v;
	case 14: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR70);
		if (n) break; a = ++v;
	case 15: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR78);
		if (n) break; a = ++v;
	case 16: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR80);
		if (n) break; a = ++v;
	case 17: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR88);
		if (n) break; a = ++v;
	case 18: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR90);
		if (n) break; a = ++v;
	case 19: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR98);
		if (n) break; a = ++v;
	case 20: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRA0);
		if (n) break; a = ++v;
	case 21: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRA8);
		if (n) break; a = ++v;
	case 22: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRB0);
		if (n) break; a = ++v;
	case 23: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRB8);
		if (n) break; a = ++v;
	case 24: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRC0);
		if (n) break; a = ++v;
	case 25: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRC8);
		if (n) break; a = ++v;
	case 26: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRD0);
		if (n) break; a = ++v;
	case 27: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRD8);
		if (n) break; a = ++v;
	case 28: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRE0);
		if (n) break; a = ++v;
	case 29: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRE8);
		if (n) break; a = ++v;
	case 30: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRF0);
		if (n) break; a = ++v;
	case 31: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRF8);
	}
}

DECLFR(CartBR)
{
	return Page[A>>11][A];
}

DECLFW(CartBW)
{
	uint32		AB = A>>11;

	if (PRGIsRAM[AB] && Page[AB])
		Page[AB][A] = V;
}

DECLFR(CartBROB)
{
	uint32		AB = A>>11;

	if (!Page[AB])
		return A>>8; //XFrame.DB
	return Page[AB][A];
}

#define setpageptr(s,A,p,r) do{ \
	uint8 *_p; uint32 _AB=(A>>11)+((s>>1)-1); int _x=s>>1,_r; \
	if((_p=p)==0) _r=0; else _r=CartPRGram[r],_p-=A; \
	do \
		PRGIsRAM[_AB]=_r,Page[_AB]=_p; \
	while(--_AB,--_x); \
}while(0)

static uint8	nothing[8192];

void ResetCartMapping(void)
{
	int		x;

	x = 31;
	do
		Page[x] = nothing-(x<<11),
		PRGptr[x] = CHRptr[x] = 0,
		PRGsize[x] = CHRsize[x] = 0;
	while(--x>=0);
	x = 7;
	do
		MMC5SPRVPage[x] = MMC5BGVPage[x] = VPageR[x] = nothing-(x<<10);
	while(--x>=0);
}

void SetupCartPRGMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram)
{
	PRGptr[chip] = p;
	PRGsize[chip] = size;

	PRGmask2[chip] = (size>>11)-1;
	PRGmask4[chip] = (size>>12)-1;
	PRGmask8[chip] = (size>>13)-1;
	PRGmask16[chip] = (size>>14)-1;
	PRGmask32[chip] = (size>>15)-1;

	CartPRGram[chip] = ram ? 1:0;
}

void SetupCartCHRMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram)
{
	chip&= 31; /* say NO to crash happiness; 2010.04.12 */
	CHRptr[chip] = p;
	CHRsize[chip] = size;

	CHRmask1[chip] = (size>>10)-1;
	CHRmask2[chip] = (size>>11)-1;
	CHRmask4[chip] = (size>>12)-1;
	CHRmask8[chip] = (size>>13)-1;

	CartCHRram[chip] = ram;
}

//added these 3; 2010.03.12
int NewCHRRAM(int w,
	uint32 z)
{
	uint8		*r;

	if ((r=(uint8*)FCEU_gmalloc(z))==0)
		return 0;
	SetupCartCHRMapping(w,r,z,1);

	return 1;
}

void DeleteCHRRAM(int w)
{
	if (CHRptr[w] && CartCHRram[w])
		{FCEU_free(CHRptr[w]); CHRptr[w] = nothing, CartCHRram[w] = 0;}
}

void EraseCHRRAM(int w,
	int32 v)
{
	if (CHRptr[w] && CartCHRram[w])
		memset(CHRptr[w],v,CHRsize[w]);
}

void FASTAPASS(3) setprg4r(int r,
	unsigned int A,
	unsigned int V)
{
	V&= PRGmask4[r];
	setpageptr(4,A,!PRGptr[r] ? 0:&PRGptr[r][V<<12],r);
}

void FASTAPASS(3) setprg8r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<13)
	{
		V&= PRGmask8[r];
		setpageptr(8,A,!p ? 0:&p[V<<13],r);
	} else
	{
		uint32		VA = V<<2;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,!p ? 0:&p[((VA+x)&m)<<11],r);
		while(A+=1<<11,++x<4);
	}
}

void FASTAPASS(3) setprg16r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<14)
	{
		V&= PRGmask16[r];
		setpageptr(16,A,!p ? 0:&p[V<<14],r);
	} else
	{
		uint32		VA = V<<3;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,!p ? 0:&p[((VA+x)&m)<<11],r);
		while(A+=1<<11,++x<8);
	}
}

void FASTAPASS(3) setprg32r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<15)
	{
		V&= PRGmask32[r];
		setpageptr(32,A,!p ? 0:&p[V<<15],r);
	} else
	{
		uint32		VA = V<<4;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,!p ? 0:&p[((VA+x)&m)<<11],r);
		while(A+=1<<11,++x<16);
	}
}

void FASTAPASS(3) setchr1r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p;
	uint32		AB;

	if ((p=CHRptr[r])==0)
		return;
	FCEUPPU_LineUpdate();
	/* treated non-volatile(!) */
	p-= A, p+= (V&CHRmask1[r])<<10, VPageR[AB=A>>10] = p;

	if (CartCHRram[r])
		PPUCHRRAM|= 1u<<AB;
	else
		PPUCHRRAM&=~(1u<<AB);
}

void FASTAPASS(3) setchr1(uint32 A,
	uint8 V)
{
	uint32		AB;

	if (!*CHRptr[0])
		return;
	FCEUPPU_LineUpdate();

	VPageR[AB=A>>10] = *CHRptr+((V&*CHRmask1)<<10)-A;

	if (CartCHRram[0])
		PPUCHRRAM|= 1u<<AB;
	else
		PPUCHRRAM&=~(1u<<AB);
}

void FASTAPASS(3) setchr2r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (!CHRptr[r])
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask2[r];
	VPageR[AB] = VPageR[AB+1] = &CHRptr[r][V<<11]-(A);

	if (CartCHRram[r])
		PPUCHRRAM|= 3u<<AB;
	else
		PPUCHRRAM&=~3u<<AB;
}

void FASTAPASS(3) setchr4r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (!CHRptr[r])
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask4[r];
	VPageR[AB] = VPageR[AB+1] =
		VPageR[AB+2] = VPageR[AB+3] = &CHRptr[r][V<<12]-(A);

	if (CartCHRram[r])
		PPUCHRRAM|= 15u<<AB;
	else
		PPUCHRRAM&=~(15u<<AB);
}

void FASTAPASS(2) setchr8r(int r,
	unsigned int V)
{
	if (!CHRptr[r])
		return;

	FCEUPPU_LineUpdate();

	V&= CHRmask8[r], V<<= 13;
	setvram8(&CHRptr[r][V]);

	if (CartCHRram[r])
		PPUCHRRAM|= 0xFF;
	else
		PPUCHRRAM&=~0xFF;
}

void FASTAPASS(2) setvram4(uint32 A,
	uint8 *p)
{
	uint32		AB = A>>10;
	register uint8	**w = AB+VPageR;

	p-= A;
	*w++ = p, *w++ = p, *w++ = p, *w = p;
	PPUCHRRAM|= 15u<<AB;
}

void FASTAPASS(1) setvram8(uint8 *p)
{
	register uint8	**w = VPageR;

	*w++ = p, *w++ = p, *w++ = p, *w++ = p,
	*w++ = p, *w++ = p, *w++ = p, *w = p;
	PPUCHRRAM|= 0xFF;
}

void FASTAPASS(3) setvramb1(uint8 *p,
	uint32 A,
	uint32 b)
{
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10;
	VPageR[AB] = p+(b<<10)-A;
	PPUCHRRAM|= 1<<AB;
}

void FASTAPASS(3) setvramb2(uint8 *p,
	uint32 A,
	uint32 b)
{
	uint8		**w;
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10, w = AB+VPageR;
	w[0] = w[1] = p+(b<<11)-A;
	PPUCHRRAM|= 3<<AB;
}

void FASTAPASS(3) setvramb4(uint8 *p,
	uint32 A,
	uint32 b)
{
	uint8		**w;
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10, w = AB+VPageR;
	w[0] = w[1] = w[2] = w[3] = p+(b<<12)-A;
	PPUCHRRAM|= 15<<AB;
}

void FASTAPASS(2) setvramb8(uint8 *p,
	uint32 b)
{
	uint8		**w;

	FCEUPPU_LineUpdate();

	w = VPageR;
	w[0] = w[1] = w[2] = w[3] = w[4] = w[5] = w[6] = w[7] = p+(b<<13);

	PPUCHRRAM|= 0xFF;
}

/* This function can be called without calling SetupCartMirroring(). */

void FASTAPASS(3) setntamem(uint8 *p,
	int ram,
	uint32 b)
{
	FCEUPPU_LineUpdate();

	vnapage[b] = p;
	PPUNTARAM&=~(1<<b);
	if (ram)
		PPUNTARAM|= 1<<b;
}

static int	mirrorhard = 0;

void setmirrorw(int a,
	int b,
	int c,
	int d)
{
	FCEUPPU_LineUpdate();

	vnapage[0] = NTARAM+a*0x400;
	vnapage[1] = NTARAM+b*0x400;
	vnapage[2] = NTARAM+c*0x400;
	vnapage[3] = NTARAM+d*0x400;
}

void FASTAPASS(1) setmirror(int t)
{
	FCEUPPU_LineUpdate();

	if (mirrorhard)
		return;
//	t = 0xF053>>(t<<2&12)&3^(t&12);
//	vnapage[0] = NTARAM+(t<<10&0x400);
//	vnapage[1] = NTARAM+(t<<9&0x400);
//	vnapage[2] = NTARAM+(t<<8&0x400);
//	vnapage[3] = NTARAM+(t<<7&0x400);

	if ((t-=MI_H)==0)
		vnapage[0] = vnapage[1] = NTARAM, vnapage[2] = vnapage[3] = NTARAM+0x400;
	else if ((t-=MI_V-MI_H)==0)
		vnapage[0] = vnapage[2] = NTARAM, vnapage[1] = vnapage[3] = NTARAM+0x400;
	else if ((t-=MI_0-MI_V)==0)
		vnapage[0] = vnapage[1] = vnapage[2] = vnapage[3] = NTARAM;
	else if ((t-=MI_1-MI_0)==0)
		vnapage[0] = vnapage[1] = vnapage[2] = vnapage[3] = NTARAM+0x400;
	PPUNTARAM = 15;
}

void SetupCartMirroring(int m,
	int hard,
	uint8 *extra)
{
	if (m<4)
	{
		mirrorhard = 0;
		setmirror(m);
	} else
	{
		vnapage[0] = NTARAM;
		vnapage[1] = NTARAM+0x400;
		vnapage[2] = extra;
		vnapage[3] = extra+0x400;
		PPUNTARAM = 15;
	}

	mirrorhard = hard;
}

void FCEU_SaveGameSave(CartInfo *LocalHWInfo)
{
	FILE		*sp;
	char		*soot;

	if (!LocalHWInfo->battery || !LocalHWInfo->SaveGame[0])
		return;
	if ((soot=FCEU_MakeFName(FCEUMKF_SAV,0,"sav"))==0)
		return;
	sp = FCEUD_UTF8fopen(soot,"wb");
	free(soot);
	if (sp)
	{
		int	x = 0;

		do
			if (LocalHWInfo->SaveGame[x])
				fwrite(LocalHWInfo->SaveGame[x],1,
					LocalHWInfo->SaveGameLen[x],sp);
		while(++x<4);
		fclose(sp);
	} else
		FCEU_PrintError("WRAM file \"%s\" cannot be written to.\n",soot);
}

void FCEU_LoadGameSave(CartInfo *LocalHWInfo)
{
	FILE		*sp;
	char		*soot;

	if (!LocalHWInfo->battery || !LocalHWInfo->SaveGame[0])
		return;

	if ((soot=FCEU_MakeFName(FCEUMKF_SAV,0,"sav"))==0)
		return;
	sp = FCEUD_UTF8fopen(soot,"rb");
	free(soot);
	if (sp)
	{
		int	x = 0;

		do
			if (LocalHWInfo->SaveGame[x])
				fread(LocalHWInfo->SaveGame[x],1,
					LocalHWInfo->SaveGameLen[x],sp);
		while(++x<4);
		fclose(sp);
	}
}
