/*
 cart.h

 * MacFCEU 0.8a2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

typedef struct CartInfo
{
	/* Set by mapper/board code: */
	void (*Power)(void);
	void (*Reset)(void);
	void (*Close)(void);
	uint8	*SaveGame[4];	/* Pointers to memory to save/load. */
	uint32	SaveGameLen[4];	/* How much memory to save/load. */

	/* Set by iNES/UNIF loading code. */
	int	mirror;
	int	battery;	/* Presence of an actual battery. */
	uint8	MD5[16];
	uint32	CRC32;		/* Used by mapper/board code. */
} CartInfo;

extern uint8 *Page[32], *VPage[8], **VPageR;
extern uint8 *MMC5SPRVPage[8], *MMC5BGVPage[8];

extern uint32 VROM_size, ROM_size;

extern uint8 PRGIsRAM[32];
extern uint8 *PRGptr[32], *CHRptr[32];
extern uint32 PRGsize[32], CHRsize[32];

extern uint32 PRGmask2[32];
extern uint32 PRGmask4[32];
extern uint32 PRGmask8[32];
extern uint32 PRGmask16[32];
extern uint32 PRGmask32[32];

extern uint32 CHRmask1[32];
extern uint32 CHRmask2[32];
extern uint32 CHRmask4[32];
extern uint32 CHRmask8[32];

void FCEU_SaveGameSave(CartInfo *LocalHWInfo);
void FCEU_LoadGameSave(CartInfo *LocalHWInfo);

void ResetCartMapping(void);
void SetupCartPRGMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram);
void SetupCartCHRMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram);
void SetupCartMirroring(int m,
	int hard,
	uint8 *extra);

//added; 2010.03.12
int NewCHRRAM(int w,
	uint32 z);
void DeleteCHRRAM(int w);
void EraseCHRRAM(int w,
	int32 v);

DECLFR(CartBR);
DECLFR(CartBROB);
DECLFW(CartBW);

/* faster CartBR calls; 2010.04.28 */
DECLFR(CartBR00); DECLFR(CartBR08);
DECLFR(CartBR10); DECLFR(CartBR18);
DECLFR(CartBR20); DECLFR(CartBR28);
DECLFR(CartBR30); DECLFR(CartBR38);
DECLFR(CartBR40); DECLFR(CartBR48);
DECLFR(CartBR50); DECLFR(CartBR58);
DECLFR(CartBR60); DECLFR(CartBR68);
DECLFR(CartBR70); DECLFR(CartBR78);
DECLFR(CartBR80); DECLFR(CartBR88);
DECLFR(CartBR90); DECLFR(CartBR98);
DECLFR(CartBRA0); DECLFR(CartBRA8);
DECLFR(CartBRB0); DECLFR(CartBRB8);
DECLFR(CartBRC0); DECLFR(CartBRC8);
DECLFR(CartBRD0); DECLFR(CartBRD8);
DECLFR(CartBRE0); DECLFR(CartBRE8);
DECLFR(CartBRF0); DECLFR(CartBRF8);

void FASTAPASS(2) SetCartRead(uint16 s,
	uint16 e);

#define setprg4(A,V)	setprg4r(0,A,V)
#define setprg8(A,V)	setprg8r(0,A,V)
#define setprg16(A,V)	setprg16r(0,A,V)
#define setprg32(A,V)	setprg32r(0,A,V)

void FASTAPASS(3) setprg4r(int r,
	unsigned int A,
	unsigned int V);
void FASTAPASS(3) setprg8r(int r,
	unsigned int A,
	unsigned int V);
void FASTAPASS(3) setprg16r(int r,
	unsigned int A,
	unsigned int V);
void FASTAPASS(3) setprg32r(int r,
	unsigned int A,
	unsigned int V);

//#define setchr1(A,V)	setchr1r(0,A,V)
#define setchr2(A,V)	setchr2r(0,A,V)
#define setchr4(A,V)	setchr4r(0,A,V)
#define setchr8(V)	setchr8r(0,V)

void FASTAPASS(2) setchr1(uint32 A,
	uint8 V);
void FASTAPASS(3) setchr1r(int r,
	unsigned int A,
	unsigned int V);
void FASTAPASS(3) setchr2r(int r,
	unsigned int A,
	unsigned int V);
void FASTAPASS(3) setchr4r(int r,
	unsigned int A,
	unsigned int V);
void FASTAPASS(2) setchr8r(int r,
	unsigned int V);

void FASTAPASS(2) setvram4(uint32 A, uint8 *p);
void FASTAPASS(1) setvram8(uint8 *p);

void FASTAPASS(3) setvramb1(uint8 *p,
	uint32 A,
	uint32 b);
void FASTAPASS(3) setvramb2(uint8 *p,
	uint32 A,
	uint32 b);
void FASTAPASS(3) setvramb4(uint8 *p,
	uint32 A,
	uint32 b);
void FASTAPASS(2) setvramb8(uint8 *p,
	uint32 b);

#define MI_H	0
#define MI_V	1
#define MI_0	2
#define MI_1	3

void FASTAPASS(1) setmirror(int t);
void setmirrorw(int a,
	int b,
	int c,
	int d);
void FASTAPASS(3) setntamem(uint8 *p,
	int ram,
	uint32 b);
