/*
 cheat.c

 * MacFCEU 0.6.1
 *
 * Copyright notice for this file:
 *  cheat.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "cheat.h"
#include "general.h"
#include "cart.h"
#include "memory.h"

#ifdef env68k
 #pragma segment cheat
#endif

static int readhexfields(const char *str,
	uint8 *d,
	int z,
	int f)
{
	const char	*s;
	int		t = 0;

	if ((s=str)==0 || d==0)
		return 0;

	do
	{
		unsigned	b, c;

		if ((b=(unsigned int)*((const unsigned char*)s)++-'0'<<4)>0x90)
			if ((b-='A'-'0'-10<<4)-160>'F'-'A'<<4
					&& (b+='A'-'a'<<4)-160>'F'-'A'<<4)
				return 0;
		if ((c=(unsigned int)*((const unsigned char*)s)++-'0')>9)
			if ((c-=('A'-'0'-10))-10>('F'-'A')
					&& (c+=('A'-'a'))-10>('F'-'A'))
				return 0;
		c|= b,*d++ = c;
		if (f&1)
			while(*s==' ' || *s==':')
				++s;
	} while(++t,f>>=1,--z!=0);

	return t;
}

uint8		*CheatRPtrs[64];

void FCEU_CheatResetRAM(void)
{
	uint8		**d = CheatRPtrs;
	int		c = sizeof(CheatRPtrs)/sizeof(CheatRPtrs[0]);

	do
		*d++ = 0;
	while(--c!=0);

	X6502_DebugLevel(~2);
}

void FCEU_CheatSetRAM(int c,
	uint32 A,
	uint8 *p) /* will clear if NULL */
{
	uint8		**d;

	if (c<0)
		return;
	d = CheatRPtrs+(A>>10), p-= A;
	do
		*d++ = p;
	while(--c!=0);
}

typedef struct CHEATF
{
	struct CHEATF	*next;
	char	*name;

	uint16	addr;
	uint8	val;
	int	compare; /* -1 for no compare. */
	int	type;    /* 0 for replace, 1 for substitute(GG). */
	int	status;
} CHEATF;

typedef struct CHEATF_SUBFAST
{
	uint16		addr;
	uint8		val;
	int		compare;
	readfunc	PrevRead;
} CHEATF_SUBFAST;


static CHEATF_SUBFAST	SubCheats[256];

static int	numsubcheats = 0;
CHEATF		*cheats = 0, *cheatsl = 0;

#define CHEATC_NONE	0x8000
#define CHEATC_EXCLUDED	0x4000
#define CHEATC_NOSHOW	0xC000

static uint16	*CheatComp = 0;
static uint8	savecheats;

static DECLFR(SubCheatsRead)
{
	CHEATF_SUBFAST	*s;
	int		x = numsubcheats;
	uint8		pv;

	if ((s=SubCheats)==0)
		goto end;
	do
		if (s->addr==A)
			if (s->compare>=0)
				if ((pv=s->PrevRead(A))==s->compare)
					return s->val;
				else
					return pv;
			else
				return s->val;
	while(++s,--x!=0);
end:
	return 0; /* We should never get here. */
}

void RebuildSubCheats(void)
{
	CHEATF	*c;
	int	x;

	for(x=0; x<numsubcheats; x++)
		SetReadHandler(SubCheats[x].addr,SubCheats[x].addr,SubCheats[x].PrevRead);

	X6502_DebugLevel(~2);

	numsubcheats = 0;
	if ((c=cheats)==0)
		return;
	do
		if (c->type==1 && c->status)
		{
			unsigned	A = c->addr;

			if (GetReadHandler(A)==SubCheatsRead)
			{
				/* Prevent a catastrophe by this check. */
				//FCEU_DispMessage("oops");
			} else
			{
				SubCheats[numsubcheats].PrevRead = GetReadHandler(A);
				SubCheats[numsubcheats].addr = A;
				SubCheats[numsubcheats].val = c->val;
				SubCheats[numsubcheats].compare = c->compare;
				if ((A&0x1FFF)==0) /* RAM cheat */
					X6502_DebugLevel(2);
				SetReadHandler(A,A,SubCheatsRead);
				numsubcheats++;
			}
		}
	while((c=c->next)!=0);
}

void FCEU_PowerCheats()
{
	numsubcheats = 0; /* Quick hack to prevent setting of ancient read addresses. */
	RebuildSubCheats();
}

static void CheatMemErr(void)
{
	FCEUD_PrintError("Error allocating memory for cheat data.");
}

/* This function doesn't allocate any memory for "name" */
static int AddCheatEntry(char *name,
	uint32 addr,
	uint8 val,
	int compare,
	int status,
	int type)
{
	CHEATF		*ent;

	if ((ent=(CHEATF*)malloc(sizeof(CHEATF)))==0)
	{
		CheatMemErr();
		return 0;
	}
	ent->name = name;
	ent->addr = addr;
	ent->val = val;
	ent->status = status;
	ent->compare = compare;
	ent->type = type;
	ent->next = 0;

	if (cheats)
		cheatsl->next = cheatsl = ent;
	else
		cheats = cheatsl = ent;

	return 1;
}

void FCEU_LoadGameCheats(FILE *override)
{
	uint8		boo[2+1+1];
	FILE		*fp = 0;
	unsigned int	status;
	unsigned int	type;
	int		x;

	char		linebuf[2048];
	char		*namebuf, *fn;
	int		tc = 0;

	numsubcheats = savecheats = 0;

	if (override)
		fp = override;
	else
	{
		if ((fn=FCEU_MakeFName(FCEUMKF_CHEAT,0,0))==0)
			fp = FCEUD_UTF8fopen(fn,"rb");
		free(fn);
		if (!fp)
			return;
	}

	while(fgets(linebuf,2048,fp)>0)
	{
		char	*tbuf = linebuf;
		int	doc = 0;

		*(int32*)boo = status = type = 0;

		if (*tbuf=='S')
			++tbuf, type = 1;
		else
			type=0;

		if (*tbuf=='C')
		{
			++tbuf, doc = 1;
		}

		if (*tbuf==':')
			++tbuf, status = 0;
		else
			status = 1;

		if (doc)
		{
			char	*neo = &tbuf[4+2+2+1+1+1];
			/*if(sscanf(tbuf,"%04x%*[:]%02x%*[:]%02x",&addr,&val,&compare)!=3)*/
			if (readhexfields(tbuf,boo,3,2|4)!=3)
				continue;
			namebuf=malloc(strlen(neo)+1);
			strcpy(namebuf,neo);
		} else
		{
			char	*neo = &tbuf[4+2+1+1];
			/*if(sscanf(tbuf,"%04x%*[:]%02x",&addr,&val)!=2)*/
			if (readhexfields(tbuf,boo,3,2)!=2)
				continue;
			namebuf=malloc(strlen(neo)+1);
			strcpy(namebuf,neo);
		}

		for(x=0; x<strlen(namebuf); ++x)
			if (namebuf[x]==10 || namebuf[x]==13)
			{
				namebuf[x] = 0;
				break;
			} else if (namebuf[x]<0x20)
				namebuf[x] = ' ';

		AddCheatEntry(namebuf,(boo[0]<<8)|boo[1],boo[2],
			doc ? boo[3]:-1,status,type);
		++tc;
	}
	RebuildSubCheats();

	if (!override)
		fclose(fp);
}

void FCEU_FlushGameCheats(FILE *override,
	int nosave)
{
	if (CheatComp)
		{free(CheatComp); CheatComp = 0;}
	/* Always save cheats if we're being overridden. */
	if ((!savecheats || nosave) && !override)
	{
		if (cheats)
		{
			CHEATF	*next = cheats;

			while(next)
			{
				CHEATF	*last = next;
				next = next->next;
				free(last->name);
				free(last);
			}
			cheats = cheatsl = 0;
		}
	} else
	{
		char	*fn = 0;
	
		if (!override)
			fn = FCEU_MakeFName(FCEUMKF_CHEAT,0,0);
		if (cheats)
		{
			CHEATF	*next = cheats;
			FILE	*fp = 0;

			if (override)
				fp = override;
			else if (fn)
				fp = FCEUD_UTF8fopen(fn,"wb");
			if (fp)
			{
				for(;;)
				{
					CHEATF	*t;

					if (next->type)
						fputc('S',fp);
					if (next->compare>=0)
						fputc('C',fp);
					if(!next->status)
						fputc(':',fp);
					if (next->compare>=0)
						fprintf(fp,"%04x:%02x:%02x:%s\n",
							next->addr,
							next->val,
							next->compare,
							next->name);
					else
						fprintf(fp,"%04x:%02x:%s\n",
							next->addr,
							next->val,
							next->name);
	
					free(next->name);
					t = next;
					next = next->next;
					free(t);
					if (!next)
						break;
				}
				if (!override)
					fclose(fp);
			} else
				FCEUD_PrintError("Error saving cheats.");
			cheats = cheatsl = 0;
		} else if (!override)
			remove(fn);
		if (!override)
			free(fn);
	}

	RebuildSubCheats();  /* Remove memory handlers. */
}

int FCEUI_AddCheat(const char *name, uint32 addr, uint8 val, int compare, int type)
{
	char	*t;

	if ((t=(char*)malloc(strlen(name)+1))==0)
	{
		CheatMemErr();
		return 0;
	}
	strcpy(t,name);
	if (!AddCheatEntry(t,addr,val,compare,1,type))
	{
		free(t);
		return 0;
	}
	savecheats = 1;
	RebuildSubCheats();

	return 1;
}

int FCEUI_DelCheat(uint32 which)
{
	CHEATF		*prev, *cur;
	uint32		x = 0;

	for(prev=0,cur=cheats; cur; prev=cur,cur=prev->next,++x)
		if (x==which)	// remove
		{
			if (prev) // in chain; update pointer
				if (cur->next)	// More
					prev->next = cur->next;
				else		// Set previous as last.
					prev->next = 0, cheatsl = prev;
			else if (cur->next)	// More
				cheats = cur->next;
			else			// No more
				cheats = cheatsl = 0;
			// Now that all references to this cheat are removed,
			//  free the memory.
			free(cur->name);
			free(cur);
			savecheats = 1;
			RebuildSubCheats();
			return 1;
		}

	return 0;
}

void FCEU_ApplyPeriodicCheats(void)
{
	CHEATF		*cur;
	uint8		*d;

	if ((cur=cheats)==0)
		return;
	do
		if (cur->status && !cur->type)
			if ((d=CheatRPtrs[cur->addr>>10])!=0)
				d[cur->addr] = cur->val;
	while((cur=cur->next)!=0);
}


void FCEUI_ListCheats(int (*callb)(char *name, uint32 a, uint8 v, int compare, int s,
		int type, void *data),
	void *data)
{
	CHEATF		*next = cheats;

	if ((next=cheats)==0)
		return;
	do
		if (!callb(next->name,next->addr,next->val,next->compare,
				next->status,next->type,data))
			break;
	while((next=next->next)!=0);
}

int FCEUI_GetCheat(uint32 which,
	char **name,
	uint32 *a,
	uint8 *v,
	int *compare,
	int *s,
	int *type)
{
	CHEATF		*next;
	uint32		x = 0;

	if ((next=cheats)==0)
		return 0;

	do
		if (x==which)
		{
			if (name)
				*name = next->name;
			if (a)
				*a = next->addr;
			if (v)
				*v = next->val;
			if (s)
				*s = next->status;
			if (compare)
				*compare = next->compare;
			if (type)
				*type = next->type;
			return 1;
		}
	while(++x,(next=next->next)!=0);

	return 0;
}

static int GGtobin(char c)
{
	char		*lets = "APZLGITYEOXUKSVN";
	int		x;

	x = 0;
	do
		if (lets[x]==toupper(c))
			return x;
	while(++x<16);

	return 0;
}

/* Returns 1 on success, 0 on failure. Sets *a,*v,*c. */
int FCEUI_DecodeGG(const char *str,
	uint16 *a,
	uint8 *v,
	int *c)
{
	int		s, C = 0;
	uint16		A = 0x8000;
	uint8		V = 0;
	uint8		t;

	if (str==0)
		return 0;
	s = strlen(str);
	if (s!=6 && s!=8)
		return 0;

	t = GGtobin(*str++);
	V|= t&7;
	V|= (t&8)<<4;

	t = GGtobin(*str++);
	V|= (t&7)<<4;
	A|= (t&8)<<4;

	t = GGtobin(*str++);
	A|= (t&7)<<4;
	//if(t&0x08) return(0); /* 8-character code?! */

	t = GGtobin(*str++);
	A|= (t&7)<<12;
	A|= t&8;

	t = GGtobin(*str++);
	A|= t&7;
	A|= (t&8)<<8;

	if (s==6)
	{
		t = GGtobin(*str++);
		A|= (t&7)<<8;
		C = -1;
		goto got;
	} else
	{
		t = GGtobin(*str++);
		A|= (t&7)<<8;
		C|= t&8;

		t = GGtobin(*str++);
		C|= t&7;
		C|= (t&8)<<4;

		t = GGtobin(*str++);
		C|= (t&7)<<4;
	got:
		V|= t&8;
		if (a)
			*a = A;
		if (v)
			*v = V;
		if (c)
			*c = C;
		return 1;
	}

	return 0;
}

int FCEUI_DecodePAR(const char *str,
	uint16 *a,
	uint8 *v,
	int *c,
	int *type)
{
	uint8	boo[4];

/* removed any refs to external code for this function; 2009.09.21 */
/*sscanf(str,"%02x%02x%02x%02x",boo,boo+1,boo+2,boo+3);*/
	if (readhexfields(str,boo,4,0)!=4)
		return 0;

	if (c)
		*c = -1;
#if 1
	if (a)
		*a = (boo[3]<<8)|(boo[2]+0x7F);
	if (v)
		*v = 0;
#else
	if (a)
		*a = boo[2]|(boo[1]<<8);
	if (v)
		*v = boo[3];
#endif
	/* Zero-page addressing modes don't go through the normal read/write handlers
		in FCEU, so we must do the old hacky method of RAM cheats.
	*/
	if (type)
		if (*a<0x0100)
			*type = 0;
		else
			*type = 1;

	return 1;
}

/* name can be NULL if the name isn't going to be changed. */
/* same goes for a, v, and s(except the values of each one must be <0) */

int FCEUI_SetCheat(uint32 which,
	const char *name,
	int32 a,
	int32 v,
	int compare,
	int s,
	int type)
{
	CHEATF		*next;
	uint32		x = 0;

	for(next=cheats; next; next=next->next,++x)
		if (x==which)
		{
			if (name)
			{
				char	*t;

				if ((t=(char*)realloc(next->name,strlen(name+1)))==0)
					return 0;
				strcpy(next->name=t,name);
			}
			if (a>=0)
				next->addr = a;
			if (v>=0)
				next->val = v;
			if (s>=0)
				next->status = s;
			next->compare = compare;
			next->type = type;
	
			savecheats = 1;
			RebuildSubCheats();
			return 1;
		}

	return 0;
}

/* Convenience function. */
int FCEUI_ToggleCheat(uint32 which)
{
	CHEATF		*next;
	uint32		x;

	for(next=cheats,x=0; next; next=next->next,++x)
		if (x==which)
		{
			next->status = !next->status;
			savecheats = 1;
			RebuildSubCheats();
			return next->status;
		}

	return -1;
}

static int InitCheatComp(void)
{
	uint32		x;

	if ((CheatComp=(uint16*)malloc(65536*sizeof(uint16)))==0)
	{
		CheatMemErr();
		return 0;
	}

	for(x=0; x<65536; ++x)
		CheatComp[x] = CHEATC_NONE;

	return 1;
}

void FCEUI_CheatSearchSetCurrentAsOriginal(void)
{
	uint32		x;
	uint8		*d;

	for (x=0; x<0x10000; ++x)
		if (!(CheatComp[x]&CHEATC_NOSHOW))
			if ((d=CheatRPtrs[x>>10])!=0)
				CheatComp[x] = d[x];
			else
				CheatComp[x]|= CHEATC_NONE;
}

void FCEUI_CheatSearchShowExcluded(void)
{
	uint32		x;

	for(x=0; x<0x10000; ++x)
		CheatComp[x]&=~CHEATC_EXCLUDED;
}

int32 FCEUI_CheatSearchGetCount(void)
{
	uint32		x;
	int32		c;

	if (CheatComp==0)
		return -1; /* returned 0 before; 2010.03.31 */
	for(x=0,c=0; x<0x10000; ++x)
		if (!(CheatComp[x]&CHEATC_NOSHOW) && CheatRPtrs[x>>10])
			++c;

	return c;
}

//This function'll give the initial value of the search and current value at a location.
void FCEUI_CheatSearchGet(int (*callb)(uint32 a, uint8 last, uint8 current, void *data),
	void *data)
{
	const uint8	*d;
	uint32		x;

	if (!CheatComp)
	{
		InitCheatComp();
		return;
	}

	for(x=0; x<0x10000; ++x)
		if (!(CheatComp[x]&CHEATC_NOSHOW) && (d=CheatRPtrs[x>>10])!=0)
			if (!callb(x,CheatComp[x],d[x],data))
				break;
}

void FCEUI_CheatSearchGetRange(uint32 first,
	uint32 last,
	int (*callb)(uint32 a, uint8 last, uint8 current))
{
	const uint8	*d;
	uint32		x;
	uint32		cin = 0; //was `in` (conflicting C++ term); 2010.03.31

	if (!CheatComp)
	{
		InitCheatComp();
		return;
	}

	for(x=0; x<0x10000; ++x)
		if (!(CheatComp[x]&CHEATC_NOSHOW) && (d=CheatRPtrs[x>>10])!=0)
		{
			if (cin>=first)
				if (!callb(x,CheatComp[x],d[x]))
					break;
			if (++cin>last)
				return;
		}
}

void FCEUI_CheatSearchBegin(void)
{
	const uint8	*d;
	uint32		x;

	if (!CheatComp)
		if (!InitCheatComp())
		{
			//CheatMemErr(); --this would be called twice!
			return;
		}
	for(x=0; x<0x10000; ++x)
		if ((d=CheatRPtrs[x>>10])!=0)
			CheatComp[x] = d[x];
		else
			CheatComp[x] = CHEATC_NONE;
}

//NOTE: the type order is different between this and that of FCEUX
// FCEUX does "known" for 4, then "greater," "lesser," "greater by known"...
void FCEUI_CheatSearchEnd(int type,
	uint8 v1,
	uint8 v2)
{
	uint32		x, t;

	if (!CheatComp)
		if (!InitCheatComp())
			return;

#define CheatSearchIterator \
 for(x=0;x<0x10000;++x) \
  if(!(CheatComp[x]&CHEATC_NOSHOW))

	if (type==0)		//0: Change to a specific value.
	{
		CheatSearchIterator
			if (CheatComp[x]!=v1 || CheatRPtrs[x>>10][x]!=v2)
				CheatComp[x]|= CHEATC_EXCLUDED;
	} else if (--type==0)	//1: Search for relative change (between values).
	{
		CheatSearchIterator
			if (CheatComp[x]!=v1 || ((t=CheatComp[x]-CheatRPtrs[x>>10][x])<0
					? -t:t)!=v2)
				CheatComp[x]|= CHEATC_EXCLUDED;
	} else if (--type==0)	//2: Purely relative change.
	{
		CheatSearchIterator
			if (((t=CheatComp[x]-CheatRPtrs[x>>10][x])<0 ?
					-t:t)!=v2)
				CheatComp[x]|= CHEATC_EXCLUDED;
	} else if (--type==0)	//3: Any change.
	{
		CheatSearchIterator
			if (CheatComp[x]==CheatRPtrs[x>>10][x])
				CheatComp[x]|= CHEATC_EXCLUDED;
	} else if (--type==0)	//4: Value decreased.
	{
		CheatSearchIterator
			if (CheatRPtrs[x>>10][x]>=CheatComp[x])
				CheatComp[x]|= CHEATC_EXCLUDED;
	} else if (--type==0)	//5: Value increased.
	{
		CheatSearchIterator
			if (CheatRPtrs[x>>10][x]<=CheatComp[x])
				CheatComp[x]|= CHEATC_EXCLUDED;
		//FCEUI_CheatSearchSetCurrentAsOriginal();
	} else if (--type==0)	//6: Greater than by known.
	{
		CheatSearchIterator
			if ((CheatRPtrs[x>>10][x]-CheatComp[x])!=v2)
				CheatComp[x]|= CHEATC_EXCLUDED;
	} else if (--type==0)	//7: Lesser than by known.
	{
		CheatSearchIterator
			if ((CheatComp[x]-CheatRPtrs[x>>10][x])!=v2)
				CheatComp[x]|= CHEATC_EXCLUDED;
	} else if (--type==0)	//8: Known.
	{
		CheatSearchIterator
			if (CheatRPtrs[x>>10][x]!=v1)
				CheatComp[x]|= CHEATC_EXCLUDED;
	}
#undef CheatSearchIterator
}
