/*
 debug.h

 * MacFCEU 0.8a7
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __DEBUG_H

/* ---- constants */

/* soft breakpoint opcodes */
#define subsopcode_breakpoint	0x92 /* STX line, rel. mode (65816: STA-ind) */

/* breakpoint type flags/attributes */
#define BPOINT_READ		0x01
#define BPOINT_WRITE		0x02
#define BPOINT_PC		0x04
#define BPOINT_SOFT		0x40 /* don't write to code */
#define BPOINT_ONCE		0x80 /* clear breakpoint when hit */

/* ---- typedefs */

typedef struct bpoint/*(16+)*/
{
	uint16		address;
	uint8		original, flags;
	char		*expression; /* only break if parse_parameter(..)!=0 */
	readfunc	olda;
	writefunc	oldb;
} bpoint;

typedef int (*bpfunc)(int32 index,
	int32 validcount,
	const bpoint*);

/* ---- globals */

extern uint8 breakpoint_bitmap[0x10000>>3]; /* use _bittst (general.h) */

/* ---- prototypes */

int32 DebugReadAddr(int32 a);

/* platform needs to define this to use BPBreak */
const char *FCEUD_getvalue(const char *w,
	const char *e,
	int32 *i,
	void *u);

int BPList(int32,
	int32,
	const bpoint*);
int BPBreak(int32,
	int32,
	const bpoint*);

void FCEUI_DumpMem(const char *fn,
	uint32 start,
	uint32 end);
void FCEUI_LoadMem(const char *fn,
	uint32 start,
	int hl);

/* called when a soft code breakpoint or illegal opcode */
int FCEUI_DebuggerBreak(int instruction); /* pass negative value for no instruction */

void FCEUI_MemDump(uint16 a,
	int32 len,
	void (*callb)(uint16 a, uint8 v));
uint8 FCEUI_MemSafePeek(uint16 A);
void FCEUI_MemPoke(uint16 a,
	uint8 v,
	int hl);

void DebugGenSetByte(unsigned a,
	unsigned char v);

int DebugGetAddrType(uint16 a,
	char *name); /* NULL or C-string of 31 char capacity */

/* MacFCEU breakpoints */
int32 GetBreakPointCount(void);
int32 DebugBreakPointsReset(int32 index, /* single, or negative for all */
	int set);
int DebugBreakPointsLoop(int32 index, /* single, or negative for all */
	uint8 targetflags,
	bpfunc callb);
int32 DebugBreakPointsClear(int32 index); /* single, or negative for all */
int32 DebugBreakPointGet(uint16 address);
int32 DebugBreakPointSet(uint16 address,
	uint8 flags,
	const char *expression); /* a copy is made */

/* these used to be in driver.h */

#define FCEUI_IRQ()		X6502_IRQBegin(FCEU_IQTEMP)

void FCEUI_GetIVectors(uint16 *reset,
	uint16 *irq,
	uint16 *nmi);

/* disassembler extensions (prev. in macos/debugger.h) */

/* print P flag letters, capital for on */
int DebugPrintP(char *d, /* should buffer >=9 */
	int u); /* print unused flags (U and D) */

/* trace flag bit constants, in order */
#define tfaddCPU_A		1
#define tfaddCPU_X		2
#define tfaddCPU_Y		4
#define tfaddCPU_P		8
#define tfaddCPU_S		0x10
#define tfaddCPU_intlow		0x20
#define tfaddCPU_CYC		0x40
#define tfaddPPU_scanline	0x80
#define tfaddstack8		0x100	/*8 bytes*/
#define tftrace_always		0x8000
#define tfaddnewline		0x80000000u

uint16 DebugDisassemble65xx(const struct X6502struct *f, /* can be NULL */
	uint16 PC,
	int instruction_count,
	int (*routine)(const char*, /* can be NULL (FCEUD_Message) */
		int i),
	uint32 traceflags);

#define __DEBUG_H
#endif
