/*
 filter.c
  sound filter

 * MacFCEU 0.8a9
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <stdio.h>

#include "types.h"
#include "x6502.h"
#include "FCEU.h"
#include "fsound.h"
#include "filter.h"

#include "ftrig.h" /* for ufract32mul */

#ifdef env68k
 #pragma segment Sound
#endif

#if HAVE_HQ
 #include "fcoeffs.h"
#endif

#if HAVE_HQ
static uint32	mrindex;
static uint32	mrratio;
#endif

static void SexyFilter2(int16 *in, int32 count)
{
	static int64	acc = _int64xval(0);
#ifdef moo
	double		x, p;
	int64		c;

	x = 2*M_PI*6000/FSettings.SndRate;
	p = (double)2-cos(x)) - sqrt(pow((double)2-cos(x),2) -1;

	_int64xset(c,p*0x100000);
//printf("%f\n",(double)c/0x100000);
#endif

	if (count<=0)
		return;
	do
	{
		int64	dropcurrent; /*=((*in<<16)-acc)>>3;*/

		int64setl(dropcurrent,*in<<16);
		(void)int64sub(dropcurrent,acc);
		int64shr(dropcurrent,3);
		/**in++=(acc+=dropcurrent)>>16;*/
		(void)int64add(acc,dropcurrent);
		*in++ = int64lo(acc)>>16;
//acc=((int64)0x100000-c)* *in + ((c*acc)>>20);
//*in=acc>>20;
//in++;
	} while(--count!=0);
}

int64		sfacc1 = _int64xval(0), sfacc2 = _int64xval(0);

void SexyFilter(int32 *in,
	int16 *out,
	int32 count)
{
	int32		mul1, mul2, vmul;

//vmul=(FSettings.SoundVolume<<16)*3/4/100;
//if(FSettings.soundq) vmul/=4;
//else vmul*=2; /* TODO:  Increase volume in low quality sound rendering code itself */

	if (count<=0)
		return;
	mul1 = FSettings.SndRateF; /* (94<<24)/Rate for (94<<16)/ and (24<<16)/ */
	mul2 = ufract32mul(mul1,(4294967296.0*24*2+93)/94)>>8;
	mul1>>= 7;
	vmul = FSettings.SndVolF;
#if HAVE_HQ
	if (FSettings.soundq)
		vmul>>= 3;
#endif
	do
	{
		int64	ino, t;/*ino=(int64)*in*vmul;*/
		(void)muls64s32(ino,*in,vmul);

//sfacc1+=((ino-sfacc1)*mul1)>>16;
//sfacc2+=((ino-sfacc1-sfacc2)*mul2)>>16;
		t = ino; (void)int64sub(t,sfacc1); (void)int64muls32(t,mul1);
		int64shr(t,16); (void)int64add(sfacc1,t);
		t = ino; (void)int64sub(t,sfacc1); (void)int64sub(t,sfacc2); (void)int64muls32(t,mul2);
		int64shr(t,16); (void)int64add(sfacc2,t);

//t=(sfacc1-ino+sfacc2)>>16;
  //if(t>32767 || t<-32768) printf("Flow: %d\n",t);
//if(t>32767) t=32767;
//if(t<-32768) t=-32768;
		t = sfacc1; (void)int64sub(t,ino); (void)int64add(t,sfacc2);
		*in++ = 0;
		*out++ = int64toi(t)>>16;
	} while(--count!=0);
}

#if HAVE_HQ
/* Returns number of samples written to out. */
/* leftover is set to the number of samples that need to be copied 
   from the end of in to the beginning of in.
*/

//static uint32 mva=1000;

/* This filtering code assumes that almost all input values stay below 32767.
   Do not adjust the volume in the wlookup tables and the expansion sound
   code to be higher, or you *might* overflow the FIR code.
*/

/* this part appears to be broken */

#define interpolate64(a,b,x)	do{int64 t,u; \
	(void)muls64s32(t,a,65536+-(x&65535)); \
	(void)muls64s32(u,b,x&65535); (void)int64add(t,u); \
	a=_int64shr32(t,16+11);}while(0)

int32 NeoFilterSound(int32 *in,
	int16 *out,
	uint32 inlen,
	int32 *leftover)
{
	uint32		x, max;
	int16		*outsave = out;
	int32		count = 0;

//	for(x=0;x<inlen;x++)
//	{
//		if(in[x]>mva){ mva=in[x]; printf("%ld\n",in[x]);}
//	}

	max = (inlen-1)<<16;

	if (FSettings.soundq==2)
		for(x=mrindex; x<max; x+=mrratio,++count)
		{
			int32		*S, *D, acc = 0, acc2 = 0;
			unsigned int	c;

			S = &in[(x>>16)-SQ2NCOEFFS], D = sq2coeffs;
			if((c=SQ2NCOEFFS)!=0)
				do
					acc+= (*D*S[c])>>6,
					acc2+= (*D++*S[1+c])>>6;
				while(--c!=0);

 //acc=((int64)acc*(65536-(x&65535))+(int64)acc2*(x&65535))>>(16+11);
			interpolate64(acc,acc2,x);
			*out++ = acc;
		}
	else
		for(x=mrindex; x<max; x+=mrratio,++count)
		{
			int32		*S, *D, acc = 0, acc2 = 0;
			unsigned int	c;
		
			S = &in[(x>>16)-NCOEFFS], D = coeffs;
			if ((c=NCOEFFS)!=0)
				do
					acc+= (*D*S[c])>>6,
					acc2+= (*D++*S[1+c])>>6;
				while(--c!=0);
			interpolate64(acc,acc2,x);
			*out++ = acc;
		}

	mrindex = x-max;

	if (FSettings.soundq==2)
		mrindex+= SQ2NCOEFFS*65536, *leftover = SQ2NCOEFFS+1;
	else
		mrindex+= NCOEFFS*65536, *leftover = NCOEFFS+1;

	if (GameExpSound.NeoFill)
		GameExpSound.NeoFill(outsave,count);

/*SexyFilter(outsave,outsave,count);*/
	if (count>0)
	{
		int32 mul1 = FSettings.SndRateF>>7;
		int32 mul2 = ufract32mul(FSettings.SndRateF,(4294967296.0*24*2+93)/94)>>8;
		int32 vmul = FSettings.SndVolF>>3;
		int32 c = count;

		out = outsave;
		do
		{
			int64	ino, t;
			(void)muls64s32(ino,*out,vmul);

			t=ino; (void)int64sub(t,sfacc1); (void)int64muls32(t,mul1);
			int64shr(t,16); (void)int64add(sfacc1,t);
			t=ino; (void)int64sub(t,sfacc1); (void)int64sub(t,sfacc2);
			(void)int64muls32(t,mul2);
			int64shr(t,16); (void)int64add(sfacc2,t);

			t=sfacc1; (void)int64sub(t,ino); (void)int64add(t,sfacc2);
			*out++ = int64toi(t)>>16;
		} while(--c!=0);
	}

	if (FSettings.lowpass)
		SexyFilter2(outsave,count);
	return count;
}

void MakeFilters(int32 rate)
{
	int32 *tabs[6] = {C44100NTSC,C44100PAL,C48000NTSC,C48000PAL,
		C96000NTSC,C96000PAL};
	int32 *sq2tabs[6] = {SQ2C44100NTSC,SQ2C44100PAL,SQ2C48000NTSC,SQ2C48000PAL,
		SQ2C96000NTSC,SQ2C96000PAL};

	int32	*tmp, x;
	uint32	nco;

	if (FSettings.soundq==2)
		nco = SQ2NCOEFFS;
	else
		nco = NCOEFFS;

	mrindex = (nco+1)<<16;
//mrratio=(PAL?(int64)(PAL_CPU*65536):(int64)(NTSC_CPU*65536))/rate;
	{
		uint64	t;
		if (PAL)
			_int64xset(t,PAL_CPU*65536);
		else
			_int64xset(t,NTSC_CPU*65536);
		mrratio = uint64divs32(t,rate);
	}

	if (FSettings.soundq==2)
		tmp = sq2tabs[PAL|(rate==48000 ? 2:rate==96000 ? 4:0)];
	else
		tmp = tabs[PAL|(rate==48000 ? 2:rate==96000 ? 4:0)];

	if (FSettings.soundq==2)
		for(x=0; x<SQ2NCOEFFS>>1; ++x)
			sq2coeffs[x] = sq2coeffs[SQ2NCOEFFS-1-x] = tmp[x];
	else
		for(x=0; x<NCOEFFS>>1; ++x)
			coeffs[x] = coeffs[NCOEFFS-1-x] = tmp[x];
#ifdef MOO
 /* Some tests involving precision and error. */
	{
		static int64	acc = 0;
		int	x;

		for(x=0; x<SQ2NCOEFFS; ++x)
			acc+= (int64)32767*sq2coeffs[x];
		printf("Foo: %lld\n",acc);
	}
#endif
}
#endif
