/*
 general.c
  general file system rountines, whereabouts

 * MacFCEU 0.6.6
 *
 * Copyright notice for this file:
 *  general.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>

#ifndef mac
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
#endif

#include "driver.h"
#include "FCEU.h"
#include "general.h"
#include "state.h"
#include "movie.h"

#include "utility/MD5.h"

#ifdef env68k
 #pragma segment files
#endif

#ifdef mac
 #define PATHMAXLENGTH			1024
#else
 #define PATHMAXLENGTH			2048
#endif

static char	BaseDirectory[PATHMAXLENGTH];
static char	FileBase[PATHMAXLENGTH];
static char	FileExt[PATHMAXLENGTH]; /* includes suffix type */

static char	FileBaseDirectory[PATHMAXLENGTH];


void FCEUI_SetBaseDirectory(char *dir)
{
	strncpy(BaseDirectory,dir,sizeof(BaseDirectory)-1);
	BaseDirectory[sizeof(BaseDirectory)-1] = 0;
}

static const char	*odirs[FCEUIOD__COUNT] = {0,0,0,0,0,0};	// odirs, odors. ^_^

void FCEUI_SetDirOverride(int which,
	char *n)
{
	odirs[which] = n;

	if (FCEUGameInfo)	/* Rebuild cache of present states/movies. */
		if (which==FCEUIOD_STATE)
			FCEUSS_CheckStates();
		else if (which==FCEUIOD_MOVIE)
			FCEUMOV_CheckMovies();
}

#ifndef HAVE_ASPRINTF
static int asprintf(char **strp,
	const char *fmt,
	...)
{
	va_list		ap;
	int		ret;

	va_start(ap,fmt);
	if (strp==0)
		return 0;
 #ifdef mac
	if ((*strp=malloc(2560))==0)
		return 0;
	ret = vsprintf(*strp,fmt,ap); /* overflow would be fatal */
 #else
	if ((*strp=malloc(2048))==0)
		return 0;
	ret = vsnprintf(*strp,2048,fmt,ap);
 #endif
	va_end(ap);
	return ret;
}
#endif

int CheckFileOrDirExists(const char *path)
{
#ifdef mac
	FILE	*tmpfile; /* a clearly stupid method */

	if (path==NULL || (tmpfile=fopen(path,"rb"))==0)
	{
		if (errno==EISDIR)
			return 2;
		return 0;
	}
	fclose(tmpfile);
	return 1;
#else
	struct stat tmpstat;

	if (path==0)
		return 0;
	return stat(path,&tmpstat)!=-1; /* dir- wha? */
#endif
}

#ifdef mac
 /* directory and file names prior to HFS+ are limited to 31 characters */
 /* file base is be capped to 26 for a 4-char extension and dot */
 #define PSSFB PSS"%.26s"
 #define PSSFBX PSS"%.19s" /* when including an extra index */
#else
 #define PSSFB PSS"%.999s"
 #define PSSFBX PSS"%.999s"
#endif

char *FCEU_MakeFName(int type,
	int id1,
	const char *cd1) /* as of 2010.05.16, can be NULL */
{
	char		tempstr[512];
	char		*ret = 0;

	switch(type)
	{
	case FCEUMKF_DEFAULT:
		if ((ret=malloc(strlen(FileBase)+1))!=0)
			strcpy(ret,FileBase);
	default:
		break;
	case FCEUMKF_NPTEMP:
		asprintf(&ret,"%s"PSS"m590plqd94fo.tmp",
			BaseDirectory);
		break;
	case FCEUMKF_MOVIE:
		sprintf(tempstr,"%.450s.%s.",
			FileBase,md5_asciistr(FCEUGameInfo->MD5,1));
		if (odirs[FCEUIOD_STATE])
			asprintf(&ret,"%s"PSSFBX"%d.fcm",
				odirs[FCEUIOD_STATE],tempstr,id1);
		else
			asprintf(&ret,"%s"PSS"fcs"PSSFBX"%d.fcm",
				BaseDirectory,tempstr,id1);
		break;
	case FCEUMKF_STATE:
		sprintf(tempstr,"%.450s.%s",
			FileBase,md5_asciistr(FCEUGameInfo->MD5,1));
		if (odirs[FCEUIOD_STATE])
			asprintf(&ret,"%s"PSSFB".fc%d",
				odirs[FCEUIOD_STATE],tempstr,id1);
		else
			asprintf(&ret,"%s"PSS"fcs"PSSFB".fc%d",
				BaseDirectory,tempstr,id1);
		break;
	case FCEUMKF_SNAP:
		if (FSettings.SnapName)
		{
			sprintf(tempstr,"%.450s-%d",
				FileBase,id1);
			if (odirs[FCEUIOD_SNAPS])
				asprintf(&ret,cd1 ?
					"%s"PSSFB".%s":
					"%s"PSSFB,
					odirs[FCEUIOD_SNAPS],tempstr,cd1);
			else
				asprintf(&ret,cd1 ?
					"%s"PSS"snaps"PSSFB".%s":
					"%s"PSS"snaps"PSSFB,
					BaseDirectory,tempstr,cd1);
		} else if (odirs[FCEUIOD_SNAPS])
			asprintf(&ret,cd1 ?
				"%s"PSS"%d.%s":
				"%s"PSS,
				odirs[FCEUIOD_SNAPS],id1,cd1);
		else
			asprintf(&ret,cd1 ?
				"%s"PSS"snaps"PSS"%d.%s":
				"%s"PSS"snaps"PSS,
				BaseDirectory,id1,cd1);
		break;
	case FCEUMKF_FDS:
		sprintf(tempstr,"%.450s.%s",
			FileBase,md5_asciistr(FCEUGameInfo->MD5,1));
		if (odirs[FCEUIOD_NV])
			asprintf(&ret,"%s"PSSFB".fds",
				odirs[FCEUIOD_NV],tempstr);
		else
			asprintf(&ret,"%s"PSS"sav"PSSFB".fds",
				BaseDirectory,tempstr);
		break;
	case FCEUMKF_SAV:
		sprintf(tempstr,"%.450s.%s",
			FileBase,md5_asciistr(FCEUGameInfo->MD5,1));
		if (odirs[FCEUIOD_NV])
			asprintf(&ret,cd1 ?
				"%s"PSSFB".%s":
				"%s"PSSFB,
				odirs[FCEUIOD_NV],tempstr,cd1);
		else
			asprintf(&ret,cd1 ?
				"%s"PSS"sav"PSSFB".%s":
				"%s"PSS"sav"PSSFB,
				BaseDirectory,tempstr,cd1);
		break;
	case FCEUMKF_CHEAT:
		if (odirs[FCEUIOD_CHEATS])
			asprintf(&ret,"%s"PSSFB".cht",
				odirs[FCEUIOD_CHEATS],FileBase);
		else
			asprintf(&ret,"%s"PSS"cheats"PSSFB".cht",
				BaseDirectory,FileBase);
		break;
	case FCEUMKF_IPS:
		sprintf(tempstr,"%.450s%s",
			FileBase,FileExt);
		asprintf(&ret,"%s"PSSFB".ips",
			FileBaseDirectory,tempstr);
		break;
	case FCEUMKF_GGROM:
		asprintf(&ret,"%s"PSS"gg.rom",
			BaseDirectory);
		if (!CheckFileOrDirExists(ret))
		{ /* added; 2010.03.26, 03.29 */
			free(ret);
			asprintf(&ret,"%s"PSS"GENIE.ROM",
				BaseDirectory);
		}
		break;
	case FCEUMKF_FDSROM:
		asprintf(&ret,"%s"PSS"disksys.rom",
			BaseDirectory);
		break;
	case FCEUMKF_PALETTE:
		if (odirs[FCEUIOD_MISC])
			asprintf(&ret,"%s"PSSFB"%s.pal",
				odirs[FCEUIOD_MISC],FileBase);
		else
			asprintf(&ret,"%s"PSS"gameinfo"PSSFB"%s.pal",
				BaseDirectory,FileBase);
	}

	return ret;
}

void MakeFileBase(const char *f)
{
	const char	*tp1, *tp3;

 #if PSS_STYLE==4
	tp1 = (char*)strrchr(f,':');
 #elif PSS_STYLE==1
	tp1 = (char*)strrchr(f,'/');
 #else
	tp1 = (char*)strrchr(f,'\\');
  #if PSS_STYLE!=3
	tp3 = (char*)strrchr(f,'/');
	if (tp1<tp3)
		tp1 = tp3;
  #endif
 #endif
	if (tp1==0)
	{
		tp1 = f;
		strcpy(FileBaseDirectory,".");
	} else
	{
		memcpy(FileBaseDirectory,f,tp1-f);
		FileBaseDirectory[tp1-f] = 0;
		++tp1;
	}

	if ((tp3=strrchr(f,'.'))!=0 && tp3>tp1)
	{
		memcpy(FileBase,tp1,tp3-tp1);
		FileBase[tp3-tp1] = 0;
		strcpy(FileExt,tp3);
	} else
	{
		strcpy(FileBase,tp1);
		FileExt[0] = 0;
	}
}

uint32 uppow2(uint32 n)
{
	unsigned	x = 31, v = 0x80000000UL;

	do
		if (n&v)
		{
			if (v!=n)
				return v+=v;
			break;
		}
	while(v>>=1,--x!=0);

	return n;
}
