/*
 iNES.c

 * MacFCEU 0.8a8
 *
 * Copyright notice for this file:
 *  ines.c Copyright (C) 1998 BERO
 *  ines.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "cart.h"
#include "PPU.h"

#define INESPRIV
#include "iNES.h"
#include "general.h"
#include "state.h"
#include "file.h"
#include "memory.h"
#include "utility/MD5.h"
#include "cheat.h"
#include "VSUni.h"

#ifdef env68k
 #pragma segment files
#endif

/* ---- globals */

extern SFORMAT FCEUVSUNI_STATEINFO[];

/* Kevin Horton discovered that very few existing ROMs have a trainer,
 and most of these aren't real trainers at all but subroutines used by a mapper hack. */
uint8		*iNESTrainerData = 0;
uint8		*iNESPRGROM = 0;
uint8		*iNESCHRROM = 0;

static CartInfo	iNESCart;

uint8		iNESMirroring;
uint16		iNESCHRBankList[8];
int32		iNESIRQLatch, iNESIRQCount;
uint8		iNESIRQa;

int		iNESMapper;
uint8		iNESSubMapper; /* Rarely used, submappers define mostly how CHR is wired up */
uint8		iNESVersion, iNESSecurity;
static uint32	PPUtype;

uint32		iNESPRGROMSize = 0;
uint32		iNESPRGRAMSize = 0;
uint32		iNESCHRROMSize = 0;
uint32		iNESCHRRAMSize;

/* non-volatile Work RAM */
uint32		iNESPRGWRAMSize, iNESCHRWRAMSize;

static iNES_HEADER	head;

void (*MapClose)(void);
void (*MapperReset)(void);

/* ---- protos */

static void iNESPower(void);
static int NewiNES_Init(int num);

/* ---- code */

/*
	MapperReset() is called when the NES is reset(with the reset button).
	Mapperxxx_init is called when the NES has been powered on.
*/

static DECLFR(TrainerRead)
{
	return iNESTrainerData[A-0x7000]; //&0x1FF
}

void iNESGI(int h)
{
	switch(h)
	{
	case GI_RESETM2:
		if (MapperReset)
			MapperReset();
		if (iNESCart.Reset)
			iNESCart.Reset();
		break;
	case GI_POWER:
		if (iNESCart.Power)
			iNESCart.Power();
		if (iNESTrainerData)
		{
			int	x = 0;

			do
			{
				X6502_DMW(0x7000+x,iNESTrainerData[x]);
				if (X6502_DMR(0x7000+x)!=iNESTrainerData[x])
				{
					SetReadHandler(0x7000,0x71FF,TrainerRead);
					break;
				}
			} while(++x<512);
		}
		break;
	case GI_CLOSE:
		FCEU_SaveGameSave(&iNESCart);

		if (iNESCart.Close)
			iNESCart.Close();
		if (iNESPRGROM)
			{FCEU_free(iNESPRGROM); iNESPRGROM = 0;}
		if (iNESCHRROM)
			{FCEU_free(iNESCHRROM); iNESCHRROM = 0;}
		if (MapClose)
			MapClose();
		if (iNESTrainerData)
			{FCEU_free(iNESTrainerData); iNESTrainerData = 0;}
	}
}

static uint32	iNESGameCRC32;

struct CRCMATCH
{
	uint32	crc;
	char	*name;
};

struct INPSEL
{
	uint32	crc32;
	int	input1;
	int	input2;
	int	inputfc;
};

/* This is mostly for my personal use.  So HA. */
static void SetInput(void)
{
	static struct INPSEL moo[] =
	{
 {0x3a1694f9,SI_GAMEPAD,SI_GAMEPAD,SIFC_4PLAYER}, /* Nekketsu Kakutou Densetsu */

 {0xc3c0811d,SI_GAMEPAD,SI_GAMEPAD,SIFC_OEKAKIDS}, /* The two "Oeka Kids" games */
 {0x9d048ea4,SI_GAMEPAD,SI_GAMEPAD,SIFC_OEKAKIDS}, /*			     */

 {0xaf4010ea,SI_GAMEPAD,SI_POWERPADB,-1}, /* World Class Track Meet */
 {0xd74b2719,SI_GAMEPAD,SI_POWERPADB,-1}, /* Super Team Games */
 {0x61d86167,SI_GAMEPAD,SI_POWERPADB,-1}, /* Street Cop */
 {0x6435c095,SI_GAMEPAD,SI_POWERPADB,-1}, /* Short Order/Eggsplode */


 {0x47232739,SI_GAMEPAD,SI_GAMEPAD,SIFC_TOPRIDER}, /* Top Rider */

 {0x48ca0ee1,SI_GAMEPAD,SI_GAMEPAD,SIFC_BWORLD}, /* Barcode World */
 {0x9f8f200a,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERA}, /* Super Mogura Tataki!! - Pokkun Moguraa */
 {0x9044550e,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERA}, /* Rairai Kyonshizu */
 {0x2f128512,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERA}, /* Jogging Race */
 {0x60ad090a,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERA}, /* Athletic World */

 {0x8a12a7d9,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Totsugeki Fuuun Takeshi Jou */
 {0xea90f3e2,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Running Stadium */
 {0x370ceb65,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Meiro Dai Sakusen */
 // Bad dump? {0x69ffb014,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Fuun Takeshi Jou 2 */
 {0x6cca1c1f,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Dai Undoukai */
 {0x29de87af,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Aerobics Studio */
 {0xbba58be5,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Family Trainer: Manhattan Police */
 {0xea90f3e2,SI_GAMEPAD,SI_GAMEPAD,SIFC_FTRAINERB}, /* Family Trainer:  Running Stadium */

 {0xd9f45be9,SI_GAMEPAD,SI_GAMEPAD,SIFC_QUIZKING}, /* Gimme a Break ... */
 {0x1545bd13,SI_GAMEPAD,SI_GAMEPAD,SIFC_QUIZKING}, /* Gimme a Break ... 2 */

 {0x7b44fb2a,SI_GAMEPAD,SI_GAMEPAD,SIFC_MAHJONG}, /* Ide Yousuke Meijin no Jissen Mahjong 2 */
 {0x9fae4d46,SI_GAMEPAD,SI_GAMEPAD,SIFC_MAHJONG}, /* Ide Yousuke Meijin no Jissen Mahjong */

 {0x980be936,SI_GAMEPAD,SI_GAMEPAD,SIFC_HYPERSHOT}, /* Hyper Olympic */
 {0x21f85681,SI_GAMEPAD,SI_GAMEPAD,SIFC_HYPERSHOT}, /* Hyper Olympic (Gentei Ban) */
 {0x915a53a7,SI_GAMEPAD,SI_GAMEPAD,SIFC_HYPERSHOT}, /* Hyper Sports */
 {0xad9c63e2,SI_GAMEPAD,-1,SIFC_SHADOW},	/* Space Shadow */

 {0x24598791,-1,SI_ZAPPER,0},	/* Duck Hunt */
 {0xff24d794,-1,SI_ZAPPER,0},   /* Hogan's Alley */
 {0xbeb8ab01,-1,SI_ZAPPER,0},	/* Gumshoe */
 {0xde8fd935,-1,SI_ZAPPER,0},	/* To the Earth */
 {0xedc3662b,-1,SI_ZAPPER,0},	/* Operation Wolf */
 {0x2a6559a1,-1,SI_ZAPPER,0},	/* Operation Wolf (J) */

 {0x23d17f5e,SI_GAMEPAD,SI_ZAPPER,0},	/* The Lone Ranger */
 {0xb8b9aca3,-1,SI_ZAPPER,0},  /* Wild Gunman */
 {0x5112dc21,-1,SI_ZAPPER,0},  /* Wild Gunman */
 {0x4318a2f8,-1,SI_ZAPPER,0},  /* Barker Bill's Trick Shooting */
 {0x5ee6008e,-1,SI_ZAPPER,0},  /* Mechanized Attack */
 {0x3e58a87e,-1,SI_ZAPPER,0},  /* Freedom Force */
 {0x851eb9be,SI_GAMEPAD,SI_ZAPPER,0},	/* Shooting Range */
 {0x74bea652,SI_GAMEPAD,SI_ZAPPER,0},	/* Supergun 3-in-1 */
 {0x32fb0583,-1,SI_ARKANOID,0}, /* Arkanoid(NES) */
 {0xd89e5a67,-1,-1,SIFC_ARKANOID}, /* Arkanoid (J) */
 {0x0f141525,-1,-1,SIFC_ARKANOID}, /* Arkanoid 2(J) */

 {0x912989dc,-1,-1,SIFC_FKB},	/* Playbox BASIC */
 {0xf7606810,-1,-1,SIFC_FKB},	/* Family BASIC 2.0A */
 {0x895037bc,-1,-1,SIFC_FKB},	/* Family BASIC 2.1a */
 {0xb2530afc,-1,-1,SIFC_FKB},	/* Family BASIC 3.0 */
 {0,-1,-1,-1}
	};

	int	x = -1;

	while(moo[++x].input1>=0 || moo[x].input2>=0 || moo[x].inputfc>=0)
		if (moo[x].crc32==iNESGameCRC32)
		{
			FCEUGameInfo->input[0] = moo[x].input1;
			FCEUGameInfo->input[1] = moo[x].input2;
			FCEUGameInfo->inputfc = moo[x].inputfc;
			break;
		}
}

#define INESB_INCOMPLETE	1
#define INESB_CORRUPT		2
#define INESB_HACKED		4

typedef struct BADINF
{
	uint64	md5partial;
	char	*name;
	uint32	type;
} BADINF;

static BADINF BadROMImages[] =
{
#include "iNES-bad.h"
 {_int64val(0,0), 0, 0},
};

void CheckBad(uint64 md5partial);
void CheckBad(uint64 md5partial)
{
	int	x = -1;

	while(BadROMImages[++x].name)
		if (int64eq(BadROMImages[x].md5partial,md5partial))
		{
			FCEU_PrintError(
"The copy game you have loaded, \"%s\", is bad,\n"
" and will not work properly on FCE Ultra.",
				BadROMImages[x].name);
			return;
		}
}

typedef struct CHINF
{
	uint32	crc32;
	int32	mapper;
	int32	mirror;
} CHINF;

#define partialmdIx(d,s,x,h)	do{ \
	((uint8*)&(d))[BIG_ENDIAN?7-(x):x]|=(s).MD5[h-(x)]; \
}while(0)

static void FormPartialMD5(uint64 *d,
	int h)
{ //unrolled here
	_int64xset(*d,0);
	partialmdIx(*d,iNESCart,0,h);
	partialmdIx(*d,iNESCart,1,h);
	partialmdIx(*d,iNESCart,2,h);
	partialmdIx(*d,iNESCart,3,h);
	partialmdIx(*d,iNESCart,4,h);
	partialmdIx(*d,iNESCart,5,h);
	partialmdIx(*d,iNESCart,6,h);
	partialmdIx(*d,iNESCart,7,h);
}
#undef partialmdIx

static void CheckHInfo(void)
{
/* ROM images that have the battery-backed bit set in the header that really
 don't have battery-backed RAM is not that big of a problem, so I'll
 treat this differently by only listing games that should have battery-backed RAM.

 Lower 64 bits of the MD5 hash.
*/
	static uint64 savie[] =
	{
		_int64val(0x498c10dcL,0x463cfe95L), /* Battle Fleet */
		_int64val(0x6917ffcaL,0xca2d8466L), /* Famista '90 */

		_int64val(0xd63dcc68L,0xc2b20adcL), /* Final Fantasy J */
		_int64val(0x012df596L,0xe2b31174L), /* Final Fantasy 1+2 */
		_int64val(0xf6b359a7L,0x20549ecdL), /* Final Fantasy 2 */
		_int64val(0x5a30da1dL,0x9b4af35dL), /* Final Fantasy 3 */

		_int64val(0x2ee3417bL,0xa8b69706L), /* Hydlide 3*/

		_int64val(0xebbce5a5L,0x4cf3ecc0L), /* Justbreed */

		_int64val(0x6a858da5L,0x51ba239eL), /* Kaijuu Monogatari */
		_int64val(0xa4066674L,0x0b7d22feL), /* Mindseeker */

		_int64val(0x77b811b2L,0x760104b9L), /* Mouryou Senki Madara */

		_int64val(0x11b69122L,0xefe86e8cL), /* RPG Jinsei Game */

		_int64val(0xa70b4953L,0x14f4d075L), /* Ys 3 */

		_int64val(0xc04361e4L,0x99748382L), /* AD&D Heroes of the Lance */
		_int64val(0xb72ee233L,0x7ced5792L), /* AD&D Hillsfar */
		_int64val(0x2b7103b7L,0xa27bd72fL), /* AD&D Pool of Radiance */

		_int64val(0x854d7947L,0xa3177f57L), /* Crystalis */

		_int64val(0xb0bcc02cL,0x843c1b79L), /* DW */
		_int64val(0x4a1f5336L,0xb86851b6L), /* DW */

		_int64val(0x2dcf3a98L,0xc7937c22L), /* DW 2 */
		_int64val(0x733026b6L,0xb72f2470L), /* Dw 3 */
		_int64val(0x98e55e09L,0xdfcc7533L), /* DW 4*/
		_int64val(0x8da46db5L,0x92a1fcf4L), /* Faria */
		_int64val(0x91a6846dL,0x3202e3d6L), /* Final Fantasy */
		_int64val(0xedba17a2L,0xc4608d20L), /* ""		*/

		_int64val(0x94b94848L,0x62a26cbaL), /* Legend of Zelda */
		_int64val(0x04a31647L,0xde80fdabL), /*      ""      */

		_int64val(0x9aa1dc16L,0xc05e7de5L), /* Startropics */
		_int64val(0x1b084107L,0xd0878bd0L), /* Startropics 2*/

		_int64val(0x836c0ff4L,0xf3e06e45L), /* Zelda 2 */

	/* Abandon all hope if the game has 0 in the lower 64-bits of its MD5 hash */
		_int64val(0L,0L)
	};

	static CHINF moo[] =
	{
	#include "iNES-correct.h"
	};

	uint64	partialmd5;
	int	x, tofix;

	FormPartialMD5(&partialmd5,15);
	CheckBad(partialmd5);

	x = 0, tofix = 0;
	do
	{
		if (moo[x].crc32!=iNESGameCRC32)
			continue;
		if (moo[x].mapper>=0)
		{
			if (moo[x].mapper&0x800 && VROM_size)
			{
				VROM_size = 0;
				FCEU_free(iNESCHRROM); iNESCHRROM = 0;
				tofix|= 8;
			}
			if (iNESMapper!=moo[x].mapper&0xFF)
				iNESMapper = moo[x].mapper&0xFF, tofix|= 1;
		}
		if (moo[x].mirror<0)
			break;
		if (moo[x].mirror==8)
		{
			if (Mirroring==2)
			{ // Anything but hard-wired(four screen).
				tofix|=2;
				Mirroring=0;
			}
		} else if (Mirroring!=moo[x].mirror)
		{
			/* Don't complain if one-screen mirroring needs to be set
			   (the iNES header can't hold this information).
			*/
			if (Mirroring!=(moo[x].mirror&~4))
				if ((moo[x].mirror&~4)<=2)
					tofix|= 2;
			Mirroring = moo[x].mirror;
		}
		break;
	} while(++x,moo[x].mirror>=0 || moo[x].mapper>=0);

	x = 0;
	do
	{
		if (int64eq(savie[x],partialmd5))
			if (!(head.ROM_type&B6_Battery))
				head.ROM_type|= 2, tofix|= 4;
	} while(!int64z(savie[++x]));

	/* Games that use these iNES mappers tend to have the four-screen bit set
		when it should not be.
	*/
	if ((iNESMapper==118 || iNESMapper==24 || iNESMapper==26) && Mirroring==2)
		Mirroring = 0, tofix|= 2;

	/* Four-screen mirroring implicitly set. */
	if (iNESMapper==99)
		Mirroring = 2;

	if (tofix)
	{
		char	gigastr[384];

		strcpy(gigastr,
"  The iNES header contains incorrect information.\n"
"For now, the information will be corrected in RAM.\n");

		if (tofix&1)
			sprintf(gigastr+strlen(gigastr),
"  The mapper number should be set to %d.",
				iNESMapper);
		if (tofix&2)
		{
			static char	*mstr[4] =
				{"Horizontal","Vertical","Four-screen","???"};

			sprintf(gigastr+strlen(gigastr),
"  Mirroring should be set to \"%s\".",
				mstr[Mirroring&3]);
		}

		if (tofix&4)
			strcat(gigastr,"  The battery-backed bit should be set.");
		if (tofix&8)
			strcat(gigastr,"  This game should not have any CHR ROM.");

		FCEU_printf("%s\n",
			gigastr);
	}
}

static void CheckVSUni(const iNES_HEADER *h)
{
	uint64		partialmd5;

	if (h->ROM_type2&B7_VSUni)
	{
		iNESSecurity = 0;
		if (iNESVersion && h->expansion[5]<(4<<4))
			iNESSecurity = h->expansion[5]>>4;
	}
	FormPartialMD5(&partialmd5,7);
	FCEU_VSUniCheck(partialmd5,&iNESMapper,&Mirroring,&PPUtype);
}

int iNESFixHeader(uint8 *h)
{
	if (!memcmp(h+7,"DiskDude",8) || !memcmp(h+7,"demiforce",9))
		goto clear9;
	else if (!memcmp(h+10,"Ni03",4))
	{
		if (!memcmp(h+7,"Dis",3))
			goto clear9;
	} else
	{
		if (iNESVersion)
		{
			h[6]&=~B6_Battery;
			if ((uint32)(h[10]-(1<<4))<(14<<4))
				h[6]|= B6_Battery;
		} else if (h[10]|h[11]|h[12]|h[13]|h[14]|h[15])
		{
		clear9:
			h[7] = h[8] = h[9] = 0;
			goto end;
		}
		return 1; /* no problems */
	}
end:
	h[10] = h[11] = h[12] = h[13] = h[14] = h[15] = 0;
	return 0;
}

int iNESReadHeader(const iNES_HEADER *h)
{
	unsigned	t;

	VROM_size = h->VROM_size;
	ROM_size = h->ROM_size;

	iNESMapper = h->ROM_type>>4;
	iNESMapper|= h->ROM_type2&0xF0;
	iNESSubMapper = 0;

	if (h->ROM_type&B6_FourScreen)
		Mirroring = 2; /* FourScreen */
	else
		Mirroring = h->ROM_type&1; /* V/H */

	if (iNESVersion)
	{
		iNESMapper|= h->expansion[0]<<8&0x100;
		iNESSubMapper = h->expansion[0]>>4;
		VROM_size|= h->expansion[1]<<8&0xF00;
		ROM_size|= h->expansion[1]<<4&0xF00;
		/* PRGRAM prev. took from exp[0] */
		iNESPRGRAMSize = (t=h->expansion[2]&0xF)-1u<14u ? 64u<<t:0;
		iNESPRGWRAMSize = (t=h->expansion[2]>>4)-1u<14u ? 64u<<t:0;
		iNESCHRRAMSize = (t=h->expansion[3]&0xF)-1u<14u ? 64u<<t:0;
		iNESCHRWRAMSize = (t=h->expansion[3]>>4)-1u<14u ? 64u<<t:0;
	} else
	{
		iNESPRGRAMSize = h->expansion[0]*8192;
		iNESPRGWRAMSize = (h->expansion[0] ? h->expansion[0]:1)*8192;
		iNESCHRRAMSize = VROM_size ? 0:iNESMapper==13 ? 16384:8192;
		iNESCHRWRAMSize = 0;
	}

	if (h->ROM_type&B6_Battery)
		iNESPRGRAMSize = 0;
	else
		iNESPRGWRAMSize = 0;

	if ((iNESPRGROMSize=uppow2(ROM_size))==0)
	{ /* in olden days, this just shouted "No PRG ROM!", then quit */
		FCEU_PrintError(
"Warning: PRG ROM size is stored as zero.\n"
" Setting to 1024KB.");
		iNESPRGROMSize = 1024<<10;
	} else
		iNESPRGROMSize<<= 14;

	if (VROM_size)
		VROM_size = uppow2(VROM_size);
	iNESCHRROMSize = VROM_size<<13;

	return 1;
}

static uint32 iNESDetectPPU(const char *name,
	const iNES_HEADER *h)
{
	uint32		t = 0;

	if (iNESVersion && h->expansion[4]&2)
		t = PPU2C02|PPU2C07;
	else if (h->expansion[iNESVersion ? 4:1]&1)
		t = PPU2C07;

	if (iNESVersion)
		if (h->ROM_type2&B7_PC10)
			t = RP2C03B;
		else if ((t=h->expansion[5]&0xF)<13) /* expansion[6] = VSMode */
			t = 1L<<t;

	if (t==0 && name)
	{ /* have to go by name... */
		if (strstr(name,"(u)") || strstr(name,"(U)")
				|| strstr(name,"(j)") || strstr(name,"(J)")) /* NTSC, right? */
			t = PPU2C02;
		else if (strstr(name,"(e)") || strstr(name,"(E)")
				|| strstr(name,"(F)") || strstr(name,"(f)") 
				|| strstr(name,"(G)") || strstr(name,"(g)") 
				|| strstr(name,"(I)") || strstr(name,"(i)"))
			t = PPU2C07;
	}

	switch(t&(PPU2C02|PPU2C07))
	{
	case PPU2C02: FCEUGameInfo->vidsys = GIV_NTSC; break;
	case PPU2C07: FCEUGameInfo->vidsys = GIV_PAL;
	}

	return t;
}

void iNESPrintInfo(uint8 submap)
{
	if (iNESVersion)
		FCEUD_Message(" iNES Version 2.0\n");

	if (iNESPRGRAMSize)
		FCEU_printf(
" PRG RAM: %4d bytes\n",
			iNESPRGRAMSize);
	if (iNESCHRRAMSize)
		FCEU_printf(
" CHR RAM: %4d bytes\n",
			iNESCHRRAMSize);
	FCEU_printf(
" PRG ROM:  %3d x 16KiB\n"
" CHR ROM:  %3d x  8KiB\n"
" ROM CRC32:  0x%08lx\n",
		ROM_size,VROM_size,iNESGameCRC32);
	FCEU_printf(
" ROM MD5:    0x%s\n",
		md5_asciistr(iNESCart.MD5,0)); /* two calls, smaller & faster */
	FCEU_printf(submap ?
" Mapper:  % 4d, Sub-mapper %d":" Mapper:  % 4d",
		iNESMapper,submap);
	if (iNESVersion && head.ROM_type2&B7_PC10)
		FCEU_printf(
"\n PlayChoice-10");

	FCEU_printf(
"\n Mirroring:  %s\n",
		Mirroring==2 ? "None(Four-screen)"
		:Mirroring ? "Vertical":"Horizontal");

	if (head.ROM_type&B6_Battery)
		FCEUD_Message(
" Battery-backed.\n");
	if (head.ROM_type&B6_Trainer)
		FCEUD_Message(
" Trained.\n");
}

static void PrintSystemInfo(void)
{
	static const char	*strModeS[3] =
	{
		"RBI Baseball", "TKO Boxing", "Super Xevious"
	};
	static const char	*strPPU[14] =
	{
		"RP2C03B", "RP2C03G",
		"RP2C04-0001", "RP2C04-0002", "RP2C04-0003", "RP2C04-0004",
		"RC2C03B", "RC2C03C",
		"RC2C05-01", "RC2C05-02", "RC2C05-03", "RC2C05-04",
		"RC2C05-05"
	};

	unsigned	t, x;

	if ((t=iNESSecurity-1)<3)
		FCEU_printf(
" VS Mode '%s'.\n",
			strModeS[t]);

	if (PPUtype)
	{
		char	buf[170];
		long	z = 0;

		x = 0, t = 1;
		*buf  = 0; /* 2012.03.17 */
		do
			if (t&PPUtype)
				z+= sprintf(z+buf," %s",
					strPPU[x]);
		while(t+=t,++x<13);

		if ((t=PPUtype&(PPU2C02|PPU2C07))!=0)
			FCEU_printf(
" Video...%s%s\n",
				buf,t==PPU2C02 ?
					" NTSC":t==PPU2C07 ?
					" PAL":
					" NTSC/PAL");
	}
}

int iNESRead(const char *name,
	FCEUFILE *fp)
{
	struct md5_context	md5;

	int		result;

	if (fp==NULL)
		return -3;

	if (FCEU_fread(&head,1,16,fp)!=16 || memcmp(&head,"NES\x1A",4))
		return 0;
	memset(&iNESCart,0,sizeof(iNESCart));

	/* version has to be read first */
	if ((iNESVersion=head.ROM_type2>>2&3)!=2)
		iNESVersion = 0;
	if ((result=iNESFixHeader((uint8*)&head))==0)
		FCEU_printf("** Non-standard iNES header data removed.\n");

	iNESReadHeader(&head);
	PPUtype = iNESDetectPPU(name,&head);

	if ((iNESPRGROM=(uint8*)FCEU_malloc(iNESPRGROMSize))==0)
		goto memfail;
	dwmemset(iNESPRGROM,-1,iNESPRGROMSize);

	if (VROM_size)
	{
		if ((iNESCHRROM=(uint8*)FCEU_malloc(iNESCHRROMSize))==0)
			goto memfail;
		dwmemset(iNESCHRROM,-1,iNESCHRROMSize);
	}

	if (head.ROM_type&B6_Trainer)
	{
		if ((iNESTrainerData=(uint8*)FCEU_gmalloc(512))==0)
		{
		memfail:
			result = -2;
		fail:
			if (iNESPRGROM)
				{FCEU_free(iNESPRGROM); iNESPRGROM = 0;}
			if (iNESCHRROM)
				{FCEU_free(iNESCHRROM); iNESCHRROM = 0;}
			return result;
		}
		if (FCEU_fread(iNESTrainerData,512,1,fp)!=1)
			return -1;
	}

	result = FCEU_fread(iNESPRGROM,0x4000,ROM_size,fp);

	if (VROM_size)
	{
		if (result!=ROM_size)
			goto fail; /* last read didn't meet up; we're gettin' nada here */
		FCEU_fread(iNESCHRROM,0x2000,VROM_size,fp);
	}

	md5_starts(&md5);
	md5_update(&md5,iNESPRGROM,iNESPRGROMSize);

	iNESGameCRC32 = FCEUI_CRC32(0,iNESPRGROM,iNESPRGROMSize);

	if (VROM_size)
	{
		iNESGameCRC32 = FCEUI_CRC32(iNESGameCRC32,
			iNESCHRROM,iNESCHRROMSize);
		md5_update(&md5,iNESCHRROM,iNESCHRROMSize);
	}

	md5_finish(&md5,iNESCart.MD5);
	memcpy(FCEUGameInfo->MD5,iNESCart.MD5,sizeof(iNESCart.MD5));

	iNESCart.CRC32 = iNESGameCRC32;

	return 1;
}

int iNESLoad(const char *name,
	FCEUFILE *fp)
{
	int		result;

	if ((result=iNESRead(name,fp))!=1)
		return result;
	GameInterface = iNESGI;

	ResetCartMapping();
	ResetExState(0,0);

	SetupCartPRGMapping(0,iNESPRGROM,iNESPRGROMSize,0);
	SetupCartPRGMapping(1,WRAM,iNESPRGWRAMSize,1);

	iNESPrintInfo(iNESSubMapper);

	SetInput();
	CheckHInfo();
	CheckVSUni(&head);

	PrintSystemInfo();

	// Must remain here; some functions may change value of VROM_size and free(VROM).
	if (iNESCHRROMSize)
		SetupCartCHRMapping(0,iNESCHRROM,iNESCHRROMSize,0);

	if (Mirroring==2)
		SetupCartMirroring(4,1,ExtraNTARAM);
	else if (Mirroring>=0x10)
		SetupCartMirroring(2+(Mirroring&1),1,0);
	else
		SetupCartMirroring(Mirroring&1,Mirroring>>2&1,0);

	iNESCart.battery = head.ROM_type&B6_Battery;
	iNESCart.mirror = Mirroring;

	//if(MapperNo!=18) {
	//	if(ROM) free(ROM);
	//	if(VROM) free(VROM);
	//	ROM=VROM=0;
	//	return(0);
	//}

	if (iNESMapper!=0 && (result=NewiNES_Init(iNESMapper))!=0)
	{
		if (result<0)
		{
			iNESCart.battery = 0; /* belay saving */
			iNESGI(GI_CLOSE);
		}
	} else
	{ //old-style (maybe)
		iNESCart.Power = iNESPower;
		if (iNESCart.battery)
		{
			iNESCart.SaveGame[0] = WRAM;
			iNESCart.SaveGameLen[0] = iNESPRGWRAMSize;
		/*
			if (iNESCHRWRAMSize)
			{
				iNESCart.SaveGame[1] = CHRRAM;
				iNESCart.SaveGameLen[1] = iNESCHRWRAMSize;
			}
		*/
		}
	}

	FCEU_LoadGameSave(&iNESCart);

	FCEU_printf("\n");
	return 1;
}

/* bank setters */

void FASTAPASS(2) VRAM_BANK1(uint32 A,
	uint8 V)
{
	uint32	AB = A>>10;

	V&= 7;
	VPage[AB] = &CHRRAM[V<<10]-A;
	PPUCHRRAM|= 1<<AB;
	CHRBankList[AB] = V;
}

void FASTAPASS(2) VRAM_BANK4(uint32 A,
	uint32 V)
{
	uint32	AB = A>>10;

	V = (V&1)<<2;
	VPage[AB] = &CHRRAM[V<<8]-A;
	PPUCHRRAM|= 15<<AB;

	CHRBankList[AB] = V;
	CHRBankList[++AB] = ++V;
	CHRBankList[++AB] = ++V;
	CHRBankList[++AB] = ++V;
}

void FASTAPASS(2) VROM_BANK1(uint32 A,
	uint32 V)
{
	setchr1(A,V);
	CHRBankList[A>>10] = V;
}

void FASTAPASS(2) VROM_BANK2(uint32 A,
	uint32 V)
{
	setchr2(A,V);
	CHRBankList[A>>=10] = V+=V;
	CHRBankList[++A] = ++V;
}

void FASTAPASS(2) VROM_BANK4(uint32 A,
	uint32 V)
{
	setchr4(A,V);
	CHRBankList[A>>=10] = V<<=2;
	CHRBankList[++A] = ++V;
	CHRBankList[++A] = ++V;
	CHRBankList[++A] = ++V;
}

void FASTAPASS(1) VROM_BANK8(uint32 V)
{
	setchr8(V);
	CHRBankList[0] = V<<=3;
	CHRBankList[1] = ++V;
	CHRBankList[2] = ++V;
	CHRBankList[3] = ++V;
	CHRBankList[4] = ++V;
	CHRBankList[5] = ++V;
	CHRBankList[6] = ++V;
	CHRBankList[7] = ++V;
}

void FASTAPASS(2) ROM_BANK8(uint32 A,
	uint32 V)
{
	setprg8(A,V);
	if ((signed)(A-=0x8000)>=0)
		PRGBankList[A>>13] = V;
}

/*AJP: added; 2010.03.09*/
DECLFW(ROM_BANK8W)
{
	ROM_BANK8(A&=0xE000,V);
}
DECLFW(ROM_BANK16W)
{
	ROM_BANK16(A&=0xC000,V);
}

void FASTAPASS(2) ROM_BANK16(uint32 A,
	uint32 V)
{
	setprg16(A,V);
	if ((signed)(A-=0x8000)>=0)
	{
		PRGBankList[A>>=13] = V+=V;
		PRGBankList[++A] = ++V;
	}
}

void FASTAPASS(1) ROM_BANK32(uint32 V)
{
	setprg32(0x8000,V);
	PRGBankList[0] = V<<=2;
	PRGBankList[1] = ++V;
	PRGBankList[2] = ++V;
	PRGBankList[3] = ++V;
}

void FASTAPASS(1) onemir(uint8 V) /*single screen*/
{
	if (Mirroring==2)
		return;
	if (V>1)
		V = 1;
	Mirroring = 0x10|V;
	setmirror(MI_0+V);
}

void FASTAPASS(1) MIRROR_SET2(uint8 V) /*0:H,1:V*/
{
	if (Mirroring==2)
		return;
	Mirroring = V;
	setmirror(V);
}

void FASTAPASS(1) MIRROR_SET(uint8 V) /*0:V,1:H*/
{
	if (Mirroring==2)
		return;
	Mirroring = V^=1;
	setmirror(V);
}

static void NONE_init(void)
{
	ROM_BANK16(0x8000,0);
	ROM_BANK16(0xC000,~0);

	if (VROM_size)
		VROM_BANK8(0);
	else
		setvram8(CHRRAM);
}

static DECLFW(FCEU_BWRAM)
{
	WRAM[A-0x6000] = V;
}

static DECLFR(FCEU_AWRAM)
{
	return WRAM[A-0x6000];
}

/* ---- iNES mappers */

void (*MapInitTab[256])(void) =
{
	0,0,0,0,
	0,0,Mapper6_init,0,
	Mapper8_init,Mapper9_init,Mapper10_init,0,
	0,0,0,Mapper15_init,
	Mapper16_init,Mapper17_init,Mapper18_init,0,
	0,Mapper21_init,Mapper22_init,Mapper23_init,
	Mapper24_init,Mapper25_init,Mapper26_init,0,
	0,0,0,0,
	Mapper32_init,Mapper33_init,0,0,
	0,0,0,0,
	Mapper40_init,Mapper41_init,Mapper42_init,0,
	0,0,Mapper46_init,0,
	Mapper48_init,0,Mapper50_init,0,
	0,0,0,0,
	0,Mapper57_init,Mapper58_init,0,
	0,Mapper61_init,Mapper62_init,0,
	0,Mapper65_init,0,Mapper67_init,
	Mapper68_init,Mapper69_init,0,Mapper71_init,
	Mapper72_init,Mapper73_init,0,Mapper75_init,
	Mapper76_init,Mapper77_init,0,0,
	0,0,Mapper82_init,Mapper83_init,
	0,Mapper85_init,Mapper86_init,0,
	Mapper88_init,Mapper89_init,0,Mapper91_init,
	Mapper92_init,0,0,0,
	Mapper96_init,Mapper97_init,0,Mapper99_init,
	0,0,0,0,
	0,0,0,Mapper107_init,
	0,0,0,0,
	Mapper112_init,0,Mapper114_init,0,
	0,Mapper117_init,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,Mapper151_init,
	0,Mapper153_init,Mapper154_init,0,
	Mapper156_init,Mapper157_init,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	Mapper180_init,0,Mapper182_init,0,
	Mapper184_init,Mapper185_init,0,0,
	0,Mapper189_init,0,0,
	0,Mapper193_init,0,0,
	0,0,0,0,
	Mapper200_init,Mapper201_init,Mapper202_init,Mapper203_init,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,Mapper225_init,Mapper226_init,Mapper227_init,
	0,Mapper229_init,Mapper230_init,Mapper231_init,
	Mapper232_init,0,Mapper234_init,0,
	0,0,0,0,
	0,Mapper241_init,Mapper242_init,0,
	Mapper244_init,0,Mapper246_init,0,
	Mapper248_init,0,0,0,
	0,0,0,Mapper255_init
};

void (*MapStateRestore)(int version);
void iNESStateRestore(int version)
{
	int		x;

	if (iNESMapper==0)
		return;
	x = 0;
	do
		setprg8(4+x<<13,PRGBankList[x]);
	while(++x<4);

	if (VROM_size)
	{
		x = 0;
		do
			setchr1(x<<10,CHRBankList[x]);
		while(++x<8);
	}

if(0) switch(Mirroring)
 {
   case 0:setmirror(MI_H);break;
   case 1:setmirror(MI_V);break;
   case 0x12:
   case 0x10:setmirror(MI_0);break;
   case 0x13:
   case 0x11:setmirror(MI_1);break;
 }

	if (MapStateRestore)
		MapStateRestore(version);
}

static void iNESPower(void)
{
	int		x, type = iNESMapper;

	SetCartRead(0x8000,0xFFFF);

	GameStateRestore = iNESStateRestore;
	MapClose = 0;
	MapperReset = 0;
	MapStateRestore = 0;

	setprg8r(1,0x6000,0);

	SetReadHandler(0x6000,0x7FFF,FCEU_AWRAM);
	SetWriteHandler(0x6000,0x7FFF,FCEU_BWRAM);
	FCEU_CheatSetRAM(8,0x6000,WRAM);

	/* This statement represents atrocious code.  I need to rewrite
	  all of the iNES mapper code...
	*/
	IRQCount = IRQLatch = IRQa = 0;
	dwmemset(GameMemBlock+iNESPRGWRAMSize,0,sizeof(GameMemBlock)-iNESPRGWRAMSize);

	/* set to defaults */
	NONE_init();

	ResetExState(0,0);
	if (FCEUGameInfo->type==GIT_VSUNI)
		AddExState(FCEUVSUNI_STATEINFO,~0,0,0);

	if (type==19 || type==6 || type==69 || type==85 || type==96)
		iNESPRGRAMSize = 0x8000;
	if (iNESPRGWRAMSize)
		AddExState(WRAM,iNESPRGWRAMSize,0,"WRAM");
	if (iNESPRGRAMSize)
		AddExState(MapperExRAM,iNESPRGRAMSize,0,"MEXR");
	if ((!VROM_size /*&& type!=13*/ && type!=96) || type==6 || type==19)
		AddExState(CHRRAM,iNESCHRRAMSize,0,"CHRR");
	if (head.ROM_type&B6_FourScreen)
		AddExState(ExtraNTARAM,2048,0,"EXNR");

	/* Exclude some mappers whose emulation code handle save state stuff themselves. */
	if (type /*&& type!=13*/ && type!=96)
	{
		char	tak[8];

		AddExState(mapbyte1,32,0,"MPBY");
		AddExState(&Mirroring,1,0,"MIRR");
		AddExState(&IRQCount,4,1,"IRQC");
		AddExState(&IRQLatch,4,1,"IQL1");
		AddExState(&IRQa,1,0,"IRQA");
		AddExState(PRGBankList,4,0,"PBL");
		x = 0, tak[0] = 'C', tak[1] = 'B', tak[2] = 'L', tak[4] = 0;
		do
		{
			tak[3] = '0'+x;
			AddExState(&CHRBankList[x],2,1,tak);
		} while(++x<8);
	}

	if (MapInitTab[type])
		MapInitTab[type]();
	else if (type)
		FCEU_PrintError("iNES mapper #%d is not supported at all.",
			type);
}

#include "board.h"

static int NewiNES_Init(int num)
{
	BMAPPING	*l;

	if (FCEUGameInfo->type==GIT_VSUNI)
		AddExState(FCEUVSUNI_STATEINFO,~0,0,0);
	for(l=FCEU_bmap; l->init; ++l)
	{
		if (num!=l->mapper)
			continue;
		if (l->name)
			FCEU_printf(" %s board.\n",
				l->name);
		if (!VROM_size)
		{
			if ((iNESCHRROM=(uint8*)FCEU_malloc(iNESCHRRAMSize))==0)
				return -2;
			SetupCartCHRMapping(0,iNESCHRROM,iNESCHRRAMSize,1);
			AddExState(iNESCHRROM,iNESCHRRAMSize,0,"CHRR");
		}
		if (head.ROM_type&B6_FourScreen)
			AddExState(ExtraNTARAM,2048,0,"EXNR");
		l->init(&iNESCart);
		return 1;
	}
	return 0;
}
