/*
 iNES.h

 * MacFCEU 0.6.6
 *
 * Copyright notice for this file:
 *  ines.h Copyright (C) 1998 Bero
 *  ines.h Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

int iNESLoad(const char *name,
	struct FCEUFILE *fp);

typedef struct iNES_HEADER/*(16)*/
{
	char	ID[4]; /*NES\32*/
/*04*/	uint8	ROM_size;
	uint8	VROM_size;
/*06*/	uint8	ROM_type;
	uint8	ROM_type2;	//bits #2-3: version
/*08*/	uint8	expansion[8];	//submap, chrhi.4, prghi.4, chrram/prgram
} iNES_HEADER;

#define B6_FourScreen	8
#define B6_Trainer	4
#define B6_Battery	2
#define B7_PC10		2	/* PlayChoice-10 handled like a console... 'that so wrong? */
#define B7_VSUni	1

/* made these public so info can be printed externally; 2010.04.09 */
int iNESRead(const char *name,
	struct FCEUFILE *fp);
void iNESGI(int cmd);
int iNESFixHeader(uint8 *h);
int iNESReadHeader(const iNES_HEADER *h);
void iNESPrintInfo(uint8 submap);
void iNESStateRestore(int v);

extern int iNESMapper;
extern uint8 iNESSubMapper, iNESVersion, iNESSecurity, iNESMirroring;
extern uint32 iNESPRGRAMSize, iNESPRGROMSize, iNESCHRROMSize, iNESCHRRAMSize;
extern uint32 iNESPRGWRAMSize, iNESCHRWRAMSize;
extern uint16 iNESCHRBankList[8];
extern int32 iNESIRQLatch, iNESIRQCount;
extern uint8 iNESIRQa;

#ifdef INESPRIV
extern void (*MapStateRestore)(int v);
extern void (*MapClose)(void);
extern void (*MapperReset)(void);

/* This order is necessary */
#define WRAM          (GameMemBlock)
#define sizeofWRAM    8192

#define MapperExRAM   (GameMemBlock+sizeofWRAM)
#define sizeofMapperExRAM  32768
/* for the MMC5 code to work properly.  It might be fixed later... */

#define CHRRAM        (GameMemBlock+sizeofWRAM+sizeofMapperExRAM)
#define sizeofCHRRAM 8192

#define ExtraNTARAM   (GameMemBlock+sizeofWRAM+sizeofMapperExRAM+sizeofCHRRAM)
#define sizeofExtraNTARAM 2048

#define PRGBankList    (ExtraNTARAM+sizeofExtraNTARAM)

#define mapbyte1       (PRGBankList+4)
#define mapbyte2       (mapbyte1+8)
#define mapbyte3       (mapbyte2+8)
#define mapbyte4       (mapbyte3+8)

#define IRQa iNESIRQa
#define Mirroring iNESMirroring
#define IRQCount iNESIRQCount
#define IRQLatch iNESIRQLatch
#define CHRBankList iNESCHRBankList
#endif

void FASTAPASS(2) VRAM_BANK1(uint32 A, uint8 V);
void FASTAPASS(2) VRAM_BANK4(uint32 A, uint32 V);

void FASTAPASS(2) VROM_BANK1(uint32 A, uint32 V);
void FASTAPASS(2) VROM_BANK2(uint32 A, uint32 V);
void FASTAPASS(2) VROM_BANK4(uint32 A, uint32 V);
void FASTAPASS(1) VROM_BANK8(uint32 V);

void FASTAPASS(2) ROM_BANK8(uint32 A, uint32 V);
void FASTAPASS(2) ROM_BANK16(uint32 A, uint32 V);
void FASTAPASS(1) ROM_BANK32(uint32 V);

//2010.03.09
DECLFW(ROM_BANK8W);
DECLFW(ROM_BANK16W);

extern uint8 vmask;
extern uint32 vmask1;
extern uint32 vmask2;
extern uint32 vmask4;
extern uint32 pmask8;
extern uint8 pmask16;
extern uint8 pmask32;

void FASTAPASS(1) onemir(uint8 V);
void FASTAPASS(1) MIRROR_SET2(uint8 V);
void FASTAPASS(1) MIRROR_SET(uint8 V);

//Mapper2_init
//Mapper3_init

void Mapper6_init(void);
//Mapper7_init
void Mapper8_init(void);
void Mapper9_init(void);
void Mapper10_init(void);
//Mapper11_init

//Mapper13_init

void Mapper15_init(void);
void Mapper16_init(void);
void Mapper17_init(void);
void Mapper18_init(void);

void Mapper21_init(void);
void Mapper22_init(void);
void Mapper23_init(void);
void Mapper24_init(void);
void Mapper25_init(void);
void Mapper26_init(void);

void Mapper32_init(void);
void Mapper33_init(void);
//Mapper34_init

void Mapper40_init(void);
void Mapper41_init(void);
void Mapper42_init(void);
//Mapper43_init

void Mapper46_init(void);

void Mapper48_init(void);

void Mapper50_init(void);
//Mapper51_init

void Mapper57_init(void);
void Mapper58_init(void);

//Mapper60_init
void Mapper61_init(void);
void Mapper62_init(void);

//Mapper64_init
void Mapper65_init(void);
//Mapper66_init
void Mapper67_init(void);
void Mapper68_init(void);
void Mapper69_init(void);
//Mapper70_init
void Mapper71_init(void);
void Mapper72_init(void);
void Mapper73_init(void);

void Mapper75_init(void);
void Mapper76_init(void);
void Mapper77_init(void);
//Mapper78_init
//Mapper79_init
//Mapper80_init

void Mapper82_init(void);
void Mapper83_init(void);

void Mapper85_init(void);
void Mapper86_init(void);
//Mapper87_init
void Mapper88_init(void);
void Mapper89_init(void);

void Mapper91_init(void);
void Mapper92_init(void);
//Mapper93_init
//Mapper94_init

void Mapper96_init(void);
void Mapper97_init(void);

void Mapper99_init(void);

void Mapper107_init(void);

void Mapper112_init(void);
//Mapper113_init
void Mapper114_init(void);

void Mapper117_init(void);

//Mapper140_init

//Mapper144_init

void Mapper151_init(void);
//Mapper152_init
void Mapper153_init(void);
void Mapper154_init(void);

void Mapper156_init(void);
void Mapper157_init(void);
//Mapper158_init
//Mapper159_init

void Mapper180_init(void);

void Mapper182_init(void);

void Mapper184_init(void);
void Mapper185_init(void);

void Mapper189_init(void);

void Mapper193_init(void);

void Mapper200_init(void);
void Mapper201_init(void);
void Mapper202_init(void);
void Mapper203_init(void);

//Mapper207_init

void Mapper225_init(void);
void Mapper226_init(void);
void Mapper227_init(void);
//Mapper228_init
void Mapper229_init(void);
void Mapper230_init(void);
void Mapper231_init(void);
void Mapper232_init(void);

void Mapper234_init(void);
void Mapper235_init(void);

void Mapper241_init(void);
void Mapper242_init(void);

void Mapper244_init(void);
//Mapper245_init
void Mapper246_init(void);

void Mapper248_init(void);

void Mapper255_init(void);

void NSFAY_Init(void);
void NSFVRC6_Init(void);
void NSFVRC7_Init(void);
