/*
 macos/config.c
  configuration file processing (include'd by shell.c)
  requires ini.c v1.8.4 or later

  Created: 2009.01.21
  Revised: 2015.08.22

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

extern int	rec_wavedepth;
extern char	recentpaths[RECENTPATHSIZE][RECENTCOUNT];

#include "../video.h"
#include "../palette.h"
#include "../NSF.h"
#include "macinput.h"

static trec	ntsc_trec[] =
{
	{_int32,0*4,"col"},
	{_float,1*4,";tint"},
	{_float,1*4,"sat"},
	{_int32,2*4,"hue"},
	{_int32,3*4,"brt"},
	{_ifixed,4*4,"r"},
	{_ifixed,5*4,"g"},
	{_ifixed,6*4,"b"},
	{0}
};

static trec	tileview_trec[] =
{
	{sizeof(tileviewcolors), 0, "colors"},
	{0}
};

static trec	wavelog_trec[] =
{
	{_int32, 0, "depth"},
	{0}
};

static trec	log_trec[] =
{
	{_uint8, 0, "max"},
	{0}
};

static trec	recentpaths_trec[] =
{
	{RECENTPATHSIZE,0,RECENTCOUNTSTR},
	{0}
};

static tcon	macfceu_tcon[] =
{
	{emu_trec, "emu", &emuset},
	{snd_trec, "sound", &sndset},
	{vid_trec, "video", &vidset},
	{input_trec, "input", &inputset},
	{ntsc_trec, "ntsc", ntsccthb},
	{debug_trec, "debug", &debugset},

	//NACA(ndiq,CalcPDecodeIQ),

	{tileview_trec, "tileview", tileviewcolors},
	{wavelog_trec, "wavelog", &rec_wavedepth},
	{log_trec, "log", &logtofile},

	O_TCON(recentpaths_trec, recentpaths),

	{0}
};

static void configfile_save(char *filename)
{
	FILE		*f;

	ntscsat = FCEUI_GetNTSC_HSB(&ntschue,&ntscbrt)/2.; /*update from internal*/

	if (filename && (f=fopen(filename,"w"))!=NULL)
	{
		vidset.bgc = PPUvoidcolor;
		memcpy(ntsccthb+4,CalcPOutscales,sizeof(CalcPOutscales));
		sndset.nsfmif = NSFMaxInitFrames;
		eoptions&=~(EO_NRATEFIX|EO_PRATEFIX|EO_DRATEFIX|EO_SHOWFPS|EO_SHOWFCT);
		if (RateFixFlags&0x80)
			eoptions|= EO_NRATEFIX;
		if (RateFixFlags&1)
			eoptions|= EO_PRATEFIX;
		if (RateFixFlags&0x40)
			eoptions|= EO_DRATEFIX;
		if (show_fps)
			eoptions|= EO_SHOWFPS;
		if (show_fct)
			eoptions|= EO_SHOWFCT;

		fprintf(f,
"#\n"
"# %s\n"
"#  v" VERSION_STR "\n"
"#\n",
			filename);
		dump_ini(f,macfceu_tcon);
		fclose(f);
	}
}

static void configfile_load(char *filename)
{
	FILE		*f;

	ntscsat = FCEUI_GetNTSC_HSB(&ntschue,&ntscbrt)/2.;

	if (filename && (f=fopen(filename,"r"))!=NULL)
	{
		fread_ini(f,filename,macfceu_tcon,0);
		fclose(f);

		memcpy(CalcPOutscales,ntsccthb+4,sizeof(CalcPOutscales));
		PPUvoidcolor = vidset.bgc;
		NSFMaxInitFrames = sndset.nsfmif;
		RateFixFlags = 0;
		if (eoptions&EO_NRATEFIX)
			RateFixFlags = 0x80;
		if (eoptions&EO_PRATEFIX)
			RateFixFlags|= 1;
		if (eoptions&EO_DRATEFIX)
			RateFixFlags|= 0x40;
		if (eoptions&EO_SHOWFPS)
			show_fps = 1;
		if (eoptions&EO_SHOWFCT)
			show_fct = 1;
	}
	/* validate */
	if (!afbskip)
		afbskip = 14;
	if (!ffbskip)
		ffbskip = 4;

	FCEUI_SetNTSC_HSB(ntsccol,ntschue,ntscsat*2,ntscbrt);
}
