/*
 macos/debugger.c
  and to logwindow - yes, to logwindow

  MacFCEU 0.9d3

Sunday, July 20, 2008 7:48 PM EDST
  rewrote list mechanisms.
  free and malloc should be realloc(..,0) and realloc(NULL,..) to minimize code space.
Wednesday, August 25, 2009 12:39 AM
  finished preliminary rewrite (argc,argv).
Saturday, October 3, 2009 4:49 AM
  fixed commandline option-key bug (all characters were being put in after the cursor).
Sunday, December 27, 2009 9:57 AM
  added debuggerwindow_click.  should use commandline-function for click loc.
Saturday, March 27, 2010 1:23 PM
  added trace flags to disassembly, added user bits step_traceflags ("tf").
Tuesday, March 30, 2010 4:05 AM
  eliminated fill_values, since getvalue is now defined here.
Tuesday, February 8, 2011 6:18 PM
  MIN/MAX widths expanded (cmd key no longer needed);
 tried to max. based on command line.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <Events.h>
#include <MacWindows.h>
#include <Dialogs.h>

#include "shell.h"
#include "../FCEU.h"
#include "../x6502.h"
#include "../general.h"
#include "../common/dface.h"
#include "../PPU.h"
#include "../fsound.h"
//#include "../FDS.h"
//#include "../boards/Namcot.h"

#include "shellmenus.h"
#include "macvideo.h"
#include "macinput.h"
#include "debugger.h"
#include "usercommands.h"
#include "tileview.h" /* for window and function */
#include "memview.h"
#include "messagelog.h" /* for cls, added in 2009 */
#include "../WAVE.h" /* for FCEU_WaveUpdateHeader */
#include "../palette.h" /* for DIQ (CalculatePalette) values... oh, trial and error */
#include "../NSF.h" /* for its public variables */

#ifdef env68k
 #pragma segment Debug
#endif

/* ---- constants */

#define PASCAL_DEBUGGER_TITLE		"\pDebugger"

#define TEXT_SIDEWIDTH			42
#define DISASMTEXT_INDENT_H		3
#define DISASMTEXT_INDENT_V		0

#define TEXT_LINEHEIGHT			12
#define DEFAULT_INSCOUNT		11

#define SIDESEPARATOR_COUNT		9

#define TEXT_SIDESEPARATOR_V		(TEXT_LINEHEIGHT*SIDESEPARATOR_COUNT+2)
#define TEXT_CHARWIDTH			6

#define INIT_DBGWINDOW_WIDTH		(5+44*TEXT_CHARWIDTH)
#define INIT_DBGWINDOW_HEIGHT		(DEBUGGERWINDOW_MINIMUM_HEIGHT+TEXT_LINEHEIGHT*2)

enum
{
	DEBUGGERWINDOW_MINIMUM_WIDTH	= 4+38*TEXT_CHARWIDTH, //34
	DEBUGGERWINDOW_MINIMUM_HEIGHT	= TEXT_LINEHEIGHT*DEFAULT_INSCOUNT-7,
	DEBUGGERWINDOW_MAXIMUM_WIDTH	= (4+104*TEXT_CHARWIDTH),
	DEBUGGERWINDOW_MAXIMUM_HEIGHT	= TEXT_LINEHEIGHT*DEFAULT_INSCOUNT*99-7
};

/* ---- globals */

extern struct FCEUGI *GI;

WindowRef	debuggerwindow = NULL;

boolean		debug_open;

debug_settings	debugset = {0, 1, 1, 0, 1};

static boolean	pendupdate = false; /* pending window update? */
static boolean	break_reenter = false; /* true if at least one breakpt is hard */

/* ---- value set/get */

enum
{ /* the order here must be in sync with the array that follows */
	_CPU_A,
	_CPU_X,
	_CPU_Y,
	_CPU_P,
	_CPU_S,
	_CPU_PC,
	_CPU_IRL,
	_CPU_IA,
	_video_SRLN,
	_video_ERLN,
	_video_SRLP,
	_video_ERLP,
	_PPU_SLVN,
	_PPU_SLVP,
	_PPU_BGBL,
	_input_UDLR,
	_sound_SFCC,
	_sound_SIRQ,
	_sound_RECAMP,
	_sound_TNDAMP,
	_sound_SNDOUT,
	_debug_BI,
	_debug_BB,
	_debug_TF,
	_debug_ECHO,
	_debug_DMT,
	_message_LOG,
	_shell_FFSKIP,
	_NTSCcol_DIQR,
	_NTSCcol_DIQG,
	_NTSCcol_DIQB,
	_NSFInitFrames,

	_endofvars_, _PPU_scanline = _endofvars_,
	_SL,
	_L,
	_intlow,
	_CYC,
	_C, /* 2013.06.04 */
	_CYC32, /* 2015.08.20 */
	_VFC,
	_BP,

	_TFA,
	_TFX,
	_TFY,
	_TFP,
	_TFS,
	_TFIRL,
	_TFI, /* 2013.05.14, 6.04 */
	_TFCYC,
	_TFC,
	_TFSL,
	_TFL,
	_TFS8,
	_TFalways,
	_nmi,
	_reset,
	_irq,
	_CPUrate
};

static const char	*debugIDs[] =
{ /*variables*/
	"A",
	"X",
	"Y",
	"P",
	"S",
	"PC",
	"IRL",
	"IA",
	"srln",
	"erln",
	"srlp",
	"erlp",
	"slvn",
	"slvp",
	"bgbl",
	"UDLR",
	"sfcc",
	"sirq",
	"recamp",
	"tndamp",
	"sndout",
	"bi",
	"bb",
	"tf",
	"echo",
	"dmt",
	"log",
	"ffskip",
	"ndiqr",
	"ndiqg",
	"ndiqb",
	"nsfmif",
  /*read-only*/
	"scanline",
	"SL",		/* added the shorter; 2010.04.19 */
	"L",		/* and shorter; 2013.06.04 */
	"I",
	"CYC",
	"C",
	"CYC32",
	"VFC",
	"BP",
  /*constants*/
	"TFA",
	"TFX",
	"TFY",
	"TFP",
	"TFS",
	"TFIRL",
	"TFI",
	"TFCYC",
	"TFC",
	"TFSL",
	"TFL",
	"TFS8",
	"TFalways",
	"nmi",
	"reset",
	"irq",
	"CPUrate",
	NULL
};

static long _getvalue(int x)
{
extern int32 SFStatL;
	switch(x)
	{
	case _CPU_A: return X6502F.u32.A;
	case _CPU_X: return X6502F.u32.X;
	case _CPU_Y: return XFrame.Y;
	case _CPU_P: return XFrame.P;
	case _CPU_S: return XFrame.S;
	case _CPU_PC: return X6502F.u32.PC;
	case _CPU_IA: return X6502F.u32.IA;
	case _CPU_IRL: case _intlow: return XFrame.intlow;
	case _video_SRLN: return srendlinen;
	case _video_ERLN: return erendlinen;
	case _video_SRLP: return srendlinep;
	case _video_ERLP: return erendlinep;
	case _PPU_SLVN: return vblslinesn;
	case _PPU_SLVP: return vblslinesp;
	case _PPU_BGBL: return PPUvoidcolor;
	case _input_UDLR: return allowUDLR;
	case _sound_SFCC: return Soundfccnt;
	case _sound_SIRQ: return SFStatL;
	case _sound_RECAMP: return rectangle_amp;
	case _sound_TNDAMP: return tnd_amp;
	case _sound_SNDOUT: return soundavailable;
	case _debug_BI: return break_illegal;
	case _debug_BB: return break_breakpoint;
	case _debug_TF: return step_traceflags;
	case _debug_ECHO: return debug_echo;
	case _debug_DMT: return debug_dmt;
	case _message_LOG: return logtofile;
	case _shell_FFSKIP: return ffbskip;
	case _NTSCcol_DIQR: x = 0; goto diq;
	case _NTSCcol_DIQG: x = 1; goto diq;
	case _NTSCcol_DIQB: x = 2;
	diq: return (CalcPDecodeIQ[x][0]<<16)|(uint16)CalcPDecodeIQ[x][1];
	case _NSFInitFrames: return NSFMaxInitFrames;

	case _PPU_scanline: case _L: case _SL: return scanline;
	case _CYC: case _C: return timestamp;
	case _CYC32: return int64lo(timestampbase)+timestamp;
	case _VFC: return PPUVFrameCounter;
	case _BP: return XFrame.BP;
	case _TFA: return tfaddCPU_A;
	case _TFX: return tfaddCPU_X;
	case _TFY: return tfaddCPU_Y;
	case _TFP: return tfaddCPU_P;
	case _TFS: return tfaddCPU_S;
	case _TFIRL: case _TFI: return tfaddCPU_intlow;
	case _TFCYC: case _TFC: return tfaddCPU_CYC;
	case _TFSL: case _TFL: return tfaddPPU_scanline;
	case _TFS8: return tfaddstack8;
	case _TFalways: return tftrace_always;
	case _nmi: return 0xFFFA;
	case _reset: return 0xFFFC;
	case _irq: return 0xFFFE;
	case _CPUrate: return (PAL ? (RateFixFlags&1 ? 1.6e6/1e9:PAL_CPU/1e9)
		:RateFixFlags&0x80 ? 1.6e6/1e9:NTSC_CPU/1e9)*emuset.speed+.5;

	default: return INT_NO_VALUE;
	}
}

static int setvalue(const char *w,
	long z,
	long i)
{
	const char	**l = debugIDs, *s;

	for(; (s=*l)!=NULL; ++l)
		if (!strncasecmp(w,s,z) && !s[z])
		{
			switch(l-debugIDs)
			{
			case _CPU_A: XFrame.A = i; break;
			case _CPU_X: XFrame.X = i; break;
			case _CPU_Y: XFrame.Y = i; break;
			case _CPU_P: XFrame.P = i; break;
			case _CPU_S: XFrame.S = i; break;
			case _CPU_PC: XFrame.PC = i, XFrame.intlow&=~FCEU_JAMMED; break;
			case _CPU_IRL: case _intlow: XFrame.intlow = i; break;
			case _debug_BI: break_illegal = i; break;
			case _debug_BB: break_breakpoint = i; break;
			case _debug_TF: step_traceflags = i/*&0x3F*/; break;
			case _debug_ECHO: debug_echo = i; break;
			case _debug_DMT: debug_dmt = !!i; break;
			case _message_LOG: logtofile = i; break;
			case _shell_FFSKIP: ffbskip = PIN(i,1,19); break;
			case _input_UDLR: allowUDLR = i; break;
			case _video_SRLN: srendlinen = i; goto refresh;
			case _video_ERLN: erendlinen = i; goto refresh;
			case _video_SRLP: srendlinep = i; goto refresh;
			case _video_ERLP: erendlinep = i; goto refresh;
			case _PPU_SLVN:
				if (i>150) /* if too high, broken, imcomplete sound, slow */
					i = 150;
				vblslinesn = i;
				goto refresh;
			case _PPU_SLVP:
				if (i>150)
					i = 150;
				vblslinesp = i;
				goto refresh;
			case _PPU_BGBL: PPUvoidcolor = i; break;
			case _sound_SFCC: Soundfccnt = i; break;
			case _sound_RECAMP: FCEUI_SetRectFilterAmp(rectangle_amp=i); break;
			case _sound_TNDAMP: FCEUI_SetTNDFilterAmp(tnd_amp=i); break;
			case _sound_SNDOUT:
				if ((soundavailable=(!i ? 0:1/*adequatesoundmanager()*/))==0)
					KillSound();
				else if (GI!=NULL)
					InitSound(GI);
				break;
			case _NTSCcol_DIQR: z = 0; goto setdiq;
			case _NTSCcol_DIQG: z = 1; goto setdiq;
			case _NTSCcol_DIQB: z = 2;
			setdiq: CalcPDecodeIQ[z][0] = i>>16;
				CalcPDecodeIQ[z][1] = i&0xFFFF;
				break;
			case _NSFInitFrames: NSFMaxInitFrames = i;
				break;

			default: return 0;

			refresh: videorefresh();
			}
			return 1;
		}
	return -1;
}

const char *FCEUD_getvalue(const char *w,
	const char *e,
	int32 *i,
	void *u)
{
#pragma unused(u)
	const char	**l = debugIDs, *s;
	long		z;

	if ((e=parse_wordbreak(w,e,NULL))==0)
		return 0;
	for(z=e-w; (s=*l)!=0; ++l)
		if (!strncasecmp(w,s,z) && !s[z])
		{
			*i = _getvalue(l-debugIDs);
			return e;
		}
	return 0;
}

/* ---- state */

static void debuggerwindow_drawrules(const Rect *bounds)
{
	/* low seperator between user entry and disassembly */
	MoveTo(TEXT_SIDEWIDTH,bounds->bottom-15);
	MacLineTo(bounds->right,bounds->bottom-15);

	/* left side seperator between state and disassembly */
	MoveTo(TEXT_SIDEWIDTH,bounds->top);
	MacLineTo(TEXT_SIDEWIDTH,bounds->bottom);

	/* middle seperator between state registers and stack dump */
	MoveTo(TEXT_SIDEWIDTH,TEXT_SIDESEPARATOR_V);
	MacLineTo(0,TEXT_SIDESEPARATOR_V);

	/* indicator for low point of stack */
	MoveTo(0,TEXT_SIDESEPARATOR_V+TEXT_LINEHEIGHT);
	MacLineTo(4,TEXT_SIDESEPARATOR_V+TEXT_LINEHEIGHT);
}

static void debuggerwindow_drawstate(const X6502 *f,
	int c)
{
	char		buf[64];
	int		y;
	/* registers */
	MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT);
	DrawText(buf,0,sprintf(buf,"A  %02X",
		f->A));
	MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT*2);
	DrawText(buf,0,sprintf(buf,"X  %02X",
		f->X));
	MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT*3);
	DrawText(buf,0,sprintf(buf,"Y  %02X",
		f->Y));
	MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT*4);
	DrawText(buf,0,sprintf(buf,"P  %02X",
		f->P));
	MoveTo(DISASMTEXT_INDENT_H,TEXT_LINEHEIGHT*5);
	DrawText(buf,0,DebugPrintP(buf,0));
	MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT*6);
	DrawText(buf,0,sprintf(buf,"S  %02X",
		f->S));
	MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT*7);
	DrawText(buf,0,sprintf(buf,f->intlow>0xFFF ? "I%04X":"I %03X",
		f->intlow));
	MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT*8);
	if (GI && GI->type==GIT_NSF)
		DrawText(" NSF ",0,5);
	else
	{
		DrawText(buf,0,sprintf(buf,"L#%03d",
			scanline));
		MoveTo(DISASMTEXT_INDENT_H+6,TEXT_LINEHEIGHT*9);
		y = FCEUPPU_GetPixel();
		DrawText(buf,0,sprintf(buf,y<0 ? " n/a ":"@#%03d",
			y));
	}
	if ((c-=SIDESEPARATOR_COUNT)>0)
	{ /* stack dump */
		uint8		s = f->S;

		if (c>199) /* y'know doing over 255'll roll, right? */
			c = 200;
		y = TEXT_SIDESEPARATOR_V-2;
		do
		{
			uint32		A = 0x100+s;

			MoveTo(DISASMTEXT_INDENT_H,y+=TEXT_LINEHEIGHT);
			DrawText(buf,0,sprintf(buf,"%03X %02X", A,ARead[A](A)));
		} while(++s,--c);
	}
}

/* ---- interface */

extern uint8	PPU[4];

volatile int	debug_stepframe =-1;
volatile int	debug_stepcount =-1;
volatile int	debug_stepovercount = 0;
volatile long	debug_stepoverpoint =-1; /* PC address or <0 (none) */
uint8		somask;

#define USERLINE_BUFFERSIZE		100
static char	userline_text[USERLINE_BUFFERSIZE];
static short	userline_length = 0;
static short	userline_position =-1;
static Rect	userline_frame;
static uint8	insert_set = 1;

static void userline_calcframe(void)
{
	Rect		*bounds = &userline_frame;

	if (!debuggerwindow)
		return;
	GetPortBounds(GetWindowPort(debuggerwindow),bounds);

	bounds->top = bounds->bottom-14;
	bounds->left+= TEXT_SIDEWIDTH+1;
	bounds->right-= 15;
}

static WindowRef debuggerwindow_create(WindowRef behind)
{
	Rect		bounds;
	WindowRef	window;
	GrafPtr		oldport;
	Ptr		storage;

	GetPort(&oldport);

	/* downward center-right of screen */
	bounds = (**gamegdevice).gdRect;
	bounds.top = bounds.top+bounds.bottom+bounds.bottom-INIT_DBGWINDOW_HEIGHT>>2;
	bounds.bottom = bounds.top+INIT_DBGWINDOW_HEIGHT;
	bounds.left = (bounds.right-=4)-INIT_DBGWINDOW_WIDTH;
	if ((storage=NewPtr(256))==NULL)
		return NULL;
	*(uint32*)(252+storage) = 0;
	if ((window=NewCWindow(storage,&bounds,PASCAL_DEBUGGER_TITLE,
			false,documentProc,behind,true,'xdbg'))==NULL)
	{
		DisposePtr(storage);
		return NULL;
	}

	SetPortWindowPort(window);

	{
		SInt16		fontid;

		GetFNum("\pMonaco",&fontid);
		TextFont(fontid);
		TextSize(9);
		TextMode(srcCopy);
	}

	MacSetPort(oldport);

	return window;
}

boolean debuggerwindow_visible()
{
	return getvisiblewindowport(debuggerwindow) ? 1:0;
}

int debuggerwindow_grow(const EventRecord *event)
{
	Rect		bounding;
	long		newdimensions;

	if (!debuggerwindow_visible())
		return -1;
	if (!event)
		return 0;

	/* GrowWindow subtracts one when min=max, can adjust wholey via command key */
	bounding.top = DEBUGGERWINDOW_MINIMUM_HEIGHT;
	bounding.left = DEBUGGERWINDOW_MINIMUM_WIDTH;
	bounding.bottom = DEBUGGERWINDOW_MAXIMUM_HEIGHT;
	bounding.right = DEBUGGERWINDOW_MAXIMUM_WIDTH+1;
	if ((newdimensions=GrowWindow(debuggerwindow,event->where,&bounding))!=0)
	{
		*(uint32*)&bounding = 0;
		bounding.right = (int16)newdimensions;

		/* make height divisible by the line height */
		newdimensions+= 7+(TEXT_LINEHEIGHT/2)<<16;
		newdimensions/= TEXT_LINEHEIGHT<<16;
		newdimensions*= TEXT_LINEHEIGHT;
		newdimensions-= 7;
		bounding.bottom = newdimensions;

		SizeWindow(debuggerwindow,bounding.right,bounding.bottom,false);
		/* because of bugs in SizeWindow, invalidation should be done manually */
		{
			GrafPtr		oldport;

			GetPort(&oldport);
			SetPortWindowPort(debuggerwindow);

			bounding.left = TEXT_SIDEWIDTH+1;
			bounding.top = bounding.bottom-14;
			bounding.right-= 15;
			EraseRect(&bounding);

			MacSetPort(oldport);
		}

		userline_calcframe();
		debuggerwindow_update(true);

		return 1;
	}

	return 0;
}

int debuggerwindow_click(const EventRecord *event)
{
	Point		where;

	if (!debuggerwindow_visible())
		return -1;
	if (!event)
		return 0;
	showhide(debuggerwindow,true);

	where = event->where;
	localwindowpoint(debuggerwindow,&where);

	if (!PtInRect(where,&userline_frame))
		return 0;

	if ((userline_position=(where.h-userline_frame.left)/TEXT_CHARWIDTH)>userline_length)
		userline_position = userline_length;

	commandline_update(debuggerwindow,&userline_frame,userline_text,
		userline_length,userline_position,insert_set);

	return 1;
}

int debuggerwindow_clear(boolean update)
{
	int		had = GetBreakPointCount()|userline_length;

	DebugBreakPointsClear(-1);
	break_reenter = false;

	userline_text[0] = 0;
	userline_length = userline_position = 0;

	debuggerwindow_update(update);

	if (update)
		debug_stepcount =-1;

	return had>0;
}

static void debuggerwindow_erase(const Rect *bounds)
{
	Rect		well;

	well.left = TEXT_SIDEWIDTH+1;
	well.top = (well.bottom=bounds->bottom)-14;
	well.right = bounds->right-15;
	EraseRect(&well);

	well.top = 0;
	well.right+= 15;
	well.bottom-= 15;
	EraseRect(&well);
}

static int disAsmDrawRoutine(const char *s,
	int i)
{
	MoveTo(TEXT_SIDEWIDTH+DISASMTEXT_INDENT_H,(i+1)*TEXT_LINEHEIGHT+DISASMTEXT_INDENT_V);
	DrawText(s,0,strlen(s));

	return 0; /* only first line to show info */
}

int debuggerwindow_update(boolean erase)
{
	Rect		bounds, growrect;
	GrafPtr		oldport;
	int		linecount;

	if (!debuggerwindow_visible())
		return 0;
	getwindowbounds(debuggerwindow,&bounds);
	linecount = (bounds.bottom-bounds.top)/TEXT_LINEHEIGHT;

	GetPort(&oldport);
	setportwindow(debuggerwindow);

	if (erase)
		debuggerwindow_erase(&bounds);

	growrect.top = (growrect.bottom=bounds.bottom)-15;
	growrect.left = (growrect.right=bounds.right)-15;
	ClipRect(&growrect); /* so lines don't show up under pre-Appearance Mgr */
	if (erase) /*IsWindowUpdatePending(debuggerwindow))*/
		DrawGrowIcon(debuggerwindow);

	bounds.bottom-= 15; /* so text isn't drawn over the grow icon */
	ClipRect(&bounds);

	if (GI!=NULL)
		DebugDisassemble65xx(&XFrame,X6502F.u32.PC,linecount-1,
			disAsmDrawRoutine,0);

	commandline_update(debuggerwindow,&userline_frame,userline_text,
		userline_length,userline_position,insert_set);

	bounds.bottom+= 15;
	ClipRect(&bounds);

	debuggerwindow_drawrules(&bounds);

	if (GI!=NULL)
		debuggerwindow_drawstate(&XFrame,linecount);

	MacSetPort(oldport);

	return 1;
}

void debuggerwindow_activate(boolean activate)
{
	Rect		bounds, growrect;
	GrafPtr		oldport;

	if (!debuggerwindow_visible())
		return;
	getwindowbounds(debuggerwindow,&bounds);

	GetPort(&oldport);
	setportwindow(debuggerwindow);

	growrect.top = (growrect.bottom=bounds.bottom)-15;
	growrect.left = (growrect.right=bounds.right)-15;
	ClipRect(&growrect); /* so lines don't show up under pre-Appearance Mgr */

	if (activate)
		DrawGrowIcon(debuggerwindow);
	else
		EraseRect(&bounds);

	ClipRect(&bounds); /* so lines don't show up (Classic OS) */
}

/* ---- command level */

/* @ 6 pixels in width, 100 would cover 600 pixels (more than enough) */
static char	linebuf[264];
static uint8	linebuflength;

static uint16	prevaddr = 0, newaddr = 0;
static uint32	prevcount = 7;

static keyword_proc	prevkeyproc = NULL;

static void requiredmsg(const char *name, int count)
{
	FCEU_printf(count==1 ?
		"'%s' requires a parameter\n":count<=0 ?
		"'%s' requires parameters\n":
		"'%s' requires %u parameters\n",
		name,count);
}

static const char *sreadanyinteger(const char *s,
	long int *v)
{
	return !s ? NULL:readanyinteger(s,strlen(s),v);
}

static void strprinter(void *u,
	char v)
{
	int		a;

	if (u && (a=*(uint16*)u)<sizeof(linebuf)-1)
		linebuf[a] = v, ++(*(uint16*)u);
}

static int help_proc(int argc,
	char const*const*argv);

static int go_proc(int argc,
	char const*const*argv)
{
	long int	count = 1;

	if (argc<0)
	{ /* help message */
		FCEU_printf(
"%s\n"
" - go (close debugger)\n",
			*argv);
		return 0;
	}
	debuggerwindow_show(0);

	return 0;
}

static int breakpoints_proc(int argc,
	char const*const*argv)
{
	const char	*name = *argv;
	int32		elen = 0;
	uint8		gt = !strcasecmp(name,"gt");
	uint8		flags = gt ? BPOINT_ONCE:0;

	if (argc<=0)
	{
		if (argc<0) /* help message */
			if (gt)
				FCEU_printf(
"%s address\n"
" - goto; breakpoint once, and go\n",
					name);
			else
				FCEU_printf(
"%s [once] [hard] address\n"
" - set a breakpoint, or list all (no parameter)\n",
					name);
		return 0;
	}

	if (!--argc)
	{
		/* no parameters; list all */
		if (DebugBreakPointsLoop(-1,flags,BPList)<=0)
			FCEU_printf(!flags ?
				"%s: there are no breakpoints set.\n":
				"%s: there are no such breakpoints set.\n",
				name);
		return 0;
	}

	do
	{
		const char	*s = *++argv;
		long int	address;
		long		v;
		uint16		a = 0;
//FCEU_printf("breakpoints_proc: param %s\n",s);
		if (sreadanyinteger(s,&address))
		{
			if ((unsigned)address>0xFFFF)
			{
				FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
					name,s);
				return argc;
			}
		} else if (*s=='\'' || *s=='"')
		{ /* expression */
			if (elen)
			{
				FCEU_printf("%s: only one expression allowed.\n",
					name,s);
				return 1;
			}
			if (argc==1)
				FCEU_printf(
"%s: expr. won't be used (wasn't before an address)\n",
					name);
			if (parse_textloop(s+1,s+strlen(s),*s,&elen,NULL,NULL,&a,strprinter)[0])
			{
				FCEU_printf("%s: bad parameter.  ('%s')\n",
					name,s);
				return 1;
			}
			if (elen>=sizeof(linebuf))
				FCEU_printf("%s: expr. truncated to 256 chars.\n",
					name);
			continue;
		} else if (!strcasecmp(s,"once"))
		{
			flags|= BPOINT_ONCE;
			continue;
		} else if (!strcasecmp(s,"soft"))
		{
			flags|= BPOINT_SOFT;
			continue;
		} else if (!strcasecmp(s,"w") || !strcasecmp(s,"write"))
		{
			flags|= BPOINT_WRITE;
			continue;
		} else if (tolower(*s)=='r')
		{
			if (tolower(s[1])=='w')
				flags|= BPOINT_WRITE;
			else if (s[1] && strcasecmp(s+1,"ead"))
				goto bad;
			flags|= BPOINT_READ;
			continue;
		} else
		{ /* no other options implemented */
		bad:	FCEU_printf("%s: unrecognized option '%s'.\n",
				name,s);
			return 1;
		}
		if ((v=DebugBreakPointSet(address,flags|BPOINT_PC,elen ? linebuf:NULL))==-1)
		{
			FCEU_printf("%s: failed to set breakpoint.\n",
				name);
			return 1;
		} else
		{
			if (gt)
				debuggerwindow_show(0);
			else /* no message */
				FCEU_printf("%s: breakpoint #%u @ $%04X set.\n",
					name,(uint32)(v&0x7FFFFFFFU),(uint32)address);
			if (v>=0 && !(flags&BPOINT_READ))
				FCEU_printf(
" **Warning: %s address; requires stepping through**\n",
					address>=0x2000 ? "hard":"work-RAM");
			pendupdate = true;
		}
		elen = 0, flags = gt ? BPOINT_ONCE:0;
	} while(--argc);

	return 0;
}

static int breakpoints_clear_proc(int argc,
	char const*const*argv)
{
	const char	*name = *argv;
	long int	address;

	if (argc<=0)
	{
		if (argc<0) /* help message */
			FCEU_printf(
"%s [address]\n"
" - clear a breakpoint, or all (no parameter)\n",
				name);
		return 0;
	}

	if (!--argc)
	{ /* no parameters; clear all */
		int	v;

		break_reenter = false;

		if ((v=DebugBreakPointsClear(-1))<=0)
		{
			FCEU_printf("%s: no breakpoints.\n",
				name);
			return 0;
		}
		if (v<0)
			FCEU_printf("%s: an error occured (%d).\n",
				name,v);
		else
			FCEU_printf("%s: all %d breakpoint(s) cleared.\n",
				name,v);
		pendupdate = true;
		return 0;
	}

	do
	{
		int32		index;

		if (!sreadanyinteger(*++argv,&address)
				|| (unsigned)address>0xFFFF)
		{
			FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
				name,*argv);
			return argc;
		}
		if ((index=DebugBreakPointGet(address))<0
				|| DebugBreakPointsClear(index)<0)
			FCEU_printf("%s: no breakpoint @ $%04X.\n",
				name,(uint32)address);
		else
		{
			FCEU_printf("%s: breakpoint #%u @ $%04X cleared.\n",
				name,(uint32)index,(uint32)address);
			pendupdate = true;
		}
	} while(--argc);

	return 0;
}

static int stepframe_proc(int argc,
	char const*const*argv)
{
	long int	count = 1;

	if (argc<0)
	{ /* help message */
		FCEU_printf(
"%s [count]\n"
" - step frames by count (default: 1)\n",
			*argv);
		return 0;
	} else if (!argc)
		count = prevcount;
	else if (--argc)
	{
		if (!sreadanyinteger(*++argv,&count)
				|| (unsigned)(count-1)>=999999)
		{
			FCEU_printf("%s: count must be in the range [1,999999].\n",
				*argv);
			return argc;
		}
		if (debug_echo)
			FCEU_printf(count==1 ? "Step 1 frame\n":"Step %u frames\n",
				count);
	} else if (debug_echo)
		FCEUD_Message("Step frame\n");
	if (debug_echo)
		DebugDisassemble65xx(&XFrame,X6502F.u32.PC,0,
			0,step_traceflags|tfaddnewline);
	somask = 0;
	debug_stepframe = PPU[2];
	debug_stepoverpoint =-1; /* 2013.06.02 */
	debug_stepovercount = prevcount=count, debug_stepcount = 0x20000000;

	return argc==1 ? 0:argc;
}

static int step_proc(int argc,
	char const*const*argv)
{
	long int	count = 1;

	if (argc<0)
	{ /* help message */
		FCEU_printf(
"%s [count]\n"
" - step by count (default: 1)\n",
			*argv);
		return 0;
	} else if (!argc)
		count = prevcount;
	else if (--argc)
	{
		if (!sreadanyinteger(*++argv,&count)
				|| (unsigned)(count-1)>=999999)
		{
			FCEU_printf("%s: count must be in the range [1,999999].\n",
				*argv);
			return argc;
		}

		if (debug_echo)
			FCEU_printf("Step by %u\n",
				count);
	} else if (debug_echo)
		FCEUD_Message("Step\n");
	if (debug_echo && !(step_traceflags&tftrace_always)) /*'always' taken in debugHook*/
		DebugDisassemble65xx(&XFrame,X6502F.u32.PC,0,
			0,step_traceflags|tfaddnewline);

	if (debug_stepcount<0)
		debug_stepcount = 0;
	debug_stepcount+= (prevcount=count);

	return argc==1 ? 0:argc;
}

static int stepover_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	count = 1;

	if (argc<0)
	{ /* help message */
		FCEU_printf(
"%s [count]\n"
" - step over by count (default: 1), zero to clear step point\n",
			name);
		return 0;
	} else if (!argc)
		count = prevcount;
	else if (--argc)
	{
		if (!sreadanyinteger(*++argv,&count)
				|| (unsigned)count>999999)
		{
			FCEU_printf("%s: count must be in the range [0,999999].\n",
				name);
			return argc;
		}

		if (debug_echo)
			FCEU_printf("Step over by %u\n",
				count);
	} else if (debug_echo)
		FCEUD_Message("Step over\n");
	if (debug_echo && !(step_traceflags&tftrace_always)) /*'always' taken in debugHook*/
		DebugDisassemble65xx(&XFrame,X6502F.u32.PC,0,
			0,step_traceflags|tfaddnewline);
	somask = 0;
	if (debug_stepovercount>0x10000000) /* 2013.06.02 */
		debug_stepovercount&= debug_stepovercount;
	if ((prevcount=count)==0)
	{
		debug_stepovercount = 0; /* 2013.06.02 */
		prevcount = 1, debug_stepoverpoint =-1;
		FCEUD_Message("Debug: step point cleared.\n");
	} else
		debug_stepovercount+= prevcount, debug_stepcount = 0x20000000;

	return argc==1 ? 0:argc;
}

static int stepout_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	count = 1;

	if (argc<0)
	{ /* help message */
		FCEU_printf(
"%s [count]\n"
" - step out (of a subroutine)\n",
			name);
		return 0;
	} else if (argc && debug_echo)
		FCEUD_Message("Step out\n");
	if (debug_echo && !(step_traceflags&tftrace_always)) /*'always' taken in debugHook*/
		DebugDisassemble65xx(&XFrame,X6502F.u32.PC,0,
			0,step_traceflags|tfaddnewline);
	somask = 0x40; /* RTS/RTI - step out */
	debug_stepovercount = debug_stepcount = 0x20000000;

	return argc==1 ? 0:argc;
}

static int disasm_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	address = X6502F.u32.PC, count = prevcount;
	int		c;

	if ((c=argc)<0)
	{ /* help message */
		if (name)
			FCEU_printf(
"%s|%cp%s [addr [count]]\n"
" - disassemble at/prior to PC or addr, # of lines (default: 7)\n",
				name,(char)*name,2+name);
		return 0;
	} else if (!c)
		address = newaddr;
	else if (count=7,--c)
	{
		if (!sreadanyinteger(*++argv,&address)
				|| (unsigned)address>0xFFFF)
		{
			FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
				name,*argv);
			return 1;
		}

		if (--c)
			if (!sreadanyinteger(*++argv,&count)
					|| (unsigned)(count-1)>=9999) /*up to 29997 bytes*/
			{
				FCEU_printf("%s: count must be in [1,9999].\n",
					name);
				return c;
			}
	}
	if (name && tolower(name[1])=='p' && (address-=8)<0)
		address = 0;
	if (argc)
		FCEU_printf("Debug: disassembling @ $%04X:\n",
			(uint32)address);
	newaddr = DebugDisassemble65xx(&XFrame,prevaddr=address,~(prevcount=count),
		0,tfaddnewline);

	return c==1 ? 0:c;
}

static int clear_proc(int argc,
	char const*const*argv)
{
	if (argc<=0)
	{
		if (argc<0) /* help message */
			FCEU_printf("%s: clear screen (Message Log)\n",
				*argv);
		return 0;
	}
	logwindow_clear();

	return argc==1 ? 0:argc;
}

static void print_hexbuf(boolean addnewline)
{
	if (linebuflength>199)
		linebuflength = 200;
	if (linebuf[128])
	{ /* copy as string */
		linebuf[127] = 47-linebuflength; /* add whitespace for <16; 2013.07.13 */
		do
			linebuf[linebuflength++] = ' ';
		while(--linebuf[127]>=0);
		memcpy(linebuf+linebuflength,linebuf+129,--linebuf[128]), linebuflength+= linebuf[128];
		linebuf[128] = 1;
	}
	if (addnewline)
		linebuf[linebuflength++] = '\n', linebuf[linebuflength] = '\0';
	FCEUD_Message(linebuf);
	linebuf[0] = linebuflength = 0;
}

static void memoryprinter(uint16 a,
	uint8 v)
{
	int		o;

	if (!((a-prevaddr)&15))
	{
		if (linebuflength)
			print_hexbuf(a!=prevaddr);
		linebuflength = (a&1)-1+sprintf(linebuf," %04X  %02X ",
			a,v);
	} else if (linebuflength<64)
		linebuflength+= (a&1)-1+sprintf(linebuflength+linebuf,"%02X ",v);
	/* buffer as text */
	if ((o=(uint8)linebuf[128])!=0)
		linebuf[128+o] = (uint8)(v-20)>=127-20 ? '.':v, ++linebuf[128];
}

static int dumpmem_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	address = prevaddr, count = prevcount;
	int		c;

	if ((c=argc)<0)
	{ /* help message */
		FCEU_printf(
"%s|%cp%s addr [-t] [count]\n"
" - dump memory/page at addr or prev., no. of bytes (default: 16/256)\n",
			name,(char)*name,2+name);
		return 0;
	}
	linebuf[128] = debug_dmt;
	if (!c)
		address = newaddr;
	else if (count=(tolower(name[1])=='p' ? 256:16),--c)
	{
		if (!sreadanyinteger(*++argv,&address)
				|| (unsigned)address>0xFFFF)
		{
			FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
				name,*argv);
			return 1;
		}
		if (--c)
			if (!sreadanyinteger(*++argv,&count)
					|| (unsigned)count>9999)
			{
				FCEU_printf("%s: count must be in [0,9999].\n",
					name);
				return 1;
			}
	}

	if ((prevcount=count)!=0)
	{
		if (argc)
			FCEU_printf("Debug: dumping memory @ $%04X:\n",
				(uint32)address);
		linebuflength = 0;
		FCEUI_MemDump(prevaddr=address,count,memoryprinter);
		newaddr = count+prevaddr;
		print_hexbuf(true); /* dump what remains (plus end-of-line) */
	}

	return c==1 ? 0:c;
}

static int find_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	address, count, v;
	int		word = !!name[4];
	uint8		vh;

	if (argc<=0)
	{
		if (argc<0) /* help message */
			FCEU_printf(
"%s addr size value\n"
" - search for a %s\n",
				name,word ? "word":"byte");
		return 0;
	}

	if ((argc-=3)<=0)
		{requiredmsg(name,3); return 1;}
	if (!sreadanyinteger(*++argv,&address)
			|| (unsigned)address>0xFFFF)
	{
		FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
			name,*argv);
		return 1;
	}
	if (!sreadanyinteger(*++argv,&count)
			|| (unsigned)(count-1-word)>65536-1)
	{
		FCEU_printf("%s: count must be in the range [%u,65536].\n",
			name,1+word);
		return argc;
	}
	if (!sreadanyinteger(*++argv,&v))
	{
		FCEU_printf("%s: value must be an integer.\n",
			name);
		return argc;
	}
	v&= word ? 0xFFFF:0xFF, vh = v>>8;
	prevaddr = address, newaddr = address+count;
	if (word)
		--count;
	do
	{
		if (ARead[address](address)!=(uint8)v
				|| (word && ARead[address+1](address+1)!=vh))
			continue;
		FCEU_printf("%s: %s found.\n",
			name,word ? "word":"byte");
		linebuflength = 0;
		FCEUI_MemDump(prevaddr=address,16,memoryprinter);
		newaddr = 16+prevaddr;
		print_hexbuf(true); /* dump what remains (plus end-of-line) */
		goto end;
	} while(++address,--count);

	FCEU_printf("%s: 0x%0*X in [$%X,$%X] not found.\n",
		name,word+word+2,v,prevaddr,newaddr-1);

end:	pendupdate = true;

	return argc==1 ? 0:argc;
}

static void setbyte_loopproc(void *u,
	char v)
{
	if (u)
		DebugGenSetByte((*(uint16*)u)++,v);
}

static int setbytes_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	address;
	long		current;

	if (argc<=0)
	{
		if (argc<0) /* help message */
			FCEU_printf(
"%s addr ...\n"
" - set bytes\n",
				name);
		return 0;
	}

	if ((argc-=2)<=0)
		{requiredmsg(name,2); return 1;}
	if (!sreadanyinteger(*++argv,&address)
			|| (unsigned)(current=address)>0xFFFF)
	{
		FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
			name,*argv);
		return 1;
	}

	do
	{
		long int	value;
		uint16		a = 0;
		int32		amt;

		if (sreadanyinteger(*++argv,&value))
			{DebugGenSetByte(current,value); amt = 1;}
		else if (parse_textloop(*argv+1,*argv+strlen(*argv),**argv,
				&amt,NULL,NULL,&a,setbyte_loopproc)[0])
		{
			FCEU_printf("%s: bad parameter.  ('%s')\n",
				name,*argv);
			return 1;
		}
		current+= amt;
	} while(--argc);

	FCEU_printf(address==--current ?
		"Debug: byte set @ $%04X.\n":
		"Debug: byte(s) set @ $%04X - $%04X.\n",
		(uint32)address,current);

	pendupdate = true;

	return argc==1 ? 0:argc;
}

static int poke_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	address;
	long		current;
	int		saveindbg, peek = tolower(name[1])=='e';

	if (argc<=0)
	{
		if (argc<0) /* help message */
			FCEU_printf(
"%s addr ...\n"
" - %s bytes (be careful)\n",
				name,peek ? "read":"write");
		return 0;
	}
	if ((argc-=2-peek)<=0)
		{requiredmsg(name,2-peek); return 1;}

	saveindbg = fceuindbg, fceuindbg = 0;
	do
	{
		if (!sreadanyinteger(*++argv,&address)
				|| (unsigned)(current=address)>0xFFFF)
		{
			FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
				name,*argv);
			return 1;
		}
		if (!peek)
			break;
		FCEU_printf(" [$%04X] = $%02X\n",
			address,ARead[address](address));
	} while(--argc);

	fceuindbg = saveindbg;
	if (!argc)
		return 0;

	do
	{
		long int	value;
		uint16		a = 0;

		if (!sreadanyinteger(*++argv,&value))
		{
			FCEU_printf("%s: bad parameter.  ('%s')\n",
				name,*argv);
			return 1;
		}
		a = current;
		BWrite[a](a,value);
	} while(++current,--argc);

	FCEU_printf(address==--current ?
		"Debug: byte written @ $%04X.\n":
		"Debug: bytes written @ $%04X - $%04X.\n",
		(uint32)address,current);

	pendupdate = true;

	return 0;
}

static void XorSetByte(unsigned a,
	unsigned char v)
{
	DebugGenSetByte(a,FCEUI_MemSafePeek(a)^v);
}

static void OrSetByte(unsigned a,
	unsigned char v)
{
	DebugGenSetByte(a,FCEUI_MemSafePeek(a)|v);
}

static void AndSetByte(unsigned a,
	unsigned char v)
{
	DebugGenSetByte(a,FCEUI_MemSafePeek(a)&v);
}
/* reveal the secret message in CMC80s with `xb $F048 $FF $88D` */
static int fillbytes_proc(int argc,
	char const*const*argv)
{
	void		(*proc)(unsigned a,
				unsigned char v) = DebugGenSetByte;
	const char	*name = !argv ? NULL:*argv;
	long int	address, value, count;
	int32		current, length;
	uint16		a;
	uint8		t = tolower(*name);

	if (argc<=0)
	{
		if (argc<0) /* help message */
			FCEU_printf(
"%s addr byte [count]\n"
" - %s bytes\n",
				name,t=='x' ? "ex-or":t=='o' ?
				"bit-or":t=='a' ? "bit-and":"fill");
		return 0;
	}

	if (t=='x')
		proc = XorSetByte;
	else if (t=='o')
		proc = OrSetByte;
	else if (t=='a')
		proc = AndSetByte;
	if ((argc-=2)<=0)
		{requiredmsg(name,2); return 1;}
	if (!sreadanyinteger(*++argv,&address)
			|| (unsigned)(current=address)>0xFFFF)
	{
		FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
			name,*argv);
		return 1;
	}
	length =-1, a = 0;
	if (!sreadanyinteger(*++argv,&value)
			&& parse_textloop(*argv+1,*argv+strlen(*argv),**argv,
			&length,NULL,NULL,&a,strprinter)[0])
	{
		FCEU_printf("%s: bad parameter.  ('%s')\n",
			name,*argv);
		return 1;
	}
	if (length==1)
		value = *linebuf;
	else if (length>=0)
	{
		FCEU_printf("%s: expected single byte.  ('%s')\n",
			name,*argv);
		return 1;
	}
	count = 1;
	if (--argc)
		if (!sreadanyinteger(*++argv,&count)
				|| (unsigned)(count-1)>65536-1)
		{
			FCEU_printf("%s: count must be in the range [1,65536].\n",
				name);
			return argc;
		}
	do
		proc(current,value);
	while(++current,--count);

	FCEU_printf(address==--current ?
		"Debug: byte set @ $%04X.\n":
		"Debug: byte(s) set @ $%04X - $%04X.\n",
		(uint32)address,current);

	pendupdate = true;

	return argc==1 ? 0:argc;
}

static int setr_proc(int argc,
	char const*const*argv)
{
	const char	*name = !argv ? NULL:*argv;
	long int	value;
	int		result;

	if (argc<=0)
	{
		if (argc<0)
		{ /* help message */
			FCEU_printf(
"%s value\n"
" - set a setting or register (or just 'set' to print all)\n",
				name);
			if (++argc<0)
				FCEUD_Message(
" - CPU:      A, X, Y, P, S, PC (16 bit), I/IRL (32 bit), IA (16 bit)\n"
"  note: setting the Program Counter unlocks if jammed\n"
"             C/CYC (cycles since video frame), CYC32 (since reset/power)\n"
"             CPUrate (adjusted CPU rate, rounded)\n"
" - Video:    srln, erln, srlp, erlp (start/end render line NTSC/PAL)\n"
" - PPU:      slvn, slvp (scanline ct. VBL NTSC/PAL) -- capped to 150"
"             bgbl (blank color index; 0xFF=default), VFC (video frame ctr)\n"
" - Sound:    sndout, wdepth (next WAVE file bitdepth)\n"
"             sfcc (APU frame remaining cycle counter)\n"
"             recamp (rect wave filter amplitude), tndamp (triangle/noise/DMC)\n"
" - Input:    UDLR (simultaneous Up+Down/Left+Right enable for gamepads)"
" - Debugger: bi, bb (break on illegal ins, breakpoint)\n"
"             echo (echo level), dmt (dumpmem prints text)\n"
"             tf (step trace flags); TFxx: A,X,Y,P,S,I,C,S8,SL,always\n"
" - Message:  log (to 'Message Log' file); *512KB wrap, 255/-1 for no wrap\n");
		}
		return 0;
	}

	if (!name[3])
	{
		const char	**l;
		long		v;
		int		i;

		if (debug_echo)
			FCEUD_Message("Debug - showing all settings/registers:\n");
		i = 0, l = debugIDs;
		if (*l)
			do
			{
				v = _getvalue(i);
				if (i<_NTSCcol_DIQR)
					FCEU_printf("%6s: 0x%X (%d)\n",
						*l++,v,v);
				else
					FCEU_printf("%6s: 0x%X (%d,%d)\n",
						*l++,v,v>>16,(int16)v);
			} while(++i<_endofvars_);
		return argc==1 ? 0:argc;
	}
	if (!--argc)
	{ /* error message */
		requiredmsg(name,1);
		return 1;
	}

	if (!sreadanyinteger(argv[1],&value))
	{
		FCEU_printf("%s: value must be an integer.  ('%s')\n",
			name,argv[1]);
		return 1;
	}

	name+= 3;
	if ((result=setvalue(name,strlen(name),value))==0)
		FCEUD_Message("Debug: ID is read-only.\n");
	else if (result<0)
		FCEU_printf("Debug: unknown ID '%s'.\n",
			name);
	else if (debug_echo) //yes, mondo el cheapo; 2010.03.27
		FCEU_printf("Debug: set %s to 0x%X.\n",
			name,(uint32)value);

	pendupdate = 1;

	return argc==1 ? 0:argc;
}

static int totaldisplay_proc(int argc,
	char const*const*argv)
{
	char		t[9];

	if (argc<0)
	{ /* help message */
		FCEU_printf("%s: total CPU display (registers, statuses)\n",
			*argv);
		return 0;
	}

	DebugPrintP(t,1);
	FCEU_printf(
"Debug - printing all CPU/Bus values:\n"
" A: $%02X      X: $%02X      Y: $%02X      P: $%02X (%s)\n"
" S: $[1]%02X  PC: $%04X   IA: $%04X  Int: $%03X\n", /* IRQ/Int was %08X */
		X6502F.u32.A,X6502F.u32.X,(uint32)XFrame.Y,(uint32)XFrame.P,t,
		(uint32)XFrame.S,X6502F.u32.PC,X6502F.u32.IA,XFrame.intlow);

	return argc<=1 ? 0:argc;
}

static int where_proc(int argc,
	char const*const*argv)
{
	static const char	*types[4] =
		{"hardware","in PRG ROM","in RAM","in PRG RAM"};
	static const char	*ggstr[4] =
		{"","GG read bypassed","GG write bypassed","GG read & write bypassed"};

	char		namebuf[32];
	const char	*name = !argv ? NULL:*argv;
	long int	address;
	int		x;

	if (argc<=0)
	{
		if (argc<0) /* help message */
			FCEU_printf("%s: locate a given address (or use PC)\n",
				name);
		return 0;
	}

	if (!--argc)
		address = X6502F.u32.PC;
	else if (!sreadanyinteger(*++argv,&address)
			|| (unsigned)address>0xFFFF)
	{
		FCEU_printf("%s: address must be in [0,$FFFF].  ('%s')\n",
			name,*argv);
		return 1;
	}
	namebuf[0] = 0;
	if ((x=DebugGetAddrType(address,namebuf))<0)
	{
		FCEU_printf("%s: address $%04X is unmapped\n",
			name,address);
		goto end;
	}
	/* WRAM fix; 2013.05.10 */
	FCEU_printf("%s: address $%04X is %s%s\n",
		name,address,address<0x2000 ? "in WRAM":types[x&3],ggstr[x>>2&3]);
	if (x&=3)
	{ /* (crudely) scan forward and backward to get how far into this space the addr is */
		uint16		hi = address, lo = address;
		int		v;

		while(--lo)
			if ((v=DebugGetAddrType(lo,NULL))<0 || (v&=3)!=x)
				{++lo; break;}
		while(++hi)
			if ((v=DebugGetAddrType(hi,NULL))<0 || (v&=3)!=x)
				break; /* fixed..; 2010.04.04 */
		--hi;
		FCEU_printf(" %u bytes in [$%04X,$%04X]\n",
			(uint32)(address-lo),(uint32)lo,(uint32)hi);
	} else if (namebuf[0])
		FCEU_printf(" %s\n",
			namebuf);
end:	return argc<=1 ? 0:argc;
}

#include "debugcheat.c"

static const commandproc_keyword	debug_keywords[] =
{
	{"ab", fillbytes_proc},
	{"br", breakpoints_proc},
	{"brc", breakpoints_clear_proc},
	{"cls", clear_proc},
	{"dm", dumpmem_proc},
	{"dp", dumpmem_proc},
	{"fb", fillbytes_proc},
	{"find", find_proc},
	{"findw", find_proc},
	{"g", go_proc},
	{"gt", breakpoints_proc},
	{"help",help_proc},
	{"il", disasm_proc},
	{"ip", disasm_proc},
	{"ob", fillbytes_proc},
	{"peek", poke_proc},
	{"poke", poke_proc},
	{"s", step_proc},
	{"sf", stepframe_proc},
	{"so", stepover_proc},
	{"sout", stepout_proc},
	{"sb", setbytes_proc},
	{"set*", setr_proc},
	{"td", totaldisplay_proc},
	{"xb", fillbytes_proc},
	{"wh", where_proc},
	/* cheatez */
	{"cheat", cheat_proc},
	{"delc", delcheat_proc},
	{"delcheat", delcheat_proc},
	{NULL, NULL}
};

#ifdef KEY_PROCNAME
static const char *getkeyprocname(keyword_proc proc)
{
	const commandproc_keyword	*keyword = debug_keywords;

	int		i;

	if ((i=sizeof(debug_keywords)/sizeof(debug_keywords[0]))==0)
		goto end;
	do
	{
		if (keyword->proc==proc)
			return keyword->name;
	} while(++keyword,--i);
end:	return "???";
}
#endif

static int help_proc(int argc,
	char const*const*argv)
{
	keyword_proc	proc;
	const char	*name = *argv;

	if (argc<=0)
	{ /* help message */
		if (argc<0)
			FCEU_printf("%s: display help for a command, or all (no parameter)\n",
				name);
		return 0;
	}
	if (!--argc)
	{ /* no parameters; do all - will call this function */
		const commandproc_keyword	*keyword;

		keyword_proc	prev = NULL;

		if (debug_echo)
			FCEU_printf("%s - showing all commands:\n",
				name);
		for(keyword=debug_keywords; keyword->name!=NULL
				&& (proc=keyword->proc)!=NULL; prev=proc,++keyword)
			if (keyword->proc!=prev) /* ignore duplicates; 2009.09.05 */
				proc(-1,&keyword->name);
		return 0;
	}

	do
	{
		int		length = strlen(*++argv);

		if ((proc=keywords_getproc(*argv,length,debug_keywords))==NULL)
		{
			FCEU_printf("%s: unable to find help for %s.\n",
				name,*argv);
			return argc;
		} else
			proc(-2,argv);
	} while(--argc);

	return 0;
}

char *err_strs[] =
{
	"syntax error",
	"memory error",
	"read error",
	"write error",
	"incomplete expression",
	"interrogator already open",
	"no such identifier",
	"handler not supplied",
	"stack over/underflow",
	"value out of range",
	"comment still open",
	"unterminated string",
	"ID already defined"
};

/* return negative for a proc for ~failed param, or +1 for too many parameters */
static int keyword_default(int argc,
	char const*const*argv)
{
	char		buf[24];
	const char	*name;

	if (!argv || (name=*argv)==NULL)
		if (argc==EVAL_HUGEVAL)
		{
			FCEU_printf("Debug: unrecognized command\n");
			return 1;
		} else
			name = "Debug";
	else if (argc==EVAL_HUGEVAL) /* blank */
	{
		FCEU_printf("Debug: unrecognized '%s'\n",
			name);
		return 1;
	}

	if (--argc>=0)
	{
		if (argc)
			FCEU_printf("%s: %u parameter(s) were not used\n",
				name,argc);
		return 1;
	}
	if ((argc+=2)==0)
	{
		FCEU_printf("%s: invalid parameter(s)\n",
			name);
		return 1;
	}
	if (argc<0)
	{
		FCEU_printf("%s: error at parameter %u\n",
			name,-argc);
		return 1;
	}

	if (!argv || !*argv)
		{FCEUD_Message("Debug: unknown error\n"); return 1;}
	else
	{
		const char	*s = *argv, *n;
		long int	value;
		long		total;
		int		err;

		if (*s=='=')
			++s; /* forced evaluation; 2009.08.26 08:57 EST */
		total = strlen(s);
		if ((n=parse_parameter(s,total,buf,&err,FCEUD_getvalue,NULL,
				NULL,DebugReadAddr))==NULL)
			{synt: FCEUD_Message("Debug: parse error\n"); return 1;}
		if (buf[0]==' ')
			return 0; /* blank */
		if (err<0)
			if (~err<GET_ARRAY_COUNT(err_strs))
			{
				FCEU_printf(
"Debug: %s\n"
"# %s\n"
"# % *^\n",
					err_strs[~err],s,n-s);
				return 1;
			} else
				goto synt;

		n = parse_skipwsp(n,total+*argv);
		if (n-s!=total && *n!='#')
			FCEU_printf("Debug - what was not parsed: '%s'\n",
				n);
		if (!sreadanyinteger(buf,&value))
			if (buf[0]=='\0') //&& *argv!='"' && *argv!='\'')
				FCEU_printf("Debug: unknown symbol '%.*s'\n",
					total,s);
			else
				FCEU_printf("Debug: %s\n",
					buf);
		else
		{ /* remove ending whitespace and print the result */
			total = parse_rskipwsp(total+*argv,n)-*argv; /* direction; 2011.08.22 */
			FCEU_printf(value<0 ?
				"Debug: '%.*s' = 0x%lX, %s or %lu\n":
				"Debug: '%.*s' = 0x%lX, %s\n",
				total,s,(unsigned)value,buf,(unsigned)value);
		}
	}
	return 0;
}

/* ---- debugger window hook */

static void debuggerwindow_key(EventRecord *event)
{ /* user can do stuff by text prompt; no dumb buttons! */
extern WindowRef myfrontwindow;
	WindowRef	front;
	int		result = 0, length = 0;
	char		charcode = event->message&charCodeMask;

	if (!debuggerwindow_visible() || !event)
		return;
	front = myfrontwindow;//FrontWindow();
	if (front!=debuggerwindow)
	{
		if (!(event->modifiers&cmdKey) && charcode>=' ' && MemViewWindow && front==MemViewWindow)
			return;
		//BringToFront(debuggerwindow);
		SelectWindow(debuggerwindow);
	}

	if (event->modifiers&cmdKey)
	{
		/* MacsBug-style convenience; 2010.03.21 */
		if (charcode==kLeftArrowCharCode)
			{charcode = kHomeCharCode; goto keyin;}
		else if (charcode==kRightArrowCharCode)
			{charcode = kEndCharCode; goto keyin;}
		else if (charcode==kUpArrowCharCode) /* top & bottom; 2010.04.05 */
			{logwindow_scroll(0,-10000); return;}
		else if (charcode==kDownArrowCharCode)
			{logwindow_scroll(0,10000); return;}
		if (charcode==kBackspaceCharCode) /* remove everything before cursor */
		{
			if (userline_position>0)
			{
				/* **system lock-up** fix (neg. as giant pos.); 2011.02.13 */
				if (userline_position>userline_length)
					userline_position = userline_length;
				memmove(userline_text,userline_position+userline_text,
					length=userline_length-=userline_position);
				userline_position = result = 0;
				goto update;
			}
		} else if (charcode=='v') /* previous */
		{
			commandline_scroll(userline_text,sizeof(userline_text),&userline_length,-1);
			length = userline_length;
			goto update;
		} else if (charcode=='b') /* next */
		{
			commandline_scroll(userline_text,sizeof(userline_text),&userline_length,1);
			length = userline_length;
			goto update;
		}
		return;
	}
keyin:	/* key in charcode */
	if ((result=commandline_key(userline_text,sizeof(userline_text),
			&userline_length,&userline_position,&insert_set,charcode,
			(event->modifiers&optionKey) && (unsigned)(charcode-kBackspaceCharCode)
				<=(unsigned)kDownArrowCharCode-kBackspaceCharCode))!=0)
		length =-1; /* erase visual before execution, erase actual at end */
	else
		length = userline_length;
update:
	commandline_update(debuggerwindow,&userline_frame,userline_text,length,
		userline_position,insert_set);

	if (result>0)
	{
		if (userline_length>0)
		{
			if ((prevkeyproc=commandline_execute(userline_text,userline_length,
					debug_keywords,keyword_default,FCEUD_getvalue,//NULL,
					NULL,DebugReadAddr))!=NULL)
			{
				commandline_storenext(userline_text,userline_length);

				//now necessary to handle this case; 2010.03.20
				if (prevkeyproc==keyword_default)
					prevkeyproc = NULL;
			}
			userline_text[0] = 0;
			userline_length = userline_position = 0;
		} else if (prevkeyproc!=NULL)
		{
#ifdef KEY_PROCNAME
			if (debug_echo && prevkeyproc!=step_proc
					&& prevkeyproc!=stepover_proc
					&& prevkeyproc!=setr_proc)
				FCEU_printf(prevaddr!=newaddr ?
					"#%s [C:%d, P:%d, N:%d]\n":"#%s [C:%d, P:%d]\n",
					getkeyprocname(prevkeyproc),
					prevcount,prevaddr,newaddr);
#endif
			prevkeyproc(0,NULL);
		}
	}
}

static int32	savebpaddr =-1;
static int32	savebpix =-1;

static void savecurrent(int count)
{ /* all in one! */

	if (count<0)
	{
		if ((savebpix=DebugBreakPointGet(X6502F.u32.PC))>=0)
			DebugBreakPointsReset(savebpix,0);
		if (!DebugStepHook)
			DebugStepHook = savecurrent;
	} else
	{
		if (savebpix>=0)
			DebugBreakPointsReset(savebpix,1);
		if (DebugStepHook==savecurrent)
			DebugStepHook = NULL;
		savebpix =-1;
	}
}

/*
 JSR =	0x20 (~0xDF)
 RTI =	0x40 (~0xBF)
 RTS =	0x60 (~0x9F)
*/
#define stepInsCheck()			do{unsigned _a=X6502F.u32.PC; uint8 _v; \
	if(!((_v=ARead[_a](_a))&0x9F)) \
		if(_v==0x20) /* JSR - step over */ \
			debug_stepcount=0x20000000,debug_stepoverpoint=3+_a; \
		else if(_v&somask) \
			debug_stepcount=1,debug_stepovercount=0; \
}while(0)

static void debughook(int c)
{
	int		saveindbg;

	if (debug_stepcount>0)
	{ /* step-over implementation; 2009.10.03 */
		saveindbg = fceuindbg;
		if (debug_stepovercount>0)
		{
			if (debug_stepframe>=0)
			{ /* debug_stepovercount serves as frame count; 2013.05.09 */
				if ((debug_stepframe^PPU[2])&0x80)
					if (debug_stepframe&0x80)
						debug_stepframe = 0; /* first visible line */
					else
					{ /* VBlank */
						debug_stepframe = 0x80;
						if (!--debug_stepovercount)
							debug_stepframe =-1, debug_stepcount = 0;
					}
			} else if (debug_stepoverpoint<0
					|| (debug_stepoverpoint==X6502F.u32.PC
					&& (debug_stepoverpoint=-1)!=0))
				if (!--debug_stepovercount)
					{debug_stepcount = 0; goto first;}
				else
				{
					fceuindbg = 1;
					stepInsCheck();
					goto end;
				}
			goto step;
		}
		goto stepone;
	} else if (debug_stepcount<0 && break_reenter)
	{
		if (!_bittst(breakpoint_bitmap,X6502F.u32.PC))
			return;
	//	debuggerwindow_show(1);
		break_reenter = false;
	}
first: /* new-style breakpoint handling: call break here, ignore second */
	if ((!c || savebpaddr<0) && _bittst(breakpoint_bitmap,X6502F.u32.PC))
	{ /* crap, forgot how _DBreak_bypassbpt applied here */
		FCEUI_DebuggerBreak(subsopcode_breakpoint/*|_DBreak_bypassbpt|_DBreak_noshow*/
			|_DBreak_noillegal|_DBreak_nosave);
		if (!debug_open)
		{
			if ((break_reenter=DebugBreakPointsReset(-1,1))==0)
				if (DebugStepHook==debughook)
					DebugStepHook = NULL;
			goto stepone;
		}
	} else
		debuggerwindow_show(1);

	saveindbg = fceuindbg;
	fceuindbg = 1;

	DebugBreakPointsReset(-1,0);

	debuggerwindow_update(true);
	FCEUPPU_LineUpdate(); /* 2013.07.13 */
	invalidatewindow(gamewindow);
	if (!tilewindow)
		BlitTileWindow(0);
	invalidatewindow(tilewindow);
	invalidatewindow(MemViewWindow);

	do
	{
		EventRecord	event;
		boolean		iskey;

		if (WaitNextEvent(everyEvent,&event,1,NULL))
		{
			InitCursor();

			iskey = event.what==keyDown || event.what==autoKey;
			/* ignore step for Enter if there is commandline text */
			if ((iskey && (event.message&charCodeMask)==kEnterCharCode
						&& userline_length>0)
					|| (!handleevent(&event,NULL) && iskey))
				debuggerwindow_key(&event);

			if (pendupdate)
			{
				pendupdate = false;
				debuggerwindow_update(2);
			}
		}

		if (!debuggerwindow_visible())
			{debug_stepcount = 0; break;}
		if (scexit(0))
			debug_stepcount = 0x40000000; /* large value to prevent reentry */
	} while(debug_stepcount<=0);

	if (userline_length>0 || userline_position>=0)
	{
		userline_text[0] = 0;
		userline_length = 0, userline_position =-1;
		/* -1 to hide the cursor in update */
		commandline_update(debuggerwindow,&userline_frame,userline_text,0,-1,
			insert_set);
	}

	if (debug_stepovercount>0)
		stepInsCheck();

	if ((break_reenter=DebugBreakPointsReset(-1,1)>0)!=0 && !debug_stepcount)
	{ /* re-entry fix; 2010.04.05, 2010.04.11 */
		debug_stepcount =-1;
		DebugStepHook = debughook;
	}
	/* break function re-entry fix; 2010.04.17 */
	if (!c && _bittst(breakpoint_bitmap,X6502F.u32.PC))
		savebpaddr = X6502F.u32.PC;
	else
		savebpaddr =-1;

end:	fceuindbg = saveindbg;
stepone:
	if (step_traceflags&tftrace_always && ((unsigned)debug_stepcount<=0x10000000
			|| debug_echo>1 || debug_stepcount==0x20000000))
	{
		savecurrent(-1); /* more resets... what a mess */
		DebugDisassemble65xx(&XFrame,X6502F.u32.PC,0,
			0,step_traceflags|tfaddnewline);
		savecurrent(0);
	}
step:
	--debug_stepcount;
}

int debuggerwindow_show(int w)
{
	if (!debuggerwindow
			&& (debuggerwindow=debuggerwindow_create((WindowRef)-1))==NULL)
		return -1;
	if (w<0)
		w = !debuggerwindow_visible();
	else
		w = w!=0;
	if (w)
	{
		ShowMenuBar(); /* since event handling here won't show it; 2009.01.25 */

		FCEU_WaveUpdateHeader(); /*added; 2009.09.02*/

		showhide(debuggerwindow,true);
		BringToFront(debuggerwindow);
	//	HiliteWindow(debuggerwindow,true);
		userline_calcframe(); /* window global may have just been initialized */

		DebugStepHook = debughook;
		debug_stepcount =-1, break_reenter = false;
	} else
	{
		HideWindow(debuggerwindow);

		if (!break_reenter)
			DebugStepHook = NULL;
	}
	debug_open = w;
	return w;
}

int debuggerwindow_step(int count)
{
	if (!debuggerwindow && debuggerwindow_show(1)<=0)
		return 0;
	if (count>0)
	{
		if (debug_stepcount<0)
			debug_stepcount = 0;
		debug_stepcount+= count;
	}
	return debuggerwindow_visible();
}

/* ---- interface public */

int FCEUI_DebuggerBreak(int opc)
{
	int32		ix =-1;

	if (scexit(0)) /* in the middle of closing? */
		return -1;
	if (opc<0)
		FCEU_printf("Debugger break @ $%04X\n",
			X6502F.u32.PC);
	else if (/*(opc&0xFF)!=subsopcode_breakpoint
			||*/ (ix=DebugBreakPointGet(X6502F.u32.PC))<0)
	{
		if (opc&_DBreak_noillegal) /* external call */
			return _DBreak_noillegal;
		/* no reentry when bailing; 2009.08.25 22:42 EST */
		if (!break_illegal || debug_stepcount>=0x30000000)
			return -1;
		FCEU_printf("Illegal instruction @ $%04X\n",
			X6502F.u32.PC);
	} else
	{ /* breakpoint hit */
		if (DebugBreakPointsLoop(ix,0,BPBreak)<=0
				|| !break_breakpoint
				|| savebpaddr==X6502F.u32.PC /* don't do current twice */
				|| opc&_DBreak_nobreakpoint)
			goto end;
		FCEU_printf("Breakpoint @ $%04X\n",
			savebpaddr=X6502F.u32.PC); /* savebpaddr set */
		if (opc&_DBreak_bypassbpt)
			return _DBreak_bypassbpt;
	}
	if (!(opc&_DBreak_noshow))
	{
		debuggerwindow_show(1);
		debughook(-1);
	}
end: /* in case we need to execute the current instruction with breakpoint */
	if (!(opc&_DBreak_nosave))
		savecurrent(-1);
	savebpaddr =-1; /* ignored *this* time */

	return FCEUI_MemSafePeek(X6502F.u32.PC);
}
