/*
 macos/files.c

  Created: 2008.07.17
  Revised: 2015.08.17

2012.02.15 - added NSFe type.
*/

#include <FSpio.h>
#include <stdlib.h>
#include <ctype.h>

#include <Dialogs.h>

#include "shell.h"
#include "../FCEU.h"
#include "../GameGenie.h"

#include "shellmenus.h"
#include "macdialogs.h"
#include "macvideo.h"
#include "debugger.h" /* for window update on loading state/movie files */

#ifdef env68k
 #pragma segment shell
#endif

/* ---- constants */

#define SHELL_CREATOR		FOUR_CHAR_CODE(CREATOR)
#define MOVIEPLAYER_CREATOR	FOUR_CHAR_CODE('TVOD')

/* ---- globals */

extern struct FCEUGI	*GI;

static FSSpec	gamespec;
static char	*gamepath = NULL, *oldpath = NULL;

/* ---- code */

static const void *getextension(const uint8 *s,
	unsigned long *z)
{
	unsigned long	c;

	if (!s || !z || (c=*z)>0xFFFF)
		return NULL;
	s+= c;
	do
	{
		uint8	v;

		if ((v=*--s)==0 || v==':' || v=='\\' || v=='/')
			break;
		if (v=='.')
		{
			*z = *z-c;
			return s+1;
		}
	} while(--c!=0);

	return NULL;
}

unsigned long extensiontotype(const uint8 *s,
	int c)
{
	unsigned long	t, z;

	if (c==INT_NO_VALUE)
		c = *s++;
	else if (c<0)
		c = strlen((const char*)s);
	if ((z=c)<=1)
		return 0;
	if ((s=getextension(s,&z))==NULL || --z>9) /* length outside [1,7] */
		return 0;
	t = toupper(*s)*FOUR_CHAR_CODE('\1\0\0\0');
	if (z!=0)
		if (t+=toupper(s[1])*FOUR_CHAR_CODE('\0\1\0\0'),--z!=0)
			if (t+=toupper(s[2])*FOUR_CHAR_CODE('\0\0\1\0'),--z!=0)
				t+= toupper(s[3])*FOUR_CHAR_CODE('\0\0\0\1');
			else
				t+= FOUR_CHAR_CODE('\0\0\0 ');
		else
			t+= FOUR_CHAR_CODE('\0\0  ');
	else
		t+= FOUR_CHAR_CODE('\0   ');

	if (t==FOUR_CHAR_CODE('NES ')
			|| t==FOUR_CHAR_CODE('INES')) /* 2013.05.25 */
		t = FOUR_CHAR_CODE('Cart');
	else if (t==FOUR_CHAR_CODE('UNF '))
		t = FOUR_CHAR_CODE('UNIF');
	else if (t==FOUR_CHAR_CODE('WAV '))
		t = FOUR_CHAR_CODE('WAVE');
	else if (t==FOUR_CHAR_CODE('FCM '))
		;
	else if ((t&FOUR_CHAR_CODE('\377\377\000\377'))==FOUR_CHAR_CODE('FC\000 '))
		t = FOUR_CHAR_CODE('FCS ');

	return t;
}

FILE *FCEUD_UTF8fopen(const char *fn,
	const char *mode)
{
#ifdef mac
	FSSpec		tmpspec, *specptr;
	unsigned long	t;

	if (!strcmp(fn,getlastfilepath(&tmpspec.vRefNum,&tmpspec.parID,tmpspec.name)))
		goto tmps;
	else if (!strcmp(fn,gamepath))
		specptr = &gamespec;
	else
	{ /* try to set the file/creator code on any creation; 2011.03.28 */
		if (path2spec(fn,strlen(fn),&tmpspec)!=noErr)
			goto last;
	tmps:	specptr = &tmpspec;
	}
	if (*specptr->name)
	{
		FILE		*p;

		if ((p=FSp_fopen(specptr,mode))!=NULL)
		{
			if (*mode=='w' && (t=extensiontotype(specptr->name,INT_NO_VALUE))!=0)
			{ /* sometimes it doesn't make it this far */
				FSp_fsetfileinfo(specptr,t==FOUR_CHAR_CODE('WAVE') ?
					MOVIEPLAYER_CREATOR:SHELL_CREATOR,t);
			}
			return p;
		}
	}
#endif
last:	return fopen(fn,mode);
}

void takefilepath(const char *path)
{
	long		z = !path ? 0:1+strlen(path);
	int		t = _sc_reopengame|_sc_closegame;

	if (z && !strcmp(gamepath,path))
		if (oldpath)
			t|= _sc_reopensame;
		//else /* we've lost the actual path; take it */
		//	t = _sc_reopengame;
	if (z || gamepath) /* no unnecessary allocations */
		gamepath = realloc(gamepath,z);
	if (z)
	{
		memcpy(gamepath,path,z);
		scexit(t);
	}
	oldpath = NULL;
}

int takefile(const struct FSSpec *file)
{
	int		r = !gamepath;

	takefilepath(NULL); /* clear */
	if ((gamepath=makepathforfsspec(file,NULL))==NULL)
		return -1;
	gamespec = *file;
	/* will open on next event */

	return r;
}

/* ---- paths */

#define RECENTMENUITEMSIZE	60 /* fits in 512-width screen */

typedef char	rpath[RECENTPATHSIZE];
rpath		recentpaths[RECENTCOUNT]; /* 2880 bytes */

static int renderrecentfilesmenu(void)
{
	uint8		b[RECENTMENUITEMSIZE];
	int		i;

	for(i=0; i<RECENTCOUNT; ++i)
	{
		const char	*s = recentpaths[i], *t;
		int		z, m;

		if ((t=memchr(s,0,RECENTPATHSIZE))==NULL)
			 t = s+RECENTPATHSIZE;
		m = i+(submenuid_recent-menuid_apple<<6)+1;
		b[1] = ''; /* truncate if we need to */
		if ((z=t-s)==0)
			*b = 1;
		else if (z<sizeof(b))
			*b = z;
		else
			s+= z, z = sizeof(b)-2, s-= z, *b = z+1;
		memcpy(b+1+(z!=*b),s,z);
		enablemenuitem(m,!!z);
		setmenuitemstr(m,b);
	}
	return i;
}

int clearrecent(boolean clear)
{
	if (clear)
	{
		memset(recentpaths,0,sizeof(recentpaths));
		renderrecentfilesmenu();
		return 0;
	}
	return RECENTCOUNT;
}

const char *getrecentpath(int x)
{
	if (x<0)
		x+= RECENTCOUNT;
	if ((unsigned)x>=RECENTCOUNT)
		return NULL;
	return recentpaths[x];
}

int addrecent(const char *path)
{
	long		z;
	int		i;

	if (path) /* remove duplicates; one will become first */
		for(i=0; i<RECENTCOUNT; ++i)
			if (!strcmp(recentpaths[i],path))
				*recentpaths[i] = 0;
	/* pack */
	for(i=0; i<RECENTCOUNT-1; ++i)
		if (!*recentpaths[i])
		{ /* no need to memmove if no data overlaps */
			z = strlen(recentpaths[i+1]);
			memset(recentpaths[i],0,RECENTPATHSIZE);
			memcpy(recentpaths[i],recentpaths[i+1],MIN(z,RECENTPATHSIZE));
			memset(recentpaths[i+1],0,RECENTPATHSIZE); /* swapping with empty */
		}
	if (path && (z=strlen(path))<=RECENTPATHSIZE)
	{ /* this does however overlap */
		memmove(recentpaths[1],*recentpaths,
			sizeof(recentpaths)-RECENTPATHSIZE);
		/* wipe unuset space */
		if (z!=RECENTPATHSIZE)
			memset(*recentpaths,0,RECENTPATHSIZE);
		memcpy(*recentpaths,path,z);
	}

	return renderrecentfilesmenu();
}

int tracknewpath(char **store,
	const char *path,
	const char *initdir,
	uint16 nameoffset,
	int priority)
{
#pragma unused(priority)
	char		*p = NULL;

	if (!path)
		return -3; //idiot proofing

	if (nameoffset!=0)
	{
		if ((p=realloc(NULL,nameoffset+1))==NULL)
			return -1;
		strncpy(p,path,nameoffset);
		p[nameoffset] = 0;
	}

	if (!nameoffset)
		goto end;
	if (store && !initdir)
	{
		if (*store)
			*store = realloc(*store,0);
		*store = p;
		return 1;
	}
	p = realloc(p,0);
end:	return 0;
}

/* ---- game code */

const char *getnextgamepath()
{ /* const to prevent attempts to modify the string */
	return gamepath;
}

int cleargamepath()
{
	if (scexit(0)==(_sc_reopengame|_sc_closegame))
		return 0; /* pending file... */
	takefilepath(NULL);

	return 1;
}

/* should only be called by shell.c */
int loadgame(const char *nameo)
{
	scexit(~_sc_reopengame); /* make sure this's cleared; 2010.04.08 */

	if (nameo)
	{
		void	*p;
		long	z = 1+strlen(nameo);

		if ((p=realloc(gamepath,z))==NULL)
			return -1;
		gamepath = p;
		memcpy(gamepath,nameo,z);
		nameo = gamepath;
	} else if ((nameo=gamepath)==NULL)
		return 0;
	oldpath = gamepath;
	setcursor(4,false); /* set to watch cursor */

	if ((GI=FCEUI_LoadGame(nameo))!=NULL)
	{
		addrecent(nameo); //not called automatically, it seems, w/ dialog
		if (eoptions&EO_FSAFTERLOAD)
			initializescreen(true);
		videorefresh();
	} else
	{
		stopsound();
		if (nameo==gamepath)
			takefilepath(NULL);
	}
	parseinputgi(GI);
	redomenusgi();
#ifdef NETWORK
	FCEUD_NetworkConnect();
#endif
	return 1;
}

void opengame(const char *initdir)
{
	static char	*gfsdir = NULL;

	const char	*filter =
"All usable files(*.nes,*.nsf[e],*.fds,*.unf,*.unif,*.zip,*.gz)\0"
"*.nes;*.nsf;*.nsfe;*.fds;*.unf;*.unif;*.zip;*.gz\0"
"All non-compressed usable files(*.nes,*.nsf[e],*.fds,*.unf,*.unif)\0"
"*.nes;*.nsf;*.nsfe;*.fds;*.unf;*.unif\0All files (*.*)\0*.*\0";
	uint16		nameoffset = 0;

	if (GI)
	{ /* delayed call for reopen */
		scexit(_sc_reopengame);
		return;
	}

	getpstr((uint8*)temporary,strlistid_common,commonstr_opengame);

	if ((gamepath=getfilepath(gfp_getfile,
			(uint8*)temporary,
			NULL,!initdir ? gfsdir:initdir,NULL,
			filter,
			&nameoffset,
			NULL,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
		tracknewpath(&gfsdir,gamepath,initdir,nameoffset,0);

		/* file will be automatically opened at next event */
	}
}

/* ---- state files */

static char	*movie_gfsdir = NULL;
static char	*state_gfsdir = NULL;
static char	fsm[] = "FCE Ultra Movie(*.fcm)\0*.fcm\0";
static char	fss[] = "FCE Ultra Save State(*.fc?)\0*.fc?\0";

static char	cns[256];

void SaveStateOrMovie(const char *initdir,
	boolean movie)
{
	static char	cnm[256];

	const char	*filter;
	char		*filepath, **gfsdir, *currentname;
	unsigned	length;
	uint16		nameoffs;

	if (GenieStage==1)
	{ /* do internal error msg */
		FCEUI_SaveState(NULL);
		return;
	}
	if (!GI)
		return;

	if (movie)
		currentname = cnm, filter = fsm, gfsdir = &movie_gfsdir;
	else
		currentname = cns, filter = fss, gfsdir = &state_gfsdir;
	nameoffs = length = 0;
	if (!currentname[0])
		getcstr(currentname,strlistid_common,commonstr_defaultstatename);
	getpstr((uint8*)temporary,strlistid_common,
		movie ? commonstr_savemovie:commonstr_savestate);

	if ((filepath=getfilepath(gfp_putfile,
			(uint8*)temporary,
			NULL,initdir==NULL ? *gfsdir:initdir,currentname,
			filter,
			&nameoffs,
			NULL,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
		if (movie)
			FCEUI_SaveMovie(filepath);
		else
			FCEUI_SaveState(filepath);

		tracknewpath(gfsdir,filepath,initdir,nameoffs,7);

		if (nameoffs>0 && (length=1+strlen(filepath)-nameoffs)<=256)
			memcpy(currentname,nameoffs+filepath,length);
	}

	if (filepath)
		filepath = realloc(filepath,0);
}

void LoadStateOrMovie(const char *initdir,
	boolean movie,
	boolean nodialog)
{
	const char	*filter;
	char		*filepath = NULL, **gfsdir;
	unsigned	length;
	uint16		nameoffs;

	if (GenieStage==1)
	{ /* do internal error msg */
		FCEUI_LoadState(NULL);
		return;
	}
	if (!GI)
		return;
	if (nodialog)
		{gfsdir = NULL; goto load;}

	if (movie)
		filter = fsm, gfsdir = &movie_gfsdir;
	else
		filter = fss, gfsdir = &state_gfsdir;
	nameoffs = 0;
	getpstr((uint8*)temporary,strlistid_common,
		movie ? commonstr_loadmovie:commonstr_loadstate);

	if ((filepath=getfilepath(gfp_getfile,
			(uint8*)temporary,
			NULL,initdir==NULL ? *gfsdir:initdir,NULL,
			filter,
			&nameoffs,
			NULL,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
	load:	if (movie)
			FCEUI_LoadMovie(filepath);
		else
		{
			FCEUI_LoadState(filepath);
			/* 2015.08.09 */
			if (nameoffs>0 && (length=1+strlen(filepath)-nameoffs)<=256)
				memcpy(cns,nameoffs+filepath,length);
		}
		tracknewpath(gfsdir,filepath,initdir,nameoffs,1);
	}

	if (filepath)
		filepath = realloc(filepath,0);

	debuggerwindow_update(true); /* necessary (if open); 2011.03.29 */
}

/* ---- sound logging */

int		rec_wavedepth;

int RecordSoundWave(const char *initdir)
{
	static char	*gfsdir = NULL;
	static char	currentname[256];
	static boolean	warned;

	char		*filepath;
	int		result;
	unsigned	length;
	uint16		nameoffs, index = rec_wavedepth==8;

	if (!*currentname)
		getcstr(currentname,strlistid_common,commonstr_defaultwavename);

	if ((result=FCEUI_EndWaveRecord())!=0)
	{
		if (result<0)
			FCEUD_PrintError("## Recording overflowed - results may be raw.\n");
		return 0;
	}
	nameoffs = length = result = 0;
	getpstr((uint8*)temporary,strlistid_common,commonstr_logsound);

	if (sndset.rate>65535 && !warned)
	{
		if (showdialog(dialogid_note,confirm_yes,"\pSound Log - high sample rate",
"Note: most Classic Mac Wave players don't recognize rates above 65535.",
			NULL,NULL,NULL)==confirm_no)
		return 0;
		warned = 1;
	}
	if ((filepath=getfilepath(gfp_putfile,
			(uint8*)temporary,
			NULL,!initdir ? gfsdir:initdir,currentname,
			"16 bit MS WAVE(*.wav)\0*.wav\0U-8 bit MS WAVE(*.wav)\0*.wav\0",
			&nameoffs,
			&index,
			gfpf_filemustexist|gfpf_hidereadonly))!=NULL)
	{
		result = FCEUI_BeginWaveRecord(filepath,rec_wavedepth=(index ? 8:16));

		tracknewpath(&gfsdir,filepath,initdir,nameoffs,9);

		if (nameoffs>0 && (length=1+strlen(filepath)-nameoffs)<=256)
			memcpy(currentname,nameoffs+filepath,length);
	}

	if (filepath)
		filepath = realloc(filepath,0);

	return result;
}

/* ---- archive selection */

#include <Sound.h>

#include "../file.h"
#include "listwindow.h"

#define NAMESEL_WIDTH		80*6
#define NAMESEL_HEIGHT		(14*11+18)

/* implemented - not; 2012.04.02 */
int FCEUI_SelectName(const char *l,
	int count)
{
	EventRecord	event;
	Rect		bounds;
	WindowRef	window;
	long		z;
	int		r = IMIN(int);

	if (count<=0)
		return 0;
	/* far from finished */
//	return 0;
	/* bottom-center of screen */
	bounds = (**gamegdevice).gdRect;
	bounds.top = bounds.top+bounds.bottom-NAMESEL_HEIGHT>>1;
	bounds.left = bounds.left+bounds.right-NAMESEL_WIDTH>>1;
	bounds.bottom = bounds.top+NAMESEL_HEIGHT;
	bounds.right = bounds.left+NAMESEL_WIDTH;

	if ((window=listwindow_create((WindowRef)-1,&bounds,"\pFile Select","\pMonaco",9,
			300,0,~0,0,FOUR_CHAR_CODE('nSel')))==NULL)
		return 0;
	z = strlen(l);
	if (z>1999)
		setcursor(4,false);
	listwindow_append(window,l,z,true);
	listwindow_maxline(window,true);
	listwindow_updatescrolls(window,false,false);
	listwindow_scroll(window,0,-32768); /* start at beginning */

	ShowWindow(window);
	reenablemenus(false);

	while(r==IMIN(int))
		if (WaitNextEvent(everyEvent,&event,15,NULL))
		{
			DialogRef	fakedlg = NULL;
			WindowRef	clkwindow;
			short		part;

			InitCursor();
			switch(event.what)
			{
			case mouseDown:
				switch(part=FindWindow(event.where,&clkwindow))
				{
				case inDrag:
				case inGrow:
				case inContent:
					if (clkwindow==window)
					{
						if (part==inDrag)
							{fakedlg = (DialogRef)window; break;}
						if (listwindow_click(window,&event,part,
								NULL)==lwDoubleClick)
							goto sel;
						continue;
					}
					SysBeep(20);
					continue;
				case inGoAway:
					if (clkwindow==window)
					{
						if (TrackGoAway(window,event.where))
							r =-1;
						continue;
					}
				}
			case keyDown:
			case autoKey:
				{
					int32		i, j;
					int		cmdcode;
					char		charcode = event.message&charCodeMask;

					if ((cmdcode=event.modifiers&cmdKey)!=0)
						if (charcode=='.' || charcode=='w')
							{r =-1; break;}
					if (charcode==kPageUpCharCode
							|| charcode==kPageDownCharCode
							|| ((charcode==kUpArrowCharCode
							|| charcode==kDownArrowCharCode) && cmdcode))
					{
						int		vcnt = charcode;

						if ((cmdcode<<=16)!=0 && charcode==kUpArrowCharCode)
							cmdcode = 0, vcnt =-1<<16; /* hack scroll fix; 2010.04.10 */
						else
							vcnt|= cmdcode, vcnt|= 0x10000;
						listwindow_scroll(window,cmdcode>>8,vcnt);
						continue;
					}
					if (charcode==kUpArrowCharCode
							|| charcode==kDownArrowCharCode)
					{
						listwindow_getsel(window,&i,&j,NULL,NULL);
						if (i<0 || j<0 || (i+=charcode==kUpArrowCharCode ? -1:1)<0)
							i = 0;
						listwindow_update(window,listwindow_select(window,i,i,
							IMIN(int),0,true)&4);
						continue;
					}
					if (charcode==kReturnCharCode
							|| charcode==kEnterCharCode)
					{
					sel:	r = listwindow_getsel(window,&i,NULL,NULL,NULL);
						r = i;
						continue;
					}
					if (charcode==kEscapeCharCode)
						{r =-1; continue;}
				}
				break;
			case updateEvt:
				if ((WindowRef)event.message==window)
				{
					BeginUpdate(window);
					listwindow_update(window,true);
					EndUpdate(window);
					continue;
				}
			}
			handleevent(&event,NULL);
		}
	listwindow_dispose(window);
	reenablemenus(true);

	return r;
}
