/*
 macos/inputdialog.c

  Created: 2009.01.20
  Revised: 2010.04.01
*/
/* TO DO: Expand this into a smaller *single* four-controller on left */
/*whatever else on right.  Switch-able panels?  Non-dialog dialog? */

#include <Dialogs.h>

#include "macdialogs.h"

#include "../input.h"
#include "../common/dface.h"

#include "macinput.h"
#include "shellmenus.h"

#include "keyscan.h"

/* ---- code */

static int getanything(EventRecord *event)
{
	KeyMap		keymap, oldmap;
	int		i;

	/* wait until depress of return/enter */
	do
	{
		GetKeys(oldmap);
		WaitNextEvent(0,event,1,NULL);
	} while((((uint8*)oldmap)[MAP_RETURN>>3]&(0x80>>(MAP_RETURN&7)))
			|(((uint8*)oldmap)[MAP_KP_ENTER>>3]&(0x80>>(MAP_KP_ENTER&7))));

	do
	{
		GetKeys(keymap);
		WaitNextEvent(0,event,1,NULL);

		if (Button())
			return 256; /* mouse trumps keyboard */
		if (((uint8*)keymap)[MAP_COMMAND>>3]&(0x80>>(MAP_COMMAND&7)))
			goto end; /* command key escapes */
	} while(memcmp(keymap,oldmap,sizeof(keymap))==0);

	keymap[3] &= ~0x80; /* clear Power key */

	/* return bit index in map */
	for(i=0; i<128; ++i)
		if (((uint8*)keymap)[i>>3]&(0x80>>(i&7)))
			return i;
end:
	return -1; /* got nothing */
}

enum
{ /* dialog items */
	_gamepad1_start			= 3 /* 8, 8 */
};

enum
{ /* panels (currently only one) */
	_panel_gamepad,

	_select_previous		= 65536,
	_select_next
};

static void render1button(DialogRef dialog, int index, ButtConfig *bc)
{
	uint8			*str = (uint8*)temporary;

	if (bc->NumC<=0)
		str = "\p<unset>";
	else /* only first button configuration is recognized for now */
		switch(bc->ButtType[0])
		{
		case BUTTC_KEYBOARD:
			getpstr(temporary,strlistid_keyscan,bc->ButtonNum[0]);
			str = (uint8*)temporary;
			break;
		case BUTTC_JOYSTICK:
			str = "\p<joy>";
			break;
		case BUTTC_MOUSE:
			str = "\p<mouse>";
			break;

		default:
			str = "\p???";
			break;
		}

	setdialogitem_pstr(dialog,index,str);
}

static void set1button(ButtConfig *bc, int value)
{
	bc->NumC = 0;
	if (value<=127)
	{
		if (value<0)
			return;
		bc->ButtType[0] = BUTTC_KEYBOARD;
		bc->ButtonNum[0] = value;
	} else if (value==256)
		bc->ButtType[0] = BUTTC_MOUSE;
	else
		return;
	bc->NumC = 1;
}

static void dopanel_gamepad(DialogRef dialog,
	gpbutttype *gamepads,
	int onlyindex,
	int message)
{
	int				i, j, x;

	if (onlyindex>0)
		onlyindex -= _gamepad1_start;
	for(j=0,x=0; j<2; ++j)
	{
		ButtConfig	*bc = gamepads[j];

		for(i=0; i<8; ++i,++x) /* only the first eight for each are in the dialog */
		{
			if (x!=onlyindex && onlyindex>=0)
				continue;
			if (message==INT_NO_VALUE) /* render */
				render1button(dialog,_gamepad1_start+x,i+bc);
			else if (message==_select_previous)
				SelectDialogItemText(dialog,
					((GetDialogKeyboardFocusItem(dialog)-_gamepad1_start-1)&15)+_gamepad1_start,0,32767);
			else if (message==_select_next)
				SelectDialogItemText(dialog,
					((GetDialogKeyboardFocusItem(dialog)-_gamepad1_start+1)&15)+_gamepad1_start,0,32767);
			else if (message<=65535)
				set1button(i+bc,message);
		}
	}
}

int input_dialog(gpbutttype *gamepads,
	int side)
{
	gpbutttype	gamepads_copy[2];

	DialogRef	dialog = NULL;
	GrafPtr		oldport;
	int		panel = _panel_gamepad;
	int		hit = -1, index;

	GetPort(&oldport);

	memcpy(gamepads_copy,gamepads,sizeof(gamepads_copy));

	if ((dialog=getnewdialog(dialogid_inputsetup,1,(WindowRef)-1))==NULL)
		goto end;

	SetPortDialogPort(dialog);

	reenablemenus(false);
	SetDialogTracksCursor(dialog,true);
	SetWTitle(GetDialogWindow(dialog),side==0 ?
		"\pConfig GamePads 1/2":"\pConfig GamePads 3/4");
draw:
	if (panel==_panel_gamepad)
		dopanel_gamepad(dialog,gamepads_copy,hit,INT_NO_VALUE);
	if (hit<0)
		MacShowWindow(GetDialogWindow(dialog));

	SelectDialogItemText(dialog,_gamepad1_start,0,32767);

	while(1)
	{
		EventRecord	event;

		WaitNextEvent(everyEvent,&event,1,NULL);
		if (event.what!=keyDown)
		{ /* cmd-Return to hit default button */
			if ((hit=handleevent(&event,dialog))!=0)
			{
				if ((unsigned)hit<=std_cancel || hit==-2)
					break;
				if ((unsigned)(hit-_gamepad1_start)<=15)
					SelectDialogItemText(dialog,hit,0,32767);
				continue;
			}
			hit = standardmodal(dialog,&event);
			continue;
		}

		switch(event.message&charCodeMask)
		{
		case kUpArrowCharCode:
		case kLeftArrowCharCode:
		previous:
			/* previous */
			if (panel==_panel_gamepad)
				dopanel_gamepad(dialog,gamepads_copy,hit,_select_previous);
			break;
		case kTabCharCode:
			if (event.modifiers&shiftKey)
				goto previous;
		case kDownArrowCharCode:
		case kRightArrowCharCode:
			/* next */
			if (panel==_panel_gamepad)
				dopanel_gamepad(dialog,gamepads_copy,hit,_select_next);
			break;
		case kEscapeCharCode:
			hit = 2;
			goto animate;
		case kReturnCharCode:
		case kEnterCharCode:
			if (event.modifiers&cmdKey)
			{
				hit = 1;
			animate:
				animatebuttonitem(dialog,hit);
				goto end;
			}
			if (Button()) /* ignore if mouse button is down */
				continue;
			hit = GetDialogKeyboardFocusItem(dialog);

			/* with enough text to make selection afterward unnecessary */
			setdialogitem_pstr(dialog,hit,"\p<press>   ");
			SelectDialogItemText(dialog,hit,0,32767);

			if ((index=getanything(&event))>=0)
			{
				switch(index)
				{
				case MAP_F1:
				case MAP_F2:
				case MAP_F3:
				case MAP_F4:
				case MAP_F5:
				case MAP_F6:
				case MAP_F7:
				case MAP_F8:
				case MAP_F9:
				case MAP_F10:
				case MAP_F11:
				case MAP_F12:
				case MAP_F13:
				case MAP_F14:
				case MAP_F15:
					/*goto flush*/; /* ahhh.... let it happen */
				}

				if (panel==_panel_gamepad)
					dopanel_gamepad(dialog,gamepads_copy,hit,index);
			}
		flush:
			FlushEvents(everyEvent,0); /* autoKey will get ignored in next pass */
			goto draw;
		}
	}
end:
	if (hit==std_ok)
		memcpy(gamepads,gamepads_copy,sizeof(gamepads_copy));
	reenablemenus(true);

	if (dialog!=NULL)
	{
		SetDialogTracksCursor(dialog,false);
		DisposeDialog(dialog);
		InitCursor();
	}

	FlushEvents(everyEvent,0);

	SetPort(oldport);

	return hit==2 ? 0:hit;
}
