/*
 macos/listwindow.h
  MacFCEU 0.8.5

  Created: 2013.05.06
*/

#ifndef __LIST_WINDOW_H
 #pragma once
 #include <MacWindows.h>
 #include "../types.h"
 #include "macgeneral.h"

/* ---- constants */

enum
{ /* listwindow_click return values */
	lwWithinBounds		= 3,	/* out of range after drag */
	lwSingleClick,
	lwDoubleClick
};

/* ---- typedefs */

typedef const char* (*lwTextHook)(char *s,
	int32 *z,
	int i);
typedef void (*lwClickHook)(int i,
	int x,
	int *s,
	int *e);

/* ---- prototypes */

int32 rcalclines(int32 *list,
	int32 *count, /* required */
	const char *text,
	const char *e);

WindowRef listwindow_create(WindowRef behind,
	const Rect *bounds,
	const uint8 *title,
	const uint8 *fontfamily,
	int16 textsize,
	int16 maxlsize, /* pass zero for no hscroll */
	int32 maxlcount, /* pass negative for no vscroll; zero = unlimited (UNTESTED) */
	int16 bottompad, /* pass negative for a modal window */
	int sel, /* pass negative = no sel. clicking; zero = single; pos. = multiple */
	long refCon);
int listwindow_dispose(WindowRef window);

int listwindow_append(WindowRef window,
	const char *str,
	int z,
	Boolean update);
int listwindow_clear(WindowRef window);

int listwindow_getlheight(WindowRef window);
int listwindow_getbounds(WindowRef window,
	Rect *bounds);

int listwindow_getscrolls(WindowRef window,
	int *h,
	int *v);

int listwindow_drawtext(WindowRef window,
	int c);

int listwindow_scroll(WindowRef window,
	int hcount,
	int vcount);

/* returns the longest line length */
int listwindow_maxline(WindowRef window,
	Boolean set); /* update the scroll maxline */

/* updates the values */
int listwindow_updatescrolls(WindowRef window,
	Boolean doBounds,
	Boolean advance);

int listwindow_update(WindowRef window,
	Boolean erase);

int listwindow_grow(WindowRef window,
	const EventRecord *event,
	Rect *bounding);
/* calls listwindow_grow on Grow event */
int listwindow_click(WindowRef window,
	EventRecord *event,
	short part,
	Rect *growbounding);

int listwindow_setactive(WindowRef window,
	Boolean active);

int listwindow_select(WindowRef window,
	int32 firstLine,
	int32 endLine,
	int32 oStart,
	int32 oEnd,
	Boolean adjust);
int listwindow_getsel(WindowRef window,
	int32 *firstLine, /* line - can be NULL */
	int32 *endLine, /* can be NULL */
	int32 *oStart, /* can be NULL */
	int32 *oEnd); /* can be NULL */

int listwindow_copy(WindowRef window);

int listwindow_settexthook(WindowRef window,
	lwTextHook hook,
	int lineCount); /* optional */

int listwindow_setclickhook(WindowRef window,
	lwClickHook hook);

/* returns temporary pointer to stored text, possibly generated if hook is installed */
const char *listwindow_gettext(WindowRef window,
	int32 *z, /* can be NULL */
	int i); /* line index - pass negative for all */
#endif
