/*
 macos/macdialogs.h

  Created: 2008.07.23
  Revised: 2010.05.21

Tuesday, September 29, 2009 4:42 PM
  added getlastfilepath.
Monday, April 19, 2010 3:48 AM
  getfilepath now features filter index return, even now useful for putfile.
*/

#ifndef __MACDIALOGS_H
 #define __MACDIALOGS_H

 /* ---- constants */

 #define alertid_fatal			27309 /* calls ExitToShell */

 #define dialogid_error			80
 #define dialogid_note			81
 #define dialogid_accept		82
 #define dialogid_confirm		83
 #define dialogid_about			84
 #define dialogid_value			85
 #define dialogid_inputsetup		160 /* was _getfilepath */
 #define dialogid_palettesetup		161
 #define dialogid_customget		320
 #define dialogid_customput		321

 #define popupmenuid_myfilter		4751

 #define std_ok				1
 #define std_cancel			2 /* only sometimes */
 #define confirm_yes			1
 #define confirm_no			2

 #ifndef __SHELL_H
  #include "shell.h"
 #endif

 #if !defined(rez) && !defined(derez)

 /* ---- prototypes */

  void splash(void);

  short showalert(short alertid,
	const uint8 *p1,
	const uint8 *p2,
	const uint8 *p3);

/* only passing a non-negative value for defaultitem keeps the dialog in front */
/* defaultbits starts at item #2, bits 0b10 hides an item, 0b11 disables */
/* passed pointer within defaulttext is not disposed; dispose manually */
/* returned pointer in defaulttext is a realloc pointer; dispose when done */
  short showdialog(short dialogid,
	short defaultitem, /* which item to put a round rectangle on; pass zero for n/a */
	const uint8 *title, /* sets the title of the dialog window; can be NULL */
	const char *caption, /* C string that fills the first text box; can be NULL */
	const uint8 *p1, /* extra parameter for ^0; can be NULL */
	uint32 *controlbits, /* for checkboxes, etc. (two bits per item); can be NULL */
	char **defaulttext); /* default text result (in second text box); can be NULL */

  double valuedialog(const uint8 *title,
	double current,
	double lower,
	double higher,
	boolean real);

  typedef enum
  { /* getfilepath types */
	gfp_putfile,
	gfp_getfile,
	gfp_getdirectory,
	gfp_getany
  } gfptype;

  enum
  { /* getfilepath flags */
	gfpf_hidereadonly	= 1L<<7,  /* hide read-only files */
	gfpf_filemustexist	= 1L<<15, /* hide failed aliases, offline files, etc. */
	gfpf_showinvisiblefiles	= 1L<<16, /* otherwise invisible files are hidden */
	gfpf_showinvisibledirs	= 1L<<17  /* otherwise invisible dirs are hidden */
	/* other bits are reserved for future use */
  };

/* returns path allocated with realloc; dispose with realloc(...,0) */
  char *getfilepath(gfptype type,
	const uint8 *title, /* title of the dialog window titlebar; can be NULL */
	const uint8 *select, /* sets the select button title; can be NULL */
	const char *initialdir, /* starting directory; can be NULL */
	const char *initialname, /* starting name in field; can be NULL */
	const char *filterlist, /* null delimited parameter (similar to Win32); can be NULL */
	uint16 *nameoffset, /* offset to object name (file or directory); can be NULL */
	uint16 *filterindex, /* initial selected list index, returned; can be NULL */
	uint32 flags);

  #ifdef mac
/* uses FSSpec data obtained in getfilepath, returns the original copy (do NOT dispose) */
   const char *getlastfilepath(short *vrefnum, /* can be NULL */
	long *parid, /* can be NULL */
	uint8 *name); /* can be NULL */
   const char *setlastfilepath(short vrefnum,
	long parid,
	const uint8 *name); /* can be NULL */
  #endif

/* palettedialog.c */
  int PaletteDialog(
	uint8 *pickpel, /* NULL for config */
	int colc);

/* ---- program routines */

  #ifdef __DIALOGS__
   int dialogupdate(DialogRef dialog);

   int standardmodal(DialogRef dialog,
	struct EventRecord *event);

   /* may call standardmodal */
   int dialogselect(DialogRef dialog,
	struct EventRecord *event);

   DialogRef getnewdialog(int dialogid,
	int defaultitem,
	WindowRef frontwindow);

   int setcontrolitem(DialogRef dialog,
	int item,
	int value,
	int minimum,
	int maximum,
	int hilite,
	const UInt8 *title);

   void setdialogitem_pstr(DialogRef dialog,
	short item,
	const UInt8 *str);

   int setbinaryitem(DialogRef dialog,
	short item,
	int value);
   int getbinaryitem(DialogRef dialog,
	short item);

   int animatebuttonitem(DialogRef dialog,
	int item);
  #endif

 #endif/*!defined(rez) && !defined(derez)*/
#endif/*!defined(__MACDIALOGS_H)*/
