/*
 macos/macgeneral.c
  cross-platform stuff

  Created: 2008.06.30
  Revised: 2014.01.21
*/

#include <stdio.h>
#include <stdlib.h>
#include <FSpio.h>
#include <errno.h>

#include <Resources.h> /* #include's MacMemory.h */
#include <Menus.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Processes.h>

#include "shell.h"

#include "../common/dface.h"

#include "shellmenus.h"

/* ---- filesystem */

char *makepathforfsspec(const FSSpec *file,
	int *length)
{
	char		*buf = NULL;
	int		z;

	if ((buf=realloc(buf,512))==NULL)
		{z = 0; goto fail;}

	if ((z=spec2path(file,buf,511))<=0)
	{
		buf = realloc(buf,0);
		if (z!=0)
			fprintf(stderr,"## makepathforfsspec: file system error (#%d)\n",
				z);
		return NULL;
	}

	buf[511] = 0; /* manual termination is necessary */

	{
		void		*t;

		if ((t=realloc(buf,z+1))==NULL)
		{
		fail:
			fprintf(stderr,"## makepathforfsspec: (re)alloc failed (#%d)\n",
				MemError());
		} else
			buf = t;
	}

	if (length!=NULL)
		*length = z;

	return buf;
}

#if !TARGET_API_MAC_CARBON && __MRC__
 #undef fileno
int fileno(FILE *iop);
int fileno(FILE *iop)
{
	return iop->_file;
}
 #undef errno
int *__libGlue_errno(void);
int *__libGlue_errno()
{
	return &errno;
}

FILE *__libGlue_getStdStreams(int d);
FILE *__libGlue_getStdStreams(int d)
{
	if (!d)
		return stdin;
	if (!--d)
		return stdout;
	return stderr;
}

FILE *FSp_fopen(const FSSpec *spec,
	const char *mode)
{
	Str255		name;

	FILE		*file;
	long		dir;
	short		vol;
	OSErr		err;

	if ((err=HGetVol(name,&vol,&dir))==noErr)
		HSetVol(NULL,spec->vRefNum,spec->parID);

	memcpy(name,1+spec->name,*spec->name);
	name[*spec->name] = 0;
	file = fopen((char*)name,mode);

	if (err==noErr)
		HSetVol(NULL,vol,dir);

	return file;
}

void FSp_fsetfileinfo(const FSSpec *spec,
	unsigned long newcreator,
	unsigned long newtype)
{
	FInfo		info;

	if (FSpGetFInfo(spec,&info)!=noErr)
		return;
	info.fdType = newtype;
	info.fdCreator = newcreator;
	FSpSetFInfo(spec,&info);
}
#endif

/* ---- strings (from MyCSeries 0.4.1) */

#define scanstrlist(s,z,i,x)		do{Handle _h; int _x; \
	if((_x=x)<0 || (_h=GetResource('STR#',i))==NULL) {z=0; break;} \
	s=(uint8*)*_h; if(*((uint16*)s)++<(unsigned)++_x) {z=0; break;} \
	z=*s++; if(--_x==0) break; do{s+=z,z=*s++;}while(--_x!=0);}while(0)

unsigned char *getpstr(void *d,
	short r,
	int x)
{
	long		z;
	uint8		*s = NULL; /* not necessary considering unset -> zero length */

	scanstrlist(s,z,r,x);
	_acpy(1+(uint8*)d,s,z);
	*(uint8*)d = z;

	return d;
}

int getcstr(char *d,
	short r,
	int x)
{
	long		z;
	uint8		*s = NULL;

	scanstrlist(s,z,r,x);
	_acpy(d,s,z);
	d[z] = 0;

	return z;
}

int countstrlist(short r)
{
	Handle		h;
	uint16		i = 0;

	SetResLoad(0);
	h = GetResource('STR#',r);
	SetResLoad(1);
	if (!h)
		return 0;
	ReadPartialResource(h,0,&i,2);

	return i;
}

#ifndef env68k
const void *memrchr(const void *s,
	int c,
	size_t z)
{
	long int	r = 1;

	if ((signed)z<=0)
		return NULL;
	do
	{
		if (*((const uint8*)s)++==(uint8)c)
			r = (long int)s;
	} while(--z!=0);

	return (const void*)(r-1);
}

unsigned char *cpstrcpy(void *dst,
	const void *src,
	unsigned short max)
{
	const uint8	*s;
	uint8		*d;
	int		x, len;

	if ((d=(uint8*)dst)==NULL)
		return NULL;
	*d++ = 0;
	if ((s=src)==NULL)
		return NULL;
	if ((len=max)==0)
		goto end;
	if (max>255)
		len = 255;
	x = len;

	do
	{
		if (0==(*d++=*s++))
			break;
	} while(--x!=0);

	*(uint8*)dst = len-=x;
end:
	return dst;
}

void memrem(void *p,
	long c,
	const unsigned char *t)
{
	uint8		*d;

	if ((d=p)==NULL || c<=0)
		return;
	d+= c;
	if (c&1)
		*d = t[*--d];
	if ((c>>=1)==0)
		return;
	do{*d = t[*--d]; *d = t[*--d];} while(--c!=0);
}
#endif

/* memcpy16 exists for performance reasons; it ignores 16-byte boundaries */
#ifdef envppc
 #define COPY16(d,s) \
	do *((Float64*)(d))++=*((Float64*)(s))++, \
		*((Float64*)(d))++=*((Float64*)(s))++; \
	while(0)
#else
 #define COPY16(d,s) \
	(void)(*(d)++=*(s)++,*(d)++=*(s)++,*(d)++=*(s)++,*(d)++=*(s)++)
#endif
void memcpy16(void *dst,
	const void *src,
	long z)
{
	const int32	*s = (const int32*)((~15)&(long int)src);
	int32		*d;

	if ((d=(int32*)((~15)&(long int)dst))==NULL || z<=0 || (z+=15)&(long)0x80000000)
		return;
	if (z&16)
		COPY16(d,s);
	z >>= 5;
	do
	{
		COPY16(d,s); COPY16(d,s);
	} while(--z!=0);
}
#undef COPY16

/* ---- windows */

Point *localwindowpoint(WindowRef w,
	Point *p)
{
	if (!w)
		return p;

	return QDGlobalToLocalPoint(GetWindowPort(w),p);
}

int drawgrowicon(WindowRef w)
{
	GrafPtr		oldport;

	if (!w)
		return -1;
	else if (TARGET_API_MAC_CARBON)
		return 0;

	GetPort(&oldport);
	SetPortWindowPort(w);
	DrawGrowIcon(w);
	MacSetPort(oldport);

	return 1;
}

int setportwindow(WindowRef w)
{
	if (!w)
		return -1;
	SetPortWindowPort(w);

	return 1;
}

CGrafPtr getvisiblewindowport(WindowRef w)
{
	return !w || !IsWindowVisible(w) ? NULL:GetWindowPort(w);
}

Rect *getwindowbounds(WindowRef w,
	Rect *rect)
{
	if (!w)
		return rect;

	return GetPortBounds(GetWindowPort(w),rect);
}

int invalidatewindow(WindowRef w)
{ /* force a screen update */
	Rect		bounds;

	if (getvisiblewindowport(w)==NULL)
		return !w ? -1:0;
	GetPortBounds(GetWindowPort(w),&bounds);
	InvalWindowRect(w,&bounds);

	return 1;
}

void updatewindows(RgnHandle updateRgn,
	Boolean paint)
{
	WindowRef	w;

	if ((w=FrontWindow())!=NULL)
		do
		{
			CalcVis(w);
			if (updateRgn && paint)
				PaintOne(w,updateRgn);
		} while((w=GetNextWindow(w))!=NULL);
	if (updateRgn)
		PaintBehind(NULL,updateRgn);
}

#if !TARGET_API_MAC_CARBON
EXTERN_API(OSStatus)
InvalWindowRect(WindowRef w,
	const Rect *bounds)
{
	GrafPtr		oldport;

	if (!w || w==(WindowRef)-1)
		return errInvalidWindowRef;
	GetPort(&oldport);
	SetPortWindowPort(w);
	InvalRect(bounds);
	MacSetPort(oldport);

	return noErr;
}

#undef SetPortWindowPort
EXTERN_API(void)
SetPortWindowPort(WindowRef w)
{
	SetPort(w); //same ptr type
}

#undef GetWindowPort
EXTERN_API(CGrafPtr)
GetWindowPort(WindowRef w)
{
	return (CGrafPtr)w;
}

/* ---- controls */

EXTERN_API(MenuRef)
GetControlPopupMenuHandle(ControlRef control)
{
	return (**(PopupPrivateDataHandle)(**control).contrlData).mHandle;
}

EXTERN_API(UInt16)
GetControlHilite(ControlRef control)
{
	return (**control).contrlHilite;
}

EXTERN_API(Rect*)
GetControlBounds(ControlRef control,
	Rect *bounds)
{
	if (!control || !bounds)
		return NULL;
	*bounds = (**control).contrlRect;

	return bounds;
}

EXTERN_API(void)
SetControlBounds(ControlRef control,
	const Rect *bounds)
{ /* uses same method as MoveControl */
	UInt8		visible;

	if (!control || !*control || !bounds)
		return;
	visible = (**control).contrlVis;
	HideControl(control);

	(**control).contrlRect = *bounds;

	if (visible)
		ShowControl(control);
}

EXTERN_API(WindowRef)
GetControlOwner(ControlRef control)
{
	if (!control)
		return NULL;
	return (**control).contrlOwner;
}

/* ---- QuickDraw[TM] */

EXTERN_API(Rect*)
GetPortBounds(CGrafPtr port,
	Rect *bounds)
{
	if (!bounds)
		return NULL;
	*bounds = port->portRect;

	return bounds;
}

EXTERN_API(Rect*)
GetRegionBounds(RgnHandle handle,
	Rect *bounds)
{
	if (!bounds)
		return NULL;
	*bounds = (**handle).rgnBBox;

	return bounds;
}

EXTERN_API(PixMapHandle)
GetPortPixMap(CGrafPtr port)
{
	if (!(((GrafPtr)port)->portBits.rowBytes&0x8000))
		return NULL;
	return port->portPixMap;
}

EXTERN_API(const BitMap*)
GetPortBitMapForCopyBits(CGrafPtr port)
{
	if (!port)
		return NULL;
	return &((const GrafPtr)port)->portBits;
}

EXTERN_API(void)
GetGlobalMouse(Point *p)
{
	*p = *(volatile Point*)0x0830; //Mouse (Cursor)
}

EXTERN_API(Point*)
QDGlobalToLocalPoint(CGrafPtr port,
	Point *point)
{
	CGrafPtr	oldport;
	GDHandle	olddevice;

	if (!point)
		return NULL;
	GetGWorld(&oldport,&olddevice);
	SetGWorld(port,NULL);

	GlobalToLocal(point);

	SetGWorld(oldport,olddevice);

	return point;
}

EXTERN_API(Pattern*)
GetQDGlobalsGray(Pattern *pattern)
{
	if (!pattern)
		return NULL;
	*pattern = qd.gray;

	return pattern;
}

/* ---- menus */

#undef CheckMenuItem
EXTERN_API(void)
CheckMenuItem(MenuRef m,
	short i,
	Boolean c)
{
	CheckItem(m,i,c);
}

#undef CountMenuItems
EXTERN_API(UInt16)
CountMenuItems(MenuRef m)
{
	return CountMItems(m);
}

 #define EnableMenuItem EnableItem
 #define DisableMenuItem DisableItem

static UInt8 *getmenuitemptr(UInt8 *data,
	UInt8 *end,
	int index)
{
	do
	{
		if (data>=end)
			return NULL;
		if (*data==0)
			return ++data;
		data+= *data+5;
	} while(--index!=0);

	return data;
}

EXTERN_API(OSStatus)
DeleteMenuItems(MenuRef menu,
	MenuItemIndex first,
	ItemCount count)
{
	uint8		*data, *next, *end;
	long		size;
	OSStatus	result = noErr;

	if (!menu || (size=GetHandleSize((Handle)menu))<16)
		return paramErr;
	end = size+(UInt8*)*menu;
	data = (**menu).menuData;
	data+= *data++;
	if (--first && (data=getmenuitemptr(data,end,first))==NULL)
		return menuItemNotFoundErr;
	if ((next=getmenuitemptr(data,end,count))==NULL)
		next = size-1+(UInt8*)*menu, result = menuItemNotFoundErr;
	else if (next-size==(UInt8*)*menu)
		--next, result = menuItemNotFoundErr;
	_bcpy(data,next,size-(long int)next+(long int)*menu);
	size-= (long int)next, size+= (long int)data;
	SetHandleSize((Handle)menu,size);

	return result;
}

EXTERN_API(Boolean)
	IsMenuItemEnabled(MenuRef menu,
	MenuItemIndex item)
{
	if (!menu)
		return false;
	return ((**menu).enableFlags&1L<<item)!=0;
}

static void restoremenubar(void)
{
	ShowMenuBar();
}

#define DeviceList			*(volatile GDHandle*)0x8A8
#define WMgrPort			*(volatile GrafPtr*)0x9DE
#define MBarHeight			*(volatile short*)0xBAA

RgnHandle		mbargrayrgn = NULL;
GDHandle		mbargdevice = NULL;
short			savembarheight = 20;
boolean			hashidden = 0;

static GDHandle getmenubardevice(void)
{
/*	GDHandle		gdh;

	if ((gdh=DeviceList)==NULL)
		return NULL;
	do
	{
		if (!(~(**gdh).gdFlags&(1L<<screenActive|1L<<roundedDevice)))
			break;
	} while((gdh=(**gdh).gdNextGD)!=NULL);

	return gdh;
*/
	return GetMainDevice();
}

static Boolean isfrontprocess(void)
{
	ProcessSerialNumber	front, current;

	Boolean		same;

	if (GetFrontProcess(&front)!=noErr
			|| MacGetCurrentProcess(&current)!=noErr
			|| SameProcess(&front,&current,&same)!=noErr)
		same = false;
	return same;
}

void HideMenuBar()
{
	RgnHandle	argn, grayrgn = GetGrayRgn();

	if (mbargrayrgn || (mbargrayrgn=NewRgn())==NULL)
		return;
	if (!hashidden)
		atexit(restoremenubar);
	CopyRgn(grayrgn,mbargrayrgn);

	if (isfrontprocess() && (mbargdevice=getmenubardevice())!=NULL)
		if ((argn=NewRgn())!=NULL)
		{
			//RectRgn(argn,&(**argn).gdRect);
			//(**argn).rgnBBox.top -= MBarHeight;
			RectRgn(argn,&(**mbargdevice).gdRect);
			UnionRgn(argn,grayrgn,grayrgn);

			/* redraw everything underneath menubar and round rectangle */
			MacXorRgn(argn,mbargrayrgn,argn);

			updatewindows(argn,true);
			DisposeRgn(argn);
		}
	savembarheight = MBarHeight;
	MBarHeight = 0;

	hashidden = true;
}

void ShowMenuBar()
{
	RgnHandle	argn, activegrayrgn = GetGrayRgn();

	if (!mbargrayrgn)
		return;
	/* try to fill in the holes; update windows and use the WMgr for the corners */
	if ((argn=NewRgn())!=NULL)
		MacXorRgn(mbargrayrgn,activegrayrgn,argn);
	/* restore the gray region so mbargrayrgn can be used as scratch */
	CopyRgn(mbargrayrgn,activegrayrgn);
	MBarHeight = savembarheight;

	if (argn && mbargdevice && (**mbargdevice).gdFlags&1L<<roundedDevice)
	{
		GDHandle	olddevice;
		CGrafPtr	oldport, wmgr;

		GetCWMgrPort(&wmgr);
		GetGWorld(&oldport,&olddevice);
		SetGWorld(wmgr,mbargdevice);

		ClipRect(&(**mbargdevice).gdRect);
		/*paintrounddevice((**argn).rgnBBox);*/
		PaintRgn(argn);

		SetGWorld(oldport,olddevice);
	}
	/* finally dispose */
	DisposeRgn(mbargrayrgn); mbargrayrgn = NULL;

	if (argn)
	{
		updatewindows(argn,false);
		DisposeRgn(argn);
	}

	MacDrawMenuBar();
}

EXTERN_API(RgnHandle) GetDragGrayBounds()
{
	if (mbargrayrgn)
		return mbargrayrgn;
	return GetGrayRgn();
}
#else
//TARGET_API_MAC_CARBON
RgnHandle GetDragGrayBounds()
{
	return GetGrayRgn();
}
#endif

boolean ismenubarvisible()
{
#if TARGET_API_MAC_CARBON
	return IsMenuBarVisible();
#else
	return MBarHeight!=0;
#endif
}

void drawmenubar()
{
#if !TARGET_API_MAC_CARBON
	if (*(volatile Handle*)0x0A1C==NULL) /* MenuList */
		return;
#endif
	MacDrawMenuBar();
}

/* would have been setmenuitemtext, but the old routine name still exists in Menus.h */
boolean setmenuitemstr(uint16 itemcode, const uint8 *text)
{
	MenuHandle	menu;

	if ((menu=GetMenuHandle(menuid_apple+(itemcode>>6)))==NULL)
		return false;
	SetMenuItemText(menu,itemcode&63,text);
	return true;
}

static void SetMenuItemEnable(MenuRef menu,
	MenuItemIndex ix,
	Boolean enable)
{
	if (enable)
		EnableMenuItem(menu,ix);
	else
		DisableMenuItem(menu,ix);
}

boolean enablemenuitem(uint16 itemcode, boolean enabled)
{
	MenuHandle	menu;

	if ((menu=GetMenuHandle(menuid_apple+(itemcode>>6)))==NULL)
		return false;
	SetMenuItemEnable(menu,itemcode&63,enabled);

	return true;
}

boolean enablemenu(uint16 itemcode, uint8 count, boolean enabled)
{
	MenuHandle	menu;
	uint8		i = itemcode&63;

	if ((menu=GetMenuHandle(menuid_apple+(itemcode>>6)))==NULL)
		return false;
	if (enabled)
		for(; i<count; ++i)
			SetMenuItemEnable(menu,i,enabled);
	else
		for(; i<count; ++i)
			SetMenuItemEnable(menu,i,enabled);
	return true;
}

boolean checkmenuitem(uint16 itemcode, uint16 check)
{
	MenuHandle	menu;

	if ((menu=GetMenuHandle(menuid_apple+(itemcode>>6)))==NULL)
		return false;
	itemcode&= 63;
	if (check&0x8000U)
		SetItemMark(menu,itemcode,check&1 ? kMenuDiamondGlyph:0xD7);
	else
		MacCheckMenuItem(menu,itemcode,!!check);

	return true;
}

boolean checkmenu(uint16 itemcode, uint8 index, boolean invert)
{
	MenuHandle	menu;
	int		i = itemcode&63, v = index, count;

	if ((menu=GetMenuHandle(menuid_apple+(itemcode>>6)))==NULL)
		return false;
	count = CountMenuItems(menu);

	if (invert)
		for(; i<=count; ++i)
			MacCheckMenuItem(menu,i,i!=index);
	else
		for(; i<=count; ++i)
			MacCheckMenuItem(menu,i,i==index);
	return true;
}
