/*
 macos/macinput.c
  input for MacFCEU 0.9

  Created: 2008.07.02

Tuesday, September 29, 2009 10:55 PM EDST
  added screenshift to updatemouse (previous versions didnt work right!).
Tuesday, March 30, 2010 6:01 AM
  doing some clean-up (removed unused code).
  only Button is called now for direct mouse poll.
  UDLR...
Wednesday, April 14, 2010 8:12 PM
  to get movies working properly I have to use the form that was there before, storing the
 mouse data, only setting when called for (don't write final data when a movie is playing).
Thursday, April 15, 2010 2:25 AM
  shift key is now reserved for input; command key will take over for its former.
Tuesday, May 11, 2010
  added F13 cycle of user input default.
*/

#include <string.h>

#include <Events.h>
#include <MacWindows.h>

#include "shell.h"

#include "../FCEU.h"
#include "../input.h"
#include "../general.h"
#include "../video.h"
#include "../movie.h"
#include "../common/dface.h"
#include "../common/cheat.h"

#include "macinput.h"
#include "macvideo.h"
#include "macdialogs.h"
#include "config.h"
#include "memview.h"
#include "../ftrig.h"

#include "keyscan.h"

#define MK(x)				{{BUTTC_KEYBOARD},{0},{MKK(x)},1}
#define MK2(x1,x2)			{{BUTTC_KEYBOARD},{0},{MKK(x1),MKK(x2)},2}
#define MKZ()				{{0},{0},{0},0}

/* ---- constants */

static const char	*inputName[] =
{
	"<none>",
	"GamePad",
	"Zapper",
	"PwrPadA",
	"PwrPadB",
	"Arkanoid"
};

static const char	*inputNameFC[] =
{
	"<none>",
	"Arkanoid",
	"Shadow",
	"4-Player",
	"FamilyKB",
	"HyprShot",
	"Mahjong",
	"QuizKing",
	"TrainerA",
	"TrainerB",
	"OEKAKids",
	"BWorld",
	"TopRider"
};

/* ---- globals */

uint8		NoWaiting = 1;
boolean		inputCmdDown;

static uint32	JSreturn;

input_settings	inputset =
{
	0, SI_GAMEPAD,SI_GAMEPAD,SIFC_NONE,
	{ /* PowerPad A */
		MK(O), MK(P), MK(BRACKET_LEFT), MK(BRACKET_RIGHT),
		MK(K), MK(L), MK(SEMICOLON), MK(APOSTROPHE),
		MK(M), MK(COMMA), MK(PERIOD), MK(SLASH)
	},
	{ /* PowerPad B */
		MK(O), MK(P), MK(BRACKET_LEFT), MK(BRACKET_RIGHT),
		MK(K), MK(L), MK(SEMICOLON), MK(APOSTROPHE),
		MK(M), MK(COMMA), MK(PERIOD), MK(SLASH)
	},
	{ /* QuizKing */
		MK(Q), MK(W), MK(E), MK(R), MK(T), MK(Y),
	},
	{ /* FTrainer */
		MK(O), MK(P), MK(BRACKET_LEFT), MK(BRACKET_RIGHT),
		MK(K), MK(L), MK(SEMICOLON), MK(APOSTROPHE),
		MK(M), MK(COMMA), MK(PERIOD), MK(SLASH),
	},
	{ /* HyperShot */
		MK(Q), MK(W), MK(E), MK(R),
	},
	{ /* Mahjong */
		MK(Q), MK(W), MK(E), MK(R), MK(T),
		MK(A), MK(S), MK(D), MK(F), MK(G), MK(H), MK(J), MK(K), MK(L),
		MK(Z), MK(X), MK(C), MK(V), MK(B), MK(N), MK(M),
	},
	/* A, B, Select, Start, Up, Down, Left, Right */
	{ /* Gamepad 1 - was Opt, Ctl, Tab, Ret, Up, Down, Left, Right */
		MK(Z), MK(X), MK(TAB), MK(Q),
		MK(R), MK(F), MK(D), MK(G),
	}, /* FCEUX 2.1 does KP3, KP2, TAB, ENTER, W, Z, A, S */
	{ /* Gamepad 2 */
		MK(N), MK(M), MK(APOSTROPHE), MK(RETURN),
		MK(O), MK(L), MK(K), MK(SEMICOLON),
	},
	{ /* Gamepad 3 */
		MK(INSERT), MK(PAGEUP), MK(BACKSLASH), MK(BACKSPACE),
		MK(HOME), MK(END), MK(DELETE), MK(PAGEDOWN),
	},
	{ /* Gamepad 4 */
		MK(KP0), MK(KP_PERIOD), MK(KP_SLASH), MK2(KP_ASTERISK,KP_ENTER),
		MK(KP8), MK(KP5), MK(KP4), MK(KP6),
	},
	{ /* FKeyboard - TAB for left ALT? */
		MK(F1), MK(F2), MK(F3), MK(F4), MK(F5), MK(F6), MK(F7), MK(F8),
		MK(1), MK(2), MK(3), MK(4), MK(5), MK(6), MK(7), MK(8), MK(9), MK(0),
			MK(MINUS), MK(EQUAL), MK(BACKSLASH), MK(BACKSPACE), MK(ESCAPE),
		MK(Q), MK(W), MK(E), MK(R), MK(T), MK(Y), MK(U), MK(I), MK(O), MK(P),
			MK(GRAVE), MK(BRACKET_LEFT), MK(RETURN),
		MK(CONTROL), MK(A), MK(S), MK(D), MK(F), MK(G), MK(H), MK(J), MK(K), MK(L),
			MK(SEMICOLON), MK(APOSTROPHE), MK(BRACKET_RIGHT), MK(INSERT),
		MK(SHIFT), MK(Z), MK(X), MK(C), MK(V), MK(B), MK(N), MK(M), MK(COMMA),
			MK(PERIOD), MK(SLASH), MK(OPTION), MK(SHIFT), MK(OPTION), MK(SPACE),
		MK(DELETE), MK(END), MK(PAGEDOWN),
		MK(CURSORUP), MK(CURSORLEFT), MK(CURSORRIGHT), MK(CURSORDOWN),
	}
};

int		inputtype[3], inputtypegi[3];
static int	cspec;

int		gametype;

static uint8	QuizKingData;
static uint8	HyperShotData;
static uint32	MahjongData;
static uint32	FTrainerData;
static uint8	TopRiderData;

static uint8	BWorldData[1+13+1];

/* ---- keyboard */

static KeyMap	mykeys;

int InitKeyboard()
{
	memset(mykeys,0,sizeof(mykeys));
	return 1;
}

char *GetKeyboard()
{
	GetKeys(mykeys);
	return (char*)mykeys;
}

static int getkeystate(int x)
{
	return _bittst(mykeys,x)!=0;
}

#ifdef env68k
//use KEY and _keyonly for the constant parameter
 #define KEY(__a)			((((const uint8*)mykeys)[MKK(__a)>>3]&1<<(~MKK(__a)&7))!=0)
 #define getkeystate(x)			_bittst(mykeys,x)
#else
 #define KEY(__a)			getkeystate(MKK(__a))
#endif

static int	DIPS = 0;
static uint8	keyonce[MKK_COUNT];

#define _keyonly(__a)			(KEY(__a) ? \
	(!keyonce[MKK(__a)]?(keyonce[MKK(__a)]=1):0):(keyonce[MKK(__a)]=0))
#define keyonly(__a)			(getkeystate(MKK(__a)) ? \
	(!keyonce[MKK(__a)]?(keyonce[MKK(__a)]=1):0):(keyonce[MKK(__a)]=0))

uint8		cidisabled = 0;

static void initializeinput(void);

static uint8	recordSub = 1, recordFPF = 0;
uint8		inputSnapped; /* oh, he snapped */

int inputDoSnap(int record)
{
	static uint8	lastFPF = 3;
	int		r;

	if (inputSnapped || cidisabled)
		return 0; /* shell may have done this already; will be cleared */
	inputSnapped = 1;
	if (record)
	{
		if (recordFPF)
			goto stop;
		if ((r=valuedialog("\pSnap 1 out of every x frames...",
				lastFPF,0,255,false))==INT_NO_VALUE)
			r = 0;
		if (r!=0)
		{
			lastFPF = r;
			FCEUI_DispMessage("Snap FPF recording...");
		}
		return recordFPF=r;
	} else if (recordFPF)
	{
	stop:
		recordFPF = 0, recordSub = 1;
		FCEUI_DispMessage("Snap FPF stopped.");
		FCEU_ResetSnaps();
		return 0;
	} else
		FCEUI_SaveSnapshot();

	return 1;
}

static void dokeycommands(void)
{
extern WindowRef myfrontwindow;
	if (inputtype[2]==SIFC_FKB)
	{
		if (_keyonly(SCROLLLOCK))
			FCEUI_DispMessage("Family Keyboard %sabled.",
				(cidisabled^=1)!=0 ? "en":"dis");
		if (cidisabled)
		{
			/* remove any kind of menu-cmd key event */
			FlushEvents(keyDownMask|keyUpMask|autoKeyMask,everyEvent);
			return;
		}
	}

	if ((inputCmdDown=KEY(COMMAND))!=0)
	{
		int	delta;

		if ((delta=_keyonly(MINUS)*2+_keyonly(EQUAL))==2) /* prev. KP; 2010.04.25 */
			delta =-1; /* down only */
		else if (delta==1)
			delta = 1; /* up only */
		if (delta)
		{
			if (KEY(SHIFT))
				delta+= delta<<2; /* blast it */
			if (adjust_soundvolume(delta)>=0)
				FCEUI_DispMessage("Volume: %d%%",
					soundvolume);
		}
		if (_keyonly(F3))
			FCEUI_SetRenderDisable(-1,4);
		if (gametype!=GIT_NSF)
			if (_keyonly(F5))
				FCEUI_SaveMovie(NULL);
			else if (_keyonly(F7))
				LoadStateOrMovie(NULL,1,1); //FCEUI_LoadMovie(NULL);
		if (_keyonly(F12)) //was F13
		{
			if (KEY(SHIFT))
			{ /* cycle #2 */
				if (++userinputtype[2]>SIFC_TOPRIDER)
					userinputtype[2] = 0;
				initializeinput();
				FCEUI_DispMessage("UsrFC-%s;%s",
					inputNameFC[userinputtype[2]],inputNameFC[inputtype[2]]);
			} else
			{ /* cycle #1 */
				if (++userinputtype[1]>SI_ARKANOID)
					userinputtype[1] = 0;
				initializeinput();
				FCEUI_DispMessage("Usr#1-%s;%s",
					inputName[userinputtype[1]],inputName[inputtype[1]]);
			}
		}
		//if (_keyonly(F9))
		//	inputDoSnap(1);
	} else
	{
		if (_keyonly(F3))
			FCEUI_SetRenderDisable(-1,2);
		if (_keyonly(F4)) /* now since Shift's reserved, the Cmd does bkgnd (above) */
			FCEUI_SetRenderDisable(2,-1);
		if (_keyonly(F12)) //was F13
		{ /* cycle #0 */
			if (++userinputtype[0]>SI_ARKANOID)
				userinputtype[0] = 0;
			initializeinput();
			FCEUI_DispMessage("Usr#0-%s;%s",
				inputName[userinputtype[0]],inputName[inputtype[0]]);
		}
		if (gametype!=GIT_NSF)
			if (_keyonly(F5))
				FCEUI_SaveState(NULL);
			else if (_keyonly(F7))
				LoadStateOrMovie(NULL,0,1); //FCEUI_LoadState(NULL);
		//if (_keyonly(F9))
		//	inputDoSnap(0);
	}

	NoWaiting&=~1;
	if (KEY(GRAVE) || KEY(0D)) //&& nocount)
		NoWaiting|= 1;

	if (gametype==GIT_FDS)
		if (_keyonly(F6))
			FCEUI_FDSSelect();
		else if (_keyonly(F8))
			FCEUI_FDSInsert(0);

	if (_keyonly(F11))
		FCEU_ResetMessages(-1); /* to manually toggle; 2010.04.14 */
	if (MemViewWindow && myfrontwindow==MemViewWindow)
		goto clear;
	if (gametype==GIT_VSUNI)
	{
		if (_keyonly(F8))
			FCEUI_VSUniCoin();
		if (_keyonly(F6))
		{
			DIPS^= 1;
			FCEUI_VSUniToggleDIPView();
		}

		if (!(DIPS&1))
			goto DIPSless;
		if (_keyonly(1))
			FCEUI_VSUniToggleDIP(0);
		if (_keyonly(2))
			FCEUI_VSUniToggleDIP(1);
		if (_keyonly(3))
			FCEUI_VSUniToggleDIP(2);
		if (_keyonly(4))
			FCEUI_VSUniToggleDIP(3);
		if (_keyonly(5))
			FCEUI_VSUniToggleDIP(4);
		if (_keyonly(6))
			FCEUI_VSUniToggleDIP(5);
		if (_keyonly(7))
			FCEUI_VSUniToggleDIP(6);
		if (_keyonly(8))
			FCEUI_VSUniToggleDIP(7);
	} else
	{
		static uint8	bbuf[32];
		static int	bbuft;
		static int	barcoder = 0;

		if (!inputCmdDown) /* this part got in the way quickly; 2010.04.25 */
		{
			if (_keyonly(T))
				FCEUI_NTSCSelect(0);
			else if (_keyonly(H))
				FCEUI_NTSCSelect(1);
			else if (_keyonly(B)) /* added; 2010.05.09 */
				FCEUI_NTSCSelect(2);
			if (KEY(MINUS))
				FCEUI_NTSCDEC();
			if (KEY(EQUAL))
				FCEUI_NTSCINC();
		}

		if (inputtype[2]!=SIFC_BWORLD && cspec!=SIS_DETACH)
			barcoder = 0;
		else if (_keyonly(F8))
			if (barcoder^=1)
				{bbuft = 0; FCEUI_DispMessage("Enter Barcode");}
			else
			{
				if (inputtype[2]==SIFC_BWORLD)
				{
					strcpy((char*)&BWorldData[1],(const char*)bbuf);
					BWorldData[0] = 1;
				} else
					FCEUI_DatachSet(bbuf);
				FCEUI_DispMessage("Barcode Entered");
			}
	DIPSless:
	#define SSMi(x)		do{ \
		if(!keyonly(x)) \
			break; \
		if(barcoder) { \
			if(bbuft<13) \
				bbuf[bbuft++]='0'+x, bbuf[bbuft]=0; \
			FCEUI_DispMessage("Barcode: %s",bbuf); \
		} else if(inputCmdDown) \
			FCEUI_SelectMovie(x); \
		else \
			FCEUI_SelectState(x); \
	}while(0)
		SSMi(0);
		SSMi(1);
		SSMi(2);
		SSMi(3);
		SSMi(4);
		SSMi(5);
		SSMi(6);
		SSMi(7);
		SSMi(8);
		SSMi(9);
	#undef SSMi
	}

	if (inputCmdDown)
	clear:	InitKeyboard(); /* other menus in menubar... */
}

/* ---- general */

int DTestButton(ButtConfig *bc)
{
	int		x, t;

	for(x=0; x<bc->NumC; ++x)
		if ((t=bc->ButtType[x])==BUTTC_KEYBOARD)
		{
			if (getkeystate(bc->ButtonNum[x]))
				return 1;
		} else if (t==BUTTC_MOUSE) /*whatever*/
			return Button();

	return 0;
}

/* ---- hypershot */

static void UpdateHyperShot(void)
{ /* unrolled */
	HyperShotData = 0;

	if (DTestButton(&HyperShotButtons[0]))
		HyperShotData|= 1;
	if (DTestButton(&HyperShotButtons[1]))
		HyperShotData|= 2;
	if (DTestButton(&HyperShotButtons[2]))
		HyperShotData|= 4;
	if (DTestButton(&HyperShotButtons[3]))
		HyperShotData|= 8;
}

/* ---- mahjong */

static void UpdateMahjong(void)
{
	unsigned	x, b;

	x = 0, b = 1, MahjongData = 0;
	do
		if (DTestButton(&MahjongButtons[x]))
			MahjongData|= b;
	while(b+=b,++x<21);
}

/* ---- quizking */

static void UpdateQuizKing(void)
{ /* unrolled */
	QuizKingData = 0;

	if (DTestButton(&QuizKingButtons[0]))
		QuizKingData|= 1;
	if (DTestButton(&QuizKingButtons[1]))
		QuizKingData|= 2;
	if (DTestButton(&QuizKingButtons[2]))
		QuizKingData|= 4;
	if (DTestButton(&QuizKingButtons[3]))
		QuizKingData|= 8;
	if (DTestButton(&QuizKingButtons[4]))
		QuizKingData|= 0x10;
	if (DTestButton(&QuizKingButtons[5]))
		QuizKingData|= 0x20;
}

/* ---- toprider */

ButtConfig	TopRiderButtons[8] =
{
	MK(Q), MK(W), MK(E), MK(R), MK(T), MK(Y), MK(U), MK(I),
};

static void UpdateTopRider(void)
{ /* unrolled */
	TopRiderData = 0;

	if (DTestButton(&TopRiderButtons[0]))
		TopRiderData|= 1;
	if (DTestButton(&TopRiderButtons[1]))
		TopRiderData|= 2;
	if (DTestButton(&TopRiderButtons[2]))
		TopRiderData|= 4;
	if (DTestButton(&TopRiderButtons[3]))
		TopRiderData|= 8;
	if (DTestButton(&TopRiderButtons[4]))
		TopRiderData|= 0x10;
	if (DTestButton(&TopRiderButtons[5]))
		TopRiderData|= 0x20;
	if (DTestButton(&TopRiderButtons[6]))
		TopRiderData|= 0x40;
	if (DTestButton(&TopRiderButtons[7]))
		TopRiderData|= 0x80;
}

/* ---- [f]trainer */

static void UpdateFTrainer(void)
{
	unsigned	x, b;

	x = 0, b = 1, FTrainerData = 0;
	do
		if (DTestButton(&FTrainerButtons[x]))
			FTrainerData|= b;
	while(b+=b,++x<12);
}

/* ---- emulated keyboard */

static uint8	fkbkeys[0x48];

static void UpdateFKB(void)
{
	int		x;

	x = 0;
	do
		fkbkeys[x] = DTestButton(&fkbmap[x]);
	while(++x<0x48);
}

/* ---- mouse */

Point		winmousept;

boolean updatemouse()
{
	CGrafPtr	port;
	const PixMap	**pm;
	const BitMap	*windowmap;
	boolean		result;

	if ((port=getvisiblewindowport(gamewindow))==NULL)
		return 0;

	GetGlobalMouse(&winmousept);

	if ((pm=GetPortPixMap(port))==NULL)
		windowmap = GetPortBitMapForCopyBits(port);
	else
		windowmap = (const BitMap*)*pm;

	/* map top/left for window are negative (covers remaining screen space) */
	winmousept.h = (winmousept.h+windowmap->bounds.left)>>screenshift;
	winmousept.v = (winmousept.v+windowmap->bounds.top)>>screenshift;

	/* gameinitbounds holds the true size of the window (dont just use 256x240) */
	result = (unsigned)winmousept.v<<screenshift<gameinitbounds.bottom-gameinitbounds.top
		&& (unsigned)winmousept.h<<screenshift<gameinitbounds.right-gameinitbounds.left;

	/* for 7/6 ratio; 2011.03.05 */
	if (screenshift&0x80)
		winmousept.h = ufract32mul(winmousept.h,(6*4294967296.0+6)/7);

	return result;
}

/* ---- gamepad */

static void UpdateGamepad(void)
{
	static uint8	rapid = 0;
	uint32		JS = 0, wg = 0, wgb = 0;

	rapid^= 1;

#define TestAddPadButton(x)	do{ \
	if(DTestButton(&GamePadConfig[wg][x])) \
		JS|=(1<<(x&7))<<wgb; \
}while(0)
	do
	{ /* unrolled */
		TestAddPadButton(0);
		TestAddPadButton(1);
		TestAddPadButton(2);
		TestAddPadButton(3);
		TestAddPadButton(4);
		TestAddPadButton(5);
		TestAddPadButton(6);
		TestAddPadButton(7);

		// rapid-fire A and B
		if (rapid)
		{
			TestAddPadButton(8);
			TestAddPadButton(9);
		}
	} while(wgb+=8,++wg<4);
#undef TestAddPadButton
	/* correct conflicting keys (simultaneous up+down) -- process unrolled */
	if (!allowUDLR)
	{
		wgb = ~JS;
		if (!(wgb&0x30))
			JS&=~0x30;
		if (!(wgb&0xC0))
			JS&=~0xC0;
		if (!(wgb&0x3000))
			JS&=~0x3000;
		if (!(wgb&0xC000))
			JS&=~0xC000;
		if (!(wgb&0x300000))
			JS&=~0x300000;
		if (!(wgb&0xC00000))
			JS&=~0xC00000;
		if (!(wgb&0x30000000))
			JS&=~0x30000000;
		if (!(wgb&0xC0000000u))
			JS&=~0xC0000000u;
	}

	JSreturn = JS;
}

/* ---- powerpad */

static uint32	powerpadbuf[2];

static uint32 UpdatePPadData(int w)
{
	ButtConfig	*ppadtsc = powerpadsc[w];
	uint32		r, b;
	int		x;

	x = 0, r = 0, b = 1;
	do
		if (DTestButton(&ppadtsc[x]))
			r|= b;
	while(b+=b,++x<12);

	return r;
}

/* ---- input GI */

static uint32	MouseData[3];

void FCEUD_UpdateInput()
{ /* Apple's input always active (interrupt time) */
	int		x, t = 0;

	if (recordFPF && !--recordSub)
		recordSub = recordFPF, FCEUvideoShot = 2;

	if (suspended) /* background fix; 2010.04.03 */
		return;

	GetKeys(mykeys);

	dokeycommands();

	if (FCEUMOV_IsPlaying())
		return;

	x = 0;
	do
		switch(inputtype[x])
		{
		case SI_GAMEPAD: t|= 1; break;
		case SI_ARKANOID:
		case SI_ZAPPER: t|= 2; break;
		case SI_POWERPADA:
		case SI_POWERPADB: powerpadbuf[x] = UpdatePPadData(x);
		}
	while(++x<2);

	switch(inputtype[2])
	{
	case SIFC_ARKANOID:
	case SIFC_OEKAKIDS:
	case SIFC_SHADOW: t|= 2; break;
	case SIFC_FKB:
		if (cidisabled)
			UpdateFKB();
		break;
	case SIFC_HYPERSHOT: UpdateHyperShot(); break;
	case SIFC_MAHJONG: UpdateMahjong(); break;
	case SIFC_QUIZKING: UpdateQuizKing(); break;
	case SIFC_FTRAINERB:
	case SIFC_FTRAINERA: UpdateFTrainer(); break;
	case SIFC_TOPRIDER: UpdateTopRider();
	}

	if (t&1)
		UpdateGamepad();
	if (t&2)
		MouseData[0] = winmousept.h,
		MouseData[1] = winmousept.v,
		MouseData[2] = Button();
}

static void initializeinput()
{
	void		*dataptr;
	int		x;

	inputtype[0] = userinputtype[0];
	inputtype[1] = userinputtype[1];
	inputtype[2] = userinputtype[2];

	if (inputtypegi[0]>=0)
		inputtype[0] = inputtypegi[0];
	if (inputtypegi[1]>=0)
		inputtype[1] = inputtypegi[0];
	if (inputtypegi[2]>=0)
		inputtype[2] = inputtypegi[0];

	x = 0;
	do
	{
		switch(inputtype[x])
		{
		case SI_POWERPADA:
		case SI_POWERPADB: dataptr = x+powerpadbuf; break;
		case SI_GAMEPAD: dataptr = &JSreturn; break;
		case SI_ZAPPER: /*attrib = fullscreen;*/
		case SI_ARKANOID: dataptr = MouseData; break;

		default: dataptr = 0;
		}
		FCEUI_SetInput(x,inputtype[x],dataptr,0);
	} while(++x<2);

	dataptr = 0;
	switch(inputtype[2])
	{
	case SIFC_SHADOW:
	case SIFC_OEKAKIDS: dataptr = MouseData; break;
	case SIFC_FKB: dataptr = fkbkeys; break;
	case SIFC_HYPERSHOT: dataptr = &HyperShotData; break;
	case SIFC_MAHJONG: dataptr = &MahjongData; break;
	case SIFC_QUIZKING: dataptr = &QuizKingData; break;
	case SIFC_TOPRIDER: dataptr = &TopRiderData; break;
	case SIFC_BWORLD: dataptr = BWorldData; break;
	case SIFC_FTRAINERA:
	case SIFC_FTRAINERB: dataptr = &FTrainerData;
	}

	FCEUI_SetInputFC(inputtype[2],dataptr,0);
	FCEUI_DisableFourScore(eoptions&EO_NOFOURSCORE);
}

void parseinputgi(FCEUGI *gi)
{
	if (gi!=NULL)
	{
		inputtypegi[0] = gi->input[0];
		inputtypegi[1] = gi->input[1];
		inputtypegi[2] = gi->inputfc;
		cspec = gi->cspecial;
		gametype = gi->type;

		initializeinput();
	} else
		inputtypegi[0] = inputtypegi[1] = inputtypegi[2] =-1,
		cspec = gametype = 0;
}

void config_inputdialog()
{ //hey! the select and start buttons are in the wrong visual order...
	if (input_dialog(GamePadConfig,0)>0)
		input_dialog(GamePadConfig+2,1); //just damn ugly.  but i have to.
}
