/*
 macos/macsound.c
  sound for MacFCEU

  Created: 2008.07.02
  Revised: 2015.08.29

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Tuesday, September 16, 2008 10:52 PM EDST
  because playback end starts at buffered end, double buffering can be removed
Tuesday, January 19, 2009 11:43 PM
  sound rate NTSC mod (ConsoleDownRate) moved to sound.c.
Saturday, October 24, 2009 9:08 PM
  officially getting rid of the non-68k callback stuff.
*/

#include <stdio.h>
//#include <stdlib.h>

#include <Sound.h>

#include "shell.h"

#include "../FCEU.h"
#include "../ftrig.h"
#include "../common/dface.h"
#include "../WAVE.h"  /* for WAVE_DefaultRate */
#include "../memory.h"

#include "shellmenus.h"

#ifdef env68k
 #pragma segment Sound
#endif

#define HQ_PLAYBACK		1

//#ifndef USE_DIRECT_COMPONENT_SOUND -- faster, less compatible

/* ---- macros */

#define _bcpy2(d,s,c)		do{void *__d=d; const void *__s=s; long __z=c; \
	if(__z&1) *((uint16*)__d)++=*((uint16*)__s)++; \
	if(__z&2) *((uint32*)__d)++=*((uint32*)__s)++; \
	if(__z&4) *((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++; \
	if(__z&8) *((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++, \
		*((uint32*)__d)++=*((uint32*)__s)++; \
	if ((__z&=-16)!=0) do{ \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
		*((uint32*)__d)++=*((uint32*)__s)++,*((uint32*)__d)++=*((uint32*)__s)++; \
	} while((__z+=-16)!=0);}while(0)

#if TARGET_CPU_68K
 #define __FIXEDEXT80(x)	((x)/65536.0)
#else
 #define __FIXEDEXT80(x)	{0x400E, x, 0}
#endif

/* ---- globals */

#define BUFFER_SIZE		16384 //13472

static int16	playbuffer[2][BUFFER_SIZE];

static volatile int	buftotal;

static SndCallBackUPP	userroutine;

static SndChannel	mychannel =
{
	NULL, NULL, NULL, /* next channel, first mod, callback proc */
	0, 0, /* user info (buffer index and mask),  wait */
	{0}, /* command in progress */
	0, 2, 0, 0, /* flags, qLength, qHead, qTail */
};

static CmpSoundHeader	sndheader =
{
	0, /* sample data */
	1, rate22khz, 0, 0, /* channel count, sample rate, loop start, end */
	cmpSH, 60, /* encoding, base note */
	0, __FIXEDEXT80(rate22khz), /* frame count, floating sample rate */
	NULL, /* marker sync track */
	k16BitNativeEndianFormat,
	0, NULL, NULL, 0, /* reserved, state block ptr, leftover samples, codec ID */
	0, 0, 16, /* packet size, synth ID, sample size */
};
#define SAMPLES			(int16*)sndheader.samplePtr

int		minsndcount;

#if HQ_PLAYBACK
static uint32	halfds;
#endif

/* ---- code */

int KillSound()
{
	SndCommand	command;

	command.cmd = bufferCmd;
	command.param1 = 0;
	command.param2 = (long int)&sndheader;

	if (!mychannel.firstMod || !SAMPLES)
		return 0;
	mychannel.userInfo = 0; /* no callback play */
	SAMPLES = playbuffer[0];
	memset(playbuffer,sndheader.encode==stdSH ? 0x80:0,32);
	sndheader.numFrames = 16;
	if (sndheader.encode==stdSH)
		((SoundHeader*)&sndheader)->length = 16;
	SndDoImmediate(&mychannel,&command);

	SndDisposeChannel(&mychannel,true);
	mychannel.firstMod = 0;
#if CALL_NOT_IN_CARBON
	if (VMMask&1)
		UnlockMemory(playbuffer,sizeof(playbuffer));
#endif
	DisposeSndCallBackUPP(userroutine); userroutine = NULL;

	return 1;
}

static pascal void mychannelcb(SndChannelPtr chan,
	SndCommand *cmd)
{
#pragma unused(cmd)
	if (!chan->userInfo || !SAMPLES)
		return;
	if (!buftotal)
		{chan->userInfo = 0; return;} /* need samples! */
	if (sndheader.encode==stdSH)
		((SoundHeader*)&sndheader)->length = buftotal;
	else
		sndheader.numFrames = buftotal>>(sndheader.numChannels-1);
	cmd->cmd = bufferCmd;
	cmd->param2 = (long int)&sndheader;
	SndDoCommand(&mychannel,cmd,true);
	cmd->cmd = callBackCmd;
	cmd->param2 = 0;
	SndDoCommand(&mychannel,cmd,true);
	/* swap buffers, and reset counters */
	if (SAMPLES==playbuffer[0])
		SAMPLES = playbuffer[1];
	else
		SAMPLES = playbuffer[0];
	buftotal = 0;
}

/* null git now taken as normal */
/* (soundon/soundvolume determine if channel should be created) */
int InitSound(FCEUGI *git)
{
	SndChannel	*mychannelptr = &mychannel;
	UnsignedFixed	ratemultiplier = 0x10000u;
	int		samplerate, chancnt, shift;
	OSErr		err;

	if (mychannel.firstMod)
		KillSound();

	if (!soundon || !soundvolume) /*git==NULL)*/
	{
		FCEUI_SetSoundActiveChannels(0);
		FCEUI_Sound(4998);
		return 0;
	}
	FCEUI_SetSoundActiveChannels(sndactivechannels);

	/* fix sample rate variables */
	samplerate = sndset.rate*16+.5;
	if (!git)
		;
	else if (!git->soundrate)
		git->soundrate = samplerate;
	else if (!samplerate)
		sndset.rate = (samplerate=git->soundrate)/16.;

	if (samplerate<1000*16 || samplerate>MAXIMUM_SRATE)
	{ /* fix range */
		sndset.rate = samplerate<1000*16 ? 1000.:MAXIMUM_SRATE/16.;
		samplerate = sndset.rate*16+.5;
		updatemenus();
	#if HAVE_HQ
		FCEUI_SetSoundQuality(0);
	#endif
	}

	chancnt = git && git->soundchan>=2 ? 2:1;

	samplerate = FCEUI_Sound2(samplerate);
	/* note that the rate now has a 4 bit fraction! */
	WAVE_DefaultRate = (samplerate+8)>>4;

	if (!soundavailable)
		return 1;
	if (!s16baud)
		sndheader.encode = stdSH;
#if CALL_NOT_IN_CARBON
	if (VMMask&1)
		LockMemory(playbuffer,sizeof(playbuffer));
#endif
	err =-1;
	if ((userroutine=NewSndCallBackUPP(mychannelcb))==NULL
			|| (err=SndNewChannel(&mychannelptr,sampledSynth,chancnt==2 ?
			initStereo:initMono,userroutine))!=noErr)
	{
		if (err==-32767 || err==noHardwareErr || err==handlerNotFoundErr)
		{
			soundavailable = 0;
			mychannel.firstMod = 0; /* not automatically cleared */
			FCEU_printf("## InitSound: no sound %s for sound channel\n",
				err==handlerNotFoundErr ? "handler":"device");
		} else
			FCEU_printf(
"## InitSound: failed to create sound channel (%s#%d) - refer to MacErrors.h\n",
				err/1000==-8 ? "codec error ":err/100==-2 ? "SM error ":"",err);
		return -1;
	}
	/* support for sample rates >= 65536 */
	shift = cntlz32(samplerate);
	sndheader.sampleRate = samplerate<<MIN(16-4,shift);
#if HQ_PLAYBACK
	halfds = 0;
	if (samplerate>=88200u*16) /* Classic SM rate converter has only 16-bit fractional precision */
		halfds = 0x40000000u, ++shift, samplerate = samplerate+1>>1;
#endif
	if (s16baud)
	{ /* rateMultiplier support (better quality than my bints16) */
		if (shift<16-4 && (ratemultiplier<<=16-4+-shift)==0)
			{ratemultiplier =-1ul; /*DebugStr("\pDemon, be Gone!");*/}
	} else if (samplerate>=65536*16)
		if (samplerate==65536*16)
			sndheader.sampleRate =~0; /* ==65536 in rounded fixed muldiv */
		else /* old system setup - likely doesn't carry 44100 rate */
			halfds = 0x40000000u;
	SAMPLES = playbuffer[0];
	sndheader.numChannels = chancnt;

	if (ratemultiplier!=0x10000ul)
	{
		SndCommand	command;

		command.cmd = rateMultiplierCmd;
		command.param1 = 0;
		command.param2 = ratemultiplier;
		err = SndDoImmediate(&mychannel,&command);
	}
	/* throttle always on now; 2012.01.03 04:46 ET */
	RefreshThrottleFPS();
	buftotal = 0;

	return 1;
}

void bds16_8(uint8 *d,
	const int16 *s,
	int32 c)
{
	int32		i;

	if ((i=c>>2)!=0)
		do
			*((int32*)d)++ = ((((uint32)((const uint8*)s)[1-BIG_ENDIAN]<<8
				|(uint32)((const uint8*)s)[3-BIG_ENDIAN])<<8
				|(uint32)((const uint8*)s)[5-BIG_ENDIAN])<<8
				|((const uint8*)s)[7-BIG_ENDIAN])^0x80808080u;
		while(s+=8,--i);
	if (c&2)
		*d++ = ((uint32)((const uint8*)s)[1-BIG_ENDIAN]<<8
			|((const uint8*)s)[3-BIG_ENDIAN])^0x8080, s+= 4;
	if (c&=1)
		*(int8*)d = 0x80+((const uint8*)s)[1-BIG_ENDIAN];
}

#if HQ_PLAYBACK
static int32 bhalf16(uint32 c,
	const int16 *s,
	int16 *_d,
	uint32 *ds)
{
	int16		*d = _d;
	int32		r = c, x, v, a, b;
	/* removed the pseudo-RMS correction (it didn't work) */
	if (*ds&1<<31 && c)
		--c, a = *s++, b = (int16)*ds, v = a+b, *d++ = v>>1;
	x = (c&1);
	if (c>>=1)
		do
			a = *s, b = s[1], v = a+b, *d++ = v>>1;
		while(s+=2,--c);
	*ds&= 1<<30;
	if (x)
		*ds|= (uint16)*s|1u<<31;
	return d-_d;
}
#endif

void writesound(const int16 *data,
	int count)
{
	SndCommand	command;
	int16		*wr;
	int32		total;

	if (!mychannel.firstMod || count<=0 || !data)
		return;
	mychannel.userInfo = 0; /* busy */
	total = buftotal;
	wr = SAMPLES;
	/* copy the whole thing, play it in chunks */
#if HQ_PLAYBACK
	if (halfds)
	{
		count+= total*2;
		if (count>BUFFER_SIZE*2-2)
			count = BUFFER_SIZE*2-2;
		count-= total*2;
		count = bhalf16(count,data,wr+total,&halfds);
	} else
#endif
	{
		count+= total;
		if (count>BUFFER_SIZE)
			count = BUFFER_SIZE;
		count-= total;
		_bcpy2(wr+total,data,count);
	}
	if (sndheader.encode==stdSH)
		bds16_8((uint8*)wr+total,wr+total,count);

	buftotal = (total+=count);
	mychannel.userInfo = 1;

	if (total<minsndcount)
		return;
	if (mychannel.qHead==-1)
	{
		command.cmd = callBackCmd;
		command.param1 = 0;
		command.param2 = 0;
		mychannelcb(&mychannel,&command);
	}// else if ((ix=((ix-bufindexread)&BUFFER_MASK)-2)>0)
	//	sndoverrun = ix*4000*sndset.latency;
}

void stopsound()
{
	SndCommand	command;

	if (!mychannel.firstMod)
		return;
	mychannel.userInfo = 0;
	buftotal = 0;
	command.cmd = quietCmd;
	command.param1 = 0;
	command.param2 = 0;
	SndDoImmediate(&mychannel,&command);
	command.cmd = flushCmd;
	SndDoImmediate(&mychannel,&command);
}

/*boolean adequatesoundmanager()
{
	NumVersion	version = SndSoundManagerVersion();

	return version.majorRev>=3;
}*/

int adjust_soundvolume(int displacement)
{
	int		volume; /* shouldnt directly set the global */

	if (!soundon)
		return -1;
	if ((volume=soundvolume/10)<0)
		volume = 16;
	else if ((volume+=displacement)<0)
		volume = 0;
	else if (volume>15)
		volume = 16;
	return domenu(_subvolume_off+volume); /* should call InitSound during 0% volume game */
}
