/*
 macos/macvideo.c
  system and shell functions dealing with video/graphics

  MacFCEU 0.9d2

  Created: 2008.07.02

Friday, January 30, 2009 9:13 PM EST
  fixed the systemtable12 table (the red value was 9 for some 6's).
Thursday, October 8, 2009 8:39 PM EDST
  after adding EmptyRgn to make sure BlitScreen doesn't get called when app. switching yesterday,
  now GetWindowList is used so BlitScreen isn't used when other windows are up in fullscreen.
Thursday, March 11, 2010 4:18 AM EST
  scaling back on the "additional colors" since emphasis covers 0x40-0x7F.
Monday, March 15, 2010 8:30 AM EDT
  removed the actual performance loser, initializebuffers.
Tuesday, March 30, 12:44 PM
  added VSync code... the new performance loser.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <MacWindows.h>
//#include <Gestalt.h>
#include <Menus.h>
#include <QDOffscreen.h>
#include <Palettes.h>
#include <Devices.h>
#include <Video.h>

#include <LowMem.h>

#include "shell.h"

#include "../video.h"
#include "../common/dface.h"

#include "shellmenus.h"
#include "macvideo.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- globals/macros */

extern WindowRef myfrontwindow;

#if !TARGET_API_MAC_CARBON
 QDGlobals	qd;

/* dont even waste our time in this case */
 #define GetPortVisibleRegion(p,r)	(0+(p)->visRgn)
 #define GetWindowList()		(0+*(WindowRef*)0x09D6)
#endif

static RgnHandle	gprgn = NULL;

GDHandle	gamegdevice;

/* ---- general code */

static WindowRef GetTopWindow(void)
{
	WindowRef	window;

	if ((window=GetWindowList())==NULL)
		return NULL;
	do
	{
		if (IsWindowVisible(window))
			break;
	} while((window=GetNextWindow(window))!=NULL);

	return window;
}

#if 0
static int LSetEntries(ColorSpec *colors, short first, short count)
{
	GDHandle	gdh;

	if ((gdh=gamegdevice)==NULL)
		return -1;
	if ((**gdh).gdType==directType)
		return cDevErr;
#if TARGET_API_MAC_CARBON
 	{
 		GDHandle	oldgd = LMGetTheGDevice();

		LMSetTheGDevice(gdh);
		SetEntries(first,count,colors);
		LMSetTheGDevice(oldgd);
	}
	return QDError();
#else
 	{
		VDSetEntryRecord	record;

		ParamBlockRec	pb;

		*(VDSetEntryRecord**)&pb.cntrlParam.csParam = &record;
		record.csStart = first;
		record.csCount = count;
		record.csTable = colors;

		if (first<0 || first+count>(**(**(**gdh).gdPMap).pmTable).ctSize)
			return cRangeErr;

		pb.cntrlParam.ioNamePtr = NULL;
		pb.cntrlParam.ioVRefNum = 0;
		pb.cntrlParam.ioCRefNum = (**gdh).gdRefNum;
		pb.cntrlParam.csCode = cscSetEntries;
		return PBControlSync(&pb);
	}
#endif
}
#endif

static void expandcolors12(ColorSpec *outcolors, uint8 *s, uint16 i, uint16 c)
{
	uint16		*d = (uint16*)outcolors;
	unsigned	step = 0x1111;

	if (c<=0)
		return;
	do
	{
		*d++ = i;
		*d++ = (*s++&0x0F)*step;
		*d++ = ((*s&0xF0)>>4)*step;
		*d++ = (*s++&0x0F)*step;
	} while(++i,--c);
}

static void paintwindow(WindowRef window)
{
	if (!gprgn)
		if ((gprgn=NewRgn())==NULL)
			return;
	if (!window || !IsWindowVisible(window))
		return;
#if TARGET_API_MAC_CARBON
	if (GetWindowRegion(window,kWindowStructureRgn,gprgn)!=noErr)
		MacCopyRgn(GetGrayRgn(),gprgn);
#else
	GetWindowStructureRgn(window,gprgn);
#endif
	PaintOne(window,gprgn);
}

static void redraweverything(GDHandle gdh)
{
	WindowRef	window;

	if (!gprgn)
		if ((gprgn=NewRgn())==NULL)
			return;
	RectRgn(gprgn,&(**gdh).gdRect);
	if ((window=FrontWindow())!=NULL)
		do
			if (IsWindowVisible(window))
			{
				PaintOne(window,gprgn);
				PaintBehind(NULL,gprgn); /* 2011.03.04 */
			}
		while((window=GetNextWindow(window))!=NULL);
	PaintBehind(NULL,gprgn);

	drawmenubar();
}

/* ---- VSync code */

#if TARGET_API_MAC_CARBON
int VideoRetraceSync(GDHandle gdh)
{
#pragma unused(gdh)
	return paramErr;
}
#else
#include <Retrace.h>

static short GDeviceSlot(GDHandle gdh)
{
	AuxDCEHandle	dctl;

	if ((dctl=(AuxDCEHandle)GetDCtlEntry((**gdh).gdRefNum))==NULL)
		return 0;
	return (**dctl).dCtlSlot;
}

typedef struct VRSyncData
{
	VBLTask		VBL;
	VBLUPP		vUPP;
	GDHandle	vGDH;
	int		vCnt;
} VRSyncData;

#ifdef env68k
 #pragma parameter __A0 GetA0()
 extern void *GetA0(void) = {0x4E71}; /*NOP*/
static pascal void VRTask(void)
{
	VRSyncData	*vr = (VRSyncData*)GetA0();
#else
static void VRTask(VRSyncData *vr)
{
#endif
	static long	sig = 0xABCDEF01;

	/* ptr and signature check (if this task wasn't disposed of...) */
	if ((long int)vr->vUPP&((((unsigned long int)-1U)>>1)+2)
			|| sig!=0xABCDEF01)
		return; /* not setting vblCount will cause SlotVRemove */
	++vr->vCnt;
	vr->vCnt&= ((unsigned int)-1U)>>1;
	vr->VBL.vblCount = 1;
}

/* pass -1 to return count, same to return count and clears, NULL to dispose */
int VideoRetraceSync(GDHandle gdh)
{
	static VRSyncData	*vrsync = NULL;

	int		result;

	if (!gdh)
	{
		if (!vrsync)
			return INT_NO_VALUE;
		result = SlotVRemove((QElemPtr)&vrsync->VBL,GDeviceSlot(vrsync->vGDH));
	fail:
		DisposePtr((Ptr)vrsync);
		vrsync = NULL;
		return result;
	}
	if (gdh==(GDHandle)-1)
	{
		if (!vrsync)
			return INT_NO_VALUE;
		return vrsync->vCnt;
	}

	if (vrsync)
		if (vrsync->vGDH==gdh)
		{
			result = vrsync->vCnt;
			vrsync->vCnt = 0;
			return result;
		} else /* dispose of current first */
			result = VideoRetraceSync(NULL);

	if ((vrsync=(VRSyncData*)NewPtrSys(sizeof(VRSyncData)))==NULL)
		return -2; /* crap. */
	if ((vrsync->vUPP=NewVBLUPP(VRTask))==NULL)
	{
		result = -1;
		goto fail;
	}

	vrsync->VBL.qLink = NULL;
	vrsync->VBL.qType = vType;
	vrsync->VBL.vblAddr = vrsync->vUPP;
	vrsync->VBL.vblCount = 1;
	vrsync->VBL.vblPhase = 0;
	vrsync->vCnt = 0;
	vrsync->vGDH = gdh;

	return SlotVInstall((QElemPtr)&vrsync->VBL,GDeviceSlot(gdh));
}
#endif /*!TARGET_API_MAC_CARBON*/

/* ---- video code */

#define getwindowextra()		(screenshift&0x80 ? 299-256:0)

struct
{ /* cant set type as CTab cause size of ctTable doesnt cover 256 entries */
	long		seed;
	short		flags, size;
	ColorSpec	entries[256];
} myctable =
{
	0, (int16)0x8000, 255,
	0, {-1,-1,-1},
};

CTabPtr		gameclut = (CTabPtr)&myctable;

/* for direct to screen LUTs */
uint32		myctable32[256];
uint32		myctable15[256];
uint8		mycolorsinvalidate = 2;

#define BLITBPP_LOW			8
#define BLITBPP_HIGH			32

#ifndef env68k
void FCEUD_SetPalette(uint8 index,
	uint8 r,
	uint8 g,
	uint8 b)
{
	ColorSpec	*spec = index+myctable.entries;

	spec->rgb.red = (r<<8)+r;
	spec->rgb.green = (g<<8)+g;
	spec->rgb.blue = (b<<8)+b;

	myctable15[index] = (((r<<15-8)|(b>>8-5))&0x7C1F)|((g<<10-8)&0x3E0);
	myctable32[index] = (((r<<8)|g)<<8)|b;

	/*directsetcolors(spec,index,1,true);*/
	mycolorsinvalidate|= 1;
}
#endif

void FCEUD_GetPalette(uint8 index,
	uint8 *r,
	uint8 *g,
	uint8 *b)
{
	ColorSpec	*spec = index+myctable.entries;

	/* lower byte of each should be equal to higher byte */
	*r = spec->rgb.red;
	*g = spec->rgb.green;
	*b = spec->rgb.blue;
}

Rect		gameinitbounds; /* bounds on-screen set by initializescreen */
WindowRef	gamewindow; /* myfrontwindow moved to shell.c */

video_settings	vidset = {8, 231, 20, 0, 239, 70, 0, 255};
int		srendline = 0, erendline = 239, vblslines;

static int makegametitle(uint8 *str)
{
	uint8		*d;
	const char	*s;
	long		z;

	if ((d=str)==NULL)
		return 0;
	*d++ = 0;
	if ((s=getnextgamepath())==NULL)
		return 0;
	if ((z=strlen(s))>99) /* clip at end, 99 characters */
		memcpy(d,s+z-99,*str=99);
	else
		memcpy(d,s,*str=z);

	return TruncString(256-36,str,truncMiddle); //was 256-32; 2010.03.30
}

static void restorescreen(void)
{
	ShowMenuBar();
	if (gamewindow)
		{DisposeWindow(gamewindow); gamewindow = NULL;}
	myfrontwindow = NULL;
	VideoRetraceSync(NULL);
}

int initializescreen(boolean fs)
{
	static Boolean	installed = false;

	Rect		selectbounds;
	int		gdheight, mintop;
	int		windowwidth, windowheight;

	if (!gamegdevice)
		{fputs("initializescreen: no device for bounds!",stderr); return -1;}
	if (!installed)
	{
		atexit(restorescreen);
		installed = true;
	}

	gameinitbounds = (**gamegdevice).gdRect;
	gdheight = gameinitbounds.bottom-gameinitbounds.top;
	mintop = gdheight>499 ? 20:0;

//	if (screenshift>1) /* 3x+ not supported (think possible 512x342 screen) */
//		screenshift = 1;
	windowwidth = 256+getwindowextra()<<screenshift;
	windowheight = 240<<screenshift;

	/* center, adjust for log window, below, on small screens */
	/* align left to 16 points (for MOVE16 in 8 bpp) */
	gameinitbounds.left = (gameinitbounds.left+gameinitbounds.right-windowwidth>>1)&-16;
	if ((gameinitbounds.top=gameinitbounds.top+gameinitbounds.bottom-windowheight-64>>1)<mintop)
		gameinitbounds.top = mintop;

	gameinitbounds.right = gameinitbounds.left+windowwidth;
	gameinitbounds.bottom = gameinitbounds.top+windowheight;

	if (fs)
	{
		selectbounds = (**gamegdevice).gdRect;
		HideMenuBar();
		HideCursor();
	} else
	{
		selectbounds = gameinitbounds;
		ShowMenuBar();
	}

	if (!gamewindow)
	{
		Str255		title;
		Ptr		storage;

		if ((storage=NewPtr(256))!=NULL)
			*(SInt32*)(252+storage) = 0;
		if ((gamewindow=GetNewCWindow(windowid_screen,storage,NULL))==NULL)
		{
			if (storage)
				DisposePtr(storage);
			return -1;
		}

		makegametitle(title);
		SetWTitle(gamewindow,title);
	}

	if (fs!=vidset.fullscreen)
		HideWindow(gamewindow);
	MacMoveWindow(gamewindow,selectbounds.left,selectbounds.top,false);
	SizeWindow(gamewindow,
		selectbounds.right-=selectbounds.left,
		selectbounds.bottom-=selectbounds.top,
		false);
	MacShowWindow(gamewindow);
	HiliteWindow(gamewindow,true);

	*(uint32*)&selectbounds.top = 0; /* topLeft = 0L */

	{ /* make sure the fore/back colors are black/white, set the origin */
		RGBColor	color;
		CGrafPtr	oldport;
		GDHandle	olddevice;

		GetGWorld(&oldport,&olddevice);
		SetGWorld(GetWindowPort(gamewindow),gamegdevice);

		*(UInt32*)&color.red = 0UL, color.blue = 0;
		RGBForeColor(&color);
		*(UInt32*)&color.red = -1UL, color.blue = -1U;
		RGBBackColor(&color);
/*		PaintRect(&selectbounds);*/

		if (fs)
			SetOrigin(-gameinitbounds.left,-gameinitbounds.top);
		else
			SetOrigin(0,0);

		SetGWorld(oldport,olddevice);
	}

	vidset.fullscreen = fs;
	return 0;
}

static void invaltopwindow(void)
{
	Rect		b;
	WindowRef	win;
	/* DO NOT USE myfrontwindow - stale pointer after dialog/window disposal */
	/* includes getfile/putfile dialogs that use the shell event routine */
	if ((win=FrontWindow())!=NULL || !IsWindowVisible(win))
		return;
	SetRect(&b,-16000,-16000,16000,16000);
	InvalWindowRect(win,&b);
}

void FCEUD_BlitScreen(uint8 *XBuf);
void FCEUD_BlitScreen(uint8 *XBuf)
{ /* probably should make sure the fg and bg colors are black and white (you, not me) */
	static PixMap	srcpixmap =
	{
		NULL, /* base address */
		0, /* row size */
		{0, 0, 0, 256}, /* bounds */
		0, /* PixMap version (0=fixed/direct) */
		0, 0, /* pack type, size (both unused in this case) */
		72<<16, 72<<16, /* resolution h, v */
		clutType, 8, /* pixel type, size */
		1, 8, /* component count, size per */
		0, /* plane size or pixel format */
		&gameclut /* fake CTabHandle */
	};

	Rect		dstbounds, srcblitrect; /* source no longer a global! */
	PixMapHandle	dstblitmap;
	GDHandle	olddevice;
	CGrafPtr	oldport, port;
	unsigned	pixelsize = 0, ssh = screenshift;

	if ((port=GetWindowPort(gamewindow))==NULL)
		{XBuf = NULL, dstblitmap = NULL; goto color;}
	if ((dstblitmap=GetPortPixMap(port))!=NULL)
		pixelsize = (**dstblitmap).pixelSize;
	/* update color table regardless of the image's existence; 2010.03.20 */
	if (pixelsize<=8)
	{
		if (!suspended)
			myctable.seed = (**(**dstblitmap).pmTable).ctSeed; /* dickey move */
	color:
		if (mycolorsinvalidate)
			SetEntries(0,254,myctable.entries);
		if (!port && gamegdevice)
			pixelsize = (**(**gamegdevice).gdPMap).pixelSize;
		if (mycolorsinvalidate&2) /* set and update (manually) */
			if (pixelsize<15)
				redraweverything(gamegdevice);
			else
				invaltopwindow();
	} else if (!myctable.seed)
		myctable.seed = GetCTSeed();

	mycolorsinvalidate = 0;

	if ((srcpixmap.baseAddr=(Ptr)XBuf)==NULL)
		return;

	srcblitrect.right = 256-(srcblitrect.left=eoptions&EO_CLIPSIDES);
	srcblitrect.top = srendline<<ssh;
	srcblitrect.bottom = (erendline+1)<<ssh;
	srcpixmap.bounds.bottom = 256<<ssh;
	srcpixmap.rowBytes = FCEUSCRN_ROWSIZE>>ssh|0x8000;

	if (screenrefreshsync && VideoRetraceSync(gamegdevice)>=0)
	{ /* same as sound; wait for a limited period so there is no chance of a hang */
		uint32		t = TickCount()+2; /* 1/15-1/30 sec */

		do
			if (VideoRetraceSync((GDHandle)-1))
				break;
		while((int8)(TickCount()-t)<0);
	}

	if (ssh) /* 0.8.5+: interlaced */
		srcpixmap.baseAddr = (Ptr)((long int)XBuf&~511), ssh = 0;
	#if !TARGET_API_MAC_CARBON
	if (!ismenubarvisible() && !suspended && GetTopWindow()==gamewindow
			&& (unsigned)(pixelsize-BLITBPP_LOW)<=BLITBPP_HIGH-BLITBPP_LOW)
	{
		if (!EmptyRgn(GetPortVisibleRegion(port,gprgn)))
			BlitScreen((const uint8*)srcpixmap.baseAddr,srcpixmap.rowBytes&0x3FFF,&srcblitrect,dstblitmap);
		return;
	}
	#endif
	GetGWorld(&oldport,&olddevice);
	SetGWorld(port,NULL);

	dstbounds.top = srcblitrect.top<<ssh;
	dstbounds.left = srcblitrect.left<<screenshift;
	dstbounds.bottom = srcblitrect.bottom<<ssh;
	dstbounds.right = srcblitrect.right+getwindowextra()<<screenshift;

	CopyBits((BitMap*)&srcpixmap,GetPortBitMapForCopyBits(port),
		&srcblitrect,&dstbounds,srcCopy,NULL);

	myctable.seed = 0;
	SetGWorld(oldport,olddevice);
}

#if 0
/* additional tables so the icon mgr will have confidence drawing in color */
/* revised: 2010.03.11 */
static uint8	scaletables12[] =
{ /* gray */
/*20*/	0xE,0xEE, 0xD,0xDD, 0xC,0xCC, 0xB,0xBB,
/*24*/	0xA,0xAA, 0x9,0x99, 0x8,0x88, 0x7,0x77,
/*28*/	0x6,0x66, 0x5,0x55, 0x4,0x44, 0x3,0x33,
/*2C*/	0x2,0x22, 0x1,0x11,
 /* yellow */
/*2E*/	0xC,0xC0, 0x3,0x30,
 /* cyan */
/*30*/	0x0,0xCC, 0x0,0x33,
 /* magenta */
/*32*/	0xC,0x0C, 0x3,0x03,
 /* red */
/*34*/ /*0xC,0x0*/0x9,0x00, 0x6,0x00, 0x3,0x00,
 /* green */
/*38*/	0x0,0xC0, 0x0,0x90, 0x0,0x60, 0x0,0x30,
 /* blue */
/*3C*/	0x0,0x0C, 0x0,0x09, 0x0,0x06, 0x0,0x03,
};

static uint8	icontable12[] =
{
	0x0,0x0F, /*0x0,0xF0,*/ 0x0,0xFF, 0xF,0x00, 0xF,0x0F, 0xF,0xF0,
  /* Apple Icon colors, excluding those already accounted for in other tables */
	/*0x0F,0xFC,*/ 0xF,0xC9, 0xC,0x96, 0xF,0x63,
	0x3,0xF9, 0x3,0x96, 0x3,0x66,
	0x9,0xFF, 0x0,0x9F, /*0x03,0x09,*/
	0xF,0x6C, 0x9,0x06, 0x9,0x69,
	0xC,0xCF, 0x9,0x9F, 0x6,0x6C, 0x3,0x36,
};
#endif

static uint8	systemtable12[256*2] =
{ /* subbing for clut ID 128 */
	0xF,0xFF, 0xF,0xFC, 0xF,0xF9, 0xF,0xF6, 0xF,0xF3, 0xF,0xF0,
	0xF,0xCF, 0xF,0xCC, 0xF,0xC9, 0xF,0xC6, 0xF,0xC3, 0xF,0xC0,
	0xF,0x9F, 0xF,0x9C, 0xF,0x99, 0xF,0x96, 0xF,0x93, 0xF,0x90,
	0xF,0x6F, 0xF,0x6C, 0xF,0x69, 0xF,0x66, 0xF,0x63, 0xF,0x60,
	0xF,0x3F, 0xF,0x3C, 0xF,0x39, 0xF,0x36, 0xF,0x33, 0xF,0x30,
	0xF,0x0F, 0xF,0x0C, 0xF,0x09, 0xF,0x06, 0xF,0x03, 0xF,0x00,

	0xC,0xFF, 0xC,0xFC, 0xC,0xF9, 0xC,0xF6, 0xC,0xF3, 0xC,0xF0,
	0xC,0xCF, 0xC,0xCC, 0xC,0xC9, 0xC,0xC6, 0xC,0xC3, 0xC,0xC0,
	0xC,0x9F, 0xC,0x9C, 0xC,0x99, 0xC,0x96, 0xC,0x93, 0xC,0x90,
	0xC,0x6F, 0xC,0x6C, 0xC,0x69, 0xC,0x66, 0xC,0x63, 0xC,0x60,
	0xC,0x3F, 0xC,0x3C, 0xC,0x39, 0xC,0x36, 0xC,0x33, 0xC,0x30,
	0xC,0x0F, 0xC,0x0C, 0xC,0x09, 0xC,0x06, 0xC,0x03, 0xC,0x00,

	0x9,0xFF, 0x9,0xFC, 0x9,0xF9, 0x9,0xF6, 0x9,0xF3, 0x9,0xF0,
	0x9,0xCF, 0x9,0xCC, 0x9,0xC9, 0x9,0xC6, 0x9,0xC3, 0x9,0xC0,
	0x9,0x9F, 0x9,0x9C, 0x9,0x99, 0x9,0x96, 0x9,0x93, 0x9,0x90,
	0x9,0x6F, 0x9,0x6C, 0x9,0x69, 0x9,0x66, 0x9,0x63, 0x9,0x60,
	0x9,0x3F, 0x9,0x3C, 0x9,0x39, 0x9,0x36, 0x9,0x33, 0x9,0x30,
	0x9,0x0F, 0x9,0x0C, 0x9,0x09, 0x9,0x06, 0x9,0x03, 0x9,0x00,

	0x6,0xFF, 0x6,0xFC, 0x6,0xF9, 0x6,0xF6, 0x6,0xF3, 0x6,0xF0,
	0x6,0xCF, 0x6,0xCC, 0x6,0xC9, 0x6,0xC6, 0x6,0xC3, 0x6,0xC0,
	0x6,0x9F, 0x6,0x9C, 0x6,0x99, 0x6,0x96, 0x6,0x93, 0x6,0x90,
	0x6,0x6F, 0x6,0x6C, 0x6,0x69, 0x6,0x66, 0x6,0x63, 0x6,0x60,
	0x6,0x3F, 0x6,0x3C, 0x6,0x39, 0x6,0x36, 0x6,0x33, 0x6,0x30,
	0x6,0x0F, 0x6,0x0C, 0x6,0x09, 0x6,0x06, 0x6,0x03, 0x6,0x00,

	0x3,0xFF, 0x3,0xFC, 0x3,0xF9, 0x3,0xF6, 0x3,0xF3, 0x3,0xF0,
	0x3,0xCF, 0x3,0xCC, 0x3,0xC9, 0x3,0xC6, 0x3,0xC3, 0x3,0xC0,
	0x3,0x9F, 0x3,0x9C, 0x3,0x99, 0x3,0x96, 0x3,0x93, 0x3,0x90,
	0x3,0x6F, 0x3,0x6C, 0x3,0x69, 0x3,0x66, 0x3,0x63, 0x3,0x60,
	0x3,0x3F, 0x3,0x3C, 0x3,0x39, 0x3,0x36, 0x3,0x33, 0x3,0x30,
	0x3,0x0F, 0x3,0x0C, 0x3,0x09, 0x3,0x06, 0x3,0x03, 0x3,0x00,

	0x0,0xFF, 0x0,0xFC, 0x0,0xF9, 0x0,0xF6, 0x0,0xF3, 0x0,0xF0,
	0x0,0xCF, 0x0,0xCC, 0x0,0xC9, 0x0,0xC6, 0x0,0xC3, 0x0,0xC0,
	0x0,0x9F, 0x0,0x9C, 0x0,0x99, 0x0,0x96, 0x0,0x93, 0x0,0x90,
	0x0,0x6F, 0x0,0x6C, 0x0,0x69, 0x0,0x66, 0x0,0x63, 0x0,0x60,
	0x0,0x3F, 0x0,0x3C, 0x0,0x39, 0x0,0x36, 0x0,0x33, 0x0,0x30,
	0x0,0x0F, 0x0,0x0C, 0x0,0x09, 0x0,0x06, 0x0,0x03,

	0xE,0x00, 0xD,0x00, 0xB,0x00, 0xA,0x00, 0x8,0x00,
	0x7,0x00, 0x5,0x00, 0x4,0x00, 0x2,0x00, 0x1,0x00,

	0x0,0xE0, 0x0,0xD0, 0x0,0xB0, 0x0,0xA0, 0x8,0x80,
	0x0,0x70, 0x0,0x50, 0x0,0x40, 0x0,0x20, 0x0,0x10,

	0x0,0x0E, 0x0,0x0D, 0x0,0x0B, 0x0,0x0A, 0x0,0x08,
	0x0,0x07, 0x0,0x05, 0x0,0x04, 0x0,0x02, 0x0,0x01,

	0xE,0xEE, 0xD,0xDD, 0xB,0xBB, 0xA,0xAA, 0x8,0x88,
	0x7,0x77, 0x5,0x55, 0x4,0x44, 0x2,0x22, 0x1,0x11
};

static void setclut12(uint8 *s, int c)
{
	ColorSpec	*store = NULL;

	if (c>256)
		c = 256;
	if (c<=0 || (store=realloc(store,c*sizeof(*store)))==NULL)
		return;
	expandcolors12(store,s,0,c);

	SetEntries(0,--c,store);

	store = realloc(store,0);
}

void videocolorsinit()
{
	ColorSpec	*colors = myctable.entries;

//	expandcolors12(6+colors,icontable12,9,sizeof(icontable12)/2);
//	expandcolors12(0x20+colors,scaletables12,0x20,sizeof(scaletables12)/2);

	{ /* fill in the rest of the entry index values */
		int		x = 0;

		do
			colors->value = x;
		while(++colors,++x<256);
	}

	/* white and black need this flag */
	myctable.entries[0xFF].value|= 0x800;
}

int InitVideo(FCEUGI *gi)
{
#pragma unused(gi)
	if (initializescreen(vidset.fullscreen)!=0)
		return 0;
	videocolorsinit();

	if (!gprgn)
		gprgn = NewRgn();
	return 1;
}

int KillVideo()
{
	restorescreen();

	if (gdbitdepth()<=8)
	{ /* ires of 5 will actually increase precision after exit until update */
		setclut12(systemtable12,sizeof(systemtable12)/2);
		MakeITable((**(**gamegdevice).gdPMap).pmTable,(**gamegdevice).gdITable,5);
		redraweverything(gamegdevice);
	}

	FlushEvents(mDownMask|mUpMask|keyDownMask|keyUpMask|autoKeyMask,0); /* yes, here */
	InitCursor();

	return 1;
}

void videorefresh()
{
extern uint64 videofps_boop[];
	memset(videofps_boop,0,80);
	RefreshThrottleFPS();

	FCEUI_DisableSpriteLimitation(eoptions&EO_NOSPRLIM);

	FCEUI_SetRenderedLines(srendlinen,erendlinen,srendlinep,erendlinep);
	FCEUI_GetCurrentVidSystem(&srendline,&erendline,&vblslines);
	FCEUI_SetVBLScanlines(vblslinesn,vblslinesp); //this was missing in 0.5.4!; 2010.03.13

	if (myfrontwindow!=gamewindow)
		invaltopwindow();
	paintwindow(gamewindow);
}
