/*
 macos/memview.c

  Created: 2015.07.12
  Revised: 2015.08.05

Sunday, July 12, 2015
  created.  based on ideas with the FCEUX RAM editor.
*/

#include <stdio.h>

#include <MacWindows.h>
#include <Events.h>

#include "shell.h"
#include "macvideo.h"
#include "memview.h"
#include "listwindow.h"
#include "../utility/disasm6502.h"

/* ---- globals */

WindowRef	MemViewWindow;

/* ---- code */

#define PASCAL_RAMVIEW_TITLE		"\pMemory View"
#define PASCAL_RAMVIEW_FONTNAME		"\pMonaco"
#define TEXT_FONT_SIZE			9

#define MAXIMUM_LINECOUNT		4096
#define INIT_RAMVIEW_HEIGHT		(11*11+15)

#define RAMVIEW_MINIMUM_HEIGHT		(INIT_RAMVIEW_HEIGHT-24)
#define RAMVIEW_MAXIMUM_HEIGHT		(MAXIMUM_LINECOUNT/2*12+16)
#define RAMVIEW_WIDTH			(9*5*6+6+18)

/* ---- code */

static const char *print_line(char *d,
	int32 *z,
	int i)
{
	static char	b[9*5]; //+1];
	char		*s;
	int		c;
#pragma unused(d)
	i<<= 4;
	Disasm_HexStr(b,i,4);
	c = 8, s = b+5, s[-1] = ' ';
	do
		*s = ' ', Disasm_HexStr(s+1,FCEUI_MemSafePeek(i)<<8|FCEUI_MemSafePeek(i+1),4);
	while(s+=5,i+=2,--c);
//	*s = 0;
	*z = sizeof(b);
	return b;
}

static void click_line(int i,
	int x,
	int *ss,
	int *se)
{
#pragma unused(i)
	x-= 6, x = x*((131071+4)/5)>>16;
	if (x>15u)
		return;
	x+= x, x+= x>>2, x+= 6;
	*ss = x;
	*se = x+2;
}

static WindowRef memview_create(void)
{
	Rect		bounds;
	WindowRef	win;
	/* bottom-left of screen, above message log */
	bounds = (**gamegdevice).gdRect;
	bounds.top = (bounds.bottom-=9+INIT_RAMVIEW_HEIGHT)-INIT_RAMVIEW_HEIGHT;
	bounds.left = bounds.left+9;
	bounds.right = bounds.left+RAMVIEW_WIDTH;

	if ((win=listwindow_create((WindowRef)-1,&bounds,PASCAL_RAMVIEW_TITLE,
			PASCAL_RAMVIEW_FONTNAME,TEXT_FONT_SIZE,
			0,MAXIMUM_LINECOUNT,0,1,FOUR_CHAR_CODE('ramv')))==NULL)
		return NULL;
	listwindow_settexthook(win,print_line,MAXIMUM_LINECOUNT);
	listwindow_setclickhook(win,click_line);
	return win;
}

int memview_show(int w)
{
	if (!MemViewWindow && w!=IMIN(w) && (MemViewWindow=memview_create())==NULL)
		return -1;
	return showhide(MemViewWindow,w);
}

void memview_update(boolean erase)
{
	listwindow_update(MemViewWindow,erase);
}

int memview_click(EventRecord *event,
	short what)
{
	Rect		bounding;

	bounding.top = RAMVIEW_MINIMUM_HEIGHT;
	bounding.left = RAMVIEW_WIDTH+1;
	bounding.bottom = RAMVIEW_MAXIMUM_HEIGHT;
	bounding.right = RAMVIEW_WIDTH+1;

	return listwindow_click(MemViewWindow,event,what,&bounding);
}

void memview_activate(boolean active)
{
	listwindow_setactive(MemViewWindow,active);
}

int memview_key(int k)
{
	int32		i, x, e;
	int		a;

	if (k==kEscapeCharCode)
	{
		if (MemViewWindow)
			HiliteWindow(MemViewWindow,false);
		return 0;
	}
	if (k-'0'>9u && k-'A'>15u && k-'a'>15u)
		return 0;
	if ((a=listwindow_getsel(MemViewWindow,&i,NULL,&x,&e))<0)
		return a;
	if (e-x>2u)
		return 0; /* no byte selected */
	x-= 6, x-= x*((131071+4)/5)>>17;
	if (x>31u)
		return 0;
	if (k>='A')
		k+= 9;
	k&= 15;
	i = i*16u+(x>>1);
	a = FCEUI_MemSafePeek(i);
	if (x&1)
		a = (a&0xF0)|k;
	else
		a = (a&15)|k<<4;
	FCEUI_MemPoke(i,a,0);
	/* advance and update selection */
	++x, i = (i+(x>>5))>>4, x&= 31, a = x&1;
	x+= x>>2, x+= 6;
	listwindow_select(MemViewWindow,i,i,x,x+1+!a,true);
	memview_update(false);
	return 1;
}
