/*
 macos/menus.c

  Created: 2008.07.08
  Revised: 2015.08.29

Wednesday, August 19, 2009 1:10 AM
  increased/maxed out some option limits (1000-65535 srate, 24-429% for speed).
Friday, October 9, 2009 1:48 PM
  finishing the menubar show/hide code to work properly.
Tuseday, March 23, 2010 1:26 PM
  moved all the rest of the sound stuff declared here out, externs are now in shell.h.
Saturday, March 27, 2010 1:51 PM
  a run-in with a falsely rewritten "Set Enabled" routine prompts me to check validity.
Thursday, April 29, 2010 8:24 AM
  watch cursor before any InitSound (some carts may call some slow init re-process).
Saturday, May 4, 2013 5:19 AM
  max. volumes/scales was +/-10000, now 2000.
*/

#include <stdlib.h>

#include "shell.h"

#include "../FCEU.h"
#include "../common/dface.h"
#include "../WAVE.h"

#include "shellmenus.h"
#include "messagelog.h"
#include "macdialogs.h"
#include "macvideo.h"
#include "debugger.h"
#include "config.h"
#include "tileview.h"
#include "memview.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- globals */

extern struct FCEUGI *GI;

#define MAX_RATE

static const uint32	sndqualities[] =
{
	8000*16, /* was 8192 */
	11025*16,
	15000*16,
	22050*16,
	36000*16,
	44100*16, /* added 2013.05.14 */
	66763.63*16,
	88200*16, /* added 2015.08.21 */
	96000*16,
	MAXIMUM_SRATE
};

static const uint32	emulationspeeds[] =
{
	500000000,
	629960525,
	707106781,
	793700526,
	890898718,
	1000000000,
	1259921050,
	1587401052,
	1781797436,
	2000000000,
};

/* ---- code */

static int getvidtype(void)
{
	if (!GI)
		return 0;
	if (GI->vidsys==GIV_USER && palyo || GI->vidsys==GIV_PAL)
		return 1;
	return 0x80;
}

void redomenusgi()
{
	static uint8	simpled[] =
		{/*_file_close, */_file_savestate, _file_loadstate,
		_emulator_reset, _emulator_power, _emulator_step, 0};

	int		x =-1;
	uint8		i;

	while((i=simpled[++x])!=0)
		enablemenuitem(i,GI!=NULL);
}

void reenablemenus(boolean enable)
{
	uint8		*simpled = (uint8*)"\1\4\2\3\0";

	int		x =-1;
	uint8		i;

	while((i=simpled[++x])!=0)
		enablemenuitem(i<<6,enable);
	enablemenuitem(_apple_about,enable); /* was the whole menu; 2010.04.11 */

	drawmenubar();
}

/* in case of file overflow; 2012.02.15 */
void updatewavestatus()
{
	if (FCEU_WaveUpdateHeader()<0)
		setmenuitemstr(_file_logsound,getpstr(temporary,strlistid_common,
			commonstr_logsound));
}

void updatemenus()
{
	static uint8	*polo[] = {&genie,&vidset.fullscreen,
		&screenshift, &screenrefreshsync,
		&fceutileview};
	uint16		simpled[] = {_game_Genie, _subvideo_fullscreen,
		_subvideo_2x_scale, _subvideo_refresh_sync,
		_subvideo_tile_view, 0};

	int		x =-1, m;
	uint16		i;

	while((i=simpled[++x])!=0)
		checkmenuitem(i,*polo[x]&1);

	//partial switch added to tell whether it's applicable; 2010.03.23, 03.24
	m = palyo&1;
	if (GI && GI->vidsys!=GIV_USER
			&& ((GI->vidsys!=GIV_NTSC && !m)
			|| (GI->vidsys!=GIV_PAL && m)))
		m|= 0x8000U;
	checkmenuitem(_game_PAL,m);

	if ((m=getvidtype())==0)
	{
		x = commonstr_all_1_6Mhz;
		m = !RateFixFlags ? 0:~RateFixFlags&0x81 ?
			RateFixFlags&1|0x8000U:1;
	} else
	{
		if (m&0x80)
			x = commonstr_NTSC_1_6MHz;
		else
			x = commonstr_PAL_1_6MHz;
		m&= RateFixFlags;
	}
	setmenuitemstr(_game_force_1_6MHz,getpstr(temporary,strlistid_common,x));
	checkmenuitem(_game_force_1_6MHz,m);

	for(x=0,m=sndset.rate*16; x<GET_ARRAY_COUNT(sndqualities); ++x)
		if (sndqualities[x]==m)
			break;
	checkmenu(_subsound_telephone,x+(_subsound_telephone&63),0); /* if loop fails, Other will automatically be checked */

	x = soundvolume/10, x|= (unsigned)(soundvolume-x*10)<<8, ++x;
	checkmenu(_subvolume_off,MIN((unsigned)x,17),0);

	checkmenuitem(_subvideo_5to4,screenshift&0x80);
	checkmenuitem(_subvideo_clip_sides,eoptions&EO_CLIPSIDES);
	checkmenuitem(_subvideo_no_sprite_limit,(eoptions&EO_NOSPRLIM)!=0);
	//checkmenuitem(_subvideo_skip_odd_frames,(eoptions&EO_ODDFRAMESKIP)!=0);
	checkmenuitem(_subvideo_auto_skip_frames,(eoptions&EO_AUTOFRAMESKIP)!=0);

	for(x=_subchannels_rect_1,m=1; x<=_subchannels_DMC; ++x,m+=m)
		checkmenuitem(x,sndactivechannels&m);
	checkmenuitem(_subexpchannels_enable,sndactivechannels&0x20);
	for(x=_subexpchannels_VRC6,m+=m; x<=_subexpchannels_SFME; ++x,m+=m)
	{
		enablemenuitem(x,sndactivechannels&0x20);
		checkmenuitem(x,sndactivechannels&m);
	}

	checkmenuitem(_subspeed_adj_dmc,!(RateFixFlags&0x40));
	for(x=0; x<GET_ARRAY_COUNT(emulationspeeds); ++x)
		if (emulationspeeds[x]==emuset.speed)
			break;
	checkmenu(_subspeed_50,x+(_subspeed_50&63),0);

	checkmenu(_subbackground_off,1+(eoptions&EO_BGRUN)+(eoptions&EO_MUTE ? 2:0),0);

	enablemenuitem(_file_close,forewindow(-1));

	checkmenuitem(_edit_memview,memview_show(IMIN(int)));

	updatewavestatus();
}

static void setfvolume(int16 *v,
	const uint8 *t)
{
	int		tmp;

	stopsound();
	if ((tmp=valuedialog(t,*v,-2000,2000,false))!=INT_NO_VALUE)
		*v = tmp, update_soundsettings(1);
}

boolean domenu(uint16 itemcode)
{
	double		td;
	int		tmp;

	switch(itemcode&0x7FFF)
	{ /* menubar toggle is automatic during fullscreen */
	/* apple */
	case _apple_about:
		stopsound();
		splash();
		break;

	/* file */
	case _file_open:
		stopsound();
		opengame(NULL);
		break;
	case _file_close:
		if (GI)
			scexit(_sc_closegame);
		else
			forewindow(0);
		break;
	case _file_savestate:
		if (!GI)
			return false;
		stopsound();
		SaveStateOrMovie(NULL,itemcode>>15);
		break;
	case _file_loadstate:
		if (!GI)
			return false;
		stopsound();
		LoadStateOrMovie(NULL,itemcode>>15,false);
		break;
	case _file_logsound:
		setmenuitemstr(_file_logsound,getpstr(temporary,strlistid_common,
			RecordSoundWave(NULL)>0 ? commonstr_stopwave:commonstr_logsound));
		break;
	case _file_saveconfig:
		SaveConfig();
		break;
	case _file_quit:
		scexit(_sc_quit);
		break;

	/* emulator */
	case _emulator_reset:
		FCEUI_ResetNES();
		goto subreset;
	case _emulator_power:
		FCEUI_PowerNES();
	subreset:
		if (debuggerwindow_visible())
			debuggerwindow_step(1); /* start */
		break;
	case _emulator_debug:
		checkmenuitem(_emulator_debug,debuggerwindow_show(-1)>0);
		break;
	case _emulator_step:
		checkmenuitem(_emulator_debug,debuggerwindow_step(1)>0);
		break;

	/* game */
	case _game_force_1_6MHz:
		setcursor(4,false);
		stopsound();
		if ((tmp=getvidtype())==0) /* All if not none */
			RateFixFlags = (0x81&~RateFixFlags) ? 0x81:0;
		else
			RateFixFlags^= tmp;
		RefreshThrottleFPS();
		goto rsnd;
	case _game_PAL:
		setcursor(4,false);
		stopsound();
		FCEUI_SetVidSystem(palyo^=1);
		RefreshThrottleFPS(); //first to clear
		videorefresh();
		RefreshThrottleFPS(); //second to ensure timing update
		goto rsnd;
	case _game_Genie:
		FCEUI_SetGameGenie(genie^=1);
		break;
	case _game_input:
		stopsound();
		config_inputdialog();
		break;
	case _game_palette:
		stopsound();
		gamewindowtofront();
		PaletteDialog(NULL,0);
		break;
	case _game_netplay:
		stopsound();
/*		ShowNetplayConsole();*/
		break;
	/* cheat item replaced with commandline commands (will be implemented eventually) */

	/* edit */
	case _edit_copy:
		if (logwindow_show(IMIN(int)))
			logwindow_scrap(0);
		break;
	case _edit_select_all:
		if (logwindow_scrap(IMAX(int)))
			enablemenuitem(_edit_copy,true);
		break;
	case _edit_log:
		tmp = logwindow_show(-1)>0;
		checkmenuitem(_edit_log,tmp);
		enablemenuitem(_edit_select_all,tmp);
		enablemenuitem(_edit_copy,tmp && logwindow_scrap(-1));
		break;
	case _edit_memview:
		memview_show(-1);
		break;

	case _subsound_latency:
		if ((td=valuedialog("\pSound Latency (in milliseconds)",
				sndset.latency,4,125,true))
				==INT_NO_VALUE)
			break;
		sndset.latency = td;
		RefreshThrottleFPS();
		break;
	/* sound quality */
	case _subsound_telephone:
	case _subsound_lower:
	case _subsound_low:
	case _subsound_default:
	case _subsound_moderate:
	case _subsound_low_cd:
	case _subsound_triple_22:
	case _subsound_double_cd:
	case _subsound_better:
	case _subsound_best:
	case _subsound_other:
		stopsound();

		if ((itemcode+=-_subsound_telephone)>=GET_ARRAY_COUNT(sndqualities))
		{
			if ((td=valuedialog("\pCustom Sample Rate",
					(int32)(sndset.rate*100)/100.,1000,(MAXIMUM_SRATE+8)/16.,true))
					==INT_NO_VALUE)
				goto end;
			/* rate warning moved to files.c; 2012.03.18 */
			tmp = (int32)(td*16+.5);
			if (tmp>MAXIMUM_SRATE)
				tmp = MAXIMUM_SRATE;
			sndset.rate = tmp/16.;
		} else
			sndset.rate = sndqualities[itemcode]/16.;
		if (GI)
			GI->soundrate = 0;
	#if HAVE_HQ
		FCEUI_SetSoundQuality(soundquality=0); /*(qualities[itemcode]>>24)&7);*/
	#endif
		goto rsnd;

	/* sound volume submenu (sorry, no slider) */
	case _subvolume_off:
	case _subvolume_10:
	case _subvolume_20:
	case _subvolume_30:
	case _subvolume_40:
	case _subvolume_50:
	case _subvolume_60:
	case _subvolume_70:
	case _subvolume_80:
	case _subvolume_90:
	case _subvolume_100:
	case _subvolume_110:
	case _subvolume_120:
	case _subvolume_130:
	case _subvolume_140:
	case _subvolume_150:
	case _subvolume_other:
		if ((itemcode-=_subvolume_off)==0)
			goto zerovolume;

		if (itemcode>=_subvolume_other-_subvolume_off)
		{
			if ((tmp=valuedialog("\pMaster Volume (percent)",
					soundvolume,-2000,2000,false))==INT_NO_VALUE)
				goto end;
			if ((soundvolume=tmp)==0)
			{
			zerovolume:
				soundvolume = 0; /* now must set ahead of InitSound */
				stopsound();
				InitSound(NULL);
				RefreshThrottleFPS(); /* 2015.08.04 */
			}
		} else
			soundvolume = itemcode*10;
		FCEUI_SetSoundVolume(soundvolume);
		if (soundon && soundvolume>0)
		{
		rsnd:	setcursor(4,false); /* set to watch cursor; 2010.04.29 */
			InitSound(GI);
			setcursor(-1,false);
		}
		break;

	/* video */
	case _subvideo_fullscreen:
		if (!GI)
			vidset.fullscreen^= 1;
		else
			initializescreen(vidset.fullscreen^1);
		checkmenuitem(_subvideo_fullscreen,vidset.fullscreen);
		goto end;
	case _subvideo_2x_scale:
		checkmenuitem(_subvideo_2x_scale,screenshift^=1);
		if (GI)
			initializescreen(vidset.fullscreen);
		videorefresh();
		goto end;
	case _subvideo_5to4:
		screenshift^= 0x80;
		if (GI)
			initializescreen(vidset.fullscreen);
		videorefresh();
		break;
	case _subvideo_clip_sides:
		eoptions^= EO_CLIPSIDES;
		videorefresh();
		break;
	case _subvideo_no_sprite_limit:
		eoptions^= EO_NOSPRLIM;
		videorefresh();
		break;
	//case _subvideo_skip_odd_frames:
	//	eoptions^= EO_ODDFRAMESKIP;
	//	break;
	case _subvideo_auto_skip_frames:
		eoptions^= EO_AUTOFRAMESKIP;
		break;
	case _subvideo_refresh_sync:
		screenrefreshsync^= 1;
		RefreshThrottleFPS(); /* 2015.07.31 */
		break;
	case _subvideo_tile_view:
		ShowHideTileView(-1);
		break;

	/* sound channels */
	case _subchannels_rect_1:
	case _subchannels_rect_2:
	case _subchannels_triangle:
	case _subchannels_noise:
	case _subchannels_DMC:
		sndactivechannels^= 1<<itemcode-_subchannels_rect_1;
	sche:	FCEUI_SetSoundActiveChannels(sndactivechannels);
		break;
	case _subchannels_rect_amp:
		stopsound();
		if ((tmp=valuedialog("\pRectangle Filter Amp (FCEU 1.0 = 256)",
				rectangle_amp,-2000,2000,false))!=INT_NO_VALUE)
			FCEUI_SetRectFilterAmp(rectangle_amp=tmp);
		goto end;
	case _subchannels_tnd_amp: /* teenage nusaince minja durdles */
		stopsound();
		if ((tmp=valuedialog("\pTri/Noise/DMC Amp (FCEU 1.0 = 256)",
				tnd_amp,-2000,2000,false))!=INT_NO_VALUE)
			FCEUI_SetTNDFilterAmp(tnd_amp=tmp);
		goto end;
	case _subchannels_tri_scale:
		stopsound();
		if ((tmp=valuedialog("\pTriangle Scale (3 = 1.0)",
				tri_scale,0,9,false))!=INT_NO_VALUE)
			FCEUI_SetTriangleScale(tri_scale=tmp);
		goto end;
	case _subchannels_noise_scale:
		stopsound();
		if ((tmp=valuedialog("\pNoise Scale (2 = 1.0)",
				noise_scale,0,9,false))!=INT_NO_VALUE)
			FCEUI_SetNoiseScale(noise_scale=tmp);
		goto end;
	case _subchannels_dc_filter:
		stopsound();
		if ((td=valuedialog("\pDC Filter Loss (over 94; default: 93)",
				dcfilter_loss/512.,0,65535/512.,true))!=INT_NO_VALUE)
			FCEUI_SetDCFilterLoss(dcfilter_loss=td*512+.5);
		goto end;
	case _subexpchannels_enable:
		sndactivechannels^= 0x20;
		goto sche;
	case _subexpchannels_VRC6:
	case _subexpchannels_VRC7:
	case _subexpchannels_FDS:
	case _subexpchannels_MM5:
	case _subexpchannels_N163:
	case _subexpchannels_SFME:
		sndactivechannels^= 64u<<itemcode-_subexpchannels_VRC6;
		FCEUI_SetSoundActiveChannels(sndactivechannels);
		break;
	case _subexpchannels_vrc6_volume:
		setfvolume(&vrc6_vol,"\pVRCVI Volume (%)");
		goto end;
	case _subexpchannels_vrc7_volume:
		setfvolume(&vrc7_vol,"\pVRCVII Volume (%)");
		goto end;
	case _subexpchannels_fds_volume:
		setfvolume(&fds_vol,"\pFDS Volume (%)");
		goto end;
	case _subexpchannels_mmc5_volume:
		setfvolume(&mmc5_vol,"\pMMC5 Volume (%)");
		goto end;
	case _subexpchannels_n163_volume:
		setfvolume(&n163_vol,"\pNamco 163 Volume (%)");
		goto end;
	case _subexpchannels_sfme_volume:
		setfvolume(&sfme_vol,"\pFME-07 Volume (%)");
		goto end;

	/* speed */
	case _subspeed_adj_dmc:
		RateFixFlags^= 0x40;
		FCEUI_SetTriangleScale(tri_scale); /* to rebuild the tables */
		break;
	case _subspeed_50:
	case _subspeed_63:
	case _subspeed_71:
	case _subspeed_79:
	case _subspeed_89:
	case _subspeed_100:
	case _subspeed_126:
	case _subspeed_159:
	case _subspeed_178:
	case _subspeed_200:
	case _subspeed_other:
		if ((itemcode+=-_subspeed_50)>=GET_ARRAY_COUNT(emulationspeeds))
		{
			stopsound();
			/* sound buffer clipping occurs for values lower than 17.3% */
			if ((td=valuedialog("\pCustom Emulation Speed",
					emuset.speed/10000000.,
					#if HAVE_HQ
						soundquality ? 66.6:
					#endif
						17.3,429,true))==INT_NO_VALUE)
				goto end;
			FCEUI_SetSpeed(emuset.speed=td*10000000.0+.5);
		} else
			FCEUI_SetSpeed(emuset.speed=emulationspeeds[itemcode]);
		RefreshThrottleFPS();
		break;

	/* background */
	case _subbackground_off:
	case _subbackground_full:
	case _subbackground_muted:
	case _subbackground_bg_muted:
		eoptions&=~(EO_BGRUN|EO_MUTE);
		if (--itemcode&1)
			eoptions|= EO_BGRUN;
		if (itemcode&2)
			eoptions|= EO_MUTE;
		break;

	default:
		if ((itemcode&0x7FC0)!=(submenuid_recent-menuid_apple<<6))
			break;
		itemcode&= 63;
		if (--itemcode==clearrecent(false))
			clearrecent(true);
		else
			takefilepath(getrecentpath(itemcode));
	}
	updatemenus();
end:
	return true;
}
