/*
 macos/messagelog.c
  and not to stderr... (no longer stderr)

  Created: 2008.07.12
  Revised: 2015.07.19

Sunday, July 20, 2008 7:48 PM EDST
  rewrote list mechanisms.
  free and malloc should be realloc(..,0) and realloc(NULL,..) to minimize code space.
Monday, April 5, 2010 1:54 PM
  removed unused setlogwindowscrolls.
  logwindow_scroll now caps to technical Classic limits (as it should've for a long time),
 also multiplies by magnitude only if applied (INT_NO_VALUE must be preserved if not).
*/

#include <stdio.h>

#include <MacWindows.h>
#include <Events.h>

#include "shell.h"
#include "macvideo.h"
#include "messagelog.h"
#include "listwindow.h"

/* ---- globals */

WindowRef	logwindow = 0;

/* note that a message log window can exist while the logfile doesnt have to */
static FILE	*logfile = 0;

uint8		logtofile = 1; /* *512KB clipped */

/* ---- code */

#define PASCAL_LOG_FONT_NAME		"\pMonaco"
#define TEXT_FONT_SIZE			9

#define MAXIMUM_LINESIZE		512 /* should be <=16000 */
#define MAXIMUM_LINECOUNT		200 /* 130 can dm all RAM; was 64, 100, 128, 130, 150 */
#define MAXIMUM_LOGSIZE			(512*1024) /* was 64KB */

#define INIT_LOGWINDOW_WIDTH		480
#define INIT_LOGWINDOW_HEIGHT		(9*11+18)

#define LOGWINDOW_MINIMUM_HEIGHT	(INIT_LOGWINDOW_HEIGHT-24)
#define LOGWINDOW_MINIMUM_WIDTH		196
#define LOGWINDOW_MAXIMUM_HEIGHT	(MAXIMUM_LINECOUNT/2*12+16)
#define LOGWINDOW_MAXIMUM_WIDTH		(MAXIMUM_LINESIZE/2*6)

void logwindow_scroll(int hcount,
	int vcount)
{
	listwindow_scroll(logwindow,hcount,vcount);
}

static WindowRef logwindow_create(void)
{
	Str255		title;
	Rect		bounds;

	getpstr(title,strlistid_names,namesstr_logfilename);
	/* bottom-center of screen */
	bounds = (**gamegdevice).gdRect;
	bounds.top = (bounds.bottom-=9)-INIT_LOGWINDOW_HEIGHT;
	bounds.left = bounds.left+bounds.right-INIT_LOGWINDOW_WIDTH>>1;
	bounds.right = bounds.left+INIT_LOGWINDOW_WIDTH;

	return listwindow_create((WindowRef)-1,&bounds,title,PASCAL_LOG_FONT_NAME,TEXT_FONT_SIZE,
		MAXIMUM_LINESIZE,MAXIMUM_LINECOUNT,0,1,FOUR_CHAR_CODE('mLog'));
}

int logmessage(const char *str,
	long length)
{
	long		position;
	int		result = 0;

	if (length<0)
		length = strlen(str);
	if (!logwindow)
		logwindow = logwindow_create();

	if (!logfile && logtofile)
	{
		getcstr(temporary,strlistid_names,namesstr_logfilename);
		if (temporary[0]) /* option: no log file by removing string (or strlist) */
			logfile = fopen(temporary,"w"); /* may fail (especially on a locked vol.) */
	}

	if (logfile && (uint32)(logtofile-1)>1
			&& (position=ftell(logfile))+length>=logtofile*MAXIMUM_LOGSIZE)
		fseek(logfile,0,SEEK_SET); /* overwriting self! */

	listwindow_append(logwindow,str,length,true);

	if (logfile && logtofile)
		result = fwrite(str,1,length,logfile);

	return result;
}

int logwindow_show(int w)
{
	if (!logwindow && (logwindow=logwindow_create())==NULL)
		return -1;
	return showhide(logwindow,w);
}

void logwindow_clear()
{
	listwindow_clear(logwindow);
}

int logwindow_scrap(int w)
{
	if (w<0)
		return listwindow_getsel(logwindow,NULL,NULL,NULL,NULL)>0;
	if (!w)
		return listwindow_copy(logwindow);
	if (w==IMAX(int))
		if (listwindow_select(logwindow,0,w,IMIN(int),0,false))
			return listwindow_update(logwindow,false);
	return 0;
}

void logwindow_update(boolean erase)
{
	listwindow_update(logwindow,erase);
}

int logwindow_click(EventRecord *event,
	short what)
{
	Rect		bounding;

	bounding.top = LOGWINDOW_MINIMUM_HEIGHT;
	bounding.left = LOGWINDOW_MINIMUM_WIDTH;
	bounding.bottom = LOGWINDOW_MAXIMUM_HEIGHT;
	bounding.right = LOGWINDOW_MAXIMUM_WIDTH;

	return listwindow_click(logwindow,event,what,&bounding);
}

void logwindow_activate(boolean active)
{
	listwindow_setactive(logwindow,active);
}
