/*
 macos/shell.c
  main code for MacFCEU

2012.01.03 - in fixing input throttle, what was throttle.c is now here.
*/
#define REVISION_DATE		"2015.08.29"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <Files.h>
#include <Script.h>
#include <Dialogs.h>
#include <Gestalt.h>
#include <AppleEvents.h>
#include <Sound.h>
#include <Timer.h>
#include <Devices.h>

#define TREC_STRUCTS

#include "../utility/ini.h"

#include "shell.h"
#include "../FCEU.h"
#include "../common/dface.h"
#include "keyscan.h"

#include "../PPU.h"
#include "../movie.h"
#include "../FDS.h"
#include "../fsound.h"
#include "../boards/Namcot.h"
#include "../boards/MMC5.h"


#include "shellmenus.h"
#include "macdialogs.h"
#include "messagelog.h"
#include "macvideo.h"
#include "debugger.h"
#include "config.h"
#include "tileview.h"
#include "memview.h"
#include "listwindow.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- constants */

#ifndef NO_CONFIG
 #define CONFIGFILE_PATH	"macfceu.ini"
#endif

/* ---- globals */

WindowRef	myfrontwindow;

char		temporary[1024] = {0};

static sc_type	softclose = 0;
boolean		suspended = false, mousehidden = false;
boolean		soundavailable, dragdropavail;

short		vmenubar = 20;

uint8		VMMask, s16baud, CPUMask = 0x80, sysVer = 0x71;

static char	defaultdirs[][256] =
{
	":cheats",
	":gameinfo",
	":sav",
	":fcs",
	":snaps",
	""	//was ":", couldn't really work as result; 2010.03.27
};

FCEUGI		*GI = NULL;

#ifdef env68k
 #define DEFAULT_FLAGS		EO_SNAPNAME|EO_AUTOFRAMESKIP
#else
 #define DEFAULT_FLAGS		EO_SNAPNAME
#endif
emu_settings	emuset = {1e9, DEFAULT_FLAGS};

snd_settings	sndset = {1, 0, 100, 22050, 0xFFF, 256, 256, 2, 3, 93<<9, 100, 100, 100, 100, 100, 100};

int		ntsccthb[7];

uint8		*xbsave = NULL; /* for manual updates */

/* timing */

/* ---- messages */

void FCEUD_Message(const char *s)
{
	logmessage(s,strlen(s));
}

void FCEUD_PrintError(const char *s)
{
	logmessage(s,strlen(s));
	logmessage("\n",1);
	if (mousehidden)
		{InitCursor(); mousehidden = false;}
	showdialog(dialogid_error,std_ok,"\pFCE Ultra Error",s,"\p",NULL,NULL);
	/* mouse cursor will hide automatically, if need be */
}

void splash()
{
	char		buf[200];

	sprintf(buf,
"MacFCEU " VERSION_STR "\n"
"Revised:   " REVISION_DATE "\n" /* __DATE__/__TIME__ are inconsistent (e.g., compiler date) */
"Compiler: " CVERSIONSTR "\n"
"http://macfceu.sourceforge.net\n"
"Initially based on FCE Ultra 0.98.13.\n"
"http://fceultra.sourceforge.net\n"
#ifdef __NVERS
		,__NVERS
#endif
		);
	showdialog(dialogid_about,1,"\pAbout MacFCEU",buf,"\p",NULL,NULL);
}

sc_type scexit(sc_type what)
{
	if (!what || softclose==_sc_quit) /* status, or if already terminating */
		goto end;
	if (what&0x80 && what!=_sc_quit)
		{softclose&= what; goto end;}
	if ((what==_sc_quit || what==(_sc_reopengame|_sc_closegame)) && GI)
	{ /* only if a game is active */
		stopsound();
		if (what==_sc_quit)
		{
			int		strc;

			if ((strc=countstrlist(strlistid_exit))==0)
				temporary[512] = 0;
			else
				getcstr(512+temporary,strlistid_exit,rand()%strc);
		} else
		{ /* now ordered instead of random like above */
			getcstr(512+temporary,strlistid_close,what&_sc_reopensame ? 1:0);
			if (GetKeyboard()[MAP_COMMAND>>3]&0x80>>(MAP_COMMAND&7))
				goto skip;
		}
		/* if no str (or its empty), use default string(s) */
		if (showdialog(dialogid_confirm,confirm_no,
				!temporary[512] ? NULL:
					what!=_sc_quit ? "\pClose game?":"\pQuit during game?",
				!temporary[512] ? "Are you sure you want to exit?":
					512+temporary,
				"\p",NULL,NULL)!=confirm_yes)
			goto end;
	}
skip:	if (what!=_sc_quit)
		what&= _sc_reopengame|_sc_closegame;
	softclose|= what;
end:	return softclose;
}

static void printinfo(void)
{
	FCEUD_Message(
"  MacFCEU " VERSION_STR //"  -- WARNING: DEBUG build!" --time to retire this
#ifdef env68k
"\n  -- for 680x0 Macs, performance may be unacceptably slow"
#endif
			);
	/*if (!soundavailable)
		FCEUD_Message("\n  -- Sound Manager 3.0+ not present (no sound playback)");
	else*/ if (s16baud)
		FCEUD_Message("\n  -- 16-bit sound support detected");
	/* update this if the default keys are changed */
	FCEUD_Message(
"\n\n"
"Default key assignments (see menus):\n"
"   `    Fast forward.\n"
"  H/V   Select NTSC hue or tint.\n"
"  -/=   Adjust hue/tint (cmd=volume adjust).\n"
"  0-9   Show state/movie slot (cmd=movie).\n"
"  F1    Show/hide PPU Tile Palette.\n"
"  F3    Show/hide background layer (cmd=fast).\n"
"  F4    Show/hide sprite layer.\n"
"  F5    Save state (cmd=movie).\n"
"  F6    Show/hide dips/disks.\n"
"  F7    Load state (cmd=movie).\n"
"  F8    Insert/Eject coin | enter barcode.\n"
"  F9    Take snapshot.\n"
"  F10   Toggle FPS (cmd=frame count).\n"
"  F11   Toggle last message.\n"
"  F12   Cycle input type #0, (cmd=#1,cmd+shift=FC).\n"
"\n"
			);
}

/* ---- private game code */

static void closegame()
{
	/* added in case a movie's active; 2011.04.16 */
	FCEUMOV_Stop();
	FCEUI_SelectMovie(0);

	debuggerwindow_clear(false);
	cleargamepath();
	if (!GI)
		return;
/*	debuggerwindow_show(0);*/
	FCEUI_CloseGame();
	GI = NULL;
	redomenusgi();
	debuggerwindow_clear(true);
}

/* ---- general code */

static void setdirectories(void)
{
	static int	jlist[6] =
	{
		FCEUIOD_CHEATS,FCEUIOD_MISC,FCEUIOD_NV,FCEUIOD_STATE,FCEUIOD_SNAPS,INT_NO_VALUE
	};
	char		*n;
	int		i;
	short		x = namesstr_cheatsdir;

	FCEUI_SetSnapName(eoptions&EO_SNAPNAME);

	do
	{
		n = defaultdirs[x-namesstr_cheatsdir];

		getcstr(n,strlistid_names,x);
		if ((i=jlist[x-1])==INT_NO_VALUE)
			break;
		mkdir(n); /* create directory */
		FCEUI_SetDirOverride(i,n);
	} while(++x);

	if (x==0)
		return;
	FCEUI_SetBaseDirectory(n);
}

int setcursor(int ID,
	boolean color)
{
	if (ID<0)
		{InitCursor(); return 0;}
	if (color)
	{
		CCrsrHandle	crsr;

		if ((crsr=GetCCursor(ID))==NULL)
			return -1;
		SetCCursor(crsr);
		DisposeCCursor(crsr);
	} else
	{
		CursHandle	cursor;

		if ((cursor=GetCursor(ID))==NULL)
			return -1;
		SetCursor(*cursor);
	}

	return 1;
}

/* previously in menus.c */
int gamewindowtofront()
{
	if (!gamewindow)
		return -1;
	if (vidset.fullscreen) /* in window mode only; 2009.10.06 22:02 EDT */
		return 0;

	BringToFront(gamewindow);

	return 1;
}

/* ---- shell menus */

static void initrecentmenu(void)
{
	MenuRef		menu;
	int		x;

	if ((menu=NewMenu(submenuid_recent,"\pOpen Recent"))==NULL)
		return;
	MacInsertMenu(menu,kInsertHierarchicalMenu);

	for(x=clearrecent(false); x>0; --x)
		AppendMenu(menu,"\p(...");
	getpstr((uint8*)temporary,strlistid_common,commonstr_clear);
	if (*temporary)
		AppendMenu(menu,(uint8*)temporary);
}

/* used to be public */
static boolean reinitmenus(void)
{ /* apple, file, edit, emulator, special, [hierachical] */
	MenuHandle	menu;
	int		i;
	boolean		gotall = true;

	ClearMenuBar();

	initrecentmenu();

	for(i=menuid_apple; i<=submenuid_background; ++i)
	{
		if ((menu=MacGetMenu(i))==NULL)
			{gotall = false; continue;}
		MacInsertMenu(menu,(!!(i>=submenuid_sound))*kInsertHierarchicalMenu);

		if (i==menuid_apple)
			AppendResMenu(menu,'DRVR');
	}
	drawmenubar();

	return gotall;
}

static boolean handlemenu(long itemhit,
	uint16 modifiers)
{
	HiliteMenu(0);

	/* sometimes the mouse remains hidden for dumb reasons; 2009.01.29 */
	mousehidden = true;

	/* the platform-independent form here does 9 bits menu ID, 6 bits menuitem */
	if ((itemhit+=-menuid_apple<<16)&0xFC00FFC0)
		return false;

	if ((unsigned)(itemhit-2)<=32765)
	{
#if TARGET_API_MAC_CARBON
		return false; /* shouldnt come here */
#else
		GetMenuItemText(GetMenuHandle(menuid_apple),itemhit,(UInt8*)temporary);
		OpenDeskAcc((UInt8*)temporary);
		return true;
#endif
	}

	return domenu((itemhit>>10&~63)|(itemhit&63)|(modifiers<<15-optionKeyBit&0x8000));
}

static void fixmenus(uint16 modifiers)
{
	int		d = modifiers&optionKey ? commonstr_savemovie-commonstr_savestate:0;

	setmenuitemstr(_file_savestate,getpstr(temporary,strlistid_common,commonstr_savestate+d));
	setmenuitemstr(_file_loadstate,getpstr(temporary,strlistid_common,commonstr_loadstate+d));
	updatewavestatus();
}

/* ---- keys */

static short getEventChar(const EventRecord *event)
{
	UInt32		state = 0;
	long		cache;
	short		accrued = (event->modifiers&0xF7FF)|(event->message>>8&0xFF);

	/* how can I tell if an error happened? */
	if ((cache=GetScriptManagerVariable(smKCHRCache))==0)
		return event->message&charCodeMask;
	return KeyTranslate((const void*)cache,accrued,&state);
}

static int handlekey(EventRecord *event,
	DialogRef dialog)
{
extern uint8 show_fps, show_fct, inputSnapped;
	long		menuitem;
	char		charcode = event->message&charCodeMask;
	boolean		did_snap;

	did_snap = inputSnapped;
	inputSnapped = 0;

	if (event->modifiers&cmdKey)
	{
		if (charcode==kFunctionKeyCharCode && (event->message&0xFF00)==0x6500)
			goto snap;
		if (!dialog && (menuitem=MenuKey(getEventChar(event)))!=0)
			{handlemenu(menuitem,event->modifiers); return -1;}
		/* special (unreliable) cmd-keys: home, top of list; end, end of list */
		if (charcode==kHomeCharCode)
			return 0x7F20;
		else if (charcode==kEndCharCode)
			return 0x7FFF;
		else
			switch(event->message>>8&255)
			{
			case MAP_KP0^7:
				FCEUI_DispMessage("Volume: 0%%");
				domenu(_subvolume_off);
				return -1;
			case MAP_KP1^7:
				FCEUI_DispMessage("Volume: 100%%");
				domenu(_subvolume_100);
				return -1;
			case MAP_KP2^7:
				domenu(_subvideo_2x_scale);
				return -1;
			}
	}

	if (dialog) /* dialog function is to handle its own key events */
		return 0;

	if (charcode==kFunctionKeyCharCode)
	{
		switch(event->message>>8&255)
		{
		case MAP_F1^7:
			domenu(_subvideo_tile_view); /* toggle */
			return -1;
		case MAP_F9^7:
		snap:	if (event->what==keyDown && !did_snap) /* don't repeat */
				inputDoSnap(event->modifiers&cmdKey);
			return -1;
		case MAP_F10^7:
			if (event->modifiers&cmdKey)
				show_fct = !show_fct;
			else
				show_fps = !show_fps;
			return -1;
		case MAP_F15^7:
			if (IsMenuItemEnabled(GetMenuHandle(menuid_emulator),_emulator_debug))
				{checkmenuitem(_emulator_debug,debuggerwindow_show(-1)>0); return -1;}
		}
	} else if (charcode==kEnterCharCode)
	{ /* note that enter is ignored when the debugger window is hidden/uninitialized */
		if (getvisiblewindowport(debuggerwindow))
		{ /* command-enter: step 16 */
			checkmenuitem(_emulator_debug,debuggerwindow_step(event->modifiers&cmdKey ? 16:1)>0);
			return -1;
		}
	} else if (charcode==kPageUpCharCode
			|| charcode==kPageDownCharCode
			|| charcode==kUpArrowCharCode
			|| charcode==kDownArrowCharCode)
	{
		if (!gamewindow || getvisiblewindowport(debuggerwindow))
		{
			int		cmdcode, vcnt = charcode;
	
			if ((cmdcode=(event->modifiers&cmdKey)<<16)!=0
					&& charcode==kUpArrowCharCode)
				cmdcode = 0, vcnt =-1<<16; /* hack scroll fix; 2010.04.10 */
			else
				vcnt|= cmdcode, vcnt|= 0x10000;
			listwindow_scroll(myfrontwindow==MemViewWindow ? MemViewWindow:logwindow,
				cmdcode>>8,vcnt);
			return -1;
		}
	} else if (myfrontwindow==MemViewWindow && getvisiblewindowport(MemViewWindow))
		memview_key(charcode);
	return 0; /* invoke standard proc */
}

/* ---- windows */

static int closewindow(WindowRef window)
{
	int		r = 1;

	if (!window)
		return -1;
	if (window==gamewindow)
		softclose|= _sc_closegame;
	else if (window==logwindow)
		{domenu(_edit_log); r = 0;}
	else if (window==debuggerwindow)
		{checkmenuitem(_emulator_debug,debuggerwindow_show(0)>0); r = 0;}
	else if (window==tilewindow)
		domenu(_subvideo_tile_view);
	else if (window==MemViewWindow)
		domenu(_edit_memview);
	//else
	//	HideWindow(window);
	return r;
}

int forewindow(int w)
{
	if (w<0)
		return !!FrontWindow();
	if (!w)
		return closewindow(FrontWindow());
	return 0;
}

static void selLowerWindow(WindowRef window,
	boolean hilite)
{
	while((window=MacGetNextWindow(window))!=NULL)
		if (IsWindowVisible(window))
		{
			if (window!=gamewindow)
			{
				if (hilite)
					SelectWindow(window);
				HiliteWindow(window,hilite);
			}
			break;
		}
}

int showhide(WindowRef win,
	int w)
{
	if (w==IMIN(w))
		return !!getvisiblewindowport(win);
	HiliteWindow(win,true);

	if (w<0)
		w = !getvisiblewindowport(win);
	else
		w = !!w;

	if (!w)
		selLowerWindow(win,true);
	ShowHide(win,w);
	if (!w)
		return 0;
	HiliteWindow(win,true);

	/* otherwise, youd have to do window mode; 2009.09.27 09:11 EDST */
	if (w&vidset.fullscreen)
		SelectWindow(win);
	selLowerWindow(win,false);
	return w;
}

/* ---- events */

static void suspend(void)
{
	suspended = true;
	ShowMenuBar();

	logwindow_activate(0); /* updated manually; 2010.05.27 */
	debuggerwindow_activate(0);
	memview_activate(0);
}

static void resume(void)
{
	WindowRef	win;

	suspended = mousehidden = false;
	mycolorsinvalidate = 2; /* update all colors upon return */
	win = FrontWindow();
	if (win==logwindow) /* updated manually; 2010.05.27 */
		logwindow_activate(0);
	else if (win==MemViewWindow)
		memview_activate(0);
}

int handleevent(EventRecord *event,
	DialogRef dialog)
{
	Rect		bounds;
	WindowRef	window;
	int		hit, part;
	/* cursor show/hide moved here so menubar shows for all callers; 2009.10.08 08:55 EDT */
	if (!vidset.fullscreen)
	{
		if (mousehidden)
			{mousehidden = false; goto show;}
	} else if (gamewindow)
		if (dialog || event->where.v<vmenubar)
		{
		show:
			ShowMenuBar();
			if (mousehidden)
				{InitCursor(); mousehidden = false;}
		} else if (!suspended)
			{HideMenuBar(); mousehidden = false;} /* to hide cursor */

	myfrontwindow = NULL;
	if (!suspended)
	{
		if ((myfrontwindow=FrontWindow())!=NULL)
			if (!IsWindowVisible(myfrontwindow)
					|| !IsWindowHilited(myfrontwindow))
				myfrontwindow = MacGetNextWindow(myfrontwindow);
	}
	if ((long)myfrontwindow>0x10000000)
		exit(1);
	if (!event)
		return 0;

	hit =-1;
	switch(event->what)
	{
	case mouseDown:
		switch(part=FindWindow(event->where,&window))
		{
		/*case inNoWindow:
			return -1;*/
		case inMenuBar:
			fixmenus(event->modifiers);
			InitCursor(); //2011.03.05
			handlemenu(MenuSelect(event->where),event->modifiers);
			break;
		case inDrag:
			if (dialog && GetDialogWindow(dialog)!=window
					&& (event->modifiers&cmdKey)==0)
				{SysBeep(20); break;}
			/*if (fullscreen)
				break;*/
			/* the Mac OS writers made a mistake in taking the keys without option
			  in DragWindow; this is the very reason palette windows need extra code */
			/* heres somethin: letting go of the commandkey at the right time with
			  a dialog in front can give you the ability to front other windows */
			GetRegionBounds(GetGrayRgn(),&bounds);
			DragWindow(window,event->where,&bounds);
			hit =-3;
			break;
		case inGrow:
		case inContent:
			if (dialog && GetDialogWindow(dialog)!=window)
				{SysBeep(20); break;}
			if (window==logwindow)
				enablemenuitem(_edit_copy,logwindow_click(event,part)>2);
			else if (dialog && GetDialogWindow(dialog)==window
					&& (hit=dialogselect(dialog,event))==0)
				hit =-1; /* dont call standardmodal twice */
			else if (window==tilewindow)
				TileViewClick(event);
			else if (window==MemViewWindow)
				memview_click(event,part);
			else if (window==debuggerwindow)
				if (part==inGrow)
					debuggerwindow_grow(event);
				else
					debuggerwindow_click(event);
			if (!vidset.fullscreen)
				SelectWindow(window);
			break;
		case inGoAway:
			if (dialog && GetDialogWindow(dialog)!=window)
				{SysBeep(20); break;}
			if (!TrackGoAway(window,event->where))
				break;
			if (closewindow(window)>0)
				hit =-2; /* signal close */
		}
		break;
	case keyDown:
	case autoKey:
		hit = handlekey(event,dialog);
		break;
	case activateEvt:
		HiliteMenu(0);

		/* one thing to piss people off; some dialogs that cut in front may
		  modal-disable the buttons on the program dialog */
		hit = 0;

		window = (WindowRef)event->message;
		if (!dialog)
			if (window==logwindow)
				logwindow_activate(event->modifiers&activeFlag);
			else if (window==debuggerwindow)
				debuggerwindow_activate(event->modifiers&activeFlag);
			else if (window==MemViewWindow)
				memview_activate(event->modifiers&activeFlag);
		break;
	case updateEvt:
		if ((window=(WindowRef)event->message)==NULL)
			break;
		BeginUpdate(window);
		if (window==gamewindow)
			FCEUD_BlitScreen(xbsave); /* update whatever we can */
		else if (window==logwindow)
			logwindow_update(true);
		else if (window==debuggerwindow)
			debuggerwindow_update(true);
		else if (window==tilewindow)
			BlitTileWindow(1);
		else if (window==MemViewWindow)
			memview_update(1);
		else if (GetDialogWindow(dialog)==window)
			dialogupdate(dialog);
		EndUpdate(window);
		break;
	case osEvt:
		if ((event->message&(long)0xFF000000)!=suspendResumeMessage<<24)
			break;
		InitCursor();

		if (event->message&resumeFlag)
			resume();
		else
			suspend();
		break;
	case kHighLevelEvent:
		AEProcessAppleEvent(event);
		break;

	default:
		hit = 0;
	}

	return hit;
}

#include "config.c"

/* ---- Apple Events[TM] */

boolean		handlersinstalled = false;

static OSErr eventcheck(const AppleEvent *event)
{
	DescType	type;
	Size		actual;
	OSErr		err;

	if ((err=AEGetAttributePtr(event,keyMissedKeywordAttr,typeWildCard,
			&type,0,0,&actual))==errAEDescNotFound)
		return noErr;

	return err==noErr ? errAEEventNotHandled:err;
}

static boolean openany(const FSSpec *file,
	boolean force)
{
	uint32		ftype;

	if ((ftype=extensiontotype(file->name,INT_NO_VALUE))!=0)
	{
		const char	*spath;

		if (ftype==FOUR_CHAR_CODE('FCM '))
		{
			if (!GI)
				opengame(NULL);
			else if ((spath=setlastfilepath(file->vRefNum,file->parID,
					file->name))!=NULL)
				FCEUI_LoadMovie(spath);
			return false;
		} else if (ftype==FOUR_CHAR_CODE('FCS '))
		{
			if (!GI)
				opengame(NULL);
			else if ((spath=setlastfilepath(file->vRefNum,file->parID,
					file->name))!=NULL)
				FCEUI_LoadState(spath);
			return false;
		}
	}
	if (!takefile(file))
	{
		if (force) /* a hang without this */
			softclose = _sc_reopengame|_sc_closegame;
		else
			scexit(_sc_reopengame|_sc_closegame);
		return true;
	}
	return false;
}

static pascal OSErr odocevent(const AppleEvent *event,
	AppleEvent *reply,
	long refcon)
{
#pragma unused(reply,refcon)
	AEDescList	list;
	long		index, count;
	OSErr		err;
	boolean		gotone = false;

	if ((err=AEGetParamDesc(event,keyDirectObject,typeAEList,&list))!=noErr
			|| (err=eventcheck(event))!=noErr
			|| (err=AECountItems(&list,&count))!=noErr)
		return err;

	for(index=1; index<=count; ++index)
	{
		FSSpec		file;
		Size		actual;
		DescType	type;
		AEKeyword	keyword;

		if ((err=AEGetNthPtr(&list,index,typeFSS,&keyword,
				&type,(Ptr)&file,sizeof(FSSpec),&actual))!=noErr)
			continue;
		if (!gotone)
			gotone = openany(&file,false);
	}
	return noErr;
}

static pascal OSErr oappevent(const AppleEvent *event,
	AppleEvent *reply,
	long refcon)
{
#pragma unused(reply,refcon)
	return eventcheck(event);
}

static pascal OSErr quitevent(const AppleEvent *event,
	AppleEvent *reply,
	long refcon)
{
#pragma unused(reply,refcon)
	OSErr		err;

	if ((err=eventcheck(event))!=noErr)
		return err;
	softclose = _sc_quit;

	return noErr;
}

static pascal OSErr pdocevent(const AppleEvent *event,
	AppleEvent *reply,
	long refcon)
{
#pragma unused(reply,refcon)
	OSErr		err;

	if ((err=eventcheck(event))!=noErr)
		return err;
	return errAEEventNotHandled;
}

static void installhandlers(void)
{
	AEEventHandlerUPP	handler;

	if (handlersinstalled)
		return;

	/* ordered in importance in case any fail */
	if ((handler=NewAEEventHandlerUPP(odocevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEOpenDocuments,handler,0,false);
	if ((handler=NewAEEventHandlerUPP(oappevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEOpenApplication,handler,0,false);
	if ((handler=NewAEEventHandlerUPP(pdocevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEPrintDocuments,handler,0,false);
	if ((handler=NewAEEventHandlerUPP(quitevent))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEQuitApplication,handler,0,false);

	handlersinstalled = true;
}

static void tryevents(void)
{
	EventRecord	event;
	int		retries = 60;

	do
		if (WaitNextEvent(highLevelEventMask,&event,0,NULL))
			AEProcessAppleEvent(&event);
	while(--retries);
}

#ifdef USE_DRAG_AND_DROP
/* where dragging is seen */
static pascal OSErr dragtracker(short msg,
	WindowRef window,
	void *refcon,
	DragReference dref)
{
#pragma unused(window,refcon)
	if (IsMenuItemEnabled(GetMenuHandle(menuid_file),0))
		if (msg==kDragTrackingEnterWindow)
		{
			Rect		bounds;
			RgnHandle	rgn;

			if ((rgn=NewRgn())!=NULL)
			{
				SetRect(&bounds,0,0,
					256<<screenshift,240<<screenshift);
				RectRgn(rgn,&bounds);
				ShowDragHilite(dref,rgn,true);
				DisposeRgn(rgn);
			}
		} else if (msg==kDragTrackingLeaveWindow)
			HideDragHilite(dref);
	return noErr;
}

/* where we actually take the frickin' file */
static pascal OSErr dragreceive(WindowRef window,
	void *refcon,
	DragReference dref)
{
#pragma unused(window,refcon)
	ItemReference	item;
	Size		z;
	HFSFlavor	r;
	uint16		c, x;
	boolean		gotone = false;

	if (IsMenuItemEnabled(GetMenuHandle(menuid_file),0)
			&& !CountDragItems(dref,&c))
		for(x=1; x<=c; ++x)
			if (!GetDragItemReferenceNumber(dref,x,&item)
					&& !GetFlavorDataSize(dref,item,flavorTypeHFS,&z)
					&& z<=sizeof(HFSFlavor)
					&& !GetFlavorData(dref,item,flavorTypeHFS,(Ptr)&r,&z,0))
				if (!gotone)
					gotone = openany(&r.fileSpec,true);
	return noErr;
}

static DragTrackingHandlerUPP	wtrackerupp;
static DragReceiveHandlerUPP	wreceiveupp;

static void uninstallreceiver(WindowRef window)
{
	if (wtrackerupp)
	{
		RemoveTrackingHandler(wtrackerupp, window);
		DisposeDragTrackingHandlerUPP(wtrackerupp);
		wtrackerupp = NULL;
	}
	if (wreceiveupp)
	{
		RemoveReceiveHandler(wreceiveupp,window);
		DisposeDragReceiveHandlerUPP(wreceiveupp);
		wreceiveupp = NULL;
	}
}

static int installreceiver(WindowRef window)
{
	if (!dragdropavail)
		return -1;
	if (!wtrackerupp)
		wtrackerupp = NewDragTrackingHandlerUPP(dragtracker);
	if (!wreceiveupp)
		wreceiveupp = NewDragReceiveHandlerUPP(dragreceive);
	if (wtrackerupp && wreceiveupp
			&& !InstallTrackingHandler(wtrackerupp,window,NULL)
			&& !InstallReceiveHandler(wreceiveupp,window,NULL))
		return 1;
	return 0;
}
#endif

/* ---- system check */

#define SysVersion		(*(const int16*)0x15A)
#define CPUFlag			(*(const int8*)0x12F)

static void prerequisites(long *val)
{
	/* endian macro test; 2010.04.12
	*val = 1;
	if (((uint8*)val)[3]!=BIG_ENDIAN)
		showalert(alertid_fatal,
"\pBIG_ENDIAN in compilation is not consistent with actual!",
			"\p","\p");
	*/
	/* there is no PPC Sys. 6 and lower */
#ifdef env68k
	/* no more printing of current */
	//if (Gestalt(gestaltSystemVersion,val))
	*val = SysVersion;
	if (*val<0x700)
	{ /* since we might be starting on a 68000, sprintf should be avoided here */
		getpstr(temporary,strlistid_errors,errorstr_oldsystem);
		goto dlg;
	} else
		sysVer = *val>>4;
	/* CPU selector should exist under 7.1+ */
	if (sysVer>=0x70)
		*val =-1, Gestalt(gestaltNativeCPUtype,val);
	else if ((*val=CPUFlag)&0x10)
		*val = gestaltCPU68040;
	if (*val<gestaltCPU68040)
		if (*val>gestaltCPU68010)
		{
			CPUMask = 1u<<(*val-gestaltCPU68000);
			//if (Gestalt(gestaltAddressingModeAttr,val)!=noErr
			//		|| !(*val&1<<gestalt32BitAddressing))
			//	goto incomp;
		} else
		{
		incomp:	getpstr(temporary,strlistid_errors,errorstr_oldcpu);
		dlg:	showalert(alertid_fatal,(UInt8*)temporary,"\p","\p");
		}
#endif
	if (!Gestalt(gestaltSoundAttr,val))
		s16baud = *val>>gestalt16BitAudioSupport&1;
	if (!Gestalt(gestaltVMAttr,val))
		VMMask = *val;

	/* install AE handlers for dragging in files */
	if (Gestalt(gestaltAppleEventsAttr,val)==noErr
			&& *val&1u<<gestaltAppleEventsPresent)
		installhandlers();
	if (Gestalt(gestaltDragMgrAttr,val)==noErr
			&& *val&1u<<gestaltDragMgrPresent)
		dragdropavail = true;
}

#if !TARGET_API_MAC_CARBON
extern void installPatches(void);
#endif

static void initialize(long int stackhigh)
{
#pragma unused(argc,argv)
	long		value = 0;

	MoreMasters();
	MoreMasters();
	MoreMasters();

# if TARGET_API_MAC_CARBON /* Carbon should already add more than enough stack space */
#  pragma unused(stackhigh)
# else
	SetApplLimit((Ptr)((-64*1024)+stackhigh)); /* adds about 64KB of stack space */
	MaxApplZone();
	InitGraf((Ptr)&qd.thePort);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0);
# endif
	prerequisites(&value);

//	remove("stderr");

	InitCursor();

	gamegdevice = GetMainDevice();

	srand(TickCount());
#if !TARGET_API_MAC_CARBON
	installPatches();
#endif
}

#ifndef TERM_CLRALL
 #define TERM_CLRALL		1
#endif
#if TERM_CLRALL
 #include "usercommands.h"
#endif

void SaveConfig()
{
#ifdef CONFIGFILE_PATH
	configfile_save(CONFIGFILE_PATH);
#endif
}

static void cleanup(void)
{
#ifdef CONFIGFILE_PATH
	SaveConfig();
#endif
	closegame();
	FCEUI_Kill();
	KillSound();
	FCEUI_EndWaveRecord();
#if TERM_CLRALL
	commandline_free();
#endif
	InitCursor();
	FlushEvents(mDownMask|mUpMask|keyDownMask|keyUpMask|autoKeyMask,0);
}

int update_soundsettings(int v)
{
	if (v<0)
	{ /* fixed: first set-up for variables (if off, called again with 0); 2013.05.05 */
		FCEUI_Sound(22050);

		InitSound(NULL); /* sometimes logged a zero sample rate without this; 2010.04.22 */
		FCEUI_SetSoundVolume(soundvolume);
	#if HAVE_HQ
		if (soundon && soundquality)
			FCEUI_SetSoundQuality(soundquality);
	#endif
		v = 0;
	}
	if (v<=0)
	{
		FCEUI_SetRectFilterAmp(rectangle_amp); //2010.03.23
		FCEUI_SetTNDFilterAmp(tnd_amp); //2013.05.04
		FCEUI_SetNoiseScale(noise_scale); //2011.02.04
		FCEUI_SetTriangleScale(tri_scale); //2013.05.05
		FCEUI_SetDCFilterLoss(dcfilter_loss); //2015.06.29
	}
	FCEU_FDSSetVolume(fds_vol); //2013.05.25
	FCEU_N163SetVolume(n163_vol); //2013.05.28
	FCEUI_SetVRC7Volume(vrc7_vol); //2013.06.02
	FCEU_MMC5SetVolume(mmc5_vol); //2013.06.04
	FCEUI_SetSFMEVolume(sfme_vol);
	FCEUI_SetVRC6Volume(vrc6_vol); //2013.06.05

	if (sndset.latency<4 || sndset.latency>125)
		sndset.latency = 8;
	return v;
}

/* ---- main */

static uint16	afskipc, afskipm;

void FCEUD_Update(uint8 *XBuf,
	const int16 *Buffer,
	int Count);

static int eventloop(void)
{
	int		result = 0;

	//if (!getnextgamepath()) /* fullscreen by default when loading by HLE (Finder) */
	//	vidset.fullscreen = 0;

	updatemenus();
	enablemenuitem(_subvideo_refresh_sync,
		VideoRetraceSync((GDHandle)-1)!=paramErr);

	printinfo(); /* print info regardless (now); 2009.01.18 */
	if (!getnextgamepath())
	{
		if (!vidset.fullscreen)
			domenu(_edit_log); /* show info (message log) if not in fullscreen */
		if (GetKeyboard()[6]&0x80) /* command key; was (eoptions&EO_FOAFTERSTART) */
			opengame(NULL);
	}
	redomenusgi();

	while(softclose!=_sc_quit)
	{
		EventRecord	event;

		if (!GI)
		{ /* no game */
			if (gamewindow)
				KillVideo(); /* dispose of window and reset palette */
			if (WaitNextEvent(everyEvent,&event,15,NULL))
			{
				InitCursor();
				handleevent(&event,NULL);
			}

			if (getnextgamepath())
			{
				mycolorsinvalidate|= 2;
				loadgame(NULL);
				softclose = 0; /* it kept closing & reopening...; 2010.04.10 */
			}
			continue;
		}

		if (!gamewindow)
		{ /* should have no window for no game */
			if (InitVideo(GI)<=0)
				break;
			InitSound(GI);
		}

		videorefresh(); /* for FPS; 2011.04.01 */
		updatemenus(); /* anything may have changed; 2010.03.23 */
		vmenubar = 20+(**gamegdevice).gdRect.top;
		InitCursor(); /* watch cursor... */

#ifdef USE_DRAG_AND_DROP
		installreceiver(gamewindow);
#endif
		/* a hack to fix square_timer_div2 */
		FCEUSND_Reset();
		do
		{
			static uint8	eventskip = 1, force;

			uint8		*gfx;
			int16		*sound;
			int32		ssize, f;

			if (inputCmdDown || !--eventskip)
			{ /* must take event regardless of return value; 2009.10.08 10:31 EDT */
				event.what = nullEvent;
				GetNextEvent(everyEvent,&event);
				handleevent(&event,NULL);
				eventskip = 2;
				force+= 0x20; //each 8
			}

			if (suspended)
			{
				if (eoptions&EO_BGRUN)
					goto emulate;
				continue;
			}
			if ((updatemouse() || vidset.fullscreen) && gamewindow==myfrontwindow)
			{ /* hide cursor when within window (window covers screen in fs) */
				if (!mousehidden && event.where.v>=vmenubar)
					{HideCursor(); mousehidden = true;}
			} else if (mousehidden || !force)
				{InitCursor(); mousehidden = false;}
		emulate:
			f = 0;
			if (!debug_open) /*debug_stepframe>=0 .. debug_stepcount>0 */
			{
				//f = eoptions&EO_ODDFRAMESKIP;
				if (NoWaiting&4 && ++afskipc<afskipm)
					f = 0x100;
				else
					afskipc = 0;
			}
			if (FCEUI_Emulate(&gfx,&sound,&ssize,screenshift ? 0:f))
				xbsave = gfx;
			if (softclose)
			{
			close:
#ifdef USE_DRAG_AND_DROP
				uninstallreceiver(gamewindow);
#endif
				closegame();
				if (softclose==_sc_reopengame)
					opengame(NULL);
				if (~softclose&(_sc_reopengame|_sc_closegame))
					softclose = 0;
				continue;
			}
			FCEUD_Update(f ? NULL:gfx,sound,ssize);
		} while(GI);

		xbsave = NULL;
		stopsound();

		//everythin' showin' up lookin' all fresh
		updatemenus();
		ShowMenuBar();
		InitCursor(); mousehidden = false;
	}
exit: /* restoration is automatic (atexit) */
	cleanup();

	return result;
}

int main(int argc,
	char **argv)
{
#pragma unused(argc)
	initialize((long int)&argv);

	if (!FCEUI_Initialize())
	{ /* memory problem (no need for dialog) */
		/*showalert(alertid_error,"\pFCEUI_Initialize failed!","\p","\p","\p");*/
		return 1;
	}

	if (!reinitmenus())
	{
		SysBeep(20);
		FCEUD_Message("## Some menus are missing?\n");
	}

	soundavailable = 1;//adequatesoundmanager();

	/* allow sound to be generated anyway; 2009.01.25 */

	if (handlersinstalled)
		tryevents();
#ifdef CONFIGFILE_PATH
	configfile_load(CONFIGFILE_PATH);
#endif
	FCEUI_SetVidSystem(palyo&=1);
	FCEUI_SetGameGenie(genie&=1);

	setdirectories();
	addrecent(NULL);

	/* update everything that may have been changed */

	if (eoptions&EO_CPALETTE)
		FCEUI_SetPaletteArray(cpalette);
	else
		FCEUI_SetPaletteArray(NULL);
	FCEUI_SetNTSC_HSB(ntsccol,ntschue,ntscsat*2,ntscbrt);

	update_soundsettings(-1);
	/* needs to be done after initsndvariables */
	emuset.speed = FCEUI_SetSpeed(emuset.speed);

	/* and lastly... */
	videorefresh();

	return eventloop();
}

/* ---- timing */

#include "../ftrig.h"

extern int minsndcount; /* in samples */

static uint64	updatetime; /* in microseconds */
int32		frametime, throttletime, sampletime; /* in 32MHz */

#if USETICKS
 #if USETICKS==2
  #include <time.h>
  boolean	usetickcount = 1; /* UNIX time on OS X is 100/sec */
  #define readtimer()		do{ \
	if(usetickcount) \
		(void)mulu64u32(updatetime,clock(),32000099/100); \
	else \
		Microseconds((UnsignedWide*)&updatetime); \
  }while(0)
 #else
  #define readtimer()		(void)mulu64u32(updatetime,TickCount(),32000059/60)
 #endif
#else
 #define readtimer()		Microseconds((UnsignedWide*)&updatetime)
#endif

uint64 FCEUD_GetTime()
{
	//Microseconds((UnsignedWide*)&time);
	return updatetime;
}

uint32 FCEUD_GetTimeFreq()
{
	return 1000000;
}

/* ---- throttle */

void RefreshThrottleFPS()
{
	uint64		numer;
	uint32		denom;
	/* because of mute, calc this anyway; 2012.03.02 */
	if ((denom=FCEUI_GetDesiredFPS())==0)
		denom = 1;
	_int64xset(numer,32000000*15625000.0);
	throttletime = uint64divu32(numer,denom);
	if (FSettings.OrgRate)
		sampletime = (32000000*16-15+(FSettings.OrgRate>>1))/FSettings.OrgRate;
	else
		sampletime = (32000000*16-15+2499)/4998;
	minsndcount = sndset.latency*32000/sampletime;
	frametime = 0; /* no first delay; determined in Update */
	afskipm = afbskip+(GI && GI->type==GIT_NSF ? 24:0); /* NSFs have less visual priority */
	NoWaiting&=~4; /* to fix switching off auto-skip on too-slow machines; 2013.06.02 */
	readtimer();
}

void FCEUD_Update(uint8 *XBuf,
	const int16 *sampledata,
	int samplecount)
{
	static int32	ltime = 0; /* allow wrap in case anything goes wrong */
	static uint8	ffskipc;
	int32		diff;

	if (!soundon || eoptions&EO_MUTE && (!(eoptions&EO_BGRUN) || suspended))
		samplecount = 0;
	if (NoWaiting&1)
	{
		if (ffskipc<ffbskip)
		{
			++ffskipc, XBuf = 0;
			readtimer();
			goto stime;
		}
		if (!XBuf && !afskipc)
			XBuf = xbsave; /* frame skipping fix; 2010.04.29 */
	}
	ffskipc = 0;
	writesound(sampledata,samplecount); /* does the effective waiting */
	/* throttle always on */
waiter:	readtimer();
	/* downshift the difference by two to know whether frametime is 4x
	  (waaay behind or ahead)
	*/
	if ((diff=((((signed)int64lo(updatetime)<<5)-ltime)>>2))<0)
	{
		diff =-diff, NoWaiting&=~(4|8);
		if (diff>frametime)
			goto set;
		if (!(NoWaiting&16))
			goto waiter;
	}
	if (eoptions&EO_AUTOFRAMESKIP)
	{ /* consistent slowdown; overrun sound playback to catch up */
		if (NoWaiting&8)
			NoWaiting|= 4; /* machine may be too slow for real-time */
		NoWaiting|= 8;
	}
	if (diff>frametime) /* four frames ahead?  reset */
	set:	ltime = int64lo(updatetime)<<5;
	else /* otherwise advance frame time */
		ltime+= frametime; //+sndoverrun;
stime:	frametime = throttletime;
	if (samplecount)
		frametime = samplecount*sampletime;
	if (XBuf)
	{
		FCEUD_BlitScreen(XBuf);
		if (GI)
		{
			if (fceutileview && GI->type!=GIT_NSF)
				BlitTileWindow(0);
			if (getvisiblewindowport(MemViewWindow))
				memview_update(0);
		}
	}
	FCEUD_UpdateInput();
}
