/*
 macos/shell.h

  MacFCEU 0.9d3

  Created: 2008.07.02

Thursday, April 1, 2010 2:11 PM EDT
  __GNUC__ appeared twice with the same code...
*/

#ifndef __SHELL_H

 /* ---- universal constants */

 #define INT_NO_VALUE			(long)0x80000000

 #if DOWNSAMPLE_HIGH
  #define MAXIMUM_SRATE			(100000*16)
 #else
  #define MAXIMUM_SRATE			(111860.75*16)
 #endif

 #define strlistid_names		80
 #define strlistid_errors		81
 #define strlistid_exit			82
 #define strlistid_close		83
 #define strlistid_common		84 /* prev. 161 */

 #if !defined(rez) && !defined(derez)
  #ifndef __FCEU_DRIVER_H
   #include "../driver.h"
  #endif

  #if defined(__MACTYPES__) && !defined(__MACOS7_9_H)
   #include "../zlib/contrib/macos/macos7-9.h"
  #endif

  #ifndef __MACGENERAL_H
   #include "macgeneral.h"
  #endif

/* ---- constants */

/* __VERSION__ isnt reliable; a string for CVERSIONSTR should show up with this */
#ifdef __MRC__
 #define CVERSIONSTR "MrC %X"
 #define __NVERS __MRC__
#elif defined(__SC__)
 #define CVERSIONSTR "SC %X"
 #define __NVERS __SC__
#elif defined(__MWERKS__)
 #define CVERSIONSTR "CodeWarrior %X"
 #define __NVERS __MWERKS__
#elif defined(SYMANTEC_CPLUS)
 #define CVERSIONSTR "Symantec C++ %X"
 #define __NVERS SYMANTEC_CPLUS
#elif defined(SYMANTEC_C)
 #define CVERSIONSTR "Symantec C %X"
 #define __NVERS SYMANTEC_C
#elif defined(SC_PLUS_THINK)
 #define CVERSIONSTR "Symantec C %X"
 #define __NVERS SC_PLUS_THINK
#elif defined(THINK_C)
 #define CVERSIONSTR "THINK C %X"
 #define __NVERS THINK_C
#elif defined(__PPCC__)
 #define CVERSIONSTR "PPCC %X"
 #define __NVERS __PPCC__
#elif defined(applec)
 #define CVERSIONSTR "MPW C %X"
 #define __NVERS applec
#elif defined(__GNUC__)
 #define CVERSIONSTR "gcc %X"
 #define __NVERS __GNUC__
#elif defined(__xlc)
 #define CVERSIONSTR "xlc %X"
 #define __NVERS __xlc
#elif defined(__xlC)
 #define CVERSIONSTR "xlc %X"
 #define __NVERS __xlC
#elif defined(__xlC__)
 #define CVERSIONSTR "xlc %X"
 #define __NVERS __xlC__
#elif defined(__XLC121__)
 #define CVERSIONSTR "xlc %X"
 #define __NVERS __XLC121__
#elif defined(_MSC_VER)
 #define CVERSIONSTR "Visual Studio C/C++ %X"
 #define __NVERS _MSC_VER
#elif defined(__MOTO__)
 #define CVERSIONSTR "mcc %X"
 #define __NVERS __MOTO__
#elif defined(_MIPS_ISA)
 #define CVERSIONSTR "MIPSpro %X"
 #define __NVERS _MIPS_ISA
#elif defined(__sparc)
 #define CVERSIONSTR "SPARCompiler %X"
 #define __NVERS __sparc
#else
 #define CVERSIONSTR "???" __VERSION__
#endif

enum
{
	EO_BGRUN		= 0x1,
	EO_NOSPRLIM		= 0x2,
	EO_CPALETTE		= 0x4,
	EO_CLIPSIDES		= 0x8, /* deliberately set to 8 for simple AND */

	EO_ODDFRAMESKIP		= 0x10, /* 2010.04.26 */

	EO_FSAFTERLOAD		= 0x20,
/*	EO_FOAFTERSTART = 0x0040, // replaced with key modifiers */
	EO_SHOWFPS		= 0x40,
	EO_SHOWFCT		= 0x80,
	EO_MUTE			= 0x100, /* 2009.12.28 */
	EO_SNAPNAME		= 0x200,

	EO_AUTOFRAMESKIP	= 0x800,

	EO_NRATEFIX		= 0x1000, /* just copies */
	EO_PRATEFIX		= 0x2000,
	EO_DRATEFIX		= 0x4000,

	/* other bits are either not applicable or always set */
	EO_NOFOURSCORE		= 0x8000
};

enum
{ /* strlist indices */
	errorstr_oldsystem	= 0,
	errorstr_oldcpu,

	namesstr_logfilename	= 0,
/*	namesstr_logtitle,
	namesstr_logfont,*/
	namesstr_cheatsdir,
	namesstr_miscdir,
	namesstr_nvdir,
	namesstr_statedir,
	namesstr_snapsdir,
	namesstr_defaultdir,

	commonstr_defaultwavename = 0,
	commonstr_defaultstatename,
	commonstr_defaultpalettename,
	commonstr_pickcolor, /* prompts */
	commonstr_savestate,
	commonstr_loadstate,
	commonstr_savemovie,
	commonstr_loadmovie,
	commonstr_logsound,
	commonstr_stopwave,
	commonstr_savepalette,
	commonstr_loadpalette,
	commonstr_opengame,
	commonstr_PAL_1_6MHz,
	commonstr_NTSC_1_6MHz,
	commonstr_all_1_6Mhz,
	commonstr_clear
};

enum
{ /* scexit bits (any non-zero value will close a game) */
	_sc_closegame		= 0x01,
	_sc_reopengame		= 0x02,
	_sc_reopensame		= 0x04,
	_sc_quit		= 0xFF
};

/* ---- typedefs */

typedef unsigned char	boolean;
typedef uint8		sc_type;

/* ---- globals */

/* temporary is intended for API usage (using it at higher levels may have it
  overwritten at lower levels */
extern char		temporary[1024];

extern uint8		NoWaiting, sysVer, VMMask, s16baud;
extern boolean		suspended, mousehidden, soundavailable;

extern int		ntsccthb[4+3];
#define ntsccol				ntsccthb[0]
#define ntscsat				*(float*)&ntsccthb[1]
#define ntschue				ntsccthb[2]
#define ntscbrt				ntsccthb[3]

typedef struct emu_settings
{
	uint32		speed;
	int		options;
	uint8		fskip, afskip;
	uint8		usegenie, pal;
	uint8		palette[192];
} emu_settings;

#ifdef TREC_STRUCTS
static trec	emu_trec[] =
{
	O_TREC(_uint32,emu_settings,speed),
	O_TREC(_hint32,emu_settings,options),
	O_TREC(_uint8,emu_settings,fskip),
	O_TREC(_uint8,emu_settings,afskip),
	O_TREC(_bool,emu_settings,usegenie),
	O_TREC(_bool,emu_settings,pal),
	C_TREC(emu_settings,palette),
	{0}
};
#endif

extern emu_settings	emuset;
#define eoptions			emuset.options
#define ffbskip				emuset.fskip
#define afbskip				emuset.afskip
#define genie				emuset.usegenie
#define palyo				emuset.pal
#define cpalette			emuset.palette

typedef struct snd_settings
{
	uint8		on;
	int8		nsfmif;
	int16		volume; //percent
	float		rate;
	uint16		activechannels;
	int16		rectamp, tndamp; //1.0=256
	uint8		nscale, tscale; //1.0=2,3
	uint16		dcfloss;
	int16		vrc6vol, vrc7vol, fdsvol, mmc5vol, n163vol, sfmevol; //percent
	float		latency; //in milliseconds
} snd_settings;

#ifdef TREC_STRUCTS
static trec	snd_trec[] =
{
	O_TREC(_bool,snd_settings,on),
	O_TREC(_int16,snd_settings,volume),
	O_TREC(_float,snd_settings,rate),
	O_TREC(_float,snd_settings,latency),
	O_TREC(_hint16,snd_settings,activechannels),
	O_TREC(_int16,snd_settings,rectamp),
	O_TREC(_int16,snd_settings,tndamp),
	O_TREC(_uint8,snd_settings,nscale),
	O_TREC(_uint8,snd_settings,tscale),
	O_TREC(_uint16,snd_settings,dcfloss),
	O_TREC(_int16,snd_settings,vrc6vol),
	O_TREC(_int16,snd_settings,vrc7vol),
	O_TREC(_int16,snd_settings,fdsvol),
	O_TREC(_int16,snd_settings,mmc5vol),
	O_TREC(_int16,snd_settings,n163vol),
	O_TREC(_int16,snd_settings,sfmevol),
	O_TREC(_int8,snd_settings,nsfmif),
	{0}
};
#endif

extern snd_settings	sndset;
#define soundon				sndset.on
//#define soundquality			sndset.quality
#define soundvolume			sndset.volume
#define rectangle_amp			sndset.rectamp
#define tnd_amp				sndset.tndamp
#define noise_scale			sndset.nscale
#define tri_scale			sndset.tscale
#define dcfilter_loss			sndset.dcfloss
#define sndactivechannels		sndset.activechannels
#define vrc6_vol			sndset.vrc6vol
#define vrc7_vol			sndset.vrc7vol
#define fds_vol				sndset.fdsvol
#define mmc5_vol			sndset.mmc5vol
#define n163_vol			sndset.n163vol
#define sfme_vol			sndset.sfmevol

/* ---- macros */

 /* alternative to memcpy and BlockMove so no function call is made
   (useful for dealing with unlocked handles) */
#define _acpy(d,s,c)			do{void *__d=d; const void *__s=s; long __z=c; \
	if(__z&1) *((int8*)__d)++=*((int8*)__s)++; \
	if(__z&2) *((int16*)__d)++=*((int16*)__s)++; \
	if(__z&4) *((int32*)__d)++=*((int32*)__s)++; \
	if (__z&=-8) do \
		*((int32*)__d)++=*((int32*)__s)++,*((int32*)__d)++=*((int32*)__s)++; \
	while(__z+=-8); \
}while(0)
 /* longer version */
 #define _bcpy(d,s,c)			do{void *__d=d; const void *__s=s; long __z=c; \
	if(__z&1) *((int8*)__d)++=*((int8*)__s)++; \
	if(__z&2) *((int16*)__d)++=*((int16*)__s)++; \
	if(__z&4) *((int32*)__d)++=*((int32*)__s)++; \
	if(__z&8) *((int32*)__d)++=*((int32*)__s)++, \
		*((int32*)__d)++=*((int32*)__s)++; \
	if(__z&16) *((int32*)__d)++=*((int32*)__s)++, \
		*((int32*)__d)++=*((int32*)__s)++, \
		*((int32*)__d)++=*((int32*)__s)++, \
		*((int32*)__d)++=*((int32*)__s)++; \
	if (__z&=-32) do \
		*((int32*)__d)++=*((int32*)__s)++,*((int32*)__d)++=*((int32*)__s)++, \
		*((int32*)__d)++=*((int32*)__s)++,*((int32*)__d)++=*((int32*)__s)++, \
		*((int32*)__d)++=*((int32*)__s)++,*((int32*)__d)++=*((int32*)__s)++, \
		*((int32*)__d)++=*((int32*)__s)++,*((int32*)__d)++=*((uint32*)__s)++; \
	while(__z+=-32); \
}while(0)

/* ---- prototypes */

int gamewindowtofront(void);
int forewindow(int w);

/* general */

void splash(void);
sc_type scexit(sc_type); /* pass zero to only get current status */

#ifdef __DIALOGS__
int handleevent(struct EventRecord *event,
	DialogRef dialog);
#endif
#ifdef __MACWINDOWS__
int showhide(WindowRef win,
	int w);
#endif

int setcursor(int ID,
	boolean color);

/* sound */

void writesound(const int16 *buffer,
	int count);
void stopsound();

int RecordSoundWave(const char *initdir);

int adjust_soundvolume(int displacement);

int update_soundsettings(int v);

void bds16_8(uint8*,
	const int16*,
	int32 c);

/* files */

char *makepathforfsspec(const struct FSSpec *file,
	int *outlength); /* can be NULL */

unsigned long extensiontotype(const uint8 *s,
	int z); /* pass negative to treat `s` as C-string, INT_NO_VALUE for P-string */

int tracknewpath(char **store, /* where to store for next getfilepath; can be NULL */
	const char *path,
	const char *initdir, /* can be NULL (add recent only) */
	uint16 nameoffset,
	int priority); /* lower = higher; negative = ignore; zero adds filepath */

int takefile(const struct FSSpec *file);
void takefilepath(const char *path); /* 2012.02.16 */

int loadgame(const char *filepath); /* can pass NULL (use current path string) */
void opengame(const char *initdir); /* if game is open, _sf_reopengame is set instead */

const char *getnextgamepath(void);
int cleargamepath(void); /* returns zero if one is pending */

/* pass NULL to only condense the list */
#define RECENTPATHSIZE		288
#define RECENTCOUNT		15
#define RECENTCOUNTSTR		"\\15"

int addrecent(const char *path);
const char *getrecentpath(int); /* temporary - may change */
int clearrecent(boolean clear); /* returns max. count */

/* these call getfilepath */
void SaveStateOrMovie(const char *initdir,
	boolean movie);
void LoadStateOrMovie(const char *initdir,
	boolean movie,
	boolean nodialog);

void SaveConfig(void);

/* input */

boolean updatemouse(void); /* returns true if within window content (game screen) */
int inputDoSnap(int record);

/* this was simply extern'd.  tsk, tsk. */
void parseinputgi(struct FCEUGI *gi);

/* throttle */

void RefreshThrottleFPS(void);
void SpeedThrottle(void);

 #endif /*!defined(rez) && !defined(derez)*/
 #define __SHELL_H
#endif
