/*
 macos/shellmenus.h

  Created: 2008.08.04
  Revised: 2015.08.21
*/

#ifndef __SHELLMENUS_H
 #define __SHELLMENUS_H

 #define menuid_apple			80
 #define menuid_file			81
 #define menuid_edit			82
 #define menuid_emulator		83
 #define menuid_special			84
 #define submenuid_sound		85
 #define submenuid_volume		86
 #define submenuid_video		87
 #define submenuid_channels		88
 #define submenuid_expchannels		89
 #define submenuid_speed		90
 #define submenuid_background		91
 #define submenuid_recent		92

 /* 'cause Apple made the Rez template mark char a `char`, and not a literal */
 #define submenuchr_sound		"\125"
 #define submenuchr_volume		"\126"
 #define submenuchr_video		"\127"
 #define submenuchr_channels		"\130"
 #define submenuchr_expchannels		"\131"
 #define submenuchr_speed		"\132"
 #define submenuchr_background		"\133"
 #define submenuchr_recent		"\134"

 #if !defined(rez) && !defined(derez)
  #ifndef __SHELL_H
   #include "shell.h"
  #endif

enum
{
	_apple_about		= 0001,

	_file_open		= (menuid_file-menuid_apple<<6)+1,
	_file_openrecent,
	_file_close,
	_filesep1,
	_file_savestate,
	_file_loadstate,
	_filesep2,
	_file_logsound,
	_filesep3,
	_file_saveconfig,
	_filesep4,
	_file_quit,

	_emulator_reset		= (menuid_emulator-menuid_apple<<6)+1,
	_emulator_power,
	_emulatorsep1,
	_emulator_debug,
	_emulator_step,

	_gamesub_sound		= (menuid_special-menuid_apple<<6)+1,
	_gamesub_volume,
	_gamesub_channels,
	_gamesub_expchannels,
	_gamesub_video,
	_gamesub_speed,
	_gamesep1,
	_game_force_1_6MHz,
	_game_PAL,
	_game_Genie,
	_gamesub_background,
	_gamesep2,
	/*_game_directories, -- eh, Windows-only, really */
	_game_input,
	_game_palette,
	_game_netplay,

	_edit_undo		= (menuid_edit-menuid_apple<<6)+1,
	_editsep1,
	_edit_cut,
	_edit_copy,
	_edit_paste,
	_edit_clear,
	_editsep2,
	_edit_select_all,
	_editsep3,
	_edit_log,
	_edit_memview,

	_subsound_latency = (submenuid_sound-menuid_apple<<6)+1,
	_subsound_sep1,
	_subsound_telephone,
	_subsound_lower,
	_subsound_low,
	_subsound_default,
	_subsound_moderate,
	_subsound_low_cd,
	_subsound_triple_22,
	_subsound_double_cd,
	_subsound_better,
	_subsound_best,
	_subsound_other,

	_subvolume_off		= (submenuid_volume-menuid_apple<<6)+1,
	_subvolume_10,
	_subvolume_20,
	_subvolume_30,
	_subvolume_40,
	_subvolume_50,
	_subvolume_60,
	_subvolume_70,
	_subvolume_80,
	_subvolume_90,
	_subvolume_100,
	_subvolume_110,
	_subvolume_120,
	_subvolume_130,
	_subvolume_140,
	_subvolume_150,
	_subvolume_other,

	_subvideo_fullscreen	= (submenuid_video-menuid_apple<<6)+1,
	_subvideo_2x_scale,
	_subvideo_refresh_sync,
	_subvideo_5to4, /* still a separator - not implemented properly */
	_subvideo_clip_sides,
	_subvideo_no_sprite_limit,
	//_subvideo_skip_odd_frames,
	_subvideo_auto_skip_frames,
	_subvideosep2,
	_subvideo_tile_view,

	_subchannels_rect_1	= (submenuid_channels-menuid_apple<<6)+1,
	_subchannels_rect_2,
	_subchannels_triangle,
	_subchannels_noise,
	_subchannels_DMC,
	_subchannelssep1,
	_subchannels_rect_amp,
	_subchannels_tnd_amp,
	_subchannels_tri_scale,
	_subchannels_noise_scale,
	_subchannels_dc_filter,

	_subexpchannels_enable	= (submenuid_expchannels-menuid_apple<<6)+1,
	_subexpchannelssep1,
	_subexpchannels_VRC6,
	_subexpchannels_VRC7,
	_subexpchannels_FDS,
	_subexpchannels_MM5,
	_subexpchannels_N163,
	_subexpchannels_SFME,
	_subexpchannelssep2,
	_subexpchannels_vrc6_volume,
	_subexpchannels_vrc7_volume,
	_subexpchannels_fds_volume,
	_subexpchannels_mmc5_volume,
	_subexpchannels_n163_volume,
	_subexpchannels_sfme_volume,

	_subspeed_adj_dmc	= (submenuid_speed-menuid_apple<<6)+1,
	_subspeedssep1,
	_subspeed_50,
	_subspeed_63,
	_subspeed_71,
	_subspeed_79,
	_subspeed_89,
	_subspeed_100,
	_subspeed_126,
	_subspeed_159,
	_subspeed_178,
	_subspeed_200,
	_subspeed_other,

	_subbackground_off	= (submenuid_background-menuid_apple<<6)+1,
	_subbackground_full,
	_subbackground_muted,
	_subbackground_bg_muted
};

/* ---- menu prototypes */

void reenablemenus(boolean enable);

void updatemenus(void);
void updatewavestatus(void); /* called by updatemenus */
boolean domenu(uint16 itemcode);
void redomenusgi(void);

void reenablemenus(boolean enable); /* for fake modal routines */

boolean setmenuitemstr(uint16 itemcode, const uint8 *text);
boolean enablemenuitem(uint16 itemcode, boolean enabled);
boolean enablemenu(uint16 itemcode, uint8 count, boolean enabled);
boolean checkmenuitem(uint16 itemcode, uint16 check); //bit 15: use diamond
boolean checkmenu(uint16 firstitemcode, uint8 index, boolean invert);
void drawmenubar(void);

void ShowMenuBar(void);
void HideMenuBar(void);
boolean ismenubarvisible(void);

#ifdef __MENUS__
EXTERN_API(Boolean) IsMenuItemEnabled(MenuRef menu, MenuItemIndex item);
EXTERN_API(RgnHandle) GetDragGrayBounds(void); /* added; 2009.10.09 05:25 EDT */
#endif

 #endif /*!defined(rez) && !defined(derez)*/
#endif /*!defined(__SHELLMENUS_H)*/
