/*
 macos/tileview.c
  replaces the FCE Ultra PPU.c function

  Created: 2009.09.28
  Revised: 2015.08.15

Monday, September 28, 2009 1:59 PM
  things have gotten pretty stupid here.
Tuesday, October 6, 2009 7:54 PM
  things have gotten smarter here.
Tuesday, October 6, 2009 10:28 PM
  TileView window now moves down if menubar clips it.
*/

#include <stdlib.h>

#include <MacWindows.h>

#include "shell.h"

#include "../PPU.h"
#include "../video.h"

#include "macdialogs.h"
#include "macvideo.h"
#include "tileview.h"

/* ---- globals */

WindowRef		tilewindow = NULL;
/* white, bright green, mid-red, black -- refer to palettes.h for the unv colors */
uint8			tileviewcolors[4] = {PAL_WHITE, PAL_LTGRAY, PAL_DRED, PAL_BLACK};
static uint32		lut8bit[256]; /* 1KB */
static uint8		tileconvert[37*8*128]; /* BG/FG VRAM, vna, tileview colors; 37KB */

/* ---- processing code */

static void rendertiles(void)
{
extern uint8 *VPage[8],*vnapage[4];
	uint32		*d = (uint32*)(1*8*128+tileconvert);
	uint32		A = 0, i = 32;

	do
	{
		const uint8	*C;
		int		j = 8;

		if ((C=VPage[A>>10])==NULL)
			continue;
		else
			C += A;
		do
		{
			int		k = 16;

			do
			{
				*d++ = lut8bit[(*C&0xF0)|(C[8]>>4)];
				*d++ = lut8bit[((*C<<4)&0xF0)|(C[8]&15)];
			} while(C+=16,--k);
		} while(C+=1-256,--j);
	} while(A+=256,--i);

	A = 0;
	do
	{
		const uint32	*s;
		int		j = 32;

		if ((s=(const uint32*)vnapage[A])==NULL)
			continue;
		do
		{
			*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
			*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
		} while(--j);
	} while(++A<4);
}

static void makelut8bit(unsigned m2) /* pass 1 for an *any* table (1 bit) */
{
	uint8		*d = (uint8*)lut8bit;
	uint8		bg = PALRAM[0]|0x80;
	unsigned	i = 0;

	do
	{ /* probably in LSB order (00 10 01 11) */
		uint8		v;

		if ((v=tileviewcolors[(i>>7)|((i>>2)&m2)])==0xFF)
			v = bg;
		*d++ = v;
		if ((v=tileviewcolors[((i>>6)&1)|((i>>1)&m2)])==0xFF)
			v = bg;
		*d++ = v;
		if ((v=tileviewcolors[((i>>5)&1)|(i&m2)])==0xFF)
			v = bg;
		*d++ = v;
		if ((v=tileviewcolors[((i>>4)&1)|((i+i)&m2)])==0xFF)
			v = bg;
		*d++ = v;
	} while(++i<sizeof(lut8bit)/4);
}

static void initcolors(void)
{
	makelut8bit(2);

	/* paint clickable colors */
	{
		uint32		*d = (uint32*)tileconvert;
		/* gray pattern for 0xFF */
		uint32		c0 = tileviewcolors[0]==0xFF ? 0xFF00FF00:lut8bit[0];
		uint32		c1 = tileviewcolors[1]==0xFF ? 0xFF00FF00:lut8bit[0xF0];
		uint32		c2 = tileviewcolors[2]==0xFF ? 0xFF00FF00:lut8bit[0x0F];
		uint32		c3 = tileviewcolors[3]==0xFF ? 0xFF00FF00:lut8bit[0xFF];
		uint32		c4 = c0==0xFF00FF00 ? 0x00FF00FF:c0;
		uint32		c5 = c1==0xFF00FF00 ? 0x00FF00FF:c1;
		uint32		c6 = c2==0xFF00FF00 ? 0x00FF00FF:c2;
		uint32		c7 = c3==0xFF00FF00 ? 0x00FF00FF:c3;
		int		i = 4; /* c,w,h 4x16x8 */

		do
		{ /* only do left half for active colors on right; 2009.12.27 04:37 EST */
			*d++ = *d++ = *d++ = *d++ = c0;
			*d++ = *d++ = *d++ = *d++ = c1;
			*d++ = *d++ = *d++ = *d++ = c2;
			*d++ = *d++ = *d++ = *d++ = c3;
			d += 16;
			*d++ = *d++ = *d++ = *d++ = c4;
			*d++ = *d++ = *d++ = *d++ = c5;
			*d++ = *d++ = *d++ = *d++ = c6;
			*d++ = *d++ = *d++ = *d++ = c7;
			d += 16;
		} while(--i);
	}
}

/* added; 2009.12.27 04:38 EST */
static void paintactivecolors(void)
{
	uint32		p[32];
	uint32		*d, v;
	int		i;

	/* make table of four for each (c,w,h 32x4x4) */
	i = 0, d = p;
	do
		v = epal_inv[PALRAM[i]], v|= v<<8, v|= v<<16, *d++ = v;
	while(++i<sizeof(p)/sizeof(p[0]));

	d = 16+(uint32*)tileconvert;

	/* background-layer colors */
	i = 4;
	do
	{
		const uint32	*s = p;

		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d = *s;
		d += 17;
	} while(--i);

	/* sprite-layer colors */
	i = 4;
	do
	{
		const uint32	*s = 16+p;

		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
		*d++ = *s++, *d++ = *s++, *d++ = *s++, *d = *s;
		d += 17;
	} while(--i);
}

/* ---- window code */

void SetTileViewColors(uint8 *p)
{
	*(uint32*)tileviewcolors = *(uint32*)p;
	initcolors();
}

static int initializetileview(void)
{
	if (lut8bit[1]==0)
		initcolors();

	if (tilewindow==NULL)
	{ /* create window, make sure fore/back color is black/white, respectively */
		RGBColor	color;
		GrafPtr		oldport;
		Ptr		storage;

		if ((storage=NewPtr(256))!=NULL)
			*(SInt32*)(252+storage) = 0;
		/* yes, relying on the definition's default position */
		if ((tilewindow=GetNewCWindow(windowid_screen,storage,(WindowRef)-1))==NULL)
		{
			if (storage!=NULL)
				DisposePtr(storage);
			return -1;
		}

		SetWRefCon(tilewindow,FOUR_CHAR_CODE('tile'));
		SetWTitle(tilewindow,"\pTile View");

		GetPort(&oldport);
		SetPortWindowPort(tilewindow);

		*(int32*)&color.red = 0UL, color.blue = 0;
		RGBForeColor(&color);
		*(int32*)&color.red = -1UL, color.blue = -1U;
		RGBBackColor(&color);

		SetPort(oldport);
	}

	if (!fceutileview)
		return 0;

	MacShowWindow(tilewindow);

	if (vidset.fullscreen)
		BringToFront(tilewindow);

	return 1;
}

static int		lastvoidcolor = -1;

/* void color painted on right, above vna (where nothing was before) */
static void paintvoidcolor(uint8 pel)
{
	static const Rect	bounds = {0, 128, 8, 256};

	RGBColor	color;

	if (pel==lastvoidcolor)
		return;

	if (pel!=0xFF)
	{
		color = gameclut->ctTable[pel].rgb;
		RGBForeColor(&color);

		PaintRect(&bounds);

		*(int32*)&color.red = 0UL, color.blue = 0;
		RGBForeColor(&color);
	} else
	{ /* gray pattern for n/a */
		Pattern		graypattern;

		PenPat(GetQDGlobalsGray(&graypattern));

		*(int32*)&color.red = 0UL, color.blue = 0;
		RGBForeColor(&color);

		PaintRect(&bounds);

		PenNormal();
	}
}

/* at one point, this was split into two windows, but it was never compiled */
int BlitTileWindow(int update)
{
	static PixMap	tilemap =
	{ /* everything is just two 1 bit 8x8 OR'd together! */
		(Ptr)tileconvert, /* base address */
		0x8080, /* row size */
		{0, 0, 264, 128}, /* bounds */
		0, /* PixMap version (0=fixed/direct) */
		0, 0, /* pack type, size (both unused in this case) */
		72<<16, 72<<16, /* resolution h, v */
		clutType, 8, /* pixel type, size */
		1, 8, /* component count, size per */
		0, /* plane size or pixel format */
		&gameclut /* fake CTabHandle */
	};
	static BitMap	vnamap =
	{
		33*8*128+(Ptr)tileconvert, /* base address */
		0x0010, /* row size */
		{0, 0, 256, 128} /* bounds */
	};

	/* back should be white; make sure the fore color is black for CopyBits */
	Rect		dstbounds;

	GDHandle	olddevice;
	CGrafPtr	oldport, port;
	const PixMap	**dstblitmap;
	const BitMap	*portmap;
	long		oldseed = 0;

	if (tilewindow==NULL)
		initializetileview();

	if ((port=getvisiblewindowport(tilewindow))==NULL)
		return 0;

	GetGWorld(&oldport,&olddevice);
	SetGWorld(port,NULL);

	/* paint any colors that've changed */
	if (tileviewcolors[0]==0xFF
			|| tileviewcolors[1]==0xFF
			|| tileviewcolors[2]==0xFF
			|| tileviewcolors[3]==0xFF)
		initcolors();

	paintactivecolors();

	if (update==0)
		lastvoidcolor = -1;
	paintvoidcolor(PPUvoidcolor);

	/* render! */
	rendertiles();

	/* adjust window's top if too high, fix the seed (faster) */
	if ((dstblitmap=GetPortPixMap(port))!=NULL)
	{
		const PixMap	*dstmap = *dstblitmap;

		if (gamewindow!=NULL && dstmap->pixelSize>=8)
		{
			oldseed = gameclut->ctSeed;
			gameclut->ctSeed = (**dstmap->pmTable).ctSeed;
		}

		dstbounds = dstmap->bounds;
		if (dstbounds.top>-32) /* above the menubar (sometimes happens while in fs) */
			MacMoveWindow(tilewindow,-dstbounds.left,32,false);
	}

	portmap = GetPortBitMapForCopyBits(port);

	dstbounds = tilemap.bounds;
	CopyBits((BitMap*)&tilemap,portmap,&tilemap.bounds,&dstbounds,srcCopy,NULL);

	dstbounds.top = 8, dstbounds.left = 128, dstbounds.right = 256;
	CopyBits(&vnamap,portmap,&vnamap.bounds,&dstbounds,srcCopy,NULL);

	if (dstblitmap!=NULL)
		gameclut->ctSeed = oldseed;

	SetGWorld(oldport,olddevice);

	return 1;
}

int TileViewClick(const EventRecord *event)
{
	Point		ml;

	if (event==NULL || getvisiblewindowport(tilewindow)==NULL)
		return -1;

	ml = event->where;
	localwindowpoint(tilewindow,&ml);

	if ((unsigned)(ml.v>>=2)<2)
	{
		uint8		*color, pel, mask = 0xFF;

		if ((unsigned)(ml.h>>=2)<16)
			color = (ml.h>>2)+tileviewcolors;
		else if (ml.h<32)
		{
			color = (PALRAM-16)+ml.v*16+ml.h, mask = 0x3F;
			if (color<PALRAM || color>=PALRAM+sizeof(PALRAM))
				color = PALRAM; /* make sure it's in bounds */
		} else
			color = &PPUvoidcolor;
		pel = *color&mask;
		if (PaletteDialog(&pel,mask+1)==std_ok)
		{
			*color = pel;
			if (color>=PALRAM && color<PALRAM+sizeof(PALRAM))
				FCEUPPU_UpdatePalette();

			initcolors();
			BlitTileWindow(0); /* finally, blit it */
			return 1;
		}
	}

	return 0;
}

int ShowHideTileView(int show)
{
	if (tilewindow==NULL)
		initializetileview();

	if (show<0)
		if (show==INT_NO_VALUE)
			return show; /* current value */
		else
			show = fceutileview^1;
	fceutileview = show&1;

	ShowHide(tilewindow,show);
	if (vidset.fullscreen&show)
		BringToFront(tilewindow);

	return show;
}
