/*
 macos/usercommands.h
  for commandline.c

  Created: 2008.08.11
  Revised: 2012.02.09

Wednesday, August 19, 2009 10:08 PM EDST
 rewriting the strict nature of the 'command line' interpreter (i.e., commandline_parseparameter).
*/

#ifndef __USERCOMMANDS_H
 #define __USERCOMMANDS_H

 #include "../utility/interpreter.h"

/* ---- constants */

#define LONGINT_MIN			((long int)1<<(sizeof(long int)<<3)-1)
#define LONGINT_MAX			(~LONGINT_MIN)

/* ---- typedefs/structures */

/* argc<0 means print help text, <-1 to print info items on it */
typedef int (*keyword_proc)(int argc, char const*const*argv);

typedef struct commandproc_keyword
{
	const char	*name;
	keyword_proc	proc;
} commandproc_keyword;

/* ---- prototypes */

 #ifdef __MACWINDOWS__
int commandline_update(WindowRef window, /* can be NULL (render nothing) */
	const Rect *inset, /* can be NULL (render nothing) */
	const char *text, /* can be NULL (no text) */
	short length,
	short position,
	unsigned char overwriting);
 #endif

int commandline_key(char *tbuf, /* can be NULL (do nothing) */
	short bufsize,
	short *tlength, /* can be NULL (do nothing) */
	short *tposition, /* can be NULL */
	unsigned char *insertmode, /* can be NULL */
	char charcode,
	unsigned char option);

int commandline_scroll(char *tbuf, /* can be NULL (do nothing) */
	short bufsize,
	short *tlength,
	short direction);

int commandline_storenext(const char *text, /* can be NULL (do nothing) */
	short length);

int commandline_free(void); /* free everything allocated; 2010.03.07 */

keyword_proc commandline_execute(const char *text,
	int length,
	const commandproc_keyword *keywords,
	keyword_proc defaultproc,
	const char *(*getvalue)(const char *s, const char *e,
		int32 *i, void*), /* can be NULL */
	unsigned char *userdata,
	evalvalue (*evalindir)(evalvalue)); /* can be NULL */

keyword_proc keywords_getproc(const char *name,
	long length,
	const commandproc_keyword *keywords);

#endif /*!defined(__USERCOMMANDS_H)*/
