/*
 mappers/BTLSMB2.c
  bootleg "Super Mario Bros. 2"

 * MacFCEU 0.6.4
 *
 * Copyright notice for this file:
 *  40.c, 43.c, 50.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"

static void FP_FASTAPASS(1) BTLSMB2IRQ(int a)
{
	if (IRQa)
		if ((IRQCount+=a)>=0x1000) //prev. added regardless of IRQa
		{
			IRQa = 0;
			X6502_IRQBegin(FCEU_IQEXT);
		}
}

/* ---- mapper 40: BTL SMB2 type A */

static DECLFW(M40W8)
{
(void)A; (void)V;
	IRQa = 0, IRQCount = 0;
	X6502_IRQEnd(FCEU_IQEXT);
}

static DECLFW(M40WA)
{
(void)A; (void)V;
	IRQa = 1;
}

static DECLFW(M40WE)
{
(void)A;
	ROM_BANK8(0xC000,V&7); //0
}

void Mapper40_init(void)
{
	ROM_BANK8(0x6000,~1); //6
	ROM_BANK8(0x8000,~3); //4
	ROM_BANK8(0xA000,~2); //5
//	ROM_BANK8(0xE000,7);
	SetWriteHandler(0x8000,0x9FFF,M40W8);
	SetWriteHandler(0xA000,0xBFFF,M40WA);
	SetWriteHandler(0xC000,0xDFFF,0);
	SetWriteHandler(0xE000,0xFFFF,M40WE);
	SetCartRead(0x6000,0x7FFF);
	MapIRQHook = BTLSMB2IRQ;
}

/* ---- mapper 43: BTL SMB2 type C */

#define reg	GameMemBlock

static SFORMAT M43StateRegs[] =
{
	{&IRQCount,4|FCEUSTATE_RLSB,"IRQC"}, /* should be 4 for backward-compatibility */
	{&IRQa,1,"IRQA"},
	{reg,1,"REG"},
	{0}
};

static void M43Sync(int v)
{
(void)v;
	setprg8(0xC000,reg[0]);
}

static DECLFW(M43W4022)
{
	uint8		transo[8] = {4,3,4,4,4,7,5,6};
(void)A;
	reg[0] = transo[V&7]; //Nestopia would just do =V&1
	M43Sync(0);
}

static DECLFW(M43W4122)
{
(void)A;
	X6502_IRQEnd(FCEU_IQEXT);
	IRQa = V&3, IRQCount = 0;
}

static void M43Power(void)
{
	reg[0] = ROM_size>=64>>4;
	setprg4(0x5000,16);
	setprg8(0x6000,2);
	setprg8(0x8000,1);
	setprg8(0xA000,0);
	setprg8(0xE000,9);
	setchr8(0);
	M43Sync(0);
	SetWriteHandler(0x4022,0x4022,M43W4022);
	SetWriteHandler(0x4122,0x4122,M43W4122); //do 8122 as well?
	SetCartRead(0x6000,0xFFFF);
}

void BTL_SMB2jC_Init(CartInfo *info)
{
	info->Power = M43Power;
	GameStateRestore = M43Sync;
	MapIRQHook = BTLSMB2IRQ;
	AddExState(&M43StateRegs,~0,0,0);
}

/* ---- mapper 50: BTL SMB2 type B */

static void M50Restore(int v)
{
(void)v;
	setprg8(0xC000,mapbyte1[0]);
}

static DECLFW(M50W)
{
	if ((A&0xD060)==0x4020)
		if (A&0x100)
		{
			IRQa = V&1;
			if (!IRQa)
				IRQCount = 0;
			X6502_IRQEnd(FCEU_IQEXT);
		} else
		{
			V = (V<<2&4)|(V>>1&1)|(V>>1&2)|(V&8);
			mapbyte1[0] = V;
			setprg8(0xC000,V);
		}
}

void Mapper50_init(void)
{
	SetWriteHandler(0x4020,0x5FFF,M50W);
	SetCartRead(0x6000,0xFFFF);
	MapStateRestore = M50Restore;
	MapIRQHook = BTLSMB2IRQ;

	setprg8(0x6000,15);
	setprg8(0x8000,8);
	setprg8(0xA000,9);
	setprg8(0xC000,0);
	setprg8(0xE000,11);
}
