/*
 mappers/FFE.c
  FFE mappers 6, 8, 17

 * MacFCEU 0.5.5
 *
 * Copyright notice for this file:
 *  6.c, 8.c, 17.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"

#define FFEmode mapbyte1[0]

/* ---- mapper 6: F3xxx / F4xxx */

#define FVRAM_BANK8(A,V) do{ \
	VPage[0]=VPage[1]=VPage[2]=VPage[3]= \
		VPage[4]=VPage[5]=VPage[6]=VPage[7]= \
		V?&MapperExRAM[(V)<<13]-(A):&CHRRAM[(V)<<13]-(A); \
	CHRBankList[0]=((V)<<3); \
	CHRBankList[1]=((V)<<3)+1; \
	CHRBankList[2]=((V)<<3)+2; \
	CHRBankList[3]=((V)<<3)+3; \
	CHRBankList[4]=((V)<<3)+4; \
	CHRBankList[5]=((V)<<3)+5; \
	CHRBankList[6]=((V)<<3)+6; \
	CHRBankList[7]=((V)<<3)+7; \
	PPUCHRRAM=0xFF; \
}while(0)

static void FP_FASTAPASS(1) FFEIRQHook(int a)
{
	if (IRQa)
		if ((IRQCount+=a)>=0x10000)
		{
			X6502_IRQBegin(FCEU_IQEXT);
			IRQa=IRQCount=0;
		}
}

static DECLFW(M6_write)
{
	if (A<0x8000)
	{
		switch(A)
		{
		case 0x42FE:onemir(V>>3&2);FFEmode=V; break;
		case 0x42FF:MIRROR_SET(V>>4&1); break;
		case 0x4501:IRQa=0;X6502_IRQEnd(FCEU_IQEXT); break;
		case 0x4502:IRQCount&=0xFF00;IRQCount|=V; break;
		case 0x4503:IRQCount&=0xFF;IRQCount|=V<<8;IRQa=1; break;
		}
	} else if (FFEmode&0x80)
		setchr8(V);
	else
	{
		ROM_BANK16(0x8000,V>>2);
		FVRAM_BANK8(0x0000,V&3);
	}
}

void M6Restore(int v)
{
	int x;
(void)v;
	x = 0;
	do
	{
		if ((PPUCHRRAM&(1<<x))==0)
			continue;
		if (CHRBankList[x]>7)
			VPage[x] = &MapperExRAM[(CHRBankList[x]&31)*0x400]-(x*0x400);
		else
			VPage[x] = &CHRRAM[(CHRBankList[x]&7)*0x400]-(x*0x400);
	} while(++x<8);
}

void Mapper6_init(void)
{
	ROM_BANK16(0xC000,7);
	MapStateRestore = M6Restore;
	MapIRQHook = FFEIRQHook;
	SetWriteHandler(0x4020,0x5fff,M6_write);
	SetWriteHandler(0x8000,0xffff,M6_write);
}

/* ---- mapper 8: F3xxx type B */

static DECLFW(M8_write)
{
(void)A;
        ROM_BANK16(0x8000,V>>3);
        VROM_BANK8(V&7);
}

void Mapper8_init(void)
{
	ROM_BANK32(0);
        SetWriteHandler(0x8000,0xFFFF,M8_write);
}

/* ---- mapper 17: F8xxx / SMxxxx */

static DECLFW(M17_write)
{
	switch(A)
	{
	case 0x42FE:
		onemir((V>>4)&1);
		break;
	case 0x42FF:
		MIRROR_SET(V>>4&1);
		break;
	case 0x4501:IRQa=0;X6502_IRQEnd(FCEU_IQEXT);break;
	case 0x4502:IRQCount&=0xFF00;IRQCount|=V;break;
	case 0x4503:IRQCount&=0x00FF;IRQCount|=V<<8;IRQa=1;break;
	case 0x4504: ROM_BANK8(0x8000,V);break;
	case 0x4505: ROM_BANK8(0xA000,V);break;
	case 0x4506: ROM_BANK8(0xC000,V);break;
	case 0x4507: ROM_BANK8(0xE000,V);break;
	case 0x4510: VROM_BANK1(0x0000,V);break;
	case 0x4511: VROM_BANK1(0x0400,V);break;
	case 0x4512: VROM_BANK1(0x0800,V);break;
	case 0x4513: VROM_BANK1(0x0C00,V);break;
	case 0x4514: VROM_BANK1(0x1000,V);break;
	case 0x4515: VROM_BANK1(0x1400,V);break;
	case 0x4516: VROM_BANK1(0x1800,V);break;
	case 0x4517: VROM_BANK1(0x1C00,V);break;

	default: break;
	}
}

void Mapper17_init(void)
{
	MapIRQHook = FFEIRQHook;
	SetWriteHandler(0x4020,0x5fff,M17_write);
}
