/*
 mappers/Sunsoft5B.c

 * MacFCEU 0.8a9
 *
 * Copyright notice for this file:
 *  69.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"

static void AYSound(int32 Count);
static void AYSoundHQ(void);
static void DoAYSQ(int x);
static void DoAYSQHQ(int x);

#define sunselect mapbyte1[0]
#define sungah    mapbyte1[1]

static uint8 sunindex;

static DECLFW(SUN5BWRAM)
{
	if (!(~sungah&0xC0))
		(WRAM-0x6000)[A] = V;
}

static DECLFR(SUN5AWRAM)
{
	if ((sungah&0xC0)==0x40)
		return A>>8;	//XFrame.DB
	return CartBR(A);
}

static DECLFW(M69_SWL)
{
(void)A;
	sunindex = V%14;
}

static DECLFW(M69_SWH)
{
	GameExpSound.Fill[SCSFME] = AYSound;
	GameExpSound.HiFill[SCSFME] = AYSoundHQ;
	if (FSettings.SndRate)
		switch(sunindex)
		{
		case 0:
		case 1:
		case 8: A = 0; goto doAY;
		case 7:
		#if HAVE_HQ
			if (FSettings.soundq) /* both 0 and 1 */
				DoAYSQHQ(0);
			else
		#endif
				DoAYSQ(0);
		case 2:
		case 3:
		case 9: A = 1;
		doAY:
		#if HAVE_HQ
			if (FSettings.soundq)
				DoAYSQHQ(A);
			else
		#endif
				DoAYSQ(A);
			break;
		case 4:
		case 5:
		case 10: A = 2; goto doAY;
		}
	MapperExRAM[sunindex] = V; 
}

static DECLFW(M69_write)
{
	if ((A&=0xE000)==0x8000)
		{sunselect = V; return;}
	else if (A!=0xA000)
		return;
	switch(sunselect&15)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:	VROM_BANK1(sunselect<<10,V);
		break;
	case 8:	sungah = V;
		if (V&0x40)
		{
			if (V&0x80) // Select WRAM
				setprg8r(0x10,0x6000,0);
		} else
			setprg8(0x6000,V);
		break;
	case 9: ROM_BANK8(0x8000,V); break;
	case 10: ROM_BANK8(0xA000,V); break;
	case 11: ROM_BANK8(0xC000,V); break;
	case 12: if (V&2)
			onemir(V&1);
		else
			MIRROR_SET(V&1);
		break;
	case 13: IRQa = V; X6502_IRQEnd(FCEU_IQEXT); break;
	case 14: IRQCount&= 0xFF00, IRQCount|= V; X6502_IRQEnd(FCEU_IQEXT); break;
	case 15: IRQCount&= 0x00FF, IRQCount|= V<<8; X6502_IRQEnd(FCEU_IQEXT); break;
	}
}

static int32 vcount[3];
static int32 dcount[3];
static int CAYBC[3];

static int32	volume = 96<<8; /*x<<=6, x+=x>>1*/

void FCEUI_SetSFMEVolume(int16 v)
{
#if HAVE_HQ
	if (v<0 && FSettings.soundq)
		v =-v;
#endif
	volume = v*(24<<8)/25;
}

static void DoAYSQ(int x)
{
	int32	freq, amp;
	int32	V, start = CAYBC[x], end = (SOUNDTS<<16)/soundtsinc;

	if (end<=(V=start) || (amp=MapperExRAM[8+x]&15)==0
			/*|| MapperExRAM[7]&1u<<x*/)
		return;
	amp = amp*volume>>8;
	freq = ((MapperExRAM[x<<1]|((MapperExRAM[x+x+1]&15)<<8))+1)<<4+17;
	CAYBC[x] = end;
	do
	{
		if (dcount[x])
			FCEUWave[V>>4]+= amp;
		vcount[x]-= nesincsize;
		while(vcount[x]<=0)
			dcount[x]^= 1, vcount[x]+= freq;
	} while(++V<end);
}

static void DoAYSQHQ(int x) 
{
	int32	freq, amp, V;

	if ((amp=MapperExRAM[8+x]&15)!=0) /* && !(MapperExRAM[7]&1u<<x))*/
	{
		amp = amp*volume>>8;
		freq = ((MapperExRAM[x<<1]|((MapperExRAM[x+x+1]&15)<<8))+1)<<4;
		for(V=CAYBC[x]; V<SOUNDTS; ++V)
		{
			if (dcount[x])
				FCEUWave[V]+= amp;
			if (--vcount[x]<=0)
				dcount[x]^= 1, vcount[x] = freq;
		}
	}
	CAYBC[x] = SOUNDTS;
}

static void AYSound(int32 Count)
{
	DoAYSQ(0);
	DoAYSQ(1);
	DoAYSQ(2);
	CAYBC[0] = CAYBC[1] = CAYBC[2] = Count;
}

static void AYSoundHQ(void)
{
	DoAYSQHQ(0);
	DoAYSQHQ(1);
	DoAYSQHQ(2);
}

static void AYHiSync(int32 ts)
{
	CAYBC[0] = CAYBC[1] = CAYBC[2] = ts;
}

static void FP_FASTAPASS(1) SunIRQHook(int a)
{
	if (IRQa)
		if ((IRQCount-=a)<=0)
		{
			X6502_IRQBegin(FCEU_IQEXT);
			IRQa = 0, IRQCount = 0xFFFF;
		}
}

void M69Restore(int v)
{
(void)v;
	if (mapbyte1[1]&0x40)
	{
		if (mapbyte1[1]&0x80) // Select WRAM
			setprg8r(0x10,0x6000,0);
	} else
		setprg8(0x6000,mapbyte1[1]);
}

void Mapper69_ESI(void)
{
	GameExpSound.RChange[SCSFME] = Mapper69_ESI;
	GameExpSound.HiSync[SCSFME] = AYHiSync;
	memset(dcount,0,sizeof(dcount));
	memset(vcount,0,sizeof(vcount));
	memset(CAYBC,0,sizeof(CAYBC));
}

void NSFAY_Init(void)
{ /* end changed from FFFF to EFFF to resolve conflict with N163 board */
	sunindex = 0;
	SetWriteHandler(0xC000,0xDFFF,M69_SWL);
	SetWriteHandler(0xE000,0xEFFF,M69_SWH); /* Nestopia does only E000 */
	Mapper69_ESI();
}

void Mapper69_init(void)
{
	NSFAY_Init();

	SetupCartPRGMapping(0x10,WRAM,8192,1);

	SetWriteHandler(0x8000,0xBFFF,M69_write);
	SetWriteHandler(0x6000,0x7FFF,SUN5BWRAM);
	SetReadHandler(0x6000,0x7FFF,SUN5AWRAM);
	MapIRQHook = SunIRQHook;
	MapStateRestore = M69Restore;
}
