/*
 mappers/VRC246.c
  Konami mappers 21, 23, 24, 25, 26

 * MacFCEU 0.8a9
 *
 * Copyright notice for the original files:
 *  21.c, 23.c, 24and26.c, 25.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"

static uint8	ktype;
static uint16	acount = 0;

static void FP_FASTAPASS(1) KonamiIRQHook(int a)
{
	if (!IRQa)
		return;
	if ((acount+=a*3)>=LCYCS) //appears to be correct (-3+3)
		do
		{
			acount-= LCYCS;
			if (++IRQCount==0x100) //was &0x100
			{
				X6502_IRQBegin(FCEU_IQEXT);
				IRQCount = IRQLatch;
			}
		} while(acount>=LCYCS);
}

#define K4buf mapbyte2
#define K4IRQ mapbyte1[1]
#define K4sel mapbyte1[0]

static DECLFW(VRCIVwrite8)
{
(void)A;
	ROM_BANK8((K4sel<<13&0x4000)|0x8000,V);
}

static DECLFW(VRCIVwrite9)
{
	if (ktype!=0x41)
	{
		if (ktype==0x35)
			A|= A>>2;
		else
			A|= (A>>2)|(A>>4)|(A>>6);
		if ((A&=3)==ktype>>6) //VRCIV does mirror at line 1 instead of 0
		{
			if (V==0xFF) //???
				return;
		mir:
			if (V&2)
				onemir(V&1);
			else
				MIRROR_SET(V&1);
			return;
		}
	} else
	{
		A|= A>>5;
		if (!(A&=6))
			goto mir;
	}
	if ((K4sel^V)&2)
	{
		uint8	swa = PRGBankList[0];

		ROM_BANK8(0x8000,PRGBankList[2]);
		ROM_BANK8(0xC000,swa);
	}
	K4sel = V;
}

static DECLFW(VRCIVwriteBCDE)
{
	int		x;

//	if (UNIFchrrama) {VROM_BANK8(0); return;}
	if (ktype!=0x41)
	{
		A|= (A>>2)|(A>>4)|(A>>6&3), A&= 0xF003;
		x = (A>>1&1)|(A-0xB000>>11);
		A = A<<2&4;
	} else
	{
		if (ktype==0x35)
			A = (A&0xF003)|(A>>2&3), x = (A&1)|(A-0xB000>>11);
		else
			A|= A>>5&15, x = (A>>2&1)|(A-0xB000>>11);
		A&= 2, A+= A;
	}
	K4buf[x]&= 0xF0>>A, V&= 15;
	VROM_BANK1(x<<10,K4buf[x]|=V<<A);
}

static DECLFW(VRCIVwriteF)
{
	V&= 15;
	if (ktype==0x41)
		A|= A>>5;
	else if (ktype==0x35)
		A|= A>>2;
	else
		A|= (A>>2)|(A>>4)|(A>>6);
	if ((A&=3)==0)
		IRQLatch&= 0xF0, IRQLatch|= V&15;
	else if (--A==0) //1
		IRQLatch&= 0x0F, IRQLatch|= V<<4;
	else if (--A==0) //2
	{
		IRQCount = IRQLatch, IRQa = V&2, K4IRQ = V&1, acount = 0;
		goto irqend;
	} else //3
	{
		IRQa = K4IRQ;
	irqend:
		X6502_IRQEnd(FCEU_IQEXT);
	}
}

static void VRCIV_init(void)
{
	SetWriteHandler(0x8000,0x8FFF,VRCIVwrite8);
	SetWriteHandler(0x9000,0x9FFF,VRCIVwrite9);
	SetWriteHandler(0xA000,0xAFFF,ROM_BANK8W);
	SetWriteHandler(0xB000,0xEFFF,VRCIVwriteBCDE);
	SetWriteHandler(0xF000,0xFFFF,VRCIVwriteF);
	MapIRQHook = KonamiIRQHook;
	AddExState(&acount,sizeof(acount),1,"ACNT");
	ktype = 0;
}

/* ---- mapper 21: modified VRCIV-2 (I think) */

static void FP_FASTAPASS(1) KonamiIRQHook4(int a)
{
#define LCYCS4 ((227*2)+1) //I've no reason to think that this's correct...
	if (!IRQa || (acount+=a*4)<LCYCS4) //acount+=a*3;
		return;
	do
	{
		acount-= LCYCS4;
		if (++IRQCount==0x100)
		{
			X6502_IRQBegin(FCEU_IQEXT);
			IRQCount = IRQLatch;
		}
	} while(acount>=LCYCS4);
}

void Mapper21_init()
{
	VRCIV_init();
	MapIRQHook = KonamiIRQHook4;
	ktype = 0x41;
}

/* ---- mapper 23: VRCII/VRCIV-2; has WRAM */

void Mapper23_init()
{
	VRCIV_init();
	ktype = ROM_size>=512>>4 ? 0x43:0x23;
}

/* ---- mapper 24: VRCVI type a */

static void (*sfun[3])(void);

#define VRCtemp mapbyte1[0]
#define VPSG2 mapbyte3
#define VPSG mapbyte2

static void DoSQV1(void);
static void DoSQV2(void);
static void DoSawV(void);

static DECLFW(VRC6SW)
{
	A&=0xF003;
	if ((A-=0x9000)<=2)
	{
		VPSG[A]=V;
		if (sfun[0])
			sfun[0]();
	} else if ((A-=0xA000-0x9000)<=2)
	{
		VPSG[4|A]=V;
		if (sfun[1])
			sfun[1]();
	} else if ((A-=0xB000-0xA000)<=2)
	{
		VPSG2[A]=V;
		if (sfun[2])
			sfun[2]();
	}
}

static DECLFW(VRCVI_write9AB)
{
	if (ktype==0x61)
		A = (A&~3)|(A>>1&1)|((A+A)&2);
	if ((A&=0xF003)==0xB003)
		if ((V>>=2)&2) //was case 0,4,8,12
			onemir(V&1);
		else
			MIRROR_SET2(~V&1);
	else
		VRC6SW(A,V);
}

static DECLFW(VRCVI_writeDE)
{
	if (ktype==0x61)
		A = (A&~3)|(A>>1&1)|((A+A)&2);
	VROM_BANK1((A<<10&0xC00)|(~A&0x1000),V); //evaluated correct
}

static DECLFW(VRCVI_writeF)
{
	if (ktype==0x61)
		A = (A&~3)|(A>>1&1)|((A+A)&2);
	if ((A&=3)==0)
		IRQLatch = V;
	else if (--A==0) //1
	{
		VRCtemp = V&1;
		if ((IRQa=V&2)!=0)
			IRQCount = IRQLatch, acount = 0;
		X6502_IRQEnd(FCEU_IQEXT);
	} else if (--A==0) //2
	{
		IRQa = VRCtemp;
		X6502_IRQEnd(FCEU_IQEXT);
	}
}

/* ---- VRCVI sound */

static int32	CVBC[3];
static int32	vcount[3];
static int32	dcount[2];
static int32	phaseacc = 0;
static uint8	b3 = 0;

static int32	volume = 3<<(8-2);

void FCEUI_SetVRC6Volume(int16 v)
{
#if HAVE_HQ
	if (v<0 && FSettings.soundq)
		v =-v;
#endif
	volume = v*(3<<(8-4))/25;
}

static INLINE void DoSQV(int x)
{
	int32	V, start = CVBC[x], end = (SOUNDTS<<16)/soundtsinc;
	int32	amp;

	if (end<=(V=start))
		return;
	CVBC[x] = end;

	if (!(VPSG[(x<<2)|2]&0x80))
		return;
	amp = (VPSG[x<<2]&15)*volume;
	if (VPSG[x<<2]&0x80)
		do
			FCEUWave[V>>4]+= amp;
		while(++V<end);
	else
	{
		int32	thresh = VPSG[x<<2]>>4&7;
		int32	freq = (VPSG[(x<<2)|1]|((VPSG[(x<<2)|2]&15)<<8))+1<<17;

		do
		{
			if (dcount[x]>thresh) //Greater than, not >=.  Important.
				FCEUWave[V>>4]+= amp;
			vcount[x]-= nesincsize;
			while(vcount[x]<=0) //Should only be <0 in a few circumstances.
			{
				vcount[x]+= freq;
				dcount[x]=(dcount[x]+1)&15;
			}
		} while(++V<end);
	}
}

static void DoSQV1(void)
{
	DoSQV(0);
}

static void DoSQV2(void)
{
	DoSQV(1);
}

static void DoSawV(void)
{
	int	V;
	int32	start = CVBC[2], end = (SOUNDTS<<16)/soundtsinc;

	if (end<=start)
		return;
	CVBC[2] = end;

	if (VPSG2[2]&0x80)
	{
		static int32	saw1phaseacc = 0, duff = 0;
		uint32		freq3 = (VPSG2[1]+((VPSG2[2]&15)<<8)+1)<<18;

		for(V=start; V<end; ++V)
		{
			saw1phaseacc-= nesincsize;
			if (saw1phaseacc<=0)
			{
				do
				{
					saw1phaseacc+= freq3;
					phaseacc+= VPSG2[0]&0x3F;
					if (++b3==7)
						b3 = phaseacc = 0;
				} while(saw1phaseacc<=0);
				duff = ((phaseacc>>3)&0x1F)*volume;
			}
			FCEUWave[V>>4]+= duff;
		}
	}
}

static INLINE void DoSQVHQ(int x)
{
	int32		V;
	int32		amp;

	if ((VPSG[(x<<2)|2]&0x80)==0)
		goto endo;
	amp = (VPSG[x<<2]&15)*volume;
	if (VPSG[x<<2]&0x80)
		for(V=CVBC[x]; V<SOUNDTS; ++V)
			FCEUWave[V]+= amp;
	else
	{
		int32	thresh = VPSG[x<<2]>>4&7;

		for(V=CVBC[x]; V<SOUNDTS; ++V)
		{
			if (dcount[x]>thresh) //Greater than, not >=.  Important.
				FCEUWave[V]+= amp;
			if (--vcount[x]<=0) //Should only be <0 in a few circumstances.
			{
				vcount[x] = (VPSG[(x<<2)|1]|((VPSG[(x<<2)|2]&15)<<8))+1;
				dcount[x] = (dcount[x]+1)&15;
			}
		}
	}
endo:	CVBC[x] = SOUNDTS;
}

static void DoSQV1HQ(void)
{
	DoSQVHQ(0);
}

static void DoSQV2HQ(void)
{
	DoSQVHQ(1);
}

static void DoSawVHQ(void)
{
	int32		V, duff;

	if (!(VPSG2[2]&0x80))
		goto endo;
	duff = ((phaseacc>>3)&0x1F)*volume;
	for(V=CVBC[2]; V<SOUNDTS; ++V)
	{
		FCEUWave[V]+= duff;
		if (--vcount[2]<=0)
		{
			vcount[2] = VPSG2[1]+((VPSG2[2]&15)<<8)+1<<1;
			phaseacc+= VPSG2[0]&0x35;
			if (++b3==7)
				b3 = phaseacc = 0;
				duff = ((phaseacc>>3)&0x1F)*volume;
		}
	}
endo:	CVBC[2] = SOUNDTS;
}


void VRC6Sound(int32 Count)
{
	int	x;

	DoSQV1();
	DoSQV2();
	DoSawV();
	x = 0;
	do
		CVBC[x] = Count;
	while(++x<3);
}

void VRC6SoundHQ(void)
{
	DoSQV1HQ();
	DoSQV2HQ();
	DoSawVHQ();
}

void VRC6SyncHQ(int32 ts)
{
	int	x;

	x = 0;
	do
		CVBC[x] = ts;
	while(++x<3);
}

static void VRC6_ESI(void)
{
	GameExpSound.RChange[SCVRC6] = VRC6_ESI;
	GameExpSound.Fill[SCVRC6] = VRC6Sound;
	GameExpSound.HiFill[SCVRC6] = VRC6SoundHQ;
	GameExpSound.HiSync[SCVRC6] = VRC6SyncHQ;

	memset(CVBC,0,sizeof(CVBC));
	memset(vcount,0,sizeof(vcount));
	memset(dcount,0,sizeof(dcount));
	/* 2015.06.29 */
	memset(VPSG,0,8);
	memset(VPSG2,0,3);

	if (FSettings.SndRate)
	{
	#if HAVE_HQ
		if (FSettings.soundq)
		{
			sfun[0] = DoSQV1HQ;
			sfun[1] = DoSQV2HQ;
			sfun[2] = DoSawVHQ;
		} else
	#endif
		{
			sfun[0] = DoSQV1;
			sfun[1] = DoSQV2;
			sfun[2] = DoSawV;
		}
	} else
		memset(sfun,0,sizeof(sfun));
}

void Mapper24_init(void)
{
        SetWriteHandler(0x8000,0x8FFF,ROM_BANK16W);
        SetWriteHandler(0x9000,0xBFFF,VRCVI_write9AB);
        SetWriteHandler(0xC000,0xCFFF,ROM_BANK8W);
        SetWriteHandler(0xD000,0xEFFF,VRCVI_writeDE);
        SetWriteHandler(0xF000,0xFFFF,VRCVI_writeF);
        VRC6_ESI();
        MapIRQHook = KonamiIRQHook;
	AddExState(&acount,sizeof(acount),1,"ACNT");
	ktype = 0x60;
}

/* ---- mapper 25: VRCIV-2 */

void Mapper25_init(void)
{
	VRCIV_init();
	ktype = 0x35;
}

/* ---- mapper 26: VRCVI type B */

void Mapper26_init(void)
{
        Mapper24_init();
	ktype = 0x61; //pins 9 and 10 are in opposite order
}

void NSFVRC6_Init(void)
{
	VRC6_ESI();
	SetWriteHandler(0x8000,0xBFFF,VRC6SW);
}
