/*
 mappers/VRCVII.c

 * MacFCEU 0.8a9
 *
 * Copyright notice for this file:
 *  85.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"

#define VRCtemp mapbyte1[0]
static uint8	indox;

#include "emu2413.h"

static int32	VRCVIIAmp;

static int	acount = 0;

static OPLL	*VRC7Sound = NULL;
static int	dwave = 0;

void FCEUI_SetVRC7Volume(int16 v)
{
#if HAVE_HQ
	if (FSettings.soundq)
		v = ABS(v);
#endif
	VRCVIIAmp = (v<<8)/5;
}

void DoVRC7Sound(void)
{
	int32		z, a;
#if HAVE_HQ
	if (FSettings.soundq)
		return; /* rendered by UpdateOPLNEO */
#endif
	z = ((SOUNDTS<<16)/soundtsinc)>>4;
	a = z-dwave;

	OPLL_fill(VRC7Sound,&FCEUWave[dwave],a,VRCVIIAmp);

	dwave+= a;
}

void UpdateOPLNEO(int16 *Wave,
	int Count)
{
	OPLL_fillAbsolute(VRC7Sound,Wave,Count,VRCVIIAmp);
}

void UpdateOPL(int32 Count)
{
	int32		z, a;
(void)Count;
	z = ((SOUNDTS<<16)/soundtsinc)>>4;

	if ((a=z-dwave)!=0 && VRC7Sound)
		OPLL_fill(VRC7Sound,&FCEUWave[dwave],a,VRCVIIAmp);
	dwave = 0;
}

static INLINE void DaMirror(uint8 V)
{
	static int8	salpo[4] = {MI_V,MI_H,MI_0,MI_1};

	setmirror(salpo[V&3]);
}

DECLFW(M85_writeABCD) /* A>=0xA000 && A<=0xDFFF */
{
	int		x;

	A|= (A&8)*2;
	// printf("$%04x, $%04x\n",XFrame.PC,A);
	A&= 0xF010;
	x = (A>>4&1)|(A-0xA000)>>11;
	mapbyte3[x] = V;
	setchr1(x<<10,V);
}

DECLFW(M85_write9)
{
	A|= (A&8)*2;
	if (A==0x9030)
	{
		if (FSettings.SndRate)
		{
			OPLL_writeReg(VRC7Sound,indox,V);
			GameExpSound.Fill[SCVRC7] = UpdateOPL;
			GameExpSound.NeoFill = UpdateOPLNEO;
		}
	} else if (!(A&0x10)) /*0x9000*/
		{mapbyte2[2] = V; setprg8(0xC000,V);}
	else /*0x9010*/
		indox = V;
}

DECLFW(M85_write8)
{
	/*A|= (A&8)*2;*/
	if (!(A&0x18)) /*0x8000*/
		{mapbyte2[0] = V; setprg8(0x8000,V);}
	else /*0x8010*/
		{mapbyte2[1] = V; setprg8(0xA000,V);}
}

DECLFW(M85_writeE)
{
	if (!(A&0x18)) /*0xE000*/
		{mapbyte2[3] = V; DaMirror(V);}
	else /*0xE010*/
		{IRQLatch = V; X6502_IRQEnd(FCEU_IQEXT);}
}

DECLFW(M85_writeF)
{
	if (!(A&0x18)) /*0xF000*/
	{
		VRCtemp = V&1;
		if ((IRQa=V&2)!=0)
			IRQCount = IRQLatch;
		acount = 0;
		X6502_IRQEnd(FCEU_IQEXT);
	} else /*0xF010*/
		{IRQa = !!VRCtemp; X6502_IRQEnd(FCEU_IQEXT);}
}

static void FP_FASTAPASS(1) KonamiIRQHook(int a)
{
	if (!IRQa)
		return;
	if ((acount+=a*3)>=LCYCS)
		do
		{
			acount-= LCYCS;
			if (++IRQCount&0x100)
			{
				X6502_IRQBegin(FCEU_IQEXT);
				IRQCount = IRQLatch;
			}
		} while(acount>=LCYCS);
}

void M85Restore(int version)
{
	int		x;

	if (version<7200)
	{
		x = 0;
		do
			mapbyte3[x] = CHRBankList[x];
		while(++x<8);
		x = 0;
		do
			mapbyte2[x] = PRGBankList[x];
		while(++x<3);
		mapbyte2[3] = Mirroring<0x10 ? Mirroring:Mirroring-0xE;
	}

	x = 0;
	do
		setchr1(x*0x400,mapbyte3[x]);
	while(++x<8);
	x = 0;
	do
		setprg8(0x8000+x*8192,mapbyte2[x]);
	while(++x<3);
	DaMirror(mapbyte2[3]);
	//LoadOPL();
}

static void M85SC(void)
{
	if (VRC7Sound)
		OPLL_set_rate(VRC7Sound,FSettings.SndRate);
}

static void M85SKill(void)
{
	if (VRC7Sound)
		OPLL_delete(VRC7Sound);
	VRC7Sound = NULL;
}

static void VRC7SI(void)
{
	GameExpSound.RChange[SCVRC7] = M85SC;
	GameExpSound.Kill = M85SKill;
	/* 3579552/72 = 49716 */
	VRC7Sound = OPLL_new(3579545,FSettings.SndRate ? FSettings.SndRate:44100);
	OPLL_reset(VRC7Sound);
	OPLL_reset(VRC7Sound);
}

void NSFVRC7_Init(void)
{
	SetWriteHandler(0x9010,0x901F,M85_write9);
	SetWriteHandler(0x9030,0x903F,M85_write9);
	VRC7SI();
}

void Mapper85_init(void)
{
	MapIRQHook = KonamiIRQHook;
	SetWriteHandler(0x8000,0x8FFF,M85_write8);
	SetWriteHandler(0x9000,0x9FFF,M85_write9);
	SetWriteHandler(0xA000,0xDFFF,M85_writeABCD);
	SetWriteHandler(0xE000,0xEFFF,M85_writeE);
	SetWriteHandler(0xF000,0xFFFF,M85_writeF);
	GameStateRestore = M85Restore;
	if (!VROM_size)
		SetupCartCHRMapping(0,CHRRAM,8192,1);
	//AddExState(VRC7Instrument, 16, 0, "VC7I");
	//AddExState(VRC7Chan, sizeof(VRC7Chan), 0, "V7CH");
	VRC7SI();
}
