/*
 emu2413.h
*/

#ifndef __EMU2413_H
 #include "../types.h"

#ifdef EMU2413_DLL_EXPORTS
  #define EMU2413_API __declspec(dllexport)
#elif defined(EMU2413_DLL_IMPORTS)
  #define EMU2413_API __declspec(dllimport)
#else
  #define EMU2413_API
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define PI 3.14159265358979323846

enum {OPLL_VRC7_TONE=0};

/* voice data */
typedef struct {
  uint32 TL,FB,EG,ML,AR,DR,SL,RR,KR,KL,AM,PM,WF;
} OPLL_PATCH;

/* slot */
typedef struct {

  OPLL_PATCH patch;  

  int32 type;          /* 0 : modulator 1 : carrier */

  /* OUTPUT */
  int32 feedback;
  int32 output[2];   /* Output value of slot */

  /* for Phase Generator (PG) */
  uint16 *sintbl;    /* Wavetable */
  uint32 phase;      /* Phase */
  uint32 dphase;     /* Phase increment amount */
  uint32 pgout;      /* output */

  /* for Envelope Generator (EG) */
  int32 fnum;          /* F-Number */
  int32 block;         /* Block */
  int32 volume;        /* Current volume */
  int32 sustine;       /* Sustine 1 = ON, 0 = OFF */
  uint32 tll;	      /* Total Level + Key scale level*/
  uint32 rks;        /* Key scale offset (Rks) */
  int32 eg_mode;       /* Current state */
  uint32 eg_phase;   /* Phase */
  uint32 eg_dphase;  /* Phase increment amount */
  uint32 egout;      /* output */

} OPLL_SLOT;

/* Mask */
#define OPLL_MASK_CH(x) (1u<<(x))

/* opll */
typedef struct {

  uint32 adr;
  int32 out;

#ifndef EMU2413_COMPACTION
  uint32 realstep;
  uint32 oplltime;
  uint32 opllstep;
  int32 prev, next;
#endif

  /* Register */
  uint8 LowFreq[6];
  uint8 HiFreq[6];
  uint8 InstVol[6];

  uint8 CustInst[8];

  int32 slot_on_flag[6 * 2];

  /* Pitch Modulator */
  uint32 pm_phase;
  int32 lfo_pm;

  /* Amp Modulator */
  int32 am_phase;
  int32 lfo_am;

  uint32 quality;

  /* Channel Data */
  int32 patch_number[6];
  int32 key_status[6];

  /* Slot */
  OPLL_SLOT slot[6 * 2];

  uint32 mask;

} OPLL;

/* Create Object */
EMU2413_API OPLL *OPLL_new(uint32 clk,
	uint32 rate);
EMU2413_API void OPLL_delete(OPLL *);

/* Setup */
EMU2413_API void OPLL_reset(OPLL *);
EMU2413_API void OPLL_set_rate(OPLL *opll,
	uint32 r);
EMU2413_API void OPLL_set_quality(OPLL *opll,
	uint32 q);

/* Port/Register access */
EMU2413_API void OPLL_writeIO(OPLL *,
	uint32 reg,
	uint32 val);
EMU2413_API void OPLL_writeReg(OPLL *,
	uint32 reg,
	uint32 val);

/* Synthsize */
EMU2413_API int16 OPLL_calc(OPLL *);

/* Misc */
EMU2413_API void OPLL_forceRefresh(OPLL *);

/* Channel Mask */
EMU2413_API uint32 OPLL_setMask(OPLL *,
	uint32 mask);
EMU2413_API uint32 OPLL_toggleMask(OPLL *,
	uint32 mask);

/* two variants (of one previously named moocow); 2008.07.21, 2013.05.14 */
EMU2413_API void OPLL_fillAbsolute(OPLL* opll,
	int16 *buf,
	int32 len,
	int32 amp);
EMU2413_API void OPLL_fill(OPLL* opll,
	int32 *buf,
	int32 len,
	int32 amp);

#ifdef __cplusplus
}
#endif

 #define __EMU2413_H
#endif
