/*
 mappers/simple.c

 * MacFCEU 0.5.5
 *
 * Copyright notice for this file:
 *  simple.c Copyright (C) 1998 BERO
 *  simple.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"

static uint8 latche;

//this stays here since there is no LATC
static DECLFW(M34_write)
{
	if (A>=0x8000 || (A-=0x7FFD)==0)
		ROM_BANK32(V);
	else //0x7FFx
		VROM_BANK4(A<<12&0x1000,V);
}

void Mapper34_init(void)
{
	ROM_BANK32(0);
	SetWriteHandler(0x7FFD,0xFFFF,M34_write);
}


/* I might want to add some code to the mapper 96 PPU hook function
   to not change CHR banks if the attribute table is being accessed,
   if I make emulation a little more accurate in the future.
*/

static uint8 M96LA;
static DECLFW(M96_write)
{
(void)A;
	latche=V;
	setprg32(0x8000,V&3);
	setchr4r(0x10,0x0000,(latche&4)|M96LA);
	setchr4r(0x10,0x1000,(latche&4)|3);
}

static void FP_FASTAPASS(1) M96Hook(uint32 A)
{
	if((A&0x3000)!=0x2000)
		return;
	//if((A&0x3ff)>=0x3c0) return;
	M96LA=A>>8&3;
	setchr4r(0x10,0x0000,(latche&4)|M96LA);
}

static void M96Sync(int v)
{
(void)v;
	setprg32(0x8000,latche&3);
	setchr4r(0x10,0x0000,(latche&4)|M96LA);
	setchr4r(0x10,0x1000,(latche&4)|3);
}

void Mapper96_init(void)
{
	SetWriteHandler(0x8000,0xffff,M96_write);
	PPU_hook=M96Hook;
	AddExState(&latche,1,0,"LATC");
	AddExState(&M96LA,1,0,"LAVA");
	SetupCartCHRMapping(0x10,MapperExRAM,32768,1);
	latche=M96LA=0;
	M96Sync(0);
	setmirror(MI_0);
	GameStateRestore=M96Sync;
}

static void M185Sync(int v)
{
(void)v;
	if (VROM_size==1)
		if ((mapbyte1[0]&3)==1)
			setchr8(0);
		else
			setchr8r(0x10,0);
	else if (VROM_size==2)
		if (mapbyte1[0]&2)
			setchr8(mapbyte1[0]&1);
		else
			setchr8r(0x10,0);
}

static DECLFW(M185_write)
{
(void)A;
	mapbyte1[0] = V;
	M185Sync(0);
}

void Mapper185_init(void)
{
	int	x;

	for(x=0; x<8192; ++x)
		MapperExRAM[x] = 0xFF;
	MapStateRestore = M185Sync;
	mapbyte1[0] = 0;
	M185Sync(0);

	SetupCartCHRMapping(0x10,MapperExRAM,8192,0);
	SetWriteHandler(0x8000,0xFFFF,M185_write);
}

static DECLFW(M156Write)
{
	if ((A-=0xC000)<=3)
		VROM_BANK1(A<<10,V); 
	else if ((A-=0xC008-0xC000)<=3)
		VROM_BANK1(4+A<<10,V);
	else if (A==0xC010-0xC008)
		ROM_BANK16(0x8000,V);
}

void Mapper156_init(void)
{
	onemir(0);
	SetWriteHandler(0xc000,0xc010,M156Write);
}

static DECLFW(m107w)
{
(void)A;
	ROM_BANK32(V>>1&3);
	VROM_BANK8(V&7);
}

void Mapper107_init(void)
{
	SetWriteHandler(0x8000,0xffff,m107w);
}
