/*
 memory.c

 * MacFCEU 0.8.5
 *
 * Copyright notice for this file:
 *  memory.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include "types.h"
#include "FCEU.h"
#include "memory.h"

/* added for file output consistency (zeroes)/speed; 2009.12.31 */
void dwmemset(void *p,
	long c, /* needs to be expanded (*0x01010101) */
	unsigned long z)
{
	int32		*d = p, x;

//	if (!d)
//		z = 0;
	if (z>7)
	{
		if ((x=z&~31)!=0)
			do
				*d++ = *d++ = *d++ = *d++ =
					*d++ = *d++ = *d++ = *d++ = c;
			while(x-=32);
		if (z&16)
			*d++ = *d++ = *d++ = *d++ = c;
		if (z&8)
			*d++ = *d++ = c;
	}
	if (z&4)
		*d++ = c;
	if (z&2)
		*((int16*)d)++ = c;
	if (z&=1)
		*(int8*)d = c;
}

/* now only called by mappers */
void *FCEU_gmalloc(unsigned long size)
{
	void		*ret;

	if ((ret=malloc(size))==0)  
	{
		FCEU_PrintError("Error allocating memory!  Doing a hard exit.");
		exit(1);
	}

	dwmemset(ret,0,size);

	return ret;
}

void *FCEU_malloc(unsigned long size)
{
	void		*ret;

	if ((ret=malloc(size))==0)
	{
		FCEU_PrintError("Unable to allocate %u bytes.",
			size);
		return 0;
	}
	dwmemset(ret,0,size);

	return ret;
}

void FCEU_free(void *ptr)	// Might do something with this and FCEU_malloc later...
{
	free(ptr);
}

void FASTAPASS(3) FCEU_memcpy(void *_d,
	const void *_s,
	unsigned long z)
{
	const int32	*s = _s;
	int32		*d, x;

	if ((d=_d)==0)
		z = 0;
	/* alignment doesn't matter (or matter much) with 68020+ */
#ifndef mac
	if ((((int)d^(int)s)&3)==0) //4-byte aligned
#endif
	{
		if (z>7)
		{
			if ((x=z&~31)!=0)
				do
					*d++ = *s++, *d++ = *s++,
						*d++ = *s++, *d++ = *s++,
						*d++ = *s++, *d++ = *s++,
						*d++ = *s++, *d++ = *s++;
				while(x-=32);
			if (z&16)
				*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
			if (z&8)
				*d++ = *s++, *d++ = *s++;
		}
		if (z&4)
			*d++ = *s++;
		if (z&2)
			*((int16*)d)++ = *((const int16*)s)++;
		if (z&=1)
			*((int8*)d)++ = *((const int8*)s)++;
		return;
	}
#ifndef mac
	if ((x=z&~7)!=0)
		do
			*((int8*)d)++ = *((const int8*)s)++,
				*((int8*)d)++ = *((const int8*)s)++,
				*((int8*)d)++ = *((const int8*)s)++,
				*((int8*)d)++ = *((const int8*)s)++,
				*((int8*)d)++ = *((const int8*)s)++,
				*((int8*)d)++ = *((const int8*)s)++,
				*((int8*)d)++ = *((const int8*)s)++,
				*((int8*)d)++ = *((const int8*)s)++;
		while(--x);
#endif
	if (z&=7)
		do
			*((int8*)d)++ = *((const int8*)s)++;
		while(--z);
}

void _memrem4(void *_p,
	uint32 c,
	const uint8 *t)
{
	uint8		*p = _p;

	c&=~3, p+= c;
	do
		*p = t[*--p], *p = t[*--p], *p = t[*--p], *p = t[*--p];
	while(c-=4);
}
