/*
 ops6502.h
  operational code switch cases for 6502 template

* MacFCEU 0.8a7
*
* Copyright notice for this file:
*  ops.h Copyright (C) 2002 Xodnizel
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

case 0x00: BRK; break;
case 0x40: RTI; break;

case 0x60: /* RTS */
	_PC = PUL(), _PC|= PUL()<<8, ++_PC, _PC&= 0xFFFF;
	break;

case 0x48: PSH(_A); break; /* PHA */
case 0x08: PSH(_P|M_FLAG|B_FLAG); break; /* PHP */
case 0x68: X_ZN(RA=PUL()); break; /* PLA */
case 0x28: _P = PUL(); break; /* PLP */
case 0x4C: /* JMP absolute */
	t = RMem(_PC), ++_PC, _PC&= 0xFFFF, _PC = t|= RMem(_PC)<<8;
	break;
case 0x6C: JIN; break; /* JMP indirect */
case 0x20: /* JSR */
	t = RMem(_PC), ++_PC,_PC&= 0xFFFF;
	PSH(_PC>>8); PSH(_PC);
	_PC = t|= RMem(_PC)<<8;
	break;

case 0xAA: X_ZN(RX=RA); break; /* TAX */
case 0x8A: X_ZN(RA=RX); break; /* TXA */

case 0xA8: X_ZN(_Y=RA); break; /* TAY */
case 0x98: X_ZN(RA=_Y); break; /* TYA */

case 0xBA: X_ZN(RX=_S); break; /* TSX */
case 0x9A: _S = _X; break; /* TXS */

case 0x88: X_ZN(--_Y); break; /* DEY */
case 0xC8: X_ZN(++_Y); break; /* INY */
case 0xCA: X_ZN(--_X); break; /* DEX */
case 0xE8: X_ZN(++_X); break; /* INX */

case 0x18: _P&=~C_FLAG; break; /* CLC */
case 0x38: _P|= C_FLAG; break; /* SEC */
case 0x58: _P&=~I_FLAG; break; /* CLI */
case 0x78: _P|= I_FLAG; break; /* SEI */

case 0xB8: _P&=~V_FLAG; break; /* CLV */
case 0xD8: _P&=~D_FLAG; break; /* CLD */
case 0xF8: _P|= D_FLAG; break; /* SED */

case 0xEA: break; /* the official NOP... INC to nothing */

case 0xC6: RMW_ZP(DEC);
case 0xD6: RMW_ZPX(DEC);
case 0xCE: RMW_AB(DEC);
case 0xDE: RMW_ABX(DEC);

case 0xE6: RMW_ZP(INC);
case 0xF6: RMW_ZPX(INC);
case 0xEE: RMW_AB(INC);
case 0xFE: RMW_ABX(INC);

case 0x0A: RMW_A(ASL);
case 0x06: RMW_ZP(ASL);
case 0x16: RMW_ZPX(ASL);
case 0x0E: RMW_AB(ASL);
case 0x1E: RMW_ABX(ASL);

case 0x4A: RMW_A(LSR);
case 0x46: RMW_ZP(LSR);
case 0x56: RMW_ZPX(LSR);
case 0x4E: RMW_AB(LSR);
case 0x5E: RMW_ABX(LSR);

case 0x2A: RMW_A(ROL);
case 0x26: RMW_ZP(ROL);
case 0x36: RMW_ZPX(ROL);
case 0x2E: RMW_AB(ROL);
case 0x3E: RMW_ABX(ROL);

case 0x6A: RMW_A(ROR);
case 0x66: RMW_ZP(ROR);
case 0x76: RMW_ZPX(ROR);
case 0x6E: RMW_AB(ROR);
case 0x7E: RMW_ABX(ROR);

case 0x69: LD_IM(ADC);
case 0x65: LD_ZP(ADC);
case 0x75: LD_ZPX(ADC);
case 0x6D: LD_AB(ADC);
case 0x7D: LD_ABX(ADC);
case 0x79: LD_ABY(ADC);
case 0x61: LD_IX(ADC);
case 0x71: LD_IY(ADC);

case 0x29: LD_IM(ANA);
case 0x25: LD_ZP(ANA);
case 0x35: LD_ZPX(ANA);
case 0x2D: LD_AB(ANA);
case 0x3D: LD_ABX(ANA);
case 0x39: LD_ABY(ANA);
case 0x21: LD_IX(ANA);
case 0x31: LD_IY(ANA);

case 0x24: LD_ZP(BIT);
case 0x2C: LD_AB(BIT);

case 0xC9: LD_IM(CPA);
case 0xC5: LD_ZP(CPA);
case 0xD5: LD_ZPX(CPA);
case 0xCD: LD_AB(CPA);
case 0xDD: LD_ABX(CPA);
case 0xD9: LD_ABY(CPA);
case 0xC1: LD_IX(CPA);
case 0xD1: LD_IY(CPA);

case 0xE0: LD_IM(CPX);
case 0xE4: LD_ZP(CPX);
case 0xEC: LD_AB(CPX);

case 0xC0: LD_IM(CPY);
case 0xC4: LD_ZP(CPY);
case 0xCC: LD_AB(CPY);

case 0x49: LD_IM(EOA);
case 0x45: LD_ZP(EOA);
case 0x55: LD_ZPX(EOA);
case 0x4D: LD_AB(EOA);
case 0x5D: LD_ABX(EOA);
case 0x59: LD_ABY(EOA);
case 0x41: LD_IX(EOA);
case 0x51: LD_IY(EOA);

case 0xA9: LD_IM(LDA);
case 0xA5: LD_ZP(LDA);
case 0xB5: LD_ZPX(LDA);
case 0xAD: LD_AB(LDA);
case 0xBD: LD_ABX(LDA);
case 0xB9: LD_ABY(LDA);
case 0xA1: LD_IX(LDA);
case 0xB1: LD_IY(LDA);

case 0xA2: LD_IM(LDX);
case 0xA6: LD_ZP(LDX);
case 0xB6: LD_ZPY(LDX);
case 0xAE: LD_AB(LDX);
case 0xBE: LD_ABY(LDX);

case 0xA0: LD_IM(LDY);
case 0xA4: LD_ZP(LDY);
case 0xB4: LD_ZPX(LDY);
case 0xAC: LD_AB(LDY);
case 0xBC: LD_ABX(LDY);

case 0x09: LD_IM(ORA);
case 0x05: LD_ZP(ORA);
case 0x15: LD_ZPX(ORA);
case 0x0D: LD_AB(ORA);
case 0x1D: LD_ABX(ORA);
case 0x19: LD_ABY(ORA);
case 0x01: LD_IX(ORA);
case 0x11: LD_IY(ORA);

case 0xEB:
case 0xE9: LD_IM(SBC);
case 0xE5: LD_ZP(SBC);
case 0xF5: LD_ZPX(SBC);
case 0xED: LD_AB(SBC);
case 0xFD: LD_ABX(SBC);
case 0xF9: LD_ABY(SBC);
case 0xE1: LD_IX(SBC);
case 0xF1: LD_IY(SBC);

case 0x85: ST_ZP(RA);
case 0x95: ST_ZPX(RA);
case 0x8D: ST_AB(RA);
case 0x9D: ST_ABX(RA);
case 0x99: ST_ABY(RA);
case 0x81: ST_IX(RA);
case 0x91: ST_IY(RA);

case 0x86: ST_ZP(RX);
case 0x96: ST_ZPY(RX);
case 0x8E: ST_AB(RX);

case 0x84: ST_ZP(_Y);
case 0x94: ST_ZPX(_Y);
case 0x8C: ST_AB(_Y);

#ifdef JR_DESTIN
	JR_DESTIN; /* for instantly removing redundant code */
#endif

case 0x10: JR(!(_P&N_FLAG)); /* BPL */
case 0x30: JR(_P&N_FLAG); /* BMI */
case 0x50: JR(!(_P&V_FLAG)); /* BVC */
case 0x70: JR(_P&V_FLAG); /* BVS */
case 0x90: JR(!(_P&C_FLAG)); /* BCC */
case 0xB0: JR(_P&C_FLAG); /* BCS */
case 0xD0: JR(!(_P&Z_FLAG)); /* BNE */
case 0xF0: JR(_P&Z_FLAG); /* BEQ */

	/* Here comes the undocumented instructions block. */
/* AAC / ANC */
case 0x2B:
case 0x0B: LD_IM(ANC);

case 0x87: ST_ZP(SAX);
case 0x97: ST_ZPY(SAX);
case 0x8F: ST_AB(SAX);
case 0x83: ST_IX(SAX);

case 0x6B: LD_IM(ARR); /* ARR - ARGH, MATEY! */
case 0x4B: LD_IM(ASR);

/* AXS / SBX */ 
case 0xCB: LD_IM(AXS);

/* DCP */
case 0xC7: RMW_ZP(DCP);
case 0xD7: RMW_ZPX(DCP);
case 0xCF: RMW_AB(DCP);
case 0xDF: RMW_ABX(DCP);
case 0xDB: RMW_ABY(DCP);
case 0xC3: RMW_IX(DCP);
case 0xD3: RMW_IY(DCP);

/* ISB */
case 0xE7: RMW_ZP(ISB);
case 0xF7: RMW_ZPX(ISB);
case 0xEF: RMW_AB(ISB);
case 0xFF: RMW_ABX(ISB);
case 0xFB: RMW_ABY(ISB);
case 0xE3: RMW_IX(ISB);
case 0xF3: RMW_IY(ISB);

/* DOP */
/* these used to just `_PC++` (change affects emulation) */
case 0x04: /* undefined 0H-operator line (similar to BIT), to ZP */
case 0x44: /* ... 2H-op. line (similar to BIT), to ZP */
case 0x64: LD_ZP(;); /* ... 3H-op. line (similar to BIT), to ZP */

case 0x14: /* undefined 0C-op. line, to ZP,X */
case 0x34: /* ... 1C-op. line, to ZP,X */
case 0x54: /* ... 2C-op. line, to ZP,X */
case 0x74: /* ... 3C-op. line, to ZP,X */
case 0xD4: /* ... a compare line to ZX (not to Y) */
case 0xF4: LD_ZPX(;); /* ... a compare line to ZX (not to X) */

case 0x80: /* appears to be STY to immediate (can jam) */

case 0x82: /* ... STX to imm... */
case 0xC2: /* ... DEC to imm... */
case 0xE2: /* ... INC to imm... */

case 0x89: ++_PC, _PC&= 0xFFFF; break; /* ... STA to immediate (can jam) */

/* KIL */

case 0x02: /* appears to be ASL to stack push (bad addressing mode) */
case 0x12: /* ... ASL to relative (bad use of addressing mode) */
case 0x22: /* ... ROL to stack pop (bad use of addressing mode) */
case 0x32: /* ... ROL to relative... */
case 0x42: /* ... LSR to stack pop.. */
case 0x52: /* ... LSR to relative.. */
case 0x62: /* ... ROR to stack pop.. */
case 0x72: /* ... ROR to relative.. */
case 0x92: /* ... STX to relative.. */
case 0xB2: /* ... LDX from relative (unimplemented use of addressing mode) */
case 0xD2: /* ... DEC to relative.. */
case 0xF2: /* ... INC to relative.. */
	--_PC, _PC&= 0xFFFF;
	XFrame.PC = _PC;
	if ((t=FCEUI_DebuggerBreak(t))>=0)
	{
		x = CycTable[t], x-= CycTable[subsopcode_breakpoint];
	//	_ADDCYC(x);
		_PC = RPC;
		goto exec; /* may be passing a breakpoint */
	}
	_ADDCYC(0xFF);
	_intlow|= FCEU_JAMMED;
	break;

case 0xBB: LD_ABY(LAS); //mode was RMW_ABY; 2010.05.27

case 0xAB: LD_IM(LAX); /* LXA */
case 0xA7: LD_ZP(LAX);
case 0xB7: LD_ZPY(LAX);
case 0xAF: LD_AB(LAX);
case 0xBF: LD_ABY(LAX);
case 0xA3: LD_IX(LAX);
case 0xB3: LD_IY(LAX);

/* single byte, no operation */
case 0x1A: /* would be be ASL to no addressing mode */
case 0x3A: /* ... ROL to no addressing mode */
case 0x5A: /* ... LSR to no addressing mode */
case 0x7A: /* ... ROR to no addressing mode */
case 0xDA: /* ... DEC to no addressing mode */
case 0xFA: break; /* ... INC to no addressing mode */

case 0x27: RMW_ZP(RLA);
case 0x37: RMW_ZPX(RLA);
case 0x2F: RMW_AB(RLA);
case 0x3F: RMW_ABX(RLA);
case 0x3B: RMW_ABY(RLA);
case 0x23: RMW_IX(RLA);
case 0x33: RMW_IY(RLA);

case 0x67: RMW_ZP(RRA);
case 0x77: RMW_ZPX(RRA);
case 0x6F: RMW_AB(RRA);
case 0x7F: RMW_ABX(RRA);
case 0x7B: RMW_ABY(RRA);
case 0x63: RMW_IX(RRA);
case 0x73: RMW_IY(RRA);

case 0x07: RMW_ZP(SLO);
case 0x17: RMW_ZPX(SLO);
case 0x0F: RMW_AB(SLO);
case 0x1F: RMW_ABX(SLO);
case 0x1B: RMW_ABY(SLO);
case 0x03: RMW_IX(SLO);
case 0x13: RMW_IY(SLO);

case 0x47: RMW_ZP(SRE);
case 0x57: RMW_ZPX(SRE);
case 0x4F: RMW_AB(SRE);
case 0x5F: RMW_ABX(SRE);
case 0x5B: RMW_ABY(SRE);
case 0x43: RMW_IX(SRE);
case 0x53: RMW_IY(SRE);

case 0x93: ST_IY(SHA);
case 0x9F: ST_ABY(SHA);

case 0x9C: ST_ABX(SHY);
case 0x9E: ST_ABY(SHX);
case 0x9B: ST_ABY(SHS);

/* TOP - read address, do nothing */
case 0x0C: GetAB(t); /* undefined 0H-operator line (similar to JMP) */
	break; /* ...groan... `break` was missing; 2010.04.23 */
case 0x1C: /* undefined 0C-op. line (similar to BIT) */
case 0x3C: /* undefined 1C-op. line (similar to BIT) */
case 0x5C: /* undefined 2C-op. line (similar to STY) */
case 0x7C: /* undefined 3C-op. line (similar to STY) */
case 0xDC: /* undefined 6C-op. line (similar to LDY) */
case 0xFC: LD_ABX(;); /* undefined 7C-op. line (similar to LDY) */

case 0x8B: LD_IM(ANE);
