#if 0
C {0}.o interpreter.o ::obj:disasm6502.c.o ::obj:ftrig.a.o -o asm6502 -d SAMPLE -d CHR_HX=36 -d PS=58 -d ALLOW_BIN_HI_LO -d NO_EBREAKS -d ALLOW_DLITERALS -d ALLOW_NAME_DOT {"Parameters"}
Exit 0
#endif
/*
 asm6502.c
  6502/6510 high and low-level assembler.  Most formats supported.
   See asm6502.txt for details.

  Revised: 2013.07.23

 * Copyright notice for this file:
 * asm6502.c Copyright (C) 2010, 2013 adamjasonp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 Preprocessing considerations:
	.macro <name> [<param1>, [<param2> [...]]]
	 .shift
	.endmacro | .endm

 DrPCjr appears to have a 512 byte header.
*/

#include "../enviro64.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "interpreter.h"
#include "ftrig.h"

#include "asm6502.h"

/* ---- macros */

#define SKIPCHR(s,e,c)		do{while((s)<e&&*(s)==(c)) ++(s);}while(0)

/* ---- code */

static const char *IDBreak(const char *str,
	const char *e,
	int a)
{
	const char	*s = str;

	if (s>=e)
		return s;
	if (*s=='@') /* one starting 'at' only */
		if (++s>=e)
			return str;
	do
		switch(*s)
		{
		case '\047': /* apostophe */
			if (!a || s==str || !IS_ALA(s[-1]))
				return s;
		case '.':
#ifndef NO_MULTI_DOT_IDS
			if (s>str && s[-1]=='.')
				return s;
#endif
		case '_':
		CASE_NUMERIC
		CASE_ALPHABETIC
		next:	++s;
			break;

		default:
#ifdef CHR_NB
			if (*s==CHR_NB)
				goto next;
#endif
			return s;
		}
	while(s<e);

	return s;
}

static const char *getLineEnd(const char *s,
	const char *e)
{
	const char	*n;
	uint8		c, p;

	for(n=s,p=0; n<e && (c=*n)!=';' && !IS_EOL(c); ++n,p=c)
		if (c=='"' || (c=='\047' && !IS_ALA(p)))
			while(++n<e)
				if (*n=='\\')
					++n; /* skip escape */
				else if (*n==c)
					break;
	return MIN(n,e);
}

static const char *getBoundsEnd(const char *n, /* first skipped */
	const char *e)
{
	long		x;
	uint8		c, p;

	for(x=1,p=0; n<e && (c=*n)!=','; ++n,p=c)
		if (c=='"' || (c=='\047' && !IS_ALA(p)))
		{ /* skip quote; 2013.06.25 */
			while(++n<e)
				if (*n=='\\')
					++n; /* skip escape */
				else if (*n==c)
					break;
		} else if (c=='(')
			++x;
		else if ((c==')' || c==']') && !--x)
			break;
	return MIN(n,e);
}

static int appendParent(char **b,
	const char *pfn)
{
	const char	*n;
	char		*p;
	long		x, z;

	if (!b || (z=strlen(p=*b))<0)
		return _e_general_error;
#if PS=='\\' /* Windows */
	if (*pfn==':' || (pfn[1] && pfn[1]==':'))
#elif PS==':' /* Mac OS */
	if (*pfn!=PS)
#else
	if (*pfn==PS)
#endif
		goto end; /* explicit path */
	if ((n=strrchr(pfn,PS))==NULL || (x=n-pfn)<=0)
		goto end;
	if (!z++ || p[z-2]!=PS)
		++x;
	/* insert before */
	if ((p=realloc(p,x+z))==NULL)
		return _e_memory_error;
	*b = p;
	memmove(p+x,p,z);
	memmove(p,pfn,x);
end:	return 1;
}

long X6502_ParseName(const char *str,
	const char *e,
	long *z,
	char **b,
	const char *parent) /* added 2013.05.26 */
{
	const char	*s;
	char		*p;
	long		r;

	if (!str || str>=e)
		return _e_general_error;
	if (*str=='\047' || *str=='"')
	{
		r = IMAX(r);
		s = parse_textloop(str+1,e,*str,&r,NULL,b,0,NULL);
		if (e-str<2 || s[-1]!=*str) /* point of error */
			return s>=e ? _e_open_string:_e_general_error;
		e = NULL; /* buffered - not directly from str */
	} else /* can expect an extension */
		r = (s=IDBreak(str,e,0))-str;
	if (r<0)
		return 0;
	if ((p=realloc(*b,r+1))==NULL)
	{
		if (*b)
			{free(*b); *b = NULL;}
		return _e_memory_error;
	}
	(*b=p)[r] = 0; /* C-string termination */
	if (z)
		*z = r;
	if (e)
		memcpy(*b,str,r);
	if (strlen(*b)!=r)
		return _e_name_null;
	if (parent && (r=appendParent(b,parent))<0)
		return r;
	return s-str;
}

static int labelScan(asm6502frame *f,
	evalvalue *v,
	int c,
	char y)
{
	const uint8	*s, *l;
	int32		t =-1, o = f->ctr;

	if (!f)
		return _e_general_error;
	if ((s=f->lab)==0 || !*(l=s))
		return 0;
	if (y=='-')
		y = '+';
	else if (y=='+')
		y = '-';
	else
		y =-1;
	/* scan forward first (reverse after getting to the closest ctr) */
	for(; *l++; l+=1+_LabDataSize)
		if ((*(l+=*l+2)&0xBF)==_Label) /* must be an established label */
			if ((t=l[5]|((l[6]|((l[7]|(l[8]<<8))<<8))<<8))>=o && y!=l[-2])
				if (c<0)
				{
					if (!(c+=o>=t))
						goto val;
					goto rev;
				} else if ((c-=t>o)<=0)
					goto val;
	if (c<0)
	{
		l-= 1+1+_LabDataSize;
		do
		{
			if ((*l&0xBF)==_Label && y!=l[-2] && !++c)
			{
			val:	t = l[1]|((l[2]|((l[3]|(l[4]<<8))<<8))<<8);
				if (v)
					*v = t;
				return 1;
			}
		rev:	;
		} while((l-=l[-1]+2+_LabDataSize+2)>s);
	}
	return 0;
}

static const char *parseBinary(const char *s,
	const char *e,
	evalvalue *i)
{
	uint32		v = 0;
	uint8		c;

	do
		if ((c=*s-'0')>1)
		{ /* can have no other following alphanumerics */
			if (c<=9 || (c+='0'-'A')<='Z'-'A'
					|| (c+='A'-'a')<='z'-'a')
				return NULL;
			break;
		} else
			v+= v, v+= c;
	while(++s<e);
	if (i)
		*i = v;
	return s;
}

static const char *parseValue(const char *s,
	const char *e,
	evalvalue *i,
	void *u)
{
	asm6502frame	*f;
	uint8		b[256], *d;
	const char	*n;
	int32		x;
	int		z, h;

	if (s>=e)
		return NULL;
	if (*s=='%')
	{ /* binary */
		if (++s>=e)
			return NULL;
		return parseBinary(s,e,i);
	}
	if ((f=u)==NULL)
	{ /* `.if xxx || yyy` default zero */
		if (i)
			*i = 0;
		return IDBreak(s,e,0);
	}
	if (*s=='*')
	{ /* current location */
		if (i)
			*i = f->org;
		return s+1;
	}
	n = IDBreak(s,e,1), h =-1;
	if (toupper((uint8)*s)=='A' && n==s+1)
		return NULL; /* register-reserved */
	if (s==n)
	{
		if (*s=='-')
			{SKIPCHR(n,e,'-'); x = s-n;}
		else if (*s=='+')
			{SKIPCHR(n,e,'+'); x = n-s, f->flg|= _PreEstID;}
		else
			return NULL;
		if (labelScan(f,i,x,*s)<=0)
			if (!(f->opt&_Prescan))
				{f->flg|= _LScanErr; return NULL;}
			else if (i)
				*i = f->org+3;
		return n;
	} else
	{
		e = n;
		if (n-s>4 && n[-4]=='.'
				&& toupper((uint8)n[-1])=='B' /* BS */
				&& toupper((uint8)n[-2])=='S')
		{ /* (somewhat inefficient) handling of .LSB/.MSB */
			uint8		c = toupper((uint8)n[-3]);

			if (c=='L')
				h = 0, e-= 4;
			else if (c=='M')
				h = 8, e-= 4;
		}
	}
	if ((x=X6502_AsmGetID(s,e,b,u,&z))>0)
	{
		d = ((uint8*)f->lab)+x;
		if (!(f->opt&_Prescan) && (*d&0x3F)==_Unestablished)
			return NULL; /* should be defined by now */
	} else if (X6502_ParseKey(s,e)>=0) /* can't be a keyword; 2013.05.21 */
		return NULL;
	else if (f->opt&_Prescan
#if 1
			)
	{
		f->flg|= _PreEstID;
		if (i)
			*i = f->org+3;
		return n;
	} else
		return NULL;
#else
			|| (x=X6502_AsmAddID(s,e,b,u,_Unestablished|0x80))<=0)
	else
	{
		d = ((uint8*)f->lab)+x;
		d[2] = f->org>>8; /* page copy */
	}
#endif
	if (*d&0x80)
		f->flg|= _PreEstID;
	if (i)
	{
		x = d[1]+(d[2]<<8);
		if (h>=0)
			*i = x>>h&0xFF;
		else
			*i = x+((d[3]+(d[4]<<8))<<16);
	}
	return n;
}

/* returns the total parsed length, and larger operand mode in `m` */
static long parseOperands(const char *str,
	const char *e,
	evalvalue *v,
	asm6502frame *f,
	int *m)
{
	const char	*s = str, *n;
	long		r;
	int		t;
	char		b = 0, c;

	if (s>=e)
		{*m = OPM_IP; return 0;}
	n = e, c = *s;
	if (c=='#')
		*m = OPM_IM, ++s; /* immediate */
	else if (c=='(') /* indirect - must scan for each evaluating bounds */
		{b = ')'; goto ib;}
	else if (c=='[')
	{
		b = ']';
	ib:	*m = OPM_IB, n = getBoundsEnd(++s,e);
	} else
		*m = OPM_AB; /* absolute (or ZP if low and opt on) */
	if ((r=evaluate_line(s,n,v,&t,parseValue,NULL,f,NULL))==0 || t)
	{ /* for pre-labels - you may get away with `[-]` */
		if (*s==':')
			++s; /* for ca65 `:-`, `:+`, etc. */
		if ((n=parseValue(s,n,v,f))==NULL)
		{
			if (parse_skipwsp(str+1,e)>=e && toupper((uint8)*str)=='A')
				{*m = OPM_IA; return e-s;} /* A register */
			return t<0 ? t==_e_incomplete_error && f->flg&_LScanErr
					? _e_labelscan_overflow:t
					:_e_general_error;
		} else
			r = n-s;
	}
	s = parse_skipwsp(s+r,e);
	r = e-str;
	if (s>=e)
		return b ? _e_general_error:r;
	if (*s==')' || *s==']')
	{ /* bracket/parenthesis must match, IB or IY, resulting */
		if (!b || *s!=b || *m!=OPM_IB)
			return _e_general_error;
		t = 1, s = parse_skipwsp(s+1,e);
	}
	if (*s==',')
	{
		s = parse_skipwsp(s+1,e);
		if (s>=e) /* only a comma? */
			return _e_general_error;
		if ((c=toupper((uint8)*s))=='X' && !t)
			if (*m==OPM_AB)
				*m = OPM_AX;
			else if (*m==OPM_IB)
				*m = OPM_IX;
			else
				return _e_general_error;
		else if (c=='Y')
		{ /* must be closed first */
			if (*m==OPM_AB)
				*m = OPM_AY;
			else if (*m==OPM_IB && t)
				*m = OPM_IY;
			else
				return _e_general_error;
		} else
			return _e_general_error;
		s = parse_skipwsp(s+1,e);
		/* need indirect closing?  IY doesn't. */
		if (s>=e)
			return *m==OPM_IX ? _e_general_error:r;
		if (*s!=b) /* IX does */
			return _e_general_error;
		s = parse_skipwsp(s+1,e);
	} else if (b && !t) /* need indirect closing */
		return _e_general_error;
	if (s<e)
		return _e_general_error;
	return r;
}

/* ---- keywords */

static const char	*keyw_list[] =
{
	"INCBIN",
	"INCLUDE",
	"ADDR",
	"ALIGN",
	"BANK",
	"BLOCK",
	"BYT",
	"BYTE",
	"DB",
	"DCB",
	"DC.B",
	"DEFCHR",
	"DW",
	"DCW",
	"DC.W",
	"DCL",
	"DC.L",
	"END",
	"ENDPROC",
	"EQU",
	"FILL",
	"INESCHR",
	"INESMAP",
	"INESMIR",
	"INESPRG",
	"MEM",
	"MODULE",
	"OPT",
	"ORG",
	"ORGA",
	"P02",
	"P10",
	"PAD",
	"PROCESSOR",
	"RES",
	"SEGMENT",
	"SUBROUTINE",
	"WORD",
	"TEXT",
	"CODE",
	"DATA",
	"BSS",
	"ZEROPAGE",
	"ZP",
	"DEFAULT",
	"ALIAS",
	"PROC",
	"SPACE",
	"ACCU",
	"ADVANCE",
	"ASCII",
	"CHECKPC",
	"INDEX",
	"PUSHSEG",
	"POPSEG",
	"PRINTT",
	"PRINTV",
	"EXPORT",
	"IMPORT"
};

#define NUMBER_OF_KEYS		(OPN_6510_COUNT+GET_ARRAY_COUNT(keyw_list))

int X6502_ParseKey(const char *str,
	const char *e)
	/*int strict)*/
{
	char		n[10];
	const char	*s = str;
	int		i, z = e-s;
	uint8		m = 0;

	if (!s || s>=e)
		return _e_general_error;
	if (*s=='.')
		++s, m = 1;
	if ((z=e-s)>10)
		goto nope;
	i = 0;
	do
		n[i] = toupper((uint8)*s++);
	while(++i,s<e);

	if (z==3 && !m)
	{
		const char	*t = *x6510_opnametable;

		i = 0;
		do
			if (*t==*n && t[1]==n[1] && t[2]==n[2])
				return i;
		while(t+=3,++i<OPN_6510_COUNT);
	} else
		i = OPN_6510_COUNT;
#ifdef KEYW_REQUIRES_DOT
	if (m)
#endif
	{ /* keyword (e.g., .equ) */
		const char	**t = keyw_list, *s, c = *n;

		--z;
		do
			if (*(s=*t)==c && !memcmp(++s,n+1,z) && !s[z])
			{
				if (/*strict &&*/ i>=_text && *str!='.')
					break; /* dot required for these */
				return i;
			}
		while(++t,++i<NUMBER_OF_KEYS);
	}
nope:	return _e_no_identifier;
}

/* ---- low-level ID access and creation */

/* returns ID type/data offset, ~end or zero on error */
int32 X6502_AsmGetID(const char *s,
	const char *e,
	uint8 *n,
	const asm6502frame *f,
	int *_z)
{
	const uint8	*l;
	uint8		h, z;

	if (!f || !n || !s || e<=s)
		return 0;
	*n = *s++, z = h = 0;
	if (f->mod)
		if (*n=='_' || *n=='@')
		{
			long	x;

			if ((x=strlen(f->mod))>255)
				return 0; /* no room for anything else? */
			n[x] = *n;
			memcpy(n,f->mod,x);
			z = x;
		}
	/* fill capitalized in 256-char buffer */
	*n = toupper((uint8)*n)&0x3F;
	while(s<e)
		if (!++z)
			return 0; /* too long */
		else
			h+= n[z] = toupper((uint8)*s++);
	if ((h=h<<6|*n)==0) /* '@' is valid */
		h = 0x80; /* hash can't be null */
	*n++ = h;
	if (_z)
		*_z = 1+z;
	if ((l=f->lab)==0)
		return ~0;
	/* hash and type match */
	for(; *l++; l+=*l,l+=1+2+_LabDataSize)
		if (*l==z && h==l[-1] && !memcmp(l+1,n,z))
			return 2+z+l-f->lab;
	return f->lab-l; /* becomes ~total */
}

/* returns ID offset, or <=0 on error */
int X6502_AsmAddID(const char *s,
	const char *e,
	uint8 *n,
	asm6502frame *f,
	uint8 t)
{
	uint8		*d;
	int32		x;
	int		z;

	if ((x=X6502_AsmGetID(s,e,n,f,&z))>=0)
	{
		if (!x)
			return 0;
		d = f->lab+x;
		if (t!=_Unestablished && !(*d&0x40) && (*d&0x3F)!=_Unestablished)
			return -x;
		if (f->opt&_Prescan)
			t|= *d&0x80; /* preserve PreEst */
		goto typ; /* can only establish once */
	}
	/* (-x) end offset including null term., 1+1 length, 1 typ, 4 val, 4 ctr */
	if ((d=realloc(f->lab,z+1+2+_LabDataSize-x))==0)
		return _e_memory_error;
	f->lab = d, d-= x;
	d[-1] = *n++; /* hash (includes first char) */
	*d++ = x = z-1; /* post-hash length */
	memcpy(d,n,x); /* post-hash name */
	d+= x, *d++ = x; /* duplicate-length for reverse scanning */
	d[1] = d[2] = d[3] = d[4] = d[9] = 0; /* init and terminate */
	x = f->ctr;
	d[5] = x;
	d[6] = x>>8;
	d[7] = x>>16;
	d[8] = x>>24;
typ:	if (f->opt&_Prescan)
		t|= 0x40;
	*d = t;
	return d-f->lab;
}

int X6502_AsmDumpIDs(FILE *f,
	const asm6502frame *af)
{
	const uint8	*l;
	long		t, v, o;
	uint8		c;

	if (!af)
		return _e_general_error;
	if ((l=af->lab)==NULL)
		return 0;
	for(t=0; (c=*l++)!=0; l+=1+2+_LabDataSize,++t)
	{
		const char	*s = (const char*)l+1;
		int		z = *l;

		l+= z;
		if (f)
		{
			const char	*lut, y = l[2];

			c&= 0x3F, c|= (~c*2)&0x40; /* unmangle */
			lut = y&0x80 ? "?dle":"?DLE";
			v = l[3]|((l[4]|((l[5]|(l[6]<<8))<<8))<<8);
			o = l[7]|((l[8]|((l[9]|(l[10]<<8))<<8))<<8);
			fprintf(f,"%c %c%.*s%* @%u = 0x%X\n",
				lut[y>_Equate ? 0:y],c,z,s,9-MIN(z,9),o,v);
		}
	}
	return t;
}

/* ---- general stack */

static int stPush(uint8 **st,
	uint8 v)
{
	uint8		*d;
	int32		x =-1;

	if ((d=*st)==NULL || (x=*(int32*)d)<0 || (x&7)==6)
	{
		if ((d=realloc(d,((x+9)&~7)+sizeof(int32)))==NULL)
			return 0;
		*st = d;
	}
	*(int32*)d = ++x;
	d[x+4] = v;

	return 1;
}

static int stPop(uint8 **st,
	uint8 *v)
{
	uint8		*d;
	int32		x;

	if ((d=*st)==NULL || (x=*(int32*)d)<0)
		return 0;
	if (x)
		{*v = d[x+4], --*(int32*)d; return 2;}
	else
		{*v = 0, *st = NULL; free(d); return 1;}
}

/* ---- segments */

static int strMatch(const char *s,
	const char *n,
	int32 z)
{
	return !s[z] && !memcmp(n,s,z);
}

static uint8 *getSeg(asm6502frame *f,
	int x)
{
	uint8		*s;

	if (x<0)
		return NULL;
	if ((x-=_zpSegment+1)<0)
		return f->sbi[x+_zpSegment+1];
	if ((s=f->sgs)==NULL)
		return NULL;
	for(s; x && *s; s+=*s,s+=1+4,--x)
		;
	if (x)
		return NULL;
	return s+*s++;
}

void swapSeg(asm6502frame *f,
	uint8 *d,
	uint8 x,
	uint8 new)
{
	int		v = 0;

	if (!d || x==f->seg)
		return;
	if (!new)
		v = d[0]|(d[1]|(d[2]|d[3]<<8)<<8)<<8;
	if ((d=getSeg(f,f->seg))!=NULL)
	{
		d[0] = f->org;
		d[1] = f->org>>8;
		d[2] = f->bnk;
		d[3] = f->bnk>>8|(f->flg&_orgInited);
	}
	if (!new)
	{
		f->flg&=~_orgInited, f->flg|= v>>24&_orgInited;
		f->bnk = v>>16&0x7FFF;
		f->org = v;
	}
	f->seg = x;
	if ((unsigned)x>_zpSegment
			|| (f->seg<=_dataSegment && f->flg&_orgInited
			&& (f->org>=f->flr || f->opt&_DataOnly)))
		f->pos = f->org;
}

/* returns parsed length or <0 on error */
static int setSeg(asm6502frame *f,
	const char *s,
	const char *e)
{
	char		*t = NULL;
	uint8		*d;
	int32		x, z, i, ix;

	if ((x=X6502_ParseName(s,e,&z,&t,NULL))<0)
	{
		if (t)
			free(t);
		return x;
	}
	if (z-1u>255u-1u)
		{x = _e_name_too_long; goto end;}
	if ((s=parse_skipwsp(s+x,e))<e)
		{x = _e_general_error; goto end;}
	/* fill capitalized in 256-char buffer */
	d = (uint8*)t, i = z;
	do
		*d++ = toupper((uint8)*d);
	while(--i);
	if (strMatch(keyw_list[_zeropage-OPN_6510_COUNT],t,z)
			|| strMatch(keyw_list[_zp-OPN_6510_COUNT],t,z))
		{ix = _zpSegment; goto bix;}
	else if (strMatch(keyw_list[_bss-OPN_6510_COUNT],t,z))
		{ix = _bssSegment; goto bix;}
	else if (strMatch(keyw_list[_data-OPN_6510_COUNT],t,z))
		{ix = _dataSegment; goto bix;}
	else
	{
		if (strMatch(keyw_list[_code-OPN_6510_COUNT],t,z)
				|| strMatch(keyw_list[_text-OPN_6510_COUNT],t,z))
		{
			ix = _textSegment;
		bix:	d = f->sbi[ix];
			goto set;
		}
	}
	ix = _zpSegment+1;
	if ((d=f->sgs)==NULL)
		{i = 0; goto add;}
	for(; *d; d+=*d,d+=1+4,++ix)
		if (*d==z && *t==d[1] && !memcmp(d+2,t+1,z-1))
		{ /* have it, swap it */
			d+= z, ++d;
			goto set;
		}
	if (ix>255)
		return _e_stack_error;
	i = d-f->sgs; /* end offset including null term., 1 length, 4 val */
add:	s = NULL;
	if ((d=realloc(f->sgs,z+1+4+i))==0)
		{x = _e_memory_error; goto end;}
	f->sgs = d, d+= i;
	*d++ = z;
	memcpy(d,t,z);
	d+= z;
	d[4] = 0;
set:	swapSeg(f,d,ix,!s);
end:	free(t);
	return x;
}

int popSeg(asm6502frame *f)
{
	uint8		v;

	if (!stPop(&f->sst,&v))
		return _e_stack_error;
	swapSeg(f,getSeg(f,v),v,0);
	return 0;
}

/* ---- disposal */

static int freeDef(preDef **p)
{
	preDef		*l;
	int		r = 0;

	if ((l=*p)!=NULL)
	{
		for(r=0; l->data; ++l)
			{free(l->data); l->data = NULL, ++r;}
		free(*p); *p = NULL, ++r;
	}
	return r;
}

int X6502_FreeScan(asm6502frame *f)
{
	int		r;

	if (!f)
		return _e_general_error;
	/* preprocess */
	r = freeDef(&f->pre);
	if (f->est)
		{free(f->est); f->est = NULL, ++r;}
	/* modules */
	r+= freeDef(&f->prc);
	if (f->mod)
		{free(f->mod); f->mod = NULL, ++r;}
	/* segment */
	if (f->sgs)
		{free(f->sgs); f->sgs = NULL, ++r;}
	if (f->sst)
		{free(f->sst); f->sst = NULL, ++r;}
	memset(f->sbi,0,sizeof(f->sbi));
	return r;
}

int X6502_FreeFrame(asm6502frame *f)
{
	int		r;

	if (!f)
		return _e_general_error;
	r = X6502_FreeScan(f);
	if (f->lab)
		{free(f->lab); f->lab = NULL, ++r;}
	return r;
}

/* ---- assembler */

static int getOpcode(uint16 t)
{
	int		i = 0;

	do /* mask used OPB_BAD - BAD. */
		if ((x6510_opn_table[i]&(OPB_FLAG-1))==t)
			return i;
	while(++i<256);

	return ~0;
}

static const char *parsePragmaParam(asm6502frame *f,
	const char *s,
	const char *e,
	int *r,
	evalvalue *v)
{
	const char	*n;
	uint8		opt, par;

	s = n = parse_skipwsp(s,e);
	if (s>=e)
		{*r = _e_general_error; goto end;}
	if (*s=='(')
		n = getBoundsEnd(++s,e), par = 1;
	else
	{
		n = s;
		if (*n=='+')
			SKIPCHR(n,e,'+');
		else
			SKIPCHR(n,e,'-');
		n = IDBreak(n,e,1), par = 0;
	}
	opt = f->opt, f->opt&=~_Prescan; /* no temporary labels */
	s+= evaluate_line(s,n,v,r,parseValue,NULL,f,NULL);
	f->opt = opt;
	if ((unsigned)*r>=_e_return)
		goto end;
	if (par && (n>=e || *n++!=')'))
		{*r = _e_general_error; goto end;}
end:	return n;
}

static long printt(asm6502frame *f,
	const char *s,
	const char *e)
{
	char		*t = NULL;
	int32		r, z;

	if ((r=X6502_ParseName(s,e,&z,&t,NULL))>=0 && !(f->opt&(_Prescan|_QuietMode)))
	{
		f->flg|= _PrintNL;
		fwrite(t,1,z,stdout);
	}
	if (t)
		free(t);
	return r;
}

/* returns source text name length or error */
static long incbin(const char *n,
	asm6502frame *f,
	unsigned long int ofs,
	unsigned long int amt)
{
	void (*bw)(asm6502frame*,
		uint8);
	long int	z;
	FILE		*nf = NULL;

	if ((nf=fopen(n,"rb"))==NULL)
		return _e_include_open;
	fseek(nf,ofs,SEEK_SET);
	if ((bw=f->writef)==NULL)
	{
		fseek(nf,0,SEEK_END);
		if ((z=ftell(nf))>ofs)
			z = MIN(z,amt)-ofs;
		else
			z = 0;
		f->org+= z;
	} else
	{
		int		c;

		for(z=amt; amt && (c=fgetc(nf))>=0; ++f->org,--amt)
			bw(f,c);
	}
	fclose(nf);
	return !!z;
}

static void incchar(void *u,
	char c)
{
	asm6502frame	*f;

	if ((f=u)==NULL)
		return;
	if (f->writef)
		f->writef(f,c);
	++f->org;
}

static int optimizeMode(int m,
	uint32 v)
{
	if (v>255)
	{ /* expand */
		if (m==OPM_ZP)
			m = OPM_AB;
		else if (m==OPM_IN)
			m = OPM_IB;
		else if (m==OPM_ZX)
			m = OPM_AX;
		else if (m==OPM_ZY)
			m = OPM_AY;
	} else if (m==OPM_AB) /* reduce */
		m = OPM_ZP;
	else if (m==OPM_IB)
		m = OPM_IN;
	else if (m==OPM_AX)
		m = OPM_ZX;
	else if (m==OPM_AY)
		m = OPM_ZY;
	return m;
}

static long makeTmpLabelName(uint8 *d,
	asm6502frame *f,
	char y)
{
	int		z;

	if (!d || !f)
		return _e_general_error;
	*d = 'A';
	z = 1+makebasestring(d+1,NULL,f->ctr,20,1,0);
	d[z++] = y;
	return z;
}

#define prepWr(f,r)	do{ \
	if((unsigned)f->seg-_bssSegment<=_zpSegment-_bssSegment) \
		{r = _e_write_in_addr_seg; goto end;} \
	if(!(f->flg&_orgInited)) { \
		f->flg|= _orgInited, f->org = f->pos; \
		if(!r) wrn=~_e_unprep_org; \
	} \
}while(0)

long X6502_Assemble(const char *str,
	const char *e,
	int *_r,
	asm6502frame *f)
{
	uint8		b[256];
	asm6502frame	af;
	void (*bw)(asm6502frame*,
		uint8);
	const char	*s, *le;
	int		r, m, x, wrn;

	if ((s=str)==NULL)
		return 0;
	if (!f) /* scratch - clear */
		memset(f=&af,0,sizeof(af));
	r = x = wrn = 0, f->flg&=~(_PreEstID|_LScanErr), bw = f->writef;
/*	k = f->sct==_zpSection ? 0xFF:0xFFFF;*/
	++f->ctr;
	le = getLineEnd(s,e);
	while(s<e)
	{
		evalvalue	v;
		const char	*n, *ls;
		long		z, l, lorg;

		if (s>=le)
			goto nl;
		n = ls = parse_skipwsp(s,le);
		if (n!=s)
			f->flg|= _IntoLine;
		m =-1, z = 0, lorg = f->org, n = IDBreak(s=n,le,1);
		if ((x=X6502_ParseKey(s,n))>=0)
		{
			/* some assemblers allow non-instruction keys at start */
			if (x<OPN_6510_COUNT && !(f->flg&_IntoLine))
				{r = _e_label_error; goto end;}
			if ((unsigned)x-_default<=(unsigned)_space-_default)
			{
				s = parse_skipwsp(n,le), n = IDBreak(s,le,1);
				goto labo;
			} else
				goto gotn;
		}
		if (*n==':' || !(f->flg&_IntoLine))
		{ /* identifier: equate/label */
			uint8		t;

		labo:	t = _Label;
			if (*s-'0'<=9u) /* can't start w. a digit */
				{r = _e_label_error; break;}
			if ((z=n-s)==0) /* try for anon. label */
			{
				char		y = *n;

				if ((x<0 && y=='*') || *n==':')
					z = 1; /* will be automatically skipped */
				else
				{
					if (y=='+')
						SKIPCHR(n,le,'+');
					else
						SKIPCHR(n,le,'-');
					if (s==n)
					{
						r = strchr("\\'\"",y) ? _e_general_error
							:_e_no_identifier;
						break;
					}
				}
				z = makeTmpLabelName(b+128,f,y);
				s = (char*)b+128;
			} else if (x==_default)
				t = _Default;
			if (toupper((uint8)*s)=='A' && z==1)
				{r = _e_identifier_conflict; break;} /* register-reserved */
			if ((m=X6502_AsmAddID(s,s+z,b,f,t))<=0)
				{r = m ? _e_identifier_conflict:_e_name_too_long; break;}
			/* _proc put here since trailing ':' is invalid */
			if (x==_proc)
				goto gotp; /* reuse for module name (indep. labels) */
			if ((*n&0xEF)==0x2A) /* ':' || *n=='*' */
				++n;
		} else if (x<0)
		{
			r = (n=parse_skipwsp(n,le))>=le ? 0
				:IDBreak(n,le,0)!=n ? _e_general_error:_e_macro_error;
			break;
		}
		s = n = parse_skipwsp(n,le);
		if ((unsigned)x-_alias<=_space-_alias)
			goto gotn;
		if ((n=IDBreak(n,le,1))<=s)
		{
			if (*s=='=')
			{
				++n;
				if (x!=_default)
				{
					if (m<0)
						{r = _e_general_error, s = str; break;}
					x = _equ;
				}
				goto gotn;
			}
			if (m>=0)
				goto lset;
			if (x<0 && ls<le)
				s = ls, r = _e_macro_error;
			break;
		} else if (x==_default) /* must follow '=' */
			{r = _e_general_error; break;}
		else if (x==_space)
			goto gotp;
		/* macro or instruction */
		if ((x=X6502_ParseKey(s,n))<0)
			{r = x==_e_no_identifier ? _e_macro_error:x; break;}
		/* keyword, possible equate, etc. */
	gotn:	s = parse_skipwsp(n,le);
	gotp:	r = 0; /* now at position of the parameter(s) */
		switch(x)
		{ /* the `r` end-value when using lset displaces org */
		case _align:
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if (v-1u>0xFFFFu-1u)
				{r = _e_range_error; goto end;}
			f->org+= v-1, f->org-= f->org%v, /*f->org&= k,*/ r = 0;
			if ((unsigned)f->seg-_bssSegment>_zpSegment-_bssSegment)
			{
				if ((l=f->org-lorg)<0)
					wrn =~_e_org_wrap;
				f->pos+= l;
			}
			break;
		case _bank: /* this may carry other params, including 'SLOT', but... */
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			/* 32 KB always mapped, even if mirrored */
			if ((unsigned)v>f->maxbnk && (unsigned)v>1)
				{r = _e_range_error; goto end;}
			f->bnk = v, r = 0;
			break;
		case _block: /* asign block, skip (add to ORG) */
		case _res:
		case _space:
			/*if (s>=e)
				{r = _e_general_error; goto end;}*/
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			f->org+= v; /*f->org&= k;*/
			goto lset;
		case _checkpc:
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if (v-1u>0xFFFFu-1u)
				{r = _e_range_error; goto end;}
			if (lorg>v)
				{r = _e_check_error; goto end;}
			break;
		case _ascii: /* strings only */
		case _byt:
		case _byte: /* no strings */
		case _db:
		case _dcb:
		case _dc_b:
			prepWr(f,r);
			l = 0;
			do
			{
				if ((l && *s++!=',') || (s=parse_skipwsp(s,le))>=le)
					{r = _e_general_error; goto end;}
				if (*s=='\047' || *s=='"')
				{
					if (x==_byte)
						{r = _e_data_type; goto end;}
					n = s;
					s = parse_textloop(s+1,le,*n,NULL,NULL,NULL,f,incchar);
					if (le-n<2 || s[-1]!=*n) /* point of error */
					{
						r = s>=le ? _e_open_string:_e_general_error;
						goto end;
					}
				} else
				{
					if (x==_ascii)
						{r = _e_data_type; goto end;}
					s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
					if ((unsigned)r>=_e_return)
						goto end;
					if (v>255 || v<-0x80)
						wrn =~_e_range_error;
					incchar(f,v);
				}
			} while(++l,(s=parse_skipwsp(s,le))<le);
			goto lset;
		case _defchr:
			prepWr(f,r);
			l = 0;
			do
			{ /* supposed to pick up 8 hex codes, proc. 2 bits of ea. nybble */
				uint16	dp;

				if (l && ((s=parse_skipwsp(s,le))>=le || *s++!=','))
					{r = _e_general_error; goto end;}
				s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
				if ((unsigned)r>=_e_return)
					goto end;
				dp = 0, x = 8;
				do
					dp+=dp, dp|= (v>>28-8)&0x100, dp|= (v>>28+1)&1;
				while(v<<=4,--x);
				b[l] = dp>>8, b[l+8] = dp;
			} while(++l<8);
			if (!bw)
				f->org+= 16;
			else
			{
				x = 0;
				do
					bw(f,b[x]);
				while(++f->org,++x<16);
			}
			x = _defchr;
			goto lset;
		case _dcl:
		case _dc_l:
			x = 4;
			goto dcw;
		case _addr:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
		case _dw:
		case _dcw:
		case _dc_w:
		case _word:
			x = 2;
		dcw:	prepWr(f,r);
			l = 0;
			do
			{
				if ((l && *s++!=',') || (s=parse_skipwsp(s,le))>=le)
					{r = _e_general_error; goto end;}
				s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
				if ((unsigned)r>=_e_return)
					goto end;
				if (x<4 && (v>65535 || v<-0x8000))
					wrn =~_e_range_error;
				if (!bw)
					f->org+= x;
				else
				{
					bw(f,v), ++f->org, bw(f,v>>8), ++f->org;
					if (x>2)
						bw(f,v>>=16), ++f->org, bw(f,v>>8), ++f->org;
				}
			} while(++l,(s=parse_skipwsp(s,le))<le);
			x = x>2 ? _dcl:_dcw;
			goto lset;
		case _end:
			f->flg|= _EndKeyw;
			if (f->flg&_IntoProc)
				{f->flg&=~_IntoProc, r = _e_open_proc; goto end;}
			break; /* cannot put a label on 'end' lines */
		case _endproc:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (f->flg&_IntoProc)
			{
				f->flg&=~_IntoProc;
				if (f->mod)
					{free(f->mod); f->mod = NULL;}
			} else
				wrn =~_e_no_proc;
			break; /* cannot put a label on 'end' lines */
		case _alias:
		case _default:
		case _equ: /* set label to explicit value */
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if (m<0)
				{r = _e_no_identifier; goto end;}
			if (f->lab)
			{
				r = x==_default ? _Default:_Equate;
				if (f->opt&_Prescan)
					r|= 0x40;
				f->lab[m] = r;
			}
			lorg = v;
			goto lset;
		case _incbin:
			{
				char		*t = NULL;
				evalvalue	ofs = 0, amt =-1;

				if ((r=X6502_ParseName(s,le,NULL,&t,f->lfn))<=0)
				{
					if (t)
						free(t);
					r = r ? r:_e_incomplete_error;
					goto end;
				}
				for(s+=r; (s=parse_skipwsp(s,le))<le; )
				{
					n = IDBreak(s,le,0), r = n-s;
					if (r==4 && !strncasecmp(s,"SKIP",4))
						s = parsePragmaParam(f,s+4,le,&r,&ofs);
					else if (r==4 && !strncasecmp(s,"READ",4))
						s = parsePragmaParam(f,s+4,le,&r,&amt);
					else
						{r = _e_unsupported_param; goto end;}
					if (r<0)
						goto end;
				}
				prepWr(f,r);
				r = incbin(t,f,ofs,amt);
				free(t);
				if (r<0)
					goto end;
			}
			x = _incbin;
			goto lset;
		case _include: /* read this file in place of line */
			r = _e_include;
			goto lset;
		case _accu: /* accumulator depth (for 65816 status flag switch) */
		case _index: /* op-index depth (for 65816 status flag switch) */
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if (v!=8)
				{r = _e_size, s = n+1; goto end;}
			r = 0;
			break;
		case _ineschr:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if ((unsigned)v>512)
				{r = _e_range_error; goto end;}
			if (f->maxchr>=0 && f->maxchr!=v)
				{r = _e_value_set; goto end;}
			f->maxchr = v, r = 0;
			break;
		case _inesmap:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if ((unsigned)v>512)
				{r = _e_range_error; goto end;}
			if (f->map>=0 && f->map!=v)
				{r = _e_value_set; goto end;}
			f->map = v, r = 0;
			break;
		case _inesmir:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if ((unsigned)v>2)
				{r = _e_range_error; goto end;}
			r = f->opt&_MirrorMask;
			if (r!=_MirrorMask && r!=v)
				{r = _e_value_set; goto end;}
			f->opt&=~_MirrorMask, f->opt|= v, r = 0;
			break;
		case _mem:
		case _inesprg:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if (x==_mem)
				{wrn =~_e_handler_error; break;}
			if ((unsigned)--v>=(0x3FFFFFF>>f->bsz) || v>=0x7FFFu)
				{r = _e_range_error; goto end;}
			if (f->maxbnk>=0 && f->maxbnk!=v)
				wrn =~_e_value_set;
			f->maxbnk = v, r = 0;
			break;
		case _module:
			if (f->flg&_IntoProc)
				{r = _e_module_in_proc; goto end;}
			goto dmd;
		case _proc:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (f->flg&_IntoProc)
				{r = _e_already_in_proc; goto end;}
			f->flg|= _IntoProc;
			wrn =~_e_cannot_optimize;
		dmd:	if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (IDBreak(s,le,0)==s) /* can't stuff *any* char into labels */
			{
				*b&= 0x3F, *b|= (~*b*2)&0x40; /* unmangle */
				if (s<le || *str=='-' || *str=='+')
					{r = _e_general_error; goto end;}
				else if ((r=X6502_ParseName((char*)b,(char*)b+z,
						&l,&f->mod,NULL))==0)
					{r = _e_general_error; goto end;}
			} else if ((r=X6502_ParseName(s,le,&l,&f->mod,NULL))==0)
				{r = _e_general_error; goto end;}
			if (r>255) /* overflow */
				r = _e_name_too_long;
			if (r<0)
				goto end;
			s+= r;
			if (f->flg&_Preprocessed)
				m =-1; /* handled externally */
			goto lset;
		case _opt:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			s = parse_skipwsp(s,le);
			n = IDBreak(s,le,0), r = n-s;
			if (r==2 && !strncasecmp(s,"ON",2))
				f->opt|= _Optimize;
			else if (r==3 && !strncasecmp(s,"OFF",3))
				f->opt&=~_Optimize;
			else
				{r = _e_off_or_on; goto end;}
			r = 0, s = n;
			break;
		case _processor:
			if ((n=IDBreak(s,le,0))==s)
				{r = _e_general_error; goto end;}
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if (n-s!=4)
				goto lang;
			if (!memcmp(s,"6502",4))
		case _p02:	f->opt|= _6502Only;
			else if (!memcmp(s,"6510",4)
					|| !strncasecmp(s,"2A03",4)
					|| !strncasecmp(s,"2A07",4))
		case _p10:	f->opt&=~_6502Only;
			else
			{
			lang:	r = _e_language_error;
				goto end;
			}
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			s = n;
			break;
		case _code:
		case _text:
		case _data:
		case _bss:
		case _zp:
		case _zeropage:
			if (m>=0)
				break;
			if (*(s=ls)=='.')
				++s;
		case _segment:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			if ((r=setSeg(f,s,le))<0)
				goto end;
			s+= r, r = 0;
			break;
		case _pushseg:
			if (!stPush(&f->sst,f->seg))
				{r = _e_memory_error; goto end;}
			break;
		case _popseg:
			if ((r=popSeg(f))!=0)
				goto end;
			break;
		case _printt:
			if ((r=printt(f,s,le))<0)
				goto end;
			s+= r, r = 0;
			break;
		case _printv:
			s = parse_skipwsp(s,le);
			n = IDBreak(s,le,0), r = n-s;
			if (r==3 && !strncasecmp(s,"DEC",3))
				l = 0;
			else if (r==3 && !strncasecmp(s,"HEX",3))
				l = 1;
			else
				{r = _e_dec_or_hex; goto end;}
			s = n;
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			r = 0;
			if (f->opt&(_Prescan|_QuietMode))
				break;
			f->flg|= _PrintNL;
			printf(l ? "%lX":"%ld",
				v);
			break;
		case _export: /* in-text name to be flagged in library */
		case _import: /* text export library name needed */
			r = _e_handler_error;
			goto end;
		case _subroutine:
			if (f->opt&_DataOnly)
				{r = _e_data_only; goto end;}
			/*wrn =~_e_cannot_optimize;*/
			goto lset;
		case _advance: /* programmer's semantical variant of .org */
			x = _org;
		case _fill: /* fill bytes of length, optional value */
		case _org: /* set origin position */
		case _orga: /* set same-bank origin position */
		case _pad:  /* pad to this next address */
			if (s>=le)
				{r = _e_general_error; goto end;}
			s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
			if ((unsigned)r>=_e_return)
				goto end;
			if ((unsigned)v>0xFFFF)
				{r = _e_range_error; goto end;}
			if (x==_fill)
				l = v;
			else
				l = v-lorg;
			if (*(s=parse_skipwsp(s,le))==',')
			{
				v = 0;
				if (++s>=le)
					{r = _e_general_error; goto end;}
				if (x==_orga)
					{r = _e_unsupported_param; goto end;}
				s+= evaluate_line(s,le,&v,&r,parseValue,NULL,f,NULL);
				if ((unsigned)r>=_e_return)
					goto end;
			} else if ((unsigned)x-_org<=_orga-_org)
			{
				if ((unsigned)f->seg>_zpSegment
						|| (f->seg<=_dataSegment && f->flg&_orgInited
						&& (v>=f->flr || f->opt&_DataOnly)))
					f->pos = v;
				f->org+= l; /*f->org&= k;*/
				f->flg|= _orgInited;
				if (x==_org)
					if (f->org==f->flr) /* lower, fixed bank */
						f->bnk = 0, f->pos = f->org;
					else if (f->org==0x10000-(1u<<f->bsz))
					{
						if (f->bnk<0)
							f->bnk = 0;
						if (/*v<f->org &&*/ ++f->bnk>f->maxbnk && f->maxbnk)
							f->bnk = 1, wrn =~_e_bank_overflow;
						f->pos = f->org;
					}
				r = 0;
				break;
			} else
				v = 0;
			prepWr(f,r);
			if (l>0)
				if (!bw)
					f->org+= l;
				else
					do
						bw(f,v);
					while(++f->org,--l);

		default: /* label at org */
		lset:	if (m>=0)
			{
				uint8	*d;

				if ((d=f->lab)==NULL)
					{r = _e_read_error; break;}
				d+= m;
				d[1] = lorg;
				d[2] = lorg>>8;
				d[3] = lorg>>16;
				d[4] = lorg>>24;
				m = f->ctr;
				d[5] = m;
				d[6] = m>>8;
				d[7] = m>>16;
				d[8] = m>>24;
			}
			if (r!=_e_include)
				r = 0;
			m =-1;
		}
		if (m>=0) /* must be cleared by now */
			{r = _e_label_error; goto end;}
		s = parse_skipwsp(s,le);
		if ((unsigned)x>=OPN_6510_COUNT)
		{ /* not encoding an instruction */
			if (s>=le)
			{ /* make sure ptr is after eol and continue */
			nl:	while(s<e && !IS_EOL(*s))
					++s;
				f->flg&=~_LineFlags, ++s;
				le = getLineEnd(s,e);
				continue;
			}
			if (r!=_e_include)
				r = _e_general_error;
			goto end;
		}
		prepWr(f,r);
		if (f->opt&_DataOnly)
			{r = _e_data_only; goto end;}
		if (f->opt&_6502Only && x>OPN_ROR)
			{r = _e_6510_instruction, s = n-2; goto end;}
		r = parseOperands(s,le,&v,f,&m);
		/* not supposed to get _e_return, but handle it anyway */
		if ((unsigned)r>=_e_return)
			goto end;
		s+= r;

		/* optimize - op size must be consistent (no unestablished IDs) */
		if ((unsigned)m<OPM_IA && f->opt&_Optimize && !(f->flg&_PreEstID))
			m = optimizeMode(m,v);
		if ((r=getOpcode(x|(m<<8)))<0)
			if (m==OPM_IP && (r=getOpcode(x|(OPM_IA<<8)))>=0)
				m = OPM_IA;
			else if ((m==OPM_ZP || m==OPM_AB)
					&& (r=getOpcode(x|(OPM_RL<<8)))>=0)
			{ /* instruction calls for a relative; convert */
				m = OPM_RL, v-= f->org, v-= 2;
				/* check range - not necessarily established in prescan */
				if (v!=(int8)v && !(f->opt&_Prescan))
					{r = _e_range_error; break;}
			} else if (f->opt&_Optimize)
			{ /* last resort: if reduced, check for a larger, alt. mode */
				m = optimizeMode(m,256);
				if ((r=getOpcode(x|(m<<8)))<0)
					{r = _e_op_error; break;} /* no such instruction */
			} else
				{r = _e_op_error; break;}
		if (bw)
			bw(f,r);
		++f->org, /*f->org&= k,*/ r = 0;
		if (m>=OPM_IA)
			continue;
		if (bw)
			bw(f,v);
		++f->org; /*f->org&= k;*/
		if ((unsigned)(m-OPM_IB)<=OPM_AY-OPM_IB)
		{ /* 2-byte mode range */
			if (bw)
				bw(f,v>>8);
			++f->org; /*f->org&= k;*/
 			if ((unsigned)v>0xFFFF) /* 6510: addresses only */
				wrn =~_e_range_error;
		} else if ((unsigned)v>0xFF && v<-0x80 && m!=OPM_RL && x!=OPN_AND)
			wrn =~_e_range_error;
	}
	f->flg&=~_IntoLine;
end:	if (f==&af)
		X6502_FreeFrame(f);
	if (_r)
		*_r = r ? r:wrn;
	return s-str+!r; /* slightly out of range in case of a non-position warning */
}

/* ---- preprocessing */

char		lastTRStr[32];

static const char *prestr[] =
{
	"__DATE__",
	"__FILE__",
	"__LINE__",
	"__TIME__",
	"__6510__",
	"DEFINE",
	"UNDEF",
	"ERROR",
	"FAIL",
	"IF",
	"IFDEF",
	"IFNDEf",
	"ELSE",
	"ELIF",
	"ENDIF"
};

static int _getPreDef(const preDef *l,
	const char *s,
	long z,
	int32 m)
{
	const preDef	*a;
	const char	*n, **p, **e;
	char		c = toupper((uint8)*s++);

	--z;
	if ((a=l)!=NULL)
		for(; (n=a->data)!=NULL; ++a)
			if (toupper((uint8)*n++)==c && !strncasecmp(s,n,z) && !n[z])
				return a-l;
	for(p=prestr,e=p+m; p<=e; ++p)
		if (toupper((uint8)*(n=*p))==c && !strncasecmp(s,++n,z) && !n[z])
			return p-prestr+_date_def;
	return l-a-1;
}

int X6502_GetPreDef(const asm6502frame *af,
	const char *s,
	const char *e)
{
	long		z;

	if (!af->pre || *s=='.' || (z=IDBreak(s,e,0)-s)==0)
		return -1;
	return _getPreDef(af->pre,s,z,_endif-_date_def);
}

static int _addPreDef(preDef **p,
	long i,
	const char *n,
	long t,
	const char *s,
	long z)
{
	char		*d;
	preDef		*a;

	if (i<0)
		return _e_general_error;
	if ((a=realloc(*p,(2+i)*sizeof(preDef)))==NULL)
		return _e_memory_error;
	*p = a;
	a+= i;
	if ((a->data=d=malloc(a->size=t+1+z))==NULL)
		return _e_memory_error;
	memcpy(d,n,t), d+= t, *d++ = 0;
	memcpy(d,s,z);
	a[1].data = NULL;
	return t;
}

int X6502_AddPreDef(asm6502frame *f,
	const char *str,
	const char *e)
{
	const char	*s;
	long		i, t;

	if (*str=='.' || X6502_ParseKey(str,e)>=0)
		return _e_no_identifier;
	if ((t=(s=IDBreak(str,e,0))-str)==0)
		return _e_no_identifier;
	if (!f->pre)
		i =-1;
	else if ((i=_getPreDef(f->pre,str,t,_endif-_date_def))>=0)
		return _e_identifier_conflict;
	s = parse_skipwsp(s,e); /* wsp after name replaced w. null */
	return _addPreDef(&f->pre,~i,str,t,s,e-s);
}

int X6502_SubPreDef(asm6502frame *f,
	const char *str,
	const char *e)
{
	const char	*s;
	preDef		*a;
	long		i;

	s = IDBreak(str,e,0);
	if (s==str || (unsigned)(i=X6502_GetPreDef(f,str,e))>=_date_def)
		return _e_no_identifier;
	a = f->pre+i;
	if (a->data)
		free(a->data);
	if (!i && !a[1].data)
	{ /* list now empty; compact */
		free(f->pre); f->pre = NULL;
		goto end;
	}
	/* move the following entries down */
	for(; a->data; ++a) /* leaves at end NULL entry */
		*a = a[1];
	if ((a=realloc(f->pre,(a-f->pre)*sizeof(preDef)))!=NULL)
		f->pre = a;
end:	return s-str;
}

static char *preStr(const char *s,
	uint32 *_z)
{
	const char	*t = "ZhSeEqkabtlvfr";
	char		*a, *d;
	long		z = strlen(s);
	uint8		c;
	/* largest size is double strlen; add first quote */
	if ((d=a=malloc(z*2+1))==NULL)
		return NULL;
	*d++ = '"';
	do
		if ((c=*s++)<14)
			*d++ = '\\', *d++ = t[c];
		else if (c=='"' || c=='\\')
			*d++ = '\\', *d++ = c;
		else
			*d++ = c;
	while(--z);

	/* minimize allocation, plus termination */
	if ((d=realloc(a,1+(*_z=d-a+1)))==NULL)
	{
		if (*_z)
			free(a);
		return NULL;
	}
	a = d, d+= *_z;
	d[-1] = '"', *d = 0;
	return a;
}

static void preUnmark(preDef *l)
{
/*	if (!l || !l->name)
		return 0;
*/
	for(; l->data; ++l)
		l->size&=~IMIN(l->size);
/*	return 1;*/
}

static char *preExpand(asm6502frame *af, /* must have valid .pre */
	char *str,
	uint32 *z, /* bit #31 may be set on syntax/param error */
	char upd)
{
	const char	*e = str+*z, *n;
	char		*p, *d, skp;
	/* replace text of all applicable define */
	for(d=str,skp=0; d<e; ++d)
	{
		preDef		*a;
		char		*na;
		uint32		t;
		long		i, x;
		char		q;

		switch(*d)
		{
		case '\047':
		case '"': /* must skip until endquote */
			q = *d;
			while(++d<e && *d!=q)
				if (*d=='\\')
					++d; /* esc-skip */
			break;
		case ';': /* with eols erased, just... skip all */
			/**z = d-str, upd|= 1;*/
			goto end;
		CASE_NUMERIC
			skp = 1;
			break;
		case'd':
			if (upd>1 && d+6<e && !memcmp(d+1,"efined",6))
			{ /* defined() */
				n = parse_skipwsp(d+7,e), x = 0;
				if (n<e && *n=='(') /* parentheses optional */
					n = parse_skipwsp(n+1,e), x = 1;
				if ((i=IDBreak(n,e,0)-n)==0)
					goto ser;
				*lastTRStr = '0'+(_getPreDef(af->pre,n,i,_6510_def-_date_def)>=0);
				n+= i;
				if (x)
				{
					n = parse_skipwsp(n,e);
					if (n>=e || *n++!=')')
					{
					ser:	*z|= IMIN(int);
						goto end;
					}
				}
				i = n-d, na = NULL, n = lastTRStr, t = 1;
				goto xfe;
			}
		case'_':case'A':case'B':
		case'C':case'D':case'E':
		case'F':case'G':case'H':
		case'I':case'J':case'K':
		case'L':case'M':case'N':
		case'O':case'P':case'Q':
		case'R':case'S':case'T':
		case'U':case'V':case'W':
		case'X':case'Y':case'Z':
			case'a':case'b':
		case'c':	case'e':
		case'f':case'g':case'h':
		case'i':case'j':case'k':
		case'l':case'm':case'n':
		case'o':case'p':case'q':
		case'r':case's':case't':
		case'u':case'v':case'w':
		case'x':case'y':case'z':
			if (skp)
				break;
			i = t = IDBreak(d,e,0)-d; /* shouldn't fail */
			if ((x=_getPreDef(af->pre,d,i,_6510_def-_date_def))>=0)
				switch(x)
				{
				case _date_def:
					time((time_t*)lastTRStr);
					t = strftime(lastTRStr,sizeof(lastTRStr)-1,
						"\"%Y:%m:%d\"",gmtime((time_t*)lastTRStr));
					na = NULL, n = lastTRStr;
					goto xfe;
				case _file_def:
					if ((n=na=preStr(af->lfn,&t))==NULL)
						{*z = _e_memory_error; goto end;}
					goto xfe;
				case _line_def:
					t = makebasestring(lastTRStr,NULL,af->lnx,10,1,'-');
					na = NULL, n = lastTRStr;
					goto xfe;
				case _time_def:
					time((time_t*)lastTRStr);
					t = strftime(lastTRStr,sizeof(lastTRStr)-1,
						"\"%H:%M:%S\"",gmtime((time_t*)lastTRStr));
					na = NULL, n = lastTRStr;
					goto xfe;
				case _6510_def:
					*lastTRStr = '0'+!(af->opt&_6502Only);
					na = NULL, n = lastTRStr, t = 1;
					goto xfe;

				default:
					if ((a=af->pre+x)->size<0)
						break;
					n = na = NULL;
					if ((t=~i+a->size)!=0)
					{ /* expand preprocessing ID */
						if ((na=malloc(t))==NULL)
							{*z = _e_memory_error; goto end;}
						memcpy(na,a->data+i+1,t);
						/* must mark for non-recursion */
						a->size|= IMIN(a->size);
						n = na = preExpand(af,na,&t,0);
					}
				xfe:	if (t!=i) /* t=new slen; i=old slen */
					{
						long		mz = e-(d+i);

						*z+= t-i;
						if (t>i)
						{
							x = d-str; /* save the next offset */
							if ((p=realloc(str,*z&IMAX(long)))==NULL)
							{
								if (na)
									free(na);
								*z = _e_memory_error;
								goto end;
							}
							str = p, d = str+x, upd&=~1;
						} else
							upd|= 1;
						memmove(d+t,d+i,mz); /* move post-text */
						e = str+(*z&IMAX(long)); /* reset end-pointer */
					}
					memcpy(d,n,t);
					if (na)
						free(na);
				}
			d+= t; /* skipping next char too */

		default:
			skp = 0;
		}
	}
	if (upd&1 && (p=realloc(str,MAX(*z&IMAX(long),1)))!=NULL)
		str = p;
end:	return str;
}

static char *preDirect(asm6502frame *af,
	char *str,
	uint32 *z,
	long *t)
{
	const char	*s = str+*t;
	const char	*e = str+*z;
	evalvalue	v;
	int		r, x, i;

	*t = IDBreak(s,e,0)-s;
	r = _getPreDef(NULL,s,*t,_endif-_date_def);
	if ((unsigned)r-_define>_endif-_define)
	{
		*t = 0;
		return str;
	}
	s = parse_skipwsp(s+*t,e), *t = s-str;
	switch(r)
	{
	case _define:
		if (af->pei&0x80)
			break;
		if ((r=X6502_AddPreDef(af,s,e))<0)
			{*t = r; return str;}
		break; /* all rest of line accounted for */
	case _undef:
		if (af->pei&0x80)
			break;
		if ((r=X6502_SubPreDef(af,s,e))<0)
			{*t = r; return str;}
		s = parse_skipwsp(s+r,e);
		r = s<e && *s!=';' && !IS_EOL(*s) ? _e_general_error:1;
		break;
	case _error:
	case _fail:
		if (af->pei&0x80)
			break;
		*t = _e_error_directive;
		return str;
	case _ifdef:
	case _ifndef:
		x = IDBreak(s,e,0)-s, *t+= x;
		/* defined() includes __LINE__, etc. */
		i = _getPreDef(af->pre,s,x,_6510_def-_date_def)>=0;
		s = parse_skipwsp(s+x,e);
		if (s<e && *s!=';' && !IS_EOL(*s))
			{*t = _e_general_error; return str;}
		if (r==_ifdef)
			i^= 1;
		goto ifp;
	case _else:
	case _elif:
		if (af->pei&0x20)
			break;
		if (!af->est || (x=*(int32*)af->est)<0)
			{*t = _e_no_if; return str;}
		if (!((af->pei^=0xC0)&0x40))
			{*t = _e_else_else; return str;}
		/* *ei|= *ei<<2&0x80; */
		if (r==_elif)
		{
	case _if:	str = preExpand(af,str,z,2);
			if ((signed)*z<0)
			{
				if (*z<-14u) /* not an error code */
					*t = _e_general_error, *z&= IMAX(long);
				else
					*t = *z;
				return str;
			}
			e = str+*z, v = 0;
			i = evaluate_line(s,getLineEnd(s,e),&v,&x,parseValue,NULL,NULL,NULL);
			if (x<0) /* && (x!=_e_no_identifier || parse_skipwsp(IDBreak(s,e,0),e)<e))*/
			{
				*t = x;
				return str;
			}
			i = !v;
		ifp:	if (!stPush(&af->est,af->pei))
				{*t = _e_memory_error; return str;}
			af->pei|= af->pei>>2; /* necessary to keep lower levels disabled */
			if (!(af->pei&0x80))
				af->pei = i<<7;
		}
		break;
	case _endif: /* ei==0 is 'if' */
		if (!stPop(&af->est,&af->pei))
			{*t = _e_no_if; return str;}
	}
	*t = s-str;
	return str;
}

static int preAppend(preDef *a,
	void *s,
	long z,
	long nl)
{
	char		*p;
	long		t;

	if (a->data && (t=a->size)>=0) /* termator entry or unprep */
	{
		if (nl<0)
			nl = 0;
		t+= nl;
		if ((p=realloc(a->data,t+z))==NULL)
			return _e_memory_error;
		a->data = p, p+= t;
		if (nl)
			memset(p-nl,'\n',nl);
		memcpy(p,s,z);
		a->size = t+z;
	}
	return 0;
}

/* ---- errors */

/* converts tabs to spaces so the pointer (^) places right */
static void untab(char *s,
	const char *e)
{
	if (!s)
		return;
	while(s<e)
		if (*s++=='\t')
			s[-1] = ' ';
}

static const char	*err_strs[] =
{
	"syntax/param error",
	"memory error",
	"read error",
	"write error",
	"incomplete expression",
	"interrogator already open",
	"ID not found or applicable",
	"operation not supported here",
	"too many/few parentheses",
	"value out of range",
	"comment still open",
	"unterminated string",
	"ID already defined",
	"no such instruction mode",
	"invalid/unsupported label asignment",
	"unknown/unsupported language",
	"unrecognized macro, keyword or instruction",
	"subroutines are always linked, flat",
	"value already set to something else",
	"destination store is number"
};

static const char	*err_strs2[] =
{
	"include failed",
	"exceeded maximum include depth",
	"could not open include file",
	"expected 'on' or 'off'",
	"attempt to write to non-code/data space",
	"could not write here",
	"auto-banking wrapped - bank set to #1",
	"no such label",
	"wrap occurred in operation",
	"hit .proc within .proc",
	"hit end within a proc",
	"hit endproc outside a proc",
	"total name too long",
	"operation does not allow this data type",
	"instructions not allowed here",
	"6510 instructions were disabled",
	"ORG not set in this segment",
	"no .endif for .ifxxx at end of file",
	"no starting .ifxxx",
	"else after else",
	"higher value in PC check",
	"fatal error",
	"only 8 supported here",
	"unknown/unsupported parameter",
	"print type dec or hex needed",
	"cannot define a module within a proc",
	"nulls not allowed in names"
};

#define MAXIMUM_DEPTH		10	/*8*/

void reportErr(asm6502frame *af,
	char *str,
	long z,
	long t,
	int r,
	uint16 lpos)
{
	const char	*s = NULL;
	long		errc;

	if (r>0 && af->opt&_Prescan)
		return;
	if (r==_e_param_separation || r==_e_return)
		r = _e_general_error;
	if ((errc=r)<0)
		errc =~errc;
	if (errc<GET_ARRAY_COUNT(err_strs))
		s = err_strs[errc];
	else if ((errc-=~_e_include)<GET_ARRAY_COUNT(err_strs2))
		s = err_strs2[errc];
	else if (r<0)
		s = "unknown error";
	if (s)
	{
		if (af->flg&_PrintNL)
			fputc('\n',stdout), af->flg&=~_PrintNL;
		if ((signed)z>0)
			if ((unsigned)t<=z)
				fprintf(stderr,"### %.*s\n### %* ^\n",
					z,str,t);
			else /* offset returned is meaningless */
				fprintf(stderr,"# > %.*s\n",
					z,str);
		fprintf(stderr,"##%s%s\n%s '%s'; Line %u\n",
			r<0 ? "# ":" Warning - ",s,af->flg&_IntoProcList
			? "# Proc":"File",af->lfn,af->lnx);
		if (r==_e_pos_write_error)
			fprintf(stderr,"### POS = [$%04X,$%04X)\n\n",
				lpos,af->pos);
		else
			fprintf(stderr,
				r==_e_check_error ? "### ORG = $%04X\n\n":"\n",
				af->org);
	}
}

/* ---- reading */

#if __POCC__
 #pragma warn(disable:2154)
#endif

static char *readLine(FILE *f,
	char *str,
	uint32 *z,
	uint32 *x)
{
	const char	*s, *e;
	char		*d, c;

	if ((d=str=freadline(f,z,str,NULL,x,'\\'))==NULL)
		return NULL;
	if (!*z)
		return d;
	/* remove escaped eols */
	/* don't need to track double-esc because freadline omits non-esc'd eols
	  (an eol would necessarily be escaped)
	*/
	s = d, e = s+*z;
	do
		if ((c=*s++)=='\\' && e-s>1 && IS_EOL(*s))
			++s;
		else
			*d++ = c;
	while(s<e);
	*z = d-str;
	return str;
}

#if 1
static int addProc(asm6502frame *af) /* mod. name copied */
{
	preDef		*a;
	uint32		i = 0;
	int		r;

	if ((a=af->prc)!=NULL) /* get the next index */
		while(a->data)
			++a, ++i;
	r = makebasestring(lastTRStr,NULL,af->seg,10,0,0);
	if ((r=_addPreDef(&af->prc,i,af->mod,strlen(af->mod),lastTRStr,r))>=0)
		af->prx = i+1, r = 0;
	return r;
}
#else
 #define addProc(af) 0
#endif

static char *processLine(asm6502frame *af,
	char *str,
	uint32 *z,
	int *r,
	const char *fn,
	uint8 depth)
{
	const char	*e = str+*z, *s = parse_skipwsp(str,e);
	long		t;
	uint16		lpos = af->pos;
	/* highest-level preprocessing must be done here */
	af->lfn = fn;
	if (*s=='.')
	{
		t = 1+s-str;
		str = preDirect(af,str,z,&t);
		if (t)
		{
			*r = t<0 ? t:0;
			goto evr;
		}
	}
	if (af->pei&0x80)
		goto end;
	/* preExpand needs to be called at least once for sake of the __xx__ constants */
	if (!(af->flg&_IntoProcList))
	{
		if (af->pre)
			preUnmark(af->pre);
		str = preExpand(af,str,z,1);
	}
	if ((signed)z<0)
		{*r = *z, *z = 0, t =-1; goto se;}
	e = str+*z, s = str;

	/* within proc scan? */
	if (af->prx>0 && af->prc)
	{ /* append line to memory-stored proc text */
		const char	*n;

		s = parse_skipwsp(s,e);
		if ((t=X6502_ParseKey(s,e))>=0)
		{
			if (t==_endproc || t==_end)
			{
				af->prx = af->lnp = 0, af->flg&=~_IntoProc;
				if (af->mod)
					{free(af->mod); af->mod = NULL;}
				if (t==_end)
					{*r = _e_open_proc; goto se;}
				goto end;
			}
		} else if ((n=IDBreak(s,e,1))<e && (s==str || *n==':'))
		{
			if (*n==':')
				++n;
			s = parse_skipwsp(n,e);
			if ((t=X6502_ParseKey(s,e))==_endproc)
				{*r = _e_label_error; goto evr;}
		}
		if (!af->lnp)
			af->lnp = af->lnx-1;
		*r = preAppend(af->prc+af->prx-1,str,*z,af->lnx-af->lnp);
		af->lnp = af->lnx;
		goto end;
	}
	t = X6502_Assemble(s,e,r,af);
	if (*r==_e_include)
	{
		FILE		*nf = NULL;
		char		*b = NULL;

		s+= t;
		if ((*r=X6502_ParseName(s,e,NULL,&b,af->lfn))>0)
			if ((s=parse_skipwsp(s+*r,e))<e && !IS_EOL(*s) && *s!=';')
				*r = _e_general_error;
			else
				nf = fopen(b,"r");
		if (nf)
		{
			af->flg&=~_IntoLine; /* for files that start w. a label */
			*r = X6502_AsmFile(nf,b,af,depth+1);
			fclose(nf);
		} else if (*r>0)
			*r = _e_include_open;
		else
			*r = *r ? *r:_e_incomplete_error, t+= *r;
		if (b)
			free(b);
		if (*r>0) /* error/warning message already handled? */
			goto end;
	} else if (*r==~_e_cannot_optimize && af->mod)
		return str;
evr:	if (*r)
	{
	se:	untab(str,e);
		reportErr(af,str,*z,t,*r,lpos);
		*r = *r<0;
	} else if (af->flg&_PWrErr)
		{*r = _e_pos_write_error, t =-1; goto se;}
end:	if (*r || af->flg&_EndKeyw)
		if (str)
			{free(str); str = NULL;}
	return str;
}

static void setModLabel(asm6502frame *af,
	const char *s,
	const char *n)
{
	uint8		b[256];
	int32		x;

	if ((x=X6502_AsmGetID(s,n,b,af,NULL))>0)
	{
		uint8		*d = af->lab+x;

		d[1] = af->org;
		d[2] = af->org>>8;
		d[3] = af->org>>16;
		d[4] = af->org>>24;
	}
}

static int processProc(asm6502frame *af,
	int *r,
	uint8 depth)
{
	preDef		*l;

	if (af->mod)
		{free(af->mod); af->mod = NULL;}
	af->flg|= _IntoProc|_IntoProcList;
	for(l=af->prc,af->lnp=0; l->data; ++l)
	{
		const char	*n, *s = l->data, *e = s+l->size;
		char		*p;
		evalvalue	v;
		uint32		x, z;

		s+= strlen(s);
		/* read segment index (decimal) */
		if ((n=readanyinteger(s+1,e-s-1,&v))==NULL)
			{*r = _e_general_error; goto end;}
		swapSeg(af,getSeg(af,v),v,0);
		/* update label */
		setModLabel(af,l->data,s);
		af->mod = l->data;
		for(x=1; (s=++n)<e; ++x)
		{
			char		*str;

			while(n<e && *n!='\n')
				++n;
			if ((str=malloc(z=n-s))==NULL)
				{*r = _e_memory_error; goto end;}
			memcpy(str,s,z);
			af->lnx = x;
			str = processLine(af,str,&z,r,l->data,depth);
			if (str)
				free(str);
			if (*r)
				goto end;
		}
		af->mod = NULL;
		/* save memory by setting allocation to minimum */
		if ((p=realloc(l->data,l->size=1))!=NULL)
			l->data = p;
	}
end:	if (*r<0)
		{reportErr(af,NULL,0,0,*r,0); *r = 1;}
	af->flg&=~(_IntoProc|_IntoProcList);
	return freeDef(&af->prc);
}

int X6502_AsmFile(FILE *f,
	const char *fn,
	asm6502frame *af,
	uint8 depth)
{
	char		*str;
	uint32		z, x = 1;
	int		r = 0;

	if (!f || !af)
		return _e_general_error;
	str = malloc(0); /* needed for first-empty line(s) case */
	if (!depth)
		af->flg|= _Preprocessed, af->ctr = 0;
	else if (depth>=MAXIMUM_DEPTH)
		return _e_include_depth;
	while(z=~0,af->lnx=x,(str=readLine(f,str,&z,&x))!=NULL)
		if ((str=processLine(af,str,&z,&r,fn,depth))==NULL)
		{
			if (r)
				goto evr;
			z = 0;
			break;
		} else if (r==~_e_cannot_optimize)
		{
			if ((r=addProc(af))!=0)
			{
				reportErr(af,str,z,0,r,0);
				if (r<0)
				{
					if (str)
						{free(str); str = NULL;}
					goto evr;
				}
			}
			continue;
		}
	if (z) /* length - realloc failed */
		r = _e_memory_error;
	else
	{ /* check termination within .proc */
		if (af->flg&_IntoProc)
			reportErr(af,NULL,0,0,_e_open_proc,0), r = 1;
	}
evr:	if (!depth)
	{
		if (af->est)
		{ /* check termination within .if */
			if (!r)
				{reportErr(af,NULL,0,0,_e_open_if,0); r = 1;}
			free(af->est);
			af->est = NULL;
		}
		freeDef(&af->pre);
		if (!r && af->prc)
			processProc(af,&r,depth);
	}
	return r;
}

#ifdef SAMPLE
static long uppow2(long v)
{
	int		x = 1u<<(31-cntlz32(v));

	if (v!=x)
		x+= x;
	return x;
}

static int copyFile(FILE *d,
	FILE *s)
{
	long		z;
	char		b[1024];

	while((z=fread(b,1,sizeof(b),s))>0)
		if (fwrite(b,1,z,d)!=z)
			return _e_write_error;
	return 0;
}

typedef struct bankData
{
	int32		k, bk;
	uint8		b[1];
} bankData;

void bwrite(asm6502frame *f,
	uint8 v)
{
	bankData		*p;
	int32		a;

	if (!f || (p=f->u)==NULL)
		return;
	if ((a=f->pos++-f->flr)<0)
		{f->flg|= _PWrErr;}
	if (f->bnk>=0)
		a&= p->bk, a+= f->bnk<<f->bsz;
	p->b[a&p->k] = v;
}

/* this is the "logical minimum" routine */
void bexpand(asm6502frame *f,
	uint8 v)
{
#pragma unused(v)
	bankData		*p;
	int32		a;

	if (!f || (p=f->u)==NULL)
		return;
	if ((a=f->pos++-f->flr)<0)
		{f->flg|= _PWrErr; return;}
	if (a>p->k)
		p->k = a;
}

/* ---- top program */

#define _SyntaxCheck		0x8000
#define _FlatBank		0x4000
#define _TextCHR		0x2000
#define _DumpIDs		0x1000

typedef struct asm6502params
{
	asm6502frame	af;
	const char	*ofn, *pfn, *cfn;
	long		prz;
	int16		opt;
	uint8		prgram, prgwram, chrram, chrwram;
	uint8		submap;
} asm6502params;

static long processText(asm6502frame *af,
	FILE *f,
	const char *fn,
	long sz,
	uint8 bsz,
	int16 opt)
{
	int		r;

	if ((af->u=calloc(1,sizeof(bankData)))==NULL) /* keep extra byte */
		return _e_memory_error;

	/* at code point, initially */
	af->bsz = bsz;
	af->pos = af->flr;
	af->bnk = af->flg = 0, af->maxbnk = af->maxchr =-1;
	af->opt = _Prescan|opt;
	((bankData*)af->u)->k = 0; /* writing to that single byte */
	((bankData*)af->u)->bk = opt&_FlatBank ? ~0:(1u<<af->bsz)-1;
	af->writef = sz ? bwrite:bexpand;

	if ((r=X6502_AsmFile(f,fn,af,0))!=0)
		fprintf(stderr,"### Error(s) prevented further processing of '%s'.\n",
			fn);
	else
	{
		fseek(f,0,SEEK_SET);

		/* reset for full scan */
		X6502_FreeScan(af);
		af->org = 0, af->pos = af->flr, af->bnk = af->flg = 0;
		/* refresh all options - code may change _Optimize, anything else */
		af->opt = opt;
		if (af->map>=0)
		{
			if (!sz)
				sz = ((bankData*)af->u)->k+1;
			if (sz>16384)
				sz = (sz+16383)&~16383;
			else /* for PRG mirroring (repeated output in .nes) */
				sz = uppow2(sz);
		} else if (!sz)
		{
			if (!(opt&(_DataOnly|_QuietMode)))
				printf("# data size set to logical minimum\n");
			if ((sz=((bankData*)af->u)->k+1)<=0)
				sz = 1, fputs("## Warning: file wrote no PRG.\n",stderr);
		}
		if (af->maxbnk<0)
			af->maxbnk = (sz-1)>>af->bsz;

		((bankData*)af->u)->k = uppow2(sz)-1;
		if (opt&_SyntaxCheck)
		{
			free(af->u); af->u = NULL;
			af->writef = NULL;
		} else
		{
			bankData	*p;

			af->writef = bwrite; /* k (mask) is one less */
			if ((p=realloc(af->u,sizeof(bankData)
					+((bankData*)af->u)->k))==NULL)
				r = _e_memory_error;
			else
			{
				af->u = p;
				memset(&p->b,0,p->k+1);
			}
		}
		if (!r)
			r = X6502_AsmFile(f,fn,af,0);
	}
	if (opt&_DumpIDs)
	{
		int		x = X6502_AsmDumpIDs(stdout,af);

		printf("# %u IDs for '%s'\n",
			x,fn);
	}
	X6502_FreeFrame(af);
	return r ? (r>0 ? -1:r):sz;
}

static int writeOutput(asm6502params *ap)
{
	asm6502frame	ac;
	bankData	*p;
	FILE		*f, *cf = NULL;
	long int	b =-1;
	long		z;
	int		r = 0, x;

	if (!ap || (p=ap->af.u)==NULL)
		return _e_general_error;
	memset(&ac,0,sizeof(ac));

	if (ap->af.flg&_PrintNL)
		fputc('\n',stdout);
	if (ap->af.map>511)
	{
		fprintf(stderr,"### bad mapper index (%d)\n",
			ap->af.map);
		return 1;
	}
	if ((f=fopen(ap->ofn,"wb"))==NULL)
	{
		fprintf(stderr,"## writeOutput: could not open '%s' for writing.\n",
			ap->ofn);
		return 1;
	}
	z = (p->k+1)>>10;
	if (!(ap->opt&_QuietMode))
		if (ap->prz&1023)
			printf("# PRG total: %u bytes / %u KB\n",
				ap->prz,z);
		else
		{
			x = ap->af.map<0 ? 0:16384/ap->prz;
			printf(x>1 ? "# PRG total: %u KB / %u KB x%u\n"
				:"# PRG total: %u KB / %u KB\n",
				ap->prz>>10,z,x);
		}
	if (ap->af.map>=0)
	{ /* iNES */
		static char	h[16] = "NES\x1A";
		int		ver = 0, mir;
		uint8		wrn = 0;

		if (ap->cfn)
		{
			if ((cf=fopen(ap->cfn,ap->opt&_TextCHR ? "r":"rb"))==NULL)
			{
				fprintf(stderr,"## writeOutput: could not open '%s' for reading.\n",
					ap->cfn);
				r = 1;
				goto end;
			}
			if (ap->opt&_TextCHR)
			{
				ac.map =-1;
				b = processText(&ac,cf,ap->cfn,0,13,_DataOnly);
				fclose(cf), cf = NULL;
				if (b<0)
					{r = b; goto end;}
				/* Nestopia doesn't like truncated CHR data */
				b = (b+8191)&~8191;
			} else
			{
				fseek(cf,0,SEEK_END);
				b = (ftell(cf)+8191)&~8191;
				fseek(cf,0,SEEK_SET);
				
			}
		}
		x = b>>13;
		if (ap->af.maxchr>=0)
		{
			if (ap->af.maxchr!=MAX(x,0))/* && x>=0)*/
				wrn|= 2;
			x = ap->af.maxchr;
		}
		if (x<0)
			x = 0;
		else if (!(ap->opt&_QuietMode))
			printf("# CHR total: %u KB / %u KB%s\n",
				x<<13-10,b<0 ? 0:b>>10,wrn ? " (!)":"");
		if ((mir=ap->af.opt&_MirrorMask)==_MirrorMask)
			mir = 0;
		/* set version two if submapper, CHR WRAM or large output size(s) */
		/* CHR RAM is 16KB with mapper 13, 8KB otherwise */
		if ((z=ap->af.maxbnk+1)>255 || ap->af.map>255 || x>255
				|| ap->submap || ap->chrwram
				|| (ap->chrram && ((ap->af.map==13 && ap->chrram>14)
				|| (ap->af.map!=13 && ap->chrram>13))))
			ver = 2<<2;
		h[4] = z; /*PRG-low*/
		h[5] = x; /*CHR-low*/
		/* set the battery bit if PRG-WRAM */
		h[6] = ap->af.map<<4|(mir&1)|(2*!!ap->prgwram); /*map-low*/
		h[7] = (ap->af.map&0xF0)|ver|(mir<<2&8); /*map-high & version*/

		if (ver)
		{
			h[8] = (ap->af.map>>8&1)|ap->submap<<4;
			h[9] = (z>>8&0xF)|(x>>4&0xF0); /*PRG-high/CHR-high*/
			h[10] = (ap->prgram ? MAX(ap->prgram,7)-6:0)
				|(ap->prgwram ? MAX(ap->prgwram,7)-6:0)<<4;
			h[11] = (ap->chrram ? MAX(ap->chrram,7)-6:0)
				|(ap->chrwram ? MAX(ap->chrwram,7)-6:0)<<4;
		} else
		{ /* no CHR WRAM; CHR RAM is automatic, and not stored in the header */
			z = MAX(ap->prgram,ap->prgwram);
			h[8] = z ? 1u<<MAX(z,13)-13:0;
		}
		if (fwrite(h,1,16,f)!=16)
			r = _e_write_error;
		x = 16384;
	} else
		x = 0;
	z = 0; /* a loop to repeat output (PRG mirror) */
	do
		if (fwrite(p->b,1,ap->prz,f)!=ap->prz)
			{r = _e_write_error; goto end;}
	while((z+=ap->prz)<x);
	/* CHR data */
	if (cf) /* raw */
		r = copyFile(f,cf), fclose(cf);
	else if (ac.u)
	{
		if (fwrite(((bankData*)ac.u)->b,1,b,f)!=b)
			r = _e_write_error;
		free(ac.u);
	}
end:	fclose(f);
	if (r==_e_write_error)
	{
		fprintf(stderr,"### writeOutput: error writing to '%s'.\n",
			ap->ofn);
		r = 1;
	}
	return r;
}

static int asm6502(asm6502params *ap)
{
	FILE		*f;
	int		r;

	if (!ap)
		return _e_general_error;
	if ((f=fopen(ap->pfn,"r"))==NULL)
	{
		fprintf(stderr,"### could not open file '%s'\n",
			ap->pfn);
		return 1;
	}
	if (ap->prz<0)
		ap->prz = 16384; /* default */
	r = ap->prz = processText(&ap->af,f,ap->pfn,ap->prz,14,ap->opt);
	if (r>0)
		r = 0;
	if (f)
		fclose(f);
	if (ap->af.u)
	{
		if (!r)
		{
			if (!ap->ofn)
				ap->ofn = ap->af.map>=0 ? "asm-out.nes":"asm-out.bin";
			r = writeOutput(ap);
		}
		free(ap->af.u); ap->af.u = NULL;
	} else if (!r && !(ap->opt&_QuietMode))
		printf("# Processing successful.\n");
	return r==-1 ? 1:r;
}

/* returns largest negative on error */
static long int parseInt(const char *s)
{
	long int	t;

	return readanyinteger(s,strlen(s),&t) ? t:IMIN(long int);
}

static int parseParams(asm6502params *ap,
	int argc,
	char **argv)
{
	char		**a = argv;
	const char	*s;

	if (!ap)
		return _e_general_error;
	if (argc<=0)
		goto disp;
	ap->af.flr = 0x8000, ap->opt = _MirrorMask;
	ap->af.map =-1, ap->prz =-1;
	while(--argc>=0 && (s=*++a)!=NULL)
		if (*s=='-')
		{
			long int	t = 0, mx = 0;

			switch(s[1])
			{
			case 'B': /*ap->opt|= _AutoBank;*/ break;
			case 'T': ap->opt|= _TextCHR;
			case 'C': if (ap->cfn)
				{
					fprintf(stderr,"### %s - one leading CHR file at a time.\n",
						*argv);
					return 1;
				}
				ap->cfn = s+2;
				if (ap->af.map<0)
					ap->af.map = 0;
				break;
			case 'D': ap->opt|= _DumpIDs;
				break;
			case 'F': mx = 0xFFFF;
				ap->af.flr = t = parseInt(s+2);
				break;
			case 'K': ap->opt|= _SyntaxCheck; break;
			case 'L': ap->opt|= _FlatBank; break;
			case 'M': mx = 511;
				ap->af.map = t = parseInt(s+2);
				break;
			case 'N': ap->ofn = s+2; break;
			case 'O': ap->opt|= _Optimize; break;
			case 'P': if (s[2]=='R')
					{ap->chrram = t = parseInt(s+3), mx = 20; break;}
				else if (s[2]=='W')
					{ap->chrwram = t = parseInt(s+3), mx = 20; break;}
				mx = 16383, t = parseInt(s+3);
				t = parseInt(s+2), ap->prz = t<<10;
				break;
			case 'Q': ap->opt|= _QuietMode;
				break;
			case 'R': mx = 2;
				ap->opt&=~_MirrorMask, ap->opt|= t = parseInt(s+2);
				break;
			case 'S': mx = 15;
				ap->submap = t = parseInt(s+2);
				break;
			case 'X': mx = 20, t = parseInt(s+3);
				if (s[2]=='R')
					ap->chrram = t;
				else if (s[2]=='W')
					ap->chrwram = t;
				else
					t = IMIN(t);
				break;

			default:
				fprintf(stderr,"### %s - unknown/unsupported option '%s'.\n",
					*argv,s);
				return 1;
			}
			if ((unsigned)t>mx)
			{
				fprintf(stderr,"### %s - expected integer after %.2s, in [0,%u].\n",
					*argv,s,mx);
				return 1;
			}
		} else
		{
			if (ap->pfn)
			{
				fprintf(stderr,"### %s - one input at a time.\n",
					*argv);
				return 1;
			}
			ap->pfn = s;
		}
	if (!ap->pfn)
	{
		fputs("### no file input\n",stderr);
	disp:	fputs(
"#\n"
/*"#  -A           auto-mirror PRG (otherwise cleared)\n" -- auto for iNES*/
/*"#  -B           auto-bank (incr. on new ORGs)\n" -- always on now*/
"#  -C<name>     load CHR file\n"
"#  -D           dump IDs\n"
"#  -F<value>    code floor (default: $8000)\n"
"#  -K           check syntax only\n"
"#  -L           flat (unrestricted 64KB positioning)\n"
"#  -M<mapper>   (iNES output)\n"
"#  -N<name>     set output name\n"
"#  -O           turn optimization on\n"
"#  -P<value>    set output PRG KB size - 0=packed 32KB\n"
"#  -PR<value>   set PRG RAM - 1<<x bytes (iNES output)\n"
"#  -PW<value>   set PRG WRAM - 1<<x bytes (iNES output)\n"
"#  -Q           quiet mode (no printf)\n"
"#  -R<value>    set mirror (iNES output) - H=0, V=1, FS=2\n"
"#  -S<value>    set submapper (iNES output)\n"
"#  -T<name>     load asm text CHR file\n"
"#  -XR<value>   set CHR RAM - 1<<x bytes (iNES output)\n"
"#  -XW<value>   set CHR WRAM - 1<<x bytes (iNES output)\n"
"\n",
			stderr);
		return 1;
	}
	return 0;
}

int main(int argc,
	char **argv)
{
	asm6502params	ap;
	int		r;

	if (argc<=0 || !argv)
		return 0;
	if (!--argc || !argv[1])
		argc = 0, fprintf(stderr,"### usage: %s filename\n",
			*argv);
	memset(&ap,0,sizeof(asm6502params));
	if ((r=parseParams(&ap,argc,argv))==0)
		r = asm6502(&ap);
	if (r<0)
	{
		fprintf(stderr,"### %s - fatal error (%d).\n",
			*argv,r);
		if (r==_e_memory_error)
			fputs("### out of memory.",stderr);
		r = 1;
	}
	return r;
}
#endif
