/*
 utility/disasm6502.h

  Revised: 2015.07.27

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

enum
{ /* index displacement alters only the low byte */
	OPM_IM, /* 1-immediate		- #$xx  ;+0/2/9	*/
	OPM_RL, /* 1-relative - r=long	- *+$xx   ;+10	*/
	OPM_ZP, /* 1-zero page		- $xx	  ;+05	*/
	OPM_ZX, /* 1-zero page, X index	- $xx,X   ;+15	*/
	OPM_ZY, /* 1-zero page, Y index	- $xx,Y   ;+16	*/
	OPM_IX, /* 1-indirect, X index	- ($xx,X) ;+01	*/
	OPM_IY, /* 1-indirect, Y index	- ($xx),Y ;+11	*/
	OPM_IN, /* 1-indirect		- ($xx)   ;+12	*/
	OPM_IB, /* 2-absolute indirect	- ($xxxx) ;6C	*/
	OPM_AB, /* 2-absolute		- $xxxx   ;+0D	*/
	OPM_AX, /* 2-absolute, X index	- $xxxx,X ;+1E	*/
	OPM_AY, /* 2-absolute, Y index	- $xxxx,Y;BE/+19*/
	OPM_IA, /* 0-implied A to self		  ;+0A	*/
	OPM_IP, /* 0-implied - r=long		  ;+1A	*/
	OPM_IC, /* 0-constant		- ; $xx   ;02	*/
	/* 15 - 65816 modes - ZP above seconds as dp	*/
	OPM_SR, /* 1-stack relative	- $xx,S     ;+03*/
	OPM_SI, /* 1-stack rel. indir.	- ($xx,S),Y ;+13*/
	OPM_DL, /* 1-dp indirect long	- [$xx]     ;+07*/
	OPM_DI, /* 1-dp indir. long	- [$xx],Y   ;+17*/
	OPM_DK, /* 2-src/dst banks	- $xx,$xx ;44/54*/
	OPM_IL, /* 2-indir. long	- [$xxxx]   ;+1C*/
	OPM_LX, /* 3-long indexed	- $xxxxxx,X ;+1F*/

	OPM_COUNT /* 22 */
};

enum
{ /* op numbers, as used in the table */
	OPN_NOP, /* 0 - official no-op */
	OPN_BRK,
	OPN_JMP,
	OPN_JSR,
	OPN_RTI, /* 4 */
	OPN_RTS,
	OPN_PHA,
	OPN_PHP,
	OPN_PLA, /* 8 */
	OPN_PLP,
	OPN_BPL,
	OPN_BMI,
	OPN_BVC, /* 12 */
	OPN_BVS,
	OPN_BCC,
	OPN_BCS,
	OPN_BNE, /* 16 */
	OPN_BEQ,
	OPN_STX,
	OPN_STY,
	OPN_STA, /* 20 */
	OPN_TXS,
	OPN_TSX,
	OPN_TXA,
	OPN_TAX, /* 24 - 6502 P affected start */
	OPN_TYA,
	OPN_TAY,
	OPN_CPX,
	OPN_CPY, /* 28 */
	OPN_CMP,
	OPN_BIT,
	OPN_SEC,
	OPN_SEI, /* 32 */
	OPN_SED,
	OPN_CLV,
	OPN_CLC,
	OPN_CLI, /* 36 */
	OPN_CLD,
	OPN_LDX,
	OPN_LDY,
	OPN_LDA, /* 40 */
	OPN_EOR,
	OPN_ORA,
	OPN_AND,
	OPN_ADC, /* 44 */
	OPN_SBC,
	OPN_DEX,
	OPN_DEY,
	OPN_INX, /* 48 */
	OPN_INY,
	OPN_DEC,
	OPN_INC,
	OPN_ASL, /* 52 */
	OPN_LSR,
	OPN_ROL,
	OPN_ROR,
	OPN_6502_COUNT, /* 56 */
	/* 6510, P unaffected start */
	OPN_SHX		= OPN_6502_COUNT,
	OPN_SHY,
	OPN_SHA,
	OPN_SHS,
	OPN_SAX, /* 60 */
	OPN_SBX, /* 6510, P affected start */
	OPN_ARR,
	OPN_ASR,
	OPN_RLA, /* 64 */
	OPN_RRA,
	OPN_LAX,
	OPN_LAS,
	OPN_SLO, /* 68 */
	OPN_SRE,
	OPN_ANE,
	OPN_ANC,
	OPN_DCP, /* 72 */
	OPN_ISB,
	OPN_6510_COUNT, /* 74 - everything's defined */
	/* 65816 */
	OPN_BRA = OPN_6510_COUNT,
	OPN_STZ, /* store zero */
	OPN_PLX, /* 76 */
	OPN_PLY,
	OPN_PLD,
	OPN_PLB,
	OPN_PHX, /* 80 */
	OPN_PHY,
	OPN_PHD, /* push direct page reg. */
	OPN_PHB, /* push data bank reg. */
	OPN_PHK, /* 84 - push program bank reg. */
	OPN_PEA, /* push effective address */
	OPN_SWA, /* swap accumulator bytes */
	OPN_XCE, /* exchange carry/6502 emulation */
	OPN_TCD, /* 88 - trans. accum. to data reg. */
	OPN_TDC, /* trans. data reg. to accum. */
	OPN_TSB, /* status set bits */
	OPN_TRB, /* status reset bits */
	OPN_REP, /* 92 - reset (clear) P bits */
	OPN_SEP, /* set (clear) P bits */
	OPN_MVN, /* move block negative */
	OPN_MVP, /* move block positive */
	OPN_WAI, /* 96 - wait for interrupt */
	OPN_STP, /* stop */
	OPN_COP, /* co-processor enable - value pushed */
	OPN_WDM, /* reserved for expansion (illegal) */
	/* 68516 contains 6502 but not 6510, and vice versa */
	OPN_65xx_COUNT /* 100 */
};

/* x6510_opn_table entry: [read.1] [write.1] [flag.1] [bad.1] [opm.4] [oph.1] [opn.7] */
#define OPB_READ		0x8000
#define OPB_WRITE		0x4000
#define OPB_FLAG		0x2000	/* instruction affects P reg. */
#define OPB_BAD			0x1000	/* will crash a 6502/6510 */
#define OPB_HI			0x80	/* 65816-designated opcode */

extern const char	x65xx_opsizes[OPM_COUNT];
extern const char	x6510_opnametable[OPN_65xx_COUNT][3]; /* 228 bytes */
extern const uint16	x6510_opn_table[256]; /* 512 bytes */

#ifndef FP_FASTAPASS
 #define FP_FASTAPASS(x)
 #define FASTAPASS(x)
 #if defined(__GNUC__) && defined(C80x86)
  #undef FP_FASTAPASS
  #undef FASTAPASS
  #define FP_FASTAPASS(x)	__attribute__((regparm(x)))
  #define FASTAPASS(x)		__attribute__((regparm(x)))
 #elif MSVC
  #undef FASTAPASS
  #define FASTAPASS(x)		__fastcall
 #endif
 typedef uint8 (FP_FASTAPASS(1) *readfunc)(uint32 A);
#endif

/* NOTE: unlike FCEUI_Disassemble, fceuindbg is not affected (set yourself) */
uint32 X6502_Disassemble(const struct X6502struct *X, /* can be NULL (no peeking) */
	const readfunc *readers, /* must be defined (pass ARead) */
	uint32 addr, /* starting address to disassemble; next one returned */
	uint32 amask, /* for 16-bit addressing, pass 0xFFFF */
	int rdownshift, /* pass zero for one addr for each "reader" */
	char *strout); /* should be able to buffer >=34 bytes */

char *Disasm_HexStr(char *d,
	unsigned v,
	unsigned c);
