/*
 utility/endian.c
  Contains file I/O functions that write/read data LSB first.

 * MacFCEU 0.6a4
 *
 * Copyright notice for this file:
 *  endian.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "../types.h"
#include "endian.h"

void FlipByteOrder(uint8 *s,
	uint32 c)
{ /* some improvements, including pointer check */
	uint8		*e, t;

	if (s==0 || c==0 || c&0x80000001U)
		return; /* This shouldn't happen. */
	e = c+s;
	do
		t = *--e, *e = *s, *s++ = t;
	while((c-=2)!=0);
}

int write32le(uint32 v,
	FILE *fp)
{
	uint8		s[4];

	s[0] = v;
	s[1] = v>>=8;
	s[2] = v>>=8;
	s[3] = v>>=8;
	return fwrite(s,1,4,fp)<4 ? 0:4;
}

int read32le(uint32 *d,
	FILE *fp)
{
	uint8		buf[4];

	if (fread(buf,4,1,fp)<=0)
		return 0;
	*d = FCEU_de32lsb(buf);

	return 1;
}

void FCEU_en32lsb(uint8 *d,
	uint32 v)
{
	*d++ = v;
	*d++ = v>>=8;
	*d++ = v>>=8;
	*d = v>>=8;
}

uint32 FCEU_de32lsb(uint8 *s)
{
	return ((((uint32)s[3]<<8)+s[2]<<8)+s[1]<<8)+*s;
}
