/*
 evaluate.h
  the evaluation switch for interpreter v1.8.1
*/
#ifndef _OPC
 #define _OPC(m)		case _op_##m&0x3F:
#endif
#ifndef _OPD
 #define _OPD(m)		case(_op_##m&0x3F)|0x20:
#endif
/* unary: op b==>a */
_OPC(stv)
	if (!(o&0x80))
		goto end;
	/* set */
	s = unpack_eval(s,o,&a);
	/* b would hold reference for the destination but the
	 current implementation doesn't allow this
	*/
	return 0; /* <xstack */
_OPC(ldv) a = b; break;
_OPC(ldi) /* load indirect */
	if (geti)
		a = geti(b);
	break;
_OPC(srt) a = esqrt(b); break; /* square root */
_OPC(abs) a = b>=0 ? b:-b; break;
_OPC(neg) a =-b; break;
_OPC(not) a =-1-b; break; /* binary compliment */
_OPC(bno) a = !b; break; /* boolean NOT */
_OPC(dno) a = !!b; break; /* double- boolean NOT */
#ifdef ALLOW_BIN_HI_LO
_OPC(blo) t = b, a = t&0xFF; break;
_OPC(bhi) t = b, a = (t>>8)&0xFF; break;
#endif
/* dyadic */
_OPC(ban) a = a && b; break; /* bool. AND */
_OPC(bor) a = a || b; break; /* bool. OR */
_OPC(and) _OPD(and) t = b, a = (long)a&t; break; /* bin. AND */
_OPC(ior) _OPD(ior) t = b, a = (long)a|t; break; /* bin. OR */
_OPC(eor) _OPD(eor) t = b, a = (long)a^t; break; /* bin. Exclusive-OR */
_OPC(asr) t = b, a = (long)a>>t; break; /* arithmetic shift right */
_OPD(asr) t = a, a = (long)b>>t; break;
_OPC(shr) /* logical shift right */
#if EVAL_FLOATING
	a = scalb(a,-(long)b);
#else
	a = (unsigned)a>>(unsigned)b;
#endif
	break;
_OPD(shr)
#if EVAL_FLOATING
	a = scalb(b,-(long)a);
#else
	a = (unsigned)b>>(unsigned)a;
#endif
	break;
_OPC(shl) /* log. shift left */
#if EVAL_FLOATING
	a = scalb(a,(long)b);
#else
	a = (unsigned)a<<(unsigned)b;
#endif
	break;
_OPD(shl)
#if EVAL_FLOATING
	a = scalb(b,(long)a);
#else
	a = (unsigned)b<<(unsigned)a;
#endif
	break;
_OPC(add) a+= b; break;
_OPD(add) a = b+a; break; /* manual order (because of NaNs) */
_OPC(sub) a-= b; break; /* subtract */
_OPD(sub) a = b-a; break;
_OPC(cml) /* function call if not number */
_OPC(mul) a = (evalsing)a*(evalsing)b; break; /* multiply */
_OPD(mul) a = (evalsing)b*(evalsing)a; break;
_OPC(pow) a = epow(a,b); break; /* exponentiation */
_OPD(pow) a = epow(b,a); break;
_OPC(div) /* division */
#if EVAL_FLOATING
	a/= b;
#else
	a = !b ? -1u:(unsigned)a/(unsigned)b;
#endif
	break;
_OPD(div)
#if EVAL_FLOATING
	a = b/a;
#else
	a = !a ? -1u:(unsigned)b/(unsigned)a;
#endif
	break;
_OPC(dvs) /* signed div. */
	a = !b ? EVAL_HUGEVAL:(signed)(a/b);
	break;
_OPD(dvs)
	a = !a ? EVAL_HUGEVAL:(signed)(b/a);
	break;
_OPC(mod) /* modulus */
#if EVAL_FLOATING
	a = fmod(a,b);
#else
	a = !b ? 0:(unsigned)a%(unsigned)b;
#endif
	break;
_OPD(mod)
#if EVAL_FLOATING
	a = fmod(b,a);
#else
	a = !a ? 0:(unsigned)b%(unsigned)a;
#endif
	break;
_OPC(mds) t = b, a = !b ? 0:(signed)a%t; break; /* signed mod. */
_OPD(mds) t = a, a = !a ? 0:(signed)b%t; break;
_OPC(clt) a = a<b; break; /* lesser */
_OPC(cge) a = a>=b; break; /* greater or equal */
_OPC(cgt) a = a>b; break; /* greater */
_OPC(cle) a = a<=b; break; /* lesser or equal */
_OPC(ceq) a = a==b; break; /* equal */
_OPC(cne) a = a!=b; break; /* not equal */
_OPC(cid) a = !memcmp(&a,&b,sizeof(a)); break; /* identical */
_OPC(cni) a = !!memcmp(&a,&b,sizeof(a)); break; /* not identical */
