#if 0
C -o ini ini.o interpreter.o var.o -s Interpreter=Main -di interpreter.h -di ini.h -i :: -d mac -d EVAL_FLOATING -d ALLOW_RUNON_WSP -d EXAMPLE -d STRICT_ESCAPES {"Parameters"}
Exit 0
#endif
/*
 ini.c
  interpreter 1.9.3

 Created: 2013.03.25
 Revised: 2014.08.05

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
2013.08.02
  working prototype for escapes in names incompatible with correct form;
it has been routed out.
*/

#include "../enviro64.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if TARGET_OS_MAC
 #include <fp.h>	/* for nan */
#else
 #include <math.h>
#endif

#include "ini.h"

/* ---- code */

static int32 get_tcount(const char *s)
{
	const char	*e;
	int64		v;
	int32		t;

	if (*s=='\\')
		++s;
	e = s+strlen(s);
	if (readanyint64(s,e,&v)!=e || (t=int64toi(v))>0x7FFFFFF0u)
		t = _e_op_error;
	return t;
}

const tcon *gettcon(const tcon *l,
	const tobj *o,
	int32 *x)
{
	const char	*e;
	int32		z = o->z&0x7FFFFFFF;

	if (x)
		*x = 0;
	for(e=o->s+z; l->d; ++l)
		if (l->z && l->c && x)
		{
			const char	*s = o->s;
			int64		v;
			int32		nz = strlen(l->s);

			if (z<nz || memcmp(s,l->s,nz))
				continue;
			if ((s+=nz)!=e)
			{
				s = parse_skipwsp(s,e);
				if (s>=e || *s++!='#' || readanyint64(s,e,&v)!=e
						|| int64toi(v)>=l->c)
					continue;
				*x = int64lo(v);
			}
			return l;
		} else if (!memcmp(o->s,l->s,z) && !l->s[z])
			return l;
	return NULL;
}

int32 gettrec(const trec *l,
	const tobj *d,
	int32 *type)
{
	const char	*s, *e;
	int32		o, z = d->z&0x7FFFFFFF;

	if (!l || (s=l->s)==NULL || !*s)
		return _e_no_field;
	if (*s==';')
		++s;
	if (*s=='\\')
	{
		int64		v;
		int32		c;

		if ((c=get_tcount(s+1))<0)
			return c;
		s = d->s, e = s+z;
		if ((z=elem_tsize(l->t))<0)
			return z;
		if (readanyint64(s,e,&v)!=e)
			return -1;
		if (int64toi(v)>=c)
			return _e_no_field;
		o = z*int64lo(v)+l->o;
	} else
	{
		do
		{
			if (!l->t || (s=l->s)==NULL)
				return _e_no_field;
			if (*s==';')
				++s;
		} while(++l,memcmp(d->s,s,z) || s[z]);
		o = (--l)->o;
	}
	if (type)
		*type = l->t;
	return o;
}

/* ---- parsing */

const int8 *read_trec_be(const trec *r,
	void *d,
	const int8 *s)
{
	const char	*n;

	if (!r || !d)
		return 0;
	for(; r->t; ++r)
	{
		int8		*w;

		if ((n=r->s)==NULL)
			break;
		if (*n==';')
			if (n[1]=='\\' && n[2]!='\\')
				break;
			else
				continue;
		w = d, w+= r->o;
		if (*n=='\\')
		{
			int32		i, c, z;

			if ((z=elem_tsize(r->t))>0)
				for(i=0,c=get_tcount(n+1); i<c; w+=z,++i)
					s = read_typ_be(w,s,r->t);
			if (n[1]!='\\')
				break;
			continue;
		}
		s = read_typ_be(w,s,r->t);
	}
	return s;
}

void read_trec_be_ns(const trec *r,
	void *d,
	const int8 *_s)
{
	const char	*n;

	if (!r || !d)
		return;
	for(; r->t; ++r)
	{
		const int8	*s;
		int8		*w;

		if ((n=r->s)==NULL)
			break;
		if (*n==';')
			if (n[1]=='\\' && n[2]!='\\')
				break;
			else
				continue;
		w = d, w+= r->o, s = _s+r->o;
		if (*n=='\\')
		{
			int32		i, c, z;

			if ((z=elem_tsize(r->t))>0)
				for(i=0,c=get_tcount(n+1); i<c; w+=z,++i)
					s = read_typ_be(w,s,r->t);
			if (n[1]!='\\')
				break;
			continue;
		}
		read_typ_be(w,s,r->t);
	}
}

static const char *getcv(const char *s,
	const char *e,
	evalvalue *v, /* may be NULL */
	void *u)
{
	const tcon	*l;
	int32		z;

	if ((l=u)==NULL)
		return NULL;
	e = parse_wordbreak(s,e,NULL);
	for(z=e-s; l->s; ++l)
		if (!l->d && !memcmp(s,l->s,z) && !l->s[z])
		{
			*v = !l->z ? (signed)l->c:cvt_tdata(l->z,l->p);
			//printf("%s=%g\n",
			//	l->s,*v);
			return e;
		}
	return NULL;
}

int ini_parse_line(const char *str,
	const char *e,
	const tcon *l,
	ini_parse_state *st)
{
	const char	*s, *n;
	int32		t, r = 0;
	int8		lit;

	if ((s=str)==NULL || !st)
		return -1;
	if (s>=e || *s=='#') /* sol-comment; 2013.08.23 */
		return 0;
	t = st->ex;
	if (st->cm)
		goto comm;
start:	lit = 0;
	while(s<e)
		switch((uint8)*s) /* try not to get stuck */
		{
		case '/':
			++s;
			if (s>=e)
			{
				if (t==1) /* accept '/' as a param name */
					{--s; goto name;}
				goto err;
			}
			if (*s=='*')
			{ /* C-style comment */
				st->cm = 1;
			comm:	while(s<e)
					if (*s++=='*' && s<e && *s++=='/')
						{st->cm = 0; goto start;}
				goto end;
			} else if (*s!='/') /* C++-style comment */
				goto err;
		case ';': /* old-style comment */
			goto end;
		case '\\':
			if (++s>=e)
				goto err;
			if (*s-10u<=13u-10u)
				break;
			--s;
			goto name;
		case 9:
		case 10:
		case 11:
		case 12:
		case 13:
		case ' ':
			s = parse_skipwsp(s+1,e);
			break;
		case '=':
			if (t==1)
				goto name;
			if (t)
				goto err; 
			++t, ++s;
			st->px = 0;
			break;
		case ',':
			if (t!=1)
				goto err;
			++s;
			deltobj(st->eb+1);
			goto name;
		case '[':
			if (t==1)
				goto name;
			if (++s<e && *s=='=')
				{++s, t = 5; break;}
			if (t)
			{
				if (t>6u)
					goto end;
				goto err;
			}
			t = 2;
			break;
		case ']':
			if (t==1)
				goto name;
			if (t==6)
			{
				evalvalue	v = 1;
				int		err = 0;

				evaluate_line(st->eb->s,
					st->eb->s+(st->eb->z&0x7FFFFFFF),
					&v,&err,getcv,NULL,(void*)l,NULL);
				r = err;
				if (v)
					t = 3;
				deltobj(st->eb);
			} else if (t!=3)
				goto err;
			++t, ++s;
			break;
		case '.':
			if (!t)
				goto err;
		name:
		case '!':
		case '#':
		case '%':
		case '+':
		case '-':
		case '_':
		case '|':
		case '`':
		case ':':
		case '@':
		case '&':
		case '?':
		case '*':
		case '(':
		case ')':
		CASE_NUMERIC
		CASE_ALPHABETIC
			if (t>1u)
			{
				if (t<5u) /* condition */
				{
					if (t!=2)
						goto err;
					/* record name */
					st->cr = NULL;
				} else if (t>6u)
					goto end;
				s = parse_skipwsp(s,e);
				if (s>=e) /* lone string in .inf for t=0 */
					goto err;
				s = parse_entrybreak(st->eb->s=(char*)s,e,t);
				st->eb->z = (s-st->eb->s)|0x80000000u;
				s = parse_skipwsp(s,e);
				if (t!=5)
				{
					st->ix = 0;
					if ((st->cr=gettcon(l,st->eb,&st->cx))==NULL)
						{r = _e_no_identifier; goto end;}
					deltobj(st->eb); deltobj(st->eb+1);
					if (!(*st->cr).p)
						{r = _e_write_error; goto end;}
				}
				++t;
				break;
			}
			if (st->eb[t].s)
				goto err;
			if ((s=parse_entrybreak(st->eb[t].s=(char*)s,e,t))==NULL)
				goto err;
			st->eb[t].z = (s-st->eb[t].s)|0x80000000u;
			if (!t)
				break;
			goto obj;
		case '\047':
			lit = 1;
		case '"':
			if (t>1u)
			{
				if (t<7)
					goto err;
				goto end;
			}
			if (st->eb[t].s || !l || !l->p)
				goto err;
			n = s;
			s = parse_textloop(s+1,e,*s,&st->eb[t].z,NULL,&st->eb[t].s,0,NULL);
			st->eb[t].z&= 0x7FFFFFFF;
			if (e-n<2 || s[-1]!=*n) /* point of error */
			{
				if (t) /* 2013.07.15 */
					{r = _e_open_string; goto end;}
				goto err;
			}
			if (!t)
				break;
		obj:	if (st->cr)
			{
				const tcon	*lc = st->cr;

				if (st->eo<0)
					{r = _e_overlong_data; goto end;}
				++st->ix;
				if (!st->et && (st->eo=gettrec(lc->d,st->eb,&st->et))<0)
					{r = st->eo; goto end;}
				deltobj(st->eb);
				if ((t=set_tdata(&st->et,st->eb+1,
						(char*)lc->p+st->eo+st->cx*lc->z,lit,!st->px++))<0)
					{r = st->eo; goto end;}
				st->eo+= t;
				if (!st->et)
					 st->eo =-1;
			}
			t = 1;
			break;

		default:
		err:	r = s-str+1;
			goto end;
		}
end:	if (st->cm)
	{ /* allocate any ref pointers that may become stale */
		int32		r2;

		if ((r2=unrtobj(st->eb))<0)
			r = r2;
		if ((r2=unrtobj(st->eb+1))<0)
			r = r2;
	} else
	{ /* end of actual line - reset fields */
		if (!r && st->eb->s && t<4u)
			r = st->cr ? _e_incomplete_error:_e_handler_error;
		st->eo = st->et = 0;
		if (t<7u)
			t = 0, st->px = 0;
		deltobj(st->eb); deltobj(st->eb+1);
	}
	st->ex = t;
	return r;
}

static void print_err(const char *n,
	long m,
	long r,
	uint32 l,
	long ix)
{
	int8		f;
	const char	*s;

	if (!m)
	{
		if (ix>0)
			fprintf(stdout,"## More than %d errors.  Take care of yourself.\n",
				ix);
		return;
	}
	if (!n)
		n = "";
	f = *n!=':' || n[1];
	switch(r)
	{
	case _e_op_error: s = "bad record data"; break;
	case _e_range_error: s = "literal too large"; break;
	case _e_open_string: s = "ill-formed string"; break;
	case _e_no_field: s = "no field"; break;
	case _e_no_identifier: s = "no record"; break;
	case _e_write_error: s = "NULL destination"; break;
	case _e_overlong_data: s = "overlong data"; break;
	case _e_incomplete_error: s = "incomplete param"; break;
	case _e_handler_error: s = "entry outside record"; break;

	default: s = "syntax/param error";
	}
	fprintf(m>0 ? stdout:stderr,r>0 ? "%s%s%sline %d; #%d: %s @ %d\n"
		:"%s%s%sline %d; #%d: %s\n",
		*n ? f ? "## ini parse error\nfile '":"":"## ini parse error - ",
		f ? n:"",*n && f ? "'; ":"",l,ix,s,r);
}

int32 fread_ini(FILE *f,
	const char *n,
	const tcon *cl,
	int32 k)
{
	ini_parse_state	st;
	char		*s;
	uint32		l, x = 1, z, ec = 101;
	int32		r, p, p2;

	if (!cl)
		return _e_read_error;
	s = malloc(0); /* needed for first-empty line(s) case */
	memset(&st,0,sizeof(st));
	for(p=p2=-1; z=~0,l=x,f && (s=freadline(f,&z,s,NULL,&x,'\\'))!=NULL; p=r)
	{
		const char	*e = s+z;

		if ((r=ini_parse_line(s,e,cl,&st))!=0
				&& 1u<<~(r<0 ? r:-1)&k)
		{
			if (r!=_e_handler_error || p!=r)
				print_err(n,k,r,l,st.ix);
			else if (p!=p2)
				fputs("...\n",k>0 ? stdout:stderr);
			n = ":", p2 = p;
			if (!--ec)
				{print_err(0,0,0,0,100); z = 0; break;}
		}
	}
	deltobj(st.eb), deltobj(st.eb+1);
	free(s);

	return z ? _e_memory_error:MIN(l,IMAX(int32));
}

int32 sread_ini(const char *s,
	const char *e,
	const char *n,
	const tcon *cl,
	int32 k)
{
	ini_parse_state	st;
	uint32		l, x = 1, z, ec = 101;
	int32		r, p, p2;

	if (!s || s>=e)
		return 0;
	if (!cl)
		return _e_read_error;
	memset(&st,0,sizeof(st));
	p = p2 =-1;
	do
	{
		const char	*le;

		l = x, z = sreadline(s,e-s,&x,'\\'), le = s+z;
		if ((r=ini_parse_line(s,le,cl,&st))!=0
				&& 1u<<~(r<0 ? r:-1)&k)
		{
			if (r!=_e_handler_error || p!=r)
				print_err(n,k,r,l,st.ix);
			else if (p!=p2)
				fputs("...\n",k>0 ? stdout:stderr);
			n = ":", p2 = p;
			if (!--ec)
				{print_err(0,0,0,0,100); break;}
		}
	} while(p=r,(s+=z+1)<e); /* skip eol */
	deltobj(st.eb), deltobj(st.eb+1);
	return MIN(l,IMAX(int32));
}

/* ---- printing */

int32 print_trec(FILE *f,
	const trec *r,
	const void *d)
{
	const char	*s;
	int32		t;

	if (!r || !d)
		return 0;
	for(t=0; r->t; ++r)
	{
		const int8	*p = d;

		if ((s=r->s)==NULL)
			break;
		if (*s==';')
			if (s[1]=='\\')
				break;
			else
				continue;
		p+= r->o;
		if (*s=='\\')
		{
			int32		i, c, z;

			if ((z=elem_tsize(r->t))<=0)
				break;
			for(i=0,c=get_tcount(s+1); i<c; p+=z,++i)
			{
				t+= fprintf(f,"%u=",
					i);
				t+= print_tdata(f,r->t,p,1), t+= fputc('\n',f)>0;
			}
			break;
		}
		t+= print_string(f,s,strlen(s),0,1);
		t+= fputc('=',f)>0;
		t+= print_tdata(f,r->t,p,1), t+= fputc('\n',f)>0;
	}
	return t;
}

int32 print_tcon(FILE *f,
	const tcon *l)
{
	const int8	*p;
	int32		t, x;

	if (!l->d || !l->s)
		return 0;
	t = x = 0, p = l->p;
	do
	{
		t+= fprintf(f,l->z && l->c ? "\n[%s #%d]\n":"\n[%s]\n",
			l->s,x);
		if (!p)
			{t+= fputs("; NULL ptr\n",f); break;}
		t+= print_trec(f,l->d,p);
	} while(p+=l->z,++x<l->c);

	return t;
}

int32 dump_ini(FILE *f,
	const tcon *l)
{
	int32		t = 0;

	for(; l->d; ++l)
		t+= print_tcon(f,l);
	return t;
}

#ifdef TESTRBE
/* this all works */
#define _angle		_int16
typedef int16		angle, world_distance;

typedef struct world_point3d
{
	int16		x, y, z;
} world_point3d;

typedef struct map_object/*(16)*/
{
	int16		type; /* _saved_monster, _saved_object, _saved_item, ... */
	int16		index;
	angle		facing;
	int16		polygon_index;
	world_point3d	location; /* z is a starting delta from floor */

	uint16		flags;
} saved_object;

#pragma warn(disable:2237)
static int8	storedobj[16] = "\0\6\0\10\1\0\1\36\120\0\100\0\377\0\0\1";

static trec	savedObject_trec[] =
{
	O_TREC(_int16,saved_object,type),
	O_TREC(_int16,saved_object,index),
	O_TREC(_angle,saved_object,facing),
	O_TREC(_int16,saved_object,polygon_index),
	O_TREC(TYP_TRIPLET(_ifrc16,_ifrc16,_ifrc16),saved_object,location),
	O_TREC(_uint16,saved_object,flags),
	{0}
};

int main(int argc,
	char **argv)
{
	saved_object	obj;

	if (argc<=0)
		return 0;
	printf("%d MSB parsed\n\n",
		read_trec_be(savedObject_trec,&obj,storedobj)-storedobj);
	print_trec(stdout,savedObject_trec,&obj);
}
#elif defined(EXAMPLE)
typedef struct example_rec
{
	double		doub;
	float		sing;
	int64		llng;
	ifract		ifrc;
	ifixed		ifix;
	int32		lint, tk32, recl[4];
	int16		shrt;
	int8		ctyp;
	uint8		ucfr, ufl8, argb[4], rgb[3];
	uint32		ufrc;
	uint32		ufix;
	uint32		uint;
	uint32		ulfr;
	uint16		usfr;
	uint16		ushr;
	uint8		uchr;
	char		bool;
	uint8		Str7[8];
	char		str8[8];
	int16		ix16, fr16, fo16, fl16, rect[4];
	uint16		recu[4];
	int8		font[80];
	char		esc1[12], esc2[8], esc3[8], esc4[8], esc5[8];
	float		rcfl[4];
} example_rec;

example_rec	example[1];

/* to pick up the virtual space in-string */
#define rec
#define type
#define hex

static trec example_context[] =
{
	P_TREC(example_rec, Str7),
	C_TREC(example_rec, str8),
	O_TREC(_double, example_rec, doub),
	O_TREC(_float, example_rec, sing),
	O_TREC(_int64, example_rec, llng),
	O_TREC(_ifract, example_rec, ifrc),
	O_TREC(_ifixed, example_rec, ifix),
	O_TREC(_int32, example_rec, lint),
	O_TREC(_int16, example_rec, shrt),
	{_int8, OFFSET(example_rec,ctyp), "char type"},
	O_TREC(_ufract, example_rec, ufrc),
	O_TREC(_ufixed, example_rec, ufix),
	O_TREC(_ufrc16, example_rec, usfr),
	O_TREC(_ufrc32, example_rec, ulfr),
	O_TREC(_uint32, example_rec, uint),
	O_TREC(_uint16, example_rec, ushr),
	O_TREC(_uint8, example_rec, uchr),
	O_TREC(_bool, example_rec, bool type),
	O_TREC(_ufrc8, example_rec, ucfr),
	O_TREC(_hint32, example_rec, ufrc hex),
	O_TREC(_flt16, example_rec, fl16),
	O_TREC(_rect16, example_rec, rect),
	O_TREC(TYP_QUADRUPLET(_uint16,_uint16,_uint16,_uint16), example_rec, recu),
	O_TREC(_ifix16, example_rec, ix16),
	O_TREC(_ifrc16, example_rec, fr16),
	O_TREC(_fnts80, example_rec, font),
	O_TREC(_argb32, example_rec, argb),
	O_TREC(_rgb24, example_rec, rgb),
	O_TREC(_itck32, example_rec, tk32),
	O_TREC(_uflt8, example_rec, ufl8),
	C_TREC(example_rec, esc1),
	C_TREC(example_rec, esc2),
	C_TREC(example_rec, esc3),
	C_TREC(example_rec, esc4),
	C_TREC(example_rec, esc5),
	{0},
};

static tcon context_list[] =
{
	A_TCON(example_context, example rec),
	{NULL, "version", NULL, 0, INTERP_VERS},
	{0}
};

#undef rec
#undef type
#undef hex

#define USE_SREAD 1

int main(int argc,
	char **argv)
{
	FILE		*f = NULL;
	char		**a, *s, *n;
	int32		r = 0;

	if (argc<=0 || (a=argv)==NULL)
		return 0;
	if (!--argc || (s=*++a)==NULL)
	{
		fprintf(stderr,"### usage: %s filename\n",
			*argv);
		fputs(
"#\n"
"#  -            use stdin\n"
"#\n",
			stderr);
		return 1;
	}

	if (*s=='-')
	{
		if (!s[1])
			f = stdin, s = "dev::stdin";
		else
		{
			fprintf(stderr,"### %s - unknown option '%s'.\n",
				*argv,s);
			return 1;
		}
	} else
		f = fopen(s,"r");
	if (!f)
	{
		fprintf(stderr,"### %s - could not get/open '%s'.\n",
			*argv,s);
		return 1;
	}
#if !TARGET_API_MAC_CARBON && TARGET_OS_MAC
	if (((volatile char*)0x174)[60>>3]&(0x80>>(60&7))) /* keymap - control */
		Debugger();
#endif
	n = s;
#if USE_SREAD
	if ((s=freadfile(f,&r,NULL))==NULL)
		r = r ? _e_memory_error:0;
	else
	{ /* a last empty line not counted */
		if (f!=stdin)
			fclose(f);
		r = sread_ini(s,s+r,n,context_list,~0);
		free(s);
	}
#else
	r = fread_ini(f,n,context_list,~0);
	if (f!=stdin)
		fclose(f);
#endif
	if (r<0)
		fprintf(stderr,"### %s - fatal error (%d).\n",
			*argv,r), r = 1;
	else
	{
		printf("%u lines parsed\n",
			r);
		dump_ini(stdout,context_list);
		r = 0;
	}
	return r;
}
#endif
