/*
 ini.h
  interpreter 1.9.3

 Revised: 2014.08.02
*/

#include "interpreter.h"

/* ---- typedefs */

typedef int32		trec_type;
typedef uint32		trec_offs;

/* record object */
typedef struct trec
{
	trec_type	t;
	trec_offs	o;
	char		*s;	/* start w. '\' for array, ';' for "don't print" */
} trec;

/* normal offset record macro */
#define O_TREC(t,s,n)			{t, (int)&((s*)0)->n, #n}
/* C and Pascal string record macros */
#define C_TREC(s,n)			{sizeof(((s*)0)->n), (int)&((s*)0)->n, #n}
#define P_TREC(s,n)			{1-sizeof(((s*)0)->n), (int)&((s*)0)->n, #n}

/* context structure */
typedef struct tcon
{
	const trec	*d;
	char		*s;
	void		*p;
	trec_type	z, c;	/* both non-zero for tcon array (xxx #n) */
} tcon;

#define O_TCON(d,s)			{d,#s,&s}
#define A_TCON(d,s)			{d,#s,&s,sizeof(*s),GET_ARRAY_COUNT(s)}
#define A_NCON(d,n,s)			{d,#n,s,sizeof(*s),GET_ARRAY_COUNT(s)}

/* zero this structure out before using */
typedef struct ini_parse_state
{
	tobj		eb[2]; /* entry name and data buffer */
	const tcon	*cr; /* current record tcon */
	int32		cx, eo, et; /* cr index, entry data offset & type */
	uint16		px; /* parameter index */
	unsigned char	ex; /* entry index (0/1=entry/data, 2..4=record) */
	unsigned char	cm; /* long-comment state */
	int32		ix; /* record entry counter */
} ini_parse_state;

/* ---- prototypes */

const tcon *gettcon(const tcon *l,
	const tobj *o,
	int32 *x); /* if !=NULL, allow tcon arrays, get index */
/* returns offset or error (negative) */
int32 gettrec(const trec *l,
	const tobj *d,
	int32 *type);

int32 print_trec(struct FILE *f,
	const trec *r,
	const void *p);
int32 print_tcon(struct FILE *f,
	const tcon *l);

/* be sure to call deltobj for eb[0] and eb[1] when finished */
int ini_parse_line(const char *str,
	const char *e,
	const tcon *l,
	ini_parse_state *st);

/* returns line count or error code (negative) */
/* `m` = error bit mask; bit #31 = print errors to stderr, otherwise stdout */
int32 fread_ini(struct FILE *f,
	const char *name,
	const tcon *cl,
	int32 m);
int32 sread_ini(const char *s,
	const char *e,
	const char *name,
	const tcon *cl,
	int32 m);
/* returns char count printed */
int32 dump_ini(struct FILE *f,
	const tcon *l);

/* reads big endian binary data into sequential fields, ignoring commented ones */
/* returns the incremented byte stream pointer */
const int8 *read_trec_be(const trec *r,
	void *d,
	const int8 *s);
/* this copies/processes data non-sequentially */
void read_trec_be_ns(const trec *r,
	void *d,
	const int8 *s);
