/*
 video.c

 * MacFCEU 0.8.5
 *
 * Copyright notice for this file:
 *  video.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <zlib.h>

#include "driver.h"
#include "FCEU.h"
#include "NSF.h"
#include "VSUni.h"
#include "video.h"
#include "general.h"
#include "memory.h"
#include "state.h"
#include "movie.h"
#include "palette.h"
#include "input.h"

#ifdef mac
 #include "../zlib/contrib/macos/macos7-9.h"
#elif defined(unix)
 #include <unistd.h>
 #include <utime.h>
#else
 #include <direct.h>
 #include <io.h>
#endif

#ifdef env68k
 #pragma segment shell
#endif

//#define SHOWFPS
#define SNAP_IMAGE_WIDTH	256

/* ---- initialization */

uint8		*XBuf, *FCEU_video;
static uint8	*xbsave;

void FCEU_KillVirtualVideo(void)
{
	if (xbsave)
		{FCEU_free(xbsave); xbsave = 0;}
}

int FCEU_InitVirtualVideo(void)
{
	if (!XBuf) /* Some driver code may allocate XBuf externally. */
	{ /* 256 bytes per scanline, * 256 scanline maximum, one extra for alignment */
		if ((XBuf=(uint8*)FCEU_malloc(256*FCEUSCRN_ROWSIZE + FCEUSCRN_ROWSIZE))==0)
			return 0;
	}

	xbsave = XBuf;

	XBuf = FCEU_video = (uint8*)(~511&(511+(long int)XBuf));
	dwmemset(XBuf,0x80808080,256*FCEUSCRN_ROWSIZE); //*240);

	return 1;
}

uint8		FCEUvideoShot = 0;

static long	howlong;
static char	msgbuf[72]; /* was 65 */

#include "drawing.h"

/* ---- counters/fps */

uint64		videofps_boop[60];
static unsigned	boopcount = 0;
uint8		show_fps, show_fct;

#include "ftrig.h"

#define OneTenth		(4294967295.0+10)/10

static void DrawCounters(uint32 framecnt,
	uint8 render)
{
	static char	buf[32], *lastd = 0;
	char		*d;
	uint32		num, quo, y;
#ifdef mac
 #define booplimit 10
 #define FCEUD_GetTimeFreq()	1000000 /* always in microseconds */
#else
	// It's not averaging FPS over exactly 1 second, but it's close enough.
	unsigned	booplimit = PAL ? 50:60;
#endif
	if (!show_fps)
		_int64xset(videofps_boop[0],0);
	else if ((unsigned)(y=FSettings.FirstSLine)<240)
	{
		uint64		da = FCEUD_GetTime();

		y+= 3;
		(void)uint64sub(da,videofps_boop[boopcount]);
		videofps_boop[boopcount] = FCEUD_GetTime();
		if (!render)
			goto skip;
		d = buf+15, *d = 0;
		if (framecnt&1 && lastd) /* skip calculation on odd frames; 2010.04.26 */
			{d = lastd; goto drawfps;}
		if (int64z(videofps_boop[0]) || int64z(videofps_boop[9])
				|| int64hi(da)!=0 || !int64lo(da)
#ifdef FCEUD_GetTimeFreq
				|| (num=(FCEUD_GetTimeFreq()*20*booplimit)/int64lo(da)+1>>1)==0)
#else
				|| (num=umuldiv(FCEUD_GetTimeFreq(),20*booplimit,
					int64lo(da))+1>>1)==0)
#endif
			*--d = '-', *--d = '-';
		else
		{
			quo = ufract32mul(num,OneTenth);
			*--d = '0'+num-quo*10;
			*--d = '.';
			num = quo;
			do
				quo = ufract32mul(num,OneTenth), *--d = '0'+num-quo*10;
			while((num=quo)!=0 && d>buf);
		}
		lastd = d;
	drawfps:
		DrawTextTrans(XBuf+y*FCEUSCRN_ROWSIZE+240-8*(buf-d+15),
			FCEUSCRN_ROWSIZE,(uint8*)d,PAL_DRED);
	skip:
		boopcount = (boopcount+1)%booplimit;
	}
	if (show_fct && render && (unsigned)(y=FSettings.LastSLine-(8+3))<240)
	{ /* frame count -- added; 2010.04.14 */
		d = buf+sizeof(buf)-1, *d = 0;
		num = framecnt;
		quo = num/10, *--d = '0'+num-quo*10; /* resolves OneTenth inaccuracy */
		if ((num=quo)!=0)
			do
				quo = ufract32mul(num,OneTenth), *--d = '0'+num-quo*10;
			while((num=quo)!=0 && d>buf);
		DrawTextTrans(XBuf+y*FCEUSCRN_ROWSIZE+240-8*(buf+sizeof(buf)-1-d),
			FCEUSCRN_ROWSIZE,(uint8*)d,PAL_DRED);
	}
}

static void ReallySnap(void);

void FCEU_PutImage(uint32 framecnt,
	uint8 skip)
{
	if (FCEUGameInfo->type==GIT_NSF)
	{
		DrawNSF(skip ? 0:XBuf);
		DrawCounters(framecnt,!skip);
	/* Save snapshot after NSF screen is drawn.  Why would we want to do it before? */
		if (skip)
			return;
		if (FCEUvideoShot)
			ReallySnap();
	} else
	{
		DrawCounters(framecnt,!skip);
	/* Save snapshot before overlay stuff is written. */
		if (skip)
			return;
		if (FCEUvideoShot)
			ReallySnap();
		if (FCEUGameInfo->type==GIT_VSUNI)
			FCEU_VSUniDraw(XBuf,PAL_WHITE);
		FCEU_DrawSaveStates(XBuf);
		FCEU_DrawMovies(XBuf);
		FCEU_DrawNTSCControlBars(XBuf);
	}

	DrawMessage();
	FCEU_DrawInput(XBuf);
}

/* ---- messages */

void FCEU_DispMessage(const char *format, ...)
{
	va_list		ap;

	va_start(ap,format);
	vsprintf(msgbuf,format,ap);
	va_end(ap);

	howlong = 180; /* about 3 seconds NTSC or 3.6 seconds PAL (at 100% speed) */
}

void FCEU_ResetMessages(int on)
{
	if (on<0) /* toggle */
		on = howlong<=0 ? 180:0;
	if ((howlong=on)==0)
	{ /* this added; 2010.04.25 */
		FCEU_DrawSaveStates(0);
		FCEU_DrawMovies(0);
		FCEU_DrawNTSCControlBars(0);
	}
}

/* ---- snapshots */

static char	*pngfp = 0;
static uint8	*compmem = 0;
static uint32	lastu = 0;
static uLongf	compmaxsize = 0;

void FCEUI_SaveSnapshot(void)
{
	FCEUvideoShot = 1; /* saved on next frame */
}

void FCEU_ResetSnaps(void)
{
	if (compmem)
		{FCEU_free(compmem); compmem = 0;}
	if (pngfp)
		{free(pngfp); pngfp = 0;}
	FCEUvideoShot = 0; /*..maybe fix?; 2010.05.19 */
}

static void ReallySnap(void)
{
	uint8		lsts = FCEUvideoShot; /*SaveSnapshot-cleared; 2011.02.06*/
	int		r = SaveSnapshot();

	if (lsts==1)
		if (r>0)
			FCEU_DispMessage("Screen snapshot %d saved.",
				r-1);
		else
			FCEU_DispMessage("Error saving screen snapshot.");
}

static int WritePNGChunk(FILE *fp,
	uint32 size,
	char *type,
	uint8 *data)
{
	uint8		tempo[4];
	uint32		crc;

	tempo[0] = size>>24;
	tempo[1] = size>>16;
	tempo[2] = size>>8;
	tempo[3] = size;

	if (fwrite(tempo,4,1,fp)!=1)
		return 0;
	if (fwrite(type,4,1,fp)!=1)
		return 0;

	if (size!=0)
		if (fwrite(data,1,size,fp)!=size)
			return 0;

	crc = FCEUI_CRC32(0,(uint8*)type,4);
	if (size!=0)
		crc = FCEUI_CRC32(crc,data,size);

	tempo[3] = crc;
	tempo[2] = crc>>=8;
	tempo[1] = crc>>=8;
	tempo[0] = crc>>=8;

	return fwrite(tempo,4,1,fp)!=1 ? 0:1;
}

static int WritePNGHeader(FILE *pp,
	int width,
	int height,
	int bitdepth)
{
	static uint8	header[8] = {137, 80, 78, 71, 13, 10, 26, 10};
	uint8		chunko[13];

	if (fwrite(header,8,1,pp)!=1)
		return 0;

	chunko[0] = chunko[1] = 0;
	chunko[2] = width>>8&0xFF;
	chunko[3] = width&0xFF;

	chunko[4] = chunko[5] = 0;
	chunko[6] = height>>8&0xFF;
	chunko[7] = height&0xFF;

	chunko[8] = bitdepth;
	chunko[9] = bitdepth<=8 ? 3:0; // color type (indexed or direct?)
	chunko[10] = 0;	// deflate method of compression
	chunko[11] = 0;	// basic adapative filter set (though none are used)
	chunko[12] = 0;	// not interlaced

	return WritePNGChunk(pp,13,"IHDR",chunko);
}

static int WritePNGPalette(FILE *pp)
{
	uint8		pdata[256*3], *pptr = pdata;
	int		x;

	x = 0;
	do
		FCEUD_GetPalette(x,pptr,1+pptr,2+pptr);
	while(pptr+=3,++x<256);

	return WritePNGChunk(pp,256*3,"PLTE",pdata);
}

/* important: starts with filter code byte (expects (1+rowsize)*height) */
static int GetPNGImage(void *pixels,
	int rowsize,
	int width,
	int height)
{
	const uint8	*a = XBuf+FSettings.FirstSLine*rowsize;
	uint8		*d = pixels;
	int		x, y = height;

	if ((x=width)<=0 || y<=0 || rowsize<=0)
		return (x|y)<0 ? -1:0;
	do
	{
		const uint8	*s = a;

		*d++ = 0; // no filter (initially uncompressed)
		if (width&1)
			*d++ = *s++;
		x = width>>2; /* unaligned, four at a time */
		if (width&2)
			goto two;
		do
		{
			*d++ = *s++, *d++ = *s++;
		two:	*d++ = *s++, *d++ = *s++;
		} while(--x);
	} while(a+=rowsize,--y);

	return 1;
}

#if defined(WIN32) || !defined(unix)
 #define mymkdir(x)		mkdir(x)
#else
 #define mymkdir(x)		mkdir(x,0775)
#endif

/* success only when result>0 */
int SaveSnapshot(void)
{
	int		linecount = 1+FSettings.LastSLine-FSettings.FirstSLine;
	int		ix = lastu, dirc = FCEUvideoShot>1 && !pngfp;
	FILE		*pp = 0;
	uLongf		maxsize = linecount*263+12, compmemsize = maxsize;

	if ((!compmem || compmaxsize!=compmemsize)
			&& (compmem=(uint8*)FCEU_malloc(compmaxsize=compmemsize))==0)
		return 0;
	if (!pngfp)
		for(; ix<32767; ++ix) /* was 99999 */
		{ /* find an available file name */
			if ((pngfp=FCEU_MakeFName(FCEUMKF_SNAP,ix,
					dirc ? 0:"png"))==0)
				return -1;
			if (FCEUvideoShot>1 || !CheckFileOrDirExists(pngfp))
				break;
		}
	if (!pngfp)
		return 0;
	if (FCEUvideoShot>1)
	{ /* I made a mess, indeed; 2010.05.16 */
		long	z = strlen(pngfp);
		char	*s = z+pngfp;

		if (dirc)
		{
			/* general.c allocates at least 2048 bytes */
			if (z>2047-10 || mkdir(pngfp)<=0)
				return 0;
			strcpy(s,PSS "00000.png");
			s+= 6;
		} else
			s-= 4;
		z = 5;
		do /* ASCII increment */
			if (++(*--s)>'9')
				*s = '0';
			else
				break;
		while(--z!=0);
	}
	lastu = ix;

	pp = FCEUD_UTF8fopen(pngfp,"wb");
	if (FCEUvideoShot<=1)
		{free(pngfp); pngfp = 0;}
	if (!pp)
		return 0;
//	rewind(pp);
	if (WritePNGHeader(pp,SNAP_IMAGE_WIDTH,linecount,8)<=0)
		goto PNGerr;

	if (WritePNGPalette(pp)<=0)
		goto PNGerr;

	{
		uint8		*pixels;

		if ((pixels=(uint8*)FCEU_malloc(linecount*SNAP_IMAGE_WIDTH+linecount))==0)
			goto PNGerr;

		GetPNGImage(pixels,FCEUSCRN_ROWSIZE,SNAP_IMAGE_WIDTH,linecount);

		if (compress(compmem,&compmemsize,pixels,
				linecount*SNAP_IMAGE_WIDTH+linecount)!=Z_OK)
		{
			if (pixels!=0)
				free(pixels);
			goto PNGerr;
		}

		if (pixels)
			FCEU_free(pixels);
		if (!WritePNGChunk(pp,compmemsize,"IDAT",compmem))
			goto PNGerr;
	}

	if (!WritePNGChunk(pp,0,"IEND",0))
	PNGerr: ix = -1;

	if (FCEUvideoShot<=1)
		FCEU_ResetSnaps();
	if (pp!=0)
		fclose(pp);

	return ix+1;
}
