/*
 x6502.h
  used by the 2A0x functions

 * MacFCEU 0.8a7
 *
 * Copyright notice for this file:
 *  x6502.h, x6502struct.h Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _X6502H

#include "6502struct.h"

/* ---- constants */

#define FCEU_IQEXT		0x1
#define FCEU_IQFDS		0x2	/* prev. IQEXT2 */
/* ... */
#define FCEU_IQRESET		0x20
#define FCEU_IQNMI2		0x40	/* Delayed NMI, gets converted to *_IQNMI */
#define FCEU_IQNMI		0x80
#define FCEU_IQDMC		0x100
#define FCEU_IQFCOUNT		0x200
#define FCEU_IQTEMP		0x800
/* ... */
#define FCEU_JAMMED		0x8000

/* ---- globals */

extern uint32 timestamp;

extern X6502U X6502F;
#define XFrame X6502F.low

extern void FP_FASTAPASS(1) (*MapIRQHook)(
	int a); /* instruction cycle count */

extern void FP_FASTAPASS(1) (*DebugStepHook)(
	int c); /* count (used by the debugger - pass zero) */

extern void FP_FASTAPASS(1) (*X6502_Run)(
	int32 cycles); /* PPU cycle count */

/* ---- prototypes */

/* positive to OR, eitherwise AND (~x) */
void X6502_DebugLevel(int bits); /* 1 is user, 2 is cheat.c, 4 is GameGenie */

void X6502_Init(void);
void X6502_Reset(void);
void X6502_Power(void);

uint8 FASTAPASS(1) X6502_DMR(uint32 A);
void FASTAPASS(2) X6502_DMW(uint32 A,
	uint8 V);
void FASTAPASS(1) X6502_IRQBegin(int w);
void FASTAPASS(1) X6502_IRQEnd(int w);

#define _X6502H
#endif
