/*
 macos7-9.h
  cross-platform routines include for MacOS 7.x - 9.x

  Created: 2008.07.04 17:54 EST
  Revised: 2014.04.02
*/

#ifndef __MACOS7_9_H
 #define __MACOS7_9_H

 #ifndef __MACTYPES__
  #include <MacTypes.h>
 #endif

 #include <string.h>
 #include <time.h>

/* ---- constants */

 #define vRefNumLocal			0x8003 /* current local directory */

/* ---- prototypes */

/* not using ZEXPORT - no DLLs allowed here */

/* result is new dirID, or if negative, the translated errno negated */
/* important: the errno is _not_ set, so use the error code returned */
int mkdir(const char *path);

/* just like in using FSMakeFSSpec (used), fnfErr means the path is applicable */
extern OSErr path2spec(const void *path, /* can be NULL (return empty spec) */
	int length, /* can be <=0 (return empty spec) */
	struct FSSpec *file);

/* returns negative for error code, otherwise length */
extern int spec2path(const struct FSSpec *file, /* can be NULL (do nothing) */
	void *outpath, /* can be NULL (do nothing) */
	int maxlength); /* maximum length of `outpath` */

/* fsiocopydata does not call SetFPos and returns noErr for EOF errors */
extern OSErr iocopydata(short dstref, /* ioRefNum for destination */
	short srcref, /* ioRefNum for source */
	long int dstpos, /* starting position of destination */
	long int srcpos, /* starting position of source */
	long length, /* length of bytes to be copied */
	void *buffer); /* can be NULL (uses stack or malloc/free for sizes >512) */

/* uses sfTC ID 0 resource to find a type/creator pair */
extern int extensionToFileAndCreator(const void *name, /* filepath or name; can be NULL*/
	int length, /* length of name; can be <=0 */
	OSType *typecreator); /* stores both file type and creator; can be NULL */

/* returns positive on completion, zero if nothing could be applied, negative on error */
extern int fsapplyinfo(short vrefnum, /* volume ref. number (valid #s have bit 15 set) */
	long parid, /* parent directory */
	const UInt8 *name, /* the name ("\p:" for the parid as _the_ directory) */
	time_t datetime); /* year 1904 formatted; pass zero to not set */

#endif /*!defined(__MACOS7_9_H)*/
