/*
 contrib/macos/typelist.r
  master suffix-to-filetype/creator list (Mac OS platform)

  Created: 2008.07.03
  Revised: 2012.02.17 - added '.nsfe'

2010.12.26 - added '.hex'
2010.05.22
*/

/* i'm gracious enough to not require the list to be presorted (or lazy enough) */

/* the text creator is now seperate from 'source' creator codes */
#define DEFAULT_TEXT_CREATOR		'TBB6'	/* Tex-Edit Plus */
#define DEFAULT_SOURCE_CREATOR		'MPS '	/* MPW Shell */
#define DEFAULT_IMAGE_CREATOR		'GKON'	/* GraphicConverter */
#define DEFAULT_MULTIMEDIA_CREATOR	'TVOD'	/* MoviePlayer/QuickTime Player */
#define DEFAULT_WEB_CREATOR		'MOSS'	/* Netscape Navigator/Communicator */
#define DEFAULT_ARCHCODEC_CREATOR	'SITx'	/* StuffIt Expander */
#define DEFAULT_BINARY_CREATOR		'hDmp'	/* HexEdit */

type 'sfTC'
{
	wide array ExtensionTCList
	{ /* empty space must be zeroed */
		string[8]; /* suffix data -- must be in lowercase */
		literal longint; /* file type */
		literal longint; /* file creator */
	};
};

resource 'sfTC' (0,"Extension List")
{
	{
	/*".0" // digit suffixes are unreliable; also used for archive segmentation (binary) */
		".1st", 'TEXT', DEFAULT_TEXT_CREATOR, /* read me text */
		".3dgf", '3DGF', DEFAULT_IMAGE_CREATOR, /* 3DGF graphic */
		".4th", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Forth text */
		".669", '6669', 'SNPL', /* Soundplayer MOD music */
		".8med", 'STrk', 'SCPL', /* Amiga OctaMed music */
		".8svx", '8SVX', 'SCPL', /* Amiga 8-bit audio */
	/*".a", 'TEXT', DEFAULT_SOURCE_CREATOR, // assembly, or i386 object */
		".abek", 'ABEK', '????', /* Abekas */
		".ac", 'TEXT', DEFAULT_SOURCE_CREATOR, /* automake text */
		".ado", 'TEXT', DEFAULT_SOURCE_CREATOR, /* ADO source */
		".afp", 'AFP ', DEFAULT_IMAGE_CREATOR, /* AFP (TIFF Variant) */
		".aif", 'AIFC', DEFAULT_MULTIMEDIA_CREATOR, /* AIFF/AIFC audio */
		".aiff", 'AIFF', DEFAULT_MULTIMEDIA_CREATOR, /* AIFF audio */
		".aifc", 'AIFC', DEFAULT_MULTIMEDIA_CREATOR, /* AIFC audio */
		".ajax", 'TEXT', DEFAULT_TEXT_CREATOR, /* source text */
		".al", 'ALAW', 'SCPL', /* ALaw audio */
		".am", 'TEXT', DEFAULT_SOURCE_CREATOR, /* automake text */
		".ani", 'ANIi', DEFAULT_IMAGE_CREATOR, /* Animated NeoChrome file */
		".arc", 'mArc', 'arc*', /* DOS ArcPC archive */
		".arj", 'BINA', 'DArj', /* DeArj archive */
		".arr", 'ARR ', DEFAULT_IMAGE_CREATOR, /* AMBER ARR file */
		".art", 'ART ', DEFAULT_IMAGE_CREATOR, /* First Publisher ART file */
		".asc", 'TEXT', DEFAULT_TEXT_CREATOR, /* ASCII (maybe) */
		".ascii", 'TEXT', DEFAULT_TEXT_CREATOR, /* ASCII text */
		".asm", 'TEXT', DEFAULT_SOURCE_CREATOR, /* assembly */
		".asp", 'TEXT', DEFAULT_TEXT_CREATOR, /* Server Protocol (web text) */
		".aspx", 'TEXT', DEFAULT_TEXT_CREATOR, /* Server Protocol (web text) */
		".apd", 'TEXT', 'ALD3', /* Aldus Printer Description */
		".app", 'BINA', DEFAULT_BINARY_CREATOR, /* Carbon application/Mac OS X app */
		".au", 'AU  ', DEFAULT_MULTIMEDIA_CREATOR, /* Sun/NeXt audio */
		".avi", 'VfW ', DEFAULT_MULTIMEDIA_CREATOR, /* AVI multimedia */
		".avr", '????', DEFAULT_MULTIMEDIA_CREATOR, /* Audio Visual Research */
		".awk", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Awk text */
		".b", 'TEXT', DEFAULT_SOURCE_CREATOR, /* ASCII text (maybe) */
		".b32", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Borland source text */
		".bar", 'BARF', 'S691', /* Unix BAR archive -- heh heh, BARF */
		".bas", 'TEXT', DEFAULT_SOURCE_CREATOR, /* BASIC source */
		".bat", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Windows batch file (text) */
		".beos", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development text */
		".bga", 'BMP ', 'ogle', /* Windows / OS/2 bitmap */
		".bib", 'TEXT', '*TEX', /* BibTex Bibliography */
		".bin", 'BINA', DEFAULT_BINARY_CREATOR, // raw binary / MacBinary 1-3 */
		".bld", 'BLD ', DEFAULT_IMAGE_CREATOR, /* BLD (MegaPaint) */
		".boo", 'TEXT', 'R*ch', /* BOO encoded text */
		".bor", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Borland source text */
		".bmp", 'BMP ', 'ogle', /* Windows / OS/2 bitmap */
		".bst", 'TEXT', '*TEX', /* BibTex Style */
	/*".buf", 'BUF ', DEFAULT_IMAGE_CREATOR, // HSI-BUF (unreliable) */
		".bw", 'SGI ', DEFAULT_IMAGE_CREATOR, /* Silicon Graphics image */
		".bz", 'Bzb2', DEFAULT_ARCHCODEC_CREATOR, /* BZip */
		".bz2", 'Bzb2', DEFAULT_ARCHCODEC_CREATOR, /* BZip */
		".c", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source text */
		".caf", '????', DEFAULT_MULTIMEDIA_CREATOR, /* Apple Core Audio File */
		".cam", 'CAM ', DEFAULT_IMAGE_CREATOR, /* Casio QV-10/100 Digital Camera image */
		".cdef", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Magic Cap text */
		".cel", 'CEL ', DEFAULT_IMAGE_CREATOR, /* KISS CEL image */
		".cfg", 'BINA', DEFAULT_BINARY_CREATOR, /* Configuration file */
		".cgi", 'TEXT', DEFAULT_TEXT_CREATOR, /* CGI text (web) */
		".cfm", 'TEXT', DEFAULT_SOURCE_CREATOR, /* CFM source text */
		".cgm", 'CGM ', DEFAULT_IMAGE_CREATOR, /* CGM graphic */
		".chm", 'BINA', DEFAULT_BINARY_CREATOR, /* CHM file */
		".class", 'Clss', 'CWIE', /* Java Class file */
		".clut", '8BCT', '8BIM', /* 24 bit (rgb) raw clut */
		".clp", 'CLP ', DEFAULT_IMAGE_CREATOR, /* Windows Clipboard */
		".cmd", 'TEXT', DEFAULT_TEXT_CREATOR, /* OS/2 Batch file */
		".cp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C+ source text */
		".cpg", 'TEXT', DEFAULT_TEXT_CREATOR, /* CPG (unregistered, unreliable, etc.) */
		".cpp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C++ source text */
		".cpt", 'PACT', 'CPCT', /* CompactPro archive */
	/*".coef", 'TEXT', DEFAULT_SOURCE_CREATOR, /* coefficient list (unreliable) */
		".com", 'PCFL', 'SWIN', /* DOS/Windows compatible Command executable */
		".contrib", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text (contributions) */
		".css", 'TEXT', DEFAULT_TEXT_CREATOR, /* Style Sheet */
		".csv", 'TEXT', 'XCEL', /* Comma Sperated Var[iable]s */
		".ct", '..CT', DEFAULT_IMAGE_CREATOR, /* Scitex */
		".cut", 'Halo', DEFAULT_IMAGE_CREATOR, /* Dr Halo image */
		".cvg", 'CVG ', DEFAULT_IMAGE_CREATOR, /* Calamus CVG file */
		".cwk", 'CWWP', 'BOBO', /* ClarisWorks document */
		".cxx", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source text */
		".d", '.D  ', DEFAULT_IMAGE_CREATOR, /* .D image (???) */
		".darwin", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development file */
	/*".dat", 'DJ10', DEFAULT_IMAGE_CREATOR, // data, DJ1000, ONCOR (unreliable) */
		".dax", 'daxA', 'daxA', /* DAKX audio */
		".dbf", 'COMP', 'FOX+', /* DBase Document */
		".dcr", 'FGDM', 'AFTB', /* Shockwave */
		".dcx", 'DCX ', DEFAULT_IMAGE_CREATOR, /* PCX image (possibly) */
		".def", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text definitions */
		".dict", 'TEXT', DEFAULT_SOURCE_CREATOR, /* dictionary text (maybe) */
		".dif", 'TEXT', 'XCEL', /* Data Interchange Format/DIF spreadsheet */
		".diff", 'TEXT', DEFAULT_SOURCE_CREATOR, /* diff text */
		".dir", 'FGDM', 'AFTB', /* Director Shockwave */
		".diz", 'TEXT', 'R*ch', /* BBS Descriptive text */
		".dj2", 'TEXT', DEFAULT_SOURCE_CREATOR, /* djgpp source text */
		".dl", 'DL  ', 'AnVw', /* DL Animation */
		".dll", 'PCFL', 'SWIN', /* DOS/Windows compatible Dynamic Linker module */
		".dmg", 'devi', 'ddsk', /* Disk Image */
		".dhtml", 'TEXT', DEFAULT_SOURCE_CREATOR, /* D-HTML text */
		".dnt", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Windows NT development text */
		".doc", 'WDBN', 'MSWD', /* MS Word document (or sometimes raw text) */
		".dom", 'TEXT', DEFAULT_SOURCE_CREATOR, /* DOM web text */
		".dot", 'TEXT', 'MSWD', /* MS Word for Windows text */
		".dsk", 'dimg', 'ddsk', /* Disk Image */
		".dsp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* MS Developer Studio Project File */
		".dsw", 'TEXT', DEFAULT_SOURCE_CREATOR, /* MS Developer Studio Workspace File */
		".dtd", 'TEXT', DEFAULT_SOURCE_CREATOR, /* DTD protocol text */
		".dv", 'dvc!', DEFAULT_MULTIMEDIA_CREATOR, /* Digital Video */
		".dvi", 'ODVI', 'OTEX', /* TeX DVI */
		".dxf", '.DXF', DEFAULT_IMAGE_CREATOR, /* DXF image / AutoCAD 3D Data */
		".dw5", 'TEXT', DEFAULT_TEXT_CREATOR, /* DisplayWrite 5 */
		".exc", 'TEXT', DEFAULT_TEXT_CREATOR, /* implied executable */
		".exe", 'PCFL', 'SWIN', /* DOS/Windows executable */
		".eps", 'EPSF', DEFAULT_IMAGE_CREATOR, /* Encapsulated PostScript */
		".epsf", 'EPSF', DEFAULT_IMAGE_CREATOR, /* Encapsulated PostScript */
		".esm", 'ESM ', DEFAULT_IMAGE_CREATOR, /* Enhanced Simplex Format file */
		".etx", 'TEXT', 'ezVu', /* Easy View SEText */
		".evy", 'EVYD', 'ENVY', /* Envoy Document */
		".f", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".f copy", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".f2k", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".f66", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".f77", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".f90", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".f95", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".fas", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Forth text */
		".faq", 'TEXT', DEFAULT_TEXT_CREATOR, /* Frequent Asked Questions (text) */
		".fax", 'FAX3', DEFAULT_IMAGE_CREATOR, /* FAX (CCITT3) */
/*		".fc", 'BINA', 'FCEU', // state */
		".fc0", 'FCS ', 'FCEU', /* state */
		".fc1", 'FCS ', 'FCEU', /* state */
		".fc2", 'FCS ', 'FCEU', /* state */
		".fc3", 'FCS ', 'FCEU', /* state */
		".fc4", 'FCS ', 'FCEU', /* state */
		".fc5", 'FCS ', 'FCEU', /* state */
		".fc6", 'FCS ', 'FCEU', /* state */
		".fc7", 'FCS ', 'FCEU', /* state */
		".fc8", 'FCS ', 'FCEU', /* state */
		".fc9", 'FCS ', 'FCEU', /* state */
		".fcm", 'FCM ', 'FCEU', /* state */
		".fcs", 'FCS ', 'FCEU', /* state */
		".fds", 'FDS ', 'FCEU', /* Famicom Disk System file */
		".ff9", '&ff9', DEFAULT_IMAGE_CREATOR, /* Acorn Sprite */
		".flac", '????', DEFAULT_MULTIMEDIA_CREATOR, /* FLAC audio */
		".fi", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".fild", 'fil\266', '52\311D', /* od28-6 recording (unregistered) */
		".fit", 'FITS', DEFAULT_IMAGE_CREATOR, /* (same as FITS) */
		".fits", 'FITS', DEFAULT_IMAGE_CREATOR, /* Flexible Image Transport System */
		".flc", 'FLI ', DEFAULT_MULTIMEDIA_CREATOR, /* (same as FLIC) */
		".flh", 'FLI ', DEFAULT_IMAGE_CREATOR, /* Autodesk FLI/FLC/FLH */
		".fli", 'FLI ', DEFAULT_MULTIMEDIA_CREATOR, /* Animator (Pro), Autodesk */
		".flic", 'FLI ', DEFAULT_MULTIMEDIA_CREATOR, /* FLIC Animation, Autodesk */
		".fm", 'FMPR', 'FMPR', /* FileMaker Pro Database */
		".fm3", 'FMPR', 'FMP3', /* FileMaker Pro 3 Database */
		".for", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Fortran source text */
		".fpx", 'FPix', DEFAULT_IMAGE_CREATOR, /* FlashPix image */
		".fth", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Forth text */
		".fts", 'FITS', DEFAULT_IMAGE_CREATOR, /* Flexible Image Transport System */
		".g", 'TEXT', DEFAULT_SOURCE_CREATOR, /* G source text */
		".g3", 'FAX3', DEFAULT_MULTIMEDIA_CREATOR, /* FAX (CCITT3) */
		".gde", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula-2 source text */
		".gem", 'GEM-', DEFAULT_IMAGE_CREATOR, /* GEM-Metafile image */
		".gfx", 'GFX ', DEFAULT_IMAGE_CREATOR, /* GFX (XGrabber) */
		".gif", 'GIFf', 'ogle', /* GIF image (Compuserve?) */
		".gl", 'GL  ', 'AnVw', /* GL Animation */
		".gmo", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula-2 source text */
		".gp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* G+ source text */
		".gpp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* G++ source text */
		".grp", 'GRPp', DEFAULT_IMAGE_CREATOR, /* GRP image */
		".guess", 'TEXT', DEFAULT_SOURCE_CREATOR, /* UNIX development text (defered) */
		".gz", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* GZip */
		".h", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source text header */
		".ham", '????', DEFAULT_IMAGE_CREATOR, /* image */
		".hcom", 'FSSD', 'SCPL', /* SoundEdit Sound ex SOX */
		".hex", 'TEXT', DEFAULT_SOURCE_CREATOR, /* hexadecimal-formatted text */
		".hfv", 'dimg', 'ddsk', /* HFS[+] raw image */
		".hp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C+ source text header */
		".hpi", 'HPI ', DEFAULT_IMAGE_CREATOR, /* HPI image */
		".hpgl", 'HPGL', DEFAULT_IMAGE_CREATOR, /* Hewlett Packard Graphics Lang./2 file */
		".hpp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C++ source text header */
		".hqx", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* BinHex */
		".htk", '????', DEFAULT_MULTIMEDIA_CREATOR, /* HMM Tool Kit file */
		".htm", 'TEXT', DEFAULT_WEB_CREATOR, /* HTML source text */
		".html", 'TEXT', DEFAULT_WEB_CREATOR, /* HTML source text */
		".howto", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development file */
		".hxx", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source text */
		".i", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Interfacer source text */
		".i3", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula 3 Interface */
		".ic1", 'IMAG', DEFAULT_IMAGE_CREATOR, /* Imagic ICx */
		".ic2", 'IMAG', DEFAULT_IMAGE_CREATOR, /* Imagic ICx */
		".ic3", 'IMAG', DEFAULT_IMAGE_CREATOR, /* Imagic ICx */
		".ico", 'ICO ', DEFAULT_IMAGE_CREATOR, /* Windows Icon */
		".icn", 'ICO ', DEFAULT_IMAGE_CREATOR, /* Windows Icon */
		".icns", 'ICNS', 'Mngl', /* Mac OS 8.5+ Icons file */
		".idl", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Yech development text */
		".ief", 'IEF ', DEFAULT_IMAGE_CREATOR, /* IEF image */
	/*".img", 'dimg', 'ddsk', // DiskCopy image */
	/*".img", 'IMGg', DEFAULT_IMAGE_CREATOR, // GEM bit image/XIMG */
	/*".img", 'KONT', DEFAULT_IMAGE_CREATOR, // KONTRON Image */
		".image", 'dImg', 'ddsk', /* DiskCopy image */
		".in", 'TEXT', DEFAULT_SOURCE_CREATOR, /* source text */
		".inc", 'TEXT', DEFAULT_TEXT_CREATOR, /* include (text) */
		".inf", 'TEXT', DEFAULT_TEXT_CREATOR, /* Inform source */
		".ini", 'TEXT', DEFAULT_TEXT_CREATOR, /* initialization sequence text */
		".inl", 'BINA', DEFAULT_BINARY_CREATOR, /* inline code file (unregistered) */
		".iff", '????', DEFAULT_MULTIMEDIA_CREATOR, /*IFF/ILBM/MAYA-IFF img, Amiga IFF snd*/
		".ilbm", 'ILBM', DEFAULT_IMAGE_CREATOR, /* Interchange File Form image */
		".img", 'IMGg', DEFAULT_IMAGE_CREATOR, /* GEM Bit img, KONTRON, or VG HB600 file */
		".ih", 'TEXT', DEFAULT_TEXT_CREATOR, /* C source header text */
		".isc", 'ISS ', DEFAULT_IMAGE_CREATOR, /* ISS graphic */
		".isk", 'ISS ', DEFAULT_IMAGE_CREATOR, /* ISS graphic */
		".ism", 'ISS ', DEFAULT_IMAGE_CREATOR, /* ISS graphic */
		".iss", 'ISS ', DEFAULT_IMAGE_CREATOR, /* ISS graphic */
		".isy", 'ISS ', DEFAULT_IMAGE_CREATOR, /* ISS graphic */
		".j6i", 'j6i ', DEFAULT_IMAGE_CREATOR, /* Ricoh Digital Camera (DC-1/2) aimage */
		".java", 'TEXT', DEFAULT_TEXT_CREATOR, /* Javascript source text */
		".jbi", 'JBI ', DEFAULT_IMAGE_CREATOR, /* Xerox JBI image */
		".jpe", 'JPEG', DEFAULT_IMAGE_CREATOR, /* JPEG/JFIF image */
		".jpg", 'JPEG', 'ogle', /* JPEG/JFIF image */
		".jpeg", 'JPEG', 'ogle', /* JPEG/JFIF image */
		".jfif", 'JPEG', 'ogle', /* JPEG/JFIF image */
		".jfx", 'TIFF', 'JFAX', /* JFax TIFF */
		".job", 'JBI ', DEFAULT_IMAGE_CREATOR, /* Xerox Image */
		".js", 'TEXT', DEFAULT_TEXT_CREATOR, /* Javascript source text */
		".kdc", 'KDC ', DEFAULT_IMAGE_CREATOR, /* Kodak Digital Camera image */
		".l", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Lex source text */
		".latex", 'TEXT', 'OTEX', /* Latex text */
		".lbm", 'ILBM', DEFAULT_IMAGE_CREATOR, /* Interchange File Form image */
		".lha", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* LHA */
		".linux", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development file */
		".log", 'TEXT', DEFAULT_TEXT_CREATOR, /* log text (maybe) */
		".lst", 'TEXT', DEFAULT_SOURCE_CREATOR, /* list text */
		".lua", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Lua script */
		".lzh", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* LHA */
		".m", 'TEXT', DEFAULT_SOURCE_CREATOR, /* configure paths (?) */
		".m15", 'MPEG', 'JESW', /* MPEG-1 audio */
		".m1a", 'MPEG', 'JESW', /* MPEG-1 audio */
		".m1s", 'MPEG', 'JESW', /* MPEG-1 audio */
		".m1v", 'MPEG', 'JESW', /* MPEG-1 audio */
		".m3u", 'M3U ', 'JESW', /* M3U audio */
		".m2", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula 2 source */
		".m2s", 'MPEG', DEFAULT_MULTIMEDIA_CREATOR, /* MPEG-2 movie */
		".m3", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula 3 source */
		".m4", 'TEXT', DEFAULT_SOURCE_CREATOR, /* configure paths (text) */
		".m75", 'MPEG', 'JESW', /* MPEG-1 audio */
		".mac", 'PNTG', DEFAULT_IMAGE_CREATOR, /* MacPaint (maybe), Mac OS PICT */
		".macosx", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development file */
		".mak", 'TEXT', DEFAULT_SOURCE_CREATOR, /* makefile text */
		".make", 'TEXT', DEFAULT_SOURCE_CREATOR, /* makefile text */
		".makeout", 'TEXT', DEFAULT_SOURCE_CREATOR, /* makefile text */
		".mask", 'APIX', DEFAULT_IMAGE_CREATOR, /* ALIAS pix */
		".mat", '????', DEFAULT_MULTIMEDIA_CREATOR, /* GNU Octave / Matlab */
		".mbm", 'MBM ', DEFAULT_IMAGE_CREATOR, /* PSION 5 (MBM) */
		".mcw", 'WDBN', 'MSWD', /* Mac Word document */
		".med", 'STrk', 'SCPL', /* Amiga MED Sound */
		".mf", 'TEXT', '*MF*', /* Metafont */
		".mid", 'MIDI', DEFAULT_MULTIMEDIA_CREATOR, /* MIDI music */
		".midi", 'MIDI', DEFAULT_MULTIMEDIA_CREATOR, /* MIDI music */
		".mid!", 'MooV', DEFAULT_MULTIMEDIA_CREATOR, /* QuickTime music */
		".mif", 'TEXT', 'Fram', /* FrameMaker MIF */
		".mil", 'CALS', DEFAULT_IMAGE_CREATOR, /* CALS/CCITT4 */
		".mime", 'TEXT', 'mPAK', /* MIME/Base64 */
		".mk", 'TEXT', DEFAULT_SOURCE_CREATOR, /* makefile text */
		".ml", 'TEXT', DEFAULT_SOURCE_CREATOR, /* ML Source */
		".mml", 'TEXT', DEFAULT_SOURCE_CREATOR, /* markup source text */
		".mms", 'TEXT', DEFAULT_SOURCE_CREATOR, /* VMS MMS development text */
		".mod", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula-2 source text */
		".mol", 'TEXT', 'RSML', /* MDL Molfile */
		".mov", 'MooV', DEFAULT_MULTIMEDIA_CREATOR, /* QuickTime movie */
		".moov", 'MooV', DEFAULT_MULTIMEDIA_CREATOR, /* QuickTime movie */
		".mp2", 'MP3 ', 'JESW', /* MPEG II */
		".mp3", 'MP3 ', 'JESW', /* MPEG III */
		".mp4", 'MP3 ', 'JESW', /* MPEG IV ('going to work?) */
		".mpa", 'MPEG', 'JESW', /* MPEG-1 audio */
		".mpg", 'MPEG', DEFAULT_MULTIMEDIA_CREATOR, /* MPEG */
		".mpe", 'MPEG', DEFAULT_MULTIMEDIA_CREATOR, /* MPEG */
		".mpeg", 'MPEG', DEFAULT_MULTIMEDIA_CREATOR, /* MPEG */
		".mpm", 'MPEG', DEFAULT_MULTIMEDIA_CREATOR, /* MPEG */
		".mps", 'MSPp', DEFAULT_IMAGE_CREATOR, /* Microsoft Paint Picture */
		".mpv", 'MPEG', DEFAULT_MULTIMEDIA_CREATOR, /* MPEG video */
		".mpw", 'TEXT', DEFAULT_SOURCE_CREATOR, /* MPW source file */
		".msc", 'TEXT', DEFAULT_SOURCE_CREATOR, /* MS C development file */
		".msp", '????', DEFAULT_IMAGE_CREATOR, /* Microsoft Paint (maybe) */
		".msvc", 'TEXT', DEFAULT_SOURCE_CREATOR, /* MS Visual C development file */
		".mtm", 'MTM ', 'SNPL', /* MultiMOD music */
		".mw", 'MW2D', 'MWII', /* MacWrite doc. */
		".mwii", 'MW2D', 'MWII', /* MacWrite II doc. */
		".neo", 'NeoC', DEFAULT_IMAGE_CREATOR, /* Atari NeoChrome image */
		".nes", 'Cart', 'FCEU', /* NES/Famicon Cartrige file */
		".nfo", 'TEXT', DEFAULT_TEXT_CREATOR, /* info text */
		".nif", 'NIF ', DEFAULT_IMAGE_CREATOR, /* NIF image */
	/*".now", 'TEXT', DEFAULT_TEXT_CREATOR, /* info text (unreliable) */
		".nsf", 'NSF ', 'FCEU', /* Nintendo Sound File */
		".nsfe", 'NSF ', 'FCEU', /* ..enhanced */
		".nst", 'STrk', 'SCPL', /* MOD music */
		".nt", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Windows NT development text */
		".o", 'BINA', DEFAULT_BINARY_CREATOR, /* object file */
		".obj", 'PCFL', 'SWIN', /* object file */
		".oda", 'ODIF', 'ODA ', /* ODA Document */
		".odef", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Magic Cap text */
		".of", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Forth text */
		".okt", 'STrk', 'SCPL', /* Oktalyser MOD music */
		".otf", 'sfnt', 'movr', /* OpenType Font */
	/*".out", 'TEXT', DEFAULT_BINARY_CREATOR, // text */
		".ovl", 'PCFL', 'SWIN', /* DOS/Windows Overlay */
		".p", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Pascal source file */
		".pac", 'STAD', DEFAULT_IMAGE_CREATOR, /* STAD PAC */
		".pal", 'BINA', DEFAULT_BINARY_CREATOR, /* Color Palette */
		".paf", '????', DEFAULT_MULTIMEDIA_CREATOR, /* Ensoniq PARIS file */
		".pas", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Pascal source file */
		".pau", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* (???) */
		".pbm", 'PPGM', DEFAULT_IMAGE_CREATOR, /* Portable Bitmap */
		".pc", 'TEXT', DEFAULT_TEXT_CREATOR, /* package control (text) */
		".pc1", 'Dega', DEFAULT_IMAGE_CREATOR, /* Degas PIx/PCx */
		".pc2", 'Dega', DEFAULT_IMAGE_CREATOR, /* Degas PIx/PCx */
		".pc3", 'Dega', DEFAULT_IMAGE_CREATOR, /* Degas PIx/PCx */
		".pcd", 'PCDI', DEFAULT_IMAGE_CREATOR, /* Kodak PCD image */
		".pcx", 'PCXx', DEFAULT_IMAGE_CREATOR, /* Paintbrush PCX/SCR image */
		".pch", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source header text */
		".pchc", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source header text */
		".pchcp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source header text */
		".pcs", 'PICT', DEFAULT_IMAGE_CREATOR, /* Mac OS PICTs animation */
		".pcz", 'PICT', 'ogle', /* Mac OS PICT (MS Office type) */
		".pdb", 'PDB ', DEFAULT_IMAGE_CREATOR, /* PDB image */
		".pdf", 'PDF ', 'CARO', /* PDF document */
		".pdx", 'TEXT', 'ALD5', /* Printer Description (Aldus PageMaker 5) */
		".pgm", 'PPGM', DEFAULT_IMAGE_CREATOR, /* Portable Graymap */
		".pgp", 'PKey', 'MPGP', /* PGP Key File (MacPGP262) */
		".ph", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Perl source text */
		".php", 'TEXT', DEFAULT_TEXT_CREATOR, /* web protocol text */
		".phy2", 'phy2', '52\311D', /* Marathon 2+ physics file */
		".pi1", 'Dega', DEFAULT_IMAGE_CREATOR, /* Degas PIx/PCx */
		".pi2", 'Dega', DEFAULT_IMAGE_CREATOR, /* Degas PIx/PCx */
		".pi3", 'Dega', DEFAULT_IMAGE_CREATOR, /* Degas PIx/PCx */
		".pic", 'IPIC', DEFAULT_IMAGE_CREATOR, /*Lotus/PICc,SOFTIMAGE,MSX2+,PSION,PORST,BioRad*/
		".pit", 'PIT ', DEFAULT_ARCHCODEC_CREATOR, /* PackIt Archive */
		".pix", 'APIX', DEFAULT_IMAGE_CREATOR, /* ALIAS pix */
		".pkg", 'HBSF', 'GEOL', /* AppleLink Package */
		".pl", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Perl source text */
		".plist", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development text */
		".plt", 'HPGL', DEFAULT_IMAGE_CREATOR, /* Hewlett Packard Graphics Lang./2 file */
	/*".pm", 'TEXT', DEFAULT_SOURCE_CREATOR, // Perl source text, Bitmap from xv */
		".pm3", 'ALB3', 'ALD3', /* PageMaker 3 Document */
		".pm4", 'ALB4', 'ALD4', /* PageMaker 4 Document */
		".pm5", 'ALB5', 'ALD5', /* PageMaker 5 Document */
		".png", 'PNGf', DEFAULT_IMAGE_CREATOR, /* Portable Network Graphics image */
		".pntg", 'PNTG', DEFAULT_IMAGE_CREATOR, /* MacPaint */
		".ppm", 'PPGM', DEFAULT_IMAGE_CREATOR, /* Portable Pixmap Image */
		".ppt", 'SLD3', 'PPT3', /* PowerPoint Presentation */
		".prn", 'TEXT', DEFAULT_TEXT_CREATOR, /* Printer Output File */
		".ps", 'TEXT', DEFAULT_TEXT_CREATOR, /* PostScript */
		".psd", '8BIM', '8BIM', /* Photoshop */
		".psw", 'TEXT', DEFAULT_TEXT_CREATOR, /* PostScript */
		".pswm", 'TEXT', DEFAULT_TEXT_CREATOR, /* PostScript */
		".pt4", 'ALT4', 'ALD4', /* PageMaker 4 Template */
		".pt5", 'ALT5', 'ALD5', /* PageMaker 5 Template */
		".pup", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Amiga powerUP development text */
		".pvf", '????', DEFAULT_MULTIMEDIA_CREATOR, /* Portable Voice Format file */
		".pw2", 'TEXT', 'BOBO', /* ProWrite 2 */
		".pxr", 'PXR ', '8BIM', /* Pixar Image */
		".py", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development file */
		".qcp", 'celp', 'Blad', /* QUALCOMM Audio (PureVoice) */
		".qdv", 'QDVf', DEFAULT_IMAGE_CREATOR, /* Giffer */
		".qif", 'QTIF', 'ogle', /* QuickTime image */
		".qnt", 'QNT ', DEFAULT_IMAGE_CREATOR, /* QNT image */
		".qt", 'MooV', DEFAULT_MULTIMEDIA_CREATOR, /* QuickTime (movie) */
		".qti", 'QTIF', 'ogle', /* QuickTime image */
		".qtif", 'QTIF', 'ogle', /* QuickTime QTIF image */
		".qxd", 'XDOC', 'XPR3', /* QuarkXpress Document */
		".qxt", 'XTMP', 'XPR3', /* QuarkXpress Template */
		".r", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Rez source file */
		".ra", 'PNRM', 'PNst', /* RealAudio */
		".ram", 'PNRM', 'PNst', /* RealAudio */
		".rc", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Windows RC resource source text */
		".rde", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula-2 source text */
		".readme", 'TEXT', DEFAULT_TEXT_CREATOR, /* readme */
		".res", 'BINA', DEFAULT_BINARY_CREATOR, /* resource file (Windows) */
		".rez", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Rez source file */
		".rft", 'TEXT', '????', /* DCARFT document */
		".rgb", '.SGI', 'ogle', /* SGI Image */
		".rgba", '.SGI', 'ogle', /* SGI Image */
		".rif", 'RIFF', DEFAULT_IMAGE_CREATOR, /* Raster Image File Format Graphic */
		".riff", 'RIFF', DEFAULT_IMAGE_CREATOR, /* Raster Image File Format Picture */
		".riscos", 'TEXT', DEFAULT_SOURCE_CREATOR, /* RISC OS source text */
		".rla", 'RLA ', DEFAULT_IMAGE_CREATOR, /* RLA image */
		".rle", 'RLE ', DEFAULT_IMAGE_CREATOR, /* Utah Raster Toolkit */
		".rlec", '????', DEFAULT_IMAGE_CREATOR, /* Compuserve RLE graphic */
	/*".rm", '.RM ', DEFAULT_IMAGE_CREATOR, // NASA Raster Metafile, RealAudio */
	/*".rme", 'TEXT', DEFAULT_TEXT_CREATOR, // another readme type */
		".rmo", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Modula-2 source text */
		".rpl", 'FRL!', 'REP!', /* Replica Document */
		".rsc", 'BINA', DEFAULT_BINARY_CREATOR, /* resource file, probably not Mac OS */
		".rsrc", 'BINA', DEFAULT_BINARY_CREATOR, /* resource file, might not be Mac OS */
		".rtf", 'TEXT', 'BOBO', /* RTF (rich) text */
		".rtx", 'TEXT', 'R*ch', /* Rich Text */
		".s", 'TEXT', DEFAULT_SOURCE_CREATOR, /* assembly text (maybe) */
		".s12", 'MSX ', DEFAULT_IMAGE_CREATOR, /* MSX Picture */
		".s3m", 'S3M ', 'SNPL', /* ScreamTracker 3 MOD */
		".sas", 'TEXT', DEFAULT_SOURCE_CREATOR, /* SMakefile text */
		".sce", 'sce2', '26.\260', /* Marathon 2+ scenario */
		".sce2", 'sce2', '26.\260', /* Marathon 2+ scenario */
		".scea", 'sce2', '26.A', /* Aleph One scenario */
		".sced", 'sce2', '52\311D', /* od28-6 scenario (unregistered) */
		".scc", 'MSX ', DEFAULT_IMAGE_CREATOR, /* MSX2+ [HBI-V1] */
		".scg", 'RIX3', DEFAULT_IMAGE_CREATOR, /* ColoRIX image */
		".sci", 'RIX3', DEFAULT_IMAGE_CREATOR, /* ColoRIX image */
		".scm", 'TEXT', DEFAULT_SOURCE_CREATOR, /* SCMake text */
		".scp", 'RIX3', DEFAULT_IMAGE_CREATOR, /* ColoRIX image */
		".scr", 'PCXx', DEFAULT_IMAGE_CREATOR, /* ScreenCapture PCX/SCR, ColoRIX image */
		".scu", 'RIX3', DEFAULT_IMAGE_CREATOR, /* ColoRIX image */
		".sd2", 'Sd2f', 'SCPL', /* SoundDeluxe II (probably needs RF) */
		".sds", '????', DEFAULT_MULTIMEDIA_CREATOR, /* Midi Sample Dump Standard file */
		".sea", 'SIT5', DEFAULT_ARCHCODEC_CREATOR, /*StuffIt self-expander; no RF, no APPL*/
		".sf", '????', DEFAULT_MULTIMEDIA_CREATOR, /* Berkeley/IRCAM/CARL SF file */
		".sfw", 'SFWI', DEFAULT_IMAGE_CREATOR, /* SFW (Filmworks) */
		".sgi", '????', DEFAULT_IMAGE_CREATOR, /* SGI image */
		".sh", 'TEXT', DEFAULT_SOURCE_CREATOR, /* shell development text */
		".sha", 'TEXT', 'UnSh', /* Unix Shell Archive */
		".shar", 'TEXT', 'UnSh', /* Unix Shell Archive */
		".shp", 'SHPp', DEFAULT_IMAGE_CREATOR, /*LightingPress img, Printmaster Icon Lib.*/
		".shp2", 'shp2', '52\311D', /* Marathon 2+ shapes file */
		".shtml", 'TEXT', DEFAULT_WEB_CREATOR, /* HTML source text */
		".sit", 'SIT5', DEFAULT_ARCHCODEC_CREATOR, /* StuffIt archive */
		".sithqx", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* BinHexed StuffIt archive */
		".six", 'SIXE', DEFAULT_IMAGE_CREATOR, /* SIXEL image */
		".sixel", 'SIXE', DEFAULT_IMAGE_CREATOR, /* SIXEL image */
		".sgi", 'SGI ', DEFAULT_IMAGE_CREATOR, /* Silicon Graphics image */
		".slk", 'TEXT', 'XCEL', /* SYLK Spreadsheet */
		".smi", 'dimg', 'ddsk', /* ShrinkWrap Self-Mounting image (no RF, no APPL) */
		".snd", 'BINA', 'SCPL', /* sound of any type */
		".snd2", 'snd2', '52\311D', /* Marathon 2+ sounds file */
		".spc", 'Spec', DEFAULT_IMAGE_CREATOR, /* Atari Spectrum 512 image */
		".sql", 'TEXT', DEFAULT_SOURCE_CREATOR, /* SQL text */
		".sr", 'SUNn', DEFAULT_IMAGE_CREATOR, /* Sun Rasterfile */
		".srd", 'BINA', DEFAULT_BINARY_CREATOR, /*eDoc Self-Reading Doc. (no RF, no APPL)*/
		".ss", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development text */
		".sty", 'TEXT', '*TEX', /* TeX textures style */
		".sub", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development text */
		".sun", 'SUNn', DEFAULT_IMAGE_CREATOR, /* Sun Rasterfile */
		".sup", 'SCRN', DEFAULT_IMAGE_CREATOR, /* StartupScreen */
		".svx", '8SVX', 'SCPL', /* Amiga IFF sound */
		".syk", 'TEXT', 'XCEL', /* SYLK Spreadsheet */
		".sylk", 'TEXT', 'XCEL', /* SYLK Spreadsheet */
		".sys", 'PCFL', 'SWIN', /* DOS/Windows system file (e.g., driver) */
		".tar", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* TAR uncompressed archive */
		".targa", 'TPIC', 'ogle', /* Truecolor File Format Targa image */
		".taz", 'ZIVU', 'LZIV', /* MacCompress Tape ARchive */
		".tc", 'TEXT', DEFAULT_SOURCE_CREATOR, /* TurboC */
		".tcl", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Tcl development text */
		".tex", 'TEXT', 'OTEX', /* TeX Document (OzTex) */
		".texi", 'TEXT', 'OTEX', /* TeX Info */
		".texinfo", 'TEXT', 'OTEX', /* TeX Info */
		".text", 'TEXT', 'TBB6', /* text */
		".tga", 'TPIC', 'ogle', /* Truecolor File Format Targa, Truevision image */
		".tgz", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* TGZip; Gnu ZIPed Tape ARchive */
		".tif", 'TIFF', DEFAULT_IMAGE_CREATOR, /* same as TIFF */
		".tiff", 'TIFF', DEFAULT_IMAGE_CREATOR, /* Tag Image File Format */
		".tim", 'TIM ', DEFAULT_IMAGE_CREATOR, /* PSX TIM image */
		".tk", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Tcl development text */
		".tny", 'TINY', DEFAULT_IMAGE_CREATOR, /* Atari TINY Bitmap */
		".tpl", 'TEXT', DEFAULT_SOURCE_CREATOR, /* template source text */
		".ts", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development text */
		".tsm", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Z80/Z1810 assembly text */
		".tsv", 'TEXT', 'XCEL', /* Tab Separated Values */
		".ttc", 'ttcf', 'movr', /* TrueType Font Collection */
		".ttf", 'sfnt', 'movr', /* TrueType Font */
		".tx8", 'TEXT', DEFAULT_TEXT_CREATOR, /* 8-bit text */
		".txt", 'TEXT', DEFAULT_TEXT_CREATOR, /* text */
		".ul", 'ULAW', 'SCPL', /* Law audio */
		".unix", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development file */
		".url", 'URL ', DEFAULT_WEB_CREATOR, /* bookmark */
		".uu", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* UU Encoded */
		".uue", 'TEXT', DEFAULT_ARCHCODEC_CREATOR, /* UU Encoded */
		".vbs", 'TEXT', DEFAULT_SOURCE_CREATOR, /* VisualBasic source text */
		".vff", 'VFFf', DEFAULT_IMAGE_CREATOR, /* DESR VFF file */
		".vob", 'MPEG', DEFAULT_MULTIMEDIA_CREATOR, /* DVD Movie */
		".voc", 'VOC ', 'SCPL', /* Creative Labs VOC file */
		".vox", '????', DEFAULT_MULTIMEDIA_CREATOR, /* VOX ADPCM format file */
		".vpb", 'VPB ', DEFAULT_IMAGE_CREATOR, /* Quantel Format VPB file */
		".vtp", 'VITR', DEFAULT_IMAGE_CREATOR, /* VITRONIC image */
		".vxd", 'PCFL', 'SWIN', /* Win9x driver */
		".w32", 'TEXT', DEFAULT_SOURCE_CREATOR, /* MS 32-bit Visual C++ source text */
		".w51", '.WP5', 'WPC2', /* WordPerfect PC 5.1 Doc */
		".w64", '????', DEFAULT_MULTIMEDIA_CREATOR, /* SoundFoundry Wave-64 */
		".wad", '.WAD', 'DOOM', /* DOOM Wad */
		".wat", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Watcom 10a source text (maybe) */
		".wav", 'WAVE', DEFAULT_MULTIMEDIA_CREATOR, /* Windows Wave */
		".wbmp", 'WBMP', DEFAULT_IMAGE_CREATOR, /* Wireless Bitmap Image */
		".wk1", 'TEXT', 'XCEL', /* Lotus Spreadsheet r2.1 */
		".wks", 'TEXT', 'XCEL', /* Lotus Spreadsheet r2.x */
		".wma", '????', DEFAULT_MULTIMEDIA_CREATOR, /* Windows Media Audio */
		".wmf", 'WMF ', DEFAULT_IMAGE_CREATOR, /* Windows MetaFile */
		".wp", '.WP5', 'WPC2', /* Word Perfect 5.1 */
		".wp4", '.WP4', 'WPC2', /* Word Perfect 4 */
		".wp5", '.WP5', 'WPC2', /* Word Perfect 5 */
		".wp6", '.WP6', 'WPC2', /* Word Perfect 6 */
		".wpg", 'WPGf', DEFAULT_IMAGE_CREATOR, /* Word Perfect Graphic */
		".wpm", 'WPD1', 'WPC2', /* Word Perfect Mac */
		".wri", 'WDBN', 'MSWD', /* MS Write/Windows */
		".ws", 'TEXT', DEFAULT_SOURCE_CREATOR, /* development text */
		".x", 'TEXT', DEFAULT_BINARY_CREATOR, /* object file */
		".xve", 'BINA', 'SCPL', /* PSION sound */
		".x10", 'XWDd', DEFAULT_IMAGE_CREATOR, /* X-Windows Dump */
		".x11", 'XWDd', DEFAULT_IMAGE_CREATOR, /* X-Windows Dump */
		".xbm", 'XBM ', DEFAULT_IMAGE_CREATOR, /* X11-Bitmap */
		".xi", '????', DEFAULT_MULTIMEDIA_CREATOR, /* FastTracker 2 XI file */
		".xif", 'TIFF', DEFAULT_IMAGE_CREATOR, /* TIFF variant */
		".xih", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source header text */
		".xl", 'TEXT', 'XCEL', /* Excel Spreadsheet */
		".xlc", 'TEXT', 'XCEL', /* Excel Chart */
		".xlm", 'TEXT', 'XCEL', /* Excel Macro */
		".xlw", 'TEXT', 'XCEL', /* Excel Workspace */
		".xpm", 'XPM ', DEFAULT_IMAGE_CREATOR, /* ASCII Bitmap */
		".xh", 'TEXT', DEFAULT_SOURCE_CREATOR, /* C source header text */
		".xhtml", 'TEXT', DEFAULT_SOURCE_CREATOR, /* X-HTML source */
		".xm", 'XM  ', 'SNPL', /* FastTracker MOD */
		".xml", 'TEXT', DEFAULT_SOURCE_CREATOR, /* markup source text */
		".xpm", 'XPM ', DEFAULT_IMAGE_CREATOR, /* X-Windows Pixmap */
		".xsd", 'XWDd', DEFAULT_IMAGE_CREATOR, /* X-Windows Dump */
		".xsl", 'TEXT', DEFAULT_SOURCE_CREATOR, /* source text */
		".xslt", 'TEXT', DEFAULT_SOURCE_CREATOR, /* source text */
		".y", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Yacc source text */
	/*".z", 'TEXT', DEFAULT_SOURCE_CREATOR, // Z80/Z1810 assembly text, UNIX Compress Archive */
		".z80", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Z80/Z1810 assembly text */
		".zip", 'ZIP ', 'ZIP ', /* Zip archive */
		".zoo", 'Zoo ', 'Booz', /* Zoo archive (MacBooz) */
		".zsm", 'TEXT', DEFAULT_SOURCE_CREATOR, /* Z80/Z1810 assembly text */
		"authors", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"readme", 'TEXT', DEFAULT_SOURCE_CREATOR, /* README text */
		"index", 'TEXT', DEFAULT_SOURCE_CREATOR, /* manifest text */
		"makefile", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"news", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"requests", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text (maybe) */
		"compile", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text (maybe) */
		"copying", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"ltconfig", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"install", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"depcomp", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text (maybe) */
		"missing", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text (maybe) */
		"todo", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"manifest", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"history", 'TEXT', DEFAULT_SOURCE_CREATOR, /* text */
		"stdout", 'TEXT', DEFAULT_TEXT_CREATOR, /* text */
		"stderr", 'TEXT', DEFAULT_TEXT_CREATOR, /* text */
	};
};
