/*
 zlib/contrib/minizip/miniunz.c
  Version 1.01e+, February 12th, 2005

   Copyright (C) 1998-2005 Gilles Vollant

Mods by Adam P.
Fixed poor gramar as well.
*/

#define LAST_REVISION_DATE      "2014.04.06"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>

#define IGNORE_DSSTORE

#ifdef mac
 #include ":::zlib.h"
 #include "::macos:macos7-9.h"
 #include <Files.h>
/* so the user can cancel */
 #define signalbreak()  ((*((volatile UInt8*)0x179)&*((volatile UInt8*)0x17A))!=0)
 #define TPATH_SEP      ':'
 #define TPATH_EXTRA    1
 #define MAXNAMELENGTH  31

 #ifdef APPTOOL
  #include <AppleEvents.h>
  #include <StandardFile.h>
  #include <TextUtils.h>
  #include <Scrap.h>
  #define stderr        stdout
  #if TARGET_CPU_68K
   #pragma segment shell
  #endif
 #endif
#else
 #define signalbreak()  0
 #define TPATH_SEP      '/'
 #define TPATH_EXTRA    0
 #define MAXNAMELENGTH  255

 #ifdef unix
 #include <unistd.h>
  #include <utime.h>
 #else
  #include <direct.h>
  #include <io.h>
 #endif
#endif

#include "unzip.h"

#ifndef local
 #define local static
#endif


#define CASESENSITIVITY 0
#define WRITEBUFFERSIZE 8192
#define MAXPATHLENGTH   256

#ifdef WIN32
 #define USEWIN32IOAPI
 #include "iowin32.h"
#endif
/*
  mini unzip, demo of unzip package

  usage :
  Usage : miniunz [-exvlo] file.zip [file_to_extract] [-d extractdir]

  list the file in the zipfile, and print the content of FILE_ID.ZIP or README.TXT
    if either exists
*/

int platformfpath           OF((char*,char*));
char *printablefpath        OF((char*));

void setfileinfo            OF((const char*,const unz_file_info*,long*));

int makedir                 OF((const char*));


#ifdef mac
local void backwritehash(unsigned char *d,
    unsigned a,
    int z,
    unsigned b,
    const char *t)
{
    unsigned       r;

    do
        r = a, a/= b, r-= a*b, *--d = t[r];
    while(--z);
    *--d = '-';
}

local void *shortenfpath(void *name,
    void *point,
    unsigned a)
{
    unsigned char      *d, *f = NULL, *e, c;

    if ((d=(unsigned char*)point)==NULL)
        return NULL;
    if (!name && d<=(unsigned char*)name)
        return d;
    /* scan to end of name, finish hash */
    while((c=*d++)!=0 && (c+=-'/') && (c+='/'-'\\') && (c+='\\'-':'))
        a = (a<<5)^c;
    e =--d;

    /* scan beginning of name, gather dot-extension */
    do
    {
        if ((c=*--d)==':')
            {++d; break;}
        if (f)
            continue;
        if (c=='.' && e-(f=d)>8)
        { /* truncate extension to 8, including the '.' */
            memmove(8+d,e,1+strlen((char*)e));
            e = 8+d;
        }
    } while(d>name);

    if (!f)
        f = e;
    d+= MAXNAMELENGTH+f-e;
    e+= d-f;
    memmove(d,f,1+strlen((char*)f));

    backwritehash(d,a,6,40,"0123456789abcdefghijklmnopqrstuvwxyz_~! ");

    return e;
}

local const char *zError(int err)
{
    return err==UNZ_PARAMERROR ? " (bad parameter)"
        :err==UNZ_BADZIPFILE ? " (bad zipfile)"
        :err==UNZ_CRCERROR ? " (CRC-32 error; zipfile appears to be corrupted)"
        :err==UNZ_INTERNALERROR ? " (internal error)"
        :err==UNZ_END_OF_LIST_OF_FILE ? " - surpassed last entry in zipfile"
        :err==Z_STREAM_ERROR ? " (decompression stream error)"
        :err==Z_DATA_ERROR ? " (bad data)"
        :err==Z_MEM_ERROR ? " (out of memory)"
        :err==Z_BUF_ERROR ? " (buffer error)"
        :err==Z_VERSION_ERROR ? " (version error)"
        :"";
}

local const char *errnoError(int err)
{
    switch(err)
    {
    case EPERM: return " (an operation was not permitted)";
    case EIO: return " (I/O error)";
    case ENXIO: return " (an operation called for an unavailable device)";
    case EBADF: return " (bad file descriptor)";
    case ENOMEM: return " (out of memory space)";
    case EACCES: return " (permission denied for type of file access)";
    case ENOENT: return " (file index or position out of range)";
    case EBUSY: return " (a system resource is busy)";
    case EEXIST: return " (file/directory of same name exists)";
    case ENOTDIR: return " (path did not lead to directory; called for)";
    case EISDIR: return " (attempted to open directory like a file)";
    case ENFILE: return " (too many open files in system)";
    case EMFILE: return " (too many open files for process)";
    case ENOTTY: return " (inappropriate I/O control operation)";
    case ETXTBSY: return " (text file busy, regarding output)";
    case EFBIG: return " (file got too large)";
    case ENOSPC: return " (no space left on drive)";
    case ESPIPE: return " (invalid seek position)";
    case EROFS: return " (attempted to modify a read-only file!)";
    case EMLINK: return " (too many links on a single file)";
    case EPIPE: return " (broken pipe; no process to read it)";
    case EDOM: return " (argument outside function domain)";
    case ERANGE: return " (a result is out of range)";

    default: return "";
    }
}

 #ifdef APPTOOL
local ModalFilterUPP	stdfilter, modalfilterupp;

char		temporary[1024];

  #if !TARGET_API_MAC_CARBON
EXTERN_API(Rect*)
GetRegionBounds(RgnHandle,
	Rect*);
EXTERN_API(Rect*)
GetRegionBounds(RgnHandle rgn,
	Rect *bounds)
{
	if (!bounds)
		return NULL;
	*bounds = (**rgn).rgnBBox;
	return bounds;
}
  #endif

void GetDirectory(const UInt8 *prompt,
	StandardFileReply *reply);

/* prototypes */

DialogRef getnewdialog(int dialogid,
	WindowRef frontwindow);
int dialogupdate(DialogRef);
int findbuttonitem(DialogRef,
	const UInt8*);
int animatebuttonitem(DialogRef,
	int item);
void setdialogitem_pstr(DialogRef,
	DialogItemIndex,
	const UInt8*);

/* code */

local pascal Boolean modalfilter(DialogRef dlg,
	EventRecord *event,
	short *item_hit)
{
	Rect		bounds;
	WindowRef	window;

	*item_hit =-1;
	if (!stdfilter)
		GetStdFilterProc(&stdfilter);
	switch(event->what)
	{
	case updateEvt:
		window = (WindowRef)event->message;
		if (window!=GetDialogWindow(dlg))
			break;
		BeginUpdate(window);
		dialogupdate(dlg);
		EndUpdate(window);
	case nullEvent: /* for text cursor to flash */
		break;
	case mouseDown:
		if (FindWindow(event->where,&window)==inDrag)
		{
			GetRegionBounds(GetGrayRgn(),&bounds);
			DragWindow(window,event->where,&bounds);
			*item_hit =-1;
			return true;
		}
		goto end;
	case keyDown:
	case autoKey:
		switch(event->message&charCodeMask)
		{
		case 'a':
			if (!(event->modifiers&cmdKey))
				*item_hit = findbuttonitem(dlg,"\paccept");
			else if ((*item_hit=GetDialogKeyboardFocusItem(dlg))>=0)
			{ /* select all */
				SelectDialogItemText(dlg,*item_hit,0,32767);
				return true;
			}
			break;
		case 'e': *item_hit = findbuttonitem(dlg,"\pextract"); break;
		case 'l': *item_hit = findbuttonitem(dlg,"\plist"); break;
		case 'y': *item_hit = findbuttonitem(dlg,"\pyes"); break;
		case 'n': *item_hit = findbuttonitem(dlg,"\pno"); break;
		case 'o': *item_hit = findbuttonitem(dlg,"\poverwrite all"); break;
		case kEscapeCharCode: *item_hit = findbuttonitem(dlg,"\pcancel"); goto esc;
		}
		if (GetDialogKeyboardFocusItem(dlg)>=0) /* fix: not when in a text box */
			break;
	esc:	if (*item_hit>0) 
			{animatebuttonitem(dlg,*item_hit); return true;}
	}
	if (stdfilter)
		return InvokeModalFilterUPP(dlg,event,item_hit,stdfilter);
end:	return false;
}

typedef struct dirfilterdata
{
	StandardFileReply	*reply;

	SInt16		oldvol;
	Boolean		setfocus;
} dirfilterdata;

local char *sizeStr(char *buf,
	unsigned long sectorCount)
{
	const char	*scale = "KB";
	char		*a;
	unsigned long	size = sectorCount>>1;
	unsigned	low = sectorCount<<9;

	if (size>921)
		low = size, size>>= 10, scale = "MB";
	if (size>921)
		low = size, size>>= 10, scale = "GB";
	if ((low=((low&1023)*25+128)>>8)==100)
		++size, low = 0;
	a = buf+sprintf(buf,low ? "%u.%02u":"%u",
		size,low);
	if (a[-1]=='0')
		--a;
	if (a[-1]=='0')
		--a;
	memcpy(a,scale,3);
	return buf;
}

local pascal Boolean dirfilter(DialogRef dlg,
	EventRecord *event,
	DialogItemIndex *hit,
	void *u)
{
	dirfilterdata	*fdata = u;

	if (event->what!=keyDown && event->what!=autoKey)
	{
		TEHandle	te;

		if ((te=GetDialogTextEditHandle(dlg))!=NULL)
			TEDeactivate(te);
		if (fdata && fdata->setfocus && event->what==nullEvent
				&& GetDialogKeyboardFocusItem(dlg)==10)
		{
			fdata->setfocus = false;
			*hit = 7;
			return true;
		}
	} else
		switch(event->message&charCodeMask)
		{
		case kTabCharCode:
		case kEnterCharCode:
		case kPageUpCharCode:
		case kPageDownCharCode:
		case kReturnCharCode:
		case kFunctionKeyCharCode:
		case kLeftArrowCharCode:
		case kRightArrowCharCode:
		case kDownArrowCharCode:
		case kUpArrowCharCode:
			break;
		default:
			if (!(event->modifiers&cmdKey))
				event->what = nullEvent;
		}
	if (fdata && fdata->oldvol!=fdata->reply->sfFile.vRefNum
			&& fdata->reply->sfFile.vRefNum<0)
	{
		HVolumeParam	voldata;
		char		buf[16];

		memset(&voldata,0,sizeof(voldata));
		voldata.ioVRefNum = fdata->oldvol = fdata->reply->sfFile.vRefNum;
		PBHGetVInfoSync((HParmBlkPtr)&voldata);
		*temporary = sprintf(1+temporary,"%s free on %svolume.",
			sizeStr(buf,voldata.ioVFrBlk*(voldata.ioVAlBlkSiz>>9)),
			voldata.ioVAtrb&(kioVAtrbHardwareLockedMask|kioVAtrbSoftwareLockedMask) ?
			voldata.ioVAtrb&kioVAtrbHardwareLockedMask ? "(locked) ":"(s/w locked) ":"");
		setdialogitem_pstr(dlg,10,(const UInt8*)temporary);
	}
	return false;
}

void GetDirectory(const UInt8 *prompt,
	StandardFileReply *reply)
{
	ModalFilterYDUPP	dirfilterupp = NULL;

	Point		where;
	dirfilterdata	fdata;

	fdata.reply = reply;
	fdata.oldvol = 1;
	fdata.setfocus = true;
	*(SInt32*)&where =-1;
	dirfilterupp = NewModalFilterYDUPP(dirfilter);
	CustomPutFile(prompt,"\p",reply,322,
		where,NULL,dirfilterupp,NULL,NULL,&fdata);
	DisposeModalFilterYDUPP(dirfilterupp);
	*reply->sfFile.name = 0;
}

DialogRef getnewdialog(int dialogid,
	WindowRef frontwindow)
{
	DialogRef	dlg;
	Ptr		storage;

	if ((storage=NewPtrClear(256))==NULL)
		return NULL;
	if ((dlg=GetNewDialog(dialogid,storage,frontwindow))==NULL)
		DisposePtr(storage);
	return dlg;
}

int dialogupdate(DialogRef dlg)
{
	Rect		bounds;
	Handle		ih;
	GrafPtr		oldport;
	short		index, type;

	if (!dlg)
		return -1;
	GetPort(&oldport);
	SetPortDialogPort(dlg);

	DrawDialog(dlg);
	if ((index=GetDialogDefaultItem(dlg))>0)
	{
		GetDialogItem(dlg,index,&type,&ih,&bounds);
		if ((type&~kItemDisableBit)==kButtonDialogItem)
		{
			PenSize(3,3);
			bounds.top-= 4, bounds.left-= 4, bounds.bottom+= 4, bounds.right+= 4;
			FrameRoundRect(&bounds,16,16);
			PenSize(1,1);
		}
	}
	MacSetPort(oldport);

	return 0;
}

int findbuttonitem(DialogRef dlg,
	const UInt8 *name) /* must be a lowercase string */
{
	Str255		title;
	Rect		bounds;
	Handle		ih;
	short		type, item;

	for(item=1; item<8; ++item)
	{
		GetDialogItem(dlg,item,&type,&ih,&bounds);
		if (!ih || (type&~kItemDisableBit)!=kButtonDialogItem)
			break; /* assumed, no more buttons */
		GetControlTitle((ControlRef)ih,title);
		LowercaseText(1+(Ptr)title,*title,0);
		if (!memcmp(name,title,*name+1))
			return item;
	}
	return -1;
}

int animatebuttonitem(DialogRef dlg,
	int item)
{
	Rect		bounds;
	Handle		ih;
	short		type, last, next;

	GetDialogItem(dlg,item,&type,&ih,&bounds);
	if (!ih || (type&~kItemDisableBit)!=kButtonDialogItem)
		return -1;
	HiliteControl((ControlRef)ih,1);

	/* extend from 16 bits in case of overflow (do 32 bits and wait 3 years?) */
	next = 6+TickCount(); /* 6/60 sec. */
	do
		last = TickCount();
	while((last-=next)<0);
	HiliteControl((ControlRef)ih,0);

	next+= 2; /* 2/60 sec. */
	do
		last = TickCount();
	while((last-=next)<0);

	return 1;
}

void setdialogitem_pstr(DialogRef dlg,
	DialogItemIndex item,
	const UInt8 *str)
{
	Rect		bounds;
	Handle		ih;
	short		type;

	GetDialogItem(dlg,item,&type,&ih,&bounds);
	if (!ih)
		return;
	SetDialogItemText(ih,str);
}

local void progressDlg(const UInt8 *str,
	int msg)
{
	local Str255	pending;
	local DialogRef	dlg;
	local short	ltick;

	EventRecord	event;
	short		item_hit;

	if (msg<0)
	{
		if (dlg)
			DisposeDialog(dlg);
		return;
	}
	if (!dlg && (dlg=getnewdialog(1002,(WindowRef)-1))==NULL)
		return;
	if (!str)
		str = "\p";
	if (msg==2)
	{
		setdialogitem_pstr(dlg,1,str);
		setdialogitem_pstr(dlg,2,NULL);
		return;
	}
	if (GetNextEvent(mDownMask|mUpMask|keyDownMask|keyUpMask|autoKeyMask
			|updateMask|activMask,&event)
			&& !modalfilter(dlg,&event,&item_hit)
			&& IsDialogEvent(&event))
		DialogSelect(&event,&dlg,&item_hit);
	if (!ltick)
		ltick = (short)event.when;
	if (msg)
		{memcpy(pending,str,*str+1); return;}
	if (*pending && (short)(event.when-ltick)>(IsWindowVisible(GetDialogWindow(dlg)) ? 31:63))
	{
		ShowWindow(GetDialogWindow(dlg));
		setdialogitem_pstr(dlg,2,pending);
		*pending = 0;
		ltick = (short)event.when;
	}
}
 #endif
#endif

/* fits the effective output path so it'll actually work in the native environment */
/* returns -1 if NULL, 0 if unchanged, bit 0 set if changed, bit 1 set if shortened */
int platformfpath(char *n,char*m)
{
    unsigned char      *d, c;
    int                 r = 0;
(void)c;
    *m = 0;
    if ((d=(unsigned char*)n)==NULL || !*d)
        return -1; /* no path */
#ifdef mac
    if (!memcmp("__MACOSX/",d+1,9))
    {
        char    *s;

        strcpy(n+1,n+10);
        if ((s=strstr(n,"._"))==NULL)
            r =-1; /* bad path */
        else
            strcpy(s,s+2);
        *m = 1;
        /* rsrc established; remove indicators */
    }
    if ((c=*d++)!=0)
    {
        unsigned        z = 0, a = 0; /* name size, hash for unique renaming */

        do
            if (!(c+=-'/') || !(c+='/'-'\\'))
                d[-1] = ':', z = 0, a = 0, r|= 1;
            else if (a=(a<<5)^c,++z>MAXNAMELENGTH)
                d = shortenfpath(n,d,a), r|= 2;
        while((c=*d++)!=0);
    }
#endif
    return r;
}

#ifdef mac
/* Mac OS X header, directly prior to rsrc offset */
local char      sigMacOSX[] =
"\x00\x05\x16\x07\x00\x02\x00\x00\x4D\x61\x63\x20\x4F\x53\x20\x58"
"\x20\x20\x20\x20\x20\x20\x20\x20\x00\x02\x00\x00\x00\x09\x00\x00"
"\x00\x32\x00\x00\x00\x20\x00\x00\x00\x02";
#endif

/* any char outside [0x20,0xFF) -> space */
char *printablefpath(char *name)
{
    char            *d;
    unsigned char   c;

    if ((d=name)==NULL)
        return NULL;
    if ((c=*d++)!=0)
        do
            if ((c+=-0x20)>=0xFF-0x20)
                d[-1] = ' ';
        while((c=*d++)!=0);
    return name;
}


void setfileinfo(const char *filepath,
    const unz_file_info *file_info,
    long *fdinf)
{
#ifdef WIN32
    HANDLE      hFile;
    FILETIME    ftm,ftLocal,ftCreate,ftLastAcc,ftLastWrite;

    hFile = CreateFile(filepath,GENERIC_READ | GENERIC_WRITE,
        0,NULL,OPEN_EXISTING,0,NULL);
    GetFileTime(hFile,&ftCreate,&ftLastAcc,&ftLastWrite);
    DosDateTimeToFileTime((WORD)(file_info->dosDate>>16),(WORD)file_info->dosDate,&ftLocal);
    LocalFileTimeToFileTime(&ftLocal,&ftm);
    SetFileTime(hFile,&ftm,&ftLastAcc,&ftm);
    CloseHandle(hFile);
#else
    struct tm newdate;
    tm_unz tmu_date = file_info->tmu_date;

    newdate.tm_sec = tmu_date.tm_sec;
    newdate.tm_min = tmu_date.tm_min;
    newdate.tm_hour = tmu_date.tm_hour;
    newdate.tm_mday = tmu_date.tm_mday;
    newdate.tm_mon = tmu_date.tm_mon;
    if (tmu_date.tm_year>1900)
        newdate.tm_year = tmu_date.tm_year-1900;
    else
        newdate.tm_year = tmu_date.tm_year;
    newdate.tm_isdst =-1;

    {
        time_t          ziptime = mktime(&newdate);
#ifdef mac
        CInfoPBRec      cinfo;
        unsigned char   name[256];
        int             len = strlen(filepath);

        if (len>255)
            return;
        memcpy(1+name,filepath,len);
        name[0] = len;
        cinfo.hFileInfo.ioCompletion = NULL;
        cinfo.hFileInfo.ioFDirIndex = 0;
        cinfo.hFileInfo.ioFlAttrib = 0;
        cinfo.hFileInfo.ioACUser = 0;
        cinfo.hFileInfo.ioNamePtr = name;
        cinfo.hFileInfo.ioVRefNum = vRefNumLocal;
        cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
        if (PBGetCatInfoSync(&cinfo)!=noErr)
            return;
        cinfo.hFileInfo.ioFlMdDat = ziptime;
        /* only override creation date if new; 2009.11.26 17:46 EST */
        if (!cinfo.hFileInfo.ioFlCrDat
                || cinfo.hFileInfo.ioFlCrDat>cinfo.hFileInfo.ioFlMdDat)
            cinfo.hFileInfo.ioFlCrDat = cinfo.hFileInfo.ioFlMdDat;
        if (!(cinfo.hFileInfo.ioFlAttrib&kioFlAttribDirMask))
        { /* directory positioning in OS X info not compatible */
            /* text!; 2009.07.11 12:32 EST */
            if (file_info->internal_fa&1)
                cinfo.hFileInfo.ioFlFndrInfo.fdType = 'TEXT';
            else
                cinfo.hFileInfo.ioFlFndrInfo.fdType = 0;
            /* Windows bit #1, read-only; bit #14, executable icon - 2012.03.02 */
            cinfo.hFileInfo.ioFlAttrib|= file_info->external_fa&kioFlAttribLockedMask;

            extensionToFileAndCreator(1+name,len,&cinfo.hFileInfo.ioFlFndrInfo.fdType);

            cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
            if (fdinf) /* file positioning in OS X info not compatible */
                memcpy(&cinfo.hFileInfo.ioFlFndrInfo.fdType,fdinf,10);
        }
        PBSetCatInfoSync(&cinfo);
#elif defined(unix)
        struct utimbuf ut;

        ut.actime = ut.modtime = ziptime;
        utime(filepath,&ut);
#endif
    }
#endif
}

#if defined(WIN32) || !defined(unix)
 #define mymkdir(x)     mkdir(x)
#else
 #define mymkdir(x)     mkdir(x,0775)
#endif

int makedir(newdir)
    const char *newdir;
{
    unsigned char      *p, hold;
    char               *buffer;
    int                 len;

    if ((len=(int)strlen(newdir))<=0 || (buffer=(char*)malloc(len+1))==NULL)
        return 0;

    memcpy(buffer,newdir,len);
    buffer[len] = 0;

    if (buffer[len-1]==TPATH_SEP)
        buffer[len-1] = 0;

    if (mymkdir(buffer)==0)
    {
        free(buffer);
        return 1;
    }

    p = (unsigned char*)(1+buffer);

    do
    {
        while((hold=*p)!=0 && hold!='\\' && hold!=TPATH_SEP)
            p++;
        *p = 0;
        if (mymkdir(buffer)==-1 && errno==ENOENT)
        {
            fprintf(stderr,"couldn't create directory '%s'\n",
                buffer);
            free(buffer);
            return 0;
        }
    } while((*p++=hold)!=0);

    free(buffer);
    return 1;
}

/* main code */

local unsigned char     *extr_buf;

local const char *password = NULL;
local char        opt_overwrite, opt_extract_without_path, opt_ignore_dsstore = 1;
local long        extr_files;
#ifdef mac
local UInt8       volnam[128];
local long        oldpar;
local short       oldvol;
#endif

local void do_banner()
{
    fputs(
"MiniUnz 1.01b+, demo of zLib + Unz package written by Gilles Vollant\n"
"more info at http://www.winimage.com/zLibDll/unzip.html\n"
" - src Mod-Date " LAST_REVISION_DATE "\n"
#ifdef APPTOOL
" - app version by Adam P.\n"
#endif
"\n",
        stdout);
}

local void do_help()
{
    fputs(
"Usage : miniunz [options] file.zip [file_to_extr.] [-d extractdir]\n"
"\n"
"  -e  Extract without paths\n"
"  -x  Extract with path\n"
"  -v  list files\n"
"  -l  list files\n"
"  -d  directory to extract into\n"
"  -o  overwrite files without prompting\n"
"  -p  extract crypted file using password\n"
"  -r  ignore '.DS_Store' files\n"
"\n",
        stdout);
}


local int do_list(uf,vb)
    unzFile uf;
    int vb;
{
    unz_global_info     gi;
    uLong               i;
    uLong               n = 0, total_compressed = 0, total_uncompressed = 0;
    int                 flagMerge = 0;
    int                 err;

    if ((err=unzGetGlobalInfo(uf,&gi))!=UNZ_OK)
        fprintf(stderr,"error %d with zipfile in unzGetGlobalInfo%s.\n",
            err,zError(err));

    fputs(vb ?
"  Length   Method   Size      Date      Time    CRC-32    Vers     Info / ExAttr  Name\n"
"---------- ------ -------- ---------- -------- -------- -------- ---------------- ----\n"
:
"  Length   Method   Size   Ratio    Date / Time    Name\n" /* non-verbose no longer prints CRC */
"---------- ------ -------- ----- ----------------  ----\n",
            stdout);

    for(i=0; i<gi.number_entry; ++i)
    {
        char            filename_inzip[256];
        unz_file_info   file_info;
        uLong           ratio = 0;
 
        /* ["Unkn. " - by default to make sure it's set]; 2009.07.11 */
        const char     *string_method = "??? ";
        char            charCrypt = ' ';

        if (signalbreak())
            return -128;
        if ((err=unzGetCurrentFileInfo(uf,&file_info,filename_inzip,
                sizeof(filename_inzip),NULL,0,NULL,0)) != UNZ_OK)
        {
            fprintf(stderr,"error %d with zipfile in unzGetCurrentFileInfo%s\n",
                err,zError(err));
            break;
        }
        if (file_info.uncompressed_size>0)
        {
            ++n;
            total_compressed+= file_info.compressed_size;
            total_uncompressed+= file_info.uncompressed_size;
            ratio = (file_info.compressed_size*100)/file_info.uncompressed_size;
        }
        /* display a '*' if the file is crypted */
        if (file_info.flag&1)
            charCrypt = '*';

        if (!file_info.compression_method)
            string_method = "Raw "; /*"Stored"*/
        else if (file_info.compression_method==Z_DEFLATED)
            switch(file_info.flag>>1&3)
            {
            default: string_method="Dfl:N"; break;
            case 1: string_method="Dfl:X"; break;
            case 2: string_method="Dfl:f"; break;
            case 3: string_method="Dfl:F"; break;
            }

        if (vb)
            printf(
"%10lu %5s%c%9lu %4.2lu.%2.2lu.%2.2lu %2.2lu:%2.2lu:%2.2lu %8.8lX %8.8lX %7.7lX%c%8.8lX %s\n",
                    file_info.uncompressed_size,
                    string_method,
                    charCrypt,
                    file_info.compressed_size,
                    (uLong)file_info.tmu_date.tm_year,
                    (uLong)file_info.tmu_date.tm_mon + 1,
                    (uLong)file_info.tmu_date.tm_mday,
                    (uLong)file_info.tmu_date.tm_hour,
                    (uLong)file_info.tmu_date.tm_min,
                    (uLong)file_info.tmu_date.tm_sec,
                    (uLong)file_info.crc,
                    (uLong)file_info.version<<16|file_info.version_needed,
                    (uLong)(file_info.flag&~15)<<12|file_info.internal_fa,
                    (char)(file_info.flag&8 ? '+':' '),
                    (uLong)file_info.external_fa,
                    filename_inzip);
        else
            printf(
"%10lu %5s%c%9lu %3lu%%  %4.2lu.%2.2lu.%2.2lu %2.2lu:%2.2lu  %s\n",
                    file_info.uncompressed_size,
                    string_method,
                    charCrypt,
                    file_info.compressed_size,
                    ratio,
                    (uLong)file_info.tmu_date.tm_year,
                    (uLong)file_info.tmu_date.tm_mon + 1,
                    (uLong)file_info.tmu_date.tm_mday,
                    (uLong)file_info.tmu_date.tm_hour,
                    (uLong)file_info.tmu_date.tm_min,
                    filename_inzip);

        if ((i+1)<gi.number_entry)
            if ((err=unzGoToNextFile(uf))!=UNZ_OK)
            {
                fprintf(stderr,"error %d with zipfile in unzGoToNextFile%s\n",
                    err,zError(err));
                break;
            }
        flagMerge|= file_info.flag;
    }

    if (n>vb) /* no point in displaying totals/ratios more than once */
        printf(
"%10lu %15u %3lu%%  Total\n",
                total_uncompressed,
                total_compressed,
                (uLong)(total_compressed*100./total_uncompressed));
    if (vb && flagMerge&1)
        fputs("\n * contains password-encrypted file(s)\n",stdout);
    return err;
}

local int ask_overwrite(char *name)
{
#ifdef APPTOOL
    Str255      pname;
    int         i;
 /* truncate from start */
    i = strlen(name);
    *pname = i>240 ? 240:i;
    memcpy(pname+1,name+i-*pname,i);
    if (*pname!=i)
        pname[1] = '';
 /* should add 'alternate name' option; StuffIt Expander adds a dot and an index */
    ParamText(pname,NULL,NULL,NULL);
    i = Alert(1000,modalfilterupp);
    return --i ? --i ? 'a':0:'n';
#else
    int         rep;

    do
    {
        char        answer[8];
        /* single line to work with MPW; 2009.07.10 */
        printf("The file '%s' exists. Overwrite ? [y]es, [n]o, [A]ll:\n",
            name);

        if (!fgets(answer,sizeof(answer)-1,stdin) || !*answer)
            exit(EXIT_FAILURE);

        rep = *answer|32;
    } while(rep!='y' && rep!='n' && rep!='a');
    return rep;
#endif
}

local char *getpathname(char *s)
{
    unsigned char  *p = (unsigned char*)s;
    unsigned char   c;

    while((c=*p++)!=0)
        if (c=='/' || c=='\\')
            s = (char*)p;
    return s;
}

local int extract_currentfile(unzFile uf,
    int archx,
    int archc)
{
#pragma unused(archx,archc)
    unz_file_info file_info;
    char         filepath[TPATH_EXTRA + MAXPATHLENGTH];
    char        *filename, osx;
    int          rtv = UNZ_OK;
    FILE        *fout = NULL;
    int          nativepath_result;
    uLong        ratio = 0;

    memset(&file_info,0,sizeof(file_info));

    if ((rtv=unzGetCurrentFileInfo(uf,&file_info,TPATH_EXTRA + filepath,
            MAXPATHLENGTH,NULL,0,NULL,0))!=UNZ_OK)
    {
        fprintf(stderr,"error %d with zipfile in unzGetCurrentFileInfo%s\n",
            rtv,zError(rtv));
        return rtv;
    }

    if (!extr_buf)
        if ((extr_buf=malloc(WRITEBUFFERSIZE))==NULL)
        {
            fputs("Error allocating decompression memory\n",stderr);
            return -2;
        }

#if TPATH_EXTRA
    *filepath = TPATH_SEP;
#endif

    filename = getpathname(filepath);
    /* fix: if item has a size, it's definitely not a directory */
    if (!*filename && filename>filepath && file_info.uncompressed_size)
        filename[-1] = 0, filename = getpathname(filepath);

    if (opt_ignore_dsstore && !strcmp(filename,".DS_Store"))
        return UNZ_OK;

    if ((nativepath_result=platformfpath(filepath,&osx))<0)
    {
        if (*filename)
            printf("bad modern path '%s'\n",
                filepath);
        return UNZ_OK;
    }

    if (!*filename)
    {
        if (!opt_extract_without_path)
        {
            if (!osx)
                printf("creating directory '%s'\n",
                    filepath);
            mymkdir(filepath);
        }
    } else
    {
        long    fdinf[6];
        char   *write_filename;
        int     skip = 0;

        memset(fdinf,0,sizeof(fdinf));

        if (!opt_extract_without_path)
            write_filename = filepath;
        else
            write_filename = filename;

        if ((rtv=unzOpenCurrentFilePassword(uf,password))!=UNZ_OK)
        {
            fprintf(stderr,"error %d opening '%s' in zipfile%s\n",
                rtv,filepath,rtv==UNZ_INTERNALERROR ? " with password"
                :zError(rtv));
            return rtv;
        }

        if (!opt_overwrite && rtv==UNZ_OK && !osx)
        {
            char        rep = 0;
            FILE        *ftestexist;

            if ((ftestexist=fopen(write_filename,"rb"))!=NULL)
            {
                fclose(ftestexist);
                rep = ask_overwrite(write_filename);
            }
            if (rep=='n')
                skip = 1;
            if (rep=='a')
                opt_overwrite = 1;
        }

        if (!skip && rtv==UNZ_OK)
        {
            /* some zipfiles don't contain directory alone before file */
        #ifdef mac
            if (osx)
                fout = fdopen(open(write_filename,O_RDWR|O_BINARY|O_RSRC),"wb");
            else
        #endif
                fout = fopen(write_filename,"wb");
            if (!fout && !opt_extract_without_path
                    && filename!=(char*)filepath)
            {
                char    c = filename[-1];

                filename[-1] = 0;
                makedir(write_filename);
                filename[-1] = c;
                fout = fopen(write_filename,"wb");
            }

            if (!fout)
                fprintf(stderr,"error %d opening '%s'%s\n",
                    errno,write_filename,errnoError(errno));
        }

        if (fout)
        {
            printf(" %s '%s'\n",
                osx ? "extracting info/rsrc":nativepath_result>1 ? "extracting (as)":"extracting",
                write_filename);
        #ifdef APPTOOL
            rtv = sprintf(1+temporary,"%3d/%d: '%s'",
                archx,archc,filepath);
            *temporary = rtv>255 ? 255:rtv;
            progressDlg((const UInt8*)temporary,1);
        #endif
            /* Mac OS X file info/resource fork processing
              (corrections aided by Paul C. Pratt's 'MiniUnZp:source:app.c',
               which isn't very easy to read - all unzip code in one file)
            */
            /* offset for rsrc fork here was assumed to always be 82
              (until I saw one rfork coming out with extra leading data,
               and the type/creator was hackishly taken from the rfork)
            */
            if (osx)
            { /* read and skip OSX signature/version */
                if ((rtv=unzReadCurrentFile(uf,extr_buf,sizeof(sigMacOSX)-1))<0)
                    goto errc;
                if (memcmp(extr_buf,sigMacOSX,sizeof(sigMacOSX)-1))
                {
                    fputs(" - *skipped* - Mac OS X sign/vers mismatch\n",
                        stdout);
                    goto skipo;
                }
                /* read file/directory info: rsrc offset/length, flags.w, icon position */
                /* Finder directory info ignored due to larger screen sizes */
                if ((rtv=unzReadCurrentFile(uf,extr_buf,24))<0)
                    goto errc;
                /* unpack in an endian-independent fashion */
                *fdinf = *extr_buf<<24|extr_buf[1]<<16|extr_buf[2]<<8|extr_buf[3];
                fdinf[1] = extr_buf[4]<<24|extr_buf[5]<<16|extr_buf[6]<<8|extr_buf[7];
                fdinf[2] = *(long*)(8+extr_buf);
                fdinf[3] = *(long*)(12+extr_buf);
                fdinf[4] = extr_buf[16]<<24|extr_buf[17]<<16|extr_buf[18]<<8|extr_buf[19];
                fdinf[5] = extr_buf[20]<<24|extr_buf[21]<<16|extr_buf[22]<<8|extr_buf[23];
                /* the creation date is not stored in this file, only the FInfo */

                if (fdinf[1]<=0 || *fdinf<=0)
                    goto skipo; /* no rsrc data */
                if ((*fdinf-=sizeof(sigMacOSX)-1+24)>WRITEBUFFERSIZE)
                {
                    fputs(" - *skipped* - bad rsrc offset\n",
                        stdout);
                    goto skipo;
                }
                if ((rtv=unzReadCurrentFile(uf,extr_buf,*fdinf))<0)
                    goto errc;
            }
            do
            {
                if ((rtv=unzReadCurrentFile(uf,extr_buf,WRITEBUFFERSIZE))<0)
                errc:
                    fprintf(stderr,"error %d in reading zipfile%s\n",
                        rtv,rtv==Z_DATA_ERROR && file_info.flag&1 ? " - password mismatch / bad data"
                        :zError(rtv));
                else if (rtv)
                {
                #ifdef APPTOOL
                    progressDlg(NULL,0); /* event cycles */
                #endif
                    if (fdinf[1])
                        if (rtv>fdinf[1])
                            rtv = fdinf[1], fdinf[1] =-1; /* final */
                        else
                            fdinf[1]-= rtv;
                    if (fwrite(extr_buf,rtv,1,fout)!=1)
                    {
                        fprintf(stderr,"error %d in writing file%s\n",
                            errno,errnoError(errno));
                        rtv = UNZ_ERRNO;
                        break;
                    }
                }
                if (fdinf[1]<0)
                {
                skipo:
                    rtv = 0;
                }
            } while(rtv>0 && fdinf[1]>=0);

            if (fout)
                fclose(fout);
            if (!rtv)
                setfileinfo(write_filename,&file_info,fdinf[2]|fdinf[3] ? 2+fdinf:NULL);
        }

        if (rtv==UNZ_OK)
        {
            if ((rtv=unzCloseCurrentFile(uf))!=UNZ_OK)
                fprintf(stderr,"error %d with zipfile in unzCloseCurrentFile%s\n",
                    rtv,zError(rtv));
        } else
            unzCloseCurrentFile(uf); /* don't lose the error */
    }
    return rtv==Z_DATA_ERROR && file_info.flag&1 ? UNZ_OK:rtv;
}


local int do_extract(unzFile uf)
{
    unz_global_info     gi;

    FILE       *fout = NULL;
    uLong       i;
    int         err;

    if ((err=unzGetGlobalInfo(uf,&gi))!=UNZ_OK)
        fprintf(stderr,"error %d with zipfile in unzGetGlobalInfo%s\n",
            err,zError(err));
    for(i=0; i<gi.number_entry; )
    {
        if (!(++i&4)) /* some manual screen update; 2010.03.03 15:21 EST */
            fflush(stdout);
        if (signalbreak()) /* if user wants to stop; 2010.04.10 15:16 EDT */
            return -128;
        if ((err=extract_currentfile(uf,i,gi.number_entry))!=UNZ_OK)
            break;
        ++extr_files;
        if (i<gi.number_entry)
            if ((err=unzGoToNextFile(uf))!=UNZ_OK)
            {
                fprintf(stderr,"error %d with zipfile in unzGoToNextFile%s\n",
                    err,zError(err));
                break;
            }
    }
    return err;
}

local int do_extract_onefile(uf,filepath)
    unzFile uf;
    const char *filepath;
{
    int         err = UNZ_OK;

    if (unzLocateFile(uf,filepath,CASESENSITIVITY)!=UNZ_OK)
    {
        printf("file '%s' not found in the zipfile\n",
            filepath);
        return 2;
    }
    return extract_currentfile(uf,1,1)!=UNZ_OK;
}

local unzFile unzAltOpen(const char *zipfilepath,
    char *altpathbuf)
{
    unzFile     uf;

    if (!zipfilepath)
        return NULL;
    {
#ifdef USEWIN32IOAPI
        zlib_filefunc_def   ffunc;
#endif
        if (!altpathbuf)
        {
#ifdef USEWIN32IOAPI
            fill_win32_filefunc(&ffunc);
            return unzOpen2(zipfilepath,&ffunc);
#else
            return unzOpen(zipfilepath);
#endif
        }

        strncpy(altpathbuf,zipfilepath,MAXPATHLENGTH-1);
        /* strncpy doesn't append the trailing null if the string is too long. */
        altpathbuf[ MAXPATHLENGTH ] = 0;

#ifdef USEWIN32IOAPI
        fill_win32_filefunc(&ffunc);
        if ((uf=unzOpen2(zipfilepath,&ffunc))==NULL)
#else
        if ((uf=unzOpen(zipfilepath))==NULL)
#endif
        {
            strcat(altpathbuf,".zip");
#ifdef USEWIN32IOAPI
            uf = unzOpen2(altpathbuf,&ffunc);
#else
            uf = unzOpen(altpathbuf);
#endif
        }
    }
    return uf;
}

local int do_zipfile(int msg,
    unzFile uf,
    const char *zipfilepath,
    const char *filename_to_extract,
    const char *dirname)
{
    int         i;
#ifdef APPTOOL
    printf("'%s'\n\n",
#else
    printf("'%s' opened\n",
#endif
        zipfilepath);
    fflush(stdout);

    i = 0;

    if (msg<2)
    {
#ifdef mac
        FSSpec  spec;

        if (dirname && (path2spec(dirname,strlen(dirname),&spec)
                || HSetVol(NULL,spec.vRefNum,spec.parID)))
#else
        if (dirname && chdir(dirname))
#endif
        {
            fprintf(stderr,"Error changing into '%s'; aborting\n",
                dirname);
            exit(-1);
        }
        opt_extract_without_path = msg&1;
        if (!filename_to_extract)
            i = do_extract(uf);
        else
            i = do_extract_onefile(uf,filename_to_extract);
    } else if (msg<4)
        i = do_list(uf,msg-2);

    if (extr_buf)
        {free(extr_buf); extr_buf = NULL;}
    unzCloseCurrentFile(uf);
#ifndef APPTOOL
    printf("\n'%s' closed.\n\n",
        zipfilepath);
#endif
    return i;
}


/* APPTOOL added 2014.04.02 */
#if defined(APPTOOL) && defined(mac)
/* makes room in a Handle, removing the range xs-xe, displacing if necessary */
/* the empty space is undefined */
local long mungeprep(Handle h,
	long xs,
	long xe,
	int kadd)
{
	long		endc, z = GetHandleSize(h), oldz = z;
	int		newsls = kadd+xs;

	if (xs>xe || xs>z)
		return -1;
	endc = z-=xe, z+= newsls<0 ? 0:newsls;
	if (endc<=0) /* from end */
		endc = 0;
	else if (newsls>=0) /* no change/negative displacement */
	{
		if (z>oldz)
		{
			SetHandleSize(h,z);
			if (MemError()<0)
				return -2;
			oldz =-1;
		}
		HLock(h);
		memmove(*h+xs+kadd,*h+xe,endc);
		HUnlock(h);
	}
	if (oldz>=0)
	{
		SetHandleSize(h,z);
		if (MemError()<0)
			return -2;
	}
	return xe;
}

/* the actual text is stored in the pw Handle;
 all chars added to the TEHandle are instead bullets
*/
local int passwordevent(EventRecord *event,
	TEHandle te,
	Handle pw)
{
	long		sels, sele, z;
	int		cc;

	if (!te || !pw || !event || (event->what!=keyDown && event->what!=autoKey))
		return -1;
	cc = event->message&charCodeMask;
	z = event->modifiers&cmdKey ? 0:cc==kBackspaceCharCode ? -1:1;
	event->what = nullEvent;
	sels = (**te).selStart, sele = (**te).selEnd;
	fflush(stdout);
	if (!z && (cc=='c' || cc=='x'))
	{ /* cut/copy */
		ZeroScrap();
		HLock(pw);
		PutScrap(sele-sels,kScrapFlavorTypeText,*pw+sels);
		HUnlock(pw);
		if (cc=='c')
			return 0;
	}
	if (z)
	{ /* key */
		if ((sele=mungeprep(pw,sels,sele,z))<0)
			return sele;
		if (cc!=kBackspaceCharCode)
			(*pw)[sels] = cc, cc = kBulletCharCode;
		TEKey(cc,te);
	}
	if (cc=='v')
	{ /* GetScrap sucks */
		Handle		h = NewHandle(z=0);

		if (!h)
			return -2;
		GetScrap(h,kScrapFlavorTypeText,&z);
		z = GetHandleSize(h);
		if ((sele=mungeprep(pw,sels,sele,z))<0)
			goto fail;
		memcpy(*pw+sels,*h,z);
		SetHandleSize(h,z);
		if ((sele=MemError())<0)
		{
		fail:	DisposeHandle(h);
			return sele;
		}
		HLock(h);
		memset(*h,kBulletCharCode,z);
		TEInsert(*h,z,te);
		DisposeHandle(h);
	}
	return 1;
}

local Handle		passwh;

local pascal Boolean passwfilter(DialogRef dlg,
	EventRecord *event,
	short *item_hit)
{
	if (event->what==keyDown || event->what==autoKey)
	{
		switch(event->message&charCodeMask)
		{
		case 'a':
		case 'e':
		case 'l':
			if (event->modifiers&cmdKey)
		case kEnterCharCode:
		case kTabCharCode:
		case kPageUpCharCode:
		case kPageDownCharCode:
		case kReturnCharCode:
		case kFunctionKeyCharCode:
		case kLeftArrowCharCode:
		case kRightArrowCharCode:
		case kDownArrowCharCode:
		case kUpArrowCharCode:
		case kEscapeCharCode:
				goto normal;
		}
		if (GetDialogKeyboardFocusItem(dlg)==4)
		{ /* don't take names (don't kick ass, either) */
			passwordevent(event,GetDialogTextEditHandle(dlg),passwh);
			return true;
		}
	}
normal:	return modalfilter(dlg,event,item_hit);
}

local int optionsdialog(short dialogid,
	char *specfile) /* up to 511 characters */
{
	Rect		bounds;
	Handle		ih;
	ModalFilterUPP	passwfilterupp = NULL;
	DialogRef	dlg = NULL;
	GrafPtr		oldport;
	long		size;
	short		index =-1, type;

	if (specfile)
		*specfile = 0;
	if ((passwfilterupp=NewModalFilterUPP(passwfilter))==NULL)
		return -1;
	if (dialogid<0)
		goto clear;
	if (passwh)
	{
		HUnlock(passwh);
		SetHandleSize(passwh,0);
	} else
		passwh = NewHandle(0);

	GetPort(&oldport);

	ParamText("\p","\p","\p","\p");

	if ((dlg=getnewdialog(dialogid,(WindowRef)-1))==NULL)
		goto end;

	GetDialogItem(dlg,6,&type,&ih,&bounds);
	if (ih)
		SetControlValue((ControlRef)ih,opt_extract_without_path);
	GetDialogItem(dlg,7,&type,&ih,&bounds);
	if (ih)
		SetControlValue((ControlRef)ih,opt_ignore_dsstore);

	MacShowWindow(GetDialogWindow(dlg));
	SetPortDialogPort(dlg);

	SetDialogTracksCursor(dlg,true);

	do
	{
		ModalDialog(passwfilterupp,&index);
		if (index-6<2u)
		{
			GetDialogItem(dlg,index,&type,&ih,&bounds);
			if (ih)
				SetControlValue((ControlRef)ih,!GetControlValue((ControlRef)ih));
		}
	} while(index-1>2u);

	GetDialogItem(dlg,5,&type,&ih,&bounds);
	if (ih && specfile)
	{
		if ((size=GetHandleSize(ih))>511)
			size = 511;
		specfile[size] = 0;
		memcpy(specfile,*ih,size);
	}
	GetDialogItem(dlg,6,&type,&ih,&bounds);
	if (ih)
		opt_extract_without_path = GetControlValue((ControlRef)ih);
	GetDialogItem(dlg,7,&type,&ih,&bounds);
	if (ih)
		opt_ignore_dsstore = GetControlValue((ControlRef)ih);

	DisposeDialog(dlg);
	InitCursor();

	SetPort(oldport);

	if (index==3)
		index =-128;
	if (index==-128 || (size=GetHandleSize(passwh))==0)
	{
	clear:	if (passwh)
			{DisposeHandle(passwh); passwh = NULL;}
		password = NULL;
		goto end;
	}
	SetHandleSize(passwh,size+1);
	if (MemError()<0)
		return -2;
	HLock(passwh);
	(*passwh)[size] = 0;
	password = *passwh;
end:	if (passwfilterupp)
		DisposeModalFilterUPP(passwfilterupp);
	if (index>0)
		index = index==1 ? 0:3;
	return index;
}

/* app */

QDGlobals	qd;
int		files_taken, arch_success;
Boolean		done_banner, fatal_error;

local void do_file(const FSSpec *file)
{
	StandardFileReply	reply;

	char		path[512], specfile[512], prompt[64];
	KeyMap		map;
	int		msg, err =-128;

	if (!done_banner)
		{do_banner(); done_banner = true;}
	if (!file)
	{
		StandardGetFile(NULL,4,(const OSType*)"ZIP BINA    \0\0\0",&reply);
		if (!reply.sfGood)
			return;
		file = &reply.sfFile;
	}
	++files_taken;
	GetKeys(map);
	msg = 0, *specfile = 0;
	if (((const UInt8*)map)[48>>3]&128) /* cmd */
		msg = optionsdialog(1001,specfile);
	if ((err=msg)>=0)
	{
		unz_global_info	gi;
		unzFile		uf;

		spec2path(file,path,511);
		if ((uf=unzAltOpen(path,NULL))==NULL)
		{
			fprintf(stderr,"Failed to open '%s' as an archive.\n",
				path);
			return;
		}
		if (msg<2)
		{
			if ((err=unzGetGlobalInfo(uf,&gi))!=UNZ_OK)
			{
				fprintf(stderr,"Error %d getting info for '%s'.\n",
					err,path,zError(err));
				unzClose(uf);
				return;
			}
			*prompt = sprintf(1+prompt,
"%d item%s in archive.\n"
"Select destination\n",
				gi.number_entry,gi.number_entry==1 ? "":"s");
			GetDirectory((const UInt8*)prompt,&reply);
			if (reply.sfGood)
				HSetVol(NULL,reply.sfFile.vRefNum,reply.sfFile.parID);
			else
				err =-128;
		}
		if (err>=0)
		{
			if (msg<2)
				progressDlg("\pExtracting...  (press cmd-period to stop)",2);
			err = do_zipfile(msg,uf,path,*specfile ? specfile:NULL,NULL);
		}
		unzClose(uf);
	}
	if (err)
	{
		if (err==-128)
			printf("User canceled (%d).\n",
				err);
		else
		{
			if (err==UNZ_ERRNO && errno==ENOSPC)
				Alert(999,modalfilterupp);
			fprintf(stderr,"### fatal error (%d)\n",
				err);
		}
		fatal_error = 1;
		return;
	}
	++arch_success;
}

local OSErr required_appleevent_check(const AppleEvent *event)
{
	DescType	typeCode;
	Size		actualSize;
	OSErr		err;

	if ((err=AEGetAttributePtr(event,keyMissedKeywordAttr,typeWildCard,
			&typeCode,0,0,&actualSize))==errAEDescNotFound)
		return noErr;
	if (!err)
		return errAEEventNotHandled;
	return err;
}

local pascal OSErr handle_open_document(const AppleEvent *event,
	AppleEvent *reply,
	long myRefCon)
{
#pragma unused(reply,myRefCon)
	AEDescList	docList;
	FSSpec		file;
	Size		actualSize;
	long		index, itemsInList;
	OSErr		err;

	if ((err=AEGetParamDesc(event,keyDirectObject,typeAEList,&docList))!=noErr)
		return err;
	if ((err=required_appleevent_check(event))!=noErr)
		return err;
	if ((err=AECountItems(&docList,&itemsInList))!=noErr)
		return err;
	for(index=1; !fatal_error && index<=itemsInList; ++index) 
	{
		DescType	typeCode;
		AEKeyword	theKeyword;

		if ((err=AEGetNthPtr(&docList,index,typeFSS,&theKeyword,&typeCode,(Ptr)&file,
				sizeof(FSSpec),&actualSize))!=noErr)
			continue;
		if (files_taken)
			fputc('\n',stdout);
		do_file(&file);
	}
	return err;
}

local pascal OSErr handle_quit_application(const AppleEvent *event,
	AppleEvent *reply, 
	long myRefCon)
{
#pragma unused(reply,myRefCon)
	OSErr		err;

	if ((err=required_appleevent_check(event))!=noErr)
		return err;
	fputs("Program exited as result of a quit event.",stdout);
	exit(1);

	return err;
}

local pascal OSErr handle_open_application(const AppleEvent *event,
	AppleEvent *reply, 
	long myRefCon)
{
#pragma unused(reply,myRefCon)
	return required_appleevent_check(event);
}

local void initialize_core_events(void)
{
	AEEventHandlerUPP	upp;

	if ((upp=NewAEEventHandlerUPP(handle_open_document))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEOpenDocuments,upp,0,false);
	if ((upp=NewAEEventHandlerUPP(handle_quit_application))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEQuitApplication,upp,0,false);
	if ((upp=NewAEEventHandlerUPP(handle_open_application))!=NULL)
		AEInstallEventHandler(kCoreEventClass,kAEOpenApplication,upp,0,false);
}

local void wait_for_highlevel_event()
{
	EventRecord	event;
	int		retry = 60;
	/* WaitNextEvent is supposed to exist in Sys 7 */
	do
		if (WaitNextEvent(highLevelEventMask,&event,0,NULL))
			{AEProcessAppleEvent(&event); break;}
	while(--retry);
}
#endif


int main(argc,argv)
    int argc;
    char *argv[];
{
#ifdef APPTOOL
 #pragma unused(argc,argv)
 #if !TARGET_API_MAC_CARBON && TARGET_OS_MAC
    MaxApplZone();
    InitGraf((Ptr)&qd.thePort);
    InitWindows();
    InitMenus();
    TEInit();
    InitDialogs(0);
 #endif
    HGetVol(volnam,&oldvol,&oldpar); /* save current directory */

    InitCursor();

    freopen("miniunz.out","w",stdout);
    modalfilterupp = NewModalFilterUPP(modalfilter);

    initialize_core_events();
    wait_for_highlevel_event();

    if (!files_taken)
        do_file(NULL);
    progressDlg(NULL,-1);

    if (extr_buf)
        {free(extr_buf); extr_buf = NULL;}

    HSetVol(NULL,oldvol,oldpar);

    if (!ftell(stdout))
        {fclose(stdout); remove("miniunz.out");}
    else if (files_taken)
    {
        if (extr_files)
        printf(
"\n%d total file%s extracted.",
		extr_files,extr_files==1 ? "":"s");
        printf(
"\n%d file%s taken; %d successfully processed.\n"
"\Done.\n",
		files_taken,files_taken==1 ? "":"s",arch_success);
    }
    return 0;
#else
    char        altpathbuf[MAXPATHLENGTH+16];

    unzFile     uf;
    const char *dirname = NULL;
    const char *zipfilepath = NULL;
    const char *filename_to_extract = NULL;
    int         msg = 0;
    int         i;
#ifdef mac
    HGetVol(volnam,&oldvol,&oldpar); /* save current directory */
#endif
    do_banner();

    if (argc==1)
    {
        do_help();
        return 0;
    }
    for(i=1; i<argc; ++i)
        if ((*argv[i])=='-')
        {
            const unsigned char *p = (const unsigned char*)(1+argv[i]);
            unsigned char        c;

            while((c=*p++)!=0)
                switch(c|32|(argc>i+1)<<8)
                {
                case 'x': msg = 0; break;
                case 'e': msg = 1; break;
                case 'l': msg = 2; break;
                case 'v': msg = 3; break;
                case 'o': opt_overwrite = 1; break;
                case 'd'|0x100: dirname = argv[++i]; break;
                case 'p'|0x100: password = argv[++i]; break;
                case 'r': opt_ignore_dsstore = 1;
                }
        } else if (!zipfilepath)
            zipfilepath = argv[i];
        else if (!filename_to_extract && !dirname)
            filename_to_extract = argv[i];
    uf = unzAltOpen(zipfilepath,altpathbuf);
    if (!uf)
    {
        fprintf(stderr,"Cannot open '%s'[.zip]\n",
            zipfilepath);
        return 1;
    }
    i = do_zipfile(msg,uf,altpathbuf,
        filename_to_extract,dirname);
    unzClose(uf);
 #ifdef mac
    HSetVol(NULL,oldvol,oldpar);
    return i;
 #endif
#endif
}
