/*
   minizip.c
   Version 1.01e, February 12th, 2005

   Copyright (C) 1998-2005 Gilles Vollant

Mods by Adam P. (2012.03.02, 5.25)
Fixed bad gramar as well.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>

#ifdef mac
# include ":::zlib.h"
# include <Files.h>
/* so the user can cancel; 2009.09.27 09:54 EDT */
# define signalbreak()  ((*((volatile UInt8*)0x179)&*((volatile UInt8*)0x17A))!=0)
# define TPATH_SEP      ':'
# define TPATH_EXTRA    1
#else
# define signalbreak()  0
# define TPATH_SEP      '/'
# define TPATH_EXTRA    0

# ifdef unix
#  include <unistd.h>
#  include <utime.h>
#  include <sys/types.h>
#  include <sys/stat.h>
# else
#  include <direct.h>
#  include <io.h>
# endif
#endif

#include "zip.h"

#ifdef WIN32
# define USEWIN32IOAPI
# include "iowin32.h"
#endif


#define WRITEBUFFERSIZE (16384)
#define MAXFILENAME (256)

char *zipfpath          OF((char*));
uLong getfileinfo       OF((const char*,tm_zip*,uLong*,uLong*,uLong*));

int check_exist_file    OF((const char*));

void do_banner          OF((void));
void do_help            OF((void));

int getFileCrc          OF((const char*,void*,unsigned long,unsigned long*));


/* translates native path to zip path */
char *zipfpath(char *name)
{
    unsigned char   c, *d;

(void)c;
    if ((d=(unsigned char*)name)==NULL)
        return NULL;
#ifdef mac
    if ((c=*d++)!=0)
        do
        {
            if ((c+=-':')==0)
                d[-1] = '/';
            else if ((c+=':'-'/')==0)
                d[-1] = '-';
        } while((c=*d++)!=0);
#endif
    return name;
}

uLong getfileinfo(f, tmzip, dt, ia, ea)
    const char *f;         /* name of file to get info of */
    tm_zip *tmzip;         /* return value: access, mod. and creation datetimes */
    uLong *dt;             /* dostime */
    uLong *ia;             /* internal attributes */
    uLong *ea;             /* external attributes */
{
  int ret = 0;
#ifdef WIN32
  FILETIME ftLocal;
  HANDLE hFind;
  WIN32_FIND_DATA  ff32;

  hFind = FindFirstFile(f,&ff32);
  if (hFind != INVALID_HANDLE_VALUE)
  {
    FileTimeToLocalFileTime(&(ff32.ftLastWriteTime),&ftLocal);
    FileTimeToDosDateTime(&ftLocal,((LPWORD)dt)+1,((LPWORD)dt)+0);
    FindClose(hFind);
    ret = 1;
  }
#elif defined(unix) || defined(mac)
# ifdef unix
  struct stat s;        /* results of stat() */
# endif
  struct tm* filedate;
  time_t tm_t=0;

  if (strcmp(f,"-")!=0)
  {
    char name[MAXFILENAME + 1 + TPATH_EXTRA];
    int len;

    if ((len=strlen(f)) > MAXFILENAME)
      len = MAXFILENAME;

    strncpy(TPATH_EXTRA+name, f, MAXFILENAME-1);
    /* strncpy doesn't append the trailing null if the string is too long. */
    name[ TPATH_EXTRA + MAXFILENAME ] = '\0';

    if (name[len + (TPATH_EXTRA-1)] == TPATH_SEP)
      name[len + (TPATH_EXTRA-1)] = '\0';
#ifdef mac
    {
      CInfoPBRec    cinfo;
      char          *p;
      long          parid = 0;

      cinfo.hFileInfo.ioCompletion = NULL;
      cinfo.hFileInfo.ioVRefNum = 0x8003; /*vRefNumLocal*/
      cinfo.hFileInfo.ioFDirIndex = 0;
      cinfo.hFileInfo.ioACUser = 0;

      if (len>=255)
      {
        if ((p=strrchr(1+name,':'))==NULL || (name[0]=p-name)<=1)
          goto skip;
        len -= name[0], ++len;
        cinfo.hFileInfo.ioFlAttrib = 0;
        cinfo.hFileInfo.ioNamePtr = (unsigned char*)p;
        cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = 0;
        if (PBGetCatInfoSync(&cinfo)!=noErr)
          goto skip;
        parid = cinfo.hFileInfo.ioFlParID;
      }
      else
        p = name;

      name[0] = len;
      cinfo.hFileInfo.ioFlAttrib = 0;
      cinfo.hFileInfo.ioNamePtr = (unsigned char*)p;
      cinfo.hFileInfo.ioDirID = cinfo.hFileInfo.ioFlParID = parid;
      if (PBGetCatInfoSync(&cinfo)==noErr)
      { /* using native localtime; no delta needed */
        tm_t = cinfo.hFileInfo.ioFlMdDat;
        ret = cinfo.hFileInfo.ioFlAttrib&kioFlAttribDirMask ? (-1):(1);
        /* text!; 2009.07.11 12:38 EST */
        if (ia && (cinfo.hFileInfo.ioFlFndrInfo.fdType=='TEXT'
            || cinfo.hFileInfo.ioFlFndrInfo.fdType=='sEXT')) /* stationery; 2009.07.11 12:48 EST */
          *ia|= 1;
        if (ea) /* read-only; 2012.03.02 */
          *ea|= cinfo.hFileInfo.ioFlAttrib&kioFlAttribLockedMask;
      }
    }
#else
    /* not all systems allow stat'ing a file with / appended */
    if (stat(name,&s)==0)
    {
      tm_t = s.st_mtime;
      ret = 1;
    }
#endif
  }
skip:
  filedate = localtime(&tm_t);

  tmzip->tm_sec  = filedate->tm_sec;
  tmzip->tm_min  = filedate->tm_min;
  tmzip->tm_hour = filedate->tm_hour;
  tmzip->tm_mday = filedate->tm_mday;
  tmzip->tm_mon  = filedate->tm_mon ;
  tmzip->tm_year = filedate->tm_year;
#endif
  return ret;
}



int check_exist_file(filename)
    const char* filename;
{
    FILE* ftestexist;

    if ((ftestexist=fopen(filename,"rb"))==NULL)
        return 0;
    else
        fclose(ftestexist);

    return 1;
}

void do_banner()
{
    printf("MiniZip 1.01b2, demo of zLib + Zip package written by Gilles Vollant\n");
    printf("more info at http://www.winimage.com/zLibDll/unzip.html\n\n");
}

void do_help()
{
    printf("Usage : minizip [-o] [-a] [-0 to -9] [-p password] file.zip [files_to_add]\n"
           "\n"
           "  -o  Overwrite existing file.zip\n"
           "  -a  Append to existing file.zip\n"
           "  -0  Store only\n"
           "  -1  Compress faster\n"
           "  -9  Compress better\n"
           "\n");
}

/* calculate the CRC32 of a file,
   we need known the CRC32 of the file before possible file encryption */
int getFileCrc(const char* filepathinzip,void*buf,unsigned long size_buf,unsigned long* result_crc)
{
   unsigned long calculate_crc=0;
   int err=ZIP_OK;
   FILE * fin = fopen(filepathinzip,"rb");
   unsigned long size_read = 0;
   unsigned long total_read = 0;

   if (fin==NULL)
   {
       err = ZIP_ERRNO;
   }

   if (err == ZIP_OK)
        do
        {
            err = ZIP_OK;
            size_read = (int)fread(buf,1,size_buf,fin);
            if (size_read < size_buf)
                if (feof(fin)==0)
                {
                    fprintf(stderr,"error in reading '%s'\n", filepathinzip);
                    err = ZIP_ERRNO;
                }

            if (size_read>0)
                calculate_crc = crc32(calculate_crc,buf,size_read);
            total_read += size_read;

        } while ((err == ZIP_OK) && (size_read>0));

    if (fin)
        fclose(fin);

    *result_crc=calculate_crc;
    printf("file '%s' crc %x\n", filepathinzip,calculate_crc);
    return err;
}

int main(argc,argv)
    int argc;
    char *argv[];
{
    int i;
    int opt_overwrite=0;
    int opt_compress_level=Z_DEFAULT_COMPRESSION;
    int zipfilenamearg = 0;
    char filename_try[MAXFILENAME+16];
    int zipok;
    int err=0;
    int size_buf=0;
    void* buf=NULL;
    const char* password=NULL;

    do_banner();
    if (argc==1)
    {
        do_help();
        return 0;
    } else
        for (i=1;i<argc;i++)
        {
            if ((*argv[i])=='-')
            {
                const char *p=argv[i]+1;

                while ((*p)!='\0')
                {
                    unsigned char   c=*p++;

                    if ((c=='o') || (c=='O'))
                        opt_overwrite = 1;
                    if ((c=='a') || (c=='A'))
                        opt_overwrite = 2;
                    if (((c=='p') || (c=='P')) && (i+1<argc))
                        password=argv[++i];
                    if ((c-='0')<='9')
                        opt_compress_level = c;
                }
            } else if (zipfilenamearg == 0)
                zipfilenamearg = i;
        }

    size_buf = WRITEBUFFERSIZE;
    if ((buf=(void*)malloc(size_buf))==NULL)
    {
        fprintf(stderr,"Error allocating memory.\n");
        return ZIP_INTERNALERROR;
    }

    if (zipfilenamearg==0)
        zipok=0;
    else
    {
        int i,len;
        int dot_found=0;

        zipok = 1;
        strncpy(filename_try, argv[zipfilenamearg],MAXFILENAME-1);
        /* strncpy doesn't append the trailing null if the string is too long. */
        filename_try[ MAXFILENAME ] = '\0';

        len=(int)strlen(filename_try);
        for (i=0;i<len;i++)
            if (filename_try[i]=='.')
                dot_found=1;

        if (dot_found==0)
            strcat(filename_try,".zip");

        if (opt_overwrite==2)
        {
            /* can't append a file that doesn't exist */
            if (check_exist_file(filename_try)==0)
                opt_overwrite = 1;
        } else if (opt_overwrite==0)
            if (check_exist_file(filename_try)!=0)
            {
                char rep=0;

                do
                {
                    char answer[128];

                    /* for it to work in MPW (and w/ little change, final space->'\n'); 2009.07.11 00:27:08 GMT */
                    printf("The file '%s' exists.  Overwrite ? [y]es, [n]o, [a]ppend :\n",
                        filename_try);

                    if (scanf("%1s", answer) != 1)
                    {
                       exit(EXIT_FAILURE);
                    }
                    rep = answer[0] ;
                    if ((rep>='a') && (rep<='z'))
                        rep -= 0x20;
                } while ((rep!='Y') && (rep!='N') && (rep!='A'));

                if (rep=='N')
                    zipok = 0;
                else if (rep=='A')
                    opt_overwrite = 2;
            }
    }

    if (zipok==1)
    {
        zipFile zf;
        int errclose;
#        ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;

        fill_win32_filefunc(&ffunc);
        zf = zipOpen2(filename_try,opt_overwrite==2 ? 2 : 0,NULL,&ffunc);
#        else
        zf = zipOpen(filename_try,opt_overwrite==2 ? 2 : 0);
#        endif

        if (zf==NULL)
        {
            fprintf(stderr,"error opening '%s'\n", filename_try);
            err= ZIP_ERRNO;
        } else
        { /* printf append when applicable; 2009.12.03 04:38 EST */
            if (opt_overwrite==2)
                printf("appending '%s'\n", filename_try);
            else
                printf("creating '%s'\n", filename_try);
            fflush(stdout); /* force flush; 2009.09.27 09:54 EDT */
        }

        for (i=zipfilenamearg+1;(i<argc) && (err==ZIP_OK);i++)
        {
            const char* filepath = argv[i];

            if (!(((*filepath=='-') || ((*filepath)=='/')) &&
                  ((filepath[1]=='o') || (filepath[1]=='O') ||
                   (filepath[1]=='a') || (filepath[1]=='A') ||
                   (filepath[1]=='p') || (filepath[1]=='P') ||
                   ((filepath[1]>='0') || (filepath[1]<='9'))) &&
                  (strlen(filepath) == 2)))
            {
                char filepathinzip[512];
                FILE *fin = NULL;
                int size_read;
                zip_fileinfo zi;
                unsigned long crcFile=0;

                memcpy(filepathinzip,(':' == *(unsigned char*)filepath) ?
                        1+filepath:filepath,511);
                filepathinzip[511] = 0; /* manual termination (in case it went too long) */
                zipfpath(filepathinzip);

                zi.tmz_date.tm_sec = zi.tmz_date.tm_min = zi.tmz_date.tm_hour =
                    zi.tmz_date.tm_mday = zi.tmz_date.tm_mon = zi.tmz_date.tm_year = 0;
                zi.dosDate = 0;
                zi.internal_fa = 0;
                zi.external_fa = 0;
                if (getfileinfo(filepath,&zi.tmz_date,&zi.dosDate,&zi.internal_fa,&zi.external_fa)<0)
                    continue;

                if ((password) && (err==ZIP_OK))
                    err = getFileCrc(filepath,buf,size_buf,&crcFile);

                if ((fin=fopen(filepath,"rb"))==NULL)
                {
                    if (errno==EISDIR)
                        continue;
                    err = ZIP_ERRNO;
                    fprintf(stderr,"error in opening '%s' for reading\n", filepathinzip);
                } else
                {
                    err = zipOpenNewFileInZip3(zf,filepathinzip,&zi,
                            NULL,0,NULL,0,NULL /* comment*/,
                            (opt_compress_level != 0) ? Z_DEFLATED : 0,
                            opt_compress_level,0,
                            /* -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY, */
                            -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY,
                            password,crcFile);

                    if (err != ZIP_OK)
                    {
                        fclose(fin);
                        fin = NULL;
                        fprintf(stderr,"error #%d in opening '%s' in zipfile\n",
                            err==ZIP_ERRNO ? errno : err,filepathinzip);
                    }
                }

                if (err == ZIP_OK)
                    do
                    {
                        if (signalbreak())
                            {err = -128; break;}
                        err = ZIP_OK;
                        size_read = (int)fread(buf,1,size_buf,fin);
                        if (size_read < size_buf)
                            if (feof(fin)==0)
                            {
                                fprintf(stderr,"error in reading '%s'\n", filepathinzip);
                                err = ZIP_ERRNO;
                            }

                        if (size_read>0)
                            if ((err=zipWriteInFileInZip (zf,buf,size_read)) < 0)
                            {
                                fprintf(stderr,"error in writing '%s' in the zipfile\n",
                                        filepathinzip);
                            }
                    } while ((err == ZIP_OK) && (size_read>0));

                if (fin != NULL)
                    fclose(fin);

                if (err<0)
                    err = ZIP_ERRNO;
                else if ((err=zipCloseFileInZip(zf)) != ZIP_OK)
                    fprintf(stderr,"error in closing '%s' in the zipfile\n",
                            filepathinzip);
            }
        }

        if ((errclose=zipClose(zf,NULL)) != ZIP_OK)
            fprintf(stderr,"error in closing '%s'\n", filename_try);
        else
            printf("'%s' closed.\n", filename_try);
    }
    else
    {
       do_help();
    }

    free(buf);
    return 0;
}
